################################################################ 
# Installation
#
# The following packages should be installed elsewhere
#     libraries of GRACE (BASES/SPRING, CHANEL etc)
#     Kinematics library
#     gracefig
#
#  1. edit Makefile
#
#     Default file uses an enviroment variable "$(PWD)".
#     In some systems shell command does not set 
#     this variable (try "echo $PWD").
#     In such a case, set this variable explicitly 
#     the "Makefile" or execute "make" command with
#     argument "make PWD=`pwd`, such as:
#
#          make all PWD=`pwd`
#
# GRACEDIR  = directory of grace system
#
# LIBDIR    = directory where runtime libraries are installed.
# BINDIR    = directory where executables are installed.
# MODELPATH = list of directories of model files
# KINEMPATH = list of directories of kinematics
#
# GRACEFIG  = full path of "gracefig" command
#
# FC        = fortran complier
# FFLAGS    = flags for fortran complier
# LDFLAGS   = flags for ld
# AR        = archive
# RANLIB    = generate index to archive.
#
# XOPTI     = -I(path of include files of X11)
# XOPTI     = -I(path of include files of X11)
#             For example:
#                 -I/usr/X11/include
#                 -I/usr/X11R6/include
#                 -I/usr/include/X11R5
#                 -I/usr/local/X11R5/include
#                 -I/kek/local/X11R6/include
#
# XOPTL     = -L(path of library files of X11)
#             For example:
#                 -L/usr/X11/lib
#                 -L/usr/X11R6/lib
#                 -L/usr/lib/X11R5
#                 -L/usr/local/X11R5/lib
#                 -L/kek/local/X11R6/lib
#
# When Athena widgets are installed
# XCOPT     = -DTOOLKIT -DX11
# XLLIB     = -lXaw -lXmu -lXt -lXext -lX11
#
# When Athena widgets are not installed
# XCOPT     = -DX11
# XLLIB     = -lX11
#
# When Motif is installed
# MOTI      = -I(path of include file of Motif) 
# MOTL      = -L(path of library file of Motif)
# MOLIB     = -lXm -lXt -lX11 -lm
#
#  2. make all
#
#  3. make install
#
################################################################ 

all: makefiles grcbas grcft ftlibs grfig ntlib

install: $(BINDIR) $(LIBDIR) allinst

grcbas: grclb grcsc grcul \
        testinst/sm/eewwa/Makefile testinst/mssm/asw1SW1/Makefile

grclb: grclib/Makefile grclib/config.h
	(cd grclib; make CDIR="$(PWD)/grclib" all)

grcsc: grcsrc/Makefile
	(cd grcsrc; make CDIR="$(PWD)/grcsrc" all)

grcul: grcutl/Makefile
	(cd grcutl; make CDIR="$(PWD)/grcutl" all)

grcft: grcfort/Makefile
	(cd grcfort; make CDIR="$(PWD)/grcfort" all)

grfig: gracefig/Makefile
	(cd gracefig; make CDIR="$(PWD)/gracefig")

ftlibs: fchanel fkinemlib fbasesv5.1

fchanel: chanel/Makefile
	(cd chanel; make MACHINE="$(MACHINE)" all)

fkinemlib: kinemlib/Makefile
	(cd kinemlib; make MACHINE="$(MACHINE)" all)

fbasesv5.1: basesv5.1/Makefile
	(cd basesv5.1; make MACHINE="$(MACHINE)" all)

ntlib: netlib/blas/Makefile netlib/lapack/double/Makefile
	(cd netlib/blas; make)
	(cd netlib/lapack/double; make)

backup: backup-copy backup-clean

backup-copy:
	rm -rf $(HOME)/export/grc/*
	cp -pR * $(HOME)/export/grc

backup-clean:
	(cd $(HOME)/export/grc; make clean)

allinst: 
	if [ ! -d ${BINDIR} ]; then mkdir -p ${BINDIR}; fi
	if [ ! -d ${LIBDIR} ]; then mkdir -p ${LIBDIR}; fi
	(cd grclib;    make install)
	(cd grcsrc;    make install)
	(cd grcutl;    make install)
	(cd grcfort;   make install)
	(cd chanel;    make install)
	(cd kinemlib;  make install)
	(cd basesv5.1; make install)
	(cd gracefig;  make install)
	(cd netlib/blas; make install)
	(cd netlib/lapack/double; make install)
	cp -pR dbkinem model ${LIBDIR}

$(BINDIR):
#	mkdir $(BINDIR)

$(LIBDIR):
#	mkdir $(LIBDIR)

CDATE = `date`

makefiles: grclib/config.h \
	   grcsrc/Makefile grcutl/Makefile grcfort/Makefile gracefig/Makefile \
	   chanel/Makefile kinemlib/Makefile basesv5.1/Makefile \
	   netlib/blas/Makefile netlib/lapack/double/Makefile \
	   testinst/sm/eewwa/Makefile testinst/mssm/asw1SW1/Makefile

grclib/config.h: Makefile
	@echo "#ifdef DEFGLOBAL" > grclib/config.h
	@echo "Global char *gracedir  =\""$(GRACEDIR)"\";"  >> grclib/config.h
	@echo "Global char *graceldir =\""$(LIBDIR)"\";" >> grclib/config.h
	@echo "Global char *modelpath =\""$(MODELPATH)"\";" >> grclib/config.h
	@echo "Global char *kinempath =\""$(KINEMPATH)"\";" >> grclib/config.h
	@echo "Global char *machine   =\""$(MACHINE)"\";"   >> grclib/config.h
	@echo "Global char *mpihome   =\""$(MPIHOME)"\";"   >> grclib/config.h
	@echo "Global char *cmpdate   =\""`date`"\";"   >> grclib/config.h
	@echo "Global char *huname     =\""`uname -snrvm`"\";"   >> grclib/config.h
	@echo "Global char *fortcmp   =\""$(FC)"\";"   >> grclib/config.h
	@echo "Global char *fflags    =\""$(FFLAGS)"\";"   >> grclib/config.h
	@echo "Global char *lflags    =\""$(LDFLAGS)"\";"   >> grclib/config.h
	@echo "Global char *cernlib   =\""$(CERNLIB)"\";"   >> grclib/config.h
	@echo "Global char *reduce    =\""$(REDUCE)"\";"   >> grclib/config.h
	@echo "Global char *vorm      =\""$(VORM)"\";"   >> grclib/config.h
	@echo "#else" >> grclib/config.h
	@echo "Global char *gracedir;"  >> grclib/config.h
	@echo "Global char *graceldir;" >> grclib/config.h
	@echo "Global char *modelpath;" >> grclib/config.h
	@echo "Global char *kinempath;" >> grclib/config.h
	@echo "Global char *machine;"   >> grclib/config.h
	@echo "Global char *mpihome;"   >> grclib/config.h
	@echo "Global char *cmpdate;"   >> grclib/config.h
	@echo "Global char *huname;"    >> grclib/config.h
	@echo "Global char *fortcmp;"   >> grclib/config.h
	@echo "Global char *fflags;"    >> grclib/config.h
	@echo "Global char *lflags;"    >> grclib/config.h
	@echo "Global char *cernlib;"   >> grclib/config.h
	@echo "Global char *reduce;"    >> grclib/config.h
	@echo "Global char *vorm;"      >> grclib/config.h
	@echo "#endif" >> grclib/config.h

grclib/Makefile: Makefile grclib/Makefile.tpl
	@echo "BINDIR	= $(BINDIR)" > grclib/Makefile
	@echo "GRACEFIG	= $(GRACEFIG)" >> grclib/Makefile
	@echo "XOPTI	= $(XOPTI)" >> grclib/Makefile
	@echo "XOPTL	= $(XOPTL)" >> grclib/Makefile
	@echo "XCOPT	= $(XCOPT)" >> grclib/Makefile
	@echo "XLLIB	= $(XLLIB)" >> grclib/Makefile
	@echo "CC	= $(CC)" >> grclib/Makefile
	@echo "COPT	= $(COPT)" >> grclib/Makefile
	@echo "FORMOPT	= $(FORMOPT)" >> grclib/Makefile
	@cat grclib/Makefile.tpl >> grclib/Makefile

grcsrc/Makefile: Makefile grcsrc/Makefile.tpl
	@echo "BINDIR	= $(BINDIR)" > grcsrc/Makefile
	@echo "GRACEFIG	= $(GRACEFIG)" >> grcsrc/Makefile
	@echo "XOPTI	= $(XOPTI)" >> grcsrc/Makefile
	@echo "XOPTL	= $(XOPTL)" >> grcsrc/Makefile
	@echo "XCOPT	= $(XCOPT)" >> grcsrc/Makefile
	@echo "XLLIB	= $(XLLIB)" >> grcsrc/Makefile
	@echo "CC	= $(CC)" >> grcsrc/Makefile
	@echo "COPT	= $(COPT)" >> grcsrc/Makefile
	@echo "FORMOPT	= $(FORMOPT)" >> grcsrc/Makefile
	@echo "EXESUFF	= $(EXESUFF)" >> grcsrc/Makefile
	@cat grcsrc/Makefile.tpl >> grcsrc/Makefile

grcutl/Makefile: Makefile grcutl/Makefile.tpl
	@echo "BINDIR	= $(BINDIR)" > grcutl/Makefile
	@echo "GRACEFIG	= $(GRACEFIG)" >> grcutl/Makefile
	@echo "XOPTI	= $(XOPTI)" >> grcutl/Makefile
	@echo "XOPTL	= $(XOPTL)" >> grcutl/Makefile
	@echo "XCOPT	= $(XCOPT)" >> grcutl/Makefile
	@echo "XLLIB	= $(XLLIB)" >> grcutl/Makefile
	@echo "CC	= $(CC)" >> grcutl/Makefile
	@echo "COPT	= $(COPT)" >> grcutl/Makefile
	@echo "FORMOPT	= $(FORMOPT)" >> grcutl/Makefile
	@echo "EXESUFF	= $(EXESUFF)" >> grcutl/Makefile
	@cat grcutl/Makefile.tpl >> grcutl/Makefile

grcfort/Makefile: Makefile grcfort/Makefile.tpl
	@echo "BINDIR	= $(BINDIR)" > grcfort/Makefile
	@echo "GRACEFIG	= $(GRACEFIG)" >> grcfort/Makefile
	@echo "XOPTI	= $(XOPTI)" >> grcfort/Makefile
	@echo "XOPTL	= $(XOPTL)" >> grcfort/Makefile
	@echo "XCOPT	= $(XCOPT)" >> grcfort/Makefile
	@echo "XLLIB	= $(XLLIB)" >> grcfort/Makefile
	@echo "CC	= $(CC)" >> grcfort/Makefile
	@echo "COPT	= $(COPT)" >> grcfort/Makefile
	@echo "FORMOPT	= $(FORMOPT)" >> grcfort/Makefile
	@echo "EXESUFF	= $(EXESUFF)" >> grcfort/Makefile
	@cat grcfort/Makefile.tpl >> grcfort/Makefile

gracefig/Makefile: Makefile gracefig/Makefile.tpl
	@echo "BINDIR	= $(BINDIR)" > gracefig/Makefile
	@echo "GRACEFIG	= $(GRACEFIG)" >> gracefig/Makefile
	@echo "XOPTI	= $(XOPTI)" >> gracefig/Makefile
	@echo "XOPTL	= $(XOPTL)" >> gracefig/Makefile
	@echo "XCOPT	= $(XCOPT)" >> gracefig/Makefile
	@echo "XLLIB	= $(XLLIB)" >> gracefig/Makefile
	@echo "MOPTI	= $(MOPTI)" >> gracefig/Makefile
	@echo "MOPTL	= $(MOPTL)" >> gracefig/Makefile
	@echo "MOLIB	= $(MOLIB)" >> gracefig/Makefile
	@echo "CC	= $(CC)" >> gracefig/Makefile
	@echo "COPT	= $(COPT)" >> gracefig/Makefile
	@echo "EXESUFF	= $(EXESUFF)" >> gracefig/Makefile
	@cat gracefig/Makefile.tpl1 >> gracefig/Makefile

chanel/Makefile: Makefile chanel/Makefile.tpl
	@echo "LIBDIR   = $(LIBDIR)" > chanel/Makefile
	@echo "FC       = $(FC)" >> chanel/Makefile
	@echo "FFLAGS   = $(FFLAGS)" >> chanel/Makefile
	@echo "AR       = $(AR)" >> chanel/Makefile
	@echo "RANLIB   = $(RANLIB)" >> chanel/Makefile
	@cat chanel/Makefile.tpl >> chanel/Makefile

kinemlib/Makefile: Makefile kinemlib/Makefile.tpl
	@echo "LIBDIR   = $(LIBDIR)" > kinemlib/Makefile
	@echo "FC       = $(FC)" >> kinemlib/Makefile
	@echo "FFLAGS   = $(FFLAGS)" >> kinemlib/Makefile
	@echo "AR       = $(AR)" >> kinemlib/Makefile
	@echo "RANLIB   = $(RANLIB)" >> kinemlib/Makefile
	@cat kinemlib/Makefile.tpl >> kinemlib/Makefile

basesv5.1/Makefile: Makefile basesv5.1/Makefile.tpl
	@echo "LIBDIR   = $(LIBDIR)"   > basesv5.1/Makefile
	@echo "FC       = $(FC)"      >> basesv5.1/Makefile
	@echo "FFLAGS   = $(FFLAGS)"  >> basesv5.1/Makefile
	@echo "AR       = $(AR)"      >> basesv5.1/Makefile
	@echo "RANLIB   = $(RANLIB)"  >> basesv5.1/Makefile
	@echo "CC       = $(CC)"      >> basesv5.1/Makefile
	@echo "MDFUNCS  = $(MDFUNCS)" >> basesv5.1/Makefile
	@cat basesv5.1/Makefile.tpl   >> basesv5.1/Makefile

netlib/blas/Makefile: Makefile netlib/blas/Makefile.tpl
	@echo "LIBDIR   = $(LIBDIR)"   > netlib/blas/Makefile
	@echo "FC       = $(FC)"      >> netlib/blas/Makefile
	@echo "FFLAGS   = $(FFLAGS)"  >> netlib/blas/Makefile
	@echo "AR       = $(AR)"      >> netlib/blas/Makefile
	@echo "RANLIB   = $(RANLIB)"  >> netlib/blas/Makefile
	@cat netlib/blas/Makefile.tpl >> netlib/blas/Makefile

netlib/lapack/double/Makefile: Makefile netlib/lapack/double/Makefile.tpl
	@echo "LIBDIR   = $(LIBDIR)"            > netlib/lapack/double/Makefile
	@echo "FC       = $(FC)"               >> netlib/lapack/double/Makefile
	@echo "FFLAGS   = $(FFLAGS)"           >> netlib/lapack/double/Makefile
	@echo "AR       = $(AR)"               >> netlib/lapack/double/Makefile
	@echo "RANLIB   = $(RANLIB)"           >> netlib/lapack/double/Makefile
	@cat netlib/lapack/double/Makefile.tpl >> netlib/lapack/double/Makefile

netlib/lapack/double/Makefile: netlib/lapack/double/Makefile.tpl

testinst/sm/eewwa/Makefile: Makefile testinst/sm/eewwa/Makefile.tpl
	@echo "BINDIR   = $(BINDIR)" > testinst/sm/eewwa/Makefile
	@cat testinst/sm/eewwa/Makefile.tpl >> testinst/sm/eewwa/Makefile

testinst/mssm/asw1SW1/Makefile: Makefile testinst/mssm/asw1SW1/Makefile.tpl
	@echo "BINDIR   = $(BINDIR)" > testinst/mssm/asw1SW1/Makefile
	@cat testinst/mssm/asw1SW1/Makefile.tpl >> testinst/mssm/asw1SW1/Makefile

test-install:
	(cd testinst/sm/eewwa; make all)
	(cd testinst/mssm/asw1SW1; make all)


clean:
	rm -f *core t tt *.bak *~
	for i in `ls`; do \
	    if test -d $$i && test $$i != CVS && \
		test $$i != bin && test $$i != lib ; \
                then (cd $$i; echo `pwd`; make clean) ; \
            fi \
	done
