C                                                                              
C ********************************************************************         
C *         V                                                        *         
C *         !                                                        *         
C *         !                                                        *         
C * F1-->---!---<---F2    SUBROUTINE FFVCCN(-FFVCT)                  *         
C *                                                                  *         
C *     Purpose: to calculate vertex amplitudes for vector boson-    *         
C *              massive fermions vertex.                            *         
C *                                                                  *         
C *      -          C                C                               *         
C *      U(II,AAM,Q) (Q(AL*PL+AR*PR))  U(I,AM,P)                     *         
C *                                                                  *         
C *                                                                  *         
C *                                                                  *         
C *     L=: polarization state of vector boson.                      *         
C *     I,II=: indices to specify fermion(I,II=3) or antifermion     *         
C *            (I,II=1) state.                                       *         
C *     AM,AAM=: masses of fermions.                                 *         
C *     AL,AR=: coupling constants for vertex.                       *         
C *     C,CC=: phase factors for massive fermions.                   *         
C *     P1,P2,Q1,Q2=: light-like vectors decomposed by subroutine    *         
C *                   SPLTQ.                                         *         
C *     Q=: polarization vector of vector boson.                     *         
C *     AALL=: calculated results of vertex amplitudes for all       *         
C *            possible helicity states.                             *         
C *                                                                  *         
C *                                                                  *         
C *                                      written by H. Tanaka        *         
C ********************************************************************         
C                                                                              
C                                                                              
C       ===========================================================            
        SUBROUTINE FFVCCN(L,II,I,AAM,AM,AL,AR,CC,C,Q1,Q2,P1,P2,Q,AALL)          
C       ===========================================================            
C                                                                              
        IMPLICIT REAL*8(A-H,O-Z)                                               
        COMPLEX*16 AALL,C,CC                                                   
        DIMENSION P1(4),P2(4),Q1(4),Q2(4),C(2),CC(2),Q(4)                      
        DIMENSION AALL(4,2,2)                                                  
                                                                               
        I2 = II - 2                                                            
        I2 = -I2                                                               
        I2 = I2 + 2                                                            
                                                                               
        CALL FFV(L,I2,I,AAM,AM,AR,AL,CC,C,Q1,Q2,P1,P2,Q,AALL)                  


        AALL(L,1,1)=-AALL(L,1,1)
        AALL(L,1,2)=-AALL(L,1,2)
        AALL(L,2,1)=-AALL(L,2,1)
        AALL(L,2,2)=-AALL(L,2,2)  


        RETURN                                                                 
        END                                                                    
