C---------------------------------------------------------------------
C   GRACE System Library File
C   KINEM No. : 1301
C   Date      : 2002.07.26
C   Author    : T.Kon, Y.Kurihara
C---------------------------------------------------------------------
      subroutine kinem(nextrn, x, pe, pp, yacob, nreg, ireg, jump) 
      implicit real* 8(a-h,o-z)
      integer nextrn 
      parameter ( mxdim = 50 )
      real*8  x(mxdim) 
      real*8  pe(4,nextrn), pp(nextrn,nextrn) 
      real*8  yacob 
      integer nreg, ireg 
      integer jump 
      include 'inclk.h'
      COMMON/CUT001/AMASCT(2),ARESNS(2)
      COMMON/KCNTL1/IRESNS
*=======================================================================
*     This subroutine calculates four momenta of external particles.
*
*     Arguments
*       NEXTRN : input  : The number of external particles.
*       X      : input  : Position in the integration volume.
*       PE     : output : Table of four momenta of external particles.
*       PP     : output : Table of inner products of four momenta
*                         between two external particles.
*       YACOB  : output : Normalization for converting square of
*                         amplitude to cross-section.
*       NREG   : in/out : Maximum number of points in the phase space
*                         which correspond to the integration point.
*                         This value is one for the first call
*                         for the same integration point.
*       IREG   : input  : Counter of calling this subroutine for the
*                         same integration point.
*                         same integration point.
*                         Function FUNC increments this variable,
*                         and call this subroutine while (IREG.LE.NREG)
*       JUMP   : output : If this integration point is rejected,
*                         one should set this variable other than one,
*                         otherwize it should be zero.
*=======================================================================
      jump = 0

        xm12x=(amass1(1)-amass1(4))**2
        xm12m=(amass1(2)+amass1(3))**2
        if(IRESNS.eq.0) then
         xm12=xm12m+(xm12x-xm12m)*x(1)
         AJACOB=xm12x-xm12m
        else
         zM    =ARESNS(1)
         zM2   =zM*zM
         zMG   =ARESNS(1)*ARESNS(2)
         THEMIN=ATAN((xm12m-zM2)/zMG)
         THEMAX=ATAN((xm12x-zM2)/zMG)
         THE   =THEMIN+(THEMAX-THEMIN)*X4
         xm12    =zMG*TAN(THE)+zM2
         AJACOB=(THEMAX-THEMIN)*((xm12-zM2)**2+zMG**2)/zMG*2
        end if

        e1s=(xm12+amass1(2)**2-amass1(3)**2)/2.d0/sqrt(xm12)
        e3s=(amass1(1)**2-xm12-amass1(4)**2)/2.d0/sqrt(xm12)

        xm13x=(e1s+e3s)**2-(sqrt(e1s**2-amass1(2)**2)-
     .        sqrt(e3s**2-amass1(4)**2))**2
        xm13m=(e1s+e3s)**2-(sqrt(e1s**2-amass1(2)**2)+
     .        sqrt(e3s**2-amass1(4)**2))**2
        xm13=xm13m+(xm13x-xm13m)*x(2)

        e1=(xm12+xm13-amass1(3)**2-amass1(4)**2)/2.d0/amass1(1)
        p1=sqrt(e1**2-amass1(2)**2)
        e2=(amass1(1)**2+amass1(3)**2-xm13)/2.d0/amass1(1)
        p2=sqrt(e2**2-amass1(3)**2)
        cost=(2.d0*e1*e2+amass1(2)**2+amass1(3)**2-xm12)/2.d0/p1/p2
        sint=sqrt(1.d0-cost**2)
*-----------------------------------------------------------------------
*     Table of four momenta.
* PE(I, J) : I = 1 -> X, 2 -> Y, ... 4 -> energy, of J-th particle.
*-----------------------------------------------------------------------
*     1: initial particle mass=amass1(1)
      pe( 1, 1) = 0.d0   
      pe( 2, 1) = 0.d0   
      pe( 3, 1) = 0.d0   
      pe( 4, 1) = amass1(1)   
*     2: final particle1 mass=amass1(2)
      pe( 1, 2) = 0.d0   
      pe( 2, 2) = 0.d0   
      pe( 3, 2) = p1   
      pe( 4, 2) = e1   
*     3: final particle2 mass=amass1(3)
      pe( 1, 3) = p2*sint   
      pe( 2, 3) = 0.d0   
      pe( 3, 3) = p2*cost   
      pe( 4, 3) = e2   
*     4: final particle3 mass=amass1(4)
      pe( 1, 4) = -p2*sint   
      pe( 2, 4) = 0.d0   
      pe( 3, 4) = -p1-p2*cost   
      pe( 4, 4) = amass1(1)-e1-e2   
*-----------------------------------------------------------------------
*     PP(I,J) = inner product between I-th  and J-th particle's momenta
*
*     This table should be calculated from invariants.
*     When it is constructed from PE, it will not so accurate.
*-----------------------------------------------------------------------
      pp( 1, 1) = amass2(1)
      pp( 1, 2) = 0.5d0*(xm12+xm13-amass2(3)-amass2(4))
      pp( 1, 3) = 0.5d0*(amass2(1)+amass2(3)-xm13)
      pp( 1, 4) = 0.5d0*(amass2(1)+amass2(4)-xm12)
*-----------------------------------------------------------------------
      pp( 2, 1) = pp( 1, 2)
      pp( 2, 2) = amass2(2)
      pp( 2, 3) = 0.5d0*(xm12-amass2(2)-amass2(3))
      pp( 2, 4) = 0.5d0*(xm13-amass2(2)-amass2(4))
*-----------------------------------------------------------------------
      pp( 3, 1) = pp( 1, 3)
      pp( 3, 2) = pp( 2, 3)
      pp( 3, 3) = amass2(3)
      pp( 3, 4) = 0.5d0*(amass2(1)+amass2(2)-xm12-xm13)
*-----------------------------------------------------------------------
      pp( 4, 1) = pp( 1, 4)
      pp( 4, 2) = pp( 2, 4)
      pp( 4, 3) = pp( 3, 4)
      pp( 4, 4) = amass2(4)
*-----------------------------------------------------------------------
*     Normalization factor
*-----------------------------------------------------------------------

      yacob = 1.d0/(2.d0*pi)**3/(32.d0*amass1(1)**3)*
     . (xm13x-xm13m)*AJACOB

      return
      end
      subroutine kinemp( ireg )
************************************************************************
      implicit real* 8(a-h,o-z)
      return
      end
