C---------------------------------------------------------------------
C   GRACE System Library File
C   KINEM No. : 2401
C   Date      : 2002.07.26
C   Author    : Y.Kurihara
C---------------------------------------------------------------------
      SUBROUTINE KINIT
      IMPLICIT REAL*8(A-H,O-Z)
 
      PARAMETER ( MXDIM = 50 )
      COMMON / LOOP0 / LOOP
      COMMON / BPARM1 / XL(MXDIM),XU(MXDIM),NDIM,NWILD,
     &                 IG(MXDIM),NCALL
      COMMON / BPARM2 / ACC1,ACC2,ITMX1,ITMX2
 
      CHARACTER XSTR*14
 
      INCLUDE 'inclk.h'
*-----------------------------------------------------------------------
      COMMON/KMCNTL/IRESNS(2),idummy(2),ICOSQ3
      COMMON/KINEM1/S,W,FACT
      COMMON/CUT001/COSCUT(2,4),ENGYCT(2,4),AMASCT(2,2),ARESNS(2,2)
      common /knmflg/ knmgvs, knmhst, knmsfp, knmisr
*-----------------------------------------------------------------------
* 4-body kinematics : fusion type (2-photon, W-fusion)
*
*  Integration variables
*        q1**2, q2**2
*        cos(the_q1), phi_q1(=0)in p1-p2 CM frame
*        cos(the_p3), phi_p3   in p3-p4 CM frame
*        cos(the_p5), phi_p5             in p5-p6 CM frame
*
*
*
*       P1 \              *------- P3
*           \            / \
*            \          /   \----  P4
*             \  ****  /
*              >******<
*             /  ****  \
*            /          \   /-----P5
*           /            \ /
*       P2 /              *------- P6
*
*--- 1. Initialize constants for kinematics.
* S = (P1+P2)**2
* W = SQRT(S)
      if( knmgvs .eq. 1 ) then
      W = 200.D0
      endif
      S = W*W
      totmas=amass1(3)+amass1(4)+amass1(5)+amass1(6)
      if(totmas.gt.w) then
	write(6,*)
     .' CM energy is less than the sum of final particle masses.'
	stop
      end if
*******************************
* Angular cuts in Lab-frame.  *
*******************************
* particle 3
* minimum cos  cut
      COSCUT(1,1)= -1
* maximum cos  cut
      COSCUT(2,1)=  1
* particle 4
* minimum cos  cut
      COSCUT(1,2)= -1
* maximum cos  cut
      COSCUT(2,2)=  1
* particle 5
* minimum cos  cut
      COSCUT(1,3)= -1
* maximum cos  cut
      COSCUT(2,3)=  1
* particle 6
* minimum cos  cut
      COSCUT(1,4)= -1
* maximum cos  cut
      COSCUT(2,4)=  1
 
*******************************
* Energy  cuts in Lab-frame.  *
*******************************
* particle 3
* minimum energy cut
      ENGYCT(1,1) = AMASS1(3)
* maximum energy cut
      ENGYCT(2,1) = W
* particle 4
* minimum energy cut
      ENGYCT(1,2) = AMASS1(4)
* maximum energy cut
      ENGYCT(2,2) = W
* particle 5
* minimum energy cut
      ENGYCT(1,3) = AMASS1(5)
* maximum energy cut
      ENGYCT(2,3) = W
* particle 6
* minimum energy cut
      ENGYCT(1,4) = AMASS1(6)
* maximum energy cut
      ENGYCT(2,4) = W
 
*************************
* Cut on invariant mass *
*************************
* Paartile 3-4
* minimum
      AMASCT(1,1)=   AMASS1(3)+AMASS1(4)
* maximum
      AMASCT(2,1)= W-AMASS1(5)-AMASS1(6)
* Paartile 5-6
* minimum
      AMASCT(1,2)=   AMASS1(5)+AMASS1(6)
* maximum
      AMASCT(2,2)= W-AMASS1(3)-AMASS1(4)
 
*******************************
* Q2 singularity treatment    *
*******************************
* no-singularity           : IRESNS= 0
* narrow resonance         : IRESNS= 1
* 1/Q2 singularity         : IRESNS= 2
* Particle 3-4
      IRESNS(1)  = 0
      IF(IRESNS(1).EQ.1) THEN
C If you want treat narrow resonance, you should set resonance mass
C and width.
c       ARESNS(1,1)=AMWB
c       ARESNS(2,1)=AGWB
        ARESNS(1,1)=0
        ARESNS(2,1)=0
      END IF
* Particle 5-6
      IRESNS(2)  = 0
      IF(IRESNS(2).EQ.1) THEN
C If you want treat narrow resonance, you should set resonance mass
C and width.
c       ARESNS(1,2)=AMWB
c       ARESNS(2,2)=AGWB
        ARESNS(1,1)=0
        ARESNS(2,1)=0
      END IF
 
* 
* For the digrma as shown below,
*
*       P1 ----------------------- P3
*               |         \
*               |           \----  P4
*               | 
*               |  t = -T
*               | 
*               |           /-----P5
*               |          /
*       P2 ---------------------- P6
*
*  T is used as integration variables instead of cos(the_q1).
*
*                 integration       singularity
*   ICOSQ3 = 0   : cos(the_q)          non
*          = 1   : T                   non
*          = 2   : T                   1/T
      ICOSQ3 = 0
      
      WRITE(6,*)'*'
      WRITE(6,*)'* Kinematics initialization '
      WRITE(6,*)'*  Code # 2401 '
      WRITE(6,*)'* CM Energy ',W
      WRITE(6,*)'* COSCUT    '
      WRITE(6,'(2F10.3)')COSCUT
      WRITE(6,*)'* ENGYCT    '
      WRITE(6,'(2F10.3)')ENGYCT
      WRITE(6,*)'* AMASCT    '
      WRITE(6,'(2F10.3)')AMASCT
      WRITE(6,*)'* Singularity treatment '
      WRITE(6,*)'*   Q2 (p3-p4)'
      IF     (IRESNS(1).EQ.2) THEN
         WRITE(6,*)'* 1/Q2 singularity '
      ELSE IF(IRESNS(1).EQ. 0) THEN
         WRITE(6,*)'* NO   singularity '
      ELSE IF(IRESNS(1).EQ. 1) THEN
         WRITE(6,*)'* Narrow resonance '
         WRITE(6,*)'*  resonance mass and width'
         WRITE(6,*) ARESNS(1,1),ARESNS(2,1)
      END IF
      WRITE(6,*)'*   Q2 (p5-p6)'
      IF     (IRESNS(2).EQ.2) THEN
         WRITE(6,*)'* 1/Q2 singularity '
      ELSE IF(IRESNS(2).EQ. 0) THEN
         WRITE(6,*)'* NO   singularity '
      ELSE IF(IRESNS(2).EQ. 1) THEN
         WRITE(6,*)'* Narrow resonance '
         WRITE(6,*)'*  resonance mass and width'
         WRITE(6,*) ARESNS(1,2),ARESNS(2,2)
      END IF
      IF     (Icosq3.EQ.0) THEN
         WRITE(6,*)'* cos(the_q) ; flat '
      ELSE IF(Icosq3.EQ.1) THEN
         WRITE(6,*)'* T ; flat '
      ELSE IF(Icosq3.EQ.2) THEN
         WRITE(6,*)'* T ; 1/T '
      END IF
      WRITE(6,*)'*'
 
* Following flux factor is for a particle-antiparticle collision
* at s/m**2 >> 1. For an e-gamma collision or low energy interactions
* you should use appropriate formulae.
      VREL = 2
      FLUX = VREL*S
      FACT = GEVPB/FLUX
 
*----- BASES RELATED INITIALIZATIONS -----------------------------------
*--- 2. Dimension of integration variables.
 
          NDIM =   8
          NWILD=   7
 
C-----------------------------------------------------------------------
*--- 3. Region of integration.
 
        DO 1 I=1,NDIM
         XL(I)  =   0.D0
         XU(I)  =   1.D0
         IG(I)  =   1
1       CONTINUE

C-----------------------------------------------------------------------
*--- 4. Number of iterations
      ITMX1  =  5
      ITMX2  =  5
 
      NCALL  = 5000
 
*--- 5. Set MXREG : the maximum number of values which are returned
*       by FUNC for one phase space point
 
      MXREG =   1
 
*--- 6. Set histograms
      if( knmhst .eq. 1 ) then
      NX = 50
 
      DO 100 I = 1, NDIM
         WRITE(XSTR, 110) I
  110    FORMAT('X(',I2,') SPECTRUM')
         CALL XHINIT(I, XL(I), XU(I), NX, XSTR)
  100 CONTINUE
         CALL XHINIT(ndim+1, 0.d0, w, NX, 'Energy of Particle 3')
         CALL XHINIT(ndim+2, 0.d0, w, NX, 'Energy of Particle 4')
         CALL XHINIT(ndim+3, 0.d0, w, NX, 'Energy of Particle 5')
         CALL XHINIT(ndim+4, 0.d0, w, NX, 'Energy of Particle 6')
         CALL XHINIT(ndim+5,-1.d0,1.d0, NX, 'cos_the of Particle 3')
         CALL XHINIT(ndim+6,-1.d0,1.d0, NX, 'cos_the of Particle 4')
         CALL XHINIT(ndim+7,-1.d0,1.d0, NX, 'cos_the of Particle 5')
         CALL XHINIT(ndim+8,-1.d0,1.d0, NX, 'cos_the of Particle 6')
      endif
      RETURN
      END
