C---------------------------------------------------------------------
C   GRACE System Library File
C   KINEM No. : 2402
C   Date      : 2002.07.26
C   Author    : Y.Kurihara
C---------------------------------------------------------------------
      SUBROUTINE KINIT
      IMPLICIT REAL*8(A-H,O-Z)
      PARAMETER ( MXDIM = 50 )
      COMMON / LOOP0 / LOOP
      COMMON / BPARM1 / XL(MXDIM),XU(MXDIM),NDIM,NWILD,
     &                 IG(MXDIM),NCALL
      COMMON / BPARM2 / ACC1,ACC2,ITMX1,ITMX2
 
      CHARACTER XSTR*14
 
      INCLUDE 'inclk.h'
*-----------------------------------------------------------------------
      COMMON/KMCNTL/IRESNS,Idummy,ICOS3,ICOS4,isr
      COMMON/KINEM1/S,W,FACT
      COMMON/CUT001/COSCUT(2,4),ENGYCT(2,4),adummy(2),
     .AMASCT(2),adumy2(2),ARESNS(2)
      common /knmflg/ knmgvs, knmhst, knmsfp, knmisr
*-----------------------------------------------------------------------
* 4-body kinematics : fusion type (2-photon, W-fusion)
*
*  Integration variables
*       -q1**2, q10 ,phi3(=0)
*       -q2**2, q20
*        q**2=(p5+p6)**2
*        COS(the5^),Phi5^    in particle 5,6 CM frame
*
*
*
*    P1---------------!---------------------- P3
*                     ! q1=p1-p3
*                     !                  /--- P5
*                     *    q=p5+p6      /
*                    ***---------------<
*                     *                 \
*                     !                  \--- P6
*                     ! q2=p2-p4
*    P2---------------!---------------------- P4
*
*--- 1. Initialize constants for kinematics.
* S = (P1+P2)**2
* W = SQRT(S)
      W = 200.D0
      S = W*W
      totmas=amass1(3)+amass1(4)+amass1(5)+amass1(6)
      if(totmas.gt.w) then
	write(6,*)
     .' CM energy is less than the sum of final particle masses.'
	stop
      end if
*******************************
* Angular cuts in Lab-frame.  *
*******************************
* particle 3
* minimum cos  cut
      COSCUT(1,1)= -1
* maximum cos  cut
      COSCUT(2,1)=  1
* particle 4
* minimum cos  cut
      COSCUT(1,2)= -1
* maximum cos  cut
      COSCUT(2,2)=  1
* particle 5
* minimum cos  cut
      COSCUT(1,3)= -1
* maximum cos  cut
      COSCUT(2,3)=  1
* particle 6
* minimum cos  cut
      COSCUT(1,4)= -1
* maximum cos  cut
      COSCUT(2,4)=  1
 
*******************************
* Energy  cuts in Lab-frame.  *
*******************************
* particle 3
* minimum energy cut
      ENGYCT(1,1) = AMASS1(3)
* maximum energy cut
      ENGYCT(2,1) = W
* particle 4
* minimum energy cut
      ENGYCT(1,2) = AMASS1(4)
* maximum energy cut
      ENGYCT(2,2) = W
* particle 5
* minimum energy cut
      ENGYCT(1,3) = AMASS1(5)
* maximum energy cut
      ENGYCT(2,3) = W
* particle 6
* minimum energy cut
      ENGYCT(1,4) = AMASS1(6)
* maximum energy cut
      ENGYCT(2,4) = W
 
*******************************
* Cut on invariant mass of 5-6*
*******************************
* minimum
      AMASCT(1)= AMASS1(5)+AMASS1(6)
* maximum
      AMASCT(2)= W
 
*******************************
* Q2 singularity treatment    *
*******************************
* no-singularity           : IRESNS= 0
* narrow resonance         : IRESNS= 1
* 1/Q2 singularity         : IRESNS= 2
      IRESNS   = 1
      IF(IRESNS.EQ.1) THEN
C If you want treat narrow resonance, you should set resonance mass
C and width.
        ARESNS(1)=AMwb
        ARESNS(2)=AGwb
c       ARESNS(1)=0
c       ARESNS(2)=0
      END IF
 
*********************************
* mass singularity treatment    *
*********************************
* no-singularity           : ICOS*=0
* mass-singularity         : ICOS*=1
*
*
      ICOS3=1
      ICOS4=0

 
      WRITE(6,*)'*'
      WRITE(6,*)'* Kinematics initialization '
      WRITE(6,*)'*  Code # 2402 '
      WRITE(6,*)'* particle mass ',amass1
      WRITE(6,*)'* CM Energy ',W
      WRITE(6,*)'* COSCUT    '
      WRITE(6,'(2F10.3)')COSCUT
      WRITE(6,*)'* ENGYCT    '
      WRITE(6,'(2F10.3)')ENGYCT
      WRITE(6,*)'* AMASCT    '
      WRITE(6,'(2F10.3)')AMASCT
      WRITE(6,*)'* Singularity treatment '
      WRITE(6,*)'*   Q2'
      IF     (IRESNS.EQ. 2) THEN
         WRITE(6,*)'* 1/Q2 singularity '
      ELSE IF(IRESNS.EQ. 0) THEN
         WRITE(6,*)'* NO   singularity '
      ELSE IF(IRESNS.EQ. 1) THEN
         WRITE(6,*)'* Narrow resonance '
         WRITE(6,*)'*  resonance mass and width'
         WRITE(6,*) ARESNS
      END IF
      IF     (ICOS3.EQ.0) THEN
         WRITE(6,*)'* NO mass-singularity for 3 '
      ELSE
         WRITE(6,*)'*    mass-singularity for 3 '
      END IF
      IF     (ICOS4.EQ.0) THEN
         WRITE(6,*)'* NO mass-singularity for 4 '
      ELSE
         WRITE(6,*)'*    mass-singularity for 4 '
      END IF
      WRITE(6,*)'*'
 
* Following flux factor is for a particle-antiparticle collision
* at s/m**2 >> 1. For an e-gamma collision or low energy interactions
* you should use appropriate formulae.
      VREL = 2
      FLUX = VREL*S
      FACT = GEVPB/FLUX
 
*----- BASES RELATED INITIALIZATIONS -----------------------------------
*--- 2. Dimension of integration variables.
 
          NDIM =   8
          NWILD=   8
 
C-----------------------------------------------------------------------
*--- 3. Region of integration.
 
        DO 1 I=1,NDIM
         XL(I)  =   0.D0
         XU(I)  =   1.D0
         IG(I)  =   1
1       CONTINUE
C-----------------------------------------------------------------------
*--- 4. Number of iterations
      ITMX1  =  5
      ITMX2  =  5
 
      NCALL  =10000
 
*--- 5. Set MXREG : the maximum number of values which are returned
*       by FUNC for one phase space point
 
      MXREG =   1
 
*--- 6. Set histograms
      if( knmgvs .eq. 1 ) then
      NX = 50
      ND = 50
 
      DO 100 I = 1, NDIM
         WRITE(XSTR, 110) I
  110    FORMAT('X(',I2,') SPECTRUM')
         CALL XHINIT(I, XL(I), XU(I), NX, XSTR)
  100 CONTINUE
 
         CALL XHINIT(ndim+1, 0.d0, w, NX, 'Energy of Particle 3')
         CALL XHINIT(ndim+2, 0.d0, w, NX, 'Energy of Particle 4')
         CALL XHINIT(ndim+3, 0.d0, w, NX, 'Energy of Particle 5')
         CALL XHINIT(ndim+4, 0.d0, w, NX, 'Energy of Particle 6')
         CALL XHINIT(ndim+5,-1.d0,1.d0, NX, 'cos_the of Particle 3')
         CALL XHINIT(ndim+6,-1.d0,1.d0, NX, 'cos_the of Particle 4')
         CALL XHINIT(ndim+7,-1.d0,1.d0, NX, 'cos_the of Particle 5')
         CALL XHINIT(ndim+8,-1.d0,1.d0, NX, 'cos_the of Particle 6')
      endif
      RETURN
      END
