C---------------------------------------------------------------------
C   GRACE System Library File
C   KINEM No. : 2403
C   Date      : 2002.07.26
C   Author    : Y.Kurihara
C---------------------------------------------------------------------
      SUBROUTINE KINIT
      IMPLICIT REAL*8(A-H,O-Z)
 
      PARAMETER ( MXDIM = 100 )
      COMMON / LOOP0 / LOOP
      COMMON / BPARM1 / XL(MXDIM),XU(MXDIM),NDIM,NWILD,
     &                 IG(MXDIM),NCALL
      COMMON / BPARM2 / ACC1,ACC2,ITMX1,ITMX2
      COMMON / BASE3 / SI,SI2,SWGT,SCHI,SCALLS,ATACC,NSU,IT,WGT
 
      CHARACTER XSTR*14
 
      INCLUDE 'inclk.h'
*-----------------------------------------------------------------------
      COMMON/KMCNTL/IRESNS,ICOS3,ICOS4,ifrad
      COMMON/KINEM1/S,W,FACT
      COMMON/CUT001/COSCUT(2,4),ENGYCT(2,4),AMASCT(2),ARESNS(2,2)
*-----------------------------------------------------------------------
* 4-body kinematics
*
*  Integration variables
*        COS(the3),Phi3(=0)
*        Q2=(p4+p5)**2
*        COS(the4^),Phi4^    in particle 4,5 CM frame
*
*                         /P4
*                        /
*           p1\         /
*              \      q/\
*               ***** /  \
*              *******    \P5    + final photon
*               ***** \
*               /      \
*              /        \
*           p2/          \p3
*
*
*--- 1. Initialize constants for kinematics.
* S = (P1+P2)**2
* W = SQRT(S)
      W = 1000.D0
      S = W*W
      totmas=amass1(3)+amass1(4)+amass1(5)
      if(totmas.gt.w) then
	write(6,*)
     .' CM energy is less than the sum of final particle masses.'
	stop
      end if
      if(amass1(6).gt.1.d-20) then
       write(6,*)' Particle 6 is NOT photon (or gluon)'
       stop
      end if
*******************************
* Angular cuts in Lab-frame.  *
*******************************
      thec=0
* particle 3
* minimum cos  cut
      COSCUT(1,1)= -cos(thec*rad)
* maximum cos  cut
      COSCUT(2,1)=  cos(thec*rad)
* particle 4
* minimum cos  cut
      COSCUT(1,2)= -cos(thec*rad)
* maximum cos  cut
      COSCUT(2,2)=  cos(thec*rad)
* particle 5
* minimum cos  cut
      COSCUT(1,3)= -cos(thec*rad)
* maximum cos  cut
      COSCUT(2,3)=  cos(thec*rad)
* particle 6 (photon)
* minimum cos  cut
      COSCUT(1,4)= -cos(thec*rad)
* maximum cos  cut
      COSCUT(2,4)=  cos(thec*rad)
 
*******************************
* Energy  cuts in Lab-frame.  *
*******************************
* particle 3
* minimum energy cut
      ENGYCT(1,1) = AMASS1(3)
* maximum energy cut
      ENGYCT(2,1) = (s+amass2(3)-(amass1(4)+amass1(5)
     .+amass1(6))**2)/2.d0/w
* particle 4
* minimum energy cut
      ENGYCT(1,2) = AMASS1(4)
* maximum energy cut
      ENGYCT(2,2) = (s+amass2(4)-(amass1(5)+amass1(3)
     .+amass1(6))**2)/2.d0/w
* particle 5
* minimum energy cut
      ENGYCT(1,3) = AMASS1(5)
* maximum energy cut
      ENGYCT(2,3) = (s+amass2(5)-(amass1(3)+amass1(4)
     .+amass1(6))**2)/2.d0/w
* particle 6 (photon)
* minimum energy cut
c     ENGYCT(1,4) = amass1(6)
      ENGYCT(1,4) = 1.d-1
* maximum energy cut
c     ENGYCT(2,4) = 1.d-1
      ENGYCT(2,4) = (s+amass2(6)-(amass1(3)+amass1(4)
     .+amass1(5))**2)/2.d0/w
 
*******************************
* Cut on invariant mass of 4-5*
*******************************
* minimum
      AMASCT(1)= AMASS1(4)+AMASS1(5)
* maximum
      AMASCT(2)= W-AMASS1(3)
 
*******************************
* Q2 singularity treatment    *
*******************************
* no-singularity           : IRESNS= 0
* narrow resonance (single): IRESNS= 1
* narrow resonance (single)
* + 1/Q2 singularit        : IRESNS=-1
* narrow resonance (double): IRESNS= 2
* narrow resonance (double)
* + 1/Q2 singularit        : IRESNS=-2
* Q2=S peak                : IRESNS= 3
* 1/Q2 singularity only    : IRESNS=-3
* narrow resonance (single)
* + Q2=S peak              : IRESNS= 4
      IRESNS   =  1
      IF( abs(IRESNS).eq.1 .or.
     .    abs(IRESNS).eq.2 .or.
     .    abs(IRESNS).eq.4 ) then
C If you want treat narrow resonance, you should set resonance mass
C and width. (M1 < M2)
        ARESNS(1,1)=AMw
        ARESNS(2,1)=AGw
c       ARESNS(1,2)=AMph
c       ARESNS(2,2)=AGph
      END IF
 
****************************************
* angular distribution of particle3    *
****************************************
* no-singularity           : ICOS3= 0
* 1/t singularity          : ICOS3= 1
* 1/u singularity          : ICOS3=-1
* 1/t + 1/u singularity    : ICOS3= 2
*
      ICOS3=-1
****************************************
* angular distribution of particle 4-5 *
****************************************
* no-singularity           : ICOS4= 0
* z-axis -> -p3
* 1/t singularity          : ICOS4= 1
* 1/t + 1/u singularity    : ICOS4= 2
* z-axis -> p1
* 1/t singularity          : ICOS4=-1
* 1/t + 1/u singularity    : ICOS4=-2
      ICOS4= 0

      ifrad=3

      WRITE(6,*)'*'
      WRITE(6,*)'* Kinematics initialization '
      WRITE(6,*)'*  Code # 2403 '
      WRITE(6,*)'* particle mass ',amass1
      WRITE(6,*)'* CM Energy ',W
      WRITE(6,*)'* COSCUT    '
      WRITE(6,'(2F10.3)')COSCUT
      WRITE(6,*)'* ENGYCT    '
      WRITE(6,'(2F10.3)')ENGYCT
      WRITE(6,*)'* AMASCT    '
      WRITE(6,'(2F10.3)')AMASCT
      WRITE(6,*)'* Singularity treatment '
      WRITE(6,*)'*   Q2'
      IF     (IRESNS.EQ.-3) THEN
         WRITE(6,*)'* 1/Q2 singularity '
      ELSE IF(IRESNS.EQ. 3) THEN
         WRITE(6,*)'* Q2=S peak        '
      ELSE IF(IRESNS.EQ. 0) THEN
         WRITE(6,*)'* NO   singularity '
      ELSE 
         WRITE(6,*)'* Narrow resonance '
         if(iresns.lt.0)then
         write(6,*)'* +1/Q2 singularity '
         else if(iresns.eq.4) then
         write(6,*)'* +Q2=s peak '
         end if
         WRITE(6,*)'*  resonance mass and width'
         WRITE(6,*) ARESNS
      END IF
      write(6,*)'* particle 3   : ',icos3
      write(6,*)'* particle 4-5 : ',icos4
      WRITE(6,*)'*'
 
* Following flux factor is for a particle-antiparticle collision
* at s/m**2 >> 1. For an e-gamma collision or low energy interactions
* you should use appropriate formulae.
      VREL = 2
      FLUX = VREL*S
      FACT = GEVPB/FLUX
 
*----- BASES RELATED INITIALIZATIONS -----------------------------------
*--- 2. Dimension of integration variables.
 
          NDIM =   7
          NWILD=   7
 
C-----------------------------------------------------------------------
*--- 3. Region of integration.
 
        DO 1 I=1,NDIM
         XL(I)  =   0.D0
         XU(I)  =   1.D0
         IG(I)  =   1
1       CONTINUE
C-----------------------------------------------------------------------
*--- 4. Number of iterations
      ITMX1  =  5
      ITMX2  =  5
 
      NCALL  = 10000
 
*--- 5. Set MXREG : the maximum number of values which are returned
*       by FUNC for one phase space point
 
      MXREG =   1
 
*--- 6. Set histograms
      NX = 50

      DO 100 I = 1, NDIM
         WRITE(XSTR, 110) I
  110    FORMAT('X(',I2,') SPECTRUM')
         CALL XHINIT(I, XL(I), XU(I), NX, XSTR)
  100 CONTINUE
         CALL XHINIT(ndim+1, 0.d0, w, NX, 'Energy of Particle 3')
         CALL XHINIT(ndim+2, 0.d0, w, NX, 'Energy of Particle 4')
         CALL XHINIT(ndim+3, 0.d0, w, NX, 'Energy of Particle 5')
         CALL XHINIT(ndim+4, 0.d0, w, NX, 'Energy of Particle 6')
         CALL XHINIT(ndim+5,-1.d0,1.d0, NX, 'cos_the of Particle 3')
         CALL XHINIT(ndim+6,-1.d0,1.d0, NX, 'cos_the of Particle 4')
         CALL XHINIT(ndim+7,-1.d0,1.d0, NX, 'cos_the of Particle 5')
         CALL XHINIT(ndim+8,-1.d0,1.d0, NX, 'cos_the of Particle 6')
      RETURN
      END
