                                                                           /*
*****************************************************************************
*          ===========================================                      *
*            Primitive Graphic routines by X window                         *
*          ===========================================                      *
*                                                                           *
*  (1) void tgdxinit( )                                                     *
*         To initialization of urx parameters                               *
*  (2) void urLine                                                          *
*         To draw a line                                                    *
*  (3) void pstrans( )                                                      *
*         To transfer the coordinates from the window to the paper          *
*  (4) void psLine( )                                                       *
*         To draw a line on a paper                                         *
*  (5) void urplot( )                                                       *
*         To draw a line                                                    *
*  (6) void urarrow( )                                                      *
*         To draw an arrow                                                  *
*  (7) void urwave( )                                                       *
*         To draw a wave                                                    *
*         urLine, pstrans, psLine                                           *
*  (8) void urspiral( )                                                     *
*         To draw a spiral                                                  *
*         urLine, pstrans, psLine                                           *
*  (9) void urdott( )                                                       *
*         To draw a dotted line                                             *
*         urLine, pstrans, psLine                                           *
*  (10) void urcurveto( )                                                   *
*         To draw Bezier 3rd order curve                                    *
*  (11) void urcircle( )                                                    *
*         To draw a filled circle                                           *
*         pstrans                                                           *
*  (12) void urcircle2( )                                                   *
*         To draw a circle                                                  *
*         pstrans                                                           *
*  (13) void FillRectangle( )                                               *
*         To draw filled rectangle                                          *
*  (14) void urtxtda( )                                                     *
*         To draw a textstring on a message board                           *
*  (15) void urtxts( )                                                      *
*         To draw a text string                                             *
*  (16) void urtxts1( )                                                     *
*         To draw a particle name in the format of "form"                   *
*         urMainTimesSize, urMainSymbolSize, urSubTimesSize,                *
*         urSubSymbolSize                                                   *
*         urMainTimesDraw, urMainSymbolDraw, urSubTimesDraw,                *
*         urSubSymbolDraw                                                   *
*         urBar                                                             *
*  (17) void urBar( )                                                       *
*         To draw an upper bar for an anti-particle                         *
*  (18) void urMainTimesSize( )                                             *
*         To get the font information of the normal size of Times           *
*  (19) void urSubTimesSize( )                                              *
*         To get the font information of the sub_size of Times              *
*  (20) void urMainSymbolSize( )                                            *
*         To get the font information of the normal size of Symbol          *
*  (21) void urSubSymbolSize( )                                             *
*         To get the font information of the sub_size of Symbol             *
*  (22) void urMainTimesDraw( )                                             *
*         To draw a character with the normal size of Times                 *
*  (23) void urMainSymbolDraw( )                                            *
*         To draw a character with the normal size of Symbol                *
*  (24) void urSubTimesDraw( )                                              *
*         To draw a character with the sub_size of Times                    *
*  (25) void urSubSymbolDraw( )                                             *
*         To draw a character with the sub_size of Symbol                   *
*  (26) void StringPosition( )                                              *
*         To determine the start position of character string               *
*  (27) void urclr( )                                                       *
*         To clear the window                                               *
*  (28) unsigned long MyColor( )                                            *
*         To get color pixel code                                           *
*                                                                           *
*****************************************************************************
                                                                           */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define   MXVTX  100
#define   MXLINE 100

  Display *d ;
  Window wfm, w0, w1 ;
  XFontStruct *f0, *f1, *f2, *f3, *f4, *f5, *fda;
  XFontStruct *g0, *g1, *g2, *g3, *g4, *g5;
  XCharStruct  chstr;
  GC gc1, gc2, gc3, gcc1, gcc2, gcc3;
  GC gcr0, gcr1, gcr2, gcr3, gcr4, gcr5, gcslct; 
  GC gcg0, gcg1, gcg2, gcg3, gcg4, gcg5;  
  GC gfm, gda00, gda0, gda1, gda2, gda3;
  GC gcrw0, gcrw1, gcrw2, gcrw3, gcrw4, gcrw5; 
int mcolor;
int size_flag, ycenter;
float arrow_size; 
float spiral_width, spiral_length, wave_width, wave_length, dott_length;
FILE *psout;
int print_ps;
float xoff_ps, yoff_ps;
float grymax;
float xxx, yyy;
int xch[MXVTX+MXLINE], ych[MXVTX+MXLINE]; 
int wch[MXVTX+MXLINE], hch[MXVTX+MXLINE], nchar;
double pi, thetat, rdeg;
double xd, yd, pi2;

int flag_dott,  flag_dtarc;
int flag_wave,  flag_wvarc;
int flag_spir,  flag_sparc;

int debug;

                                                                           /*
*****************************************************************************
*                     ========================                             */
                         void tgdxinit ( ) 
/*                    ========================
*****************************************************************************
* ( Function )                                                              *
*   To initialize several parameters for draw Feynman graphs                *
*                                                                           *
*****************************************************************************
									   */
{
   Window  root;
   unsigned int xj, yj, border_width, depth;
   unsigned long white, black, blue, red, f_da, l_da ;
   unsigned long color_fm, color_da, color_das, color_da1;
   unsigned long MyColor() ;
   XGCValues gv ;

   static char *frname[] = { 
    "-adobe-times-bold-i-normal--8-80-75-75-p-47-iso8859-1",
    "-adobe-times-bold-i-normal--10-100-75-75-p-57-iso8859-1",
    "-adobe-times-bold-i-normal--12-120-75-75-p-68-iso8859-1",
    "-adobe-times-bold-i-normal--14-140-75-75-p-77-iso8859-1",
    "-adobe-times-bold-i-normal--18-180-75-75-p-98-iso8859-1",
    "-adobe-times-bold-i-normal--24-240-75-75-p-128-iso8859-1"};
/*  "-adobe-times-bold-i-normal--34-240-100-100-p-170-iso8859-1"};*/

/* static char *frname[] = { 
    "-adobe-times-bold-i-normal--10-100-75-75-p-57-iso8859-1",
    "-adobe-times-bold-i-normal--12-120-75-75-p-68-iso8859-1",
    "-adobe-times-bold-i-normal--14-140-75-75-p-77-iso8859-1",
    "-adobe-times-bold-i-normal--18-180-75-75-p-98-iso8859-1",
    "-adobe-times-bold-i-normal--24-240-75-75-p-128-iso8859-1",
    "-adobe-times-bold-i-normal--24-240-75-75-p-128-iso8859-1"};
    "-adobe-times-bold-i-normal--34-240-100-100-p-170-iso8859-1" };*/

   static char *daname =
    "-adobe-times-medium-r-normal--12-120-75-75-p-64-iso8859-1";
   
      
   static char *fgname[] = { 
    "-adobe-symbol-medium-r-normal--10-100-75-75-p-61-adobe-fontspecific",
    "-adobe-symbol-medium-r-normal--12-120-75-75-p-74-adobe-fontspecific",
    "-adobe-symbol-medium-r-normal--14-140-75-75-p-85-adobe-fontspecific",
    "-adobe-symbol-medium-r-normal--18-180-75-75-p-107-adobe-fontspecific",
    "-adobe-symbol-medium-r-normal--24-240-75-75-p-142-adobe-fontspecific",
    "-adobe-symbol-medium-r-normal--24-240-75-75-p-142-adobe-fontspecific"};
/*  "-adobe-symbol-medium-r-normal--34-240-100-100-p-191-adobe-fontspecific"};*/

       f0 = XLoadQueryFont ( d, frname[0]) ;
       f1 = XLoadQueryFont ( d, frname[1]) ;
       f2 = XLoadQueryFont ( d, frname[2]) ;
       f3 = XLoadQueryFont ( d, frname[3]) ;
       f4 = XLoadQueryFont ( d, frname[4]) ;
       f5 = XLoadQueryFont ( d, frname[5]) ;

       fda= XLoadQueryFont ( d, daname );

       g0 = XLoadQueryFont ( d, fgname[0]) ;
       g1 = XLoadQueryFont ( d, fgname[1]) ;
       g2 = XLoadQueryFont ( d, fgname[2]) ;
       g3 = XLoadQueryFont ( d, fgname[3]) ;
       g4 = XLoadQueryFont ( d, fgname[4]) ;
       g5 = XLoadQueryFont ( d, fgname[5]) ;

       gcr0 = XCreateGC ( d, w0, 0, 0 ) ;
       gcr1 = XCreateGC ( d, w0, 0, 0 ) ;
       gcr2 = XCreateGC ( d, w0, 0, 0 ) ;
       gcr3 = XCreateGC ( d, w0, 0, 0 ) ;
       gcr4 = XCreateGC ( d, w0, 0, 0 ) ;
       gcr5 = XCreateGC ( d, w0, 0, 0 ) ;

       gcslct = XCreateGC ( d, w0, 0, 0 ) ;

       gda0 = XCreateGC ( d, w1, 0, 0 ) ;
       gda1 = XCreateGC ( d, w1, 0, 0 ) ;
       gda2 = XCreateGC ( d, w1, 0, 0 ) ;
       gda3 = XCreateGC ( d, w1, 0, 0 ) ;

       gcg0 = XCreateGC ( d, w0, 0, 0 ) ;
       gcg1 = XCreateGC ( d, w0, 0, 0 ) ;
       gcg2 = XCreateGC ( d, w0, 0, 0 ) ;
       gcg3 = XCreateGC ( d, w0, 0, 0 ) ;
       gcg4 = XCreateGC ( d, w0, 0, 0 ) ;
       gcg5 = XCreateGC ( d, w0, 0, 0 ) ;

       gcrw0 = XCreateGC ( d, w0, 0, 0 ) ;
       gcrw1 = XCreateGC ( d, w0, 0, 0 ) ;
       gcrw2 = XCreateGC ( d, w0, 0, 0 ) ;
       gcrw3 = XCreateGC ( d, w0, 0, 0 ) ;
       gcrw4 = XCreateGC ( d, w0, 0, 0 ) ;
       gcrw5 = XCreateGC ( d, w0, 0, 0 ) ;

       XSetFont ( d, gcr0, f0->fid ) ;
       XSetFont ( d, gcr1, f1->fid ) ;
       XSetFont ( d, gcr2, f2->fid ) ;
       XSetFont ( d, gcr3, f3->fid ) ;
       XSetFont ( d, gcr4, f4->fid ) ;
       XSetFont ( d, gcr5, f5->fid ) ;

       XSetFont ( d, gda1, fda->fid ) ;
       XSetFont ( d, gda2, fda->fid ) ;
       XSetFont ( d, gda3, fda->fid ) ;

       XSetFont ( d, gcg0, g0->fid ) ;
       XSetFont ( d, gcg1, g1->fid ) ;
       XSetFont ( d, gcg2, g2->fid ) ;
       XSetFont ( d, gcg3, g3->fid ) ;
       XSetFont ( d, gcg4, g4->fid ) ;
       XSetFont ( d, gcg5, g5->fid ) ;

       XSetFont ( d, gcrw0, f0->fid ) ;
       XSetFont ( d, gcrw1, f1->fid ) ;
       XSetFont ( d, gcrw2, f2->fid ) ;
       XSetFont ( d, gcrw3, f3->fid ) ;
       XSetFont ( d, gcrw4, f4->fid ) ;
       XSetFont ( d, gcrw5, f5->fid ) ;

       if( mcolor < 3 )
         {
           white = MyColor("black") ;
           blue  = MyColor("black") ;
           black = MyColor("black") ;
           red   = MyColor("black") ;
           f_da  = MyColor("black") ;
           l_da  = MyColor("black") ;
           color_fm  = MyColor("white") ;
           color_da  = MyColor("white") ;
           color_das = MyColor("SlateGrey") ;
           color_da1 = MyColor("white") ;
          }
       else
         {
           white = MyColor("white") ;
           blue  = MyColor("blue") ;
           black = MyColor("black") ;
           red   = MyColor("red") ;
           f_da  = MyColor("black") ;
           l_da  = MyColor("white") ;
           color_fm  = MyColor("ivory") ;
           color_da  = MyColor("DodgerBlue3") ;
           color_das = MyColor("SlateGrey") ;
           color_da1 = MyColor("ivory") ;
          }

       XSetForeground ( d, gcr0, f_da ) ;
       XSetForeground ( d, gcr1, f_da ) ;
       XSetForeground ( d, gcr2, f_da ) ;
       XSetForeground ( d, gcr3, f_da ) ;
       XSetForeground ( d, gcr4, f_da ) ;
       XSetForeground ( d, gcr5, f_da ) ;
       XSetBackground ( d, gcr0, color_da ) ;
       XSetBackground ( d, gcr1, color_da ) ;
       XSetBackground ( d, gcr2, color_da ) ;
       XSetBackground ( d, gcr3, color_da ) ;
       XSetBackground ( d, gcr4, color_da ) ;
       XSetBackground ( d, gcr5, color_da ) ;

       XSetForeground ( d, gcslct, color_das ) ;
       XSetBackground ( d, gcslct, color_da ) ;

       XSetForeground ( d, gda1, blue ) ;
       XSetForeground ( d, gda2, red ) ;
       XSetForeground ( d, gda3, black ) ;
       XSetBackground ( d, gda1, color_da1 ) ;
       XSetBackground ( d, gda2, color_da1 ) ;
       XSetBackground ( d, gda3, color_da1 ) ;

       XSetForeground ( d, gcg0, f_da ) ;
       XSetForeground ( d, gcg1, f_da ) ;
       XSetForeground ( d, gcg2, f_da ) ;
       XSetForeground ( d, gcg3, f_da ) ;
       XSetForeground ( d, gcg4, f_da ) ;
       XSetForeground ( d, gcg5, f_da ) ;
       XSetBackground ( d, gcg0, color_da ) ;
       XSetBackground ( d, gcg1, color_da ) ;
       XSetBackground ( d, gcg2, color_da ) ;
       XSetBackground ( d, gcg3, color_da ) ;
       XSetBackground ( d, gcg4, color_da ) ;
       XSetBackground ( d, gcg5, color_da ) ;

       gv.line_width = 1 ;
       gc1 = XCreateGC ( d, w0, GCLineWidth, &gv) ;
       gcc1 = XCreateGC ( d, w0, GCLineWidth, &gv) ;

       gv.line_width = 2 ;
       gc2 = XCreateGC ( d, w0, GCLineWidth, &gv) ;
       gcc2 = XCreateGC ( d, w0, GCLineWidth, &gv) ;

       gv.line_width = 4 ;
       gc3 = XCreateGC ( d, w0, GCLineWidth, &gv) ;
       gcc3 = XCreateGC ( d, w0, GCLineWidth, &gv) ;

       XSetForeground ( d, gcc1, f_da ) ;
       XSetForeground ( d,  gc1, l_da ) ;
       XSetForeground ( d, gcc2, f_da ) ;
       XSetForeground ( d,  gc2, l_da ) ;
       XSetForeground ( d, gcc3, f_da ) ;
       XSetForeground ( d,  gc3, l_da ) ;
       XSetBackground ( d, gcc1, color_da ) ;
       XSetBackground ( d,  gc1, color_da ) ;
       XSetBackground ( d, gcc2, color_da ) ;
       XSetBackground ( d,  gc2, color_da ) ;
       XSetBackground ( d, gcc3, color_da ) ;
       XSetBackground ( d,  gc3, color_da ) ;

       XSetForeground ( d, gcrw0, white ) ;
       XSetForeground ( d, gcrw1, white ) ;
       XSetForeground ( d, gcrw2, white ) ;
       XSetForeground ( d, gcrw3, white ) ;
       XSetForeground ( d, gcrw4, white ) ;
       XSetForeground ( d, gcrw5, white ) ;
       XSetBackground ( d, gcrw0, color_da ) ;
       XSetBackground ( d, gcrw1, color_da ) ;
       XSetBackground ( d, gcrw2, color_da ) ;
       XSetBackground ( d, gcrw3, color_da ) ;
       XSetBackground ( d, gcrw4, color_da ) ;
       XSetBackground ( d, gcrw5, color_da ) ;

   return ;
} 
                                                                           /*
*****************************************************************************
*                 =================================                        */
                    void urLine( x1, y1, x2, y2 )
/*                =================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a line from ( x1, y1 ) to ( x2, y2 )                            *
*                                                                           *
*****************************************************************************
									   */
float x1, y1, x2, y2;
{
int ix1, iy1, ix2, iy2;

 ix1 = (int) x1, iy1 = (int) y1;
 ix2 = (int) x2, iy2 = (int) y2;

 if( print_ps == 0 )
   {
     if( size_flag < 2 ) 
         XDrawLine ( d, w0, gc1, ix1, iy1, ix2, iy2 ) ;
     else if( size_flag < 4 ) 
         XDrawLine ( d, w0, gc2, ix1, iy1, ix2, iy2 ) ;
     else 
         XDrawLine ( d, w0, gc3, ix1, iy1, ix2, iy2 ) ;
    }
/*
 if( debug == 1 )
   {
     printf(" urLine  ix1 = %5d, iy1 = %5d\n", ix1, iy1 );
     printf(" urLine  ix2 = %5d, iy2 = %5d\n", ix2, iy2 );
    }
*/
}

                                                                           /*
*****************************************************************************
*                 =================================                        */
                  void urArc( xc, yc, r, th0, th1 )
/*                =================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a line from ( x1, y1 ) to ( x2, y2 )                            *
*                                                                           *
*****************************************************************************
									   */
float xc, yc, r, th0, th1;
{
int ixc, iyc, ir, ith0, ith1;
float rpi = 180./3.1415927;



 if( print_ps == 0 )
   {
     ixc = (int) ( xc - r ), iyc = (int) ( yc - r );
     r   = 2 * r;
     ir  = (int) r;
     th1 = ( th1 - th0 ) * rpi;
     th0 = th0 * rpi;
     ith0 = 64 * (int) th0;
     ith1 = 64 * (int) th1;

     if( size_flag < 2 ) 
         XDrawArc ( d, w0, gc1, ixc, iyc, ir, ir, ith0, ith1 );
     else if( size_flag < 4 ) 
         XDrawArc ( d, w0, gc2, ixc, iyc, ir, ir, ith0, ith1 );
     else 
         XDrawArc ( d, w0, gc3, ixc, iyc, ir, ir, ith0, ith1 );
    }
 else
   {
     fprintf( psout, "newpath\n");
     fprintf( psout," %6.1f  %6.1f  %6.1f  %6.1f %6.1f  arc\n",
              xc, yc, r, th0*rpi, th1*rpi );
     fprintf( psout," stroke\n");
    }
}

                                                                           /*
*****************************************************************************
*                 =================================                        */
                      void pstrans( x0, y0 )
/*                =================================
*****************************************************************************
* ( Function )                                                              *
*   To translate the x-y coordinate on a window                             *
*      to the x-y coordinate on a paper for ps file.                        *
*     (  x0,  y0 ) : x-y coordinate on a window                             *
*     ( xxx, yyy ) : x-y coordinate in a pare                               *
*                                                                           *
*****************************************************************************
									   */
float x0, y0;
{
     xxx = x0  + xoff_ps;
     yyy = grymax - y0 + yoff_ps;
}

                                                                           /*
*****************************************************************************
*                 =================================                        */
                    void psLine( x1, y1, x2, y2 )
/*                =================================
*****************************************************************************
* ( Function )                                                              *
*   To write ps command to draw a line from ( x1, y1 ) to ( x2, y2 ),       *
*      which is given by the global coordinate on a window                  *
*                                                                           *
*****************************************************************************
									   */
float x1, y1, x2, y2;
{
 if( print_ps > 0 )
   {
     xxx = x1  + xoff_ps;
     yyy = grymax - y1 + yoff_ps;
     fprintf( psout, " %9.3f %9.3f moveto",   xxx, yyy );
     xxx = x2  + xoff_ps;
     yyy = grymax - y2 + yoff_ps;
     fprintf( psout, " %9.3f %9.3f lineto\n", xxx, yyy );
    }
}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
            void    urplot( xx1, yy1, xx2, yy2, xx3, yy3 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a solid line from ( xx1, yy1 ) to ( xx3, yy3 ),                 *
*      and ( xx2, yy2 ) defines the middle point of the line.               *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2, xx3, yy3;
{                                                                          /*
 ----------------
  Local Variable
 ----------------
                                                                           */
   double x0, y0, x1, y1, x2, y2, x3, y3, xc, yc, length, div, theta;
   double sint, cost, xt,  yt;
   double a0, b0, a1, b1, theta0, theta1, r, th;
   int   ntimes, n, mo;
                                                                           /*
 ----------------
  Functions
 ----------------
                                                                           */
 void get_radius();
 void pstrans();
 void urLine();
 void urArc();
                                                                           /*
================================ Entry Point ================================
                                                                           */
   if( print_ps > 0 )
     {
       pstrans( xx1, yy1 );
       xx1 = xxx, yy1 = yyy;
       pstrans( xx3, yy3 );
       xx3 = xxx, yy3 = yyy;
      }

if( xx2 == 0.0 )
  {
   if( print_ps == 0 )
     {
       urLine( xx1, yy1, xx3,  yy3 );
      }
   else
     {
       fprintf( psout, "newpath\n");
       fprintf( psout, " %8.2f %8.2f moveto", xx1, yy1 );
       fprintf( psout, " %8.2f %8.2f lineto\n", xx3, yy3 );
       fprintf( psout, "stroke\n");
      }
  }
else
  {
   if( print_ps > 0 )
     {
       pstrans( xx2, yy2 );
       xx2 = xxx, yy2 = yyy;
      }

   x0 = (double) xx1;
   y0 = (double) yy1;
   x1 = (double) xx3;
   y1 = (double) yy3;
   x2 = (double) xx2;
   y2 = (double) yy2;

   get_radius( x0, y0, x1, y1, x2, y2,
               &xc, &yc,  &r, &theta, &theta0, &theta1 );

   if( print_ps == 0 )
     {
       theta0 = atan2( yc - y0, x0 - xc );
       theta1 = atan2( yc - y1, x1 - xc );
       theta  = theta1 - theta0;
      }
   else
     {
       theta0 = atan2( y0 - yc, x0 - xc );
       theta1 = atan2( y1 - yc, x1 - xc );
       theta  = theta1 - theta0;
      }
      /*
       printf(" print_ps =%2d\n", print_ps );
       printf(" x0  = %6.1f,   x1  = %6.1f  x2 = %6.1f\n", x0, x1, x2 );
       printf(" y0  = %6.1f,   y1  = %6.1f  y2 = %6.1f\n", y0, y1, y2 );
       printf(" xc  = %6.1f\n", *xc );
       printf(" yc  = %6.1f\n", *yc );
       printf(" r0  = %6.1f\n", *r );
     /*
       printf(" print_ps =%2d\n", print_ps );
       printf(" theta1  = %6.1f,   theta0  = %6.1f\n", 
	        theta1*rdeg, theta0*rdeg );
      */


   if( theta0 >= 0.0 )
     {
       if( theta1 >= 0.0 )
         {
           if( theta0 > theta1 ) 
             {
               theta  = theta0;
               theta0 = theta1;
               theta1 = theta;
              }
          }
        else
         {
           theta =  theta1 - theta0;
           if( theta < -pi )
             {
               theta1 = theta1 + pi2;
              }
           else if( theta < 0.0 )
             {
               theta  = theta1;
               theta1 = theta0;
               theta0 = theta;
              }
          }
      }
   else
     {
       if( theta1 >= 0.0 )
         {
           if( theta1 - theta0 > pi )
             {
               theta  = theta1;
               theta1 = pi2 + theta0;
               theta0 = theta;
              }
          }
       else
         {
           if( theta0 > theta1 ) 
             {
               theta  = theta0;
               theta0 = theta1;
               theta1 = theta;
              }
          }
      }

         /*
          theta = theta1 - theta0;
          printf(" theta1  = %6.1f,   theta0  = %6.1f  theta = %6.1f\n", 
	            theta1*rdeg, theta0*rdeg, theta*rdeg );
         */

   urArc( (float) xc, (float) yc, (float) r, (float) theta0, (float) theta1 );

   }

   return ;
}

                                                                           /*
*****************************************************************************
*                 =================================                        */
                  void  urplotx( xx1, yy1, xx2, yy2 )
/*                =================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a line from ( xx1, yy1 ) to ( xx2, yy2 ), which is              *
*      given by the global coordinate on a window.                          *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2;
{
   int x0, y0, x1, y1;

   if( print_ps == 0 )
     {
       x0 = (int) xx1 ;
       y0 = (int) yy1 ;
       x1 = (int) xx2 ;
       y1 = (int) yy2 ;
       if( size_flag < 2 ) 
           XDrawLine ( d, w0, gc1, x0, y0, x1, y1 ) ;
       else if( size_flag < 4 ) 
           XDrawLine ( d, w0, gc2, x0, y0, x1, y1 ) ;
       else 
           XDrawLine ( d, w0, gc3, x0, y0, x1, y1 ) ;
      }
   else
     {
       pstrans( xx1, yy1 );
       fprintf( psout, "newpath %9.3f %9.3f moveto", xxx, yyy );
       pstrans( xx2, yy2 );
       fprintf( psout, " %9.3f %9.3f lineto  stroke\n", xxx, yyy );
      }

   return ;
}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
             float urarrow(  xx1, yy1, xxp, yyp, xx2, yy2, direct )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To write an arrow on a line from ( xx1, yy1 ) to ( xx2, yy2 ),          *
*      whose direction is determined by the sign of "direct".               *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2, xxp, yyp;
int direct;
{                                                                          /*
 ----------------
  Local Variable
 ----------------
                                                                           */
   double x0, y0, x1, y1, xp, yp, length, half, theta;
   double sint, cost, a, xt,  yt, deposit;
   int x2, y2, n;
   float width;
   static XPoint p[5];
                                                                           /*
============================== Entry Point ==================================
                                                                           */
   if( print_ps > 0 )
     {
       pstrans( xx1, yy1 );
       xx1 = xxx,   yy1 = yyy;
       pstrans( xx2, yy2 );
       xx2 = xxx,   yy2 = yyy;
      }
   width = 0;
   x0 = (double) xx1;
   y0 = (double) yy1;
   x1 = (double) xx2;
   y1 = (double) yy2;
   
   if( direct < 0 )
       direct = -1;
   else
       direct =  1;

   length  = ( x1 - x0 )*( x1 - x0 ) + ( y1 - y0 )*( y1 - y0 );
   length  = 0.5*sqrt( length );
   deposit = 0.4*length;
   a       = (double) arrow_size;

   if( a > 0.5*length )
       a = 0.5*length;

   half  = 0.5*a;
   width = ( float ) half;

   theta = atan2( y1-y0, x1-x0 );
   sint  = sin( theta );
   cost  = cos( theta );

if( xxp == 0.0 )
  {
   xt    = length - direct*( a - half - deposit );
   yt    = -half;
   x1 =  xt*cost - yt*sint;
   y1 =  xt*sint + yt*cost;
   x1 = x1 + x0;
   y1 = y1 + y0;
   x2 = (int) x1 ;
   y2 = (int) y1 ;
   p[0].x = x2, p[0].y = y2;

   xt    = length + direct*( half + deposit );
   yt    = 0.0;
   x1 =  xt*cost - yt*sint;
   y1 =  xt*sint + yt*cost;
   x1 = x1 + x0;
   y1 = y1 + y0;
   x2 = (int) x1 ;
   y2 = (int) y1 ;
   p[1].x = x2, p[1].y = y2;

   xt    = length - direct*( a - half - deposit );
   yt    = half;
   x1 =  xt*cost - yt*sint;
   y1 =  xt*sint + yt*cost;
   x1 = x1 + x0;
   y1 = y1 + y0;
   x2 = (int) x1 ;
   y2 = (int) y1 ;
   p[2].x = x2, p[2].y = y2;

   xt    = length - direct*( half*1.2 - half - deposit );
   yt    = 0.0;
   x1 =  xt*cost - yt*sint;
   y1 =  xt*sint + yt*cost;
   x1 = x1 + x0;
   y1 = y1 + y0;
   x2 = (int) x1 ;
   y2 = (int) y1 ;
   p[3].x = x2, p[3].y = y2;

   xt    = length - direct*( a - half - deposit );
   yt    = -half;
   x1 =  xt*cost - yt*sint;
   y1 =  xt*sint + yt*cost;
   x1 = x1 + x0;
   y1 = y1 + y0;
   x2 = (int) x1 ;
   y2 = (int) y1 ;
   p[4].x = x2, p[4].y = y2;
  }
else
  {
   if( print_ps > 0 )
     {
       pstrans( xxp, yyp );
       xxp = xxx,  yyp = yyy;
      }
   xt    =  - direct*( a - half);
   yt    = -half;
   x1 =  xt*cost - yt*sint;
   y1 =  xt*sint + yt*cost;
   x1 = x1 + xxp;
   y1 = y1 + yyp;
   x2 = (int) x1 ;
   y2 = (int) y1 ;
   p[0].x = x2, p[0].y = y2;

   xt    =  direct*( half );
   yt    = 0.0;
   x1 =  xt*cost - yt*sint;
   y1 =  xt*sint + yt*cost;
   x1 = x1 + xxp;
   y1 = y1 + yyp;
   x2 = (int) x1 ;
   y2 = (int) y1 ;
   p[1].x = x2, p[1].y = y2;

   xt    =  - direct*( a - half );
   yt    = half;
   x1 =  xt*cost - yt*sint;
   y1 =  xt*sint + yt*cost;
   x1 = x1 + xxp;
   y1 = y1 + yyp;
   x2 = (int) x1 ;
   y2 = (int) y1 ;
   p[2].x = x2, p[2].y = y2;

   xt    = - direct*( half*1.2 - half );
   yt    = 0.0;
   x1 =  xt*cost - yt*sint;
   y1 =  xt*sint + yt*cost;
   x1 = x1 + xxp;
   y1 = y1 + yyp;
   x2 = (int) x1 ;
   y2 = (int) y1 ;
   p[3].x = x2, p[3].y = y2;

   xt    = - direct*( a - half );
   yt    = -half;
   x1 =  xt*cost - yt*sint;
   y1 =  xt*sint + yt*cost;
   x1 = x1 + xxp;
   y1 = y1 + yyp;
   x2 = (int) x1 ;
   y2 = (int) y1 ;
   p[4].x = x2, p[4].y = y2;
  }

   if( print_ps == 0 )
     {
       XSetFillRule( d, gc1, WindingRule );
       XFillPolygon( d, w0, gc1, p, 5, Complex, CoordModeOrigin );
      }
   else
     {
       fprintf( psout, "newpath\n");
       fprintf( psout, " %9.3f %9.3f moveto\n", 
                         (float) p[0].x, (float) p[0].y );
       xx1 = (float) p[1].x;
       yy1 = (float) p[1].y;
       fprintf( psout, " %9.3f %9.3f lineto %9.3f %9.3f lineto\n",
                xx1, yy1, (float) p[2].x, (float) p[2].y );
       xx1 = (float) p[3].x;
       yy1 = (float) p[3].y;
       fprintf( psout, " %9.3f %9.3f lineto %9.3f %9.3f lineto\n",
                xx1, yy1, (float) p[4].x, (float) p[4].y );
       fprintf( psout, " closepath fill\n");
      }

   return width;
}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
             float urwave(  xx1, yy1, xx2, yy2, xx3, yy3 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a wave on a line from ( xx1, yy1 ) to ( xx3, yy3 ),             *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2, xx3, yy3;
{                                                                          /*
 ------------------
  Local Variables
 ------------------
                                                                           */
 float width;
                                                                           /*
 ------------
  Functions
 ------------
                                                                           */
 float urwaveto( );
 float urwvarc(  );
 void  pstrans(  );
                                                                           /*
================================= Entry point ===============================
                                                                           */
  if( print_ps > 0 )
    {
      pstrans( xx1, yy1 );
      xx1 = xxx, yy1 = yyy;
      pstrans( xx3, yy3 );
      xx3 = xxx, yy3 = yyy;
     }
                                                                             /*
===============================================================================
  Draw a straight line by waves
===============================================================================
                                                                             */
if( xx2 == 0.0 )
  {
     width = urwaveto( xx1, yy1, xx3, yy3 );
   }
                                                                             /*
===============================================================================
  Draw arc by waves
===============================================================================
                                                                             */
else
  {
    if( print_ps > 0 )
      {
        pstrans( xx2, yy2 );
        xx2  = xxx,  yy2  = yyy;
       }

     width = urwvarc( xx1, yy1, xx2, yy2, xx3, yy3 );

    }  

   return width;
}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
                float urwaveto(  xx1, yy1, xx3, yy3 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a stright line in terms of waves                                *
*      from ( xx1, yy1 ) to ( xx3, yy3 )                                    *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx3, yy3;
{                                                                          /*
-------------------
  Local variables
-------------------
                                                                           */
   double x0, y0, x1, y1, x2, y2, x3, y3, length, lambda, div, div0, theta;
   double sint, cost, s, a, b, xt,  yt, xt1, yt1;
   double xc, yc, a0, b0, a1, b1, r, r1, theta0, theta1, th;
   int ix1, iy1, ix2, iy2, ntimes, n, n0, n1, nx;
   int mo;
   float xx0, yy0, xx2, yy2, width, dir;
   static int mintime1 = 5, mintime2 = 10;
                                                                           /*
-----------------------------
  Functions
-----------------------------
                                                                           */
   int mode( );
   void urcurveto( );
                                                                           /*
================================= Entry point ===============================
                                                                           */
   x0 = (double) xx1;
   y0 = (double) yy1;
   x1 = (double) xx3;
   y1 = (double) yy3;
                                                                             /*
===============================================================================
  Draw a straight line by waves
===============================================================================
                                                                             */
                                                                             /*
   +--------------------------------------------------------------------------+
   |(1) calculat the length of wave "length" and number of half_waves "ntimes"|
   |     from the given wave_length.                                          |
   +--------------------------------------------------------------------------+
                                                                             */
   length = ( x1 - x0 )*( x1 - x0 ) + ( y1 - y0 )*( y1 - y0 );
   length = sqrt( length );
   div    = length / wave_length * 2.0;
   ntimes = (int) div - 1;
   if( ntimes < mintime1 )
       ntimes = mintime1;
   else if( mode( ntimes, 2 ) == 0 )
       ntimes = ntimes + 1;

   lambda = length /( ( float ) ntimes + 1.0 );
   width = wave_width;
   width = lambda*0.6;
   dir   = lambda*0.7;
   if( print_ps == 0 ) dir = -dir;
                                                                            /*
   +-------------------------------------------------------------------------+
   |(2) If flag_wave > 0 and print_ps == 1, print ps_command                 |
   +-------------------------------------------------------------------------+
                                                                            */
   if( print_ps == 1 && flag_wave > 0 )
     {
       fprintf( psout,"gsave\nnewpath\n");
       fprintf( psout," %8.2f %8.2f  moveto\n",    xx1, yy1 );
       fprintf( psout," %8.2f %8.2f %2d waveto\n", xx3, yy3, ntimes );
       fprintf( psout,"stroke\ngrestore\n");
      }
                                                                            /*
   +-------------------------------------------------------------------------+
   |(2) If not, draw wave line explicitely.                                  |
   +-------------------------------------------------------------------------+
                                                                            */
   else
     {                                                                       /*
       +----------------------------------------------------------------------+
       |(a) calculat the transformation matrix                                |
       +----------------------------------------------------------------------+
                                                                             */
       theta = atan2( y1-y0, x1-x0 );
       sint  = sin( theta );
       cost  = cos( theta );
                                                                             /*
       +----------------------------------------------------------------------+
       |(b) Draw a line of length half of half_wave length                    |
       +----------------------------------------------------------------------+
                                                                             */
       xx0   = (float) x0,  yy0 = (float) y0;
       xt    = 0.5*lambda;
       yt    = 0.0;
       xt1 =  xt*cost - yt*sint + x0;
       yt1 =  xt*sint + yt*cost + y0;
       xx1    = (float) xt1,  yy1 = (float) yt1;

       if( print_ps == 0 )
         {
           urLine( xx0, yy0, xx1, yy1 );
          }
       else
         {
           fprintf( psout, "newpath\n");
           fprintf( psout, " %8.2f %8.2f moveto", xx0, yy0 );
           fprintf( psout, " %8.2f %8.2f lineto\n", xx1, yy1 );
          }
       xx0 = xx1, yy0 = yy1;
                                                                             /*
       +----------------------------------------------------------------------+
       |(c) Draw waves by ntimes half_waves                                   |
       +----------------------------------------------------------------------+
                                                                             */
       for( n0 = 0; n0 < ntimes; n0++ )
          {
            xt = xt + 0.4*lambda;
            yt = dir;
            xt1 =  xt*cost - yt*sint + x0;
            yt1 =  xt*sint + yt*cost + y0;
            xx1    = (float) xt1,  yy1 = (float) yt1;
            xt = xt + 0.2*lambda;
            xt1 =  xt*cost - yt*sint + x0;
            yt1 =  xt*sint + yt*cost + y0;
            xx2    = (float) xt1,  yy2 = (float) yt1;
            xt = xt + 0.4*lambda;
            xt1 =  xt*cost + x0;
            yt1 =  xt*sint + y0;
            xx3    = (float) xt1,  yy3 = (float) yt1;

           if( print_ps == 0 )
             {
               urcurveto( xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3 );
              }
           else
             {
               fprintf( psout, " %8.2f %8.2f", xx1, yy1 );
               fprintf( psout, " %8.2f %8.2f", xx2, yy2 );
               fprintf( psout, " %8.2f %8.2f curveto\n", xx3, yy3 );
              }

            xx0 = xx3,  yy0 = yy3;
            dir = -dir;
          }
                                                                             /*
       +----------------------------------------------------------------------+
       |(d) Draw a line of length half of half_wave length                    |
       +----------------------------------------------------------------------+
                                                                             */
       xt    = xt + 0.5*lambda;
       xt1 =  xt*cost + x0;
       yt1 =  xt*sint + y0;
       xx1    = (float) xt1,  yy1 = (float) yt1;

       if( print_ps == 0 )
         {
           urLine( xx0, yy0, xx1, yy1 );
          }
       else
         {
           fprintf( psout, " %8.2f %8.2f lineto\n", xx1, yy1 );
           fprintf( psout, "stroke\n");
          }
      }

 return width;

}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
             float urwvarc(  xx1, yy1, xx2, yy2, xx3, yy3 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a wave on a line from ( xx1, yy1 ) to ( xx3, yy3 ),             *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2, xx3, yy3;
{                                                                          /*
-----------------------------
  Local variables
-----------------------------
                                                                           */
   double x0, y0, x1, y1, x2, y2, x3, y3, length, lambda, div, div0, theta;
   double sint, cost, s, a, b, xt,  yt, xt1, yt1;
   double xc, yc, a0, b0, a1, b1, r, r1, theta0, theta1, th;
   int ix1, iy1, ix2, iy2, ntimes, n, n0, n1, nx;
   int mo;
   float xx0, yy0, width, dir, rpi;
   static int mintime1 = 5, mintime2 = 10;
                                                                           /*
-----------------------------
  Functions
-----------------------------
                                                                           */
   int mode( );
   void urcurveto( );
   void get_radius();
                                                                           /*
================================= Entry point ===============================
                                                                           */
   x0 = (double) xx1;
   y0 = (double) yy1;
   x1 = (double) xx3;
   y1 = (double) yy3;
                                                                             /*
===============================================================================
  Draw arc by waves
===============================================================================
                                                                             */
                                                                             /*
   +--------------------------------------------------------------------------+
   |(1) calculate the center of the arc and the opening angle                 |
   +--------------------------------------------------------------------------+
                                                                             */
   x2 = (double) xx2;
   y2 = (double) yy2;

   get_radius( x0, y0, x1, y1, x2, y2,
               &xc, &yc,  &r, &theta, &theta0, &theta1 );

   if( theta > 0.0 ) length = theta;
   else              length = -theta;

   div    = length*r / wave_length*2;
   ntimes = (int) div;
   if( ntimes < mintime2 )
       ntimes = mintime2;
   if( mode( ntimes, 2 ) == 0 )
       ntimes = ntimes + 1;

   lambda = length / (( float ) ntimes + 1.0 );
   dir   = lambda*r;
   width = dir*0.6;
   dir   = dir*0.7;
      /*
        printf(" ntimes = %d, div = %15.6f  length = %15.6f\n", 
	         ntimes, div, length );
      */
                                                                            /*
   +-------------------------------------------------------------------------+
   |(2) If flag_wvarc > 0 and print_ps == 1, print ps_command                |
   +-------------------------------------------------------------------------+
                                                                            */
   if( print_ps == 1 && flag_wvarc > 0 )
     {
       rpi = 180.0 / pi;
       fprintf( psout,"gsave\nnewpath\n");
       fprintf( psout," %8.2f %8.2f  moveto\n",    xx1, yy1 );
       fprintf( psout," %8.2f %8.2f %8.2f", 
               (float) xc, (float) yc, (float) r );
       fprintf( psout," %8.2f %8.2f %2d wvarc\n", 
                rpi* (float) theta0, rpi* (float) theta1, ntimes );
       fprintf( psout,"stroke\ngrestore\n");
      }
                                                                            /*
   +-------------------------------------------------------------------------+
   |(3) If not, draw wave arc explicitely.                                   |
   +-------------------------------------------------------------------------+
                                                                            */
   else
     {
                                                                             /*
       +----------------------------------------------------------------------+
       |(a) Draw a line of length half of half_wave length                    |
       +----------------------------------------------------------------------+
                                                                             */
       xx0    = ( float ) x0,  yy0 = ( float ) y0;
       xt    = 0.5*lambda;
       if( theta > 0.0 ) th = theta0 + xt;
       else              th = theta0 - xt;

       cost = cos( th );
       sint = sin( th );

       x3 =  r * cost + xc;
       y3 =  r * sint + yc;
       xx1    = (float) x3,  yy1 = (float) y3;

       if( print_ps == 0 )
         {
           urLine( xx0, yy0, xx1, yy1 );
          }
       else
         {
           fprintf( psout, "newpath\n");
           fprintf( psout, " %8.2f %8.2f moveto", xx0, yy0 );
           fprintf( psout, " %8.2f %8.2f lineto\n", xx1, yy1 );
          }

       xx0 = xx1, yy0 = yy1;
                                                                             /*
       +----------------------------------------------------------------------+
       |(b) Draw arc of waves by ntimes half_waves                            |
       +----------------------------------------------------------------------+
                                                                             */
       for( n0 = 0; n0 < ntimes; n0++ )
          {
            xt = xt + 0.4*lambda;
            if( theta > 0.0 ) th = theta0 + xt;
            else              th = theta0 - xt;
            cost = cos( th );
            sint = sin( th );
            r1   = r + dir;
            x3 =  r1 * cost + xc;
            y3 =  r1 * sint + yc;
            xx1 = (float) x3, yy1 = (float) y3;
            xt = xt + 0.2*lambda;
            if( theta > 0.0 ) th = theta0 + xt;
            else              th = theta0 - xt;
            cost = cos( th );
            sint = sin( th );
            x3 =  r1 * cost + xc;
            y3 =  r1 * sint + yc;
            xx2 = (float) x3, yy2 = (float) y3;
            xt = xt + 0.4*lambda;
            if( theta > 0.0 ) th = theta0 + xt;
            else              th = theta0 - xt;
            cost = cos( th );
            sint = sin( th );
            x3 =  r * cost + xc;
            y3 =  r * sint + yc;
            xx3 = (float) x3, yy3 = (float) y3;
                                                                             /*
           +------------------------------------------------------------------+
           |(b') Output the line to ps file                                   |
           +------------------------------------------------------------------+
                                                                             */
           if( print_ps == 0 )
             {
               urcurveto( xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3 );
              }
           else if( flag_wvarc == 0 )
             {
               fprintf( psout, " %8.2f %8.2f", xx1, yy1 );
               fprintf( psout, " %8.2f %8.2f", xx2, yy2 );
               fprintf( psout, " %8.2f %8.2f  curveto\n", xx3, yy3 );
              }

            xx0 = xx3,  yy0 = yy3;
            dir = -dir;

          }
                                                                             /*
       +----------------------------------------------------------------------+
       |(c) Draw a line of length half of half_wave length                    |
       +----------------------------------------------------------------------+
                                                                             */
       xt    = xt + 0.5*lambda;
       if( theta > 0.0 ) th = theta0 + xt;
       else              th = theta0 - xt;
       cost = cos( th );
       sint = sin( th );
       x3 =  r * cost + xc;
       y3 =  r * sint + yc;
       xx1 = (float) x3, yy1 = (float) y3;
                                                                             /*
       +----------------------------------------------------------------------+
       |(c') Output the line to ps file                                       |
       +----------------------------------------------------------------------+
                                                                             */
       if( print_ps == 0 )
         {
           urLine( xx0, yy0, xx1, yy1 );
          }
       else if( flag_wvarc == 0 )
         {
           fprintf( psout, " %8.2f %8.2f lineto\n", xx1, yy1 );
           fprintf( psout, "stroke\n");
          }
      }

   return width;
}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
             float urwvcircl(  xx1, yy1, xx2, yy2 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a circle by the wave, of which diameter is defined              *
*      by the points (xx1, yy1) and (xx2, yy2).                             *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2;
{                                                                          /*
-----------------------------
  Local variables
-----------------------------
                                                                           */
   double x0, y0, x1, y1, x2, y2, x3, y3, length, lambda, div, div0, theta;
   double sint, cost, s, a, b, xt,  yt, xt1, yt1;
   double xc, yc, a0, b0, a1, b1, r, r1, theta0, theta1, th;
   int ix1, iy1, ix2, iy2, ntimes, n, n0, n1, nx;
   int mo;
   float xx0, yy0, width, dir, rpi;
   static int mintime1 = 5, mintime2 = 10;

   float xx3, yy3;
                                                                           /*
-----------------------------
  Functions
-----------------------------
                                                                           */
   int mode( );
   void urcurveto( );
   void get_radius();
                                                                           /*
================================= Entry point ===============================
                                                                           */
   x0 = (double) xx1;
   y0 = (double) yy1;
   x1 = (double) xx2;
   y1 = (double) yy2;
/**
   printf(" (xx1, yy1) = ( %8.2f, %8.2f ) \n", xx1, yy1 );
   printf(" (xx2, yy2) = ( %8.2f, %8.2f ) \n", xx2, yy2 );
                                                                             /*
===============================================================================
  Draw arc by waves
===============================================================================
                                                                             */
                                                                             /*
   +--------------------------------------------------------------------------+
   |(1) calculate the center of the arc and the radious                       |
   +--------------------------------------------------------------------------+
                                                                             */
   xc = 0.5*( x0 + x1);
   yc = 0.5*( y0 + y1 );
   r  = ( x0 - xc )*( x0 - xc ) + ( y0 - yc )*( y0 - yc );
   r  = sqrt( r );
/*
   if( x0 > x1 - 1.0 && x0 < x1 + 1.0 )
     {
       if( y0 < y1 )
           theta0 = - 0.5*pi;
       else
           theta0 =   0.5*pi;
      }
   else
*/
       theta0 = atan2( y0 - yc, x0 - xc );

/**
   printf(" ( xc, yc ) = ( %8.2f, %8.2f ), r = %8.2f\n", xc, yc, r );
   printf(" theta0 = %8.2f\n", theta0 );
/**/
   length = pi2;
   div    = length*r / wave_length*2;
   ntimes = (int) div;
   if( ntimes < mintime2 )
       ntimes = mintime2;
   if( mode( ntimes, 2 ) == 1 )
       ntimes = ntimes + 1;

   lambda = length / ( float ) ntimes;
   dir   = lambda*r;
   width = dir*0.6;
   dir   = dir*0.7;
   r     = r - dir;
/**
        printf(" ntimes = %d, div = %15.6f  length = %15.6f\n", 
	         ntimes, div, length );

                                                                            /*
   +-------------------------------------------------------------------------+
   |(2) If not, draw wave arc explicitely.                                   |
   +-------------------------------------------------------------------------+
                                                                            */
       xx0    = ( float ) x0,  yy0 = ( float ) y0;
       xt    = 0.0;
                                                                             /*
       +----------------------------------------------------------------------+
       |(a) Draw arc of waves by ntimes half_waves                            |
       +----------------------------------------------------------------------+
                                                                             */
       xt =  - 0.5*lambda;
       th = theta0 + xt;
       cost = cos( th );
       sint = sin( th );
       x3 =  r * cost + xc;
       y3 =  r * sint + yc;
       xx0 = (float) x3, yy0 = (float) y3;

       if( print_ps == 1 ) 
         {
           pstrans( xx0, yy0 );
           xx1 = xxx, yy1 = yyy;
               fprintf( psout, " %8.2f %8.2f moveto\n", xx1, yy1 );
          }

       for( n0 = 0; n0 < ntimes; n0++ )
          {
/**
            printf("%3d th = %8.2f\n", n0, th*rdeg );
/**/
            xt = xt + 0.4*lambda;
            th = theta0 + xt;
            cost = cos( th );
            sint = sin( th );
            r1   = r + dir;
            x3 =  r1 * cost + xc;
            y3 =  r1 * sint + yc;
            xx1 = (float) x3, yy1 = (float) y3;

            xt = xt + 0.2*lambda;
            th = theta0 + xt;
            cost = cos( th );
            sint = sin( th );
            x3 =  r1 * cost + xc;
            y3 =  r1 * sint + yc;
            xx2 = (float) x3, yy2 = (float) y3;

            xt = xt + 0.4*lambda;
            th = theta0 + xt;
            cost = cos( th );
            sint = sin( th );
            x3 =  r * cost + xc;
            y3 =  r * sint + yc;
            xx3 = (float) x3, yy3 = (float) y3;
/*
            printf("%3d ( %8.2f, %8.2f) ( %8.2f, %8.2f ) ( %8.2f, %8.2f )\n", 
                   n0, xx0 , yy0, xx1, yy1, xx2, yy2 );
                                                                             /*
           +------------------------------------------------------------------+
           |(b') Output the line to ps file                                   |
           +------------------------------------------------------------------+
                                                                             */
           if( print_ps == 0 )
             {
               urcurveto( xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3 );
               xx0 = xx3,  yy0 = yy3;
              }
           else 
             {
               xx0 = xx3,  yy0 = yy3;
               pstrans( xx1, yy1 );
               fprintf( psout, " %8.2f %8.2f", xxx, yyy );
               pstrans( xx2, yy2 );
               fprintf( psout, " %8.2f %8.2f", xxx, yyy );
               pstrans( xx3, yy3 );
               fprintf( psout, " %8.2f %8.2f  curveto\n", xxx, yyy );
              }

            dir = -dir;

          }
       if( print_ps == 1 ) 
         {
               fprintf( psout, "stroke\n");
          }

   return width;
}

                                                                           /*
*****************************************************************************
*                       ======================                             */
                         void pswave_def(   )
/*                      ======================
*****************************************************************************
* ( Function )                                                              *
*   To draw a wave on a line from ( xx1, yy1 ) to ( xx3, yy3 ),             *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
{
 fprintf( psout,"\n/wavetodict  16 dict def\n");
 fprintf( psout," wavetodict /mtrx matrix put\n\n");

 fprintf( psout,"%%-->  x  y  n  waveto -\n\n");

 fprintf( psout,"/waveto\n");
 fprintf( psout,"{ wavetodict begin\n  /n exch def\n");
 fprintf( psout,"  /y exch def\n  /x exch def\n\n");

 fprintf( psout,"  currentpoint\n");
 fprintf( psout,"  /cy exch def\n  /cx exch def\n");
 fprintf( psout,"  /dx x cx sub def\n  /dy y cy sub def\n");
 fprintf( psout,"  /th dy dx atan def\n");
 fprintf( psout,"  /savematrix mtrx currentmatrix def\n\n");

 fprintf( psout,"  /r dx dx mul dy dy mul add sqrt n 1 add div def\n\n");

 fprintf( psout,"  cx cy translate\n  r r scale\n  th rotate\n\n");

 fprintf( psout,"  /lx  0.5 def\n  /dir 0.7 def\n   lx  0   lineto\n\n");

 fprintf( psout,"  n {\n");
 fprintf( psout,"      0.4 lx add dir 0.6 lx add");
 fprintf( psout," dir 1.0 lx add 0.0 curveto\n");
 fprintf( psout,"      /lx lx 1 add def\n      /dir dir neg def\n");
 fprintf( psout,"    } repeat\n\n");
 fprintf( psout,"  /lx  lx  0.5  add def\n   lx  0  lineto\n\n");

 fprintf( psout,"  savematrix setmatrix\n  end\n} def\n");
}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
             void pswvarc_def(   )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a wave on a line from ( xx1, yy1 ) to ( xx3, yy3 ),             *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
{
 fprintf( psout,"\n/wvarcdict  16 dict def\n");
 fprintf( psout," wvarcdict /mtrx matrix put\n\n");

 fprintf( psout,"%%-->  x  y   r  ang1  ang2  n  wvarc -\n\n");

 fprintf( psout,"/wvarc\n");
 fprintf( psout,"{ wvarcdict begin\n  /n  exch def\n");
 fprintf( psout,"  /ae exch def\n  /as exch def\n  /r  exch def\n");
 fprintf( psout,"  /y  exch def\n  /x  exch def\n\n");

 fprintf( psout,"  /thf ae as sub n 1 add div def\n");
 fprintf( psout,"  /lt  thf abs 3.1415 180 div mul r mul def\n\n");

 fprintf( psout,"  /savematrix mtrx currentmatrix def\n\n");

 fprintf( psout,"  x y translate\n\n");

 fprintf( psout,"  /x0 as cos r mul def\n  /y0 as sin r mul def\n");
 fprintf( psout,"  /th0 as thf 0.5 mul add def\n");
 fprintf( psout,"  /x1 th0 cos r mul def\n  /y1 th0 sin r mul def\n");
 fprintf( psout,"   x0 y0 moveto  x1 y1 lineto\n\n  /th th0 0 add def\n");
 fprintf( psout,"  /dir lt 0.7 mul def\n\n");

 fprintf( psout,"  n {\n");
 fprintf( psout,"      /th th 0.4 thf mul add def\n");
 fprintf( psout,"      /x1 th cos r dir add mul def\n");
 fprintf( psout,"      /y1 th sin r dir add mul def\n");
 fprintf( psout,"      /th th 0.2 thf mul add def\n");
 fprintf( psout,"      /x2 th cos r dir add mul def\n");
 fprintf( psout,"      /y2 th sin r dir add mul def\n");
 fprintf( psout,"      /th th 0.4 thf mul add def\n");
 fprintf( psout,"      /x3 th cos r mul def\n");
 fprintf( psout,"      /y3 th sin r mul def\n");
 fprintf( psout,"       x1  y1  x2  y2  x3  y3 curveto\n");
 fprintf( psout,"      /dir dir neg def\n");
 fprintf( psout,"    } repeat\n\n");
 fprintf( psout,"  /th  th thf 0.5 mul add def\n");
 fprintf( psout,"  /x1 th cos r mul def\n");
 fprintf( psout,"  /y1 th sin r mul def\n");
 fprintf( psout,"   x1  y1 lineto\n");

 fprintf( psout,"  savematrix setmatrix\n  end\n} def\n");
}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
            float urspiral(  xx1, yy1, xx2, yy2, xx3, yy3 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a spiral on a line from ( xx1, yy1 ) to ( xx3, yy3 ),           *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2, xx3, yy3;
{                                                                          /*
 ------------------
  Local Variables
 ------------------
                                                                           */
 float width;
                                                                           /*
 -------------
  Functions
 ------------
                                                                           */
 void pstrans(  );
 float urspto( );
 float ursparc( );
                                                                           /*
============================ Entry Point ===================================
                                                                           */
if( print_ps > 0 )
  {
    pstrans( xx1, yy1 );
    xx1  =  xxx,  yy1  = yyy;
    pstrans( xx3, yy3 );
    xx3  =  xxx,  yy3  = yyy;
   }
                                                                            /*
=============================================================================
    Draw a striaght line by spirals
=============================================================================
                                                                            */
if( xx2 == 0.0 )
  {
    width = urspto( xx1, yy1, xx3, yy3 );
   }
                                                                            /*
=============================================================================
    Draw an arc by spirals
=============================================================================
                                                                            */
else
  {
    if( print_ps > 0 )
      {
        pstrans( xx2, yy2 );
        xx2  =  xxx,  yy2  = yyy;
       }

     width = ursparc( xx1, yy1, xx2, yy2, xx3, yy3 );

   }

   return width;
}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
                 float urspto(  xx1, yy1, xx3, yy3 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw an straight line in terms of spirals                            *
*     from ( xx1, yy1 ) to ( xx3, yy3 )                                     *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx3, yy3;
{                                                                          /*
 -----------------
  Local Variables
 -----------------
                                                                           */
   double x0, y0, x1, y1, x2, y2, x3, y3, length, r, div; 
   double xc, yc, a0, b0, a1, b1, r1, rr, theta0, theta1, th;
   double xcrcl[17], ycrcl[17], xe[17], ye[17]; 
   double temp, lambda, a, fact;
   double xt,  yt, xt1, yt1;
   double xcbase, xebase, theta, sint, cost;
   double delta, rt;
   int ntimes, m, n, mo;
   float width, widthp, widthm, xx0, yy0, xx2, yy2, rpi;
   int ix1, iy1, ix2, iy2;
   static int mintime1 = 4, mintime2 = 6;
                                                                           /*
 -------------
   Functions
 -------------
                                                                           */
   void urcurveto( );
   void urLine( );
                                                                           /*
============================ Entry Point ====================================
                                                                           */
   x0 = (double) xx1;
   y0 = (double) yy1;
   x1 = (double) xx3;
   y1 = (double) yy3;
                                                                            /*
=============================================================================
    Draw a straight line by spirals
=============================================================================
                                                                            */
                                                                            /*
  +--------------------------------------------------------------------------+
  |(1) calculate the length of line and number of siprals                    |
  +--------------------------------------------------------------------------+
                                                                            */
   length = ( x1 - x0 )*( x1 - x0 ) + ( y1 - y0 )*( y1 - y0 );
   length = sqrt( length );
   div    = length / spiral_length;
   ntimes = (int) div;
   if( ntimes < mintime1 )
       ntimes = mintime1;
   r      = length / ( 3.0* (float) ntimes /4.0 + 1.25 );
   width  = 0.5*r;
   widthp = width;
   widthm = 0.4*r;
   a      = 0.05*r;
   if( print_ps == 0 )
     {
       widthp = -widthp;
       widthm = -widthm;
       a      = -a;
      }
                                                                            /*
   +-------------------------------------------------------------------------+
   |(2) If flag_spir > 0 and print_ps == 1, print ps_command                 |
   +-------------------------------------------------------------------------+
                                                                            */
   if( print_ps == 1 && flag_spir > 0 )
     {
       fprintf( psout,"gsave\nnewpath\n");
       fprintf( psout," %8.2f %8.2f  moveto\n",    xx1, yy1 );
       fprintf( psout," %8.2f %8.2f %2d spto\n", xx3, yy3, ntimes );
       fprintf( psout,"stroke\ngrestore\n");
      }
                                                                            /*
   +-------------------------------------------------------------------------+
   |(3) If not, draw a spiral explicitely.                                   |
   +-------------------------------------------------------------------------+
                                                                            */
   else
     {
                                                                            /*
      +----------------------------------------------------------------------+
      |(a) calculate the transformation matrix                               |
      +----------------------------------------------------------------------+
                                                                            */
       theta = atan2( y1-y0, x1-x0 );
       sint  = sin( theta );
       cost  = cos( theta );
                                                                            /*
      +----------------------------------------------------------------------+
      |(b) Draw a striaght line                                              |
      +----------------------------------------------------------------------+
                                                                            */
       xx0   = (float) x0,  yy0 = (float) y0;
       xt    = 0.5*r;
    
       xt1 =  xt*cost + x0;
       yt1 =  xt*sint + y0;
       xx1 = (float) xt1,  yy1 = (float) yt1;

       if( print_ps == 0 )
         {
           urLine( xx0, yy0, xx1,  yy1 );
          }
       else
         {
           fprintf( psout, "newpath\n");
           fprintf( psout, " %8.2f %8.2f moveto", xx0, yy0 );
           fprintf( psout, "  %8.2f %8.2f   lineto\n", xx1, yy1 );
          }
       xx0 = xx1, yy0 = yy1; 
                                                                            /*
      +----------------------------------------------------------------------+
      |(c) Draw a striaght line by ntimes spirals                            |
      +----------------------------------------------------------------------+
                                                                            */
       for( m = 0; m < ntimes; m++ )
          {                                                                 /*
           +-----------------------------------------------------------------+
           |(i) Draw straight line                                           |
           +-----------------------------------------------------------------+
                                                                            */
            yt  = a;
            xt1 =  xt*cost - yt*sint + x0;
            yt1 =  xt*sint + yt*cost + y0;
            xx1 = (float) xt1,  yy1 = (float) yt1;
            if( print_ps == 0 )
              {
                urLine( xx0, yy0, xx1,  yy1 );
               }
            else
              {
                fprintf( psout, "  %8.2f %8.2f   lineto\n", xx1, yy1 );
               }
            xx0 = xx1, yy0 = yy1; 
                                                                            /*
           +-----------------------------------------------------------------+
           |(ii) Draw Bezier curve by curveto                                |
           +-----------------------------------------------------------------+
                                                                            */
            xt  = xt + 0.1*r;
            yt  = widthp;
            xt1 =  xt*cost - yt*sint + x0;
            yt1 =  xt*sint + yt*cost + y0;
            xx1 = (float) xt1,  yy1 = (float) yt1;
            xt  = xt + 0.8*r;
            xt1 =  xt*cost - yt*sint + x0;
            yt1 =  xt*sint + yt*cost + y0;
            xx2 = (float) xt1,  yy2 = (float) yt1;
            xt  = xt + 0.1*r;
            yt  = a;
            xt1 =  xt*cost - yt*sint + x0;
            yt1 =  xt*sint + yt*cost + y0;
            xx3 = (float) xt1,  yy3 = (float) yt1;

            if( print_ps == 0 )
              {
                urcurveto( xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3 );
               }
            else
              {
                fprintf( psout, "  %8.2f %8.2f ", xx0, yy0 );
                fprintf( psout, "  %8.2f %8.2f ", xx1, yy1 );
                fprintf( psout, "  %8.2f %8.2f ", xx2, yy2 );
                fprintf( psout, "  %8.2f %8.2f curveto\n", xx3, yy3 );
               }
                                                                            /*
           +-----------------------------------------------------------------+
           |(iii) Draw straight line                                         |
           +-----------------------------------------------------------------+
                                                                            */
            xx0 = xx3, yy0 = yy3; 
            if( m == ntimes-1 ) yt = 0.0;
            else                yt  = -a;
            xt1 =  xt*cost - yt*sint + x0;
            yt1 =  xt*sint + yt*cost + y0;
            xx1 = (float) xt1,  yy1 = (float) yt1;
            if( print_ps == 0 )
              {
                urLine( xx0, yy0, xx1,  yy1 );
               }
            else
              {
                fprintf( psout, "  %8.2f %8.2f   lineto\n", xx1, yy1 );
               }
            xx0 = xx1, yy0 = yy1; 
                                                                            /*
           +-----------------------------------------------------------------+
           |(iv) Draw Bezier curve by curveto  or a straight line            |
           +-----------------------------------------------------------------+
                                                                            */
            if( m < ntimes-1 )
              {                                                             /*
               +-------------------------------------------------------------+
               | 1. Draw Bezier curve by curveto                             |
               +-------------------------------------------------------------+
                                                                            */
                xt  = xt - 0.025*r;
                yt  = -widthm;
                xt1 =  xt*cost - yt*sint + x0;
                yt1 =  xt*sint + yt*cost + y0;
                xx1 = (float) xt1,  yy1 = (float) yt1;
                xt  = xt - 0.2*r;
                xt1 =  xt*cost - yt*sint + x0;
                yt1 =  xt*sint + yt*cost + y0;
                xx2 = (float) xt1,  yy2 = (float) yt1;
                xt  = xt - 0.025*r;
                yt  = -a;
                xt1 =  xt*cost - yt*sint + x0;
                yt1 =  xt*sint + yt*cost + y0;
                xx3 = (float) xt1,  yy3 = (float) yt1;

                if( print_ps == 0 )
                  {
                    urcurveto( xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3 );
                   }
                else
                  {
                    fprintf( psout, "  %8.2f %8.2f ", xx0, yy0 );
                    fprintf( psout, "  %8.2f %8.2f ", xx1, yy1 );
                    fprintf( psout, "  %8.2f %8.2f ", xx2, yy2 );
                    fprintf( psout, "  %8.2f %8.2f curveto\n", xx3, yy3 );
                   }

                xx0 = xx3, yy0 = yy3; 
               }
            else
              {                                                             /*
               +-------------------------------------------------------------+
               | 2. Draw straight line                                       |
               +-------------------------------------------------------------+
                                                                            */
                xt  = xt + 0.5*r;
                yt  =  0.0;
                xt1 =  xt*cost - yt*sint + x0;
                yt1 =  xt*sint + yt*cost + y0;
                xx1 = (float) xt1,  yy1 = (float) yt1;
                if( print_ps == 0 )
                  {
                    urLine( xx0, yy0, xx1,  yy1 );
                   }
                else
                  {
                    fprintf( psout, "  %8.2f %8.2f   lineto\n", xx1, yy1 );
                    fprintf( psout, "stroke\n");
                   }
               }
           }
      }

 return width;

}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
            float ursparc(  xx1, yy1, xx2, yy2, xx3, yy3 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw an straight line in terms of spirals                            *
*     from ( xx1, yy1 ) to ( xx3, yy3 )                                     *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2, xx3, yy3;
{                                                                          /*
 -----------------
  Local Variables
 -----------------
                                                                           */
   double x0, y0, x1, y1, x2, y2, x3, y3, length, r, div; 
   double xc, yc, a0, b0, a1, b1, r1, rr, theta0, theta1, th;
   double xcrcl[17], ycrcl[17], xe[17], ye[17]; 
   double temp, lambda, a, fact;
   double xt,  yt, xt1, yt1;
   double xcbase, xebase, theta, sint, cost;
   double delta, rt;
   int ntimes, m, n, mo;
   float width, widthm, xx0, yy0, rpi;
   int ix1, iy1, ix2, iy2;
   static int mintime1 = 4, mintime2 = 6;
                                                                           /*
 -------------
   Functions
 -------------
                                                                           */
   void urcurveto( );
   void urLine( );
   void get_radius();
                                                                           /*
============================ Entry Point ====================================
                                                                           */
   x0 = (double) xx1;
   y0 = (double) yy1;
   x1 = (double) xx3;
   y1 = (double) yy3;
                                                                            /*
=============================================================================
    Draw an arc by spirals
=============================================================================
                                                                            */
                                                                            /*
  +--------------------------------------------------------------------------+
  |(1) calculate the length of line, number of siprals and radius of arc     |
  +--------------------------------------------------------------------------+
                                                                            */
   x2 = (double) xx2;
   y2 = (double) yy2;

   get_radius( x0, y0, x1, y1, x2, y2,
               &xc, &yc,  &r1, &theta, &theta0, &theta1 );

   if( theta > 0.0 )
       length = theta;
   else
       length = -theta;

   div    = length*r1 / spiral_length;
   ntimes = (int) div;
   if( ntimes < mintime2 )
       ntimes = mintime2;
   r       = length / ( 3.0* (float) ntimes /4.0 + 1.25 );
   a       = r*r1;
   width   = a*0.5;
                                                                            /*
   +-------------------------------------------------------------------------+
   |(2) If flag_sparc > 0 and print_ps == 1, print ps_command                |
   +-------------------------------------------------------------------------+
                                                                            */
   if( print_ps == 1 && flag_sparc > 0 )
     {
       rpi = 180.0 / pi;
       fprintf( psout,"gsave\nnewpath\n");
       fprintf( psout," %8.2f %8.2f  moveto\n",    xx1, yy1 );
       fprintf( psout," %8.2f %8.2f %8.2f", 
               (float) xc, (float) yc, (float) r1 );
       fprintf( psout," %8.2f %8.2f %2d sparc\n", 
                rpi* (float) theta0, rpi* (float) theta1, ntimes );
       fprintf( psout,"stroke\ngrestore\n");
      }
                                                                            /*
   +-------------------------------------------------------------------------+
   |(3) If not, draw spiral arc explicitely.                                 |
   +-------------------------------------------------------------------------+
                                                                            */
   else
     {
       widthm  = a*0.4;
       a       = a*0.05;
                                                                             /*
       +----------------------------------------------------------------------+
       |(a) Draw a line of length half of half_wave length                    |
       +----------------------------------------------------------------------+
                                                                             */
       xx0    = ( float ) x0,  yy0 = ( float ) y0;
       xt    = 0.5*r;
       if( theta > 0.0 ) th = theta0 + xt;
       else              th = theta0 - xt;
       cost = cos( th );
       sint = sin( th );

       x3 =  r1 * cost + xc;
       y3 =  r1 * sint + yc;
       xx1    = (float) x3,  yy1 = (float) y3;

       if( print_ps == 0 )
         {
           urLine( xx0, yy0, xx1, yy1 );
          }
       else
         {
           fprintf( psout, "newpath\n");
           fprintf( psout, " %8.2f %8.2f moveto", xx0, yy0 );
           fprintf( psout, " %8.2f %8.2f lineto\n", xx1, yy1 );
          }

       xx0 = xx1, yy0 = yy1;
                                                                            /*
      +----------------------------------------------------------------------+
      |(b) Draw an arc by ntimes spirals                                     |
      +----------------------------------------------------------------------+
                                                                            */
       for( m = 0; m < ntimes; m++ )
          {                                                                 /*
           +-----------------------------------------------------------------+
           |(i) Draw straight line                                           |
           +-----------------------------------------------------------------+
                                                                            */
            rt  = r1 + a;
            x3  =  rt * cost + xc;
            y3  =  rt * sint + yc;
            xx1    = (float) x3,  yy1 = (float) y3;
            if( print_ps == 0 )
              {
                urLine( xx0, yy0, xx1, yy1 );
               }
            else
              {
                fprintf( psout, "  %8.2f %8.2f   lineto\n", xx1, yy1 );
               }
            xx0 = xx1, yy0 = yy1; 
                                                                            /*
           +-----------------------------------------------------------------+
           |(ii) Draw Bezier curve by curveto                                |
           +-----------------------------------------------------------------+
                                                                            */
            xt  = xt + 0.1*r;
            if( theta > 0.0 ) th = theta0 + xt;
            else              th = theta0 - xt;
            cost = cos( th );
            sint = sin( th );
            rt  = r1 + width;
            x3  =  rt * cost + xc;
            y3  =  rt * sint + yc;
            xx1    = (float) x3,  yy1 = (float) y3;

            xt  = xt + 0.8*r;
            if( theta > 0.0 ) th = theta0 + xt;
            else              th = theta0 - xt;
            cost = cos( th );
            sint = sin( th );
            rt  = r1 + width;
            x3  =  rt * cost + xc;
            y3  =  rt * sint + yc;
            xx2    = (float) x3,  yy2 = (float) y3;

            xt  = xt + 0.1*r;
            if( theta > 0.0 ) th = theta0 + xt;
            else              th = theta0 - xt;
            cost = cos( th );
            sint = sin( th );
            rt  = r1 + a;
            x3  =  rt * cost + xc;
            y3  =  rt * sint + yc;
            xx3    = (float) x3,  yy3 = (float) y3;

            if( print_ps == 0 )
              {
                urcurveto( xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3 );
               }
            else
              {
                fprintf( psout, "  %8.2f %8.2f ", xx0, yy0 );
                fprintf( psout, "  %8.2f %8.2f ", xx1, yy1 );
                fprintf( psout, "  %8.2f %8.2f ", xx2, yy2 );
                fprintf( psout, "  %8.2f %8.2f curveto\n", xx3, yy3 );
               }

                                                                            /*
           +-----------------------------------------------------------------+
           |(iii) Draw straight line                                         |
           +-----------------------------------------------------------------+
                                                                            */
            xx0 = xx3, yy0 = yy3; 
            if( m == ntimes-1 ) rt = r1;
            else                rt = r1 - a;
            x3  =  rt * cost + xc;
            y3  =  rt * sint + yc;
            xx1    = (float) x3,  yy1 = (float) y3;
            if( print_ps == 0 )
              {
                urLine( xx0, yy0, xx1,  yy1 );
               }
            else
              {
                fprintf( psout, "  %8.2f %8.2f   lineto\n", xx1, yy1 );
               }
            xx0 = xx1, yy0 = yy1; 
                                                                            /*
           +-----------------------------------------------------------------+
           |(iv) Draw Bezier curve by curveto  or a straight line            |
           +-----------------------------------------------------------------+
                                                                            */
            if( m < ntimes-1 )
              {                                                             /*
               +-------------------------------------------------------------+
               | 1. Draw Bezier curve by curveto                             |
               +-------------------------------------------------------------+
                                                                            */
                xt  = xt - 0.025*r;
                if( theta > 0.0 ) th = theta0 + xt;
                else              th = theta0 - xt;
                cost = cos( th );
                sint = sin( th );
                rt  = r1 - widthm;
                x3  =  rt * cost + xc;
                y3  =  rt * sint + yc;
                xx1    = (float) x3,  yy1 = (float) y3;
                xt  = xt - 0.2*r;
                if( theta > 0.0 ) th = theta0 + xt;
                else              th = theta0 - xt;
                cost = cos( th );
                sint = sin( th );
                rt  = r1 - widthm;
                x3  =  rt * cost + xc;
                y3  =  rt * sint + yc;
                xx2    = (float) x3,  yy2 = (float) y3;
                xt  = xt - 0.025*r;
                if( theta > 0.0 ) th = theta0 + xt;
                else              th = theta0 - xt;
                cost = cos( th );
                sint = sin( th );
                rt  = r1 - a;
                x3  =  rt * cost + xc;
                y3  =  rt * sint + yc;
                xx3    = (float) x3,  yy3 = (float) y3;

                if( print_ps == 0 )
                  {
                    urcurveto( xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3 );
                   }
                else
                  {
                    fprintf( psout, "  %8.2f %8.2f ", xx0, yy0 );
                    fprintf( psout, "  %8.2f %8.2f ", xx1, yy1 );
                    fprintf( psout, "  %8.2f %8.2f ", xx2, yy2 );
                    fprintf( psout, "  %8.2f %8.2f curveto\n", xx3, yy3 );
                   }

                xx0 = xx3, yy0 = yy3; 
               }
            else
              {                                                             /*
               +-------------------------------------------------------------+
               | 2. Draw straight line                                       |
               +-------------------------------------------------------------+
                                                                            */
                xt  = xt + 0.5*r;
                if( theta > 0.0 ) th = theta0 + xt;
                else              th = theta0 - xt;
                cost = cos( th );
                sint = sin( th );
                rt  = r1;
                x3  =  rt * cost + xc;
                y3  =  rt * sint + yc;
                xx1    = (float) x3,  yy1 = (float) y3;
                if( print_ps == 0 )
                  {
                    urLine( xx0, yy0, xx1,  yy1 );
                   }
                else
                  {
                    fprintf( psout, "  %8.2f %8.2f   lineto\n", xx1, yy1 );
                    fprintf( psout, "stroke\n");
                   }
               }
           }
      }

   return width;
}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
            float urspcircl(  xx1, yy1, xx2, yy2 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw an straight line in terms of spirals                            *
*     from ( xx1, yy1 ) to ( xx3, yy3 )                                     *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2;
{                                                                          /*
 -----------------
  Local Variables
 -----------------
                                                                           */
   double x0, y0, x1, y1, x2, y2, x3, y3, length, r, div; 
   double xc, yc, a0, b0, a1, b1, r1, rr, theta0, theta1, th;
   double xcrcl[17], ycrcl[17], xe[17], ye[17]; 
   double temp, lambda, a, fact;
   double xt,  yt, xt1, yt1;
   double xcbase, xebase, theta, sint, cost;
   double delta, rt;
   int ntimes, m, n, mo;
   float width, widthm, xx0, yy0, xx3, yy3, rpi;
   int ix1, iy1, ix2, iy2;
   static int mintime1 = 4, mintime2 = 20;

   int c;
                                                                           /*
 -------------
   Functions
 -------------
                                                                           */
   void urcurveto( );
   void urLine( );
                                                                           /*
============================ Entry Point ====================================
                                                                           */
   x0 = (double) xx1;
   y0 = (double) yy1;
   x1 = (double) xx2;
   y1 = (double) yy2;
/**
   printf(" (xx1, yy1) = ( %8.2f, %8.2f ) \n", xx1, yy1 );
   printf(" (xx2, yy2) = ( %8.2f, %8.2f ) \n", xx2, yy2 );
                                                                            /*
  +--------------------------------------------------------------------------+
  |(1) calculate the length of line, number of siprals and radius of arc     |
  +--------------------------------------------------------------------------+
                                                                            */
   xc   = 0.5*(x0 + x1 ),  yc  = 0.5*(y0 + y1);
   r1   = (x0 - xc)*(x0 - xc) + (y0 - yc)*(y0 - yc);
   r1   = sqrt( r1 );
/*
   if( x0 > x1 - 1.0 && x0 < x1 + 1.0 )
     {
       if( y0 < y1 ) theta0 = -0.5*pi;
       else          theta0 =  0.5*pi;
      }
   else
*/
       theta0 = atan2( y0 - yc, x0 - xc );
/**
   printf(" ( xc, yc ) = ( %8.2f, %8.2f ), r1 = %8.2f\n", xc, yc, r1 );
   printf(" theta0 = %8.2f\n", theta0 );
/***/

   length = pi2;

   div    = length*r1 / spiral_length;
   ntimes = (int) div;
   if( ntimes < mintime2 )
       ntimes = mintime2;
   r       = length / ( 3.0* (float) ntimes /4.0 + 1.25 );
   r       = length / (float) ntimes ;
   r       = 4.0*r / 3.0;
   a       = r*r1;
   width   = a*0.5;
/**
        printf(" ntimes = %d, div = %15.6f  length = %15.6f\n",
                 ntimes, div, length );
                                                                            /*
   +-------------------------------------------------------------------------+
   |(3) If not, draw spiral arc explicitely.                                 |
   +-------------------------------------------------------------------------+
                                                                            */
       widthm  = a*0.4;
       a       = a*0.05;
       r1      = r1 - widthm;

       xt =  - 0.5*r;
       th = theta0 + xt;
       cost = cos( th );
       sint = sin( th );
       x3 =  r1 * cost + xc;
       y3 =  r1 * sint + yc;
       xx0 = (float) x3, yy0 = (float) y3;

       if( print_ps == 1 ) 
         {
           pstrans( xx0, yy0 );
           fprintf( psout, " %8.2f %8.2f moveto\n", xxx, yyy );
          }

                                                                            /*
      +----------------------------------------------------------------------+
      |(b) Draw an arc by ntimes spirals                                     |
      +----------------------------------------------------------------------+
                                                                            */
       for( m = 0; m < ntimes; m++ )
          {                                                                 /*
           +-----------------------------------------------------------------+
           |(i) Draw straight line                                           |
           +-----------------------------------------------------------------+
                                                                            */
            rt  = r1 + a;
            x3  =  rt * cost + xc;
            y3  =  rt * sint + yc;
            xx1    = (float) x3,  yy1 = (float) y3;
            if( print_ps == 0 )
              {
                urLine( xx0, yy0, xx1, yy1 );
               }
            else
              {
                pstrans( xx1, yy1 );
                fprintf( psout, "  %8.2f %8.2f   lineto\n", xxx, yyy );
               }
            xx0 = xx1, yy0 = yy1; 
                                                                            /*
           +-----------------------------------------------------------------+
           |(ii) Draw Bezier curve by curveto                                |
           +-----------------------------------------------------------------+
                                                                            */
            xt  = xt + 0.1*r;
            th = theta0 + xt;
/**
            printf(" m =%3d  th = %8.2f \n", m, th*rdeg );
/**/
            cost = cos( th );
            sint = sin( th );
            rt  = r1 + width;
            x3  =  rt * cost + xc;
            y3  =  rt * sint + yc;
            xx1    = (float) x3,  yy1 = (float) y3;

            xt  = xt + 0.8*r;
            th = theta0 + xt;
            cost = cos( th );
            sint = sin( th );
            rt  = r1 + width;
            x3  =  rt * cost + xc;
            y3  =  rt * sint + yc;
            xx2    = (float) x3,  yy2 = (float) y3;

            xt  = xt + 0.1*r;
            th = theta0 + xt;
            cost = cos( th );
            sint = sin( th );
            rt  = r1 + a;
            x3  =  rt * cost + xc;
            y3  =  rt * sint + yc;
            xx3    = (float) x3,  yy3 = (float) y3;
/**
            printf(" (xx0, yy0 ) = (%8.2f, %8.2f) (xx1, yy1 ) = (%8.2f, %8.2f)\n",
                      xx0, yy0, xx1, yy1 );
            printf(" (xx2, yy2 ) = (%8.2f, %8.2f) (xx3, yy3 ) = (%8.2f, %8.2f)\n",
                      xx2, yy2, xx3, yy3 );
/**/

            if( print_ps == 0 )
              {
                urcurveto( xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3 );
               }
            else
              {
                pstrans( xx0, yy0 );
                fprintf( psout, " %8.2f %8.2f", xxx, yyy );
                pstrans( xx1, yy1 );
                fprintf( psout, " %8.2f %8.2f", xxx, yyy );
                pstrans( xx2, yy2 );
                fprintf( psout, " %8.2f %8.2f", xxx, yyy );
                pstrans( xx3, yy3 );
                fprintf( psout, " %8.2f %8.2f  curveto\n", xxx, yyy );
               }

                                                                            /*
           +-----------------------------------------------------------------+
           |(iii) Draw straight line                                         |
           +-----------------------------------------------------------------+
                                                                            */
            xx0 = xx3, yy0 = yy3; 
            if( m == ntimes-1 ) rt = r1;
            else                rt = r1 - a;
            x3  =  rt * cost + xc;
            y3  =  rt * sint + yc;
            xx1    = (float) x3,  yy1 = (float) y3;
            if( print_ps == 0 )
              {
                urLine( xx0, yy0, xx1,  yy1 );
               }
            else
              {
                pstrans( xx1, yy1 );
                fprintf( psout, "  %8.2f %8.2f   lineto\n", xxx, yyy );
               }
            xx0 = xx1, yy0 = yy1; 
                                                                            /*
           +-----------------------------------------------------------------+
           |(iv) Draw Bezier curve by curveto  or a straight line            |
           +-----------------------------------------------------------------+
                                                                            /*
               +-------------------------------------------------------------+
               | 1. Draw Bezier curve by curveto                             |
               +-------------------------------------------------------------+
                                                                            */
                xt  = xt - 0.025*r;
                th = theta0 + xt;
                cost = cos( th );
                sint = sin( th );
                rt  = r1 - widthm;
                x3  =  rt * cost + xc;
                y3  =  rt * sint + yc;
                xx1    = (float) x3,  yy1 = (float) y3;

                xt  = xt - 0.2*r;
                th = theta0 + xt;
                cost = cos( th );
                sint = sin( th );
                rt  = r1 - widthm;
                x3  =  rt * cost + xc;
                y3  =  rt * sint + yc;
                xx2    = (float) x3,  yy2 = (float) y3;

                xt  = xt - 0.025*r;
                th = theta0 + xt;
                cost = cos( th );
                sint = sin( th );
                rt  = r1 - a;
                x3  =  rt * cost + xc;
                y3  =  rt * sint + yc;
                xx3    = (float) x3,  yy3 = (float) y3;

                if( print_ps == 0 )
                  {
                    urcurveto( xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3 );
                   }
                else
                  {
                    pstrans( xx0, yy0 );
                    fprintf( psout, " %8.2f %8.2f", xxx, yyy );
                    pstrans( xx1, yy1 );
                    fprintf( psout, " %8.2f %8.2f", xxx, yyy );
                    pstrans( xx2, yy2 );
                    fprintf( psout, " %8.2f %8.2f", xxx, yyy );
                    pstrans( xx3, yy3 );
                    fprintf( psout, " %8.2f %8.2f  curveto\n", xxx, yyy );
                   }

                xx0 = xx3, yy0 = yy3; 
           }
                                                                            /*
           +-----------------------------------------------------------------+
           |(v) Draw straight line                                           |
           +-----------------------------------------------------------------+
                                                                            */
            rt  = r1 + a;
            x3  =  rt * cost + xc;
            y3  =  rt * sint + yc;
            xx1    = (float) x3,  yy1 = (float) y3;
            if( print_ps == 0 )
              {
                urLine( xx0, yy0, xx1, yy1 );
               }
            else
              {
                pstrans( xx1, yy1 );
                fprintf( psout, "  %8.2f %8.2f   lineto\n", xxx, yyy );
               }
       if( print_ps == 1 ) 
         {
               fprintf( psout, "stroke\n" );
          }
   return width;
}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
             void psspiral_def(   )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a wave on a line from ( xx1, yy1 ) to ( xx3, yy3 ),             *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
{
 fprintf( psout,"\n/sptodict  16 dict def\n");
 fprintf( psout," sptodict /mtrx matrix put\n\n");

 fprintf( psout,"%%-->  x  y  n  waveto -\n\n");

 fprintf( psout,"/spto\n");
 fprintf( psout,"{ sptodict begin\n  /n exch def\n");
 fprintf( psout,"  /y exch def\n  /x exch def\n\n");

 fprintf( psout,"  currentpoint\n");
 fprintf( psout,"  /cy exch def\n  /cx exch def\n");
 fprintf( psout,"  /dx x cx sub def\n  /dy y cy sub def\n");
 fprintf( psout,"  /th dy dx atan def\n");
 fprintf( psout,"  /savematrix mtrx currentmatrix def\n\n");

 fprintf( psout,"  /r dx dx mul dy dy mul add sqrt");
 fprintf( psout," 3 4 div n mul 1 4 div add 1 add div def\n");
 fprintf( psout,"  /m n 1 sub def\n");

 fprintf( psout,"  cx cy translate\n  r r scale\n  th rotate\n\n");

 fprintf( psout,"  /lx  0.5 def\n  /dir 0.7 def\n   lx  0   lineto\n\n");

 fprintf( psout,"  m {\n");
 fprintf( psout,"      lx 0.1 lineto\n");
 fprintf( psout,"      0.1 lx add 0.6 0.9 lx add 0.6 1.0 lx add");
 fprintf( psout,       " 0.1 curveto\n");
 fprintf( psout,"      1.0 lx add -0.1 lineto\n");
 fprintf( psout,"      0.97 lx add -0.4 0.82 lx add -0.4 0.75 lx add");
 fprintf( psout,        " -0.1 curveto\n");
 fprintf( psout,"      /lx lx 3 4 div add def\n");
 fprintf( psout,"    } repeat\n\n");
 fprintf( psout,"  /lx  lx  0.1 lineto\n\n");
 fprintf( psout,"  0.1 lx add 0.6 0.9 lx add 0.6 1.0 lx add 0.1 curveto\n");
 fprintf( psout,"  1.0 lx add 0 lineto\n  /lx lx 1 add def\n\n");

 fprintf( psout,"  /lx lx 0.5 add def\n  lx 0 lineto\n\n");

 fprintf( psout,"  savematrix setmatrix\n  end\n} def\n");
}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
             void pssparc_def(   )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a wave on a line from ( xx1, yy1 ) to ( xx3, yy3 ),             *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
{
 fprintf( psout,"\n/sparcdict  16 dict def\n");
 fprintf( psout," sparcdict /mtrx matrix put\n\n");

 fprintf( psout,"%%-->  x  y   r  ang1  ang2  n  sparc -\n\n");

 fprintf( psout,"/sparc\n");
 fprintf( psout,"{ sparcdict begin\n  /n  exch def\n");
 fprintf( psout,"  /ae exch def\n  /as exch def\n  /r  exch def\n");
 fprintf( psout,"  /y  exch def\n  /x  exch def\n\n");

 fprintf( psout,"  /thf ae as sub 3 4 div n mul");
 fprintf( psout,"   1 4 div add 1 add div def\n");
 fprintf( psout,"  /m n 1 sub def\n");
 fprintf( psout,"  /lt  thf abs 3.1415 180 div mul r mul def\n\n");

 fprintf( psout,"  /savematrix mtrx currentmatrix def\n\n");
 fprintf( psout,"  x y translate\n\n");

 fprintf( psout,"  /x0 as cos r mul def\n  /y0 as sin r mul def\n");
 fprintf( psout,"  /th0 as thf 0.5 mul add def\n");
 fprintf( psout,"  /x1 th0 cos r mul def\n  /y1 th0 sin r mul def\n");
 fprintf( psout,"   x0 y0 moveto  x1 y1 lineto\n\n  /th th0 0 add def\n");

 fprintf( psout,"  /rdp r lt 0.1 mul add def\n");
 fprintf( psout,"  /rdm r lt 0.1 mul sub def\n");
 fprintf( psout,"  /rd  r lt 0.6 mul add def\n");
 fprintf( psout,"  /rdq r lt 0.4 mul sub def\n");
 fprintf( psout,"  /thd thf 20 div def\n  /thq thd 4 div def\n");
 fprintf( psout,"  /lx  0 def\n");

 fprintf( psout,"  m {\n");
 fprintf( psout,"      /th lx   thf mul th0 add def\n");
 fprintf( psout,"      /x0 th   cos rdp mul def\n");
 fprintf( psout,"      /y0 th   sin rdp mul def\n");
 fprintf( psout,"       x0 y0   lineto\n");
 fprintf( psout,"      /th 0.1  lx  add thf mul th0 add def\n");
 fprintf( psout,"      /x1 th   cos rd  mul def\n");
 fprintf( psout,"      /y1 th   sin rd  mul def\n");
 fprintf( psout,"      /th 0.9  lx  add thf mul th0 add def\n");
 fprintf( psout,"      /x2 th   cos rd  mul def\n");
 fprintf( psout,"      /y2 th   sin rd  mul def\n");
 fprintf( psout,"      /th 1.0  lx  add thf mul th0 add def\n");
 fprintf( psout,"      /x3 th   cos rdp mul def\n");
 fprintf( psout,"      /y3 th   sin rdp mul def\n");
 fprintf( psout,"       x1 y1   x2  y2  x3  y3  curveto\n\n");

 fprintf( psout,"      /x0 th   cos rdm mul def\n");
 fprintf( psout,"      /y0 th   sin rdm mul def\n");
 fprintf( psout,"       x0 y0   lineto\n");
 fprintf( psout,"      /th 0.97 lx  add thf mul th0 add def\n");
 fprintf( psout,"      /x1 th   cos rdq mul def\n");
 fprintf( psout,"      /y1 th   sin rdq mul def\n");
 fprintf( psout,"      /th 0.82 lx  add thf mul th0 add def\n");
 fprintf( psout,"      /x2 th   cos rdq mul def\n");
 fprintf( psout,"      /y2 th   sin rdq mul def\n");
 fprintf( psout,"      /th 0.75 lx  add thf mul th0 add def\n");
 fprintf( psout,"      /x3 th   cos rdm mul def\n");
 fprintf( psout,"      /y3 th   sin rdm mul def\n");
 fprintf( psout,"       x1 y1   x2  y2  x3  y3  curveto\n\n");
 fprintf( psout,"      /lx lx 3 4 div add def\n");
 fprintf( psout,"    } repeat\n\n");

 fprintf( psout,"  /th lx   thf mul th0 add def\n");
 fprintf( psout,"  /x0 th   cos rdp mul def\n");
 fprintf( psout,"  /y0 th   sin rdp mul def\n");
 fprintf( psout,"   x0 y0   lineto\n");
 fprintf( psout,"  /th 0.1  lx  add thf mul th0 add def\n");
 fprintf( psout,"  /x1 th   cos rd  mul def\n");
 fprintf( psout,"  /y1 th   sin rd  mul def\n");
 fprintf( psout,"  /th 0.9  lx  add thf mul th0 add def\n");
 fprintf( psout,"  /x2 th   cos rd  mul def\n");
 fprintf( psout,"  /y2 th   sin rd  mul def\n");
 fprintf( psout,"  /th 1.0  lx  add thf mul th0 add def\n");
 fprintf( psout,"  /x3 th   cos rdp mul def\n");
 fprintf( psout,"  /y3 th   sin rdp mul def\n");
 fprintf( psout,"   x1 y1   x2  y2  x3  y3  curveto\n\n");

 fprintf( psout,"  /x0 th   cos r   mul def\n");
 fprintf( psout,"  /y0 th   sin r   mul def\n");
 fprintf( psout,"   x0 y0   lineto\n\n");

 fprintf( psout,"  /x1 ae cos r mul def\n");
 fprintf( psout,"  /y1 ae sin r mul def\n");
 fprintf( psout,"   x1  y1 lineto\n\n");

 fprintf( psout,"  savematrix setmatrix\n  end\n} def\n");
}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
            void   urdott( xx1, yy1, xx2, yy2, xx3, yy3 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a dotted line from ( xx1, yy1 ) to ( xx3, yy3 ),                *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2, xx3, yy3;
{                                                                          /*
 ------------------
  Local Variables
 ------------------
                                                                           */
 float width;
                                                                           /*
 ------------
  Functions
 ------------
                                                                           */
 void urdotto( );
 void urdottarc(  );
 void pstrans(  );
                                                                           /*
================================= Entry point ===============================
                                                                           */
  if( print_ps > 0 )
    {
      pstrans( xx1, yy1 );
      xx1 = xxx, yy1 = yyy;
      pstrans( xx3, yy3 );
      xx3 = xxx, yy3 = yyy;
     }
                                                                             /*
===============================================================================
  Draw a straight line by dotted line
===============================================================================
                                                                             */
if( xx2 == 0.0 )
  {
     urdotto( xx1, yy1, xx3, yy3 );
   }
                                                                             /*
===============================================================================
  Draw dotted arc 
===============================================================================
                                                                             */
else
  {
    if( print_ps > 0 )
      {
        pstrans( xx2, yy2 );
        xx2  = xxx,  yy2  = yyy;
       }

     urdottarc( xx1, yy1, xx2, yy2, xx3, yy3 );

    }  
}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
            void   psdott_def(  )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*                                                                           *
*****************************************************************************
									   */
{
  fprintf( psout,"/dottodict  16 dict def\n dottodict /mtrx matrix put\n\n");
  fprintf( psout,"%%%%-->  x  y   n  dotto -\n\n");
  fprintf( psout," /dotto\n { dottodict begin\n");
  fprintf( psout," /n  exch def\n /y  exch def\n /x  exch def\n\n");

  fprintf( psout," currentpoint\n");
  fprintf( psout," /cy exch def\n /cx exch def\n");
  fprintf( psout," /dx x cx sub def\n /dy y cy sub def\n");
  fprintf( psout," /th dy dx atan def\n");
  fprintf( psout," /savematrix mtrx currentmatrix def\n\n");

  fprintf( psout," /r dx dx mul dy dy mul add sqrt n 0.5 add div def\n\n");

  fprintf( psout," cx cy translate\n r r scale\n th rotate\n\n");

  fprintf( psout," /lx  0.0 def\n");
  fprintf( psout," n {\n");
  fprintf( psout," 0.5 lx add 0.0 lineto\n");
  fprintf( psout," /lx lx 1 add def\n lx 0.0 moveto\n");
  fprintf( psout," } repeat\n\n");

  fprintf( psout," /lx  lx  0.5  add def\n lx  0  lineto\n\n");

  fprintf( psout," savematrix setmatrix\n");
  fprintf( psout," end\n } def\n");

}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
            void   psdtarc_def(  )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*                                                                           *
*****************************************************************************
									   */
{
  fprintf( psout," /dottarcdict  16 dict def\n");
  fprintf( psout," dottarcdict /mtrx matrix put\n\n");

  fprintf( psout,"%%%%-->  x  y   r  ang1  ang2  n  dottarc -\n\n");

  fprintf( psout," /dottarc\n { dottarcdict begin\n");
  fprintf( psout," /n  exch def\n /ae exch def\n /as exch def\n");
  fprintf( psout," /r  exch def\n /y  exch def\n /x  exch def\n\n");

  fprintf( psout," /thf ae as sub n 0.5 add div def\n");
  fprintf( psout," /lt  thf abs 3.1415 180 div mul r mul def\n\n");

  fprintf( psout," /savematrix mtrx currentmatrix def\n\n");

  fprintf( psout," x y translate\n\n /x0 as cos r mul def\n");
  fprintf( psout," /y0 as sin r mul def\n /th as def\n\n");

  fprintf( psout," n {\n");
  fprintf( psout," /th th 0.5 thf mul add def\n x0 y0 moveto\n");
  fprintf( psout," th cos r mul th sin r mul lineto\n");
  fprintf( psout," /th th 0.5 thf mul add def\n");
  fprintf( psout," /x0 th cos r mul def\n /y0 th sin r mul def\n");
  fprintf( psout," } repeat\n\n");

  fprintf( psout," x0 y0 moveto\n ae cos r mul ae sin r mul lineto\n\n");

  fprintf( psout," savematrix setmatrix\n end\n } def\n");

}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
            void   urdotto( xx1, yy1, xx3, yy3 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a dotted line from ( xx1, yy1 ) to ( xx3, yy3 ),                *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx3, yy3;
{                                                                          /*
 ----------------
  Local Variable
 ----------------
                                                                           */
   double x0, y0, x1, y1, x3, y3, xc, yc, length, div, theta;
   double sint, cost, xt,  yt;
   float xxp, yyp;
   int   ntimes, n;
   static int mintime = 4;
                                                                           /*
================================ Entry Point ===============================
                                                                            
   +-------------------------------------------------------------------------+
   |(1) Calculate the length, the number of dots                             |
   +-------------------------------------------------------------------------+
                                                                            */
   x0 = (double) xx1;
   y0 = (double) yy1;
   x1 = (double) xx3;
   y1 = (double) yy3;

  /*
   printf(" x0  = %6.1f,   x1  = %6.1f\n", x0, x1 );
   printf(" y0  = %6.1f,   y1  = %6.1f\n", y0, y1 );
  */
   length = ( x1 - x0 )*( x1 - x0 ) + ( y1 - y0 )*( y1 - y0 );
   length = sqrt( length );
   div    = length / dott_length;
   ntimes = (int) div;
   if( ntimes < mintime )
       ntimes = mintime;
                                                                            /*
   +-------------------------------------------------------------------------+
   |(2) If flag_dott > 0 print_ps == 1, print ps_command                     |
   +-------------------------------------------------------------------------+
                                                                            */
   if( print_ps == 1 && flag_dott > 0 )
     {
       fprintf( psout,"gsave\nnewpath\n");
       fprintf( psout," %8.2f %8.2f  moveto\n",    xx1, yy1 );
       fprintf( psout," %8.2f %8.2f %2d dotto\n", xx3, yy3, ntimes );
       fprintf( psout,"stroke\ngrestore\n");
      }
                                                                            /*
   +-------------------------------------------------------------------------+
   |(3) If not, draw the dotted line explicitely.                            |
   +-------------------------------------------------------------------------+
                                                                            */
   else
     {
       div    = length /( ntimes + 0.5 );
       div    = div * 0.5;
       theta = atan2( y1-y0, x1-x0 );
       sint  = sin( theta );
       cost  = cos( theta );
       xx1    = ( float ) x0,  yy1 = ( float ) y0;

       if( print_ps == 1 )
         {
           fprintf( psout, "newpath\n");
          }

       xt    = 0.0;
       yt    = 0.0;
       for( n = 0; n < ntimes; n++ )
          {
            xt =  xt + div;
            x3 =  xt*cost;
            y3 =  xt*sint;
            x3 = x3 + x0;
            y3 = y3 + y0;
            xxp = (float) x3, yyp = (float) y3;
              
            urLine( xx1, yy1, xxp, yyp );
            psLine( xx1, yy1, xxp, yyp );

            xt =  xt + div;
            x1 =  xt*cost;
            y1 =  xt*sint;
            x1 = x1 + x0;
            y1 = y1 + y0;
            xx1 = (float) x1, yy1 = (float) y1;
          }

       xx1 = (float) x1, yy1 = (float) y1;

       urLine( xx1, yy1, xx3,  yy3 );

       if( print_ps == 1 )
         {
           fprintf( psout, " %9.3f %9.3f moveto\n", xx1, yy1 );
           fprintf( psout, " %9.3f %9.3f lineto\n", xx3, yy3 );
           fprintf( psout, "stroke\n");
          }
      }
return ;
}

                                                                           /*
*****************************************************************************
*           ==============================================                 */
            void   get_radius( x0, y0, x1, y1, x2, y2, 
                               xc, yc,  r, theta, theta0, theta1 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a dotted line from ( xx1, yy1 ) to ( xx3, yy3 ),                *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
double x0, y0, x1, y1, x2, y2;
double *xc, *yc, *r, *theta, *theta0, *theta1;
{                                                                          /*
 ----------------
  Local Variable
 ----------------
                                                                           */
   double x3, y3, length, div;
   double sint, cost, xt,  yt;
   double a0, b0, a1, b1, th;
   float  rpi;
   int   ntimes, n, mo;
   static int mintime = 4;
   if( y0 == y2 )
     {
       a1 = ( x2 - x1 )/(y1 - y2);
       x3 = 0.5*( x1 + x2 ), y3 = 0.5*( y1 + y2 );
       b1 = y3 - a1*x3;
       x3 = 0.5*( x0 + x2 );
       *xc = x3;
       *yc = a1*x3 + b1;
      }
   else if( y1 == y2 )
     {
       a0 = ( x2 - x0 )/(y0 - y2);
       x3 = 0.5*( x0 + x2 ), y3 = 0.5*( y0 + y2 );
       b0 = y3 - a0*x3;
       x3 = 0.5*( x1 + x2 );
       *xc = x3;
       *yc = a0*x3 + b0;
      }
   else
     {
       a0 = ( x2 - x0 )/(y0 - y2);
       x3 = 0.5*( x0 + x2 ), y3 = 0.5*( y0 + y2 );
       b0 = y3 - a0*x3;
       a1 = ( x2 - x1 )/(y1 - y2);
       x3 = 0.5*( x1 + x2 ), y3 = 0.5*( y1 + y2 );
       b1 = y3 - a1*x3;
       x3 = ( b0 - b1 )/( a1 - a0 );
       *xc = x3;
       *yc = a0*x3 + b0;
      }
   *theta0 = atan2( y0 - *yc, x0 - *xc );
   *theta1 = atan2( y1 - *yc, x1 - *xc );
   *theta  = *theta1 - *theta0;
   *r      = ( x0 - *xc )*( x0 - *xc ) + ( y0 - *yc )*( y0 - *yc );
   *r      = sqrt( *r );

  /*
   printf(" print_ps =%2d\n", print_ps );
   printf(" x0  = %6.1f,   x1  = %6.1f  x2 = %6.1f\n", x0, x1, x2 );
   printf(" y0  = %6.1f,   y1  = %6.1f  y2 = %6.1f\n", y0, y1, y2 );
   printf(" xc  = %6.1f\n", *xc );
   printf(" yc  = %6.1f\n", *yc );
   printf(" r0  = %6.1f\n", *r );
   printf(" theta1  = %6.1f,   theta0  = %6.1f  theta = %6.1f\n", 
	    *theta1*rdeg, *theta0*rdeg, *theta*rdeg );
  */

   if( *theta >  pi )
       *theta1 = *theta1 - pi2;
   else if( *theta < -pi )
       *theta0 = *theta0 - pi2;
   *theta = *theta1 - *theta0;

    /*
      printf(" theta1  = %6.1f,   theta0  = %6.1f  theta = %6.1f\n", 
	       *theta1*rdeg, *theta0*rdeg, *theta*rdeg );
    */
}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
            void   urdottarc( xx1, yy1, xx2, yy2, xx3, yy3 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a dotted line from ( xx1, yy1 ) to ( xx3, yy3 ),                *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2, xx3, yy3;
{                                                                          /*
 ----------------
  Local Variable
 ----------------
                                                                           */
   double x0, y0, x1, y1, x2, y2, x3, y3, xc, yc, length, div, theta;
   double sint, cost, xt,  yt, r1;
   double theta0, theta1, r, th;
   float  rpi;
   int   ntimes, n;
   static int mintime = 4;
                                                                           /*
------------------
     Functions
------------------
                                                                          */
 void get_radius();
                                                                           /*
================================ Entry Point ===============================
                                                                           */
                                                                            /*
   +-------------------------------------------------------------------------+
   |(1) Calculate the center and radious of the circle.                      |
   +-------------------------------------------------------------------------+
                                                                            */
   x0 = (double) xx1;
   y0 = (double) yy1;
   x1 = (double) xx3;
   y1 = (double) yy3;
   x2 = (double) xx2;
   y2 = (double) yy2;

   get_radius( x0, y0, x1, y1, x2, y2,
               &xc, &yc,  &r, &theta, &theta0, &theta1 );

   if( theta > 0.0 )
       length = theta;
   else
       length = -theta;

   div    = length*r / dott_length;
   ntimes = (int) div;
   if( ntimes < mintime )
       ntimes = mintime;
                                                                            /*
   +-------------------------------------------------------------------------+
   |(2) If flag_dtarc > 0 and print_ps == 1, print ps_command                |
   +-------------------------------------------------------------------------+
                                                                            */
   if( print_ps == 1 && flag_dtarc >= 0 )
     {
       rpi = 180.0 / pi;
       fprintf( psout,"gsave\nnewpath\n");
       fprintf( psout," %8.2f %8.2f %8.2f", 
               (float) xc, (float) yc, (float) r );
       fprintf( psout," %8.2f %8.2f %2d dottarc\n", 
                rpi* (float) theta0, rpi* (float) theta1, ntimes );
       fprintf( psout,"stroke\ngrestore\n");
      }
   else
     {
       div    = length / ( ntimes + 0.5 );
       div    = div * 0.5;
/*
            printf(" ntimes = %d, div = %15.6f  length = %15.6f\n", 
	             ntimes, div, length );
*/
       xx1    = ( float ) x0,  yy1 = ( float ) y0;

       if( print_ps == 1 )
         {
           fprintf( psout, "newpath\n");
          }

       debug = 1;

       xt    = 0.0;
       for( n = 0; n < ntimes; n++ )
          {
            xt =  xt + div;
            if( theta > 0.0 ) th = theta0 + xt;
            else              th = theta0 - xt;

            cost = cos( th );
            sint = sin( th );
            x3 =  r * cost + xc;
            y3 =  r * sint + yc;

            xx2 = (float) x3, yy2 = (float) y3;

          /*
	    r1 = ( x3 - xc )*( x3 - xc) + ( y3 - yc )*( y3 -yc );
	    r1 = sqrt( r1 );
	    printf("n = %2d, th = %10.6f  xt = %8.6f", n, th*rdeg, xt );
            printf("  xx2 = %8.2f, yy2 = %8.2f", xx2, yy2 );
            printf("  xc = %8.2f,  yc = %8.2f  r = %8.2f\n", xc, yc, r );
	    printf("cost = %10.6f  sint = %10.6f  r1 = %8.2f\n", 
		    cost, sint, r1 );
          */

            urLine( xx1, yy1, xx2, yy2 );
            psLine( xx1, yy1, xx2, yy2 );
               /*
                printf(" xx1 = %8.2f, yy1 = %8.2f\n", xx1, yy1 );
                printf(" xx2 = %8.2f, yy2 = %8.2f\n", xx2, yy2 );
               */
            xt =  xt + div;
            if( theta > 0.0 ) th = theta0 + xt;
            else              th = theta0 - xt;

            cost = cos( th );
            sint = sin( th );
            x1 =  r * cost + xc;
            y1 =  r * sint + yc;
            xx1 =  x1, yy1 = y1;
          }

       xx1 = (float) x1, yy1 = (float) y1;
      
       urLine( xx1, yy1, xx3,  yy3 );
       
       if( print_ps == 1 )
         {
           fprintf( psout, " %9.3f %9.3f moveto\n", xx1, yy1 );
           fprintf( psout, " %9.3f %9.3f lineto\n", xx3, yy3 );
           fprintf( psout, "stroke\n");
          }
      }
   return ;
}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
            void   urdtcircl( xx1, yy1, xx2, yy2 )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a dotted line from ( xx1, yy1 ) to ( xx3, yy3 ),                *
*      and ( xx2, yy2 ) defines the middle point of the wave arc.           *
*                                                                           *
*****************************************************************************
									   */
float xx1, yy1, xx2, yy2;
{                                                                          /*
 ----------------
  Local Variable
 ----------------
                                                                           */
   double x0, y0, x1, y1, x2, y2, x3, y3, xc, yc, length, div, theta;
   double sint, cost, xt,  yt, r1;
   double theta0, theta1, r, th;
   float  rpi;
   int   ntimes, n;
   static int mintime = 10;
   float  xx3, yy3;
                                                                           /*
------------------
     Functions
------------------
                                                                          */
                                                                           /*
================================ Entry Point ===============================
                                                                           */
                                                                            /*
   +-------------------------------------------------------------------------+
   |(1) Calculate the center and radious of the circle.                      |
   +-------------------------------------------------------------------------+
                                                                            */
   x0 = (double) xx1;
   y0 = (double) yy1;
   x1 = (double) xx2;
   y1 = (double) yy2;

   xc = 0.5*( x0 + x1 ),  yc = 0.5*( y0 + y1 );
   r  = ( x0 - xc )*( x0 - xc ) + ( y0 - yc )*( y0 - yc );
   r  = sqrt( r );
/*
   if( x0 > x1 - 1.0 && x0 < x1 + 1.0 )
     {
       if( y0 < y1 ) theta0 = -0.5*pi;
       else          theta0 =  0.5*pi;
      }
   else
*/
       theta0  = atan2( y0 - yc, x0 - xc );

   length = pi2;

   div    = length*r / dott_length;
   ntimes = (int) div;
   if( ntimes < mintime )
       ntimes = mintime;
/**
   printf(" ( xc, yc ) = ( %8.2f, %8.2f ), r = %8.2f\n", xc, yc, r );
   printf(" theta0 = %8.2f\n", theta0 );
/**/
                                                                            /*
   +-------------------------------------------------------------------------+
   |(2) If flag_dtarc > 0 and print_ps == 1, print ps_command                |
   +-------------------------------------------------------------------------+
                                                                            */
       div    = length / ( ntimes + 0.5 );
       div    = div * 0.5;
/**
            printf(" ntimes = %d, div = %15.6f  length = %15.6f\n", 
	             ntimes, div, length );
/**/
       xx1    = ( float ) x0,  yy1 = ( float ) y0;

       if( print_ps == 1 )
         {
           fprintf( psout, "newpath\n");
          }

       debug = 1;

       xt    = 0.0;
       for( n = 0; n < ntimes; n++ )
          {
            xt =  xt + div;
            th = theta0 + xt;

            cost = cos( th );
            sint = sin( th );
            x3 =  r * cost + xc;
            y3 =  r * sint + yc;

            xx2 = (float) x3, yy2 = (float) y3;

          /**
	    printf("%2d %6.1f ( xx1, yy1) = ( %8.2f, %8.2f) ", 
                      n, th*rdeg, xx1, yy1 );
	    printf("( xx2, yy2) = ( %8.2f, %8.2f)\n", xx2, yy2 );
          /**/

            urLine( xx1, yy1, xx2, yy2 );
            psLine( xx1, yy1, xx2, yy2 );

            xt =  xt + div;
            th = theta0 + xt;

            cost = cos( th );
            sint = sin( th );
            x1 =  r * cost + xc;
            y1 =  r * sint + yc;
            xx1 =  x1, yy1 = y1;
          }

       xx3 = (float) x1, yy3 = (float) y1;
      
       urLine( xx1, yy1, xx3,  yy3 );

       if( print_ps == 1 )
         {
           psLine( xx1, yy1, xx3, yy3 );
           fprintf( psout, "stroke\n");
          }
   return ;
}
                                                                           /*
*****************************************************************************
*        ==========================================================       */
          void urcurveto( xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3 )
/*       ==========================================================
*****************************************************************************
* ( Function )                                                              *
*   To draw the Bezier 3rd order curve defined by                           *
*     x0, y0, x1, y1, x2, y2, x3, y3.                                       *
*                                                                           *
*****************************************************************************
									   */
float xx0, yy0, xx1, yy1, xx2, yy2, xx3, yy3;
{
float ax, bx, cx, ay, by, cy, t, t2, t3, dt;
float xp, yp, xt, yt;
int    n;

void urLine( );

  cx = 3.0*( xx1 - xx0 );
  bx = 3.0*( xx2 - xx1 ) - cx;
  ax = xx3 - xx0 - bx - cx;

  cy = 3.0*( yy1 - yy0 );
  by = 3.0*( yy2 - yy1 ) - cy;
  ay = yy3 - yy0 - by - cy;

  xp = xx0,  yp = yy0, dt = 0.1;
  for( n = 1; n < 11; n++ )
     {
       t  = dt* n;
       t2 = t * t;
       t3 = t2* t;

       xt = ax*t3 + bx*t2 + cx*t + xx0;
       yt = ay*t3 + by*t2 + cy*t + yy0;

       urLine( xp, yp, xt, yt );

       xp = xt,  yp = yt;

      }

}

                                                                           /*
*****************************************************************************
*                 =================================                        */
                      void urcircle( x1, y1, r )
/*                =================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a circle centred at ( x1, y1 ) with radius r                    *
*                                                                           *
*****************************************************************************
									   */
float x1, y1, r;
{
int x0, y0, ir;
float r0;

if( print_ps == 0 )
  {
    x0 = (int) x1 - r ;
    y0 = (int) y1 - r ;
    ir = (int) 2.0*r ;

#ifdef                                                          DFcircl
    printf("x0(%6.1f) y0(%6.1f) ir(%d) n", x0, y0, ir );
#endif

    if( size_flag < 2 ) 
        XFillArc ( d, w0, gc1, x0, y0, ir, ir, 0, 360*64 ) ;
    else if( size_flag < 4 ) 
        XFillArc ( d, w0, gc2, x0, y0, ir, ir, 0, 360*64 ) ;
    else 
        XFillArc ( d, w0, gc3, x0, y0, ir, ir, 0, 360*64 ) ;
   }
else
  {
    pstrans( x1, y1 );
    fprintf( psout, "newpath %9.3f %9.3f ", xxx, yyy );
    pstrans( 0.0, 0.0 );
    r0 = xxx;
    pstrans( r, 0.0 );
    r0 = xxx - r0;
    fprintf( psout, " %9.3f 0 360 arc closepath fill\n", r0 );
   }
}
                                                                           /*
*****************************************************************************
*                 =================================                        */
                      void urcircle2( x1, y1, r )
/*                =================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a circle centred at ( x1, y1 ) with radius r                    *
*                                                                           *
*****************************************************************************
									   */
float x1, y1, r;
{
int x0, y0, ir;
float r0;

if( print_ps == 0 )
  {
    x0 = (int) x1 - r ;
    y0 = (int) y1 - r ;
    ir = (int) 2.0*r ;
    if( size_flag < 2 ) 
        XDrawArc ( d, w0, gc1, x0, y0, ir, ir, 0, 360*64 ) ;
    else if( size_flag < 4 ) 
        XDrawArc ( d, w0, gc2, x0, y0, ir, ir, 0, 360*64 ) ;
    else 
        XDrawArc ( d, w0, gc3, x0, y0, ir, ir, 0, 360*64 ) ;
   }
else
  {
    pstrans( x1, y1 );
    fprintf( psout, "newpath %9.3f %9.3f ", xxx, yyy );
    pstrans( 0.0, 0.0 );
    r0 = xxx;
    pstrans( r, 0.0 );
    r0 = xxx - r0;
    fprintf( psout, " %9.3f 0 360 arc stroke\n", r0 );
   }
}
                                                                           /*
*****************************************************************************
*                 =================================                        */
                      void FillRectangle( xx, yy, width, height )
/*                =================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a circle centred at ( x1, y1 ) with radius r                    *
*                                                                           *
*****************************************************************************
									   */
float xx, yy, width, height;
{
int x0, y0, ww, hh;

  x0  = (int) xx,    y0 = (int) yy;
  ww  = (int) width, hh = (int) height;
         /*
	  printf(" FillRectangle (xx, yx) = (%8.1f,%8.1f )\n", xx, yy );
          printf("               (wi, hi) = (%8.1f,%8.1f )\n", width,height );
	  printf(" FillRectangle (x0, y0) = (%8d,%8d )\n", x0, y0 );
          printf("               (ww, hh) = (%8d,%8d )\n", ww,hh );
         */

  XFillRectangle( d, w0, gcslct, x0, y0, ww, hh );

}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
             void   urtxtda ( xx, yy,  nc, text, color )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a text string "text" from ( xx, yy ).                           *
*                                                                           *
*****************************************************************************
									   */
float xx, yy;
int   nc;
char  *text, *color;
{
int x2, y2;

if( nc > 0 )
  {
    x2 = (int) xx;
    y2 = (int) yy;
    if( strcmp( color, "blue") == 0 )
        XDrawString( d, w1, gda1, x2, y2, text, nc );
    else if( strcmp( color, "red") == 0 )
        XDrawString( d, w1, gda2, x2, y2, text, nc );
    else
        XDrawString( d, w1, gda3, x2, y2, text, nc );
  }
}

                                                                           /*
*****************************************************************************
*           ================================================               */
            void urtxts ( size_f, xx, yy,  nc, text, color )
/*          ================================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a text string "text" from ( xx, yy ).                           *
*                                                                           *
*****************************************************************************
									   */
float xx, yy;
int   size_f, nc;
char  *text, *color;
{
int x2, y2;

if( nc > 0 )
  {
    if( print_ps == 0 )
      {
        x2 = (int) xx;
        y2 = (int) yy;
        if( strcmp( color, "white") == 0 )
          {
            if( size_f <  0 )
                XDrawString( d, w0, gcrw0, x2, y2, text, nc );
            if( size_f == 0 )
                XDrawString( d, w0, gcrw1, x2, y2, text, nc );
            else if( size_f == 1 )
                XDrawString( d, w0, gcrw2, x2, y2, text, nc );
            else if( size_f == 2 )
                XDrawString( d, w0, gcrw3, x2, y2, text, nc );
            else if( size_f == 3 )
                XDrawString( d, w0, gcrw4, x2, y2, text, nc );
            else if( size_f == 4 )
                XDrawString( d, w0, gcrw5, x2, y2, text, nc );
           }
        else
          {
            if( size_f <= 0 )
                XDrawString( d, w0, gcr1, x2, y2, text, nc );
            else if( size_f == 1 )
                XDrawString( d, w0, gcr2, x2, y2, text, nc );
            else if( size_f == 2 )
                XDrawString( d, w0, gcr3, x2, y2, text, nc );
            else if( size_f == 3 )
                XDrawString( d, w0, gcr4, x2, y2, text, nc );
            else if( size_f == 4 )
                XDrawString( d, w0, gcr5, x2, y2, text, nc );
           }
       }
    else
      {
        fprintf( psout, "newpath\n");
        pstrans( xx, yy );
        if( size_f != size_flag )
        fprintf( psout, " %9.3f  %9.3f  moveto SubTimes setfont (%s) show\n", 
		  xxx, yyy, text );
        else
        fprintf( psout, " %9.3f  %9.3f  moveto MainTimes setfont (%s) show\n", 
		  xxx, yyy, text );
       }
   }
}

int csrbear, cslbear, csascent, csdescent;
int descent0;
int size_fl;

                                                                           /*
*****************************************************************************
*           =====================================================          */
            void urDrawString ( size_f, xx, yy, nc, code, form )
/*          =====================================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a text string code from ( xx, yy ) in the format of form.       *
*                                                                           *
*****************************************************************************
									   */
float xx, yy;
int size_f, nc;
char  code[], form[];
{
   char c[10];
/*
   char code[40], form[40]
            form[i]     meaning            form[i]     meaning
                 t  :   times                   s  :   symbol
		 u  :   times upper case        l  :   times lower case
		 a  :   times upper case after backspace
		 A  :   times upper case after backspace
		 U  :   symbol upper case       L  :   symbol lower case
		 o  :   \bar{times}             O  :   \bar{symbol} 
		 w  :   \tilde{times}           W  :   \tilde{symbol} 
*/
   int width, height, widh, heih, dh, ascnt, dscnt;
   char  f;
   int   n, m;
   int  x2, y2, x0, y0, xt, yt;
   float ddx;
   int size_fl, nn;
   static char *tilde = "~";

   void   strcopy  ( );
   void   urMainTimesSize   ( );
   void   urMainSymbolSize  ( );
   void   urSubTimesSize    ( );
   void   urSubSymbolSize   ( );
   void   urStringPosition  ( );
   void   urMainTimesDraw   ( );
   void   urMainSymbolDraw  ( );
   void   urSubTimesDraw    ( );
   void   urSubSymbolDraw   ( );
   void   urBar             ( );

   x2 = (int) xx ;
   y2 = (int) yy ;
   size_fl = size_f;

   for( n = 0; n < nc; n++)
      {
	strcopy( c, code, n, n );
        m  = strlen( c );
        f = form[n];

	size_f = size_fl;
/*
	if( strcmp( c, "H" ) == 0 )
            size_f = size_fl - 1;
	if( strcmp( c, "Z" ) == 0 )
            size_f = size_fl - 1;
	if( strcmp( c, "W" ) == 0 )
            size_f = size_fl - 1;
        if( size_f < 0 )
            size_f = 0;
*/
        switch( f ){
                                                                           /*
-----------------------------------------------------------------------------
	      Normal script with Times
-----------------------------------------------------------------------------
									   */
          case 't' :

	       urMainTimesSize( size_f, c, m );

	       width = csrbear - cslbear;

                   /*
                   printf("Before Times x2    =%5d  y2    =%5d\n", 
				   x2, y2 );
                   */

	       xt = x2;
	       x2 = x2 - cslbear;
                   /*
                   printf("After  Times x2    =%5d  y2    =%5d\n", 
				  x2, y2 );
                   */

               urMainTimesDraw( size_f, x2, y2, c, m );

               ascnt = csascent;
               dscnt = csdescent;
               x2 = x2 + csrbear;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Normal script with Symbol
-----------------------------------------------------------------------------
									   */
          case 's' :

	       urMainSymbolSize( size_f, c, m );

	       width = csrbear - cslbear;

                   /*
                   printf("Before Symbol x2    =%5d  y2    =%5d\n", 
				  x2, y2 );
                   */

	       xt = x2;
	       x2 = x2 - cslbear;

                   /*
                   printf("After  Symbol x2    =%5d  y2    =%5d\n", 
				  x2, y2 );
                   */

	       urMainSymbolDraw( size_f, x2, y2, c, m );

               ascnt = csascent;
               dscnt = csdescent;
               x2 = x2 + csrbear;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Super script with Times
-----------------------------------------------------------------------------
									   */
          case 'u' :
        /*     y2 = y2 - ascnt/2;  */
               y2 = y2 - 0.7*ascnt;

	       urSubTimesSize( size_f, c, m );

	       urSubTimesDraw( size_f, x2, y2, c, m );

               width = csrbear - cslbear;
	       xt = x2;
               x2 = x2 + width;
        /*     y2 = y2 + ascnt/2;  */
               y2 = y2 + 0.7*ascnt;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Super script with Times after back space
-----------------------------------------------------------------------------
									   */
          case 'a' :
        /*     y2 = y2 - ascnt/2;  */
               y2 = y2 - 0.7*ascnt;

	       urSubTimesSize( size_f, c, m );

	       urSubTimesDraw( size_f, xt, y2, c, m );

               width = csrbear - cslbear;
               xt = xt + width;
               if( xt > x2 ) x2 = xt;
        /*     y2 = y2 + ascnt/2;  */
               y2 = y2 + 0.7*ascnt;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Sub_script with Times
-----------------------------------------------------------------------------
									   */
          case 'l' :

               urSubTimesSize( size_f, c, m );

               width = csrbear - cslbear;
               ascnt = csascent;
               y2 = y2 + ascnt*2/3;

	       urSubTimesDraw( size_f, x2, y2, c, m );

	       xt = x2;
               x2 = x2 + width;
               y2 = y2 - ascnt*2/3;

               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Super script with Symbol
-----------------------------------------------------------------------------
									   */
          case 'U' :
        /*     y2 = y2 - ascnt/2;  */
               y2 = y2 - 0.7*ascnt;

	       urSubSymbolSize( size_f, c, m );

	       urSubSymbolDraw( size_f, x2, y2, c, m );

               width = csrbear - cslbear;
	       xt = x2;
               x2 = x2 + width;
        /*     y2 = y2 + ascnt/2;  */
               y2 = y2 + 0.7*ascnt;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Super script with Symbol after back space
-----------------------------------------------------------------------------
									   */
          case 'A' :
        /*     y2 = y2 - ascnt/2;  */
               y2 = y2 - 0.7*ascnt;

	       urSubSymbolSize( size_f, c, m );

	       urSubSymbolDraw( size_f, xt, y2, c, m );

               width = csrbear - cslbear;
               xt = xt + width;
               if( xt > x2 ) x2 = xt;
        /*     y2 = y2 + ascnt/2;  */
               y2 = y2 + 0.7*ascnt;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Sub_script with Symbol
-----------------------------------------------------------------------------
									   */
          case 'L' :

	       urSubSymbolSize( size_f, c, m );

               width = csrbear - cslbear;
               ascnt = csascent;
               y2 = y2 + ascnt*2/3;

	       urSubSymbolDraw( size_f, x2, y2, c, m );

	       xt = x2;
               x2 = x2 + width;
               y2 = y2 - ascnt*2/3;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Times normal script with "bar"
-----------------------------------------------------------------------------
									   */
          case 'o' :

	       urMainTimesSize( size_f, c, m );

               width  = csrbear - cslbear;

                  /* 
                   printf("Before T_Bar x2    =%5d  y2    =%5d  width =%5d\n", 
				   x2, y2, width );
                  */

	       xt = x2;
	       x2 = x2 - cslbear;

                   /*
                   printf("After  T_Bar x2    =%5d  y2    =%5d\n", x2, y2 );
                   */

	       urMainTimesDraw( size_f, x2, y2, c, m );

               ascnt = csascent;
               y2    = y2 - ascnt* 1.3;

	       if( print_ps == 0 )
                 {
                   urBar( x2, y2, x2+width, y2 );
                  }
                else
                 {
                   fprintf( psout, "newpath\n");
                   pstrans((float) x2,(float) y2 );
                   fprintf( psout," %8.2f %8.2f moveto", xxx, yyy );
                   pstrans((float) x2+width,(float) y2 );
                   fprintf( psout," %8.2f %8.2f lineto\n", xxx, yyy );
                   fprintf( psout, "stroke\n");
                  }
                      
	       xt = x2;
               x2 = x2 + csrbear;
               y2 = y2 + ascnt* 1.3;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Symbol normal script with "bar"
-----------------------------------------------------------------------------
									   */
          case 'O' :

	       urMainSymbolSize( size_f, c, m );

               width  = csrbear - cslbear;

                   /*
                   printf("Before S_Bar x2    =%5d  y2    =%5d\n", x2, y2 );
                   */

	       xt = x2;
	       x2 = x2 - cslbear;

                   /*
                   printf("After  S_Bar x2    =%5d  y2    =%5d\n", x2, y2 );
                   */

	       urMainSymbolDraw( size_f, x2, y2, c, m );

               ascnt = csascent;

               y2    = y2 - ascnt * 1.3;

	       if( print_ps == 0 )
                 {
                   urBar( x2, y2, x2+width, y2 );
                  }
                else
                 {
                   fprintf( psout, "newpath\n");
                   pstrans((float) x2,(float) y2 );
                   fprintf( psout," %8.2f %8.2f moveto", xxx, yyy );
                   pstrans((float) x2+width,(float) y2 );
                   fprintf( psout," %8.2f %8.2f lineto\n", xxx, yyy );
                   fprintf( psout, "stroke\n");
                  }

               x2 = x2 + csrbear;
               y2 = y2 + ascnt * 1.3;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Times normal script with "tlde"
-----------------------------------------------------------------------------
									   */
          case 'w' :

	       urMainTimesSize( size_f, c, m );

               width  = csrbear - cslbear;

                  /* 
                   printf("Before T_Bar x2    =%5d  y2    =%5d  width =%5d\n", 
				   x2, y2, width );
                  */

	       xt = x2;
	       x2 = x2 - cslbear;

                   /*
                   printf("After  T_Bar x2    =%5d  y2    =%5d\n", x2, y2 );
                   */

	       urMainTimesDraw( size_f, x2, y2, c, m );

               ascnt = csascent;
               y2    = y2 - ascnt;

               urMainTimesSize( size_f, tilde, m );
               yt    = y2 + csascent/3 ;
               urMainTimesDraw( size_f, x2, yt, tilde, m );

               x2 = x2 + csrbear;
               y2 = y2 + ascnt;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Symbol normal script with "bar"
-----------------------------------------------------------------------------
									   */
          case 'W' :

	       urMainSymbolSize( size_f, c, m );

               width  = csrbear - cslbear;

                   /*
                   printf("Before S_Bar x2    =%5d  y2    =%5d\n", x2, y2 );
                   */

	       xt = x2;
	       x2 = x2 - cslbear;

                   /*
                   printf("After  S_Bar x2    =%5d  y2    =%5d\n", x2, y2 );
                   */

	       urMainSymbolDraw( size_f, x2, y2, c, m );

               ascnt = csascent;

               y2    = y2 - ascnt;

               urMainTimesSize( size_f, tilde, m );
               yt    = y2 + csascent/3 ;
               urMainTimesDraw( size_f, x2, yt, tilde, m );

               x2 = x2 + csrbear;
               y2 = y2 + ascnt;
               break;
              }
     }
   size_f = size_fl;
   return ;
}
                                                                           /*
*****************************************************************************
*          ====================================================            */
            void urStringSize ( size_f, nc, code, form, dcs, w, h )
/*         ====================================================
*****************************************************************************
* ( Function )                                                              *
*   To calculate the half_width and half_height of the string               *
*                                                                           *
*****************************************************************************
									   */
int size_f, nc, *dcs, *w, *h;
char  code[], form[];
{
   char c[10];
   char *tilde = "~";
/*
   char code[40], form[40]
            form[i]     meaning            form[i]     meaning
                 t  :   times                   s  :   symbol
		 u  :   times upper case        l  :   times lower case
		 a  :   times upper case after backspace
		 A  :   times upper case after backspace
		 U  :   symbol upper case       L  :   symbol lower case
		 o  :   \bar{times}             O  :   \bar{symbol} 
		 w  :   \tilde{times}           W  :   \tilde{symbol} 
*/
   int width, height, widh, heih, dh, ascnt, dscnt;
   char  f;
   int   n, m;
   int  x2, y2, xt, yt, y_max, y_min;

   void   strcopy  ( );
   void   urMainTimesSize   ( );
   void   urMainSymbolSize  ( );
   void   urSubTimesSize    ( );
   void   urSubSymbolSize   ( );

   x2    =    0;
   y2    =    0;
   y_max =    0;
   y_min = 1000;

   size_fl = size_f;

   for( n = 0; n < nc; n++)
      {
	strcopy( c, code, n, n );
        m  = strlen( c );
        f = form[n];

	size_f = size_fl;
/*
	if( strcmp( c, "H" ) == 0 )
            size_f = size_fl - 1;
	if( strcmp( c, "Z" ) == 0 )
            size_f = size_fl - 1;
	if( strcmp( c, "W" ) == 0 )
            size_f = size_fl - 1;
        if( size_f < 0 )
            size_f = 0;
*/
        switch( f ){
                                                                           /*
-----------------------------------------------------------------------------
	      Normal script with Times
-----------------------------------------------------------------------------
									   */
          case 't' :

	       urMainTimesSize( size_f, c, m );

	       xt = x2;
               x2  = csrbear - cslbear + x2;

		   /*
	             printf("StringSize  heght =%7d\n", 
	                       csascent + csdescent );
	             printf("            width =%7d\n", 
			       csrbear - cslbear );
                   */

	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;
	       yt  = y2 - csascent;
               if( yt < y_min ) y_min = yt;

               ascnt = csascent;
               dscnt = csdescent;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Normal script with Symbol
-----------------------------------------------------------------------------
									   */
          case 's' :

	       urMainSymbolSize( size_f, c, m );

	       xt = x2;
               x2  = csrbear - cslbear + x2;

		   /*
	             printf("StringSize  heght =%7d\n", 
	                       csascent + csdescent );
	             printf("            width =%7d\n", 
			       csrbear - cslbear );
                   */

	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;
	       yt  = y2 - csascent;
               if( yt < y_min ) y_min = yt;

               ascnt = csascent;
               dscnt = csdescent;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Super script with Times
-----------------------------------------------------------------------------
									   */
          case 'u' :
           /*  y2 = y2 - ascnt/2; */
               y2 = y2 - 0.7*ascnt;

	       urSubTimesSize( size_f, c, m );

	       xt = x2;
               x2 = csrbear - cslbear + x2;

	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;
	       yt  = y2 - csascent;
               if( yt < y_min ) y_min = yt;

           /*  y2 = y2 + ascnt/2; */
               y2 = y2 + 0.7*ascnt;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Super script with Times after back space
-----------------------------------------------------------------------------
									   */
          case 'a' :
           /*  y2 = y2 - ascnt/2; */
               y2 = y2 - 0.7*ascnt;

	       urSubTimesSize( size_f, c, m );

               xt = csrbear - cslbear + xt;
               if( xt > x2 ) x2 = xt;

	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;
	       yt  = y2 - csascent;
               if( yt < y_min ) y_min = yt;

           /*  y2 = y2 + ascnt/2; */
               y2 = y2 + 0.7*ascnt;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Sub_script with Times
-----------------------------------------------------------------------------
									   */
          case 'l' :

               urSubTimesSize( size_f, c, m );

	       xt = x2;
               x2 = csrbear - cslbear + x2;
               ascnt = csascent;
               y2 = y2 + ascnt*2/3;

	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;

               y2 = y2 - ascnt*2/3;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Super script with Symbol
-----------------------------------------------------------------------------
									   */
          case 'U' :
           /*  y2 = y2 - ascnt/2; */
               y2 = y2 - 0.7*ascnt;

	       urSubSymbolSize( size_f, c, m );

	       xt = x2;
               x2 = csrbear - cslbear + x2;

	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;
	       yt  = y2 - csascent;
               if( yt < y_min ) y_min = yt;

           /*  y2 = y2 + ascnt/2; */
               y2 = y2 + 0.7*ascnt;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Super script with Symbol after back space
-----------------------------------------------------------------------------
									   */
          case 'A' :
           /*  y2 = y2 - ascnt/2; */
               y2 = y2 - 0.7*ascnt;

	       urSubSymbolSize( size_f, c, m );

               xt = csrbear - cslbear + xt;
               if( xt > x2 ) x2 = xt;

	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;
	       yt  = y2 - csascent;
               if( yt < y_min ) y_min = yt;

           /*  y2 = y2 + ascnt/2; */
               y2 = y2 + 0.7*ascnt;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Sub_script with Symbol
-----------------------------------------------------------------------------
									   */
          case 'L' :

	       urSubSymbolSize( size_f, c, m );

	       xt = x2;
               x2 = csrbear - cslbear + x2;
               ascnt = csascent;
               y2 = y2 + ascnt*2/3;

	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;

               y2 = y2 - ascnt*2/3;
               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Times normal script with "bar"
-----------------------------------------------------------------------------
									   */
          case 'o' :

	       urMainTimesSize( size_f, c, m );

	       xt = x2;
               x2 = csrbear - cslbear + x2;
	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;

               ascnt = csascent;
               yt    = y2 - ascnt* 1.3;
               if( yt < y_min ) y_min = yt;

               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Symbol normal script with "bar"
-----------------------------------------------------------------------------
									   */
          case 'O' :

	       urMainSymbolSize( size_f, c, m );

	       xt = x2;
               x2 = csrbear - cslbear + x2;
	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;

               ascnt = csascent;
               yt    = y2 - ascnt* 1.3;
               if( yt < y_min ) y_min = yt;

               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Times normal script with "tilde"
-----------------------------------------------------------------------------
									   */
          case 'w' :

	       urMainTimesSize( size_f, c, m );

	       xt = x2;
               x2 = csrbear - cslbear + x2;
	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;

               ascnt = csascent;
               yt    = y2 - ascnt;

               urMainTimesSize( size_f, tilde, m );
               yt    = yt - csascent*2 /3;
               if( yt < y_min ) y_min = yt;

               break;
                                                                           /*
-----------------------------------------------------------------------------
	      Symbol normal script with "tlde"
-----------------------------------------------------------------------------
									   */
          case 'W' :

	       urMainSymbolSize( size_f, c, m );

	       xt = x2;
               x2 = csrbear - cslbear + x2;
	       yt  = y2 + csdescent;
               if( yt > y_max ) y_max = yt;

               ascnt = csascent;
               yt    = y2 - ascnt;

               urMainTimesSize( size_f, tilde, m );
               yt    = yt - csascent*2 /3;
               if( yt < y_min ) y_min = yt;

               break;
              }
     }

   *dcs = y_max;
   *w   = ( x2 + 1 )/2;
   *h   = ( y_max - y_min + 1 )/2;

   return ;
}
                                                                           /*
*****************************************************************************
*                 =================================                        */
                   void urBar( ix1, iy1, ix2, iy2 )
/*                =================================
*****************************************************************************
* ( Function )                                                              *
*   To draw a line from ( ix1, iy1 ) to ( ix2, iy2 )                        *
*                                                                           *
*****************************************************************************
									   */
int ix1, iy1, ix2, iy2;
{

 if( print_ps == 0 )
   {
     if( size_flag < 3 ) 
         XDrawLine ( d, w0, gcc1, ix1, iy1, ix2, iy2 ) ;
     else if( size_flag < 4 ) 
         XDrawLine ( d, w0, gcc2, ix1, iy1, ix2, iy2 ) ;
     else 
         XDrawLine ( d, w0, gcc3, ix1, iy1, ix2, iy2 ) ;
    }
}
#define MXTIME 20
                                                                           /*
*****************************************************************************
*                ===================================                       */
                  void urMainTimesSize( fsize, c, m )
/*               ===================================
*****************************************************************************
* ( Function )                                                              *
*   To clear the window                                                     *
*                                                                           *
*****************************************************************************
									   */
int fsize, m;
char c[];
{
int dir, ascent, descent;
int i;
float fract;

static char  b[MXTIME] = {   '*',   '+',   '-',   '0',   '1',
                             '2',   '3',   'H',   'L',   'R',
                             'W',   'Z',   'b',   'c',   'd',
                             'e',   'g',   's',    't',   '~' };
static int  as[MXTIME] = {    17,    12,     6,    17,    17,
                              17,    17,    17,    17,    17,
                              17,    17,    17,    11,    17,
                              11,    11,    11,    14,     8  };
static int  ds[MXTIME] = {     0,    -1,    -4,     0,     0,
                               0,     0,     0,     0,     0,
                               0,     0,     0,     0,     0,
                               0,     5,     0,     0,    -3 };
static int  rb[MXTIME] = {    11,    14,     7,    12,    10,
                              11,    11,    20,    14,    14,
                              24,    15,    11,    11,    14,
                              10,    12,     9,     9,    13 };
static int  lb[MXTIME] = {     0,     0,     0,     0,     0,
                               0,     0,     0,     0,     0,
                               1,    -1,     0,     0,     0,
                               0,     0,     0,     1,     0 };

if( print_ps == 0 )
  {
    if( fsize == -1 )
      {
        XTextExtents( f0, c, m, &dir, &ascent, &descent, &chstr );
       }
    else if( fsize == 0 )
      {
        XTextExtents( f1, c, m, &dir, &ascent, &descent, &chstr );
       }
    else if( fsize == 1 )
      {
        XTextExtents( f2, c, m, &dir, &ascent, &descent, &chstr );
       }
    else if( fsize == 2 )
      {
        XTextExtents( f3, c, m, &dir, &ascent, &descent, &chstr );
       }
    else if( fsize == 3 )
      {
        XTextExtents( f4, c, m, &dir, &ascent, &descent, &chstr );
       }
    else if( fsize == 4 )
      {
        XTextExtents( f5, c, m, &dir, &ascent, &descent, &chstr );
       }

    csrbear   = chstr.rbearing;
    cslbear   = chstr.lbearing;
    csascent  = chstr.ascent;
    csdescent = chstr.descent;
    descent0  = descent;
   }
else
  {
    for( i = 0; i < MXTIME; i++ )
       {
         if( c[0] == b[i] )
           {
             if(      fsize == -1 ) fract =  0.3333;
             else if( fsize ==  0 ) fract =  10.0/24.0;
             else if( fsize ==  1 ) fract =  0.5;
             else if( fsize ==  2 ) fract =  14.0/24.0;
             else if( fsize ==  3 ) fract =  18.0/24.0;
             else if( fsize ==  4 ) fract =  1.0;

             csascent  = fract*as[i];
             csdescent = fract*ds[i];
             csrbear   = fract*rb[i];
             cslbear   = fract*lb[i];
             descent0  = csdescent;
             break;
            }
        }
   }
/*
 printf("Times   %s  ", c );
 if(      fsize == -1 ) printf("  8 point ");
 else if( fsize ==  0 ) printf(" 10 point ");
 else if( fsize ==  1 ) printf(" 12 point ");
 else if( fsize ==  2 ) printf(" 14 point ");
 else if( fsize ==  3 ) printf(" 18 point ");
 else if( fsize ==  4 ) printf(" 24 point ");
 printf("as(%4d) ds(%4d) rb(%4d) lb(%4d) csdes(%4d)\n", csascent,
         descent0, csrbear, cslbear, csdescent );
*/
}
                                                                           /*
*****************************************************************************
*                ===================================                       */
                  void urSubTimesSize( fsize, c, m )
/*               ===================================
*****************************************************************************
* ( Function )                                                              *
*   To clear the window                                                     *
*                                                                           *
*****************************************************************************
									   */
int fsize, m;
char c[];
{
  void urMainTimesSize();

  fsize = fsize - 1;

  urMainTimesSize( fsize, c, m );

}
#define MXSYMB 5
                                                                           /*
*****************************************************************************
*                ======================================                    */
                  void urMainSymbolSize( fsize, c, m )
/*               ======================================
*****************************************************************************
* ( Function )                                                              *
*   To clear the window                                                     *
*                                                                           *
*****************************************************************************
									   */
int fsize, m;
char c[];
{
int dir, ascent, descent;
int i;
float fract;

static char  b[MXSYMB] = {   'c',   'g',   'm',   'n',   't' };
static int  as[MXSYMB] = {    12,    12,    12,    12,    12 };
static int  ds[MXSYMB] = {     5,     5,     5,     0,     0 };
static int  rb[MXSYMB] = {    12,    12,    14,    12,    11 };
static int  lb[MXSYMB] = {     0,     0,     0,     0,     0 };
                      
 if( print_ps == 0 )
   {
     if( fsize == -1 )
       {
         XTextExtents( g1, c, m, &dir, &ascent, &descent, &chstr );
        }
     else if( fsize == 0 )
       {
         XTextExtents( g1, c, m, &dir, &ascent, &descent, &chstr );
        }
     else if( fsize == 1 )
       {
         XTextExtents( g2, c, m, &dir, &ascent, &descent, &chstr );
        }
     else if( fsize == 2 )
       {
         XTextExtents( g3, c, m, &dir, &ascent, &descent, &chstr );
        }
     else if( fsize == 3 )
       {
         XTextExtents( g4, c, m, &dir, &ascent, &descent, &chstr );
        }
     else if( fsize == 4 )
       {
         XTextExtents( g5, c, m, &dir, &ascent, &descent, &chstr );
        }

     csrbear   = chstr.rbearing;
     cslbear   = chstr.lbearing;
     csascent  = chstr.ascent;
     csdescent = chstr.descent;
     descent0  = descent;
    }
 else
   {
     for( i = 0; i < MXSYMB; i++ )
        {
          if( c[0] == b[i] )
            {
              if(      fsize == -1 ) fract =  0.3333;
              else if( fsize ==  0 ) fract =  10.0/24.0;
              else if( fsize ==  1 ) fract =  0.5;
              else if( fsize ==  2 ) fract =  14.0/24.0;
              else if( fsize ==  3 ) fract =  18.0/24.0;
              else if( fsize ==  4 ) fract =  1.0;

              csascent  = fract*as[i];
              csdescent = fract*ds[i];
              csrbear   = fract*rb[i];
              cslbear   = fract*lb[i];
              descent0  = csdescent;
              break;
             }
         }
    }
/*
 printf("Symbol  %s  ", c );
 if(      fsize == -1 ) printf("  8 point ");
 else if( fsize ==  0 ) printf(" 10 point ");
 else if( fsize ==  1 ) printf(" 12 point ");
 else if( fsize ==  2 ) printf(" 14 point ");
 else if( fsize ==  3 ) printf(" 18 point ");
 else if( fsize ==  4 ) printf(" 24 point ");
 printf("as(%4d) ds(%4d) rb(%4d) lb(%4d) csdes(%4d)\n", csascent,
         descent0, csrbear, cslbear, csdescent );
*/
}
                                                                           /*
*****************************************************************************
*                ======================================                    */
                  void urSubSymbolSize( fsize, c, m )
/*               ======================================
*****************************************************************************
* ( Function )                                                              *
*   To clear the window                                                     *
*                                                                           *
*****************************************************************************
									   */
int fsize, m;
char c[];
{
 void urMainSymbolSize();

  fsize = fsize - 1;

  urMainSymbolSize( fsize, c, m );

}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
              void urMainTimesDraw( fsize, x2, y2, c, m )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To clear the window                                                     *
*                                                                           *
*****************************************************************************
									   */
int fsize, x2, y2, m;
char c[];
{

  if( print_ps == 0 )
    {
      if( fsize == 0 )
        {
          XDrawString( d, w0, gcr1, x2, y2, c, m );
         }
      else if( fsize == 1 )
       {
          XDrawString( d, w0, gcr2, x2, y2, c, m );
         }
      else if( fsize == 2 )
       {
          XDrawString( d, w0, gcr3, x2, y2, c, m );
         }
      else if( fsize == 3 )
       {
          XDrawString( d, w0, gcr4, x2, y2, c, m );
         }
      else if( fsize == 4 )
       {
          XDrawString( d, w0, gcr5, x2, y2, c, m );
         }
     }
  else
   {
      fprintf( psout, "newpath\n");
                      pstrans( (float) x2, (float) y2 );
      if( fsize == size_fl )   
          fprintf( psout, 
          " %9.3f  %9.3f  moveto MainTimes setfont (%s) show\n", 
		          xxx, yyy, c );
      else 
          fprintf( psout, 
          " %9.3f  %9.3f  moveto SubTimes setfont (%s) show\n", 
		          xxx, yyy, c );
     }
}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
              void urMainSymbolDraw( fsize, x2, y2, c, m )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To clear the window                                                     *
*                                                                           *
*****************************************************************************
									   */
int fsize, x2, y2, m;
char c[];
{
  if( print_ps == 0 )
    {
      if( fsize == 0 )
        {
          XDrawString( d, w0, gcg1, x2, y2, c, m );
         }
      else if( fsize == 1 )
       {
          XDrawString( d, w0, gcg2, x2, y2, c, m );
         }
      else if( fsize == 2 )
       {
          XDrawString( d, w0, gcg3, x2, y2, c, m );
         }
      else if( fsize == 3 )
       {
          XDrawString( d, w0, gcg4, x2, y2, c, m );
         }
      else if( fsize == 4 )
       {
          XDrawString( d, w0, gcg5, x2, y2, c, m );
         }
     }
  else
   {
      fprintf( psout, "newpath\n");
                      pstrans( (float) x2, (float) y2 );
      fprintf( psout, 
          " %9.3f  %9.3f  moveto MainSymb setfont (%s) show\n", 
		          xxx, yyy, c );
     }
}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
              void urSubTimesDraw( fsize, x2, y2, c, m )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To clear the window                                                     *
*                                                                           *
*****************************************************************************
									   */
int fsize, x2, y2, m;
char c[];
{
  if( print_ps == 0 )
    {
      if( fsize == 0 )
        {
          XDrawString( d, w0, gcr0, x2, y2, c, m );
         }
      else if( fsize == 1 )
       {
          XDrawString( d, w0, gcr1, x2, y2, c, m );
         }
      else if( fsize == 2 )
       {
          XDrawString( d, w0, gcr2, x2, y2, c, m );
         }
      else if( fsize == 3 )
       {
          XDrawString( d, w0, gcr3, x2, y2, c, m );
         }
      else if( fsize == 4 )
       {
          XDrawString( d, w0, gcr4, x2, y2, c, m );
         }
     }
  else
   {
      fprintf( psout, "newpath\n");
                      pstrans( (float) x2, (float) y2 );
      fprintf( psout, 
          " %9.3f  %9.3f  moveto SubTimes setfont (%s) show\n", 
		          xxx, yyy, c );
     }
}
                                                                           /*
*****************************************************************************
*           ==============================================                 */
              void urSubSymbolDraw( fsize, x2, y2, c, m )
/*          ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To clear the window                                                     *
*                                                                           *
*****************************************************************************
									   */
int fsize, x2, y2, m;
char c[];
{
  if( print_ps == 0 )
    {
      if( fsize == 0 )
        {
          XDrawString( d, w0, gcg0, x2, y2, c, m );
         }
      else if( fsize == 1 )
       {
          XDrawString( d, w0, gcg1, x2, y2, c, m );
         }
      else if( fsize == 2 )
       {
          XDrawString( d, w0, gcg2, x2, y2, c, m );
         }
      else if( fsize == 3 )
       {
          XDrawString( d, w0, gcg3, x2, y2, c, m );
         }
      else if( fsize == 4 )
       {
          XDrawString( d, w0, gcg4, x2, y2, c, m );
         }
     }
  else
   {
      fprintf( psout, "newpath\n");
                      pstrans( (float) x2, (float) y2 );
      fprintf( psout, 
          " %9.3f  %9.3f  moveto SubSymb setfont (%s) show\n", 
		          xxx, yyy, c );
     }
}

                                                                           /*
*****************************************************************************
*             ==============================================               */
                void urStringPosition( xflag, yflag, x2, y2 )
/*            ==============================================
*****************************************************************************
* ( Function )                                                              *
*   To define the start position of drawing the string in the window        *
*                                                                           *
*****************************************************************************
									   */
int *x2, *y2;
float xflag, yflag;
{
int width, height, widh, heih, dh, yt, nn;
int x0, y0, xx2, yy2;


  xx2    = *x2;
  yy2    = *y2;

  width  = csrbear - cslbear;
  height = csascent + csdescent;

           /*
	   printf(" x2 =%5d  y2 =%5d   xflag =%5.2f  yflag =%5.2f\n",
                    xx2, yy2, xflag, yflag );
           */

  xflag  = xflag * ( float ) width - cslbear;
  yflag  = yflag * ( float ) height;
  xx2     = xx2 + ( int ) xflag;
  yy2     = yy2 + ( int ) yflag;
/*yy2     = yy2 - csdescent;*/
  yy2     = yy2 - descent0;

  xflag  = ( float ) width;
  xflag  = 1.0 * xflag;
  widh   = ( int ) xflag;
  xflag  = ( float ) height;
  xflag  = 1.0 * xflag;
  heih   = ( int ) xflag;
  yt     = yy2;
  for( nn = 0; nn < nchar; nn++ )
     {
       x0 = xch[nn], y0 = ych[nn];
       if( ( x0 >= xx2 - widh  && x0 <= xx2 + widh ) &&
           ( y0 >= yy2 - heih  && y0 <= yy2 + heih ) )
         {
             if( y0 >= yy2 - heih && y0 <= yy2 + heih ) 
               {
                 /*if( y0 >= yy2 - heih && yy2 >= y0 )*/ 
                       if( yflag > 0 )
                           dh = heih - yy2 + y0;
                       else
                           dh = yy2 - y0 + heih;
                       yt = yy2 + dh;
                }
          }
      }
  yy2 = yt;

  xch[nchar] = xx2;
  ych[nchar] = yy2;
  nchar++;

  *x2 = xx2;
  *y2 = yy2;
}
                                                                           /*
*****************************************************************************
*                    ===============                                       */
                      void urclr()
/*                   ===============
*****************************************************************************
* ( Function )                                                              *
*   To clear the window                                                     *
*                                                                           *
*****************************************************************************
									   */
{
           XClearWindow ( d, w0 ) ;
}

                                                                           /*
*****************************************************************************
*                =================================                         */
                  unsigned long MyColor ( color )
/*               =================================
*****************************************************************************
* ( Function )                                                              *
*   To get color pixel code                                                 *
*                                                                           *
*****************************************************************************
									   */
char *color;
{
  Colormap cmap ;
  XColor c0, c1 ;

  cmap = DefaultColormap ( d, 0 ) ;

  XAllocNamedColor (d, cmap, color, &c1, &c0 ) ;

  return (c1.pixel) ;
}
