/*****************************************************************************
*             ============================================                   *
*                  List of functions for vertex table                        *
*             ============================================                   *
*                                                                            *
*   ( 0) void clear_vtxtable( )                                              *
*             to clear the vertex table                                      *
*                                                                            *
*   ( 1) void add_vtxtable( initial, v, kind, flag, clID )                   *
*             to register a new vertex in the table                          *
*                                                                            *
*   ( 2) void shift_vtxtable( add )                                          *
*             to make a space in the table for a new vertex at add           *
*                                                                            *
*   ( 3) void vtxtable_claddr( )                                             *
*             to re-calculate the address table in the table                 *
*                                                                            *
*   ( 4) void rm_vtxtable( v )                                               *
*             to remove the vertex "v" from the table                        *
*                                                                            *
*   ( 5) int  get_vtxno( add )                                               *
*             to get the vertex number at the address "add" on the table     *
*                                                                            *
*   ( 6) int  addr_vtxtable( v )                                             *
*             to get the address of the vertex "v" in the table              *
*                                                                            *
*   ( 7) int  get_kind( v )                                                  *
*             to get the kind of vertex "v"                                  *
*               ( 1: internal, 2: external, 3: radiate IR )                  *
*   ( 8)  int  get_intline( v1, v2 )                                         *
*             to get the inernal line ID between v1 and v2.                  *
*                                                                            *
*   ( 9) void set_kind( v, kind )                                            *
*             to set the kind flag of vertex "v" in the table                *
*                                                                            *
*   (10) int  get_clID( v )                                                  *
*             to get the cluster ID of the vertex "v"                        *
*                                                                            *
*   (11) void set_clID( v, clID )                                            *
*             to set the cluster ID of the vertex "v" in the table           *
*                                                                            *
*   (12) void print_vtxtable( )                                              *
*             to print the vertex table                                      *
*                                                                            *
*   (13) void list_vtxtable( )                                               *
*             to list the vertex number                                      *
*                                                                            *
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>

#define MXLEN  80
#define MXVTX  100
#define MXLINE 100
#define MXGRF  10000
#define MXEXTV 100
                                                                            /*
==============================================================================
     Structure vertex, vtx_line, int_line, inl_vtx
==============================================================================
                                                                            */
struct vertex {
        int     flag;
        int     cntr_flag;
	char    kind[4];       
	float   x;         
	float   y;   
	int     no_line;   
        }      vtx[MXVTX];

struct vtx_line {
	int   vtxno[10];
        int   lineno[10];
        }     vtxline[MXVTX];

struct int_line {
	int   flag;
	char  kind[4];      
        float x;
        float y;
	int   direct;   
	char  species[4];  
	char  form[4];  
	int   print;  
        }     intline[MXLINE];

struct inl_vtx {
	int   vtxno[2];
	}     inlvtx[MXLINE];
                                                                            /*
==============================================================================
    Structure straight line          
==============================================================================
                                                                            */
int  no_straight;  /*  the number of straight lines                     */

struct straight_line {
        int  flag;    /* ( 1 / 0 ) = ( changed / no )                      */
        int  v0;      /*  the first vertex number on the straight line     */
        int  v1;      /*  the last  vertex number on the straight line     */
        int  no;      /*  the number of vertices on the line except v0, v1 */
       }     straight[MXLINE];

int  vtx_straight[20][MXLINE];

int  firstv_straightn[MXVTX];     /* the number of appearances vtx at firstv*/
int  firstv_straight [10][MXVTX]; 
int  endv_straightn  [MXVTX];     /* the number of appearances vtx at endv  */
int  endv_straight   [10][MXVTX];
int  interv_straightn[MXVTX];     /* the number of appearances vtx at interv*/
int  interv_straight [10][MXVTX];
int  flagv_straight  [MXVTX];
                                                                            /*
------------------------------------------------------------------------------
      global variable list
------------------------------------------------------------------------------
                                                                            */
long addr_graph[5][MXGRF];
                                                                            /*
    addr_graph[0][i] : graph number
    addr_graph[1][i] : pointer of this graph to the temp file
    addr_graph[2][i] : the number of vertecies in the graph
    addr_graph[3][i] : the array size of vtx_no[] and intline_no[]
    addr_graph[4][i] : number of internal and external lines
                                                                            */
int level[2][MXGRF];
int no_graph, graph_no;
int no_initial, no_final;
int vtx_table[4][MXEXTV];
                                                                            /*
    vtx_table[0][i] : vertex number
    vtx_table[1][i] : flag = ( 1/ 2/ 3 ) 
		      = ( internal/ final/ vtx connected to radiative photon)
    vtx_table[2][i] : cluster ID # to which the vertex belongs.
    vtx_table[3][i] : pointer of the vertex number "i" in this table.
                                                                            */
int nvtx_iniline;            /* The number of vertecies on the initial line */
int no_iniline;              /* The number of vertecies on the initial line */
int nvertex;
int no_vertex, no_line;
int cluster_table[MXEXTV], no_cluster; /* cluster table and no. of clusters */

int no_inivtx[2], no_photon[2];
int initial_vtx[2][10], IR_vtx[2][10], initial_edge[2];

int count;
int vtxbuf[100];
                                                                            
/*****************************************************************************
                     ===========================                            */
                       void clear_vtxtable( )
/*                   ===========================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
******************************************************************************
                                                                            */
{
int address, n;

for( address = 0; address < MXEXTV; address++ )
   {
     for( n = 0; n < 4; n++ )
          vtx_table[n][address] = -1;
    }
}
                                                                           
/*****************************************************************************
       =============================================================        */
         void add_vtxtable( initial, v, kind, flag, clusterID )
/*     =============================================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
******************************************************************************
                                                                            */
int initial, v, kind, flag, clusterID;
{
int address;

void shift_vtxtable( );
void vtxtable_claddr ( );

address = vtx_table[3][v];
  /*
   printf(" add_vtxtable  address[%3d ] of v[%3d ]\n", address, v );
  */
if( address == -1 )
  {
    if( initial < 0 )
      {
        address = nvertex;
        nvertex++;
       }
    else if( initial == 1 )
      {
        address = no_iniline;
                                                                            /*
        <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                     */
          shift_vtxtable( address );                                        /*
        <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                     */

        no_iniline++;
        nvtx_iniline++;
       }
    else
      {
        address = nvtx_iniline;
                                                                            /*
        <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                     */
          shift_vtxtable( address );                                        /*
        <<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>                                     */

        nvtx_iniline++;
       }

    vtx_table[0][address] = v;
    vtx_table[1][address] = kind;
    vtx_table[2][address] = clusterID;
                                                                            /*
    <<<<<<<<<<<<>>>>>>>>>>>>>                                               */
       vtxtable_claddr( );                                                  /*
    <<<<<<<<<<<<>>>>>>>>>>>>>                                               */

   }
}

/*****************************************************************************
               ================================                             */
                void shift_vtxtable( address )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

int address;
{
int m, n;

 for( n = nvertex; n > address; n-- )
    {
      for( m = 0; m < 3; m++ )
         {
           vtx_table[m][n] = vtx_table[m][n-1];
          }
     }

 nvertex++;
}

/*****************************************************************************
               ================================                             */
                void chng_vtxtable( v_old,v_new )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

int v_old, v_new;
{
int add;

 add = vtx_table[3][v_old];
 if( vtx_table[0][add] == v_old )
   {
     vtx_table[3][v_old] = -1;
     vtx_table[0][add]   = v_new;
     vtx_table[3][v_new] = add;
    }
}

/*****************************************************************************
               ================================                             */
                void vtxtable_claddr(  )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

{
int n, v;

void list_vtxtable();

 for( n = 0; n < nvertex; n++ )
    {
      vtx_table[3][n] = -1;
     }

 for( n = 0; n < nvertex; n++ )
    {
      v               = vtx_table[0][n];
      vtx_table[3][v] = n;

     }
/*
 list_vtxtable();
*/
}

/*****************************************************************************
               ================================                             */
                void rm_vtxtable( v )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

int v;
{
int k, m, n;

void vtxtable_claddr( );


k = vtx_table[3][v];
  /*
  printf(" rm_vtxtable  v[%3d ] add[%3d ]\n", v, k );
  */
if( k <= no_iniline )   no_iniline--;
if( k <= nvtx_iniline ) nvtx_iniline--;

 for( n = k; n < nvertex-1; n++ )
    {
      for( m = 0; m < 3; m++ )
         {
           vtx_table[m][n] = vtx_table[m][n+1];
          }
     }

 nvertex--;

 for( n = nvertex; n < no_vertex; n++ )
    {
      vtx_table[3][n] = -1;
     }


vtxtable_claddr( );

}

/*****************************************************************************
               ================================                             */
                   int get_vtxno( address )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

int address;
{
int returnf;

returnf = vtx_table[0][address];

return returnf;
}

/*****************************************************************************
               ================================                             */
                int addr_vtxtable( v )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

int v;
{
int returnf;

returnf = vtx_table[3][v];

return returnf;
}

/*****************************************************************************
               ================================                             */
                int get_kind( v )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

int v;
{
int returnf, address;

address = vtx_table[3][v];
if( address >= 0 )
    returnf = vtx_table[1][address];
else
    returnf = -1;

return returnf;
}

/*****************************************************************************
               ================================                             */
                int get_intline( v1, v2 )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

int v1, v2 ;
{
int returnf, i;

 returnf = -1;

 for( i = 0; i < no_line; i++ )
    {
      if(( inlvtx[i].vtxno[0] == v1 && inlvtx[i].vtxno[1] == v2 ) ||
         ( inlvtx[i].vtxno[0] == v2 && inlvtx[i].vtxno[1] == v1 ) )
        {
          returnf = i;
          break;
         }
     }

return returnf;
}

/*****************************************************************************
               ================================                             */
                int set_kind( v, kind )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

int v, kind;
{
int returnf, address;

address = vtx_table[3][v];
if( address >= 0 )
    vtx_table[1][address] = kind;

}

/*****************************************************************************
               ================================                             */
                int get_clID( v )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

int v;
{
int returnf, address;

address = vtx_table[3][v];
if( address >= 0 )
    returnf = vtx_table[2][address];
else
    returnf = -1;

return returnf;
}

/*****************************************************************************
               ================================                             */
                void set_clID( v, clID )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

int v, clID;
{
int returnf, address;

address = vtx_table[3][v];
if( address >= 0 )
    vtx_table[2][address] = clID;

}

/*****************************************************************************
               ================================                             */
                void print_vtxtable( nn, flag )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/
 int nn, flag;
{
int n, nv, address;
                                                                            /*
==============================================================================
  print all of vertex table, if nn < 0
==============================================================================
                                                                            */
if( nn < 0 )
  {
    printf("\n");
    printf("******* List all of vertex table vtx_table *************\n");
    printf("             vn  in/ex/p clustID     add\n");
    printf("     n      [0]     [1]     [2]     ");
    printf("[3]  (      x,      y )\n");
    for( n = 0; n < nvertex; n++ )
       {
         if( n == no_iniline )
             printf("  ------ no_iniline   =%3d -------\n", no_iniline );
         if( n == nvtx_iniline )
             printf("  ------ nvtx_iniline =%3d -------\n", nvtx_iniline );
         nv = vtx_table[0][n];
         printf(" %5d   %5d   %5d   %5d   %5d   ( %6.1f, %6.1f )\n",
                  n, vtx_table[0][n], vtx_table[1][n],
                  vtx_table[2][n], vtx_table[3][n],
                  vtx[nv].x, vtx[nv].y   );
        }
    printf("****** list of cluster_table ***************\n" );
    for( n = 0; n < no_cluster; n++ )
       {
         printf("   cluster ID =%4d  Its root vertex =%4d\n",
                    n, cluster_table[n] );
        }
   }
                                                                            /*
==============================================================================
  print a part of vertex table at the address nn, if nn >= 0 anf flag < 0
==============================================================================
                                                                            */
else if( flag < 0 )
  {
    printf("\n");
    printf("******* content of vertex table at address %d *************\n",
                    nn );
    printf("             vn  in/ex/p clustID     add\n");
    printf("    nn      [0]     [1]     [2]     ");
    printf("[3]  (      x,      y )\n");
    nv = vtx_table[0][nn];
    printf(" %5d   %5d   %5d   %5d   %5d   ( %6.1f, %6.1f )\n",
             nn, vtx_table[0][nn], vtx_table[1][nn],
             vtx_table[2][nn], vtx_table[3][nn],
             vtx[nv].x, vtx[nv].y   );
   }
                                                                            /*
==============================================================================
  print a part of vertex table of the vertex nn, if nn >= 0 anf flag >= 0
==============================================================================
                                                                            */
else
  {
    nv = nn;
    nn = vtx_table[3][nn];
    printf("\n");
    printf("******* content of vertex table for vertex %d *************\n",
                    nv );
    printf("             vn  in/ex/p clustID     add\n");
    printf("    nn      [0]     [1]     [2]     ");
    printf("[3]  (      x,      y )\n");
    nv = vtx_table[0][nn];
    printf(" %5d   %5d   %5d   %5d   %5d   ( %6.1f, %6.1f )\n",
             nn, vtx_table[0][nn], vtx_table[1][nn],
             vtx_table[2][nn], vtx_table[3][nn],
             vtx[nv].x, vtx[nv].y   );

   }
}

/*****************************************************************************
               ================================                             */
                void list_vtxtable(  )
/*             ================================
******************************************************************************
*( Function )                                                                *
*( input )                                                                   *
*( output )                                                                  *
*****************************************************************************/

{
int i, n, v;

 printf("vtx_list1 =");

 for( n = 0; n < nvertex; n++ )
    {
      v               = vtx_table[0][n];
         printf("%3d", v );
     }
 printf("\n");

 printf("vtx_list2 =");
 for( i = 0; i < 50; i++ )
    {
      n = vtx_table[3][i];
      if( n >= 0 )
        {
          v  = vtx_table[0][n];
          printf("%3d", v);
         }
     }
 printf("\n");

 printf("address   =");
 for( i = 0; i < 50; i++ )
    {
      n = vtx_table[3][i];
      if( n >= 0 )
        {
          printf("%3d", n );
         }
     }
 printf("\n");

}

/*****************************************************************************
              =================================================              */
                void  vtx_trace( parentv, currentv, flag )
/*            =================================================
******************************************************************************
*( Function )                                                                *
*     To shift the x values of all vertices by the magnitude of shift        *
*    which are determined as follows:                                        *
*      if shift < 0.0, all vertices with smaller x values than that of vp.   *
*      if shift > 0.0, all vertices with larger x  values than that of vp.   *
*( input )                                                                   *
*    vp     : vertex ID number                                               *
*    shift  : the magnitude of x_shift                                       *
*( output )                                                                  *
*     The vertex position in the graph                                       *
*         structure vtx[].x                                                  *
******************************************************************************
*/

int parentv, currentv, flag;

{

 int n_line, n, v;
 float xx;

 if( parentv == currentv )
     count = 0;
 
 xx = vtx[currentv].x;
      /*
       printf(" from vtx_trace  parentv[%5d]  currentv[%5d]\n",
		parentv, currentv );
       printf("      vtx[currentv].x =%6.1f\n",vtx[currentv].x );
      */
 vtxbuf[count] = currentv;
 count++;

 n_line = vtx[currentv].no_line;
 for( n = 0; n < n_line; n++ )
    {
      v = vtxline[currentv].vtxno[n];
      if( v != parentv && vtx[v].flag >= 0 )
        {
              /*
              printf("      vtx[%2d].x =%6.1f flag =%3d\n",
			    v,vtx[v].x, flag );
              */
          if( flag >= 0 )
            {
              if( vtx[v].x > xx ) vtx_trace( currentv, v, flag );
             }
          else
            {
              if( vtx[v].x < xx ) vtx_trace( currentv, v, flag );
             }
         }
    }
}
                                                                            /*
==============================================================================
==============================================================================
================================
    About the straight lines
================================

 (( Data structure ))

  int  no_straight;       the number of straight lines

  structure straight_line {
       int  flag;         ( 1 / 0 ) = ( changed / no ) 
       int  v0;           the first vertex number on the straight line 
       int  v1;           the last  vertex number on the straight line   
       int  no;           the number of vertices on the line except v0, v1
      }     straight[MXLINE];

   int  vtx_straight[20][MXLINE];
                            list of vertices on the line except for v0 and v1

   int  firstv_straightn[MXVTX];      the number of appearances vtx at firstv
   int  firstv_straight [10][MXVTX];  line# list for vtx j at [][j] as firstv
   int  endv_straightn  [MXVTX];      the number of appearances vtx at endv
   int  endv_straight   [10][MXVTX];  line# list for vtx j at [][j] as endv
   int  interv_straightn[MXVTX];      the number of appearances vtx at firstv
   int  interv_straight [10][MXVTX];  line# list for vtx j at [][j] as interv
   int  flagv_straight  [MXVTX];

==============================================================================
                                                                            */
/*****************************************************************************
                  ==============================                            */
                     void  clear_straight(  )
/*                ==============================
******************************************************************************
*( Function )                                                                *
*      To clear the straight line table                                      *
*                                                                            *
******************************************************************************/
{
 int i;

  no_straight = 0;

  for( i = 0; i < MXVTX; i++ )
     {
       firstv_straightn[i] = 0;
       endv_straightn  [i] = 0;
       interv_straightn[i] = 0;
       flagv_straight  [i] = 0;
      }
 }

/*****************************************************************************
                  ==============================                            */
                     void  reset_straight(  )
/*                ==============================
******************************************************************************
*( Function )                                                                *
*      To reset the flag of all straight lines                               *
*                                                                            *
******************************************************************************/
{
 int i;

  if( no_straight > 0 )
      for( i = 0; i < no_straight; i++ )
           straight[i].flag = 0;
}

/*****************************************************************************
                  ==============================                            */
                    void  create_straight( v0, v1 )
/*                ==============================
******************************************************************************
*( Function )                                                                *
*      To create a new straight line if it is really new one                 *
*                                                                            *
******************************************************************************/
int v0, v1;
{
int vx0, vx1;
int   i, flag;
float x0, x1;

  vx0 = v0;
  vx1 = v1;

       /*
        printf(">> create_straght  v0[%2d] v1[%2d]\n", vx0, vx1 );
       */

  if( v0 >= 0 && v1 >= 0 )
    {
      x0  = vtx[v0].x;
      x1  = vtx[v1].x;

       /*
        printf("   x0[%8.2f] x1[%8.2f]\n", x0, x1 );
       */

      if(  x1 != 0.0  &&  x0 > x1  )
        {
          vx0 = v1;
          vx1 = v0;
         }
     }
  else if( v1 >= 0 )
    {
      vx0 = v1;
      vx1 = v0;
     }

  if( no_straight > 0 )
    {
      flag   = 0;
      for( i = 0; i < no_straight; i++ )
           if( vx0 == straight[i].v0 && vx1 == straight[i].v1 )
             {
               flag = 1;
               i    = no_straight;
              }
      if( flag == 0 )
        {
          straight[no_straight].flag = 0;
          straight[no_straight].v0   = vx0;
          straight[no_straight].v1   = vx1;
          straight[no_straight].no   =  0;

          if( vx0 >= 0 )
            {
              i = firstv_straightn[vx0];
              firstv_straight[i][vx0] = no_straight;
              firstv_straightn[vx0]   = i + 1;
             }
          if( vx1 >= 0 )
            {
              i = endv_straightn[vx1];
              endv_straight[i][vx1] = no_straight;
              endv_straightn[vx1]   = i + 1;
             }
          no_straight++;
         }
     }
  else
    {
      straight[0].flag = 0;
      straight[0].v0   = vx0;
      straight[0].v1   = vx1;
      straight[0].no   =  0;

      if( vx0 >= 0 )
        {
          i = firstv_straightn[vx0];
          firstv_straight[i][vx0] = no_straight;
          firstv_straightn[vx0]   = i + 1;
         }
      if( vx1 >= 0 )
        {
          i = endv_straightn[vx1];
          endv_straight[i][vx1] = no_straight;
          endv_straightn[vx1]   = i + 1;
         }
      no_straight = 1;
     }

       /*
        printf(" + create_straght  v0[%2d] v1[%2d]\n", vx0, vx1 );
        printf("   no_str = %3d  str[%d].no =%3d\n", 
                   no_straight, no_straight-1, straight[no_straight-1].no );
       */
}

/*****************************************************************************
                  ==============================                            */
                    void  addv_straight( v0, v1, v )
/*                ==============================
******************************************************************************
*( Function )                                                                *
*      To add a vertex v as a middle vertex                                  *
*                        on the straight line beginning with v0              *
*                                                                            *
******************************************************************************/
int v0, v1, v;
{
int i, j, n;

if( no_straight > 0 )
  {
    j = -1;
    for( i = 0; i < no_straight; i++ )
         if( v0 == straight[i].v0 && v1 == straight[i].v1 )
           {
             j   = i;
             i   = no_straight;
            }
    if( j >= 0 )
      {
        n   = straight[j].no;
        vtx_straight[n][j] = v;
        n++;
        straight[j].no     = n;
        if( v >= 0 )
          {
            i = interv_straightn[v];
            interv_straight[i][v] = j;
            interv_straightn[v]   = i + 1;
           }
       }
       /*
        printf(">> add_straight    v0[%2d] v1[%2d] v[%2d]\n", v0, v1, v );
        printf("   no_str = %3d  str[%d].no =%3d\n", 
                   no_straight, no_straight-1, straight[no_straight-1].no );
       */
   }
}

/*****************************************************************************
                  ==============================                            */
                    void  addve_straight( v0, v1, v )
/*                ==============================
******************************************************************************
*( Function )                                                                *
*      To add a vertex v as an end_vertex v1                                 *
*                        on the straight line beginning with v0              *
*                                                                            *
******************************************************************************/
int v0, v1, v;
{
int i, j, n, vj, nstr, flag;

if( no_straight > 0 )
  {
    j = -1;
    for( i = 0; i < no_straight; i++ )
         if( v0 == straight[i].v0 && v1 == straight[i].v1 )
           {
             j   = i;
             i   = no_straight;
            }
    if( j >= 0 )
      {
        vj = straight[j].v1;
        if( vj >= 0 && vj == v1 )
          {
            n   = straight[j].no;
            vtx_straight[n][j] = vj;
            straight[j].v1     = v;
            n++;
            straight[j].no     = n;

            i = interv_straightn[v1];
            interv_straight[i][v1] = j;
            interv_straightn[v1]   = i + 1;

            i = endv_straightn[v];
            endv_straight[i][v] = j;
            endv_straightn[v]   = i + 1;

            flag   = 0;
            nstr   = endv_straightn[vj] - 1;
            for( i = 0; i < nstr; i++ )
               {
                 n = endv_straight[i][vj];
                 if( n    == j ) flag = 1;
                 if( flag == 1 )
                     endv_straight[i][vj] = endv_straight[i+1][vj];
                }
            endv_straightn[vj] = nstr;
           }
         else if( vj < 0 )
          {
            straight[j].v1     = v;
        
            i = endv_straightn[v];
            endv_straight[i][v] = j;
            endv_straightn[v]   = i + 1;
           }
       }
       /*
        printf(">> addve_straight   v0[%2d] v1[%2d] v[%2d]\n", v0, v1, v );
        printf("   no_str = %3d  str[%d].no =%3d\n", 
                   no_straight, no_straight-1, straight[no_straight-1].no );
       */
   }
}

/*****************************************************************************
                  ==============================                            */
                    int  find_straight(  v1  )
/*                ==============================
******************************************************************************
*( Function )                                                                *
*      To add a vertex v as an end_vertex v1                                 *
*                        on the straight line beginning with v0              *
*                                                                            *
******************************************************************************/
int  v1;
{
int i, returnf;

returnf = -1;

if( no_straight > 0 )
  {
    for( i = 0; i < no_straight; i++ )
         if( v1 == straight[i].v1 )
           {
             returnf   = straight[i].v0;
             i         = no_straight;
            }
   }
return returnf;
}

/*****************************************************************************
                  ==================================                        */
                    void  put_straight( v0, v1, vv )
/*                ==================================
******************************************************************************
*( Function )                                                                *
*      To add a vertex v between vertices v0 and v1, where                   *
*           vertices v0 and v1 must be on a straight line.                   *
*                                                                            *
******************************************************************************/
int  v0, v1, vv;
{
int i0, j0, n0;
int i1, j1, n1;
int  i,  n, nn, v;
int flag;

flag = 0;

if( no_straight > 0 )
  {                                                                         /*
 ============================================================================
 (I) Find staight line with v0 and v1
 ============================================================================
    +------------------------------------------------------------------------+
    |(1) v0 is the first vertex ?                                            |
    +------------------------------------------------------------------------+
                                                                            */
    n0 = firstv_straightn[v0];
    for( i0 = 0; i0 < n0; i0++ )
       {
         j0 = firstv_straight[i0][v0];
                                                                            /*
         +-------------------------------------------------------------------+
         |(1-1) v1 is the end vertex ?                                       |
         +-------------------------------------------------------------------+
                                                                            */
         n1 = endv_straightn[v1];
         for( i1 = 0; i1 < n1; i1++ )
            {
              j1 = endv_straight[i1][v1];
              if( j0 == j1 )
                {
                  flag = 1;
                  i0   = n0;
                  i1   = n1;

                  straight[j1].no     = 1;
                  vtx_straight[0][j1] = vv;

                  nn                       = interv_straightn[vv];
                  interv_straight[nn][vv]  = j1;
                  interv_straightn[vv]     = nn + 1;
                 }
             }
         if( flag == 0 )
           {
                                                                            /*
         +-------------------------------------------------------------------+
         |(1-2) v1 is the internal vertex ?                                  |
         +-------------------------------------------------------------------+
                                                                            */
             n1 = interv_straightn[v1];
             for( i1 = 0; i1 < n1; i1++ )
                {
                  j1 = interv_straight[i1][v1];
                  if( j0 == j1 )
                    {
                      flag = 1;
                      i0   = n0;
                      i1   = n1;

                      n    = straight[j1].no;
                      for( i = n; i > 0; i-- )
                           vtx_straight[i][j1] 
                                          = vtx_straight[i-1][j1];
                      vtx_straight[0][j1] = vv;
                      straight[j1].no     = n + 1;

                      nn                       = interv_straightn[vv];
                      interv_straight[nn][vv]  = j1;
                      interv_straightn[vv]     = nn + 1;
                     }
                 }
            }
        }
                                                                            /*
    +------------------------------------------------------------------------+
    |(2) v0 is the internal vertex ?                                         |
    +------------------------------------------------------------------------+
                                                                            */
    if( flag == 0 )
      {
        n0 = interv_straightn[v0];
        for( i0 = 0; i0 < n0; i0++ )
           {
             j0 = interv_straight[i0][v0];
                                                                            /*
         +-------------------------------------------------------------------+
         |(2-1) v1 is the end vertex ?                                       |
         +-------------------------------------------------------------------+
                                                                            */
             n1 = endv_straightn[v1];
             for( i1 = 0; i1 < n1; i1++ )
                {
                  j1 = endv_straight[i1][v1];
                  if( j0 == j1 )
                    {
                      flag = 1;
                      i0   = n0;
                      i1   = n1;

                      n    = straight[j1].no;
                      vtx_straight[n][j1] = vv;
                      straight[j1].no     = n + 1;

                      nn                       = interv_straightn[vv];
                      interv_straight[nn][vv]  = j1;
                      interv_straightn[vv]     = nn + 1;
                     }
                 }
             if( flag == 0 )
               {
                                                                            /*
         +-------------------------------------------------------------------+
         |(2-2) v1 is also the internal vertex ?                             |
         +-------------------------------------------------------------------+
                                                                            */
                 n1 = interv_straightn[v1];
                 for( i1 = 0; i1 < n1; i1++ )
                    {
                      j1 = interv_straight[i1][v1];
                      if( j0 == j1 )
                        {
                          flag = 1;
                          i0   = n0;
                          i1   = n1;

                          nn   = straight[j1].no;
                          n    = 0;
                          for( i = 0; i < nn; i++ )
                             {
                               v  = vtx_straight[i][j1];
                               if( v == v0 || v == v1 )
                                 {
                                   if( i > n ) n = i;
                                  }
                              }
                          for( i = nn-1; i >= n; i-- )
                             {
                               vtx_straight[i+1][j1] = vtx_straight[i][j1];
                              }
                          vtx_straight[n][j1] = vv;
                          straight[j1].no     = nn + 1;

                          nn                       = interv_straightn[vv];
                          interv_straight[nn][vv]  = j1;
                          interv_straightn[vv]     = nn + 1;
                         }
                     }
                 if( flag == 0 )
                   {
                                                                            /*
         +-------------------------------------------------------------------+
         |(2-3) v1 is the first vertex ?                                     |
         +-------------------------------------------------------------------+
                                                                            */
                     n1 = firstv_straightn[v1];
                     for( i1 = 0; i1 < n1; i1++ )
                        {
                          j1 = firstv_straight[i1][v1];
                          if( j0 == j1 )
                            {
                              flag = 1;
                              i0   = n0;
                              i1   = n1;

                              n    = straight[j1].no;
                              for( i = n; i > 0; i-- )
                                   vtx_straight[i0][j1] 
                                          = vtx_straight[i0-1][j1];
                              vtx_straight[0][j1] = vv;
                              straight[j1].no     = n + 1;

                              nn                      = interv_straightn[vv];
                              interv_straight[nn][vv] = j1;
                              interv_straightn[vv]    = nn + 1;
                             }
                         }
                    }
                }

            }
        if( flag == 0 )
          {
                                                                            /*
    +------------------------------------------------------------------------+
    |(3) v0 is the end vertex ?                                              |
    +------------------------------------------------------------------------+
                                                                            */
            n0 = endv_straightn[v0];
            for( i0 = 0; i0 < n0; i0++ )
               {
                 j0 = endv_straight[i0][v0];
                                                                            /*
         +-------------------------------------------------------------------+
         |(3-1) v1 is the first vertex ?                                     |
         +-------------------------------------------------------------------+
                                                                            */
                 n1 = firstv_straightn[v1];
                 for( i1 = 0; i1 < n1; i1++ )
                    {
                      j1 = firstv_straight[i1][v1];
                      if( j0 == j1 )
                        {
                          flag = 1;
                          i0   = n0;
                          i1   = n1;

                          straight[j1].no     = 1;
                          vtx_straight[0][j1] = vv;

                          nn                       = interv_straightn[vv];
                          interv_straight[nn][vv]  = j1;
                          interv_straightn[vv]     = nn + 1;
                         }
                     }
                 if( flag == 0 )
                   {
                                                                            /*
         +-------------------------------------------------------------------+
         |(3-2) v1 is the internal vertex ?                                  |
         +-------------------------------------------------------------------+
                                                                            */
                     n1 = interv_straightn[v1];
                     for( i1 = 0; i1 < n1; i1++ )
                        {
                          j1 = interv_straight[i1][v1];
                          if( j0 == j1 )
                            {
                              flag = 1;
                              i0   = n0;
                              i1   = n1;

                              n    = straight[j1].no;
                              vtx_straight[n][j1] = vv;
                              straight[j1].no     = n + 1;

                              nn                      = interv_straightn[vv];
                              interv_straight[nn][vv] = j1;
                              interv_straightn[vv]    = nn + 1;
                             }
                         }
                    }
               }
           }
       }
       /*
        printf(">> put_straight     v0[%2d] v1[%2d] vv[%2d]\n", v0, v1, vv );
        printf("   no_str = %3d  str[%d].no =%3d\n", 
                   no_straight, no_straight-1, straight[no_straight-1].no );
       */
   }
}

/*****************************************************************************
                  ==================================                        */
                   void  reverse_straight( v0, v1 )
/*                ==================================
******************************************************************************
*( Function )                                                                *
*      To reverse the order of vertices on the straight line                 *
*             from v0 to v1                                                  *
*                                                                            *
******************************************************************************/
int  v0, v1;
{
int i, j, n, nv, tmpv[20], flag,  nstr;

if( no_straight > 0 )
  {
    for( i = 0; i < no_straight; i++ )
         if( v0 == straight[i].v0 && v1 == straight[i].v1 )
           {
             straight[i].v0 = v1;
             straight[i].v1 = v0;
             nv             = straight[i].no;
             for( n = 0; n < nv; n++ )
                  tmpv[n] = vtx_straight[nv-n-1][i];
             for( n = 0; n < nv; n++ )
                  vtx_straight[n][i] = tmpv[n];

             flag   = 0;
             nstr   = endv_straightn[v1] - 1;
             for( j = 0; j < nstr; j++ )
                {
                  n = endv_straight[j][v1];
                  if( n    == i ) flag = 1;
                  if( flag == 1 )
                      endv_straight[j][v1] = endv_straight[j+1][v1];
                 }
             endv_straightn[v1] = nstr;
             nstr   = firstv_straightn[v1];
             firstv_straight[nstr][v1] = i;
             firstv_straightn[v1]      = nstr + 1;

             flag   = 0;
             nstr   = firstv_straightn[v0] - 1;
             for( j = 0; j < nstr; j++ )
                {
                  n = firstv_straight[j][v0];
                  if( n    == i ) flag = 1;
                  if( flag == 1 )
                      firstv_straight[j][v0] = firstv_straight[j+1][v0];
                 }
             firstv_straightn[v0] = nstr;
             nstr   = endv_straightn[v0];
             endv_straight[nstr][v0] = i;
             endv_straightn[v0]      = nstr + 1;

             i         = no_straight;
            }
   }
}

/*****************************************************************************
                   ==============================                           */
                      void move_others( vp, shift )
/*                 ==============================
******************************************************************************
*( Function )                                                                *
*     To shift the x values of all vertices by the magnitude of shift        *
*    which are determined as follows:                                        *
*      if shift < 0.0, all vertices with smaller x values than that of vp.   *
*      if shift > 0.0, all vertices with larger x  values than that of vp.   *
*( input )                                                                   *
*    vp     : vertex ID number                                               *
*    shift  : the magnitude of x_shift                                       *
*( output )                                                                  *
*     The vertex position in the graph                                       *
*         structure vtx[].x                                                  *
******************************************************************************
*/
int   vp;
float shift;
{
                                                                            /*
-------------------------
   Local variables
-------------------------
                                                                            */
float x0, y0, x1, y1, a, b, x, y, xx;
int   i,   j,  n, nv, kind;
int    v, v0, v1;
                                                                            /*
-------------------------
   Local variables
-------------------------
                                                                            */
int get_kind( );
                                                                            /*
=====================  entry point of shift_others============================

  +--------------------------------------------------------------------------+
  |(1)  Initialization                                                       |
  +--------------------------------------------------------------------------+
                                                                            */
  for( i = 0; i < MXEXTV; i++ )
       flagv_straight[i] = 0;

  xx = vtx[vp].x;

          /*
            printf("  >> From move_others\n");
            printf("       vp[%3d].x (%8.1f)  shift =%8.2f\n", 
                           vp, xx, shift );
          */
                                                                            /*
  +--------------------------------------------------------------------------+
  |(2)  If the internal vertices on each straight line                       |
  +--------------------------------------------------------------------------+
                                                                            */

  for( i = 0; i < no_straight; i++ )
     {
       v0   = straight[i].v0;
       v1   = straight[i].v1;
       kind = get_kind( v0 );
       x    = vtx[v0].x;
       nv   = straight[i].no;

            /*
             printf("       straight_line(%2d)  v0(%3d)  v1(%3d)",
                            i, v0, v1 );
             printf("  nv(%3d)  kind(%3d)\n", nv, kind );
            */

       if( kind == 3 && nv == 0 )
         {
           if( ( shift > 0.0 && x > xx + 5.0 ) ||
               ( shift < 0.0 && x < xx - 5.0 ) )
               {
                 if( flagv_straight[v0] == 0 )
                   {
                     vtx[v0].x  += shift;
                     flagv_straight[v0] = 1;
                    }
                 if( flagv_straight[v1] == 0 )
                   {
                     vtx[v1].x  += shift;
                     flagv_straight[v1] = 1;
                    }
                }
          }
       else
         {
           for( j = 0; j < nv + 2; j++ )
              {
                if( j < nv )        v = vtx_straight[j][i];
                else if ( j == nv ) v = v0;
                else                v = v1;

                x = vtx[v].x;
                if( x != 0.0 )
                  {
                    if( ( shift > 0.0 && x > xx + 5.0 ) ||
                        ( shift < 0.0 && x < xx - 5.0 ) )
                        {
                          if( flagv_straight[v] == 0 )
                            {
                              flagv_straight[v] = 1;

                             /*
                              printf("     * vtx[%3d].x was shfted (%8.1f ==>",
                                             v, x );
                             */

                              vtx[v].x += shift;

                             /*
                              printf("%8.1f)\n", vtx[v].x );
                             */
                             }
                         }
                   }
               }
          }

       if( nv > 0 )
         { 

           x0 = vtx[v0].x,             x1 = vtx[v1].x;
           y0 = vtx[v0].y,             y1 = vtx[v1].y;

           if( x0 != x1 )
             {
               a = ( y1 - y0 )/( x1 - x0 );
               b = y0 - a*x0;
               for( n = 0; n < nv; n++ )
                  {
                    v        = vtx_straight[n][i];

                        /*
                         printf("       vtx[%3d].x is         (%8.1f)\n",
                                        v, vtx[v].x );
                         printf("    *  vtx[%3d].y was shfted (%8.1f ==>",
                                        v, vtx[v].y );
                        */

                    x        = vtx[v].x;
                    vtx[v].y = a*x + b;

                        /*
                         printf("%8.1f)\n", vtx[v].y );
                        */
                   }
              }
          } 
      }
                                                                   /* Jan '98
=============================================================================
    Move the middle points of internal lines
=============================================================================
                                                                           */
  for( i = 0; i < no_line; i++ )
     {
       x = intline[i].x;

       if( x != 0.0 )
         {
           if( ( shift > 0.0 && x > xx + 5.0 ) ||
               ( shift < 0.0 && x < xx - 5.0 ) )
             {
               intline[i].x = x + shift;
              }
          }
      }
       /*
        printf("   < End move_others\n");
       */
}

/*****************************************************************************
                  ===========================================               */
                   void  chngv_straight( v0, v1, vold, vnew )
/*                ===========================================
******************************************************************************
*( Function )                                                                *
*      To print all information of the straight lines                        *
*                                                                            *
******************************************************************************/
int v0, v1, vold, vnew;
{
int  i,  j, nv;
int  n, nn, v;
int flag, no_st1, no_st2;

int chck_straight( );
                                                                             /*
=========================== Entry point of chngv_straight ====================
                                                                             */
if( no_straight > 0 )
  {
       /*
        printf(">> chngv_straight  v0[%2d] v1[%2d] Vold[%2d] Vnew[%2d}\n", 
                   v0, v1, vold, vnew );
       */

                                                                             /*
==============================================================================
  I   check if there is the straight line with vertices v0 and v1.
==============================================================================
                                                                             */
    if( v0 >= 0 )
      {
        flag = chck_straight( v0, v1 );

        if( flag >= 0 )
          {
            no_st1 = flag;
            no_st2 = no_st1 + 1;
                  /*
                   printf("   no_str = %3d  str[%d].no =%3d\n", 
                              no_straight, flag, straight[flag].no );
                  */
           }
        else
          {
            no_st1 = 0;
            no_st2 = no_st1;
           }
       }
    else if( v0 < 0 )
      {
        flag   = -1;
        no_st1 = 0;
        no_st2 = no_straight;
       }
                                                                             /*
==============================================================================
 (I)  replace vold by vnew for straight lines from ID = no_st1 till no_str2
==============================================================================
                                                                             */
    if( flag > -10 )
      {
        for( i = no_st1; i < no_st2; i++ )
           {
             if( straight[i].v0 == vold ) straight[i].v0 = vnew;
             if( straight[i].v1 == vold ) straight[i].v1 = vnew;

             nv = straight[i].no;
             for( j = 0; j < nv; j++ )
                {
                  if( vtx_straight[j][i] == vold ) vtx_straight[j][i] = vnew;
                 }
            }

        nv = firstv_straightn[vold];
        if( nv > 0 )
          {
            if( flag == -1 )
              {
                for( i = 0; i < nv; i++ )
                     firstv_straight[i][vnew] = firstv_straight[i][vold];
                firstv_straightn[vold] = 0;
                firstv_straightn[vnew] = nv;
               }
            else if( flag >= 0 )
              {
                nn  = -1;
                for( i = 0; i < nv; i++ )
                   {
                     if( firstv_straight[i][vold] == flag )
                       {
                         nn  = i;
                        }
                    }
                if( nn >= 0 )
                  {
                    nv--;
                    for( i = nn; i < nv; i++ )
                         firstv_straight[i][vold] = firstv_straight[i+1][vold];
                    firstv_straightn[vold] = nv;

                    n  = firstv_straightn[vnew];
                    firstv_straight[n][vnew] = flag;
                    firstv_straightn[vnew]   = n + 1;
                   }
               }
           }

        nv = endv_straightn[vold];
        if( nv > 0 )
          {
            if( flag == -1 )
              {
                for( i = 0; i < nv; i++ )
                     endv_straight[i][vnew] = endv_straight[i][vold];
                endv_straightn[vold] = 0;
                endv_straightn[vnew] = nv;
               }
            else if( flag >= 0 )
              {
                nn  = -1;
                for( i = 0; i < nv; i++ )
                   {
                     if( endv_straight[i][vold] == flag )
                       {
                         nn  = i;
                        }
                    }
                if( nn >= 0 )
                  {
                    nv--;
                    for( i = nn; i < nv; i++ )
                         endv_straight[i][vold] = endv_straight[i+1][vold];
                    endv_straightn[vold] = nv;

                    n  = endv_straightn[vnew];
                    endv_straight[n][vnew] = flag;
                    endv_straightn[vnew]   = n + 1;
                   }
               }
           }

        nv = interv_straightn[vold];
        if( nv > 0 )
          {
            if( flag == -1 )
              {
                for( i = 0; i < nv; i++ )
                     interv_straight[i][vnew] = interv_straight[i][vold];
                interv_straightn[vold] = 0;
                interv_straightn[vnew] = nv;
               }
            else if( flag >= 0 )
              {
                nn  = -1;
                for( i = 0; i < nv; i++ )
                   {
                     if( interv_straight[i][vold] == flag )
                       {
                         nn  = i;
                        }
                    }
                if( nn >= 0 )
                  {
                    nv--;
                    for( i = nn; i < nv; i++ )
                         interv_straight[i][vold] = interv_straight[i+1][vold];
                    interv_straightn[vold] = nv;

                    n  = interv_straightn[vnew];
                    interv_straight[n][vnew] = flag;
                    interv_straightn[vnew]   = n + 1;
                   }
               }
           }
       }
   }
}

/*****************************************************************************
                  ===========================================               */
                   void  changey_straight( v0, ynew, clid )
/*                ===========================================
******************************************************************************
*( Function )                                                                *
*      To print all information of the straight lines                        *
*                                                                            *
******************************************************************************/
int v0, clid;
float  ynew;
{
int  i,  j, nv;
int  n, nn, v, v1;
int flag, no_st1, no_st2;
float x0, y0, x1, y1, dydx, b, xx, yy, yold;
void set_clID();
                                                                             /*
=========================== Entry point of chngv_straight ====================
                                                                             */
 nn = firstv_straightn[v0];
      /*
       printf(">> changey firstv nn =%3d   v0 =%3d start\n", nn, v0 );
      */
 yold   = vtx[v0].y;
 if(  nn > 0 )
   {
     vtx[v0].y = ynew;
     set_clID( v0, clid );

     for( n = 0; n < nn; n++ )
        {
          i  = firstv_straight[n][v0];
          v1 = straight[i].v1;
          nv = straight[i].no;
 
          if( nv > 0 )
            {
              x0 = vtx[v0].x;
              y0 = ynew;

              x1 = vtx[v1].x;
              y1 = vtx[v1].y;

              dydx = ( y1 - y0 )/( x1 - x0 );
              b    =  y0 - dydx*x0;

              for( j = 0; j < nv; j++ )
                 {
                   v  = vtx_straight[j][i];
                   xx = vtx[v].x;
                   yy = dydx*xx + b;
               /*  vtx[v].y = yy;  */
                                                                             /*
                  <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>                         */
                   changey_straight( v, yy, clid );                          /*
                  <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>                         */
                  }
             }
          else if( get_kind( v0 ) == 3 )
            {
               y1 = vtx[v1].y;
               ynew = ynew - yold + y1;
               vtx[v1].y = ynew;
               set_clID( v1, clid );
               /*
               printf("  v[%3d ].x (%6.1f) => (%6.1f)\n", v1, y1, ynew );
               */
             }
         }
    }

 nn = endv_straightn[v0];
      /*
       printf(">> changey endv nn =%3d  v0 =%3d start\n", nn, v0 );
      */
 if(  nn > 0 )
   {
     vtx[v0].y = ynew;
     set_clID( v0, clid );

     for( n = 0; n < nn; n++ )
        {
          i  = endv_straight[n][v0];
          v1 = straight[i].v0;
          nv = straight[i].no;
 
          if( nv > 0 )
            {
              x0 = vtx[v0].x;
              y0 = ynew;

              x1 = vtx[v1].x;
              y1 = vtx[v1].y;

              dydx = ( y1 - y0 )/( x1 - x0 );
              b    =  y0 - dydx*x0;

              for( j = 0; j < nv; j++ )
                 {
                   v  = vtx_straight[j][i];
                   xx = vtx[v].x;
                   yy = dydx*xx + b;
                /* vtx[v].y = yy;  */
                                                                             /*
                  <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>                         */
                   changey_straight( v, yy, clid );                          /*
                  <<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>                         */
                  }
             }
          else if( get_kind( v0 ) == 3 )
            {
               y1 = vtx[v1].y;
               ynew = ynew - yold + y1;
               vtx[v1].y = ynew;
               set_clID( v1, clid );
               /*
               printf("  v[%3d ].x (%6.1f) => (%6.1f)\n", v1, y1, ynew );
               */
             }
         }
    }
      /*
       printf(">> changey v0 =%3d end\n", v0 );
      */
}

/*****************************************************************************
                        ==============================                      */
                         int  chck_straight( v0, v1 )
/*                      ==============================
******************************************************************************
*( Function )                                                                *
*      To check if is there the straight line with v0 and v1 in the list.    *
*          If there is such a straight line, return the line ID.             *
*          Otherwise return -10.                                             * 
*                                                                            *
******************************************************************************/
int v0, v1;
{
int  i,  j, nv;
int i0, j0, n0;
int i1, j1, n1;
int  n, nn, v;
int flag;

void print_straight();
                                                                             /*
=========================== Entry point of chck_straight ====================
                                                                             */
flag = -10;

if( no_straight > 0 )
  {
       /*
        printf(">> chck_straight  v0[%2d] v1[%2d]\n", 
                   v0, v1 );
       */
                                                                             /*
==============================================================================
  I   replace vold by vnew for the straight line with v0 and v1
==============================================================================
                                                                             */
    if( v0 >= 0 && v1 >= 0 )
      {                                                                     /*
     +-----------------------------------------------------------------------+
     |(I) Find straight line with v0 and v1                                  |
     +-----------------------------------------------------------------------+
        +--------------------------------------------------------------------+
        |(1) v0 is the first vertex ?                                        |
        +--------------------------------------------------------------------+
                                                                            */
        n0   = firstv_straightn[v0];
        for( i0 = 0; i0 < n0; i0++ )
           {
             j0 = firstv_straight[i0][v0];
                                                                            /*
             +---------------------------------------------------------------+
             |(1-1) v1 is the end vertex ?                                   |
             +---------------------------------------------------------------+
                                                                            */
             n1 = endv_straightn[v1];
             for( i1 = 0; i1 < n1; i1++ )
                {
                  j1 = endv_straight[i1][v1];
                  if( j0 == j1 )
                    {
                      flag = j1;
                      i0   = n0;
                      i1   = n1;
                     }
                 }
             if( flag == -10 )
               {
                                                                            /*
             +---------------------------------------------------------------+
             |(1-2) v1 is the internal vertex ?                              |
             +---------------------------------------------------------------+
                                                                            */
                 n1 = interv_straightn[v1];
                 for( i1 = 0; i1 < n1; i1++ )
                    {
                      j1 = interv_straight[i1][v1];
                      if( j0 == j1 )
                        {
                          flag = j1;
                          i0   = n0;
                          i1   = n1;
                         }
                     }
                }
            }
                                                                            /*
        +--------------------------------------------------------------------+
        |(2) v0 is the internal vertex ?                                     |
        +--------------------------------------------------------------------+
                                                                            */
        if( flag == -10 )
          {
            n0 = interv_straightn[v0];
            for( i0 = 0; i0 < n0; i0++ )
               {
                 j0 = interv_straight[i0][v0];
                                                                            /*
             +---------------------------------------------------------------+
             |(2-1) v1 is the end vertex ?                                   |
             +---------------------------------------------------------------+
                                                                            */
                 n1 = endv_straightn[v1];
                 for( i1 = 0; i1 < n1; i1++ )
                    {
                      j1 = endv_straight[i1][v1];
                      if( j0 == j1 )
                        {
                          flag = j1;
                          i0   = n0;
                          i1   = n1;
                         }
                     }
                 if( flag == -10 )
                   {
                                                                            /*
             +---------------------------------------------------------------+
             |(2-2) v1 is also the internal vertex ?                         |
             +---------------------------------------------------------------+
                                                                            */
                     n1 = interv_straightn[v1];
                     for( i1 = 0; i1 < n1; i1++ )
                        {
                          j1 = interv_straight[i1][v1];
                          if( j0 == j1 )
                            {
                              flag = j1;
                              i0   = n0;
                              i1   = n1;
                             }
                         }
                     if( flag == -10 )
                       {
                                                                            /*
             +---------------------------------------------------------------+
             |(2-3) v1 is the first vertex ?                                 |
             +---------------------------------------------------------------+
                                                                            */
                         n1 = firstv_straightn[v1];
                         for( i1 = 0; i1 < n1; i1++ )
                            {
                              j1 = firstv_straight[i1][v1];
                              if( j0 == j1 )
                                {
                                  flag = j1;
                                  i0   = n0;
                                  i1   = n1;
                                 }
                             }
                        }
                    }

                }
            if( flag == -10 )
              {
                                                                            /*
        +--------------------------------------------------------------------+
        |(3) v0 is the end vertex ?                                          |
        +--------------------------------------------------------------------+
                                                                            */
                n0 = endv_straightn[v0];
                for( i0 = 0; i0 < n0; i0++ )
                   {
                     j0 = endv_straight[i0][v0];
                                                                            /*
             +---------------------------------------------------------------+
             |(3-1) v1 is the first vertex ?                                 |
             +---------------------------------------------------------------+
                                                                            */
                     n1 = firstv_straightn[v1];
                     for( i1 = 0; i1 < n1; i1++ )
                        {
                          j1 = firstv_straight[i1][v1];
                          if( j0 == j1 )
                            {
                              flag = j1;
                              i0   = n0;
                              i1   = n1;
                             }
                         }
                     if( flag == -10 )
                       {
                                                                            /*
             +---------------------------------------------------------------+
             |(3-2) v1 is the internal vertex ?                              |
             +---------------------------------------------------------------+
                                                                            */
                         n1 = interv_straightn[v1];
                         for( i1 = 0; i1 < n1; i1++ )
                            {
                              j1 = interv_straight[i1][v1];
                              if( j0 == j1 )
                                {
                                  flag = j1;
                                  i0   = n0;
                                  i1   = n1;
                                 }
                             }
                        }
                   }
               }
           }
       }
   }
 /*
  printf(">> chck_straight  line_ID =%3d\n", flag );
  if( flag >= 0 )  print_straight( flag );
 */

return flag;
}

/*****************************************************************************
                  ==============================                            */
                     int getv_straight( j, flag )
/*                ==============================
******************************************************************************
*( Function )                                                                *
*      To get the vertex ID of the j-th straight line                        *
*          flag = 0: return the first vertex                                 *
*          flag = 1; return the end vertex                                   *
*                                                                            *
******************************************************************************/
int j, flag;
{
int returnf;

 returnf = -1;

 if( no_straight > 0 )
   {
     if( flag == 0 ) returnf = straight[j].v0;

     else            returnf = straight[j].v1;
    }
 return returnf;
}

/*****************************************************************************
                  ==============================                            */
                     void  print_straight( line_s )
/*                ==============================
******************************************************************************
*( Function )                                                                *
*      To print all information of the straight lines                        *
*                                                                            *
******************************************************************************/
int line_s;
{
 int i, j, n;

if( no_straight > 0 )
  {
  if( line_s < 0 )
    {
      printf("\n");
      printf("*********** List of straight lines ************************\n");
      for( j = 0; j < no_straight; j++ )
         {
           n = straight[j].no;
           printf("line number[%3d ]  v0[%3d ] ==> v1[%3d ]  no_vtx = %3d\n",
                     j, straight[j].v0, straight[j].v1, n+2 );
           if( n > 0 )
             {
               printf(" internal vertices = ");
               for( i = 0; i < n; i++ )
                    printf(" %3d", vtx_straight[i][j] );
               printf("\n");
              }
          }
      printf("\n");
      printf("*********** Inverse list of straight lines ****************\n");
      for( j = 0; j < no_vertex; j++ )
         {
           printf("vertex number[%3d ]  firstv[%3d]  endv[%3d]  interv[%3d]",
                   j, firstv_straightn[j], 
                      endv_straightn[j], interv_straightn[j] );
           printf("  flagv[%3d]\n", flagv_straight[j] );

           n = firstv_straightn[j];
           if( n > 0 )
             {
               printf("   line ( firstv ) = ");
               for( i = 0; i < n; i++ )
                    printf(" %3d", firstv_straight[i][j] );
               printf("\n");
              }

           n = endv_straightn[j];
           if( n > 0 )
             {
               printf("   line ( endv   ) = ");
               for( i = 0; i < n; i++ )
                    printf(" %3d", endv_straight[i][j] );
               printf("\n");
              }

           n = interv_straightn[j];
           if( n > 0 )
             {
               printf("   line ( interv ) = ");
               for( i = 0; i < n; i++ )
                    printf(" %3d", interv_straight[i][j] );
               printf("\n");
              }
          }
      printf("\n");
     }
  else if( line_s < no_straight )
    {
       j = line_s;
       n = straight[j].no;
       printf("line number[%3d ]  v0[%3d ] ==> v1[%3d ]  no_vtx = %3d\n",
                    j, straight[j].v0, straight[j].v1, n+2 );
       if( n > 0 )
         {
           printf(" internal vertices = ");
           for( i = 0; i < n; i++ )
                printf(" %3d", vtx_straight[i][j] );
           printf("\n");
          }
     }
   }
else
  {
      printf(" **** no striaght line is registered !! *******\n" );
   }
}
