/***************************************************************
 *  File : ftgrf.c  prefix=ft
 *
 *  Functions for generating fortran source code.
 *
 *                                                 by Kaneko Toshiaki
 *
 *  1995/08/29 TI: Color Base;"CF"
 *  1995/07/13 TI:Supress generation of "smcpol.f".
 *             (add this subrouine in CHANEL library)
 *  1995/05/30 add "cmfin.c".  See this file in detail
 *             ftincl should be called before cmbases
 *             ftmakes should be called after cmbases
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "grdata.h"
#include "putid.h"
#include "config.h"
#include "gensc.h"
#include "cfnbase.h"
#include "cfnconv.h"
#include "grcfort.h"

/*------------------------------------------------------------------*
 * Parameters
 */
/*
#define SUSYDBG
*/
/* generation of smssvv or smsvsv for CVertex (SVSV)
 *
 *    VLREORD   0   old version (no SVSV)
 *    VLREORD   1   generate as smssvv
 *    VLREORD   2   generate as smsvsv
 */
#define VLREORD 2
#ifndef VLREORD
#define VLREORD 0
#endif

#define SMCPOL
#define CINLINE
/*------------------------------------------------------------------*
 * local types and variables
 */
Local FILE   *tfp1 = NULL;

#define  MAXSSTR  256
Local char   sizestr[MAXSSTR];

#define NVARS  100
#define FT_INT     1
#define FT_REAL    2
#define FT_COMPLEX 3
#define FT_INTMAX  4
#define FT_REALMAX 5
#define FT_MAX     6
Local long vartype[FT_MAX];

#define MAXVLTBL    (MAXNODES*2)

Local struct {          /* table of local variables in fortran */
        int    vtype;
        string name;
        string dim;
        int    size;
    } vartbl[NVARS];
Local int nvars = 0;
Local int incl2r = 0;
Local int incl2i = 0;

typedef struct {        /* leg of a node */
        pindex p;       /* particle */
        int    t;       /* type of i-line */
        int    i;       /* internal line */
        int    m;       /* momentum number */
        int    d;       /* direction of momentum */
    } vlegt;

Local struct {
        vlegt **vl;
        int   size;
    } vltbl[MAXVLTBL];
Local int nvltbl = 0, nvlusd = 0;

/* Unified code generation for multi processes 981009 */
Local int nummultip=0;
Local struct {
	int	nextern;	/* Number of External particles */
	pindex	pa[MAXELINE];	/* Particle code of External particles */
	int	iorf[MAXELINE];	/* Initilal/Final of External particles */
	int	procg;		/* Number of Graph */
	int	cbase;		/* Number of Cbases */
        int	lag;		/* Lag */
} uniproc[MAXPROCS];
#define csign(l)   ((l)<0 ? '-':'+')

#define EOFSYM     '\0'

Local   int  vsign;   /* global sign form susy wave function */

Local   bool majorana;
/* S-chanel graph(1) or not (0) */
Local   bool isschan;

Local   FILE *tifp = NULL;
Local   FILE *unfp = NULL;

/* local functions */
Private Void    ftmain   ARG((int));
Private Void    ftamptbp ARG((int));
Private Void    ftgtree  ARG((NoArg));
Private Void    ftdenom  ARG((NoArg));
Private Void    ftincl   ARG((NoArg));
Private Void    ftinfo   ARG((NoArg));
Private Void    ftmuldb  ARG((NoArg)); /* Shinsaku */
Private Void    ftamporp ARG((int));
Private Void    ftampsup ARG((int));
Private Void    ftampsmp ARG((int));
Private int     ftgmask  ARG((NoArg));

Private Void    ftmffv   ARG((node,vlegt**));
Private Void    ftmffs   ARG((node,vlegt**));
Private Void    ftnffv   ARG((string,node,vlegt**));
Private Void    ftnffs   ARG((string,node,vlegt**));
Private Void    ftnvvv   ARG((string,string,node,vlegt**));
Private Void    ftnsvv   ARG((string,node,vlegt**));
Private Void    ftnssv   ARG((string,node,vlegt**));
Private Void    ftnsss   ARG((string,node,vlegt**));
Private Void    ftnvvvv  ARG((string,string,node,vlegt**));
Private Void    ftnssvv  ARG((string,node,vlegt**));
#if (VLREORD==2)
Private Void    ftnsvsv  ARG((string,node,vlegt**));
#endif
Private Void    ftnssss  ARG((string,node,vlegt**));
Private Void    ftvlts   ARG((NoArg));
Private Void    ftvlte   ARG((NoArg));
Private Void    ftvlcp   ARG((vlegt*,vlegt*));
Private Void    ftvlex   ARG((vlegt**,int,int));
Private Void    ftclpord ARG((NoArg));
Private Void    ftcfamp  ARG((NoArg));
Private Void    ftdprop  ARG((NoArg));
Private Void    ftvars   ARG((int));

Private Void    fthindx  ARG((node,vlegt**,int));
Private Void    ftncoupl ARG((node));
Private Void    ftmassl  ARG((node,vlegt**,int,int));
Private Void    ftcrans  ARG((node,vlegt**,int));
Private Void    ftmlist  ARG((node,vlegt**,int));

Private Void    ftbegin  ARG((NoArg));
Private Void    ftmconsv ARG((NoArg));
Private Void    ftnode   ARG((NoArg));
Private Void    ftconn   ARG((NoArg));
Private Void    ftiline  ARG((NoArg));
Private int     ftgauge  ARG((string,pindex));
Private Void    ftgend   ARG((NoArg));

Private Void    fttmpopen ARG((NoArg));
Private Void    fttmpcp   ARG((NoArg));
Private Void    fttmpclr  ARG((NoArg));
Private Void    fttmp     ARG((int,string,string,int));
Private Void    ftincw    ARG((NoArg));
Private Void    ftincpp   ARG((int));
Private Uint    ftnvtype  ARG((string,node,vlegt**));
Private Void    ftscfact  ARG((NoArg));
Private Void    ftacfact  ARG((NoArg));
Private Void    ftexarg   ARG((vlegt*));
Private Void    ftexsubr  ARG((string,node,vlegt**,int));
#ifdef SMCPOL
Private Void    ftsmcpol ARG((NoArg));
#endif
#if DBG
Private Void    fttmdmp   ARG((NoArg));
#endif

Private string  ftmcc     ARG((string, node, int, int*));
Private Void    ftprfl    ARG((FILE*));

#ifdef OLDSEL
/* Number of Internal Higss */
Private int     ftishiggs ARG((NoArg));
/* Number of Internal Gluon */
Private int     ftisgluon ARG((NoArg));
#else
Private Ulong  ftselect   ARG((NoArg));
#endif

/* Check S-chanel or not */
Private bool    ftisschan ARG((NoArg));

#define fparm(y, nx)  (((y)+1)%(nx)+1)
#define ilpartl(l,lg)     \
        ndparticle(iltable[l].ilnodes[lg],iltable[l].illegs[lg])
#define ilpart(l)     abs(ilpartl(l, 0))
#define ndconj(n)     ((n) >= conjstart)

/*********************************************************************
 *    1. Common part.
 */
/*==================================================================*/
Public  Void    ftprocs(proct, procg)
int     proct[], procg[];
/*------------------------------------------------------------------*
 *
 *
 * Arguments
 *    proct  : the table of process numbers.
 *             proct[l] is the process number of l-loop graphs.
 *    procg  : the table of the number of graphs in each process.
 *    dvprt  : include divergent parts or not.
 *------------------------------------------------------------------*/
{
    int       n, pa;
    int       g0;
    int       pc;
    /* CF */
    int       csgn = 1;
    int       ibase = 0, iprt = 0;

    caller  = GRCFORT;
    lang    = ROFORT;

    if(tishika) {
        if((tifp = fopen("ti.dat", "w")) == NULL) {
            fprintf(Stderr, "Cannot open \"ti.dat\"\n");
            exit(FAIL);
        }
    }
    if(ucode) {
        if((unfp = fopen("un.f", "w")) == NULL) {
            exit(FAIL);
        }
    }
    printf("vcode = %d \n",vcode);
    printf("rcode = %d \n",rcode);

    if(maxloop > 0) {
        fprintf(Stderr, "*** This version supports only tree graphs\n");
        fprintf(Stderr, "    Current loop = %d\n", maxloop);
        exit(1);
    }

    /* CF */
    if((gcolgr = (cftgraph *) malloc(sizeof(cftgraph))) == NULL) {
         fprintf(Stderr, "*** gcolgr: memory allocation failed\n");
         exit(1);
    }
    if((gcbase = (cftbase *) malloc(sizeof(cftbase))) == NULL) {
         fprintf(Stderr, "*** gcbase: memory allocation failed\n");
         exit(1);
    }

    for(pc = 0; pc < nprocs; pc++) {
        reproct[pc] = proct[pc];
        reprocg[pc] = procg[pc];
    }
    cmprfxini();

    if(nprocs == 1 && reprocp[0][0] == '\0') {
        multiprc = False;
    } else {
        multiprc = True;
    }

    /* names of graphs are:
     *   a1, a2, ...
     */

    cnprc    = reproct[0];       /* process number */
    cprefix = reprocp[0];
    cprocess = dagetp(cnprc);   /* process data */
    cngraph = 0;
    cgraph  = daread0(cnprc, cngraph);
    ilmom();
    nilines = cgraph->nilines;
    flget(cgraph);

    /* gcbase is used in ftinit */
    if(colorv == 1) {
        cfcbase(gcbase);
        cfbase(gcbase,ibase,iprt);
#if DBG
/*
        printf("call cfbdump:1\n");
        cfbdump(gcbase);
*/
#endif
        if(gcbase->nbase == 1 && gcbase->nexcolor == 0) {
            printf("No Color Particle\n");
        }
    }

    /* gauge parameters */
    cmgauge();

    /* initialization */
    ftinit();


    incl2r = 0;
    incl2i = 0;

    /* repeat by loop */
    for(pc = 0; pc < nprocs; pc++) {

        cprocs  = pc;
        cprefix = reprocp[cprocs];

        /* first process */
        cnprc    = reproct[pc];       /* process number */
        cprocess = dagetp(cnprc);   /* process data */
        nextern  = cprocess->oextern;

        printf("%% +++ process = %d (%d) +++\n", pc, nprocs);

        if(tishika) {
            fprintf(tifp,"Process: ");
            for(n = 0; n < nextern; n++) {
                if(ndiorf(n) == ELINITIAL) {
                    pa  = ndname(n);
                    fprintf(tifp,"%s ",pvename(ndiorf(n),pa));
                }
            }
            fprintf(tifp," --> ");
            for(n = 0; n < nextern; n++) {
                if(ndiorf(n) == ELFINAL) {
                    pa  = ndname(n);
                    fprintf(tifp,"%s ",pvename(ndiorf(n),pa));
                }
            }
            fprintf(tifp,"'\n");
        }

        cngraph = 0;
        cgraph  = daread0(cnprc, cngraph);
        ilmom();                    /* initialize i-line table */
        nilines = cgraph->nilines;
        flget(cgraph);

        /* CF:begin */
        if(colorv == 1) {
            cfcbase(gcbase);
            cfbase(gcbase,ibase,iprt);
#if DBG
/*
            printf("call cfbdump:2\n");
            cfbdump(gcbase);
*/
#endif
            if(gcbase->nbase == 1 && gcbase->nexcolor == 0) {
                printf("No Color Particle\n");
            }
        }
        /* CF:end */

        /* repeat graphs for process pc */
        for(g0 = 0; g0 < reprocg[pc]; g0++) {
            printf("<<GRAPH>> %d\n", g0+1);

/*
            if(g0 != 0) {
*/
                cngraph = g0;
                cgraph  = daread0(cnprc, cngraph);

                /* initialize i-line table */
/*                printf("** ilmon:1 **\n");*/
                ilmom();
                nilines = cgraph->nilines;
                flget(cgraph);
/*
            }
*/ 
            if(colorv == 1) {
                if(!(gcbase->nbase == 1 && gcbase->nexcolor == 0)) {

                    csgn = cfcproj(gcbase,gcolgr);
#if DBG
                    printf("call cfgdump:3\n");
                    cfgdump(gcolgr);
#endif
                    cfproj(gcbase, gcolgr, ibase, iprt);
                }
            }
/*
            if(g0 != 0) {
*/
                cngraph = g0;
                cgraph  = daread(cnprc, cngraph);

                cgraph->sfactor *= csgn;

                /* initialize i-line table */
/*                printf("** ilmon:2 **\n");*/
                ilmom();
                nilines = cgraph->nilines;
                flget(cgraph);
/*
            }
*/ 

            isschan = ftisschan();

            /* construct fermion lines */
            majorana = flmajo();

            /* graph selection name */
            cmselect();

            printf("graph %d\n", cngraph+1);
#if DBG
            if(dbg > 0) {
                ildump();
            }
#endif
/*
            if(isschan) {
                printf(" S-chan \n");
            } else {
                printf(" NON-S-chan \n");
            }
*/
            cprocs = pc;
            cprefix = reprocp[cprocs];
            if(mergefile && cngraph == 0) {
/*                if( (reprocg[pc] - 1)/ACUT == 0 ) {*/
                    sprintf(momstr, "a%smps.f", cprefix);
/*
                } else {
                    sprintf(momstr, "a%smps0.f", cprefix);
                }
*/
                cmopend(ROFORT, cprefix, momstr);
            }

            /* generate fortran code for a graph */
            ftgraph(cprocess, cgraph);

            if(colorv == 1) {
                if(!(gcbase->nbase == 1 && gcbase->nexcolor == 0)) {
                    cfcgfree(gcolgr);
                }
            }
            printf("\n");
        }

        ftincl();
        if(ucode) ftinfo();
        ftincw();
        ftmain(pc);
    }

    /* create makefiles. */
    cprocess = dagetp(reproct[0]);
    ftmakes();

    ftmuldb();

    /* CF */
    if(colorv == 1) {
        cfcbfree(gcbase);
    }
    if(colorv == 0) {
        if(maxloop > 0 || cprocess->extcol > 0) {
           if(cprocess->qcd > 0) {
              ftscfact();
           } else {
              ftacfact();
           }
        }
    }

    daclose();

    if(tishika) {
        fclose(tifp);
        tifp = NULL;
    }
    if(ucode) {
        fclose(unfp);
        unfp = NULL;
    }
}

/*==================================================================*/
Public  Void    ftinit(NoArg)
/*------------------------------------------------------------------*
 * Initialize and generate fortran codes.
 *
 * Arguments:
 *    proc  : the process in which the graph blongs.
 *    graph : the graph to be converted.
 *------------------------------------------------------------------*/
{
    /* setup variables */
    nextern   = cprocess->oextern;
    nilines   = cgraph->nilines;
    conjstart = cgraph->conjsrt;

    momoffs = 0;
    legoffs = 0;

    /* userin and usrout */
    cmmodel();

    cminclk();

#ifdef SMCPOL
    if(cpol) {
        ftsmcpol();
    }
#endif

}

/*==================================================================*/
Private Void    ftincl(NoArg)
/*------------------------------------------------------------------*
 * generate "incl1.f"
 *------------------------------------------------------------------*/
{
    int    p, n0, lag, il, lg, tp, i, nc;
    int    mxlen, sgn, mxl;

    node   nd;
    int    ngrf = reprocg[cprocs];
    string prf  = cprefix;

    sprintf(momstr, "incl%s1.h", prf);
    cmopend(ROFORT, prf, momstr);

    fprintf(redfp,"\n");
    fprintf(redfp,"      include 'inclc.h'\n");
    fprintf(redfp,"      include 'inclg.h'\n\n");

    fprintf(redfp,"      parameter (loutgo =  2, lincom =  1)\n");
    fprintf(redfp,"      parameter (lantip = -1, lprtcl =  1)\n");
    fprintf(redfp,"      parameter (lscalr =  1)\n");
    fprintf(redfp,"      parameter (lepexa =  2, lepexv =  3)\n");
    fprintf(redfp,"      parameter (lepina =  4, lepinv =  4)\n");
    fprintf(redfp,"      parameter (lextrn =  2, lintrn =  4)\n");

    fprintf(redfp,"\n* table of amplitudes\n");
    fprintf(redfp,"      parameter (n%sgraph =%d)\n", prf, ngrf);
    fprintf(redfp,"      parameter (n%sextn  =%d)\n", prf, nextern);
    fprintf(redfp,"\n");

/* move to "inclg.h"
    fprintf(redfp,"* Switch of gauge parameters\n");
    puinit(redfp,"      common /smgaus/",
                  "     &               ",
                  ",", ",", "", 72, 16);
    putid("igaug00");
    for(i = 0; i < ngauge; i++) {
        sprintf(momstr, "igau%s", sgauge[i]);
        putid(momstr);
    }
    putend();

    fprintf(redfp,"      common /smgaug/agauge(0:%d)\n", ngauge);
*/

    /* CF */
    if(colorv == 1) {
        printf("nexcolor %d \n",gcbase->nexcolor);
/*      if(gcbase->nexcolor != 1) gcbase->nexcolor=0;*/
        if(gcbase->nexcolor == 1) {
            fprintf(redfp,"\n* number of color base\n");
            fprintf(redfp,"      parameter (ncbase =%d)\n", gcbase->nbase);
        } else {
            fprintf(redfp,"\n* No color\n");
            fprintf(redfp,"      parameter (ncbase = 1)\n");
        }
    }

    lag = 1;
    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype != ILEXTERN) {
            continue;
        }
        nd  = iltable[il].ilnodes[0];
        lg  = iltable[il].illegs[0];
        p   = abs(ndparticle(nd, lg));
        tp  = iltable[il].iltype;
        lag *= cmindx(momstr1, tp, p);
    }

    fprintf(redfp,"\n* number of all helicity states\n");
    fprintf(redfp,"      parameter (l%sag    =%d)\n\n", prf, lag);
    fprintf(redfp,"      common /amprck/kmngr,  kmnext, kmlag\n");
    fprintf(redfp,"      common /amprcs/smpref, smproc\n");
    fprintf(redfp,"      character*80   smpref, smproc\n\n");

    fprintf(redfp,"      parameter (n%sgrpsq = n%sgraph*n%sgraph)\n",
            prf, prf, prf);
#ifdef OLDSEL
    fprintf(redfp,"      common /a%smslct/jselg(n%sgraph), jgraph, ",
            prf, prf);
    fprintf(redfp,"jhiggs,jgluon\n");
#else
    sprintf(momstr,"      common /a%smslct/", prf);
    puinit(redfp, momstr,
                  "     &               ",
                  ",", ",", "", 72, 16);
    sprintf(momstr, "jselg(n%sgraph)", prf);
    putid(momstr);
    putid("jgraph");
    for(i = 0; i < nselect; i++) {
        sprintf(momstr, "j%s", sselect[i]);
        putid(momstr);
    }
    putend();
#endif

    if(colorv == 1 && gcbase->nbase > 0) {
        fprintf(redfp,"\n* Color string information\n");
        fprintf(redfp,"      common /a%smcsti/ kmcbas, kmcbmx", prf);

        if(gcbase->nbase == 1) {
            if(gcbase->nconid != NULL && gcbase->nconid[0] > 0) {
                nc = gcbase->nconid[0];
                if(nc < 1) {
                    nc = 1;
                }
                fprintf(redfp,", kmcstr(%d)",nc);
            }
        } else if(gcbase->nbase > 1) {
            mxlen = gcbase->nconid[0];
            for(i = 1; i < gcbase->nbase; i++) {
               if(gcbase->nconid[i] > mxlen) {
                   mxlen = gcbase->nconid[i];
               }
            }
            mxl = max(mxlen, 1);
            fprintf(redfp,", kmcstr(%d,0:%d)",mxl,gcbase->nbase-1);
        }
        fprintf(redfp,", icinfo(%d), icolst\n",nextern);
    }

    /* CF */
    if(colorv == 0) {
        if(!vcode) {
            fprintf(redfp,"      common /a%smgrph/ag(0:l%sag-1,n%sgraph), ",
                    prf, prf, prf);
        } else {
            fprintf(redfp,"      common /a%smgrph/ag(n_vector,0:l%sag-1,",
                    prf, prf);
            fprintf(redfp,"n%sgraph), ", prf);
        }
        if(!vcode) {
            fprintf(redfp,"aprop(n%sgraph),\n", prf);
        } else {
            fprintf(redfp,"aprop(n_vector,n%sgraph),\n", prf);
        }
        if(!vcode) {
            fprintf(redfp,"     &               ancp(n%sgraph), ", prf);
        } else {
            fprintf(redfp,"     &               ancp(n_vector,n%sgraph), ", prf);
        }
        if(!vcode) {
            fprintf(redfp,"ansp(0:n%sgraph),\n", prf);
        } else {
            fprintf(redfp,"ansp(n_vector,0:n%sgraph)\n", prf);
        }
        fprintf(redfp,"     &               igraph(n%sgraph)", prf);
        if(cprocess->extcol > 0) {
           fprintf(redfp,",cf(n%sgraph,n%sgraph)", prf, prf);
        }
        fprintf(redfp,"\n");

        fprintf(redfp,"      complex*16 ag, aprop\n");
        fprintf(redfp,"\n");
    } else if(colorv == 1) { /* CURRENT COLOR TREATMENT*/
        if(!vcode) {
            fprintf(redfp,"      common /a%smgrph/agcwrk(0:l%sag-1),",
                    prf, prf);
        } else {
            fprintf(redfp,"      common /a%smgrph/agcwrk(n_vector,", prf);
            fprintf(redfp,"0:l%sag-1),", prf);
        }
/*      if(gcbase->nbase > 1) { */
            if(!vcode) {
                fprintf(redfp,"agc(0:l%sag-1,0:ncbase-1),\n", prf);
            } else {
                fprintf(redfp,"\n     &              ");
                if( rcode ) {
                   fprintf(redfp,"agc(0:l%sag-1,0:ncbase-1,n_vector),\n", prf);
                } else {
                   fprintf(redfp,"agc(n_vector,0:l%sag-1,0:ncbase-1),\n", prf);
                }
            }
/*      } else {
            if(!vcode) {
                fprintf(redfp,"agc(0:l%sag-1),\n", prf);
            } else {
                fprintf(redfp,"agc(n_vector,0:l%sag-1),\n", prf);
            }
        } *//* 980930 */
        if(!vcode) {
            fprintf(redfp,"     &              aprop,");
        } else {
            fprintf(redfp,"     &              aprop(n_vector),");
        }
        if(!vcode) {
            fprintf(redfp,"ancp(n%sgraph),ansp(0:n%sgraph)", prf, prf);
        } else {
            fprintf(redfp,"ancp(n_vector,n%sgraph),ansp(n_vector,0:n%sgraph)",
                    prf, prf);
        }
/*
        if(gcbase->nbase > 1) {
            fprintf(redfp,"\n     &             ");
            fprintf(redfp,",cfmtx(0:ncbase-1,0:ncbase-1)\n");
        } else {
            if(gcbase->nexcolor == 1) {
               fprintf(redfp,"\n     &             ");
               fprintf(redfp,",cfmtx\n");
            } else {
               fprintf(redfp,"\n     &             ");
               fprintf(redfp,",cfmtx\n");
            }
        }
*//* 980929 */
        fprintf(redfp,"\n     &             ");
        fprintf(redfp,",cfmtx(0:ncbase-1,0:ncbase-1)\n");

        fprintf(redfp,"      common /a%smgrpi/igraph(n%sgraph)\n",
                prf, prf);
        fprintf(redfp,"      complex*16 agc, agcwrk, aprop\n");
    }

    /* common /amextr/... */
    fprintf(redfp,"\n");
    fprintf(redfp,"* Momenta of external particles\n");

    sprintf(momstr, "      common /a%smextr/", prf);
    puinit(redfp, momstr,
                  "     &               ",
                  ",", ",", "", 72, 16);
    if(!vcode) {
        for(n0 = 1; n0 <= nextern; n0++) {
            cmecmom(momstr1,n0,1,"pe");
            sprintf(momstr,"%s(4)", momstr1);
            putid(momstr);
        }
        sprintf(momstr, "prod(n%sextn, n%sextn)", prf, prf);
        putid(momstr);
    } else {
        for(n0 = 1; n0 <= nextern; n0++) {
            cmecmom(momstr1,n0,1,"pe");
            if( rcode ) {
              sprintf(momstr,"%s(4,n_vector)", momstr1);
            } else {
              sprintf(momstr,"%s(n_vector,4)", momstr1);
            }
            putid(momstr);
        }
        if( rcode ) {
            sprintf(momstr, "prod(n%sextn, n%sextn, n_vector)", prf, prf);
        } else {
            sprintf(momstr, "prod(n_vector, n%sextn, n%sextn)", prf, prf);
        }
        putid(momstr);
    }
    putend();
    putc(EOL, redfp);

    sprintf(momstr,"      common /a%smextp/", prf);
    puinit(redfp, momstr,
                  "     &               ",
                  ",", ",", "", 72, 16);

    for(nd = 0; nd < ndnodes(); nd++) {
        if(ndtype(nd) != NDEXTERN) {
            continue;
        }
        p = abs(ndparticle(nd, 0));
        sgn = cmisgn(nd+1);
        if(ptable[p]->ptype == P_VECTOR) {
            cmicwgt(momstr, nd+1, sgn, p, "ep");
            putid(momstr);
            cmicwgt(momstr, nd+1, sgn, p, "eu");
            putid(momstr);
        } else if(ptable[p]->ptype == P_DIRAC) {
            cmicwgt(momstr, nd+1, sgn, p, "ps");
            putid(momstr);
            cmicwgt(momstr, nd+1, sgn, p, "ew");
            putid(momstr);
            cmicwgt(momstr, nd+1, sgn, p, "ce");
            putid(momstr);
        } else if(ptable[p]->ptype == P_MAJORANA) {
            cmicwgt(momstr, nd+1, sgn, p, "ps");
            putid(momstr);
            cmicwgt(momstr, nd+1, sgn, p, "ew");
            putid(momstr);
            cmicwgt(momstr, nd+1, sgn, p, "ce");
            putid(momstr);
        }
    }
    putend();
    fprintf(redfp,"\n");
    for(nd = 0; nd < ndnodes(); nd++) {
        if(ndtype(nd) != NDEXTERN) {
            continue;
        }
        p = abs(ndparticle(nd, 0));
        il  = ndiline(nd, 0);
        tp  = iltable[il].iltype;
        Discard cmindx(momstr, tp, p);
        sgn = cmisgn(nd+1);
        if(ptable[p]->ptype == P_VECTOR) {
            cmicwgt(momstr1, nd+1, sgn, p, "ep");
            cmicwgt(momstr2, nd+1, sgn, p, "eu");
            if(!vcode) {
                fprintf(redfp, "      real*8     %s(4,%s), %s(%s)\n",
                        momstr1, momstr, momstr2, momstr);
            } else {
                fprintf(redfp, "      real*8     ");
                if( rcode ) {
                  fprintf(redfp, "%s(4,%s,n_vector), %s(%s,n_vector)\n",
                          momstr1, momstr, momstr2, momstr);
                } else {
                  fprintf(redfp, "%s(n_vector,4,%s), %s(n_vector,%s)\n",
                          momstr1, momstr, momstr2, momstr);
                }
            }
        } else if(ptable[p]->ptype == P_DIRAC) {
            cmicwgt(momstr1, nd+1, sgn, p, "ps");
            cmicwgt(momstr2, nd+1, sgn, p, "ew");
            if(!vcode) {
                fprintf(redfp, "      real*8     %s(4,2), %s(1)\n",
                        momstr1, momstr2);
            } else {
                fprintf(redfp, "      real*8     ");
                if( rcode ) {
                  fprintf(redfp, "%s(4,2,n_vector), %s(1,n_vector)\n",
                          momstr1, momstr2);
                } else {
                  fprintf(redfp, "%s(n_vector,4,2), %s(n_vector,1)\n",
                          momstr1, momstr2);
                }
            }
            cmicwgt(momstr1, nd+1, sgn, p, "ce");
            if(!vcode) {
                fprintf(redfp, "      complex*16 %s(2,2)\n", momstr1);
            } else {
                if( rcode ) {
                  fprintf(redfp, "      complex*16 %s(2,2,n_vector)\n", momstr1);
                } else {
                  fprintf(redfp, "      complex*16 %s(n_vector,2,2)\n", momstr1);
                }
            }
        } else if(ptable[p]->ptype == P_MAJORANA) {
            cmicwgt(momstr1, nd+1, sgn, p, "ps");
            cmicwgt(momstr2, nd+1, sgn, p, "ew");
            if(!vcode) {
                fprintf(redfp, "      real*8     %s(4,2), %s(1)\n",
                        momstr1, momstr2);
            } else {
                fprintf(redfp, "      real*8     ");
                if( rcode ) {
                  fprintf(redfp, "%s(4,2,n_vector), %s(1,n_vector)\n",
                          momstr1, momstr2);
                } else {
                  fprintf(redfp, "%s(n_vector,4,2), %s(n_vector,1)\n",
                          momstr1, momstr2);
                }
            }
            cmicwgt(momstr1, nd+1, sgn, p, "ce");
            if(!vcode) {
                fprintf(redfp, "      complex*16 %s(2,2)\n", momstr1);
            } else {
                if( rcode ) {
                  fprintf(redfp, "      complex*16 %s(2,2,n_vector)\n", momstr1);
                } else {
                  fprintf(redfp, "      complex*16 %s(n_vector,2,2)\n", momstr1);
                }
            }
        }
    }
    fprintf(redfp,"\n");

    fprintf(redfp,"* Normalization\n");
    fprintf(redfp,"      common /a%smdbgg/fknorm,fkcall,nkcall\n", prf);
    fprintf(redfp,"\n");
    fprintf(redfp,"* Calculated table of amplitudes\n");
    fprintf(redfp,"      common /a%smatbl/av, lt, indexg\n", prf);
    if(!vcode) {
      fprintf(redfp,"      complex*16 av(0:l%sag-1)\n", prf);
    } else {
      fprintf(redfp,"      complex*16 av(n_vector,0:l%sag-1)\n", prf);
    }
    fprintf(redfp,"      integer    lt(0:n%sextn), indexg(n%sextn)\n",
            prf, prf);
    fprintf(redfp,"\n");

    fprintf(redfp,"* Spin average\n");
    sprintf(momstr,"      common/a%smspin/", prf);
    puinit(redfp, momstr,
                  "     &               ",
                  ",", ",", "", 72, 16);
    putid("aspin");
    putid("aident");
    if(cpol) {
        sprintf(momstr,"jhs(n%sextn)", prf);
        putid(momstr);

        sprintf(momstr,"jhe(n%sextn)", prf);
        putid(momstr);

        sprintf(momstr,"jcpol(n%sextn)", prf);
        putid(momstr);

        sprintf(momstr,"kperm(n%sextn)", prf);
        putid(momstr);

    } else {
        sprintf(momstr,"jhs(n%sextn)", prf);
        putid(momstr);

        sprintf(momstr,"jhe(n%sextn)", prf);
        putid(momstr);

        sprintf(momstr,"kperm(n%sextn)", prf);
        putid(momstr);

    }
    putend();
}
/*==================================================================*/
Private Void    ftinfo(NoArg)
/*------------------------------------------------------------------*
 * generate "Information" for option -u
 *------------------------------------------------------------------*/
{
    int    n, p, lag, il, lg, tp;
    node   nd;
    pindex pa;

    int    ngrf = reprocg[cprocs];
    lag = 1;
    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype != ILEXTERN) {
            continue;
        }
        nd  = iltable[il].ilnodes[0];
        lg  = iltable[il].illegs[0];
        p   = abs(ndparticle(nd, lg));
        tp  = iltable[il].iltype;
        lag *= cmindx(momstr1, tp, p);
    }
    fprintf(unfp,"*----- Process %d\n",cprocs+1);
    fprintf(unfp,"      ndbgrf(%d) = %d\n",cprocs+1,ngrf);
    fprintf(unfp,"      ndbext(%d) = %d\n",cprocs+1,nextern);
    fprintf(unfp,"      ndbcbs(%d) = %d\n",cprocs+1,gcbase->nbase);
    fprintf(unfp,"      ndblag(%d) = %d\n",cprocs+1,lag);

/* 981009 */
    uniproc[nummultip].nextern = nextern;
    uniproc[nummultip].procg   = ngrf;
    uniproc[nummultip].cbase   = gcbase->nbase;
    uniproc[nummultip].lag     = lag;
    for(n = 0; n < nextern; n++) {
        uniproc[nummultip].pa[n]   = ndname(n);
        uniproc[nummultip].iorf[n] = ndiorf(n);
    }
    nummultip++;

    for(n = 0; n < nextern; n++) {
        pa  = ndname(n);
        p   = abs(pa);
        il  = ndiline(n, 0);
        tp  = iltable[il].iltype;
        cmmass(p, 1, momstr);

        fprintf(unfp,"\n");
        fprintf(unfp,"*     %d: %s %s mass=%s\n",
                    n+1, eliorf(ndiorf(n)), pvename(ndiorf(n),pa),
                    momstr);

        Discard cmindx(momstr, tp, p);
        fprintf(unfp,"      mdbjhs(%d,%d) = 0\n", cprocs+1,n+1);
        fprintf(unfp,"      mdbjhe(%d,%d) = %s - 1\n", cprocs+1,n+1, momstr);
    }
}
/*==================================================================*/
Private Void    ftmuldb(NoArg)
/*------------------------------------------------------------------*
 * generate "Information" for option -u
 *------------------------------------------------------------------*/
{
    int i,n,mg=0,me=0,mc=0,ml=0;
    
    for(i = 0; i < nummultip; i++) {
        printf(" information %d:%d %d %d %d\n",i,
            uniproc[i].nextern, uniproc[i].procg, 
            uniproc[i].cbase, uniproc[i].lag);
        if( uniproc[i].procg   > mg ) mg = uniproc[i].procg;
        if( uniproc[i].nextern > me ) me = uniproc[i].nextern;
        if( uniproc[i].cbase   > mc ) mc = uniproc[i].cbase;
        if( uniproc[i].lag     > ml ) ml = uniproc[i].lag;
        for(n = 0; n < nextern; n++) {
        }
    }
    cmopen(ROFORT, "inclun.h");
    fprintf(redfp,"* for only grcfort -u \n");
    fprintf(redfp,"\n* table of amplitudes\n");
    fprintf(redfp,"      parameter (Mgraph   =%d)\n", mg);
    fprintf(redfp,"      parameter (Mextn    =%d)\n", me);
    fprintf(redfp,"\n* number of color base\n");
    fprintf(redfp,"      parameter (Mcbase   =%d)\n", mc);
    fprintf(redfp,"\n* number of all helicity states\n");
    fprintf(redfp,"      parameter (Mlag     =%d)\n\n", ml);
}

/*********************************************************************
 *    2. For each process
 */
/*==================================================================*/
Private Void    ftmain(pc)
int     pc;
/*------------------------------------------------------------------*
 * Generate amptbl.f
 *------------------------------------------------------------------*/
{
#if DBG
    if(dbg >= redbg) {
        printf("ftmain: begin (%d, %d)\n", cnprc, reprocg[cprocs]);
    }
#endif

    cmproc();
    if(!multiprc) {
        if(colorv == 0) {
            ftampsup(-1);
        } else {
            ftampsmp(-1);
        }
        ftamporp(-1);
        ftamptbp(-1);
        if(ilenc) {
            ftincpp(-1);
        }
    } else {
        ftamptbp(pc);
        ftincpp(pc);
        ftamporp(pc);
        if(colorv == 0) {
            ftampsup(pc);
        } else {
            ftampsmp(pc);
        }
        cmprcpc(pc);
    }
}

/*==================================================================*/
Private Void    ftamptbp(pc)
int     pc;
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    node nd;
    tnode  *ndc;
    int  i, il, tp, s, n0, ioro, g0, gaumask, c, cn, sgn;
    pindex p;
    int    gn0, pc0;
    string prf, ext;
    int    n;

#ifdef OLDSEL
    int  numhiggs, numgluon;
#else
    Ulong selmask, msk;
#endif

    if(pc < 0) {
        pc0 = cprocs;
        gn0 = reprocg[pc0];
        prf = reprocp[pc0];
        sprintf(momstr, "amptbl");
    } else {
        pc0 = pc;
        gn0 = reprocg[pc0];
        prf = reprocp[pc0];
        sprintf(momstr, "a%smptbl", prf);
    }
    sprintf(momstr1, "%s.f", momstr);
    cmopend(ROFORT, prf, momstr1);


    cmsepar('*');
    if(!vcode) {
        fprintf(redfp,"      subroutine %s\n",
                momstr);
    } else {
        fprintf(redfp,"      subroutine %s(nsampl)\n",
                momstr);
    }
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");

    if(vcode) {
        fprintf(redfp, "      include 'inclvs.h'\n");
    }
    fprintf(redfp, "      include 'incl%s1.h'\n", prf);
    fprintf(redfp, "      include 'incl2.h'\n");
    fprintf(redfp, "      include 'inclk.h'\n");
    fprintf(redfp, "      include 'incl%sp.h'\n", prf);
    cmsepar('-');

    fprintf(redfp, "      jgraph = 0\n");

    fprintf(redfp, "\n* External lines\n");
    for(nd = 0; nd < ndnodes(); nd++) {
        if(ndtype(nd) != NDEXTERN) {
            continue;
        }
        p   = ndparticle(nd, 0);
        if(p > 0) {
            ioro = -1;
        } else {
            p = -p;
            ioro = 1;
        }

        if(ptable[p]->ptype == P_VECTOR) {
            il  = ndiline(nd, 0);
            tp  = iltable[il].iltype;
            Discard cmindx(momstr, tp, p);
            Discard cmmass(p, 1, momstr1);
            sgn = cmisgn(nd+1);
            cmicmom(momstr2,nd+1,sgn,"pe");
            if(ptable[p]->sname != NULL) {
                ext = ptable[p]->sname;
            } else if(ptable[p]->chane != NULL) {
                ext = ptable[p]->chane;
            } else {
                ext = "smextv";
            }
            if(!vcode) {
                fprintf(redfp, "      call %s(%s,%s,%s,",
                        ext, momstr, momstr1, momstr2);
            } else {
                fprintf(redfp, "      call %s(nsampl,%s,%s,%s,",
                        ext, momstr, momstr1, momstr2);
            }
            cmicwgt(momstr2, nd+1, sgn, p, "ep");
            cmicwgt(momstr3, nd+1, sgn, p, "eu");
            ftgauge(momstr1, p);
            fprintf(redfp, "%s,%s,%s)\n", momstr2,momstr3,momstr1);

        } else if(ptable[p]->ptype == P_GHOST) {
            ndc  = ndnode(nd);
            if(ndc->nvkind == ELINITIAL) {
                s = 1;
            } else {
                s = -1;
            }
            Discard cmmass(p, 1, momstr1);
            sgn = cmisgn(nd+1);
            cmicmom(momstr2,nd+1,sgn,"pe");
            if(ptable[p]->sname != NULL) {
                ext = ptable[p]->sname;
            } else if(ptable[p]->chane != NULL) {
                ext = ptable[p]->chane;
            } else {
                ext = NULL;
            }
            if(ext != NULL) {
                if(!vcode) {
                    fprintf(redfp, "      call %s(%s,%s,%s)\n",
                           ext, ((s > 0)?"lincom":"loutgo"),
                           momstr1, momstr2);
                } else {
                    fprintf(redfp, "      call %s(nsampl,%s,%s,%s)\n",
                           ext, ((s > 0)?"lincom":"loutgo"),
                           momstr1, momstr2);
                }
            }
        } else if(ptable[p]->ptype == P_DIRAC) {
            ndc  = ndnode(nd);
            if(ndc->nvkind == ELINITIAL) {
                s = ioro;
            } else {
                s = -ioro;
            }
            Discard cmmass(p, 1, momstr1);
            sgn = cmisgn(nd+1);
            cmicmom(momstr2,nd+1,sgn,"pe");
            if(ptable[p]->sname != NULL) {
                ext = ptable[p]->sname;
            } else if(ptable[p]->chane != NULL) {
                ext = ptable[p]->chane;
            } else {
                ext = "smextf";
            }
            if(!vcode) {
                fprintf(redfp, "      call %s(%s,%s,%s,",
                       ext, ((ioro > 0)?"lincom":"loutgo"),
                       momstr1, momstr2);
            } else {
                fprintf(redfp, "      call %s(nsampl,%s,%s,%s,",
                       ext, ((ioro > 0)?"lincom":"loutgo"),
                       momstr1, momstr2);
            }
            cmicwgt(momstr1,nd+1,sgn,p, "ps");
            cmicwgt(momstr2,nd+1,sgn,p, "ce");
            fprintf(redfp, "%s,%s)\n", momstr1,momstr2);
            cmicwgt(momstr1,nd+1,sgn,p, "ew");
            if(!vcode) {
                fprintf(redfp, "      %s(1) = %s\n",
                        momstr1,((s > 0)?"lprtcl":"lantip"));
            } else {
                fprintf(redfp, "      do %3d j = 1, nsampl\n",100+nd);
                if( rcode ) {
                  fprintf(redfp, "         %s(1,j) = %s\n",
                          momstr1,((s > 0)?"lprtcl":"lantip"));
                } else {
                  fprintf(redfp, "         %s(j,1) = %s\n",
                          momstr1,((s > 0)?"lprtcl":"lantip"));
                }
                fprintf(redfp, "  %3d continue\n",100+nd);
            }
        } else if(ptable[p]->ptype == P_MAJORANA) {
            ndc  = ndnode(nd);
            if(ndc->nvkind == ELINITIAL) {
                s = 1;
            } else {
                s = -1;
            }
            Discard cmmass(p, 1, momstr1);
            sgn = cmisgn(nd+1);
            cmicmom(momstr2,nd+1,sgn,"pe");
            if(ptable[p]->sname != NULL) {
                ext = ptable[p]->sname;
            } else if(ptable[p]->chane != NULL) {
                ext = ptable[p]->chane;
            } else {
                ext = "smextf";
            }
            if(!vcode) {
                fprintf(redfp, "      call %s(%s,%s,%s,",
                       ext, ((s > 0)?"lincom":"loutgo"),
                       momstr1, momstr2);
            } else {
                fprintf(redfp, "      call %s(nsampl,%s,%s,%s,",
                       ext, ((s > 0)?"lincom":"loutgo"),
                       momstr1, momstr2);
            }
            cmicwgt(momstr1,nd+1,sgn,p,"ps");
            cmicwgt(momstr2,nd+1,sgn,p,"ce");
            fprintf(redfp, "%s,%s)\n", momstr1, momstr2);
            cmicwgt(momstr1,nd+1,sgn,p,"ew");
            if(!vcode) {
                fprintf(redfp, "      %s(1) = lprtcl\n", momstr1);
            } else {
                fprintf(redfp, "      do %3d j = 1, nsampl\n",100+nd);
                if( rcode ) {
                  fprintf(redfp, "         %s(1,j) = lprtcl\n", momstr1);
                } else {
                  fprintf(redfp, "         %s(j,1) = lprtcl\n", momstr1);
                }
                fprintf(redfp, "  %3d continue\n",100+nd);
            }
        } else if(ptable[p]->ptype == P_SCALAR) {
            ndc  = ndnode(nd);
            if(ndc->nvkind == ELINITIAL) {
                s = 1;
            } else {
                s = -1;
            }
            Discard cmmass(p, 1, momstr1);
            sgn = cmisgn(nd+1);
            cmicmom(momstr2,nd+1,sgn,"pe");
            if(ptable[p]->sname != NULL) {
                ext = ptable[p]->sname;
            } else if(ptable[p]->chane != NULL) {
                ext = ptable[p]->chane;
            } else {
                ext = NULL;
            }
            if(ext != NULL) {
                if(!vcode) {
                    fprintf(redfp, "      call %s(%s,%s,%s)\n",
                           ext, ((s > 0)?"lincom":"loutgo"),
                           momstr1, momstr2);
                } else {
                    fprintf(redfp, "      call %s(nsampl,%s,%s,%s)\n",
                           ext, ((s > 0)?"lincom":"loutgo"),
                           momstr1, momstr2);
                }
            }
        }
    }
    fprintf(redfp, "\n");
    /* CF */
    if(colorv == 1) {
        if(gcbase->nbase > 1) {
            fprintf(redfp, "\n* Buffer clear for color-based-amplitude\n");
            fprintf(redfp,"      do 200 ic = 0 , ncbase-1\n");
            fprintf(redfp,"      do 200 ih = 0 , l%sag-1\n", prf);
            if(!vcode) {
                fprintf(redfp,"         agc(ih,ic) = (0.0d0,0.0d0)\n");
            } else {
                fprintf(redfp,"      do 200 j = 1 , nsampl\n");
                fprintf(redfp,"         agc(j,ih,ic) = (0.0d0,0.0d0)\n");
            }
            fprintf(redfp,"  200 continue\n\n");
        } else {
            fprintf(redfp, "\n* Buffer clear for amplitude\n");
            fprintf(redfp,"      do 200 ih = 0 , l%sag-1\n", prf);
            if(!vcode) {
                fprintf(redfp,"         agc(ih,0) = (0.0d0,0.0d0)\n");
            } else {
                fprintf(redfp,"      do 200 j = 1 , nsampl\n");
                fprintf(redfp,"         agc(j,ih,0) = (0.0d0,0.0d0)\n");
            }
            fprintf(redfp,"  200 continue\n\n");
        }
    }

    /* ampt0 : graph 1 .. TCUT;  ampt1 : graph TCUT+1 .. 2*TCUT */
    cn = (gn0 + TCUT - 1)/TCUT;

    for(c = 0; c < cn; c++) {
        if(!vcode) {
            fprintf(redfp, "      call a%smpt%d\n", prf, c);
        } else {
            fprintf(redfp, "      call a%smpt%d(nsampl)\n", prf, c);
        }
    }

    fprintf(redfp, "\n");

    if( ucode && cpol ) {

        fprintf(redfp,"      lt(0) = nextn\n");
        fprintf(redfp,"      do 1100 n = 1, nextn\n");
        fprintf(redfp,"         lt(n) = maxjhe(n)+1\n");
        fprintf(redfp," 1100 continue\n");

        for(n = 0; n < nextern; n++) {

          if(cmphoton(n)) {
            fprintf(redfp,"\n      if(jcpol(%d).ne.0) ", n+1);
            if(!vcode) {
               fprintf(redfp,"call smcpol(nextn, lag, %d, lt, agc)\n", n+1);
            } else {
               fprintf(redfp,"call smcpol(nsampl, nextn, lag, %d, lt, agc)\n", n+1);
            }
         }
       }
    }
    fprintf(redfp, "      return\n");
    fprintf(redfp, "      end\n");

    c = 0;

    /* repeat graph */
    for(g0 = 0; g0 < gn0; g0++) {
        if(g0%TCUT == 0) {
            if(g0 > 0) {
                fprintf(redfp, "      return\n");
                fprintf(redfp, "      end\n");
            }
            sprintf(momstr, "a%smpt%d.f", prf, c);
         /* cmopend(ROFORT, prf, momstr); */
            cmsepar('*');

            fprintf(redfp, "      subroutine a%smpt%d", prf, c);
            if(vcode) {
                fprintf(redfp, "(nsampl)");
            }
            fprintf(redfp, "\n");

            fprintf(redfp, "      implicit real*8(a-h,o-z)\n\n");
            if(vcode) {
                fprintf(redfp, "      include 'inclvs.h'\n");
            }
            fprintf(redfp, "      include 'incl%s1.h'\n", prf);
            fprintf(redfp, "      include 'incl2.h'\n");
            fprintf(redfp, "      include 'inclk.h'\n");
            cmsepar('-');

            c++;
        }
        n0 = g0 + 1;

        cgraph  = daread(pc0, g0);

        /* initialize i-line table */
/*        printf("** ilmon:3 **\n");*/
        ilmom();
        nilines = cgraph->nilines;

#ifdef OLDSEL
        numhiggs = ftishiggs();
        if(numhiggs > 0) {
            fprintf(redfp, "      if(jhiggs .ne. 0) then\n");
        }

        numgluon = ftisgluon();
        if(numgluon > 0) {
            fprintf(redfp, "      if(jgluon .ne. 0) then\n");
        }
#else
        selmask = ftselect();
        msk = 1L;
        for(i = 0; i < nselect; i++) {
            if(selmask & msk) {
                fprintf(redfp, "      if(j%s .ne. 0) then\n", sselect[i]);
            }
            msk <<= 1;
        }
#endif

        gaumask = ftgmask();
        for(i = 0; i < ngauge; i++) {
            if(gaumask & gmask[i]) {
                fprintf(redfp, "      if(igau%s .ne. 0) then\n",
                        sgauge[i]);
            }
        }
        fprintf(redfp, "      if(jselg(%d) .ne. 0) then\n", n0);
        fprintf(redfp, "        jgraph = jgraph + 1\n");
        fprintf(redfp, "        igraph(jgraph) = %d\n", n0);

        fprintf(redfp, "        call a%sg%d", prf, n0);
        if(vcode) {
            fprintf(redfp, "(nsampl)");
        }
        fprintf(redfp, "\n");

        fprintf(redfp, "      endif\n");
        for(i = 0; i < ngauge; i++) {
            if(gaumask & gmask[i]) {
                fprintf(redfp, "      endif\n");
            }
        }
#ifdef OLDSEL
        if(numgluon > 0) {
            fprintf(redfp, "      endif\n");
        }
        if(numhiggs > 0) {
            fprintf(redfp, "      endif\n");
        }
#else
        msk = 1L;
        for(i = 0; i < nselect; i++) {
            if(selmask & msk) {
                fprintf(redfp, "      endif\n");
            }
            msk <<= 1;
        }
#endif
        fprintf(redfp, "\n");
    }

    fprintf(redfp, "      return\n");
    fprintf(redfp, "      end\n");
}

/*==================================================================*/
Private int     ftgmask(NoArg)
/*------------------------------------------------------------------*
 * generate REDUCE code for one graph g0.
 *------------------------------------------------------------------*/
{
    pindex p;
    int    il, msk, ig;

    msk = 0;
    for(il = 1; il < cgraph->nilines; il++) {
        p = abs(ilpart(il));
        if(ptable[p]->ptype != P_SCALAR) {
            continue;
        }
        if((ig = ftgauge(momstr, p)) >= 0) {
            msk |= gmask[ig];
        }
    }
    return msk;
}

/*********************************************************************
 *    3. For each graph
 */
/*==================================================================*/
Public  Void    ftgraph(proc, graph)
tproc   *proc;
tgraph  *graph;
/*------------------------------------------------------------------*
 * generate REDUCE code for one graph cngraph.
 *------------------------------------------------------------------*/
{
    string prf = cprefix;

#if DBG
    if(dbg >= redbg) {
        printf("ftgraph: begin (%d)\n", cngraph+1);
    }
#endif

    /* initialize variable */
    cprocess  = proc;
    cgraph    = graph;
    nilines   = cgraph->nilines;

    conjstart = graph->conjsrt;

    /* file name */
    if(!mergefile) {
        sprintf(outname, "a%sg%d.f", prf, cngraph+1);
    } else {
        if( cngraph != 0 && cngraph%ACUT == 0 ) {
            sprintf(outname, "a%smps%d.f", prf, cngraph/ACUT);
        }
    }

    fttmpopen();
    fttmpclr();

    ftgtree();

    /* open file */
    if(!mergefile) {
        cmopend(ROFORT, prf, outname);
    } else {
        if( cngraph != 0 && cngraph%ACUT == 0 ) {
            cmopend(ROFORT, prf, outname);
        }
    }

    ftbegin();

    ftvars(cngraph);

    fttmpcp();

    ftgend();

    /* increase offset of local variables */
/* back
    momoffs += cgraph->nilines;
    legoffs += cprocess->maxlegs;
 */

#if DBG
    if(dbg >= redbg) {
        printf("ftgraph: end (%d)\n", cngraph+1);
    }
#endif
}

/*==================================================================*/
Private Void    fttmpopen(NoArg)
/*------------------------------------------------------------------*
 * Open tempolary files.
 *------------------------------------------------------------------*/
{
    static int opn = False;

    if(opn) {
        rewind(tfp1);
        return;
    }

    opn = True;

    /* open temporary file */
#ifdef SONY_NEWS
    tfp1 = fopen("tmp1", "w+"));
#else
    tfp1 = tmpfile();
#endif

    if(tfp1 == NULL || tfp1 == NULL) {
        error("fttmpopen", "temporary file cannot open");
        exit(1);
    }

}
/*==================================================================*/
Private Void    fttmpcp(NoArg)
/*------------------------------------------------------------------*
 * Copy tempolary files.
 *------------------------------------------------------------------*/
{
    int c;

    putc(EOFSYM, tfp1);

    fflush(tfp1);

    rewind(tfp1);

    while((c = getc(tfp1)) != EOFSYM) {
        putc(c, redfp);
    }
}
/*==================================================================*/
Private Void    ftgtree(NoArg)
/*------------------------------------------------------------------*
 * tree graph
 *------------------------------------------------------------------*/
{
#if DBG
    if(dbg >= redbg) {
        printf("ftgtree: begin\n");
    }
#endif

    if(majorana) {
        vsign = 1;
    }

    iliconsv();

    /* momentum conservation */
    if(!ilenc) {
        ftmconsv();
    }

    /* denominators of propagators */
    ftdenom();

    /* internal lines */
    ftiline();

    /* initialize working table of vertices */
    ftvlts();

#ifdef SUSYDBG
    prdump();
#endif

    /* vertices */
    ftnode();

    /* connect vertices */
    ftconn();

    /* denominator of propagator */
    ftdprop();

    /* call ampord */
    ftclpord();

    /* color factor x amplitude */
    ftcfamp();

    /* clear work */
    ftvlte();

#if DBG
    if(dbg >= redbg) {
        printf("ftgtree: end\n");
    }
#endif
}

/*==================================================================*/
Private Void    ftbegin(NoArg)
/*------------------------------------------------------------------*
 * Output begining message to REDUCE code.
 *------------------------------------------------------------------*/
{
#if DBG
    if(dbg >= redbg) {
        printf("ftbegin: begin\n");
    }
#endif
    cmsepar('*');
    if(cgraph->subid == 0) {
        fprintf(redfp, "*             Graph No. %ld\n",
                cgraph->grid);
    } else {
        fprintf(redfp, "*             Graph No. %ld - %ld\n",
                cgraph->grid, cgraph->subid);
    }
    if(cgraph->subid != 0 || cngraph != cgraph->grid) {
        fprintf(redfp, "*         Generated No. %d\n", cngraph+1);
    }
    cmsepar('*');
}

/*==================================================================*/
Private Void    ftmconsv(NoArg)
/*------------------------------------------------------------------*
 *  Momentum conservation of internal momenta.
 *------------------------------------------------------------------*/
{
    int    il, ii, ce, ce0, sgn;
    int    nod, leg, p, q, f0, i, t;

#if DBG
    if(dbg >= redbg) {
        printf("ftmconsv: begin\n");
    }
#endif

    fprintf(tfp1, "* Internal momenta\n");
    fprintf(tfp1, "      do 100 i = 1, 4\n");
    if(vcode) {
        fprintf(tfp1, "      do 100 j = 1, nsampl\n");
    }

    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype == ILEXTERN) {
            continue;
        } else if(iltable[il].iltype == ILLOOP) {
            continue;
        }

        sgn = cmisgn(il);
        cmicmom(momstr, il, sgn, "pe");
        if(!vcode) {
            fprintf(tfp1, "        %s(i) =", momstr);
        } else {
            if( rcode ) {
              fprintf(tfp1, "        %s(i,j) =", momstr);
            } else {
              fprintf(tfp1, "        %s(j,i) =", momstr);
            }
        }
        if(!ilenc) {
            if(!vcode) {
                fttmp(FT_REAL, momstr, "4", 4);
            }else {
              if( rcode ) {
                fttmp(FT_REAL, momstr, "4,n_vector", 4);
              } else {
                fttmp(FT_REAL, momstr, "n_vector,4", 4);
              }
            }
        }

        /* reverse the direction of fermion propagator */
        ce0 = 1;
        nod = iltable[il].ilnodes[0];
        leg = iltable[il].illegs[0];
        p   = ndparticle(nod, leg);
        t   = ptable[abs(p)]->ptype;

        if(t == P_MAJORANA || t == P_DIRAC) {
            q = ndparticle(nod, 0);
            if(ptable[abs(q)]->ptype == P_MAJORANA ||
               ptable[abs(q)]->ptype == P_DIRAC) {
                f0 = 0;
            } else {
                f0 = 1;
            }
            /* (nod, f0), (nod, f0+1) are fermions and i = 0 or 1 */
            i = leg - f0;

            /* i = 0, u-bar : outgoing,
             * i = 1, u     : incoming
             */
            ce0 = 2*i - 1;

            if(p < 0) {
                ce0 *= -1;
            }
        }

        for(ii = 1; ii < nilines; ii++) {
            if(ii == kloop) {
                continue;
            } else if((ce = iltable[il].ilmlist[ii]) == 0) {
                continue;
            }
            cmvcmom(momstr, iltable[ii].ilnodes[0],
                    iltable[ii].illegs[0], False, "pe");

            ce *= ce0;

            if(ce == -1) {
                if(!vcode) {
                    fprintf(tfp1, " - %s(i)", momstr);
                } else {
                    if( rcode ) {
                      fprintf(tfp1, " - %s(i,j)", momstr);
                    } else {
                      fprintf(tfp1, " - %s(j,i)", momstr);
                    }
                }
            } else if(ce == 1) {
                if(!vcode) {
                    fprintf(tfp1, " + %s(i)", momstr);
                } else {
                    if( rcode ) {
                      fprintf(tfp1, " + %s(i,j)", momstr);
                    } else {
                      fprintf(tfp1, " + %s(j,i)", momstr);
                    }
                }
            } else if(ce > 0) {
                if(!vcode) {
                    fprintf(tfp1, " + %d*%s(i)", ce, momstr);
                } else {
                    if( rcode ) {
                      fprintf(tfp1, " + %d*%s(i,j)", ce, momstr);
                    } else {
                      fprintf(tfp1, " + %d*%s(j,i)", ce, momstr);
                    }
                }
            } else if(ce < 0) {
                if(!vcode) {
                    fprintf(tfp1, " - %d*%s(i)", abs(ce), momstr);
                } else {
                    if( rcode ) {
                      fprintf(tfp1, " - %d*%s(i,j)", abs(ce), momstr);
                    } else {
                      fprintf(tfp1, " - %d*%s(j,i)", abs(ce), momstr);
                    }
                }
            }
        }
        fprintf(tfp1, "\n");
    }
    fprintf(tfp1, "  100 continue\n");

#if DBG
    if(dbg >= redbg) {
        printf("ftmconsv: end\n");
    }
#endif
}

/*==================================================================*/
Private Void    ftdenom(NoArg)
/*------------------------------------------------------------------*
 *  Clear temp. var table
 *------------------------------------------------------------------*/
{
    int il, e1, e2, ce1, ce2, s, c, pc, sgn;
    pindex p, pl;
    string prop;

    fprintf(tfp1, "\n* Denominators of propagators\n");
    if(nilines < 1) {
        return;
    }
    if(colorv == 0) {
        if(!vcode) {
            fprintf(tfp1, "      aprop(jgraph) = 1.0d0\n");
        } else {
            fprintf(tfp1, "      do 200 j = 1, nsampl\n");
            fprintf(tfp1, "         aprop(j,jgraph) = 1.0d0\n");
            fprintf(tfp1, "  200 continue\n");
        }
    } else {
        if(!vcode) {
            fprintf(tfp1, "      aprop = 1.0d0\n");
        } else {
            fprintf(tfp1, "      do 200 j = 1, nsampl\n");
            fprintf(tfp1, "         aprop(j) = 1.0d0\n");
            fprintf(tfp1, "  200 continue\n");
        }
    }

    if(nilines < 1) {
        fprintf(tfp1, "* %d propagators\n", nilines);
        return;
    }
    for(il = 1; il < nilines; il++) {
        p = abs(ilpart(il));
        if(iltable[il].iltype == ILEXTERN) {
            continue;
        } else if(iltable[il].iltype == ILLOOP) {
            continue;
        }

        /* denominator of the propagator for i-line il */
        sgn = cmisgn(il);
        cmicmom(momstr, il, sgn, "vm");
        if(!ilenc) {
            if(vcode) {
                fprintf(tfp1, "      do %3d j = 1, nsampl\n",200+il);
            }
            cmicmom(momstr, il, sgn, "vm");
            puinit(tfp1, "      ", "     &      ","","","", 70, 16);
            putid(momstr);
            if(vcode) {
                putid("(j)");
            }
            putid(" = ");
            if(!vcode) {
                fttmp(FT_REAL, momstr, NULL, 1);
            } else {
                fttmp(FT_REAL, momstr, "n_vector", 1);
            }

            /* inner products of momenta */
            for(e1 = 0; e1 < nilines; e1++) {
                if(e1 == kloop) {
                    continue;
                } else if((ce1 = iltable[il].ilmlist[e1]) == 0) {
                    continue;
                }

                for(e2 = e1+1; e2 < nilines; e2++) {
                    if(e2 == kloop) {
                        continue;
                    } else if((ce2 = iltable[il].ilmlist[e2]) == 0) {
                        continue;
                    }

                    ce2 *= ce1;
                    if(ce2 > 0) {
                        s = ce2;
                        c = '+';
                    } else {
                        s = -ce2;
                        c = '-';
                    }
                    sprintf(momstr1, " %c ", c);
                    putid(momstr1);
                    s *= 2;
                    if(s != 1) {
                        sprintf(momstr1, "%d.0d0*", s);
                        putid(momstr1);
                    }
                    if(!vcode) {
                        sprintf(momstr1, "prod(%d,%d)", e1, e2);
                    } else {
                        if( rcode ) {
                          sprintf(momstr1, "prod(%d,%d,j)", e1, e2);
                        } else {
                          sprintf(momstr1, "prod(j,%d,%d)", e1, e2);
                        }
                    }
                    putid(momstr1);
                }
            }

            /* mass terms */
            pl = ilpart(il);
            pc = 0;
            for(e1 = 1; e1 < nilines; e1++) {
                if(e1 == kloop) {
                    continue;
                } else if((ce1 = iltable[il].ilmlist[e1]) == 0) {
                    continue;
                }

                p = ilpart(e1);
                Discard cmmass(p, 1, momstr2);
                if(p == pl) {
                    pc++;
                } else {
                    sprintf(momstr1, " + %s**2", momstr2);
                    putid(momstr1);
                }
            }
            if(pc != 0) {
                if(pc > 0) {
                    s = pc;
                    c = '+';
                } else {
                    s = -pc;
                    c = '-';
                }
                if(s == 1) {
                    sprintf(momstr1, " %c ", c);
                } else {
                    sprintf(momstr1, " %c %d.0d0*", c, s);
                }
                putid(momstr1);
                Discard cmmass(pl, 1, momstr2);
                sprintf(momstr1, "%s**2", momstr2);
                putid(momstr1);
            }
            putend();
            if(vcode) {
                fprintf(tfp1, "  %3d continue\n",200+il);
            }
        }

        /* product of denominators */
        pl = ilpart(il);
        Discard cmmass(pl, 1, momstr2);
        if(ptable[pl]->snami != NULL) {
            prop = ptable[pl]->snami;
        } else if(ptable[pl]->chani != NULL) {
            prop = ptable[pl]->chani;
        } else {
            prop = "snprpd";
        }
        if(colorv == 0) {
            if(!vcode) {
                fprintf(tfp1, "      call %s(pphase,aprop(jgraph),%s,",
                        prop, momstr);
            } else {
                fprintf(tfp1, "      call %s(pphase,nsampl,", prop);
                fprintf(tfp1, "aprop(1,jgraph),%s,",momstr);
            }
        } else {
            if(!vcode) {
                fprintf(tfp1, "      call %s(pphase,aprop,%s,", prop, momstr);
            } else {
                fprintf(tfp1, "      call %s(pphase,nsampl,aprop,%s,",
                        prop, momstr);
            }
        }
        if(ptable[pl]->ptype == P_SCALAR) {
            if(ftgauge(momstr1, pl) >= 0) {
                fprintf(tfp1, "agauge(%s)*", momstr1);
            }
        }
/* TORIAEZU */
        fprintf(tfp1, "\n     &      ");
/* TI 980226 */
        fprintf(tfp1, "%s**2,", momstr2);
        Discard cmwidth(pl, 1, momstr1);
        if(strcmp(momstr1, "0") == 0) {
            fprintf(tfp1, "0.0d0)\n");
        } else {
            fprintf(tfp1, "%s*%s)\n", momstr2, momstr1);
        }
    }
}

/*==================================================================*/
Private Void    fttmpclr(NoArg)
/*------------------------------------------------------------------*
 *  Clear temp. var table
 *------------------------------------------------------------------*/
{
    static bool realmax = False;
    int i;

    for(i = 0; i < nvars; i++) {
        if(vartbl[i].name != NULL) {
            free(vartbl[i].name);
        }
        if(vartbl[i].dim != NULL) {
            free(vartbl[i].dim);
        }
    }
    nvars = 0;
    vartype[FT_INT]      = 0;
    vartype[FT_REAL]     = 0;
    vartype[FT_COMPLEX]  = 0;
    vartype[FT_INTMAX]   = 2*(nextern - 3)*(nextern + 12);
    if(!realmax) {
        vartype[FT_REALMAX]  = 8;
        realmax = True;
    }
}

/*==================================================================*/
Private Void    fttmp(vtype, name, dim, size)
int    vtype;
string name;
string dim;
int    size;
/*------------------------------------------------------------------*
 *  Save temporary variables
 *------------------------------------------------------------------*/
{
    if(nvars >= NVARS) {
        fprintf(Stderr, "fttmp: var-table overflow\n");
        exit(1);
    }
    vartype[vtype]++;
    vartbl[nvars].vtype = vtype;
    vartbl[nvars].name  = (string) strdup(name);
    if(dim == NULL) {
        vartbl[nvars].dim   = NULL;
    } else {
        vartbl[nvars].dim   = (string) strdup(dim);
    }
    vartbl[nvars].size  = size;
    nvars++;

}

#if DBG
/*==================================================================*/
Private Void    fttmdmp(NoArg)
/*------------------------------------------------------------------*
 *  Dump temporary variables
 *------------------------------------------------------------------*/
{
    int v, s, c, ts = 0;

    printf("\nVairbles : compex %ld, real %ld, int %ld\n",
           vartype[FT_COMPLEX], vartype[FT_REAL],
           vartype[FT_INT]);
    for(v = 0; v < nvars; v++) {
        if(vartbl[v].vtype == FT_COMPLEX) {
            c = 'C';
            s = 16*vartbl[v].size;
            ts += s;
        } else if(vartbl[v].vtype == FT_REAL) {
            c = 'R';
            s = 8*vartbl[v].size;
            ts += s;
        } else if(vartbl[v].vtype == FT_INT) {
            c = 'I';
            s = 4*vartbl[v].size;
            ts += s;
        } else {
            c = '?';
            s = vartbl[v].size;
            ts += s;
        }
        printf("%c %6d(%4d) %6s", c,s,vartbl[v].size,vartbl[v].name);
        if(vartbl[v].dim != NULL) {
            printf("(%s)", vartbl[v].dim);
        }
        printf("\n");
    }
    printf("  %6d\n\n", ts);
}
#endif

/*==================================================================*/
Private Void    ftvars(g)
int     g;
/*------------------------------------------------------------------*
 *  Save temporary variables
 *------------------------------------------------------------------*/
{
    long int lsizer = 0, lsizei=0;
    int  v, ldim, tp, typ, p, p0;
    string str;
    string prf = cprefix;

#if DBG
    if(dbg >= redbg) {
        fttmdmp();
    }
#endif

    fprintf(redfp, "      subroutine a%sg%d", prf , g+1);
    if(vcode) {
        fprintf(redfp, "(nsampl)");
    }
    fprintf(redfp, "\n");

    fprintf(redfp, "      implicit real*8(a-h,o-z)\n\n");
    if(vcode) {
        fprintf(redfp, "      include 'inclvs.h'\n");
    }
    fprintf(redfp, "      include 'incl%s1.h'\n", prf);
    fprintf(redfp, "      include 'inclk.h'\n");
    fprintf(redfp, "      include 'incl%sp.h'\n", prf);
    cmsepar('-');

    if(vartype[FT_COMPLEX] + vartype[FT_REAL] > 0) {
        puinit(redfp, "      common /amwork/",
                      "     &               ",
                      ",", ",", "", 64, 16);
        for(v = 0; v < nvars; v++) {
            if(vartbl[v].vtype != FT_COMPLEX) {
                continue;
            }
            putid(vartbl[v].name);
            lsizer += 16*vartbl[v].size;
        }
        for(v = 0; v < nvars; v++) {
            if(vartbl[v].vtype != FT_REAL) {
                continue;
            }
            putid(vartbl[v].name);
            lsizer += 8*vartbl[v].size;
        }
        if(vartype[FT_REALMAX] > lsizer) {
            sprintf(momstr, "rdmy(%ld)", (vartype[FT_REALMAX]-lsizer)/8);
            putid(momstr);
        } else {
            vartype[FT_REALMAX] = lsizer;
        }
        putend();
    }

    /* max of lsizei = 4*#vert + sum_1^{#prop-1} (4+i) */
/*
    lsizeim = 2*(nextern - 3)*(nextern + 12);
*/
    if(vartype[FT_INT] > 0) {
        puinit(redfp, "      common /amwori/",
                      "     &          ",
                      ",", ",", "", 64, 16);
        for(v = 0; v < nvars; v++) {
            if(vartbl[v].vtype != FT_INT) {
                continue;
            }
            putid(vartbl[v].name);
            lsizei += 4*vartbl[v].size;
        }
        if(vartype[FT_INTMAX] > lsizei) {
            sprintf(momstr, "idmy(%ld)", (vartype[FT_INTMAX]-lsizei)/4);
            putid(momstr);
        } else {
            vartype[FT_INTMAX] = lsizei;
        }
        putend();
    }

    fprintf(redfp,"*     %ld (%ld) + %ld (%ld) bytes used\n\n",
            lsizer, vartype[FT_REALMAX], lsizei, vartype[FT_INTMAX]);
    incl2r = max(incl2r, lsizer);
    incl2i = max(incl2i, lsizei);

    for(typ = FT_INT; typ < FT_COMPLEX; typ++) {
        if(vartype[typ] > 0) {
            if(typ == FT_COMPLEX) {
                puinit(redfp, "      complex*16 ",
                              "     &           ",
                              ",", ",", "", 64, 16);
            } else if(typ == FT_REAL) {
                puinit(redfp, "      real*8     ",
                              "     &           ",
                              ",", ",", "", 64, 16);
            } else if(typ == FT_INT) {
                puinit(redfp, "      integer    ",
                              "     &           ",
                              ",", ",", "", 64, 16);
            } else {
                fprintf(Stderr, "*** illegal variable type = %d\n",
                                typ);
                exit(1);
            }
            for(v = 0; v < nvars; v++) {
                if(vartbl[v].vtype != typ) {
                    continue;
                }
                if(vartbl[v].dim == NULL) {
                    ldim = 0;
                } else {
                    ldim = strlen(vartbl[v].dim);
                }
                if(strlen(vartbl[v].name) + ldim + 1
                   >= MAXSSTR) {
                    fprintf(Stderr, "ftvar: too long string\n");
                    exit(1);
                }
                str = sizestr;
                strcpy(str, vartbl[v].name);
                str += strlen(vartbl[v].name);
                if(vartbl[v].dim != NULL) {
                    *str++ = '(';
                    strcpy(str, vartbl[v].dim);
                    str += strlen(vartbl[v].dim);
                    *str++ = ')';
                    *str   = '\0';
                }
                putid(sizestr);
            }
            putend();
        }
    }

    for(typ = FT_COMPLEX; typ < FT_MAX; typ++) {
        if(vartype[typ] > 0) {
            for(v = 0; v < nvars; v++) {
                if(vartbl[v].vtype != typ) {
                    continue;
                }

                if(vartbl[v].dim == NULL) {
                    ldim = 0;
                } else {
                    ldim = strlen(vartbl[v].dim);
                }

                if(typ == FT_COMPLEX) {
                    puinit(redfp, "      complex*16 ",
                                  "     &           ",
                                  "", "", "", 70, 16);
                } else if(typ == FT_REAL) {
                    puinit(redfp, "      real*8     ",
                                  "     &           ",
                                  "", "", "", 70, 16);
                } else if(typ == FT_INT) {
                    puinit(redfp, "      integer    ",
                                  "     &           ",
                                  "", "", "", 70, 16);
                } else {
                    fprintf(Stderr, "*** illegal variable type = %d\n",
                                    typ);
                    exit(1);
                }
                putid(vartbl[v].name);
                if(vartbl[v].dim != NULL) {
                    putid("(");
                    p = 0;
                    while(p < ldim) {
                        p0 = p;
                        p++;
                        while(p < ldim && isalnum(vartbl[v].dim[p])) {
                            p++;
                        }
                        tp = vartbl[v].dim[p];
                        vartbl[v].dim[p] = '\0';
                        putid(vartbl[v].dim + p0);
                        vartbl[v].dim[p] = tp;
                    }
                    putid(")");
                }
                putend();
            }
        }
    }

    /* CF  */
    if(colorv == 1) {
        fprintf(redfp,"      complex*16 atmp\n");
        if(gcbase->nbase > 1) {
#ifdef CINLINE
#else
            fprintf(redfp,"      real*8     cwgt(0:%d)\n",gcbase->nbase-1);
#endif
/* For KK-COLOR-BASE for AUG95-version
            fprintf(redfp,"      data       cwgt /");
            for(i=0;i<gcbase->nbase;i++) {
                fprintf(redfp,"%fd0%c",gcolgr->wgt[i],
                    (i==gcbase->nbase-1)?'/':',');
            }
*/
            fprintf(redfp,"\n");
        }
    }

    cmsepar('-');
}

/*==================================================================*/
Private Void    ftincw(NoArg)
/*------------------------------------------------------------------*
 *  Output incl2.h
 *------------------------------------------------------------------*/
{
    cmopen(ROFORT, "incl2.h");

    if(incl2r > 0) {
        if(!vcode) {
            fprintf(redfp, "      common /amwork/rdmi(%d)\n",
                    incl2r/8);
        } else {
            fprintf(redfp, "      common /amwork/rdmi(n_vector,%d)\n",
                    incl2r/8);
        }
    }
    if(incl2i > 0) {
        fprintf(redfp, "      common /amwori/idmi(%d)\n", incl2i/4);
    }
}

/*==================================================================*/
Private Void    ftincpp(pc)
int     pc;
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    int  i;
    long nmom;
    string prf0 = "";
    string prf;

    if(pc < 0) {
        prf = prf0;
    } else {
        prf = reprocp[pc];
    }
    sprintf(momstr1, "incl%sp.h", prf);
    cmopend(ROFORT, prf, momstr1);

    sprintf(momstr2, "      common /a%smilcd/", prf);
    puinit(redfp, momstr2,
                  "     &               ",
                  ",", ",", "", 72, 16);
    nmom = 1;
    for(i = 0; i < nextern; i++) {
        nmom *= 2;
    }
    for(i = 2; i < nmom; i++) {
        cmccmom(momstr3, i, 1, "pe");
        if(!vcode) {
            sprintf(momstr4, "%s(4)", momstr3);
        } else {
            if( rcode ) {
              sprintf(momstr4, "%s(4,n_vector)", momstr3);
            } else {
              sprintf(momstr4, "%s(n_vector,4)", momstr3);
            }
        }
        putid(momstr4);
    }
    putend();

    sprintf(momstr2, "      common /a%smilcp/", prf);
    puinit(redfp, momstr2,
                  "     &               ",
                  ",", ",", "", 72, 16);
    nmom = 1;
    for(i = 0; i < nextern; i++) {
        nmom *= 2;
    }
    for(i = 2; i < nmom; i++) {
        cmccmom(momstr, i, 1, "vm");
        if(!vcode) {
            sprintf(momstr1, "%s", momstr);
        }else {
            sprintf(momstr1, "%s(n_vector)", momstr);
        }
        putid(momstr1);
    }
    putend();
}

/*==================================================================*/
Private Void    ftconn(NoArg)
/*------------------------------------------------------------------*
 * Connect vertices by i-line.
 *------------------------------------------------------------------*/
{
    node n, nc, no, nn;
    int  size, il, c;
    int  l, lg, lo, ln, sgn;
    vlegt **vl;

    fprintf(tfp1, "\n");
    while(True) {
        if(nvltbl <= nvlusd + 1) {
            return;
        }

        /* find a node "nc" to be connected */
        nc = -1;
        for(n = 0; n < nvltbl; n++) {
            if(vltbl[n].size <= 0) {
                continue;
            }
            nc = n;
            break;
        }
        if(nc < 0) {
            return;
        }

        /* find a leg of node "nc" to be connected */
        size = vltbl[nc].size;
        vl   = vltbl[nc].vl;
        lg   = -1;
        for(l = 0; l < size; l++) {
            if(vl[l]->t != ILEXTERN) {
                lg = l;
                break;
            }
        }
        if(lg < 0) {
            fprintf(Stderr,"ftconn : no internal line\n");
            fprintf(Stderr,"nc = %d, lg = %d\n", nc, lg);
            ftvldmp();
            exit(1);
        }
        il   = vl[lg]->i;

        /* find a node and its node to be connected with (nc,lg) */
        for(n = 0; n < nvltbl; n++) {
            if(vltbl[n].size <= 0) {
                continue;
            }
            if(n == nc) {
                continue;
            }
            for(l = 0; l < vltbl[n].size; l++) {
                if(vltbl[n].vl[l]->i == il) {
                    lo = l;
                    no = n;
                    goto found;
                }
            }
        }
        fprintf(Stderr,"ftconn : inconsistent connection\n");
        fprintf(Stderr,"nc = %d, lg = %d, il = %d\n", nc, lg, il);
        ftvldmp();
        exit(1);

        found:;
        /* connect (nc, lg) with (no, lo) */
        if(nvltbl >= MAXVLTBL - 1) {
            fprintf(Stderr, "ftconn: too many vertices\n");
            ftvldmp();
            exit(1);
        }
        nn = nvltbl;
        vltbl[nn].size = vltbl[nc].size + vltbl[no].size - 2;
        if((vltbl[nn].vl =
          (vlegt**) malloc(vltbl[nn].size*sizeof(vlegt*)))
          == NULL) {
            fprintf(Stderr, "ftconn: malloc failed\n");
            ftvldmp();
            exit(1);
        }
        for(l = 0; l < vltbl[nn].size; l++) {
            if((vltbl[nn].vl[l] =
              (vlegt*) malloc(sizeof(vlegt)))
              == NULL) {
                fprintf(Stderr, "ftconn: malloc failed\n");
                ftvldmp();
                exit(1);
            }
        }

        ln = 0;
        for(l = 0; l < vltbl[nc].size; l++) {
            if(l == lg) {
                continue;
            }
            ftvlcp(vltbl[nn].vl[ln], vltbl[nc].vl[l]);
            ln++;
        }
        for(l = 0; l < vltbl[no].size; l++) {
            if(l == lo) {
                continue;
            }
            ftvlcp(vltbl[nn].vl[ln], vltbl[no].vl[l]);
            ln++;
        }
        vltbl[nc].size *= -1;
        vltbl[no].size *= -1;
        nvltbl++;
        nvlusd += 2;
        c = ptypesym[ptable[abs(ilpart(il))]->ptype];

        /* for majorana */
        if(c == 'm') {
            c = 'f';
        }
        sprintf(momstr, "      call smcon%c(", c);
        puinit(tfp1,momstr,
                    "     &           ",
                    ",", ",", ")", 72, 16);
        if(vcode) {
            putid("nsampl");
        }

        if(c == 's') {
            sprintf(momstr, "lt%d,lt%d,%d,%d,av%d,av%d",
                    nc, no, lg+1, lo+1, nc, no);
            putid(momstr);
        } else if(c == 'v') {
            sprintf(momstr, "lt%d,lt%d,%d,%d",
                    nc, no, lg+1, lo+1);
            putid(momstr);

            sgn = cmisgn(vltbl[nc].vl[lg]->i);
            cmicwgt(momstr1, vltbl[nc].vl[lg]->i, sgn,
                    vltbl[nc].vl[lg]->p, "eu");
            putid(momstr1);

            sprintf(momstr, "av%d,av%d", nc, no);
            putid(momstr);
        } else {
            sprintf(momstr, "lt%d,lt%d,%d,%d",
                    nc, no, lg+1, lo+1);
            putid(momstr);

            sgn = cmisgn(vltbl[nc].vl[lg]->i);
            cmicwgt(momstr1, vltbl[nc].vl[lg]->i, sgn,
                    vltbl[nc].vl[lg]->p, "ew");
            putid(momstr1);

            sprintf(momstr, "av%d,av%d", nc, no);
            putid(momstr);

        }
        ftcrans(nn, vltbl[nn].vl, vltbl[nn].size);

        putend();
    }
}

/*==================================================================*/
Private Void    ftclpord(NoArg)
/*------------------------------------------------------------------*
 * Connect vertices by i-line.
 *------------------------------------------------------------------*/
{
    int n, nc, l;
    string prf = cprefix;

    /* find a node "nc" to be connected */
    nc = -1;
    for(n = 0; n < nvltbl; n++) {
        if(vltbl[n].size <= 0) {
            continue;
        }
        nc = n;
        break;
    }
    if(nc < 0) {
        fprintf(Stderr,"ftclpord : inconsistent connection\n");
        ftvldmp();
        exit(1);
    }
    fprintf(tfp1,"\n");
    for(l = 0; l < vltbl[nc].size; l++) {
        fprintf(tfp1,"      indexg(%d) = %d\n",
                l+1, vltbl[nc].vl[l]->i);
    }
    if(tishika) {
       fprintf(tifp,"indexg := ");
       for(l = 0; l < vltbl[nc].size; l++) {
           fprintf(tifp,"%d ",vltbl[nc].vl[l]->i);
       }
       fprintf(tifp,"\n");
    }

    if(!ucode) {
      if(cpol) {
        for(l = 0; l < vltbl[nc].size; l++) {
            n = vltbl[nc].vl[l]->i;
            if(cmphoton(n-1)) {
                fprintf(tfp1,"\n      if(jcpol(%d).ne.0) ", n);
                if(!vcode) {
                    fprintf(tfp1,"call smcpol(%d, lt, av)\n", l+1);
                } else {
                    fprintf(tfp1,"call smcpol(nsampl, %d, lt, av)\n", l+1);
                }
            }
        }
      }
    }

    fprintf(tfp1,"\n      ");

    if(!multiprc) {
        strcpy(momstr, "ampord");
    } else {
        sprintf(momstr, "a%smpord", prf);
    }

    /*  CF  */
    if(colorv == 0) {
        if(!vcode) {
            fprintf(tfp1,"call %s(lt, av, indexg, ag(0,jgraph))\n", momstr);
        } else {
            fprintf(tfp1,"call %s(nsampl, lt, av, indexg, ag(1,0,jgraph))\n",
                    momstr);
        }
    } else {
      if(!vcode) {
        if(!ucode) {
          fprintf(tfp1,"call %s(lt, av, indexg, agcwrk)\n\n", momstr);
        } else {
          fprintf(tfp1,"call %s(nextn, lag, lt, av, indexg, agcwrk)\n\n", momstr);
        }
      } else {
        if(!ucode) {
           fprintf(tfp1,"call %s(nsampl, lt, av, indexg, agcwrk)\n\n", momstr);
        } else {
           fprintf(tfp1,"call %s(nextn, lag, nsampl, lt, av, indexg, agcwrk)\n\n", momstr);
        }
      }
    }
}
/*==================================================================*/
Public Void    ftcfamp(NoArg)
/*------------------------------------------------------------------*
 * make color-base amplitude (2000.1.28 separated from flclpord)
 *------------------------------------------------------------------*/
{
    int i;
    string prf = cprefix;

      if(!vcode) {
          fprintf(tfp1,"      ancp(jgraph) = 0.0d0\n");
      } else {
          fprintf(tfp1,"      do 400 j = 1 , nsampl\n");
          fprintf(tfp1,"         ancp(j,jgraph) = 0.0d0\n");
          fprintf(tfp1,"  400 continue\n");
      }
      fprintf(tfp1,"*     nbase = %d\n",gcbase->nbase);

      if(gcbase->nbase == 1) {
        if(gcbase->nexcolor == 1) {
/*           fprintf(tfp1,"      cwgt = %fd0\n",gcolgr->wgt[0]);*/


           if((gcolgr->wgt[0]).den == 1) {
              fprintf(tfp1,"      cwgt = %d.d0\n",
                  (gcolgr->wgt[0]).num);
           } else {
              fprintf(tfp1,"      cwgt = %d.d0/%d.d0\n",
                  (gcolgr->wgt[0]).num,(gcolgr->wgt[0]).den);
           }

           fprintf(tfp1,"      do 500 ih = 0 , l%sag-1\n", prf);
           if(!vcode) {
               fprintf(tfp1,"         atmp    = cwgt*agcwrk(ih)*aprop\n");
               if(jcolmb && isschan) {
                   fprintf(tfp1,"         agc(ih,0) = agc(ih,0) + colmbf*atmp\n");
               } else {
                   fprintf(tfp1,"         agc(ih,0) = agc(ih,0) + atmp\n");
               }
           } else {
               fprintf(tfp1,"      do 500 j = 1 , nsampl\n");
               fprintf(tfp1,"         atmp    = cwgt*agcwrk(j,ih)*aprop(j)\n");
               if(jcolmb && isschan) {
                   fprintf(tfp1,"         ");
                   fprintf(tfp1,"agc(j,ih,0) = agc(j,ih,0) + colmbf*atmp\n");
               } else {
                   fprintf(tfp1,"         ");
                   fprintf(tfp1,"agc(j,ih,0) = agc(j,ih,0) + atmp\n");
               }
           }
/*      } else if(gcbase->nexcolor == 0) {*/
        } else {
           fprintf(tfp1,"      do 500 ih = 0 , l%sag-1\n", prf);
           if(!vcode) {
               fprintf(tfp1,"         atmp    = agcwrk(ih)*aprop\n");
               if(jcolmb && isschan) {
                   fprintf(tfp1,"         agc(ih,0) = agc(ih,0) + colmbf*atmp\n");
               } else {
                   fprintf(tfp1,"         agc(ih,0) = agc(ih,0) + atmp\n");
               }
           } else {
               fprintf(tfp1,"      do 500 j = 1 , nsampl\n");
               fprintf(tfp1,"         atmp    = agcwrk(j,ih)*aprop(j)\n");
               if(jcolmb && isschan) {
                   fprintf(tfp1,"         ");
                   fprintf(tfp1,"agc(j,ih,0) = agc(j,ih,0) + colmbf*atmp\n");
               } else {
                   fprintf(tfp1,"         ");
                   fprintf(tfp1,"agc(j,ih,0) = agc(j,ih,0) + atmp\n");
               }
           }
        }
      } else {
#ifdef CINLINE
#else
        for(i=0;i<gcbase->nbase;i++) {
           if((gcolgr->wgt[i]).den == 1) {
              fprintf(tfp1,"      cwgt(%d) = %d.d0\n",
                  i,(gcolgr->wgt[i]).num);
           } else {
              fprintf(tfp1,"      cwgt(%d) = %d.d0/%d.d0\n",
                  i,(gcolgr->wgt[i]).num,(gcolgr->wgt[i]).den);
           }
        }
        fprintf(tfp1,"      do 500 ic = 0 , ncbase-1\n");
#endif
        fprintf(tfp1,"      do 500 ih = 0 , l%sag-1\n", prf);
        if(!vcode) {
#ifdef CINLINE
            fprintf(tfp1,"         atmp       =  agcwrk(ih)*aprop\n");
            for(i=0;i<gcbase->nbase;i++) {
               if((gcolgr->wgt[i]).den == 1) {
                  if( (gcolgr->wgt[i]).num != 0 ) {
                      fprintf(tfp1,"         agc(ih,%d) = agc(ih,%d)",i,i);
                      if( (gcolgr->wgt[i]).num == 1 ) {
                          fprintf(tfp1," + atmp\n");
                      } else {
                          fprintf(tfp1," + (%d)*atmp\n",gcolgr->wgt[i].num);
                      }
                  }
               } else {
                  fprintf(tfp1,"         agc(ih,%d) = agc(ih,%d)",i,i);
                  fprintf(tfp1," + (%d/%d.d0)*atmp\n",
                      (gcolgr->wgt[i]).num,(gcolgr->wgt[i]).den);
               }
            }
#else
            fprintf(tfp1,"         atmp       =  cwgt(ic)*agcwrk(ih)*aprop\n");
            if(jcolmb && isschan) {
                fprintf(tfp1,"         agc(ih,ic) = ");
                fprintf(tfp1,"agc(ih,ic) + colmbf*atmp\n");
            } else {
                fprintf(tfp1,"         agc(ih,ic) = agc(ih,ic) + atmp\n");
            }
#endif
        } else {
            fprintf(tfp1,"      do 500 j = 1 , nsampl\n");
            fprintf(tfp1,"         atmp =  cwgt(ic)*agcwrk(j,ih)*aprop(j)\n");
            if(jcolmb && isschan) {
                fprintf(tfp1,"         agc(j,ih,ic) = ");
                fprintf(tfp1,"agc(j,ih,ic) + colmbf*atmp\n");
            } else {
                fprintf(tfp1,"         agc(j,ih,ic) = agc(j,ih,ic) + atmp\n");
                fprintf(tfp1,"*        agc(j,ih,ic) = agc(j,ih,ic) + ");
                fprintf(tfp1,"cwgt(ic)*agcwrk(j,ih)*aprop(j)\n");
            }
        }
      }
      if(ucode) {
          fprintf(tfp1,"         if( mskhel(ih) .eq. 1 ) then\n");
      }
      if(!vcode) {
          fprintf(tfp1,"         ancp(jgraph) = ");
          fprintf(tfp1,"ancp(jgraph) + atmp*conjg(atmp)\n");
      } else {
          fprintf(tfp1,"         ancp(j,jgraph) = ");
          fprintf(tfp1,"ancp(j,jgraph) + atmp*conjg(atmp)\n");
      }
      if(ucode) {
          fprintf(tfp1,"         endif\n");
      }
      fprintf(tfp1,"  500 continue\n");

}

/*==================================================================*/
Private Void    ftdprop(NoArg)
/*------------------------------------------------------------------*
 * Invert the numerator of the propagators.
 *------------------------------------------------------------------*/
{
    int sgn, sym;
    long num, den;

    sym = cgraph->sfactor;
    if(majorana) {
        if(vsign < 0) {
            fprintf(tfp1, "\n      vsn = - 1.0d0");
            vsign = - vsign;
        } else {
            fprintf(tfp1, "\n      vsn = + 1.0d0");
        }
    }
    if(sym < 0) {
        fprintf(tfp1, "\n      sym = - ");
        sym = - sym;
    } else {
        fprintf(tfp1, "\n      sym = + ");
    }
    if(majorana) {
        fprintf(tfp1, "vsn*");
    }
    if(sym == 1) {
        fprintf(tfp1, "1.0d0\n");
    } else {
        fprintf(tfp1, "1.0d0/%d.0d0\n", sym);
    }

    if(cprocess->extcol <= 0) {
        cfactor(&num, &den);      /* new version ?? */
        /* reset the current graph */
        cgraph = dareadf(cnprc, cngraph);
/*
        cgraph = daread(cnprc, cngraph);
*/
        ilmom();
        nilines = cgraph->nilines;

        if(den < 0) {
            den *= -1;
            sgn = -1;
        } else {
            sgn = 1;
        }
        if(num < 0) {
            num *= -1;
            sgn *= -1;
        }
        fprintf(tfp1, "      cf  = ");
        if(sgn < 0) {
            fprintf(tfp1, "- ");
        } else {
            fprintf(tfp1, "+ ");
        }
        if(den == 1) {
            fprintf(tfp1, "%ld.0d0\n", num);
        } else {
            fprintf(tfp1, "%ld.0d0/%ld.0d0\n", num, den);
        }
        if(colorv == 0) {
            if(!vcode) {
                fprintf(tfp1, "      aprop(jgraph) = cf*sym/aprop(jgraph)\n");
            } else {
                fprintf(tfp1, "      do 300 j = 1, nsampl\n");
                fprintf(tfp1, "         aprop(j,jgraph) = cf*sym/");
                fprintf(tfp1, "aprop(j,jgraph)\n");
                fprintf(tfp1, "  300 continue\n");
            }
        } else {
            if(!vcode) {
                fprintf(tfp1, "      aprop         = cf*sym/aprop\n");
            } else {
                fprintf(tfp1, "      do 300 j = 1, nsampl\n");
                fprintf(tfp1, "         aprop(j)  = cf*sym/aprop(j)\n");
                fprintf(tfp1, "  300 continue\n");
            }
        }
    } else {
        if(colorv == 0) {
            if(!vcode) {
                fprintf(tfp1, "      aprop(jgraph) = sym/aprop(jgraph)\n");
            } else {
                fprintf(tfp1, "      do 300 j = 1, nsampl\n");
                fprintf(tfp1, "         aprop(j,jgraph) = ");
                fprintf(tfp1, "sym/aprop(j,jgraph)\n");
                fprintf(tfp1, "  300 continue\n");
            }
        } else {
            if(!vcode) {
                fprintf(tfp1, "      aprop         = sym/aprop\n");
            } else {
                fprintf(tfp1, "      do 300 j = 1, nsampl\n");
                fprintf(tfp1, "         aprop(j)  = sym/aprop(j)\n");
                fprintf(tfp1, "  300 continue\n");
            }
        }
    }

}

/*********************************************************************
 *    4. End of output
 */

/*********************************************************************
 *    5. Other files
 */
/*==================================================================*/
Private Void    ftamporp(pc)
int     pc;
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    string prf;

    if(pc < 0) {
        prf = cprefix;
        sprintf(momstr, "ampord");
    } else {
        prf = reprocp[pc];
        sprintf(momstr, "a%smpord", prf);
    }
    sprintf(momstr1, "%s.f", momstr);
    cmopend(ROFORT, prf, momstr1);

    cmsepar('*');
    if(!vcode) {
        if(!ucode) {
        fprintf(redfp,"      subroutine %s(lvt, avt, iord, at)\n",
                momstr);
        } else {
        fprintf(redfp,"      subroutine %s(nextn, lag, lvt, avt, iord, at)\n",
                momstr);
        }
    } else {
        if(!ucode) {
        fprintf(redfp,"      subroutine %s(nsampl, lvt, avt, ",
                momstr);
        fprintf(redfp,"iord, at)\n");
        } else {
        fprintf(redfp,"      subroutine %s(nsampl, nextn, lag, lvt, avt, ",
                momstr);
        fprintf(redfp,"iord, at)\n");
        }
    }
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");

    if(vcode) {
        fprintf(redfp,"      include 'inclvs.h'\n");
    }
    if(!ucode) fprintf(redfp,"      include 'incl%s1.h'\n", prf);
    fprintf(redfp,"      include 'inclk.h'\n");
    fprintf(redfp,"      integer    lvt(0:n%sextn), iord(n%sextn)\n",
            prf, prf);
    if(!vcode) {
        fprintf(redfp,"      complex*16 avt(0:l%sag-1), at(0:l%sag-1)\n",
                prf, prf);
    } else {
        fprintf(redfp,"      complex*16 avt(n_vector,0:l%sag-1), ", prf);
        fprintf(redfp,"at(n_vector,0:l%sag-1)\n", prf);
    }
    fprintf(redfp,"*\n");
    if(!ucode) {
      fprintf(redfp,"      integer j(n%sextn), jv(n%sextn)\n", prf, prf);
    } else {
      fprintf(redfp,"      integer j(MXEXTN), jv(MXEXTN)\n");
    }
    cmsepar('-');

    fprintf(redfp,"      ibas  = 1\n");
    fprintf(redfp,"      do 10 i = 1, n%sextn\n", prf);
    fprintf(redfp,"        jv(iord(i)) = ibas\n");
    fprintf(redfp,"        ibas    = ibas*lvt(i)\n");
    if(!ucode) {
        fprintf(redfp,"        j(i)    = jhs(i)\n");
    } else {
        fprintf(redfp,"        j(i)    = maxjhs(i)\n");
    }
    fprintf(redfp,"   10 continue\n");
    fprintf(redfp,"\n");

    fprintf(redfp,"      it = 0\n");
    fprintf(redfp,"  100 continue\n");
    fprintf(redfp,"        iv = 0\n");
    fprintf(redfp,"        do 110 i = 1, n%sextn\n", prf);
    fprintf(redfp,"          iv = iv +   jv(i)*j(i)\n");
    fprintf(redfp,"  110   continue\n");
    fprintf(redfp,"\n");

    if(!vcode) {
        fprintf(redfp,"        at(it) = avt(iv)\n");
    } else {
        fprintf(redfp,"        do 115 jj = 1, nsampl\n");
        fprintf(redfp,"           at(jj,it) = avt(jj,iv)\n");
        fprintf(redfp,"  115   continue\n");
    }
    fprintf(redfp,"        it = it + 1\n");
    fprintf(redfp,"\n");

    fprintf(redfp,"        ii = 1\n");
    fprintf(redfp,"  120   continue\n");
    fprintf(redfp,"          j(ii) = j(ii) + 1\n");
    if(!ucode) {
    fprintf(redfp,"          if(j(ii).gt.jhe(ii)) then\n");
    fprintf(redfp,"            j(ii) = jhs(ii)\n");
    } else {
    fprintf(redfp,"          if(j(ii).gt.maxjhe(ii)) then\n");
    fprintf(redfp,"            j(ii) = maxjhs(ii)\n");
    }
    fprintf(redfp,"            ii = ii + 1\n");
    fprintf(redfp,"            if(ii.le.n%sextn) then\n", prf);
    fprintf(redfp,"              goto 120\n");
    fprintf(redfp,"            else\n");
    fprintf(redfp,"              goto 190\n");
    fprintf(redfp,"            endif\n");
    fprintf(redfp,"          endif\n");
    fprintf(redfp,"        goto 100\n");
    fprintf(redfp,"  190 continue\n");
    fprintf(redfp,"      return\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}

#ifdef SMCPOL
/*==================================================================*/
Private Void    ftsmcpol(NoArg)
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    string prf = cprefix;

/* tk
    sprintf(momstr, "s%smcpol.f", prf);
*/
    sprintf(momstr, "smcpol.f");
    cmopend(ROFORT, prf, momstr);
    cmsepar('*');
/* old
    fprintf(redfp,"      subroutine smcpol(la, lvt, avt)\n");
*/
    fprintf(redfp,"      subroutine smcpol(nextn, lag, la, lvt, avt)\n");
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");
/*o fprintf(redfp,"      include 'incl%s1.h'\n", prf);*/
    fprintf(redfp,"      include 'inclk.h'\n");
/*o fprintf(redfp,"      integer    lvt(0:n%sextn)\n", prf);*/
    fprintf(redfp,"      integer    lvt(0:nextn)\n");
/*o fprintf(redfp,"      complex*16 avt(0:l%sag-1)\n", prf);*/
    fprintf(redfp,"      complex*16 avt(0:lag-1)\n");
    fprintf(redfp,"*\n");
/*  fprintf(redfp,"      integer j(n%sextn), jv(n%sextn)\n", prf, prf);*/
    fprintf(redfp,"      integer j(MXEXTN), jv(MXEXTN)\n");
    fprintf(redfp,"      complex*16 zi, zr, ztv, ztw\n");
    fprintf(redfp,"      data ifst/0/\n");
    fprintf(redfp,"      save ifst, zr, zi\n");
    cmsepar('-');
    fprintf(redfp,"      if(ifst.eq.0) then\n");
    fprintf(redfp,"        ifst = 1\n");
    fprintf(redfp,"        r2 = 1.0d0/sqrt(2.0d0)\n");
    fprintf(redfp,"        zr = dcmplx(1.0d0,0.0d0)*r2\n");
    fprintf(redfp,"        zi = dcmplx(0.0d0,1.0d0)*r2\n");
    fprintf(redfp,"      endif\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      ja = la\n");
    fprintf(redfp,"      if(lvt(ja) .ne. 2) then\n");
    fprintf(redfp,"        write(*,*) 'smcpol: cannot calculate ");
    fprintf(redfp,"circlar polarization'\n");
    fprintf(redfp,"        write(*,*) 'lvt = ', lvt\n");
    fprintf(redfp,"        write(*,*) 'lvt(',ja,') = ', lvt(ja), ' <> ");
    fprintf(redfp,"2'\n");
    fprintf(redfp,"        stop\n");
    fprintf(redfp,"      endif\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      ibas  = 1\n");
/*o fprintf(redfp,"      do 10 i = 1, n%sextn\n", prf);*/
    fprintf(redfp,"      do 10 i = 1, nextn\n");
    fprintf(redfp,"        jv(i) = ibas\n");
    fprintf(redfp,"        ibas  = ibas*lvt(i)\n");
    fprintf(redfp,"        j(i)  = 0\n");
    fprintf(redfp,"   10 continue\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"  100 continue\n");
    fprintf(redfp,"        iv = 0\n");
/*o fprintf(redfp,"        do 110 i = 1, n%sextn\n", prf);*/
    fprintf(redfp,"        do 110 i = 1, nextn\n");
    fprintf(redfp,"          iv = iv +   jv(i)*j(i)\n");
    fprintf(redfp,"  110   continue\n");
    fprintf(redfp,"        iw = iv + jv(ja)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"        ztv =   zr*avt(iv) + zi*avt(iw)\n");
    fprintf(redfp,"        ztw =   zr*avt(iv) - zi*avt(iw)\n");
    fprintf(redfp,"        avt(iv) = ztv\n");
    fprintf(redfp,"        avt(iw) = ztw\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"c       write(*,*) 'smcpol:', j\n");
    fprintf(redfp,"c       write(*,*) 'smcpol:', iv, iw, avt(iv)\n");
    fprintf(redfp,"c       write(*,*) 'smcpol:', iw, iv, avt(iw)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"        ii = 1\n");
    fprintf(redfp,"  120   continue\n");
    fprintf(redfp,"          if(ii .eq. ja) then\n");
    fprintf(redfp,"            ii = ii + 1\n");
/*o fprintf(redfp,"            if(ii.gt.n%sextn) then\n", prf);*/
    fprintf(redfp,"            if(ii.gt.nextn) then\n");
    fprintf(redfp,"              goto 190\n");
    fprintf(redfp,"            endif\n");
    fprintf(redfp,"          endif\n");
    fprintf(redfp,"          j(ii) = j(ii) + 1\n");
    fprintf(redfp,"          if(j(ii).ge.lvt(ii)) then\n");
    fprintf(redfp,"            j(ii) = 0\n");
    fprintf(redfp,"            ii = ii + 1\n");
/*o fprintf(redfp,"            if(ii.le.n%sextn) then\n", prf);*/
    fprintf(redfp,"            if(ii.le.nextn) then\n");
    fprintf(redfp,"              goto 120\n");
    fprintf(redfp,"            else\n");
    fprintf(redfp,"              goto 190\n");
    fprintf(redfp,"            endif\n");
    fprintf(redfp,"          endif\n");
    fprintf(redfp,"        goto 100\n");
    fprintf(redfp,"  190 continue\n");
    fprintf(redfp,"      return\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}
#endif

/*==================================================================*/
Private Void    ftampsup(pc)
int     pc;
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    string prf;

    if(pc < 0) {
        prf = cprefix;
        sprintf(momstr, "ampsum");
    } else {
        prf      = reprocp[pc];
        sprintf(momstr, "a%smpsum", prf);
    }
    sprintf(momstr1, "%s.f", momstr);
    cmopend(ROFORT, prf, momstr1);

    cmsepar('*');
    if(!vcode) {
        fprintf(redfp,"      subroutine %s(ans)\n", momstr1);
    } else {
        fprintf(redfp,"      subroutine %s(nsampl,ans)\n", momstr1);
    }
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");

    if(vcode) {
        fprintf(redfp,"      include 'inclvs.h'\n");
    }
    fprintf(redfp,"      include 'incl%s1.h'\n", prf);
    fprintf(redfp,"      include 'inclk.h'\n");
    if(!vcode) {
        fprintf(redfp,"      real*8     ans\n");
        fprintf(redfp,"      complex*16 anc, amp, ampp(n%sgraph)\n", prf);
    } else {
        fprintf(redfp,"      real*8     ans(n_vector)\n");
        fprintf(redfp,"      complex*16 anc(n_vector), amp(n_vector),");
        fprintf(redfp," ampp(n%sgraph)\n", prf);
    }
    cmsepar('-');

    fprintf(redfp,"      jsum = 1\n");
    fprintf(redfp,"      do 10 ie = 1, n%sextn\n", prf);
    fprintf(redfp,"        jsum = jsum*(jhe(ie) - jhs(ie) + 1)\n");
    fprintf(redfp,"   10 continue\n");
    fprintf(redfp,"\n");

    if(!vcode) {
        fprintf(redfp,"      anc = 0\n");
    } else {
        fprintf(redfp,"      do 15 j = 1, nsampl\n");
        fprintf(redfp,"         anc(j) = 0\n");
        fprintf(redfp,"   15 continue\n");
    }

    fprintf(redfp,"      do 20 ig = 1, jgraph\n");
    if(!vcode) {
        fprintf(redfp,"         ancp(ig) = 0.0d0\n");
    } else {
        fprintf(redfp,"      do 20 j = 1, nsampl\n");
        fprintf(redfp,"         ancp(j,ig) = 0.0d0\n");
    }
    fprintf(redfp,"   20   continue\n");
    fprintf(redfp,"\n");

    fprintf(redfp,"*     sum over helicities\n");
    fprintf(redfp,"      do 100 ind = 0, jsum - 1\n");
    fprintf(redfp,"\n");

    fprintf(redfp,"*       sum over graphs\n");
    if(!qcd) {
       if(!vcode) {
           fprintf(redfp,"        amp  = 0.0d0\n");
       } else {
           fprintf(redfp,"        do 110 j = 1, nsampl\n");
           fprintf(redfp,"           amp(j)  = 0.0d0\n");
           fprintf(redfp,"  110   continue\n");
       }
    }

    fprintf(redfp,"        do 200 ig = 1, jgraph\n");
    if(!vcode) {
        fprintf(redfp,"          ampp(ig) = aprop(ig)*ag(ind,ig)\n");
    } else {
        fprintf(redfp,"        do 200 j = 1, nsampl\n");
        fprintf(redfp,"           ampp(j,ig) = aprop(j,ig)*");
        fprintf(redfp,"ag(j,ind,ig)\n");
    }
    if(!qcd) {
       if(!vcode) {
           fprintf(redfp,"          amp = amp + ampp(ig)\n");
       } else {
           fprintf(redfp,"          amp(j) = amp(j) + ampp(j,ig)\n");
       }
    }
    fprintf(redfp,"  200   continue\n");
    fprintf(redfp,"\n");

    fprintf(redfp,"        do 300 ig = 1, jgraph\n");
    if(!vcode) {
        fprintf(redfp,"          ancp(ig) = ancp(ig) + ampp(ig)");
        fprintf(redfp,"*conjg(ampp(ig))\n");
    } else {
        fprintf(redfp,"        do 300 j = 1, nsampl\n");
        fprintf(redfp,"          ancp(j,ig) = ancp(j,ig) + ampp(j,ig)");
        fprintf(redfp,"*conjg(ampp(ig))\n");
    }
    fprintf(redfp,"  300   continue\n");

    if(cprocess->extcol) {
        fprintf(redfp,"        do 400 ig = 1, jgraph\n");
        fprintf(redfp,"        do 400 jg = 1, jgraph\n");
        if(!vcode) {
            fprintf(redfp,"          anc = anc +");
            fprintf(redfp,"  cf(igraph(ig),igraph(jg))\n");
            fprintf(redfp,"     &               ");
            fprintf(redfp,"*ampp(ig)*conjg(ampp(jg))\n");
        } else {
            fprintf(redfp,"        do 400 j = 1, nsampl\n");
            fprintf(redfp,"          anc(j) = anc(j) +");
            fprintf(redfp,"  cf(igraph(ig),igraph(jg))\n");
            fprintf(redfp,"     &               ");
            fprintf(redfp,"*ampp(j,ig)*conjg(ampp(j,jg))\n");
        }
        fprintf(redfp,"  400 continue\n");
    } else {
        if(!vcode) {
            fprintf(redfp,"      anc = anc + amp*conjg(amp)\n");
        } else {
            fprintf(redfp,"        do 400 j = 1, nsampl\n");
            fprintf(redfp,"           anc(j) = anc(j) + ");
            fprintf(redfp,"amp(j)*conjg(amp(j))\n");
            fprintf(redfp,"  400 continue\n");
        }
    }
    fprintf(redfp,"  100 continue\n");
    if(!vcode) {
        fprintf(redfp,"      ans = anc\n");
    } else {
        fprintf(redfp,"      do 500 j = 1, nsampl\n");
        fprintf(redfp,"         ans(j) = anc(j)\n");
        fprintf(redfp,"  500 continue\n");
    }
    fprintf(redfp,"\n");

    fprintf(redfp,"      return\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}

/*==================================================================*/
Private Void    ftampsmp(pc)
int     pc;
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    int ic1, ic2, nl;
    string prf;

    if(pc < 0) {
        prf = cprefix;
        sprintf(momstr, "ampsum");
    } else {
        prf = reprocp[pc];
        sprintf(momstr, "a%smpsum", prf);
    }
    sprintf(momstr1, "%s.f", momstr);
    cmopend(ROFORT, prf, momstr1);

    cmsepar('*');
    fprintf(redfp,"*     CF version \n");
    cmsepar('*');
    if(!vcode) {
        fprintf(redfp,"      subroutine %s(ans)\n", momstr);
    } else {
        fprintf(redfp,"      subroutine %s(nsampl,ans)\n", momstr);
    }
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");
    if(vcode) {
        fprintf(redfp,"      include 'inclvs.h'\n");
    }
    fprintf(redfp,"      include 'incl%s1.h'\n", prf);
    fprintf(redfp,"      include 'inclk.h'\n");
    if(!vcode) {
        fprintf(redfp,"      real*8     ans\n");
        fprintf(redfp,"      complex*16 anc\n");
    } else {
        fprintf(redfp,"      real*8     ans(n_vector)\n");
        fprintf(redfp,"      complex*16 anc(n_vector)\n");
    }
    cmsepar('-');

    if(!ucode){
    fprintf(redfp,"      jsum = 1\n");
    fprintf(redfp,"      do 10 ie = 1, n%sextn\n", prf);
    fprintf(redfp,"        jsum = jsum*(jhe(ie) - jhs(ie) + 1)\n");
    fprintf(redfp,"   10 continue\n");
    fprintf(redfp,"\n");
    }

    if(!vcode) {
        fprintf(redfp,"      anc = 0\n");
    } else {
        fprintf(redfp,"      do 15 j = 1, nsampl\n");
        fprintf(redfp,"         anc(j) = 0\n");
        fprintf(redfp,"   15 continue\n");
    }
    fprintf(redfp,"\n");

    fprintf(redfp,"*     sum over helicities\n");
    if(ucode) {
       fprintf(redfp,"      do 100 ind = 0, lag  - 1\n");
       fprintf(redfp,"       if( mskhel(ind) .eq. 1 ) then\n");
    } else {
       fprintf(redfp,"      do 100 ind = 0, jsum - 1\n");
       if( vcode ) {
       fprintf(redfp,"      do 100 j = 1, nsampl\n");
       }
    }
    fprintf(redfp,"\n");

    fprintf(redfp,"*       sum over color bases\n");

    nl = 0;
    if(gcbase->nbase > 1) {
        if(gcbase->nbase < 12) {
        if(vcode) {
            fprintf(redfp,"        do 100 j = 1, nsampl\n");
        }
        /* inline expansion */
        for(ic1=0;ic1<gcbase->nbase;ic1++) {
            for(ic2=0;ic2<gcbase->nbase;ic2++) {
                if(nl%10 == 0) {
                    if(!vcode) {
                        fprintf(redfp,"        anc = anc ");
                    } else {
                        fprintf(redfp,"          anc(j) = anc(j) ");
                    }
                } else {
                    fprintf(redfp,"     .              ");
                }
                if(!vcode) {
                    fprintf(redfp," + cfmtx(%d,%d)*agc(ind,%d)*",
                            ic1,ic2,ic1);
                    fprintf(redfp,"conjg(agc(ind,%d))\n",ic2);
                } else {
                    fprintf(redfp," + cfmtx(%d,%d)*agc(j,ind,%d)*",
                            ic1,ic2,ic1);
                    fprintf(redfp,"conjg(agc(j,ind,%d))\n",ic2);
                }
                nl++;
            }
        }
        } else {
            /* inline expansion */
            fprintf(redfp,"      do 100 ic1 = 0, ncbase-1\n");
            fprintf(redfp,"      do 100 ic2 = 0, ncbase-1\n");
            if(!vcode) {
                 fprintf(redfp,"        anc = anc ");
                 fprintf(redfp,"+ cfmtx(ic1,ic2)*agc(ind,ic1)*");
                 fprintf(redfp,"conjg(agc(ind,ic2))\n");
            } else {
                 fprintf(redfp,"      do 100 j = 1, nsampl\n");
                 fprintf(redfp,"          anc(j) = anc(j)\n");
                 fprintf(redfp,"     .  + cfmtx(ic1,ic2)*agc(j,ind,ic1)*");
                 fprintf(redfp,"conjg(agc(j,ind,ic2))\n");
            }
        }
    } else if(gcbase->nbase == 1) {
        if(!vcode) {
            fprintf(redfp,"        anc = anc ");
            fprintf(redfp," + agc(ind,0)*conjg(agc(ind,0))\n");
        } else {
            fprintf(redfp,"          anc(j) = anc(j) ");
            fprintf(redfp," + agc(j,ind,0)*conjg(agc(j,ind,0))\n");
        }
    }
    if(ucode) {
       fprintf(redfp,"       endif\n");
    }
    fprintf(redfp,"  100 continue\n");
    if(gcbase->nbase > 1) {
        if(!vcode) {
            fprintf(redfp,"      ans = anc\n");
        } else {
            fprintf(redfp,"      do 200 j = 1, nsampl\n");
            fprintf(redfp,"         ans(j) = anc(j)\n");
            fprintf(redfp,"  200 continue\n");
        }
    } else if(gcbase->nbase == 1) {
        if(gcbase->nexcolor == 0) {
            if(!vcode) {
                fprintf(redfp,"      ans = cfmtx(0,0)*anc\n");
                fprintf(redfp,"*     ans = anc\n");
            } else {
                fprintf(redfp,"      do 200 j = 1, nsampl\n");
                fprintf(redfp,"         ans(j) = cfmtx(0,0)*anc(j)\n");
                fprintf(redfp,"  200 continue\n");
            }
        } else if(gcbase->nexcolor == 1) {
            if(!vcode) {
                fprintf(redfp,"      ans = cfmtx(0,0)*anc\n");
            } else {
                fprintf(redfp,"      do 200 j = 1, nsampl\n");
                fprintf(redfp,"         ans(j) = cfmtx(0,0)*anc(j)\n");
                fprintf(redfp,"  200 continue\n");
            }
        }
    }
    fprintf(redfp,"\n");

    fprintf(redfp,"      return\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}

/*********************************************************************
 *    6. Tools
 */
/*==================================================================*/
Private Void    ftgend(NoArg)
/*------------------------------------------------------------------*
 * end of REDUCE file.
 *------------------------------------------------------------------*/
{
    fprintf(redfp, "\n");
    fprintf(redfp, "      return\n");
    fprintf(redfp, "      end\n");
}

/*==================================================================*/
Private Void    ftvlts(NoArg)
/*------------------------------------------------------------------*
 * Initialize vltbl
 *------------------------------------------------------------------*/
{
    node    n;
    int     i;

    nvltbl = cgraph->nnodes;
    nvlusd = nextern;
    for(n = 0; n < nvltbl; n++) {
        /* null node */
        if(ndnode(n) == NULL) {
            vltbl[n].size = 0;
        } else if(ndtype(n) != NDTREEV) {
            vltbl[n].size = 0;
        } else {
            vltbl[n].size = ndlegs(n);
            if((vltbl[n].vl =
              (vlegt**) malloc(vltbl[n].size*sizeof(vlegt*)))
              == NULL) {
                fprintf(Stderr, "ftvlts: malloc failed\n");
                exit(1);
            }
            for(i = 0; i < ndlegs(n); i++) {
                if((vltbl[n].vl[i] =
                  (vlegt*) malloc(sizeof(vlegt)))
                  == NULL) {
                    fprintf(Stderr, "ftvlts: malloc failed\n");
                    exit(1);
                }
            }
            ftmlist(n, vltbl[n].vl, vltbl[n].size);
        }
    }
}

/*==================================================================*/
Private Void    ftvlcp(vl1, vl2)
vlegt   *vl1, *vl2;
/*------------------------------------------------------------------*
 * Copy *vl2 to *vl1
 *------------------------------------------------------------------*/
{
    vl1->p = vl2->p;
    vl1->t = vl2->t;
    vl1->i = vl2->i;
    vl1->m = vl2->m;
    vl1->d = vl2->d;
}

/*==================================================================*/
Private Void    ftvlex(vl, l1, l2)
vlegt   **vl;
int     l1, l2;
/*------------------------------------------------------------------*
 * Exchange l1 and l2 elements of vl
 *------------------------------------------------------------------*/
{
    vlegt  *t;

    t = vl[l1];
    vl[l1] = vl[l2];
    vl[l2] = t;
}

/*==================================================================*/
Public  Void    ftvldmp(NoArg)
/*------------------------------------------------------------------*
 * Dump vltbl
 *------------------------------------------------------------------*/
{
    int n, l;

    fprintf(Stderr, "\nDump vltbl: nvltbl = %d, nvlusd = %d\n",
                    nvltbl, nvlusd);
    if(nvltbl < 0) {
        return;
    }
    for(n = 0; n < nvltbl; n++) {
        fprintf(Stderr, "%4d, size = %4d\n", n, vltbl[n].size);
        for(l = 0; l < vltbl[n].size; l++) {
            fprintf(Stderr, "p=%4d, t=%4d, i=%4d, m=%4d, d=%4d\n",
                    vltbl[n].vl[l]->p, vltbl[n].vl[l]->t,
                    vltbl[n].vl[l]->i, vltbl[n].vl[l]->m,
                    vltbl[n].vl[l]->d);
        }
    }
}

/*==================================================================*/
Private Void    ftvlte(NoArg)
/*------------------------------------------------------------------*
 * Clear vltbl
 *------------------------------------------------------------------*/
{
    node    n;

    for(n = 0; n < nvltbl; n++) {
        if(vltbl[n].size != 0) {
            free(vltbl[n].vl);
        }
    }
    nvltbl = 0;
}

/*==================================================================*/
Private Void    ftnode(NoArg)
/*------------------------------------------------------------------*
 * Output node.
 *------------------------------------------------------------------*/
{
    int   nnodes = cgraph->nnodes;
    node   n;
    vindex vt;
    int    nlg;
    vlegt  **vl;
    bool   ok;
#if DBG
    pindex p;
    int    l, sgn;
    tnode  *nd;
    tleg   *lg;
#endif


    fprintf(tfp1, "\n* Vertices (%d)\n", cgraph->nnodes);
    if(majorana) {
        ftprfl(tfp1);
    }

    nvltbl = nnodes;
    for(n = 0; n < nnodes; n++) {
        /* null node */
        if(ndnode(n) == NULL) {
            fprintf(tfp1, "      null%d", n);
        }

        /* name of the vertex */
        if(ndtype(n) != NDTREEV) {
        } else {
            vl  = vltbl[n].vl;
            nlg = vltbl[n].size;
            vt = ndvertex(n);
            if(vt == VNULL) {
                fprintf(tfp1, "       vnull(%d)", n);
                continue;
            }

            Discard ftnvtype(momstr, n, vl);

#if DBG
            fprintf(tfp1, "\n");
            for(l = 0; l < nlg; l++) {
                nd  = ndnode(n);
                lg  = ndleg(n, l);
                p   = abs(ndparticle(n, l));

                sgn = cmisgn(vl[l]->m);
                cmicmom(momstr1,vl[l]->m,sgn,"pe");
                fprintf(tfp1, "*     %d(%d): %s %s\n",
                    n, l, momstr1, pvename(nd->nvkind, vl[l]->p));
            }
#endif

            ok = True;
            if(strcmp(momstr, "ffv") == 0) {
                if(majorana) {
                    ftmffv(n, vl);
                } else {
                    ftnffv("smffv", n, vl);
                }
            } else if(strcmp(momstr, "sff") == 0) {
                if(majorana) {
                    ftmffs(n, vl);
                } else {
                    ftnffs("smffs", n, vl);
                }
            } else if(strcmp(momstr, "vvv") == 0) {
                ftnvvv("smvvv", "smggg", n, vl);
            } else if(strcmp(momstr, "svv") == 0) {
                ftnsvv("smsvv", n, vl);
            } else if(strcmp(momstr, "ssv") == 0) {
                ftnssv("smssv", n, vl);
            } else if(strcmp(momstr, "sss") == 0) {
                ftnsss("smsss", n, vl);
            } else if(strcmp(momstr, "vvvv") == 0) {
                ftnvvvv("smvvvv", "smgggg", n, vl);
            } else if(strcmp(momstr, "ssvv") == 0) {
                ftnssvv("smssvv", n, vl);
#if (VLREORD==2)
            } else if(strcmp(momstr, "svsv") == 0) {
                ftnsvsv("smsvsv", n, vl);
#endif
            } else if(strcmp(momstr, "ssss") == 0) {
                ftnssss("smssss", n, vl);
#if 1
            } else if(strcmp(momstr, "xxv") == 0) {
                ftnssv("smssv", n, vl);
#endif
            } else {
                ok = False;
            }
            if(!ok) {
                fprintf(tfp1, "      call ??%s(%d)??\n", momstr, n);
            }
        }
    }
}

/*==================================================================*/
Private Void    ftmffv(n, vl)
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output ffv vertex in case with majorana fermion.
 *------------------------------------------------------------------*/
{
    string nc;
    int    vcsign;

    nc = ftmcc("sfv", n, 0, &vcsign);

    vsign *= vcsign;

    if(strcmp(nc+3, "nnn") == 0) {
        ftnffv("smffv", n, vl);
        return;
    } else {
        ftnffv(nc, n, vl);
        return;
    }
}

/*==================================================================*/
Private Void    ftmffs(n, vl)
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output ffv vertex in case with majorana fermion.
 *------------------------------------------------------------------*/
{
    string nc;
    int    vcsign;

    nc = ftmcc("sfs", n, 1, &vcsign);

    vsign *= vcsign;

    if(strcmp(nc+3, "nnn") == 0) {
        ftnffs("smffs", n, vl);
        return;
    } else {
        ftnffs(nc, n, vl);
        return;
    }
}

/*==================================================================*/
Private Void    ftnffv(name, n, vl)
string  name;
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output ffv vertex.
 *    leg : 0=f-bar, 1=f, 2=v
 *------------------------------------------------------------------*/
{
    int    sgn;
    string nm;

    ftvlex(vl, 0, 1);

    if(vtable[ndvertex(n)]->sname != NULL) {
        ftexsubr(vtable[ndvertex(n)]->sname, n, vl, 3);
        return;
    } else if(vtable[ndvertex(n)]->chanel != NULL) {
        nm = vtable[ndvertex(n)]->chanel;
    } else {
        nm = name;
    }

    sprintf(momstr, "      call %s(", nm);
    puinit(tfp1,momstr,
                "     &           ",
                ",", ",", ")", 72, 16);
    if(vcode) {
        putid("nsampl");
    }
    fthindx(n, vl, 3);                             /* l1, l2, lv, */
    sgn = cmisgn(vl[0]->m);
    cmicwgt(momstr1, vl[0]->m, sgn, vl[0]->p, "ew");
    putid(momstr1);

    sgn = cmisgn(vl[1]->m);
    cmicwgt(momstr2, vl[1]->m, sgn, vl[1]->p, "ew");
    putid(momstr2);

    ftmassl(n, vl, 0, 1);                          /* am1, am2, */

    ftncoupl(n);                                   /* cpl, */

    sgn = cmisgn(vl[0]->m);
    cmicwgt(momstr1, vl[0]->m, sgn, vl[0]->p,"ce");
    putid(momstr1);

    sgn = cmisgn(vl[1]->m);
    cmicwgt(momstr2, vl[1]->m, sgn, vl[1]->p,"ce");
    putid(momstr2);

    sgn = cmisgn(vl[0]->m);
    cmicwgt(momstr1, vl[0]->m, sgn, vl[0]->p, "ps");
    putid(momstr1);

    sgn = cmisgn(vl[1]->m);
    cmicwgt(momstr2, vl[1]->m, sgn, vl[1]->p, "ps");
    putid(momstr2);

    sgn = cmisgn(vl[2]->m);
    cmicwgt(momstr1, vl[2]->m, sgn, vl[2]->p, "ep");
    putid(momstr1);

    ftcrans(n, vl, 3);                             /* lt, av */

    putend();
}

/*==================================================================*/
Private Void    ftnffs(name, n, vl)
string  name;
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output ffs vertex.
 *    leg : 0=f-bar, 1=f, 2=s    ???
 *------------------------------------------------------------------*/
{
    int    sgn;
    string nm;

    ftvlex(vl, 0, 2);

    if(vtable[ndvertex(n)]->sname != NULL) {
        ftexsubr(vtable[ndvertex(n)]->sname, n, vl, 3);
        return;
    } else if(vtable[ndvertex(n)]->chanel != NULL) {
        nm = vtable[ndvertex(n)]->chanel;
    } else {
        nm = name;
    }

    sprintf(momstr, "      call %s(", nm);
    puinit(tfp1,momstr,
                "     &           ",
                ",", ",", ")", 72, 16);
    if(vcode) {
        putid("nsampl");
    }

    fthindx(n, vl, 3);                             /* l1, l2, lv, */
    sgn = cmisgn(vl[0]->m);
    cmicwgt(momstr1, vl[0]->m, sgn, vl[0]->p, "ew");
    putid(momstr1);

    sgn = cmisgn(vl[1]->m);
    cmicwgt(momstr2, vl[1]->m, sgn, vl[1]->p, "ew");
    putid(momstr2);

    ftmassl(n, vl, 0, 1);                          /* am1, am2, */

    ftncoupl(n);                                   /* cpl, */

    sgn = cmisgn(vl[0]->m);
    cmicwgt(momstr1, vl[0]->m, sgn, vl[0]->p, "ce");
    putid(momstr1);

    sgn = cmisgn(vl[1]->m);
    cmicwgt(momstr2, vl[1]->m, sgn, vl[1]->p, "ce");
    putid(momstr2);

    sgn = cmisgn(vl[0]->m);
    cmicwgt(momstr1, vl[0]->m, sgn, vl[0]->p, "ps");
    putid(momstr1);

    sgn = cmisgn(vl[1]->m);
    cmicwgt(momstr2, vl[1]->m, sgn, vl[1]->p, "ps");
    putid(momstr2);

    ftcrans(n, vl, 3);                  /* lt, av */

    putend();
}

/*==================================================================*/
Private Void    ftnvvv(nam1, nam2, n, vl)
string  nam1;
string  nam2;
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output vvv vertex.
 *------------------------------------------------------------------*/
{
    int gl3 = pvgl3(n), sgn;
    string nm;

    if(vtable[ndvertex(n)]->sname != NULL) {
        ftexsubr(vtable[ndvertex(n)]->sname, n, vl, 3);
        return;
    } else if(vtable[ndvertex(n)]->chanel != NULL) {
        nm = vtable[ndvertex(n)]->chanel;
    } else if(gl3) {
        nm = nam2;
    } else {
        nm = nam1;
    }

    if(gl3) {
        sprintf(momstr, "      call %s(", nm);
        puinit(tfp1,momstr,
                    "     &            ",
                    ",", ",", ")", 72, 16);
        if(vcode) {
            putid("nsampl");
        }

        fthindx(n, vl, 3);                        /* l1, l2, lv, */

        sprintf(momstr, "%c1,%c1,%c1",
                csign(vl[0]->d), csign(vl[1]->d), csign(vl[2]->d));
        putid(momstr);

        ftncoupl(n);                                   /* cpl, */

        sgn = cmisgn(vl[0]->m);
        cmicmom(momstr1,vl[0]->m,sgn,"pe");
        putid(momstr1);

        sgn = cmisgn(vl[1]->m);
        cmicmom(momstr2,vl[1]->m,sgn,"pe");
        putid(momstr2);

        sgn = cmisgn(vl[2]->m);
        cmicmom(momstr3,vl[2]->m,sgn,"pe");
        putid(momstr3);

        sgn = cmisgn(vl[0]->m);
        cmicwgt(momstr1,vl[0]->m,sgn, vl[0]->p,"ep");
        putid(momstr1);

        sgn = cmisgn(vl[1]->m);
        cmicwgt(momstr2,vl[1]->m,sgn, vl[1]->p,"ep");
        putid(momstr2);

        sgn = cmisgn(vl[2]->m);
        cmicwgt(momstr3,vl[2]->m,sgn, vl[2]->p,"ep");
        putid(momstr3);

        ftcrans(n, vl, 3);                  /* lt, av */

        putend();
    } else {
        if(jano3v) {
            fprintf(tfp1, "      if(jano3v .eq. 0) then\n");
            sprintf(momstr, "        call %s(", nm);
            puinit(tfp1,momstr,
                        "     &             ",
                        ",", ",", ")", 72, 16);
        } else {
            sprintf(momstr, "      call %s(", nm);
            puinit(tfp1,momstr,
                        "     &           ",
                        ",", ",", ")", 72, 16);
        }
        if(vcode) {
            putid("nsampl");
        }

        fthindx(n, vl, 3);                        /* l1, l2, lv, */

        sprintf(momstr, "%c1,%c1,%c1",
                csign(vl[0]->d), csign(vl[1]->d), csign(vl[2]->d));
        putid(momstr);

        ftncoupl(n);                                   /* cpl, */

        sgn = cmisgn(vl[0]->m);
        cmicmom(momstr1,vl[0]->m,sgn,"pe");
        putid(momstr1);

        sgn = cmisgn(vl[1]->m);
        cmicmom(momstr2,vl[1]->m,sgn,"pe");
        putid(momstr2);

        sgn = cmisgn(vl[2]->m);
        cmicmom(momstr3,vl[2]->m,sgn,"pe");
        putid(momstr3);

        sgn = cmisgn(vl[0]->m);
        cmicwgt(momstr1,vl[0]->m,sgn,vl[0]->p,"ep");
        putid(momstr1);

        sgn = cmisgn(vl[1]->m);
        cmicwgt(momstr2,vl[1]->m,sgn,vl[1]->p,"ep");
        putid(momstr2);

        sgn = cmisgn(vl[2]->m);
        cmicwgt(momstr3,vl[2]->m,sgn,vl[2]->p,"ep");
        putid(momstr3);

        ftcrans(n, vl, 3);                  /* lt, av */

        putend();
        if(jano3v) {
            fprintf(tfp1, "      else\n");
            sprintf(momstr, "        call smvww(");
            puinit(tfp1,momstr,
                    "     &  ",
                    ",", ",", ")", 72, 16);
            if(vcode) {
                putid("nsampl");
            }

            fthindx(n, vl, 3);                      /* l1, l2, lv, */

            sprintf(momstr, "%c1,%c1,%c1",
                    csign(vl[0]->d), csign(vl[1]->d), csign(vl[2]->d));
            putid(momstr);

            ftncoupl(n);                                 /* cpl, */

            sgn = cmisgn(vl[0]->m);
            cmicmom(momstr1,vl[0]->m,sgn,"pe");
            putid(momstr1);

            sgn = cmisgn(vl[1]->m);
            cmicmom(momstr2,vl[1]->m,sgn,"pe");
            putid(momstr2);

            sgn = cmisgn(vl[2]->m);
            cmicmom(momstr3,vl[2]->m,sgn,"pe");
            putid(momstr3);

/* CHECK ordering of vww: 2nd,3rd are 'w' */
            if(ptable[vl[1]->p]->kfcode != 24 ||
                ptable[vl[2]->p]->kfcode != 24  ) {
                printf("Invalid ordering vww %s %s %s \n",
                       ptable[vl[0]->p]->mass, ptable[vl[1]->p]->mass,
                       ptable[vl[2]->p]->mass);
                printf("Please send a mail to tishika@minami.kek.jp\n");
                exit(1);
            }
            if(ptable[vl[0]->p]->kfcode == 22) {
                putid("1.0d0");
                putid("andka");
                putid("anlma");
                sprintf(momstr, "%s**2", ptable[vl[1]->p]->mass);
                putid(momstr);

            } else if(ptable[vl[0]->p]->kfcode == 23) {
                putid("1.0d0");
                putid("andkz");
                putid("anlmz");
                sprintf(momstr, "%s**2", ptable[vl[1]->p]->mass);
                putid(momstr);
            } else {
                printf("Invalid vector partilce vww %s\n",
                        ptable[vl[0]->p]->mass);
                printf("Please send a mail to tishika@minami.kek.jp\n");
                exit(1);
            }
            sgn = cmisgn(vl[0]->m);
            cmicwgt(momstr1,vl[0]->m,sgn,vl[0]->p,"ep");
            putid(momstr1);

            sgn = cmisgn(vl[1]->m);
            cmicwgt(momstr2,vl[1]->m,sgn,vl[1]->p,"ep");
            putid(momstr2);

            sgn = cmisgn(vl[2]->m);
            cmicwgt(momstr3,vl[2]->m,sgn,vl[2]->p,"ep");
            putid(momstr3);

/*        ftcrans(n, vl, 3);                *//* lt, av */
/*  TI 95/11/21 */
            sprintf(momstr, "lt%d,av%d",n,n);
            putid(momstr);
            putend();

            fprintf(tfp1, "      endif");
        }
    }
}

/*==================================================================*/
Private Void    ftnsvv(name, n, vl)
string  name;
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output svv vertex.
 *------------------------------------------------------------------*/
{
    int    sgn;
    string nm;

    if(vtable[ndvertex(n)]->sname != NULL) {
        ftexsubr(vtable[ndvertex(n)]->sname, n, vl, 3);
        return;
    } else if(vtable[ndvertex(n)]->chanel != NULL) {
        nm = vtable[ndvertex(n)]->chanel;
    } else {
        nm = name;
    }

    sprintf(momstr, "      call %s(", nm);
    puinit(tfp1,momstr,
                "     &            ",
                ",", ",", ")", 72, 16);
    if(vcode) {
        putid("nsampl");
    }

    fthindx(n, vl, 3);                             /* l2, l3 */

    ftncoupl(n);                                   /* cpl, */

    sgn = cmisgn(vl[1]->m);
    cmicwgt(momstr1,vl[1]->m,sgn,vl[1]->p,"ep");
    putid(momstr1);

    sgn = cmisgn(vl[2]->m);
    cmicwgt(momstr2,vl[2]->m,sgn,vl[2]->p,"ep");
    putid(momstr2);

    ftcrans(n, vl, 3);                                     /* lt, av */

    putend();
}

/*==================================================================*/
Private Void    ftnssv(name, n, vl)
string  name;
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output ssv vertex.
 *------------------------------------------------------------------*/
{
    int    sgn;
    string nm;

    if(vtable[ndvertex(n)]->sname != NULL) {
        ftexsubr(vtable[ndvertex(n)]->sname, n, vl, 3);
        return;
    } else if(vtable[ndvertex(n)]->chanel != NULL) {
        nm = vtable[ndvertex(n)]->chanel;
    } else {
        nm = name;
    }

    sprintf(momstr, "      call %s(", nm);
    puinit(tfp1,momstr,
                "     &            ",
                ",", ",", ")", 72, 16);
    if(vcode) {
        putid("nsampl");
    }

    fthindx(n, vl, 3);                             /* lv, */

    sprintf(momstr, "%c1,%c1", csign(vl[0]->d), csign(vl[1]->d));
    putid(momstr);

    ftncoupl(n);                                   /* cpl, */

    sgn = cmisgn(vl[0]->m);
    cmicmom(momstr1,vl[0]->m,sgn,"pe");
    putid(momstr1);

    sgn = cmisgn(vl[1]->m);
    cmicmom(momstr2,vl[1]->m,sgn,"pe");
    putid(momstr2);

    sgn = cmisgn(vl[2]->m);
    cmicwgt(momstr1,vl[2]->m,sgn,vl[2]->p,"ep");
    putid(momstr1);

    ftcrans(n, vl, 3);                                     /* lt, av */

    putend();
}

/*==================================================================*/
Private Void    ftnsss(name, n, vl)
string  name;
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output ssv vertex.
 *------------------------------------------------------------------*/
{
    string nm;

    if(vtable[ndvertex(n)]->sname != NULL) {
        ftexsubr(vtable[ndvertex(n)]->sname, n, vl, 3);
        return;
    } else if(vtable[ndvertex(n)]->chanel != NULL) {
        nm = vtable[ndvertex(n)]->chanel;
    } else {
        nm = name;
    }

    sprintf(momstr, "      call %s(", nm);
    puinit(tfp1,momstr,
                "     &            ",
                ",", ",", ")", 72, 16);
    if(vcode) {
        putid("nsampl");
    }

    ftncoupl(n);                                   /* cpl, */

    fthindx(n, vl, 3);                             /* lv, */

    ftcrans(n, vl, 3);                                     /* lt, av */

    putend();
}

/*==================================================================*/
Private Void    ftnvvvv(nam1, nam2, n, vl)
string  nam1;
string  nam2;
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output vvvv vertex.
 *------------------------------------------------------------------*/
{
    int    gl4 = ndgl4(n), cgl4 = pvcgl4(n), sgn;
    string nm = NULL;

    if(vtable[ndvertex(n)]->sname != NULL) {
        ftexsubr(vtable[ndvertex(n)]->sname, n, vl, 4);
        return;
    } else if(vtable[ndvertex(n)]->chanel != NULL) {
        nm = vtable[ndvertex(n)]->chanel;
    } else if(cgl4) {
        nm = nam2;
    } else if(gl4 == 0) {
        nm = nam1;
    } else if(1 <= gl4 && gl4 <= 3) {
        nm = nam2;
    } else {
        error("ftnvvvv", "illegal ndgl4");
        fprintf(Stderr, "n = %d, gl4=%d\n", n, gl4);
        erfail();
    }

    sprintf(momstr, "      call %s(", nm);
    puinit(tfp1,momstr,
                "     &            ",
                ",", ",", ")", 72, 16);
    if(vcode) {
        putid("nsampl");
    }

    fthindx(n, vl, 4);                           /* l1, l2, l3, l4 */

    ftncoupl(n);                                 /* cpl, */

    sgn = cmisgn(vl[0]->m);
    cmicwgt(momstr1, vl[0]->m,sgn, vl[0]->p, "ep");
    putid(momstr1);

    sgn = cmisgn(vl[1]->m);
    cmicwgt(momstr2, vl[1]->m,sgn, vl[1]->p, "ep");
    putid(momstr2);

    sgn = cmisgn(vl[2]->m);
    cmicwgt(momstr3, vl[2]->m,sgn, vl[2]->p, "ep");
    putid(momstr3);

    sgn = cmisgn(vl[3]->m);
    cmicwgt(momstr4, vl[3]->m,sgn, vl[3]->p, "ep");
    putid(momstr4);

    ftcrans(n, vl, 4);                                   /* lt, av */

    putend();
}

/*==================================================================*/
Private Void    ftnssvv(name, n, vl)
string  name;
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output ssss vertex.
 *------------------------------------------------------------------*/
{
    int    sgn;
    string nm;

    if(vtable[ndvertex(n)]->sname != NULL) {
        ftexsubr(vtable[ndvertex(n)]->sname, n, vl, 4);
        return;
    } else if(vtable[ndvertex(n)]->chanel != NULL) {
        nm = vtable[ndvertex(n)]->chanel;
    } else {
        nm = name;
    }

    sprintf(momstr, "      call %s(", nm);
    puinit(tfp1,momstr,
                "     &            ",
                ",", ",", ")", 72, 16);
    if(vcode) {
        putid("nsampl");
    }

    fthindx(n, vl, 4);                           /* l3, l4 */
    ftncoupl(n);                                 /* cpl, */
    sgn = cmisgn(vl[2]->m);
    cmicwgt(momstr1, vl[2]->m, sgn, vl[2]->p, "ep");
    putid(momstr1);

    sgn = cmisgn(vl[3]->m);
    cmicwgt(momstr2, vl[3]->m, sgn, vl[3]->p, "ep");
    putid(momstr2);

    ftcrans(n, vl, 4);                                   /* lt, av */

    putend();
}

#if (VLREORD==2)
/*==================================================================*/
Private Void    ftnsvsv(name, n, vl)
string  name;
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output ssss vertex.
 *------------------------------------------------------------------*/
{
    int    sgn;
    string nm;

    if(vtable[ndvertex(n)]->sname != NULL) {
        ftexsubr(vtable[ndvertex(n)]->sname, n, vl, 4);
        return;
    } else if(vtable[ndvertex(n)]->chanel != NULL) {
        nm = vtable[ndvertex(n)]->chanel;
    } else {
        nm = name;
    }

    sprintf(momstr, "      call %s(", nm);
    puinit(tfp1,momstr,
                "     &            ",
                ",", ",", ")", 72, 16);
    if(vcode) {
        putid("nsampl");
    }

    fthindx(n, vl, 4);                           /* l2, l4 */
    ftncoupl(n);                                 /* cpl, */
    sgn = cmisgn(vl[1]->m);
    cmicwgt(momstr1, vl[1]->m, sgn, vl[1]->p, "ep");
    putid(momstr1);

    sgn = cmisgn(vl[3]->m);
    cmicwgt(momstr2, vl[3]->m, sgn, vl[3]->p, "ep");
    putid(momstr2);

    ftcrans(n, vl, 4);                                   /* lt, av */

    putend();
}
#endif

/*==================================================================*/
Private Void    ftnssss(name, n, vl)
string  name;
node    n;
vlegt   *vl[];
/*------------------------------------------------------------------*
 * Output ssss vertex.
 *------------------------------------------------------------------*/
{
    string nm;

    if(vtable[ndvertex(n)]->sname != NULL) {
        ftexsubr(vtable[ndvertex(n)]->sname, n, vl, 4);
        return;
    } else if(vtable[ndvertex(n)]->chanel != NULL) {
        nm = vtable[ndvertex(n)]->chanel;
    } else {
        nm = name;
    }

    sprintf(momstr, "      call %s(", nm);
    puinit(tfp1,momstr,
                "     &            ",
                ",", ",", ")", 72, 16);
    if(vcode) {
        putid("nsampl");
    }

    ftncoupl(n);                                 /* cpl, */
    ftcrans(n, vl, 4);                                   /* lt, av */

    putend();
}

/*==================================================================*/
Private Void    ftexarg(vl)
vlegt   *vl;
/*------------------------------------------------------------------*
 */
{
    int tp, sgn;

    /* kf-code */
    sprintf(momstr, "%d", ptable[vl->p]->kfcode);
    putid(momstr);

    /* the number of helicity components */
    if(ptable[vl->p]->ptype == P_SCALAR) {
        putid("1");
    } else {
        tp  = iltable[vl->i].iltype;
        Discard cmindx(momstr1, tp, vl->p);
        putid(momstr1);
    }

    /* mass */
    cmmass(vl->p, 1, momstr3);
    putid(momstr3);

    /* momentum */
    sgn = cmisgn(vl->m);
    cmicmom(momstr3,vl->m, sgn, "pe");
    putid(momstr3);

    /* direction */
    sprintf(momstr, "%c1", csign(vl->d));
    putid(momstr);

    /* momentum**2 */
    cmicmom(momstr3,vl->m, sgn, "vm");
    putid(momstr3);

    switch (ptable[vl->p]->ptype) {
        case P_SCALAR:
            break;

        case P_VECTOR:
            sgn = cmisgn(vl->m);
/* pol.vec.*/
            cmicwgt(momstr1, vl->m, sgn, vl->p, "ep");
            putid(momstr1);
/* weight  */
            cmicwgt(momstr1, vl->m, sgn, vl->p, "eu");
            putid(momstr1);
            break;

        case P_MAJORANA:
        case P_DIRAC:
            sgn = cmisgn(vl->m);
            cmicwgt(momstr1, vl->m, sgn, vl->p, "ew");
            putid(momstr1);

            cmicwgt(momstr1, vl->m, sgn, vl->p, "ce");
            putid(momstr1);

            cmicwgt(momstr2, vl->m, sgn, vl->p, "ps");
            putid(momstr2);

            break;
    }
}

/*==================================================================*/
Private Void    ftexsubr(name, n, vl, size)
char    *name;
node    n;
vlegt   *vl[];
int     size;
/*------------------------------------------------------------------*
 */
{
    int i;


    sprintf(momstr, "      call %s(", name);
    puinit(tfp1,momstr,
                "     &            ",
                ",", ",", ")", 72, 16);

    for(i = 0; i < size; i++) {
        ftexarg(vl[i]);
    }
    ftncoupl(n);
    ftcrans(n, vl, size);

    putend();
}

/*==================================================================*/
Private Void    fthindx(n, vl, size)
node    n;
vlegt   *vl[];
int     size;
/*------------------------------------------------------------------*
 * Output degree of freedom of helicity
 *------------------------------------------------------------------*/
{
    int l, tp;

    for(l = 0; l < size; l++) {
        if(ptable[vl[l]->p]->ptype == P_SCALAR) {
            continue;
        }
        tp  = iltable[vl[l]->i].iltype;
        Discard cmindx(momstr1, tp, vl[l]->p);
        putid(momstr1);
    }
}

/*==================================================================*/
Private Void    ftncoupl(n)
node    n;
/*------------------------------------------------------------------*
 * Output coupling constant
 *------------------------------------------------------------------*/
{
    Discard cmcoupl(ndvertex(n), -9, ndloop(n), momstr);
    if(tishika) {
        if(cgraph->subid == 0) {
            fprintf(tifp, "%4ld: %s\n", cgraph->grid, momstr);
        } else {
            fprintf(tifp, "%4ld.%ld: %s\n",
                    cgraph->grid, cgraph->subid, momstr);
        }
    }
    putid(momstr);
}

/*==================================================================*/
Private Void    ftmassl(n, vl, bgn, end)
node    n;
vlegt   *vl[];
int     bgn, end;
/*------------------------------------------------------------------*
 * Output the list of mass
 *------------------------------------------------------------------*/
{
    int l;

    for(l = bgn; l <= end; l++) {
        cmmass(abs(vl[l]->p), 1, momstr);
        putid(momstr);
    }
}


/*==================================================================*/
Private Void    ftcrans(n, vl, size)
node    n;
vlegt   *vl[];
int     size;
/*------------------------------------------------------------------*
 * Output amplitude table variable.
 *------------------------------------------------------------------*/
{
    long int lsize;
    int l, lstr, ll;
    char *str;
    char vwork[MAXPATHLEN];

    if(nvltbl - nvlusd == 1) {
        putid("lt,av");
        return;
    }

    sprintf(momstr1, "lt%d", n);
    sprintf(momstr2, "0:%d", size);
    putid(momstr1);
    fttmp(FT_INT, momstr1, momstr2, size+1);

    lstr = 0;
    str = sizestr;
    for(l = 0, lsize = 1; l < size; l++) {
        if(lstr + 9 >= 256) {
            fprintf(Stderr, "ftcrans: too long string\n");
            str[1] = '\0';
            fprintf(Stderr, "<%s>\n", sizestr);
            exit(1);
        }
        lsize *= cmindx(str, vl[l]->t, vl[l]->p);
        ll = strlen(str);
        str += ll;
        str[0] = '*';
        str++;
        lstr += ll + 1;
    }
    str--;
    str[0] = '\0';

    sprintf(momstr1, "av%d", n);
    sprintf(momstr2, "%ld", lsize);
    putid(momstr1);
    if(!vcode) {
        fttmp(FT_COMPLEX, momstr1, sizestr, ((int)lsize));
    } else {
        strcpy(vwork,"n_vector,");
        strcat(vwork,sizestr);
        fttmp(FT_COMPLEX, momstr1, vwork, ((int)lsize));
    }
}

/*==================================================================*/
Private Void    ftmlist(n, vl, size)
node    n;
vlegt   *vl[];
int     size;
/*------------------------------------------------------------------*
 * Output list of coupling constants.
 *------------------------------------------------------------------*/
{
    int l, il;

    /* momentum and polarization */
    for(l = 0; l < size; l++) {

        /* particle */
        vl[l]->p = abs(ndparticle(n,l));

        /* internal line */
        vl[l]->i = il = ndiline(n, l);

        /* direction */
        if(iltable[il].ilnodes[0] == n) {
            vl[l]->d = -1;
        } else {
            vl[l]->d = 1;
        }

        /* momentum number */
        if((vl[l]->t = iltable[il].iltype) == ILEXTERN) {
            vl[l]->m = il;
        } else {
            vl[l]->m = il + momoffs;
        }
    }
}

/*==================================================================*/
Private Void    ftiline(NoArg)
/*------------------------------------------------------------------*
 * output an internal line
 *------------------------------------------------------------------*/
{
    int    il, s, lg, tp, sz, sgn;
    node   n;
    pindex p;
    char vwork[MAXPATHLEN];

/*    printf("ftiline:Output an internal line\n");*/
    if(nilines < 1) {
        fprintf(tfp1, "* %d internal momenta\n", nilines);
        return;
    }

    fprintf(tfp1, "\n* Internal momenta\n");
/*    ildump();*/
    for(il = 1; il < nilines; il++) {
/*
        printf("il:%d :iltable.iltype=%d\n",il,iltable[il].iltype);
*/
        if(iltable[il].iltype == ILEXTERN) {
            continue;
        } else if(iltable[il].iltype == ILLOOP) {
            continue;
        }

        n   = iltable[il].ilnodes[0];
        lg  = iltable[il].illegs[0];
        p   = abs(ndparticle(n, lg));
        s   = ptable[p]->ptype;
        tp  = iltable[il].iltype;

        if(s == P_VECTOR) {        /* vector */

            if(jtgamm && ptable[p]->massless) {
/*
                fprintf(tfp1, "      if(jtgamm.eq.0) then\n");
*/
                if(!vcode) {
                    sgn = cmisgn(il);
                    cmicmom(momstr2, il, sgn, "vm");
                    fprintf(tfp1, "      if(jtgamm .eq. 0 .or. ");
                    fprintf(tfp1, "%s .gt. 0.0d0 ) then\n",momstr2);

                    puinit(tfp1,"          call smintv(",
                                "     &                 ",
                                ",", ",", ")", 72, 16);
                } else {

                    sz = cmindx(momstr1, tp, p);
                    putid(momstr1);

                    cmmass(p, 1, momstr3);
                    putid(momstr3);
                    fprintf(tfp1, "      if(jtgamm .eq. 0 ) then\n");
                    puinit(tfp1,"          call smintv(nsampl,",
                                "     &                 ",
                                ",", ",", ")", 72, 16);
                }
                sgn = cmisgn(il);
                cmicmom(momstr, il, sgn, "pe");
                putid(momstr);

                cmicwgt(momstr5, il, sgn, p, "ep");
                putid(momstr5);

                cmicwgt(momstr2, il, sgn, p, "eu");
                putid(momstr2);

                cmicmom(momstr4, il, sgn, "vm");
                putid(momstr4);

                ftgauge(momstr6, p);
                putid(momstr6);
                putend();

                fprintf(tfp1, "      else\n");
                if(!vcode) {
                    sz = cmindx(momstr1, tp, p);
                    cmmass(p, 1, momstr3);
                    puinit(tfp1,"          call smintp(",
                                "     &                 ",
                                ",", ",", ")", 72, 16);
                } else {
                    puinit(tfp1,"          call smintp(nsampl,",
                                "     &                 ",
                                ",", ",", ")", 72, 16);
                }
                sz = cmindx(momstr1, tp, p);
                putid(momstr1);

                cmmass(p, 1, momstr3);
                putid(momstr3);

                sgn = cmisgn(il);
                cmicmom(momstr, il, sgn, "pe");
                putid(momstr);

                cmicwgt(momstr2, il, sgn, p, "eu");
                putid(momstr2);

                cmicmom(momstr4, il, sgn, "vm");
                putid(momstr4);

                ftgauge(momstr5, p);
                putid(momstr5);

                putend();

                fprintf(tfp1, "      endif\n");
            } else {
                if(!vcode) {
                    puinit(tfp1,"      call smintv(",
                                "     &            ",
                                ",", ",", ")", 72, 16);
                } else {
                    puinit(tfp1,"      call smintv(nsampl,",
                                "     &            ",
                                ",", ",", ")", 72, 16);
                }
                sz = cmindx(momstr1, tp, p);
                putid(momstr1);

                cmmass(p, 1, momstr3);
                putid(momstr3);

                sgn = cmisgn(il);
                cmicmom(momstr, il, sgn, "pe");
                putid(momstr);

                cmicwgt(momstr5, il, sgn, p, "ep");
                putid(momstr5);

                cmicwgt(momstr2, il, sgn, p, "eu");
                putid(momstr2);

                cmicmom(momstr4, il, sgn, "vm");
                putid(momstr4);

                ftgauge(momstr5, p);
                putid(momstr5);

                putend();
            }

            /* for declaration */
            sz = cmindx(momstr1, tp, p);
            sgn = cmisgn(il);
            cmicwgt(momstr, il, sgn, p, "eu");

            if(!vcode) {
                fttmp(FT_REAL, momstr, momstr1, sz);
            } else {
              if(rcode) {
                strcpy(vwork,momstr1);
                strcat(vwork,",n_vector");
              } else {
                strcpy(vwork,"n_vector,");
                strcat(vwork,momstr1);
              }
              fttmp(FT_REAL, momstr, vwork, sz);
            }

            sgn = cmisgn(il);
            cmicwgt(momstr,  il, sgn, p, "ep");
            cmicwgt(momstr2, il, sgn, p, "eu");
            if(!vcode) {
                sprintf(momstr2, "4,%s", momstr1);
            } else {
              if(rcode) {
                sprintf(momstr2, "4,%s,n_vector", momstr1);
               } else {
                sprintf(momstr2, "n_vector,4,%s", momstr1);
               }
            }
            fttmp(FT_REAL, momstr, momstr2, 4*sz);

        } else if(!pvboson(p)) {   /* fermion */
            sz = cmindx(momstr1, tp, p);

            if(!vcode) {
                puinit(tfp1,"      call smintf(",
                            "     &            ",
                            ",", ",", ")", 72, 16);
            } else {
                puinit(tfp1,"      call smintf(nsampl,",
                            "     &            ",
                            ",", ",", ")", 72, 16);
            }
            cmmass(p, 1, momstr3);
            putid(momstr3);

            sgn = cmisgn(il);
            cmicmom(momstr, il, sgn, "pe");
            putid(momstr);

            cmicmom(momstr2, il, sgn, "vm");
            putid(momstr2);

            cmicwgt(momstr3, il, sgn, p, "ew");
            putid(momstr3);

            cmicwgt(momstr4, il, sgn, p, "ps");
            putid(momstr4);

            cmicwgt(momstr5, il, sgn, p, "ce");
            putid(momstr5);

            putend();

            cmicwgt(momstr, il, sgn, p, "ew");
            if(!vcode) {
                fttmp(FT_REAL, momstr, "2", 2);
            } else {
              if(rcode) {
                fttmp(FT_REAL, momstr, "2,n_vector", 2);
              } else {
                fttmp(FT_REAL, momstr, "n_vector,2", 2);
              }
            }

            cmicwgt(momstr, il, sgn, p, "ps");
            if(!vcode) {
                fttmp(FT_REAL, momstr, "4,3", 12);
            } else {
              if(rcode) {
                fttmp(FT_REAL, momstr, "4,3,n_vector", 12);
              } else {
                fttmp(FT_REAL, momstr, "n_vector,4,3", 12);
              }
            }

            cmicwgt(momstr, il, sgn, p, "ce");
            if(!vcode) {
                fttmp(FT_COMPLEX, momstr, "2,4", 8);
            } else {
              if(rcode) {
                fttmp(FT_COMPLEX, momstr, "2,4,n_vector", 8);
              } else {
                fttmp(FT_COMPLEX, momstr, "n_vector,2,4", 8);
              }
            }

        } else if(s == P_SCALAR) {
            fprintf(tfp1, "*     Scalar internal line i%4.4d (%d)\n",
                il, ptable[p]->pname[0]);

        } else {
            fprintf(tfp1, "*     ?????? internal line i%4.4d (%d)\n",
                il, ptable[p]->pname[0]);
        }
    }
}

/*=================================================================*/
Private int     ftgauge(str, prt)
string  str;
pindex  prt;
/*------------------------------------------------------------------*
 * Construct variable name 'igauXX'
 *------------------------------------------------------------------*/
{
    int i;
    string sg;


    sg = ptable[prt]->gauge;
    if(sg == NULL) {
        sprintf(str, "igau??");
        return -1;
    }

    sprintf(str, "igau%s", sg);
    for(i = 0; i < ngauge; i++) {
        if(strcmp(sg, sgauge[i]) == 0) {
            return i;
        }
    }
    fprintf(Stderr, "*** ftgauge : illegal sgauge\n");
    exit(FAIL);
    return -1;
}

#if (VLREORD==2)
/*=================================================================*/
Private Uint ftnvtype(str, n, vl)
string  str;
node    n;
vlegt   **vl;
/*------------------------------------------------------------------*
 * construct vertex name to "str". (predicate)
 * return value
 *     -1 : particle with illegal spin
 *      0 : without fermion nor ghost
 *      1 : with fermion
 *      2 : with ghost
 *------------------------------------------------------------------*/
{
    int  lg, t;
    Uint rt = 0L;

    for(lg = 0; lg < ndlegs(n); lg++) {
        t = ptable[abs(ndparticle(n, lg))]->ptype;
        if(t == P_MAJORANA) {
            t = P_DIRAC;
        }
        rt |= pmask[t];
        *str++ = ptypesym[t];
    }

    *str = '\0';

    return rt;
}
#else
/*=================================================================*/
Private Uint ftnvtype(str, n, vl)
string  str;
node    n;
vlegt   **vl;
/*------------------------------------------------------------------*
 * construct vertex name to "str". (predicate)
 * return value
 *     -1 : particle with illegal spin
 *      0 : without fermion nor ghost
 *      1 : with fermion
 *      2 : with ghost
 *------------------------------------------------------------------*/
{
    int  lg, t;
    Uint rt = 0L;
#if (VLREORD==1)
    int  vltyp[MAXLEGS];
    int   i;
    vlegt *vlt;
#else
    int   reptype[PTYPEMAX];
#endif

#if (VLREORD==1)
#else
    for(lg = 0; lg < PTYPEMAX; lg++) {
        reptype[lg] = 0;
    }
#endif
    for(lg = 0; lg < ndlegs(n); lg++) {
        t = ptable[abs(ndparticle(n, lg))]->ptype;
        if(t == P_MAJORANA) {
            t = P_DIRAC;
        }
#if (VLREORD==1)
        for(i = lg; i > 0; i--) {
            if(vltyp[i-1] > t) {
                vltyp[i] = vltyp[i-1];
                vlt      = vl[i];
                vl[i]    = vl[i-1];
                vl[i-1]  = vlt;
            } else {
                break;
            }
        }
        vltyp[i] = t;
#else
        rt |= pmask[t];
        reptype[t]++;
#endif
    }
#if (VLREORD==1)
    for(lg = 0; lg < ndlegs(n); lg++) {
        *str++ = ptypesym[vltyp[lg]];
    }
#else
    for(t = 0; t < PTYPEMAX; t++) {
        for(i = 0; i < reptype[t]; i++) {
            *str++ = ptypesym[t];
        }
    }
#endif

    *str = '\0';

    return rt;
}
#endif

/*==================================================================*/
Private Void    ftscfact(NoArg)
/*------------------------------------------------------------------*
 * Calculate color factor
 */
{
    long      num, den;
    int       mprc, g0, g1;
    int       p0, p1, g1s, pc, l1;
    long      gid0, gid1;
    tgraph   *graph;
    tproc    *proc, *proc0, *proc1;
    string    prf = cprefix;

    /* repeat by loop */
    for(pc = 0; pc < nprocs; pc++) {

        /* first process */
        p0 = reproct[pc];       /* process number */
        proc0 = dagetp(p0);   /* process data */

        /* the second process */
        l1 = pc;                /* loop */
        p1 = reproct[l1];       /* process number */
        proc1 = dagetp(p1);   /* process data */

        printf("Color factors : %d x %d\n", pc, l1);

        /* multiply process data */
        mprc = damproc(p0, p1);
        proc = dagetp(mprc);

        sprintf(momstr,"c%sf%dx%d.dat", prf, p0, p1);
        cmopend(ROFORT, prf, momstr);

        /* repeat graph with */
        for(g0 = 0; g0 < reprocg[pc]; g0++) {

            /* if two process are same, g0 >= g1 */
            if(p0 == p1) {
                g1s = g0;
            } else {
                g1s = 0;
            }

            /* repeat graph with loop l1 */
/*
            for(g1 = g1s; g1 < reprocg[l1]; g1++) {
 */
            for(g1 = 0; g1 < reprocg[l1]; g1++) {

                /* multiply graphs */
                graph = damult(mprc, g0, g1, &gid0, &gid1);

                /* initialize i-line table */
                ilmom();
                nilines = cgraph->nilines;

                cfactor(&num, &den);
                fprintf(redfp, " %d %d %ld %ld\n",
                       g0+1, g1+1, num, den);
#if DBG
                if(dbg > 0) {
                    ildump();
                }
#endif
            }
        }
    }
/*
    fclose(redfp);
*/
    cmclose();
}


/*==================================================================*/
Private Void    ftacfact(NoArg)
/*------------------------------------------------------------------*
 * Calculate color factor
 */
{
    long      num, den;
    int       g0;
    int       p0, pc;
    node      nd;
    tproc    *proc;
    int       con[MAXNODES], col[MAXNODES];
    string    prf = cprefix;

    /* repeat by loop */
    for(pc = 0; pc < nprocs; pc++) {

        /* first process */
        p0 = reproct[pc];       /* process number */
        proc = dagetp(p0);   /* process data */

        printf("Color factors : %d\n", pc);

        sprintf(momstr,"c%sf%d.dat", prf, p0);
        cmopend(ROFORT, prf, momstr);

        /* repeat graph */
        for(g0 = 0; g0 < reprocg[pc]; g0++) {

            cgraph = daread(p0, g0);

            /* initialize i-line table */
            ilmom();
            nilines = cgraph->nilines;
            flget(cgraph);

            cfamp(&num, &den, con, col);

            fprintf(redfp, " %d %ld %ld ", g0+1, num, den);
            for(nd = 0; nd < cprocess->nextern; nd++) {
                fprintf(redfp, " %d %d", con[nd]+1, col[nd]);
            }
            fprintf(redfp, "\n");

#if DBG
            if(dbg > 0) {
                ildump();
            }
#endif
        }
    }

    sprintf(momstr, "cf%scalc.f", prf);
    cmopend(ROFORT, prf, momstr);
    fprintf(redfp,"      subroutine cfcalc\n");
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      include 'incl%s1.h'\n", prf);
    fprintf(redfp,"\n");
    fprintf(redfp,"      integer jcon(n%sextn,n%sgraph), jcol(n%sextn,",
            prf, prf, prf);
    fprintf(redfp,"n%sgraph+1), nn(n%sgraph)\n", prf, prf);
    cmsepar('-');
    fprintf(redfp,"\n");
    fprintf(redfp,"* read color factor\n");
    fprintf(redfp,"c     write(*,*) 'n%sextn = ', n%sextn\n", prf, prf);
    fprintf(redfp,"      open(22, file='c%sf0.dat', status='old')\n", prf);
    fprintf(redfp,"\n");
    fprintf(redfp,"      do 140 i = 1, 5\n");
    fprintf(redfp,"        read(22,*)\n");
    fprintf(redfp,"  140 continue\n");
    fprintf(redfp,"      do 100 i = 1, n%sgraph\n", prf);
    fprintf(redfp,"        read(22,*) ig, nn(i), nd, (jcon(k,i), ");
    fprintf(redfp,"jcol(k,i), k = 1, n%sextn)\n", prf);
    fprintf(redfp,"c       write(6,600) 'read:', ig, nn(i), nd,\n");
    fprintf(redfp,"c    &               (jcon(k,i), jcol(k,i), k = ");
    fprintf(redfp,"1, n%sextn)\n", prf);
    fprintf(redfp,"        if(ig.ne.i) then\n");
    fprintf(redfp,"          write(*,*) 'illegal file of color ");
    fprintf(redfp,"factors (c%sf0.dat)' \n", prf);
    fprintf(redfp,"          write(6,600) ' ', ig, nn(i), nd,\n");
    fprintf(redfp,"     &                 (jcon(k,i), jcol(k,i), k =");
    fprintf(redfp," 1, n%sextn)\n", prf);
    fprintf(redfp,"          stop\n");
    fprintf(redfp,"        else if(nd.ne.1) then\n");
    fprintf(redfp,"          write(*,*) 'illegal denominator (cf0.");
    fprintf(redfp,"dat)' \n");
    fprintf(redfp,"          write(6,600) ' ', ig, nn(i), nd,\n");
    fprintf(redfp,"     &                 (jcon(k,i), jcol(k,i), k =");
    fprintf(redfp," 1, n%sextn)\n", prf);
    fprintf(redfp,"          stop\n");
    fprintf(redfp,"        endif\n");
    fprintf(redfp,"  100 continue\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      close(22)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"*   calculate cf(i,jj)\n");
    fprintf(redfp,"      do 200 i = 1, n%sgraph\n", prf);
    fprintf(redfp,"\n");
    fprintf(redfp,"        do 210 jj = i, n%sgraph\n", prf);
    fprintf(redfp,"          if(i .eq. jj) then\n");
    fprintf(redfp,"            j = n%sgraph + 1\n", prf);
    fprintf(redfp,"            do 260 k = 1, n%sextn\n", prf);
    fprintf(redfp,"                jcol(k, j) = jcol(k, jj)\n");
    fprintf(redfp,"  260       continue\n");
    fprintf(redfp,"          else\n");
    fprintf(redfp,"            j = jj\n");
    fprintf(redfp,"          endif\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"c         write(*,600) 'mult1(con)',i, (jcon(k, ");
    fprintf(redfp,"i), k=1,n%sextn)\n", prf);
    fprintf(redfp,"c         write(*,600) 'mult1(col)',i, (jcol(k, ");
    fprintf(redfp,"i), k=1,n%sextn)\n", prf);
    fprintf(redfp,"c         write(*,600) 'mult2(con)',jj, (jcon(k,");
    fprintf(redfp,"jj), k=1,n%sextn)\n", prf);
    fprintf(redfp,"c         write(*,600) 'mult2(col)',jj, (jcol(k, ");
    fprintf(redfp,"j), k=1,n%sextn)\n", prf);
    fprintf(redfp,"\n");
    fprintf(redfp,"          icf = nn(i)*nn(jj)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"          do 220 n = 1, n%sextn\n", prf);
    fprintf(redfp,"c           write(*,*) 'node = ', n, ', jcol =',");
    fprintf(redfp,"jcol(n, i), \n");
    fprintf(redfp,"c    &                 ' icf=', icf\n");
    fprintf(redfp,"            if(jcol(n, i) .le. 1) goto 220\n");
    fprintf(redfp,"            inx = n\n");
    fprintf(redfp,"            icf = icf * jcol(n, i)\n");
    fprintf(redfp,"            do 240 k = 1, n%sextn\n", prf);
    fprintf(redfp,"*             * check (i, inx)\n");
    fprintf(redfp,"c             write(*,*) 'mult(1-0) ',i, jj, k, ");
    fprintf(redfp,"inx\n");
    fprintf(redfp,"              if(inx.lt.1 .or. jcol(inx, i) .lt. ");
    fprintf(redfp,"1) goto 900\n");
    fprintf(redfp,"              jcol(inx, i) = - jcol(inx, i)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"*             * move in i\n");
    fprintf(redfp,"              inx = jcon(inx, i)\n");
    fprintf(redfp,"c             write(*,*) 'mult(1-1) ',i, jj, k, ");
    fprintf(redfp,"inx\n");
    fprintf(redfp,"              if(inx.lt.1 .or. jcol(inx, i) .lt. ");
    fprintf(redfp,"1) goto 900\n");
    fprintf(redfp,"              jcol(inx, i) = - jcol(inx, i)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"*             * move to j\n");
    fprintf(redfp,"c             write(*,*) 'mult(2-0) ',i, jj, k, ");
    fprintf(redfp,"inx\n");
    fprintf(redfp,"              if(jcol(inx, j) .lt. 1) goto 900\n");
    fprintf(redfp,"              jcol(inx, j) = - jcol(inx, j)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"*             * move in j\n");
    fprintf(redfp,"              inx = jcon(inx, jj)\n");
    fprintf(redfp,"c             write(*,*) 'mult(2-1) ',i, jj, k, ");
    fprintf(redfp,"inx\n");
    fprintf(redfp,"              if(inx.lt.1 .or. jcol(inx, j) .lt. ");
    fprintf(redfp,"1) goto 900\n");
    fprintf(redfp,"              jcol(inx, j) = - jcol(inx, j)\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"              if(inx .eq. n) goto 230\n");
    fprintf(redfp,"  240       continue\n");
    fprintf(redfp,"            goto 910\n");
    fprintf(redfp,"  230       continue\n");
    fprintf(redfp,"  220     continue\n");
    fprintf(redfp,"          do 250 k = 1, n%sextn\n", prf);
    fprintf(redfp,"            if(jcol(k, i) .lt. 0) jcol(k, i) = - ");
    fprintf(redfp,"jcol(k, i)\n");
    fprintf(redfp,"            if(jcol(k, j) .lt. 0) jcol(k, j) = - ");
    fprintf(redfp,"jcol(k, j)\n");
    fprintf(redfp,"  250     continue\n");
    fprintf(redfp,"          cf(i, jj) = icf\n");
    fprintf(redfp,"          cf(jj, i) = icf\n");
    fprintf(redfp,"c         write(*,*) 'cf:', i, jj, icf, cf(i, jj)");
    fprintf(redfp,"\n");
    fprintf(redfp,"  210   continue\n");
    fprintf(redfp,"  200 continue\n");
    fprintf(redfp,"  600 format((' ',a,16i4))\n");
    fprintf(redfp,"      return\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"  900 continue\n");
    fprintf(redfp,"      write(*,*) 'cfcalc : inconsistent'\n");
    fprintf(redfp,"      stop\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"  910 continue\n");
    fprintf(redfp,"      write(*,*) 'cfcalc : not reach to the ");
    fprintf(redfp,"start'\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}

#ifdef OLDSEL
/*==================================================================*/
Private int    ftishiggs(NoArg)
/*------------------------------------------------------------------*
 *  Higgs appears or not in internal lines. (return number of Higgs)
 *  called in ftmain                                 (1995/08/06:TI)
 *  (Warning: this treatment depends on model definition file.)
 *------------------------------------------------------------------*/
{
    int il, higgs;
    pindex p;

    higgs = 0;

    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype == ILEXTERN) {
            continue;
        } else if(iltable[il].iltype == ILLOOP) {
            continue;
        }

        p = ilpart(il);
        if(strcmp(ptable[p]->mass,"amh") == 0) {
            higgs++;
        }
    }
    return higgs;
}

/*==================================================================*/
Private int    ftisgluon(NoArg)
/*------------------------------------------------------------------*
 *  Gluon appears or not in internal lines. (return number of Gluons)
 *  called in ftmain                                 (1995/11/18:TI)
 *  (Warning: this treatment depends on model definition file.)
 *------------------------------------------------------------------*/
{
    int il, gluon;
    pindex p;

    gluon = 0;

    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype == ILEXTERN) {
            continue;
        } else if(iltable[il].iltype == ILLOOP) {
            continue;
        }

        p = ilpart(il);
        if(strcmp(ptable[p]->mass,"amg") == 0) {
            gluon++;
        }
    }
    return gluon;
}

#else
/*==================================================================*/
Private Ulong  ftselect(NoArg)
/*------------------------------------------------------------------*
 *  Set graph selection flag
 */
{
    int il, n, i;
    Ulong mask = 0L, msk;
    string ss;

    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype == ILEXTERN) {
            continue;
        } else if(iltable[il].iltype == ILLOOP) {
            continue;
        }
        ss = ptable[abs(ilpart(il))]->pselect;
        if(ss == NULL) {
            continue;
        }
        msk = 1L;
        for(i = 0; i < nselect; i++) {
            if(strcmp(ss, sselect[i]) == 0) {
                mask |= msk;
                break;
            }
            msk <<= 1;
        }
        if(i >= nselect) {
            fprintf(Stderr, "*** ftselect : no selection name\n");
            exit(FAIL);
        }
    }
    for(n = 0; n < cgraph->nnodes; n++) {
        if(ndtype(n) != NDTREEV) {
            continue;
        }
        ss = vtable[ndvertex(n)]->vselect;
        if(ss == NULL) {
            continue;
        }
        msk = 1L;
        for(i = 0; i < nselect; i++) {
            if(strcmp(ss, sselect[i]) == 0) {
                mask |= msk;
                break;
            }
            msk <<= 1;
        }
        if(i >= nselect) {
            fprintf(Stderr, "*** ftselect : no selection name\n");
            exit(FAIL);
        }
    }

    return mask;
}
#endif

#define fltnam(typ)  (((typ)==FLOPEN)?"open  ": \
                      ((typ)==FLCLOSED)?"closed":"undef ")
/*==================================================================*/
Private bool   ftisschan(NoArg)
/*------------------------------------------------------------------*
 *  S-chanel or not.
 *  called in ftmain                                 (1995/11/14:TI)
 *------------------------------------------------------------------*/
{
    int f, n, scheck;
    bool isschn;

    isschn = False;
/*    printf("\n%% +++ %d fermion lines.\n", nflines);*/
    for(f = 0; f < nflines; f++) {
/*
        printf("%% type=%s, length=%4d : ",
               fltnam(fltable[f].type), fltable[f].length);
*/
        scheck = 0;
        for(n = 0; n < fltable[f].length; n++) {
/*
            printf("[%d,n:%d,%d] ",
                   fltable[f].lglist[FL_IN][n],
                   tovert(fltable[f].ndlist[n]),
                   fltable[f].lglist[FL_OUT][n]);
*/
            if(fltable[f].lglist[FL_OUT][n] < 0) {
                if(ndiorf(tovert(fltable[f].ndlist[n]))==ELINITIAL){
/*                    printf(" Initial ");*/
                    scheck++;
                }
            }
        }
/*        putchar(EOL);*/
        if(scheck == 2) {
            isschn = True;
            return isschn;
        }
    }
    return isschn;
}

/*==================================================================*/
Private Void    ftprfl(fp)
FILE    *fp;
/*--------------------------------------------------------------------
 *   Dump the table of f-lines.
 */
{
    int f, n;

    fprintf(fp, "\n* +++ %d fermion lines. (graph %ld)\n",
            nflines, cgraph->grid);
    for(f = 0; f < nflines; f++) {
        fprintf(fp, "* type=%s, length=%4d, class=%4d : ",
               fltnam(fltable[f].type), fltable[f].length,
               fltable[f].fclass);
        for(n = 0; n < fltable[f].length; n++) {
            fprintf(fp, "%d ", tovert(fltable[f].ndlist[n]));
        }
        putc(EOL, fp);
    }
}

/*==================================================================*/
Private string  ftmcc(pref, n, f0, vcsgn)
string  pref;
node    n;
int     f0;
int     *vcsgn;
/*------------------------------------------------------------------*
 * Construct subroutine name of ffv or ffs vertex in which
 * Majorana particles are included.
 *------------------------------------------------------------------*/
{
    static char vname[32];
    static char *uvchar = "Uuv";
    int    len, i, f, t, v, d, il, nn;
    int    uv[2];
    pindex pp, p0, p[2], vp[2];

#ifdef SUSYDBG
    printf("*** ftmcc(\"%s\",%d,%d)\n", pref, n, f0);
    printf("Process: ");
    for(nn = 0; nn < nextern; nn++) {
        if(ndiorf(nn) == ELINITIAL) {
            p0  = ndname(nn);
            printf("%s(%d:%d) ",pvename(ndiorf(nn),p0), nn, p0);
        }
    }
    printf(" --> ");
    for(nn = 0; nn < nextern; nn++) {
        if(ndiorf(nn) == ELFINAL) {
            p0  = ndname(nn);
            printf("%s(%d:%d) ",pvename(ndiorf(nn),p0), nn, p0);
        }
    }
    printf("\n");
#endif
    len = strlen(pref);
    if(len > 32 - 4) {
        fprintf(Stderr, "*** ftmcc : too long prefix \"%s\"\n", pref);
        fprintf(Stderr, "***         pref=\"%s\", n=%d, f0=%d\n",
                        pref, n, f0);
        exit(FAIL);
    }
    for(i = 0; i < len; i++) {
        vname[i] = pref[i];
    }
    vname[len + 3] = '\0';

    *vcsgn = 1;

    /* two-fermions */
    for(i = 0; i < 2; i++) {
        f = f0 + i;

        p0 = pp = ndparticle(n, f);

        il = ndiline(n, f);
        t  = ptable[abs(p0)]->ptype;
        if(iltable[il].iltype == ILEXTERN) {
            if(t == P_MAJORANA) {
                /* direction of momentum */
                if(iltable[il].ilnodes[0] == n) {
                    d = -1;       /* outgoing */
                } else {
                    d =  1;       /* incoming */
                }
                d *= 2*i-1;      /* normal direction of momentum */
                if(d > 0) { /* normal direction of momentum */
                    vname[len + 2*i] = 'n';
                } else {
                    vname[len + 2*i] = 'c';
                    *vcsgn *= -1;    /* u^c, \bar{u}^c */
                }
                /* Majorana is always particle */
                uv[i] = 1;
#ifdef SUSYDBG
                printf("*** ftmcc : external majorana ");
#endif
            } else if(t == P_DIRAC) {
                nn = ndnextn(n, f);
                if(ndiorf(nn) == ELFINAL) {
                    pp = - p0;
                } else {
                    pp =   p0;
                }

                /*                  :i: wave-f.  :p0:pp: d:
                 * initial particle :0: u^c      : +: +: -:
                 * initial particle :1: u        : +: +: +:
                 * initial anti-p.  :0: \bar{v}  : -: -: +:
                 * initial anti-p.  :1: \bar{v}^c: -: -: -:

                 * final   particle :0: \bar{u}  : -: +: +:
                 * final   particle :1: \bar{u}^c: -: +: -:
                 * final   anti-p.  :0: v^c      : +: -: -:
                 * final   anti-p.  :1: v        : +: -: +:
                 */
#ifdef SUSYDBG
                printf("*** ftmcc : external dirac    ");
#endif
                if(pp < 0) {       /* anti-particle : v, \bar{v} */
                    *vcsgn *= -1;
                    uv[i] = 2;
                } else {             /* particle : u, \bar{u} */
                    uv[i] = 1;
                }
                d = p0*(2*i-1);

                if(d > 0) {  /* normal direction of fermion # */
                    vname[len + 2*i] = 'n';
                } else {
                    vname[len + 2*i] = 'c';
                    *vcsgn *= -1;
                }
            } else {
                fprintf(Stderr, "*** ftmcc : Illegal particle type : ");
                fprintf(Stderr, "t=%d,f=%d\n", t,f);
                fprintf(Stderr, "***         pref=\"%s\", n=%d, f0=%d\n",
                                 pref, n, f0);
                prdump();
                exit(FAIL);
            }
        } else {        /* propagator */
            uv[i] = 0;
            vname[len + 2*i] = 'n';
        }
        if(t == P_MAJORANA) {
            p[i] = abs(p0);
        } else {
            p[i] = p0;
        }
#ifdef SUSYDBG
        printf("n=%d, nn=%d, p0=%d, pp=%d, iorf=%d, d=%d, s=%d\n",
                n, nn, p0, pp, ndiorf(nn), d, *vcsgn);
#endif
    }


    /* direction of vertex */
    v = ndvertex(n);
    vp[0] = vtable[v]->vlegs[f0];
    vp[1] = vtable[v]->vlegs[f0+1];

    if(vp[0] == p[0] && vp[1] == p[1]) {
        vname[len + 1] = 'n';
    } else if(vp[0] == p[1] && vp[1] == p[0]) {
        vname[len + 1] = 'c';
    } else {
        vname[len + 1] = '?';

        fprintf(Stderr, "*** ftmcc : Illegal legs : ");
        fprintf(Stderr, "t=%d,f=%d\n", t,f);
        fprintf(Stderr, "***         pref=\"%s\", n=%d, f0=%d\n",
                         pref, n, f0);
        fprintf(Stderr, "     [0],  [1]\n");
        fprintf(Stderr, " p: %4d, %4d\n",  p[0],  p[1]);
        fprintf(Stderr, "vp: %4d, %4d\n", vp[0], vp[1]);
        prdump();
        exit(FAIL);
    }

    if(majorana) {
        fprintf(tfp1, "*             ");
        fprintf(tfp1, "\\bar{%c(%d)}", uvchar[uv[0]], ndiline(n, f0));
        if(vname[len] == 'c') {
            fprintf(tfp1, "^c");
        } else {
            fprintf(tfp1, "  ");
        }
        fprintf(tfp1, " \\Gamma");
        if(vname[len+1] == 'c') {
            fprintf(tfp1, "^c");
        } else {
            fprintf(tfp1, "  ");
        }
        fprintf(tfp1, " %c(%d)", uvchar[uv[1]], ndiline(n, f0+1));
        if(vname[len+2] == 'c') {
            fprintf(tfp1, "^c");
        } else {
            fprintf(tfp1, "  ");
        }
        fprintf(tfp1, "\n");
    }
#ifdef SUSYDBG
    printf("*             ");
    printf("\\bar{%c(%d)}", uvchar[uv[0]], ndiline(n, f0));
    if(vname[len] == 'c') {
        printf("^c");
    } else {
        fprintf(tfp1, "  ");
    }
    printf(" \\Gamma");
    if(vname[len+1] == 'c') {
        printf("^c");
    } else {
        fprintf(tfp1, "  ");
    }
    printf(" %c(%d)", uvchar[uv[1]], ndiline(n, f0+1));
    if(vname[len+2] == 'c') {
        printf("^c");
    } else {
        fprintf(tfp1, "  ");
    }
    printf("\n");
#endif

    return vname;
}
