/*********************************************************************
 *  File : cfbase.c
 *                                95.11.17 (K.K.)
 *                                96.10.25 (K.K.)
 *  Calculate color factor
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "cfbase.h"
#include "cfconv.h"

/*--------------------------------------------------------------------
 * parameters
 */
/*
#define DBGFREE
*/
#define MODTK
/*#undef MODTK*/

#undef  CFKDBG          /* for Kato's debug */
#undef  CFKDBG1         /* for Kato's debug */
#define CFKMAXSTR  20   /* maximum number of strings */
#define CFKMAXGLUE  2   /* maximum number of gluons */

/* TK : CFSHIFT is moved to cfbase.h */
/* #define CFSHIFT   100 */ /* shift value for gluon reduction */
                            /* gluon=x --> quark=(x+CFSHIFT)         */
                            /*         and antiquark=(x+2*CFSHIFT)   */

#define CFVGGGP    2   /* temporal def. */
#define CFVGGGM    2   /* temporal def. */
#define CFVEXTQ   10   /* temporal def. */
#define CFVEXTAQ  11   /* temporal def. */
#define CFVEXTG   12   /* temporal def. */
#define CFVEXTGH  13   /* temporal def. */

#undef  CFGLUELMT

/*--------------------------------------------------------------------
 * Private functions
 */
Private int  cfbinit        ARG((cftbase*, int, int, int));
Private int  cfcinit        ARG((cftbase*, int));

Private int  cfproji        ARG((cftbase*,cftgraph*,cftxgraph*));
Private Void cfprojx        ARG((cftbase*,cftgraph*,cftxgraph*));
Private Void cfprojw        ARG((cftbase*,cftgraph*,cftxgraph*));
Private Void cfxgalloc      ARG((cftxgraph*,int,int));
Private Void cfxgfree       ARG((cftxgraph*));
Private Void cfxgcopy       ARG((cftxgraph*,cftxgraph*));
Private int cfkperm         ARG((int,int*));
#ifdef MODTK
Private cftxgraph *cfxhalloc      ARG((int,int));
Private Void       cfxhfree       ARG((NoArg));
#endif

/* local variables */
#ifndef MODTK
Local int nnid[CFPKIND+1];
#endif
#define   MAXCFG 10000
Local int addrcfg = 0;
Local cftxgraph xcfg[MAXCFG];

/*==================================================================*/
Public  int cfbase(cbase, ibase, iprnt)
cftbase *cbase;
int     ibase;
int     iprnt;
/*--------------------------------------------------------------------
 * cbase : pointer to color base data structure
 * ibase : input flag for the selection of bases
 * iprnt : input flag for printing information
 *
 *   template by T.Kaneko , July 1995
 *   coded by K.Kato , August 1995
 *   last update on 95/08/09
 *
 * returns status
 */
{
    int i, j, k;
    int k1;
    int ix, ixx ,  ii;
    cftfrac w ;
    int ne,nb;            /* ne=number of external particles */
                          /* nb=number of color bases        */
    int nq,nqa,ng;        /* number of q, qbar, g            */
    int nqtot;  /* number of total quark paris after gluon reduction  */
    int workq[CFKMAXSTR+1];
    int works[2][CFKMAXSTR+CFKMAXSTR];

	/* --- pickup colored particles(1) ---*/
	ne = cbase->nex ;
	nq = nqa = ng = 0 ;
	for (i=1 ; i<=ne ; i++ )
		{ j = cbase->cptype[i] ;
		  switch (j)
		   { case CFQUARK  : nq++ ; break ;
		     case CFAQUARK : nqa++ ; break ;
		     case CFGLUON  : ng++ ; break ;
		   }
		}

	/* --- error check ---*/
	if (nq==0 && nqa==0 && ng==0)
	  { printf(" (cfbase) No colored external particles.  \n") ;
	  }

	if (nq != nqa)
	  { printf(" (cfbase) *BAD* Broken quark-antiquark pair %d:%d . \n",
			nq,nqa) ;
	    return 0;
	  }

	if (nq==0 && ng==1 )
	  { printf(" (cfbase) *BAD* Quark number is 0 and ng=%d.\n",ng) ;
	    return 0;
	  }

#ifdef CFGLUELMT
	if (ng>CFKMAXGLUE)
	  { printf(" (cfbase) *BAD* %d gluons not supported.\n",ng) ;
	    return 0;
	  }
#endif

	if ((nq+ng)>CFKMAXSTR)
	  { printf(" (cfbase) *BAD* Too many strings. nq,ng=%d,%d\n",nq,ng) ;
	    return 0;
	  }

	/* --- Number of color-bases ---*/
	nqtot = nq + ng ;

        if (nqtot>0)
        {
	nb = 1 ;
	for (i=1 ; i<=nqtot ; i++ ) nb *=i ;
        }
        else
        nb = 0 ;

        /* --- allocate cbase --- */
	cfbinit(cbase, nq, ng, nb);
        if ( cbase->nexcolor == 0 )
        {  cbase->constr[0]=0 ;
           cbase->nconid[0]=0 ;
           cffrset(&cbase->cwgt[0][0],1,1) ;
           return 1 ;
        }

	/* --- pickup colored particles(2) record into cbase  ---*/
	nq = nqa = ng = 0 ;
	for (i=1 ; i<=ne ; i++ )
		{ j = cbase->cptype[i] ;
		  switch (j)
		   { case CFQUARK  : cbase->quark[nq++]=i ; break ;
		     case CFAQUARK : cbase->antiq[nqa++]=i ; break ;
		     case CFGLUON  : cbase->gluon[ng++]=i ; break ;
		   }
		}

	/* --- gluon reduction ---*/
	j = nq ;
	for (i=0 ; i<ng ; i++)
	{
	  cbase->quark[j] = cbase->gluon[i] + CFSHIFT ;
	  cbase->antiq[j] = cbase->gluon[i] + CFSHIFT + CFSHIFT ;
	  j++ ;
	}

#ifdef CFKDBG
	printf( " CFKMAXSTR =%d\n",CFKMAXSTR);
	printf( "q, qbar, g= %d, %d, %d \n",nq,nqa,ng);
	if (nq>0) { for( i=0 ; i<nqtot ; i++ ) printf(" %d",cbase->quark[i]);
	            printf( " ... cbase->quark\n"); }	
	if (nqa>0){ for( i=0 ; i<nqtot ; i++ ) printf(" %d",cbase->antiq[i]);
	            printf( " ... cbase->antiq\n"); }	
	if (ng>0) { for( i=0 ; i<ng ; i++ ) printf(" %d",cbase->gluon[i]);
	            printf( " ... cbase->gluon\n"); }	
	printf(" nbase = %d\n",cbase->nbase) ;
#endif

	/* --- construction of color bases --- */

	for (i=0 ; i<nqtot ; i++)
		workq[i+1]=cbase->quark[i] ;

	nb=-1 ;
	do
	{
		nb++ ;
		cbase->constr[nb]=nqtot ;
		cbase->nconid[nb]=2*nqtot ;
        	cfcinit(cbase, nb);
		j=0 ;
		for (i=0 ; i<nqtot ; i++)
		{	cbase->conid[nb][j++]=cbase->antiq[i] ;
			cbase->conid[nb][j++]=workq[i+1] ;
		}
	}
	while ( cfkperm( nqtot , workq ) != -1 ) ;

#ifdef CFKDBG
	for (j=0 ; j<=nb ; j++)
	{ printf(" nb=%d, constr=%d, nconid=%d\n",
		j,cbase->constr[j],cbase->nconid[j]);
	  for (i=0 ; i<cbase->nconid[j] ; i++)
		printf(" %d",cbase->conid[j][i]);
          printf(" .. conid \n"); }
#endif

	if( (nb+1) != cbase->nbase )
	{ printf(" (cfbase) base confrict %d:%d\n",nb,cbase->nbase) ;
	  return  0 ; }


	/* --- product of color bases --- */

                                           /* LOOP-->(2) */
	for( k1=0 ; k1< cbase->nbase ; k1++)
	for( k =0 ; k <=k1           ; k++ )
	{
           /* copy string into work space */
	  ix= cbase->nconid[k] ;
	  for (j=0 ; j<2 ; j++ )
	  for (i=0 ; i<ix ; i++)
	  {	if (j==0)
		works[j][i] = cbase->conid[k1][i] ;
		else
		works[j][i] = cbase->conid[k ][i] ;
	  }

#ifdef CFKDBG
	printf(" k1,k= %d, %d \n",k1,k) ;
	for (j=1 ; j>=0 ; j--) {
	for (i=0 ; i<ix ; i++) printf(" %d:(%d,%d) ",works[j][i],i,j);
	printf("\n"); }
#endif

	  cffrset( &w , 1 , 1 ) ;

	  i = 0 ; /* pointer on works[][i] */
          j = 0 ; /*            works[j][] */

	  while ( i>=0 )
	  {
	    ixx = works[j][i] ;
#ifdef CFKDBG1
	printf(" <0>ixx,i,j=%d,%d,%d, w=%f\n",ixx,i,j,w) ;
#endif

	      works[j][i] = 0 ;

              if( (i%2) == 0 )     /* even i .. cbase->antiquark */
	      {
	        if ( j==0 ) i++ ;
	        else
	        {
	          j = 0 ;
	          i = -1 ;
	          for (ii=0 ; ii<ix ; ii++)
		     if ( works[j][ii]==ixx ) { i=ii ; break ; }
                  if ( i<0 )
	          {
	            cffrsml( &w , 3) ;
#ifdef CFKDBG1
	printf(" <1>ixx,i,j=%d,%d,%d, w=%f\n",ixx,i,j,w) ;
#endif
	            i = -1 ;
	            for (ii=0 ; ii<ix ; ii++ )
	               if (works[0][ii]>0 ) { i=ii ;  break ; }
	          }
	        }
	      }
	      else                 /* odd i .. cbase->quark */
	      {
	        if ( j==1 ) i-- ;
	        else
	        {
	          j = 1 ;
	          for (ii=0 ; ii<ix ; ii++)
		     if ( works[j][ii] ==ixx ){ i=ii ; break ; }
	        }
	      }

	  }
          cffrcpy( &cbase->cwgt[k1][k] , &w ) ;
#ifdef CFKDBG
	printf(" cwgt(%d,%d)=%d / %d \n",k1,k,w) ;
#endif
	}                                /* END LOOP <--(2) */

	if ( cbase->nbase >1 )
	for( k1=0      ; k1< (cbase->nbase-1) ; k1++)
	for( k =(k1+1) ; k <  cbase->nbase    ; k++ )
	{
          cffrcpy( &cbase->cwgt[k1][k], &cbase->cwgt[k][k1] ) ;
	}

        if (ng>0)
        {
          ix = 1;
          for (i=0 ; i<ng ; i++ ) ix *=2 ;
          printf(" gluon reduction factor %d included in cwgt \n",ix );

	  for( k1=0 ; k1< cbase->nbase ; k1++)
	  for( k =0 ; k < cbase->nbase ; k++ )
            cffrsml( &cbase->cwgt[k1][k] , ix) ;
        }

        /* ---- Average for initial state ---- (NOT YET) */
        ix = 1 ;
        for (i=1 ; i<=ne ; i++ )
	{
	    if ( cbase->initial[i] == CBINITIAL )
		{
		  if( cbase->cptype[i]==CFQUARK ) ix*=3 ;
                  if( cbase->cptype[i]==CFAQUARK) ix*=3 ;
                  if( cbase->cptype[i]==CFGLUON ) ix*=8 ;
                  if( cbase->cptype[i]==CFGHOST ) ix*=8 ;
                  if( cbase->cptype[i]==CFAGHOST) ix*=8 ;
		}
	}
#ifdef CFKDBG
          printf(" .. Initial ave. :divide by  %d \n",ix);
#endif
#ifdef MODTK
        cbase->iniav = ix;    /* TK : added line */
#else
	cffrset( &w , 1, ix ) ;
	for( k1=0 ; k1< cbase->nbase ; k1++)
	for( k =0 ; k < cbase->nbase ; k++ )
	{
          cffrmul( &cbase->cwgt[k1][k], &w ) ;
	}
#endif

#ifdef CFKDBG
	for (j=0 ; j<=nb ; j++)
	{ for (i=0 ; i<=nb ; i++)
		cffrpr0(&cbase->cwgt[j][i]);
          printf(" .. cwgt : %d \n",j); }
#endif
	return 1;
}

/*==================================================================*/
Public  int cfproj(cbase, cfgraph, ibase, iprnt)
cftbase  *cbase;
cftgraph *cfgraph;
int      ibase ;
int      iprnt;
/*--------------------------------------------------------------------
 * cbase   : pointer to color base data structure
 * cfgraph : information about a graph and its weight
 * ibase   : input flag for the selection of bases
 * iprnt   : input flag for printing information
 *
 *   template by T.Kaneko , July 1995
 *   coded by K.Kato , August 1995
 *   last update on 95/08/09
 *
 */
{
    int i;
    int cleg,cvtx;
#ifdef MODTK
    int nqqg, nggg, nhhg, nvqq, nvgg, next ;
#else
    int nqqg, nggg, nhhg, nvqq, next ;
#endif
    cftxgraph *cfg  ;

   /* count colored vertices */

#ifdef MODTK
   nqqg = nggg = nhhg = nvqq = nvgg = 0 ;
#else
   nqqg = nggg = nhhg = nvqq = 0 ;
#endif

#ifdef MODTK
   for (i=1 ; i<=cfgraph->nvtx ; i++)
     if ( cfgraph->nleg[i] >= 2 )
     {
#else
   for (i=1 ; i<=cfgraph->nvtx ; i++)
     if ( cfgraph->nleg[i] == 3 )
     {
#endif
       if ( cfgraph->lvtx[i] == CFVQQG )    nqqg ++ ;
       if ( cfgraph->lvtx[i] == CFVGGGP ||
            cfgraph->lvtx[i] == CFVGGGM   ) nggg ++ ;
       if ( cfgraph->lvtx[i] == CFVHHG )    nhhg ++ ;
#ifdef MODTK
       if ( cfgraph->lvtx[i] == CFVQQ )     nvqq ++ ;
       if ( cfgraph->lvtx[i] == CFVGG )     nvgg ++ ;
#endif
     }

#ifdef MODTK
   if ( nqqg+nggg+nhhg+nvqq+nvgg == 0 )
   {
     cffrset( &cfgraph->wgt[0], 1, 1 ) ;
     return 0; }
#else
   if ( nqqg+nggg+nhhg+nvqq == 0 )
   { printf(" (cfproj) no colored vertex \n") ;
     cffrset( &cfgraph->wgt[0], 1, 1 ) ;
     return ; }
#endif

   next = 2*(cbase->nquarkp + cbase->ngluon) ; /* external colored p. */
   nqqg += cbase->ngluon ;                /* gluon reduction included */
#ifdef MODTK
   cvtx = next + nqqg + nggg + nhhg + nvqq + nvgg  ;
   cleg = (next + 3*(nqqg+nggg+nhhg) + 2*(nvqq + nvgg))/2 ;
#else
   cvtx = next + nqqg + nggg + nhhg + nvqq   ;
   cleg = (next + 3*(nqqg+nggg+nhhg) + 2*nvqq)/2 ;
#endif

#ifdef CFKDBG
#ifdef MODTK
printf("cbase->nquarkp,  cbase->ngluon=%d, %d \n",
        cbase->nquarkp,  cbase->ngluon) ;
printf(" next, nqqg,  nggg,  nhhg,  nvqq, nvgg = %d: %d, %d, %d, %d, %d\n",
         next, nqqg,  nggg,  nhhg,  nvqq, nvgg )  ;
#else
printf("cbase->nquarkp,  cbase->ngluon=%d, %d \n",
        cbase->nquarkp,  cbase->ngluon) ;
printf(" next, nqqg,  nggg,  nhhg,  nvqq= %d: %d, %d, %d, %d\n",
         next, nqqg,  nggg,  nhhg,  nvqq )  ;
printf(" ........... cvtx,cleg = %d , %d  \n",cvtx,cleg);
#endif
#endif

   /* --- clear cfgraph->wgt ---*/
   for (i=0 ; i<cbase->nbase ; i++) cffrclr(&cfgraph->wgt[i]) ;

   /* --- allocte cfg ---*/
#ifdef MODTK
   if(addrcfg != 0) {
       cfxhfree();
   }
   cfg = cfxhalloc( cvtx , cleg ) ;
#else
   addrcfg = 0 ;
   cfg = &xcfg[0] ;
   cfxgalloc( cfg , cvtx , cleg ) ;
#endif

   /* --- create cfg, reading cfgraph, referring cbase --- */
   cfproji(cbase,cfgraph,cfg) ;

   /*--- Now, cfg is ready. Do main job. ---*/
   cfprojx( cbase ,cfgraph, cfg ) ;

   /* --- result  ---*/
/*
   printf(" *weight for color base *\n");
   for (i=0 ; i<cbase->nbase; i++)
   { printf(" cfgraph->wgt[%d]=",i);
     cffrprt(" ",&cfgraph->wgt[i]);
   }
*/
   if (iprnt==1)
   {
     for (i=0 ; i<cbase->nbase; i++)
     {   printf("( %d / %d )",(cfgraph->wgt[i]).num,
                             (cfgraph->wgt[i]).den);
         if ( (i==(cbase->nbase-1)) || ((i%5)==4) )
              printf("\n");
     }
   }
#ifdef MODTK
   cfxhfree();
   return 1;
#endif
}
/*==================================================================*/
Private int cfproji(cbase, cfgraph, cfg)
cftbase  *cbase;
cftgraph *cfgraph;
cftxgraph *cfg;
/*--------------------------------------------------------------------
 * cbase  : pointer to color base data structure
 * cfgraph : information about a graph and its weight
 * cfg    : pointer to graph structure (working)
 *
 */
{
    int i, j, je, ii, j1, jj;
    int leg ;

        /* Scan cfgraph, only colored particles and vertices  */
        /* are copied into cfg.                              */
        /* Now, external ends are also counted as vertices.  */
        /* Here, mvtx records negative for internal vertex,  */
        /* and keeps particle id number  for external end.   */

   je=0 ;                                /* record colored edge */
   for (i=1 ; i<=cfgraph->nedge ; i++)
   {
       if ( cfgraph->tedge[i] == CFQUARK  ||
            cfgraph->tedge[i] == CFAQUARK ||
            cfgraph->tedge[i] == CFGLUON  ||
            cfgraph->tedge[i] == CFGHOST    )
       {  je++ ;
          cfg->tedge[je]   = cfgraph->tedge[i] ;
          cfg->edge[0][je] = cfgraph->edge[0][i] ;
          cfg->edge[1][je] = cfgraph->edge[1][i] ;
       }
   }
   /* printf(" Pass cfproji-1\n");*/

   j=0 ;                                /* record vertex */
   for (i=1 ; i<=cfgraph->nvtx ; i++)    /* internal vertex */
#ifdef MODTK
     if ( cfgraph->nleg[i] >= 2 )
#else
     if ( cfgraph->nleg[i] == 3 )
#endif
     {
       if ( cfgraph->lvtx[i] == CFVQQG  ||
            cfgraph->lvtx[i] == CFVGGGP ||
            cfgraph->lvtx[i] == CFVGGGM ||
            cfgraph->lvtx[i] == CFVHHG  ||
#ifdef MODTK
            cfgraph->lvtx[i] == CFVQQ   ||
            cfgraph->lvtx[i] == CFVGG     )
#else
            cfgraph->lvtx[i] == CFVQQ     )
#endif
       {  j++ ;
          cfg->lvtx[j] = cfgraph->lvtx[i] ;
          cfg->mvtx[j] = -j ;
                                        /* revise vertex number */
          for( jj=0 ; jj<=1  ; jj++ )
          for( j1=1 ; j1<=je ; j1++ )
          {  if ( cfg->edge[jj][j1] == i ) cfg->edge[jj][j1] = j ; }
       }
     }
                                     /* external end */
   if (cbase->nquarkp >0 )
   {
     for (i=0 ; i<cbase->nquarkp ; i++)
     {
         j++ ;
         cfg->lvtx[j] = CFVEXTQ ;
         cfg->mvtx[j] = cbase->quark[i] ;
         j++ ;
         cfg->lvtx[j] = CFVEXTAQ ;
         cfg->mvtx[j] = cbase->antiq[i] ;
     }
   }
                                     /* gluon, reduced */
   if (cbase->ngluon >0 )
   {
     for (i=0 ; i<cbase->ngluon ; i++)
     {  j++ ;
        cfg->lvtx[j] = CFVQQG ;
        cfg->mvtx[j] = -(cbase->gluon[i]+CFSHIFT) ;
        j++ ;
        cfg->lvtx[j] = CFVEXTQ ;
        cfg->mvtx[j] = cbase->gluon[i]+CFSHIFT ;
        j++ ;
        cfg->lvtx[j] = CFVEXTAQ ;
        cfg->mvtx[j] = cbase->gluon[i]+CFSHIFT+CFSHIFT ;
        je++ ;
        cfg->tedge[je]   = CFQUARK ;
        cfg->edge[0][je] = j-2 ;
        cfg->edge[1][je] = j-1 ;
        je++ ;
        cfg->tedge[je]   = CFQUARK ;
        cfg->edge[0][je] = j   ;
        cfg->edge[1][je] = j-2 ;
     }
   }
   /* printf(" Pass cfproji-2\n");*/

   if( j != cfg->nvtx )
   { printf(" (cfproji) Err. for cfg->nvtx. %d:%d \n",j,cfg->nvtx) ;
                     cfdump(cbase, cfgraph); return 0 ; }
   if( je!= cfg->nedge )
   { printf(" (cfproji) Err. for cfg->nedge. %d:%d \n",je,cfg->nedge) ;
                     cfdump(cbase, cfgraph); return 0 ; }

         /* Make cross reference for external end */
   for (ii=0 ; ii<=1   ; ii++)
   for (i =1 ; i<=cfg->nedge ; i++)
   {
     leg = cfg->edge[ii][i] ;
     if (leg<0)
     {
       if ( cbase->cptype[-leg] == CFQUARK ||
            cbase->cptype[-leg] == CFAQUARK   )
       {
         for (j=1 ; j<=cfg->nvtx ; j++)
           if( cfg->mvtx[j] == (-leg) ) cfg->edge[ii][i]=j ;
       }
       if ( cbase->cptype[-leg] == CFGLUON )
       {
         for (j=1 ; j<=cfg->nvtx ; j++)
           if( cfg->mvtx[j] == (leg-CFSHIFT) ) cfg->edge[ii][i]=j ;
       }
     }
     if ( cfg->edge[ii][i] <0 )
     { printf(" (cfproji) Err. cfg->edge(%d,%d)=%d \n",ii,i,leg) ;
                     cfdump(cbase, cfgraph); return 0 ; }
   }
   /* printf(" Pass cfproji-3\n");*/

#ifdef CFKDBG
for (i=1; i<=cfg->nvtx ; i++)
{ printf (" (%d) %d, %d",i,cfg->lvtx[i],cfg->mvtx[i]);
if(i%4==0) printf (" ...lvtx,mvtx\n") ; }
           printf (" ...lvtx,mvtx\n") ;
for (i=1; i<=cfg->nedge ; i++)
{printf (" (%d) %d, %d -> %d",
i,cfg->tedge[i],cfg->edge[0][i],cfg->edge[1][i]);
if(i%4==0) printf (" ...tedge, edge\n") ;}
printf (" ...tedge, edge\n") ;
#endif

#ifdef MODTK
    return 1;
#endif
}

/*==================================================================*/
Private Void cfprojx(cbase,cfgraph,cfg)
cftbase   *cbase;
cftgraph  *cfgraph;
cftxgraph *cfg;
/*--------------------------------------------------------------------
 * Graph reduction, called iteratively.
 *
 * cbase  : pointer to color base data structure
 * cfgraph : pointer to graph information, return with final weight.
 * cfg    : pointer to graph structure (working)
 */
{
    cftxgraph *cfg1,*cfg2;
    int i, ix, j, jj, jx, k, ja;
    int kv[3],ke[3],kjj[3],kv0[3] ;
    int nv1,nv2;
    int ng1,ng2,nge1=0,nge2=0;
    int nq1,nq2,nq3,nq4;
    int iq1=0,iq2=0,iq3=0,iq4=0;
    cftfrac cffr16,cffr12;

    cffrset( &cffr16, -1, 6 ) ;
    cffrset( &cffr12,  1, 2 ) ;


#ifdef CFKDBG
cffrprt( "cfprojx w=",&cfg->wgtx) ;
#endif

      /* 2-gluon vertex reduction */
    for (i=1 ; i<=cfg->nvtx ; i++)
    if ( cfg->lvtx[i] == CFVGG )
    {
      ix = 0 ;
      ng1= ng2=0 ;
      for (j=1 ; j<=cfg->nedge ; j++)
      for (jj=0 ; jj<=1        ; jj++)
           if ( cfg->edge[jj][j] == i
             && cfg->tedge[j]==CFGLUON  )
           {
             if (ix>1) 
                    { printf(" (cfprojx) strange 2g vertex (1) \n") ;
                      return ; }
             if (ix==0) {ng1 = j ; nge1 = jj ; }
             if (ix==1) {ng2 = j ; nge2 = jj ; }
             ix++ ;
           }
      if ( ng1==0 || ng2==0 )
                    { printf(" (cfprojx) strange 2g vertex (2) \n") ;
                      return ; }
      if ( cfg->tedge[ng1] != CFGLUON  || cfg->tedge[ng2] != CFGLUON )
                    { printf(" (cfprojx) strange 2g vertex (3) \n") ;
                      return ; }
      cfg->edge[nge1][ng1] = cfg->edge[1-nge2][ng2] ;
      cfg->tedge[ng2] =0 ;
      cfg->edge[0][ng2] =0 ;
      cfg->edge[1][ng2] =0 ;
      cfg->lvtx[i] = 0 ;
    }


      /* 3-gluon vertex reduction */
    ix = -1 ;
    for (i=1 ; i<=cfg->nvtx ; i++)
    {
	if ( cfg->lvtx[i] == CFVGGGP ||
             cfg->lvtx[i] == CFVGGGM   )
        {  ix=i ; break ; }
    }

  if( ix > 0 )
  {
      k=0;
#ifdef MODTK
      for (j=1 ; j<=cfg->nedge ; j++)
      for (jj=0 ; jj<=1        ; jj++)
#else
      for (jj=0 ; jj<=1        ; jj++)
      for (j=1 ; j<=cfg->nedge ; j++)
#endif
      {
	      if ( cfg->tedge[j]>0 && cfg->edge[jj][j] == ix )
              {
                 if( k>2 )
                    { printf(" (cfprojx) strange 3g vertex ? (1)\n") ;
                      return ; }
                 if( cfg->tedge[j] != CFGLUON )
                    { printf(" (cfprojx) strange 3g vertex ? (2)\n") ;
                      return ; }
                 ke[k]= j ;
                 kv[k] = cfg->edge[1-jj][j] ;
                 kjj[k] = jj ;
                 k++ ;
              }
      }
      if( k != 3 )
        { printf(" (cfprojx) strange 3g vertex ? (3)\n") ; return ; }

/*
#ifdef CFKDBG
printf(" 3g; ix,ke,kv,kjj %d; %d %d %d; %d %d %d; %d %d %d \n",
 ix,ke[0],ke[1],ke[2],kv[0],kv[1],kv[2],kjj[0],kjj[1],kjj[2]);
#endif
*/

#ifdef MODTK
      cfg1 = cfxhalloc(cfg->nvtx+2 , cfg->nedge+3 ) ;
      cfg2 = cfxhalloc(cfg->nvtx+2 , cfg->nedge+3 ) ;
#else
      addrcfg++  ;
      cfg1 = &xcfg[addrcfg] ;
      cfxgalloc(cfg1 , cfg->nvtx+2 , cfg->nedge+3 ) ;
      addrcfg++  ;
      cfg2 = &xcfg[addrcfg] ;
      cfxgalloc(cfg2 , cfg->nvtx+2 , cfg->nedge+3 ) ;
#endif
      cfxgcopy( cfg , cfg1 ) ;

           kv0[0] = ix ;
           kv0[1] = cfg->nvtx + 1 ;
           kv0[2] = cfg->nvtx + 2 ;
           for (ja=0 ; ja<3 ; ja++)
           {  cfg1->lvtx[kv0[ja]] = CFVQQG ;
              cfg1->mvtx[kv0[ja]] = 0 ;
              cfg1->edge[kjj[ja]][ke[ja]] = kv0[ja] ;
           }
           for (ja=1 ; ja<=3 ; ja++)
           {  cfg1->tedge[cfg->nedge + ja]=CFQUARK ;
/* TK according to KK
              cfg1->edge[0][cfg->nedge + ja]=kv0[ ja-1 ];
              cfg1->edge[1][cfg->nedge + ja]=kv0[ ja%3 ];
*/
              cfg1->edge[1][cfg->nedge + ja]=kv0[ ja-1 ];
              cfg1->edge[0][cfg->nedge + ja]=kv0[ ja%3 ];
           }

           cfxgcopy( cfg1 , cfg2 ) ;

           for (ja=1 ; ja<=3 ; ja++)
           {
              cfg2->edge[0][cfg->nedge + ja]=cfg1->edge[1][cfg->nedge + ja];
              cfg2->edge[1][cfg->nedge + ja]=cfg1->edge[0][cfg->nedge + ja];
           }

           cffrsml( &cfg1->wgtx , -2 ) ;
           cffrsml( &cfg2->wgtx ,  2 ) ;

           cfprojx(cbase,cfgraph,cfg2) ;
           cfprojx(cbase,cfgraph,cfg1) ;
  }

    else
  {

      /* gluon edge reduction */
    jx = -1 ;
    for (j=1 ; j<=cfg->nedge ; j++)
    { if ( cfg->tedge[j] == CFGLUON ) { jx=j ; break ; } }

    if( jx>0 )
    {
           nv1 = cfg->edge[0][jx] ;
           nv2 = cfg->edge[1][jx] ;

           nq1 = nq2 = nq3 = nq4 = 0 ;
           for (i=1 ; i<=cfg->nedge ; i++)
           {
             if ( cfg->tedge[i] == CFQUARK )
             {
               if (cfg->edge[0][i]==nv1)
               {  nq2=cfg->edge[1][i] ; iq2=i ; }
               if (cfg->edge[1][i]==nv1)
               {  nq1=cfg->edge[0][i] ; iq1=i ; }
               if (cfg->edge[0][i]==nv2)
               {  nq4=cfg->edge[1][i] ; iq4=i ; }
               if (cfg->edge[1][i]==nv2)
               {  nq3=cfg->edge[0][i] ; iq3=i ; }
             }
           }
           if ( nq1==0 || nq2==0 || nq3==0 || nq4==0 )
           { printf(" (cfprojx) Bad gluon bridge, tedg[%d]\n",j) ;
             return  ; }

#ifdef CFKDBG
printf(" g bridge; jx, nv1,nv2 %d %d %d \n",jx,nv1,nv2);
printf(" nq1,nq2,nq3,nq4,iq1,iq2,iq3,iq4 %d %d %d %d %d %d %d %d \n",
         nq1,nq2,nq3,nq4,iq1,iq2,iq3,iq4 );
#endif

#ifdef MODTK
       cfg1 = cfxhalloc(cfg->nvtx , cfg->nedge ) ;
       cfg2 = cfxhalloc(cfg->nvtx , cfg->nedge ) ;
#else
       addrcfg++  ;
       cfg1 = &xcfg[addrcfg] ;
       cfxgalloc(cfg1 , cfg->nvtx , cfg->nedge ) ;
       addrcfg++  ;
       cfg2 = &xcfg[addrcfg] ;
       cfxgalloc(cfg2 , cfg->nvtx , cfg->nedge ) ;
#endif

           cfxgcopy(cfg,cfg1) ;
           cfxgcopy(cfg,cfg2) ;

           cfg1->lvtx[nv1] = 0 ;
           cfg1->lvtx[nv2] = 0 ;
           cfg1->tedge[jx] = 0 ;

           /* update 97.12.20 K.K.*/
           if( iq1!=iq4 )
           {
              cfg1->edge[1][iq3]= nq4 ;
              if( iq4 != iq3 && iq4 != iq1 ) cfg1->tedge[iq4] = 0 ;
           }
           else	
              cfg1->edge[1][iq3]= nq1 ;

           if( iq2!=iq3 )
           {
              cfg1->edge[1][iq1]= nq2 ;
              if( iq2 != iq3 && iq2 != iq1 ) cfg1->tedge[iq2] = 0 ;
           }
           else	
              cfg1->edge[1][iq1]= nq3 ;

           cfg2->lvtx[nv1] = 0 ;
           cfg2->lvtx[nv2] = 0 ;
           cfg2->tedge[jx] = 0 ;

           /* update 97.12.20 K.K.*/
           if( iq1!=iq4 )
           {
              cfg2->edge[1][iq1]= nq4 ;
              if( iq4 != iq3 && iq4 != iq1 ) cfg2->tedge[iq4] = 0 ;
           }
           else	
              cfg2->edge[1][iq1]= nq1 ;

           if( iq2!=iq3 )
           {
              cfg2->edge[1][iq3]= nq2 ;
              if( iq2 != iq3 && iq2 != iq1 ) cfg2->tedge[iq2] = 0 ;
           }
           else	
              cfg2->edge[1][iq3]= nq3 ;

           cffrmul( &cfg1->wgtx ,&cffr16 ) ;
           cffrmul( &cfg2->wgtx ,&cffr12 ) ;

           cfprojx(cbase,cfgraph,cfg2) ;
           cfprojx(cbase,cfgraph,cfg1) ;
    }
    else

    /* Only quark strings remain. Calculate weight.*/
    {

#ifdef CFKDBG
for (j=1 ; j<=cfg->nedge ; j++)
    { printf(" cfg->tedge[%d] =%d; %d --> %d \n",j,cfg->tedge[j],
      cfg->edge[0][j],cfg->edge[1][j] ); }
printf(" call cfprojw \n");
#endif

    cfprojw( cbase, cfgraph, cfg ) ;

    /*  free cfg  */
    cfxgfree(cfg) ;
    /* addrcfg-- ; */
    }
  }
}

/*==================================================================*/
Private Void cfprojw(cbase,cfgraph,cfg)
cftbase   *cbase;
cftgraph  *cfgraph;
cftxgraph *cfg;
/*--------------------------------------------------------------------
 * Calculate weight of quark strings, and find the
 *
 * cfgraph : pointer to graph information, return with final weight.
 * cfg    : pointer to graph structure (working)
 */
{
    int i, j, jj, jjj, ix, jx, jxx;
    cftfrac w;
    int  nqtot,idx;
    int  qstr[CFKMAXSTR+CFKMAXSTR] ;

    /* connect between external antiquark to quark */

    nqtot = cbase->nquarkp + cbase->ngluon ;
    cffrset( &w, 1, 1 ) ;
    ix = 0 ;
    if (nqtot > 0 )
    {
      for (i=0 ; i<nqtot ; i++)
      {
          qstr[ix++]= cbase->antiq[i] ;
          jx = -1 ;
          for( j=1 ; j<=cfg->nvtx ; j++ )
          {  if ( cfg->mvtx[j] == cbase->antiq[i] ) { jx=j ; break ; } }

          if( jx<1 )
          { printf(" (cfprojw) cannot find antiq-%d(%d)\n",cbase->antiq[i],i);
            return ; }

          while ( cfg->lvtx[jx] != CFVEXTQ )
          {
            jjj=-1 ;
            for( jj=0 ; jj<=1 ; jj++)
            for( j=1 ; j<=cfg->nedge ; j++)
            {
              if ( cfg->tedge[j] !=0 && cfg->edge[jj][j] ==jx )
              {
                jjj=cfg->edge[1-jj][j] ;
                cfg->tedge[j]=0 ;
                break ;
              }
            }

              if( jjj<1 )
              { printf(" (cfprojw) not find edge for antiq-%d(%d, %d) \n",
                cbase->antiq[i],i,jx) ;
                return ; }
              jx=jjj ;
          }

#ifdef CFKDBG1
printf(" *) i,ix,jx = %d, %d \n",i,ix,jx);
#endif
            qstr[ix++] = cfg->mvtx[jx] ;
      }
    }
#ifdef CFKDBG
printf(" ix=%d \n",ix);
for(i=0;i<ix;i++) printf("(%d) %d ",i,qstr[i]);
printf(" ...qstr \n");
#endif

    /* closed quark loop */

    for( j=1 ; j<=cfg->nedge ; j++ )
    {
      if ( cfg->tedge[j] == CFQUARK )
      {
        jjj =  cfg->edge[0][j] ;
        jx  = j ;
        while ( cfg->edge[1][jx] != jjj )
        {
          jxx = -1 ;
          for( jj=1 ; jj<=cfg->nedge ; jj++)
          {
            if ( cfg->tedge[jj] !=0 &&
                (cfg->edge[0][jj] == cfg->edge[1][jx]) )
                {
                  jxx = jx= jj ; cfg->tedge[jj]=0 ; break ;
                }
          }
          if( jxx<1 )
          { printf(" (cfprojw) loop unclosed; j,jx= %d, %d\n",j,jx) ;
            return ; }
        }
        cffrsml( &w , 3)  ;
#ifdef CFKDBG
printf( " loop... w=%d/ %d, at j=%d\n",w,j) ;
#endif
      }
    }

    /* find corresponding color base */

    if ( nqtot==0 )
	idx=0 ;
    else
    {
      idx = -1 ;
      for (i=0 ; i<cbase->nbase ; i++)
      {
        for (j=0 ; j<cbase->nconid[i] ; j++)
           if ( cbase->conid[i][j] != qstr[j] ) break ;
        if ( j == cbase->nconid[i] ) idx=i ;
      }

      if ( idx <0 || idx>=cbase->nbase )
      {  printf(" (cfprojx) Corresponding color base not found ? \n") ;
         return ;  }
    }

    cffrmul( &w , &cfg->wgtx) ;
    cffradd( &cfgraph->wgt[idx] , &w) ;

#ifdef CFKDBG
printf(" End of cfprojw cfgraph->wgt[%d]=",idx);
cffrpr0(&cfgraph->wgt[idx]);
cffrprt(" w=",&w);
#endif

}

/*==================================================================*/
Private Void cfxgalloc(cfg,nv,ne)
cftxgraph *cfg   ;
int      nv,ne;
/*--------------------------------------------------------------------
 * Allocate memory for structure cftxgraph
 *
 * cfg    : pointer to graph structure
 * nv     : number of vertices
 * ne     : number of edges
 */
{
#ifdef DBGFREE
    printf("cfxgalloc: addrcfg = %d\n", addrcfg);
#endif
/*    cfg = ( struct cftxgraph * ) malloc( sizeof(struct  cftxgraph) ) ; */
/*    cfg = &xcfg ;*/
    if( cfg == NULL )
    {
        fprintf(Stderr, "*** cfxgalloc :0: memory allocation failed\n");
        exit(1);
    }
    cfg->nvtx  = nv ;
    cfg->nedge = ne ;

    cfg->mvtx = ( int * ) malloc( (nv+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->mvtx = %p->%p\n", cfg, cfg->mvtx);
#endif

    cfg->lvtx = ( int * ) malloc( (nv+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->lvtx = %p->%p\n", cfg, cfg->lvtx);
#endif
    if( cfg->mvtx == NULL || cfg->lvtx == NULL )
    {
        fprintf(Stderr, "*** cfxgalloc :1: memory allocation failed\n");
        exit(1);
    }

    cfg->tedge   = ( int * ) malloc( (ne+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->tedge = %p->%p\n", cfg, cfg->tedge);
#endif

    cfg->edge[0] = ( int * ) malloc( (ne+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->edge[0] = %p->%p\n", cfg, cfg->edge[0]);
#endif

    cfg->edge[1] = ( int * ) malloc( (ne+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->edge[1] = %p->%p\n", cfg, cfg->edge[1]);
#endif
    if( cfg->tedge == NULL    ||
        cfg->edge[0] == NULL  || cfg->edge[1] == NULL )
    {
        fprintf(Stderr, "*** cfxgalloc :2: memory allocation failed\n");
        exit(1);
    }

    cffrset(&cfg->wgtx, 1, 1) ;

#ifdef CFKDBG1
printf(" addrcfg=%d, nv,ne=%d,%d\n",addrcfg,nv,ne) ;
printf(" //cfg:%d  ->nvtx:%d  ->nedfe:%d\n",cfg,&cfg->nvtx,&cfg->nedge);
printf(" lvtx:%d -> %d ; mvtx:%d -> %d \n",
&cfg->lvtx,cfg->lvtx,&cfg->mvtx,cfg->mvtx);
printf(" ->tedge:%d  ->edge:%d , %d  ->wgtx:%d\n",
cfg->tedge,cfg->edge[0],cfg->edge[1],&cfg->wgtx);
#endif
}

/*==================================================================*/
Private Void cfxgfree(cfg)
cftxgraph *cfg   ;
/*--------------------------------------------------------------------
 * Dispose memory for structure cftxgraph
 *
 * cfg    : pointer to graph structure
 */
{
    int j;

    for (j=1 ; j>=0 ; j--) {
        if (cfg->edge[j] != NULL) {
#ifdef DBGFREE
            printf("cfxgfree:  free   cfg->edge[%d] = %p->%p\n",
                   j, cfg, cfg->edge[j]);
#endif
            free( cfg->edge[j] );
            cfg->edge[j] = NULL;
        }
    }

    if (cfg->tedge != NULL) {
#ifdef DBGFREE
        printf("cfxgfree:  free   cfg->tedge = %p->%p\n", cfg, cfg->tedge);
#endif
        free( cfg->tedge );
        cfg->tedge = NULL;
    }

    if (cfg->lvtx != NULL) {
#ifdef DBGFREE
        printf("cfxgfree:  free   cfg->lvxt = %p->%p\n", cfg, cfg->lvtx);
#endif
        free(cfg->lvtx);
        cfg->lvtx = NULL;
    }

    if (cfg->mvtx != NULL) {
#ifdef DBGFREE
        printf("cfxgfree:  free   cfg->mvtx = %p->%p\n", cfg, cfg->mvtx);
#endif
        free( cfg->mvtx );
        cfg->mvtx = NULL;
    }

#ifdef DBGFREE
    printf("cfxgfree:  addrcfg = %d\n", addrcfg);
#endif
}

#ifdef MODTK
/*==================================================================*/
Private cftxgraph *cfxhalloc(nv,ne)
int      nv,ne;
/*--------------------------------------------------------------------
 */
{
    cftxgraph *cfg;

    if(addrcfg >= MAXCFG) {
        fprintf(Stderr, "*** cfxhalloc : addrcfg = %d >= %d\n",
                addrcfg, MAXCFG);
        exit(1);
    }
    cfg = xcfg + (addrcfg++);

    cfxgalloc(cfg, nv, ne);

    return cfg;
}

/*==================================================================*/
Private Void cfxhfree(NoArg)
/*--------------------------------------------------------------------
 * Dispose memory for structure cftxgraph
 *
 * cfg    : pointer to graph structure
 */
{
    while(addrcfg > 0) {
        --addrcfg;
        cfxgfree(xcfg + addrcfg);
    }
}
#endif

/*==================================================================*/
Private Void cfxgcopy(cfg,cfg1)
cftxgraph *cfg, *cfg1;
/*--------------------------------------------------------------------
 * copy structure cftxgraph
 *
 * cfg    : pointer to graph structure (source)
 * cfg1   : pointer to graph structure (target)
*/
{
   int i ;

   for (i=0 ; i<=cfg->nvtx ; i++ )
   {  cfg1->mvtx[i] = cfg->mvtx[i] ;
      cfg1->lvtx[i] = cfg->lvtx[i] ;  }

   for (i=0 ; i<=cfg->nedge ; i++ )
   {  cfg1->tedge[i] = cfg->tedge[i] ;
      cfg1->edge[0][i] = cfg->edge[0][i] ;
      cfg1->edge[1][i] = cfg->edge[1][i] ;  }

   cffrcpy( &cfg1->wgtx, &cfg->wgtx) ;

}

/*==================================================================*/
Private int cfbinit(cbase, nq, ng, nb)
cftbase  *cbase;
int      nq,ng,nb;
/*--------------------------------------------------------------------
 * Allocate memory for color base
 *
 * cbase  : pointer to color base data structure
 * nq     : the number of quark pairs
 * ng     : the number of gluons
 * nb     : the number of color base
 */
{
    int i, j;

    cbase->nquarkp= nq;
    cbase->ngluon = ng;
    cbase->nbase  = nb;

    if (nb==0)
    {  cbase->nbase = 1 ;
       cbase->nexcolor = 0 ; }
    else
       cbase->nexcolor = 1 ;

    if ( cbase->nexcolor == 1 )
    {
        cbase->quark  = (int*) malloc(sizeof(int)*(nq+ng));
#ifdef DBGFREE
        printf("cfbinit:  malloc cbase->quark = %p->%p\n", cbase, cbase->quark);
#endif


        cbase->antiq  = (int*) malloc(sizeof(int)*(nq+ng));
#ifdef DBGFREE
        printf("cfbinit:  malloc cbase->antiq = %p->%p\n", cbase, cbase->antiq);
#endif

        cbase->gluon  = (int*) malloc(sizeof(int)*ng);
#ifdef DBGFREE
        printf("cfbinit:  malloc cbase->gluon = %p->%p\n", cbase, cbase->gluon);
#endif
    }

    cbase->constr = (int*) malloc(sizeof(int)*cbase->nbase);
#ifdef DBGFREE
    printf("cfbinit:  malloc cbase->constr = %p->%p\n", cbase, cbase->constr);
#endif

    cbase->nconid = (int*) malloc(sizeof(int)*cbase->nbase);
#ifdef DBGFREE
    printf("cfbinit:  malloc cbase->nconid = %p->%p\n", cbase, cbase->nconid);
#endif

    cbase->conid  = (int**) malloc(sizeof(int*)*cbase->nbase);
#ifdef DBGFREE
    printf("cfbinit:  malloc cbase->conid = %p->%p\n", cbase, cbase->conid);
#endif

    cbase->cwgt   = (cftfrac**) malloc(sizeof(cftfrac**)*cbase->nbase);
#ifdef DBGFREE
    printf("cfbinit:  malloc cbase->cwgt = %p->%p\n", cbase, cbase->cwgt);
#endif

    if(cbase->constr == NULL ||
       cbase->nconid == NULL ||
       cbase->conid  == NULL ||
       cbase->cwgt   == NULL)
    {
        fprintf(Stderr, "*** cfbinit 1: memory allocation failed\n");
        exit(1);
    }

    for(i = 0; i < cbase->nbase; i++) {
        cbase->constr[i] = 0;
        cbase->nconid[i] = 0;
        cbase->conid[i]  = NULL;
        cbase->cwgt[i]   = (cftfrac*) malloc(sizeof(cftfrac)*cbase->nbase);
#ifdef DBGFREE
        printf("cfbinit:  malloc cbase->cwgt[%d] = %p->%p\n",
               i, cbase, cbase->cwgt[i]);
#endif
        if(cbase->cwgt[i] == NULL) {
            fprintf(Stderr, "*** cfbinit 2: memory allocation failed\n");
            exit(1);
        }
        for(j = 0; j < cbase->nbase; j++)
            cffrclr(&cbase->cwgt[i][j]);
    }
    return 1 ;
}

/*==================================================================*/
Private int cfcinit(cbase, nc)
cftbase  *cbase;
int      nc;
/*--------------------------------------------------------------------
 * Allocate memory for color base
 *
 * cbase  : pointer to color base data structure
 * nc     : for which color base space of conid is allocated
 */
{
    int i;

    cbase->conid[nc] = (int*) malloc(sizeof(int)*cbase->nconid[nc]);
#ifdef DBGFREE
    printf("cfcinit:  malloc cbase->conid[%d] = %p->%p\n",
               nc, cbase, cbase->conid[nc]);
#endif
    if(cbase->conid[nc] == NULL) {
        fprintf(Stderr, "*** cfcinit : memory allocation failed\n");
        exit(1);
    }
    for(i = 0; i < cbase->nconid[nc]; i++)
        cbase->conid[nc][i] = 0;

#ifdef MODTK
    return 1;
#endif
}


/*********************************************************************
 * Dump data
 */
/*==================================================================*/
Public  Void cfdump(cbase, cfgraph)
cftbase  *cbase;
cftgraph *cfgraph;
/*--------------------------------------------------------------------
 * Dump data for debugging
 */
{
    cfbdump(cbase);
    cfgdump(cfgraph);
}

/*==================================================================*/
Public  Void cfbdump(cbase)
cftbase  *cbase;
/*--------------------------------------------------------------------
 * Dump data for debugging
 */
{
    int i, j;

    printf("*** dump of cbase ***\n");
    if(cbase == NULL) {
       printf("    cbase = NULL\n");
       return;
    }

    printf("    nex = %d\n", cbase->nex);
    if(cbase->cptype == NULL) {
        printf("    cptype = NULL\n");
    } else {
        for(i = 1; i <= cbase->nex; i++)
#ifdef MODTK
            printf("    cptype[%d] = %d (%d:%s)\n", i, cbase->cptype[i],
                        cfnoind(-i), cfparticle[cbase->cptype[i]]);
#else
            printf("    cptype[%d] = %d (%s)\n", i, cbase->cptype[i],
                        cfparticle[cbase->cptype[i]]);

#endif
    }

    printf("\n    nbase  = %d\n", cbase->nbase);
    if(cbase->cwgt == NULL) {
        printf("    cwgt   = NULL\n");
    } else {
        for(i = 0; i < cbase->nbase; i++) {
            printf("    cwgt[%4d] = ", i);
            if(cbase->cwgt[i] == NULL) {
                printf("    NULL ");
            } else {
                for(j = 0; j < cbase->nbase; j++) {
                    cffrpr0(&cbase->cwgt[i][j]);
                    printf(",");
                    if(j % 5 == 4) {
                        printf("\n                 ");
                    }
                }
            }
            printf("\n");
        }
    }

    if(cbase->constr == NULL && cbase->nconid == NULL &&
       cbase->conid == NULL) {
        printf("    constr = NULL\n");
        printf("    nconid = NULL\n");
        printf("    conid  = NULL\n");
    } else {
        printf("\n       i:constr:nconid: conid ...\n");
        for(i = 0; i < cbase->nbase; i++) {
            printf("    %4d :", i);
            if(cbase->constr == NULL)
                printf("NULL :");
            else
                printf("%4d :", cbase->constr[i]);
            if(cbase->nconid == NULL)
                printf("NULL :");
            else
                printf("%4d :", cbase->nconid[i]);
            if(cbase->conid == NULL)
                printf("NULL :");
            else
                for(j = 0; j < cbase->nconid[i]; j++) {
                    printf("%4d(%4d)", cbase->conid[i][j],
                        cfnoind(-(cbase->conid[i][j])));
                    if(j % 5 == 4) {
                        printf("\n                      ");
                    }
                }
            printf("\n");
        }
    }
}

/*==================================================================*/
Public  Void cfgdump(cfgraph)
cftgraph *cfgraph;
/*--------------------------------------------------------------------
 * Dump data for debugging
*/
{
    int i;

    printf("*** dump of cfgraph ***\n");
    if(cfgraph == NULL) {
       printf("    cfgraph = NULL\n");
       return;
    }

    printf("    nvtx = %d,  nedge = %d\n",
           cfgraph->nvtx, cfgraph->nedge);

    if(cfgraph->nleg == NULL && cfgraph->lvtx == NULL) {
        printf("    nleg  = NULL, lvtx = NULL\n");
    } else {
#ifdef MODTK
       /*
        * Replaced by TK in order to display the original node number
        * defined in "out.grf".
        */
        printf("    i(org), nleg[i], lvtx[i]\n");
        for(i = 1; i <= cfgraph->nvtx; i++) {
            printf(" %4d(%d)", i, cfnoind(i));
#else
        printf("    i, nleg[i], lvtx[i]\n");
        for(i = 1; i <= cfgraph->nvtx; i++) {
            printf(" %4d", i);
#endif
            if(cfgraph->nleg == NULL)
                printf(" NULL");
            else
                printf(" %8d", cfgraph->nleg[i]);
            if(cfgraph->lvtx == NULL)
                printf(" NULL");
            else
                printf(" %8d (%s)", cfgraph->lvtx[i],
                                   cfvertex[cfgraph->lvtx[i]]);
            printf("\n");
        }
    }
    printf("\n");

    if(cfgraph->tedge == NULL && cfgraph->edge == NULL) {
        printf("    tedge = NULL, edge = NULL\n");
    } else {
        printf("    i, edge[0][i], edge[1][i], tedge[i]\n");
        for(i = 1; i <= cfgraph->nedge; i++) {
                printf(" %4d", i);
#ifdef MODTK
            if(cfgraph->edge[0] == NULL)
                printf("     NULL");
            else
                printf(" %4d(%3d)", cfgraph->edge[0][i],
                                    cfnoind(cfgraph->edge[0][i]));
            if(cfgraph->edge[1] == NULL)
                printf("      NULL");
            else
                printf(" %4d(%3d)", cfgraph->edge[1][i],
                                    cfnoind(cfgraph->edge[1][i]));
#else
            if(cfgraph->edge[0] == NULL)
                printf("     NULL");
            else
                printf("  %8d", cfgraph->edge[0][i]);
            if(cfgraph->edge[1] == NULL)
                printf("      NULL");
            else
                printf("  %8d", cfgraph->edge[1][i]);
#endif
            if(cfgraph->tedge == NULL)
                printf(" NULL");
            else
                printf("  %8d (%s)", cfgraph->tedge[i],
                                    cfparticle[cfgraph->tedge[i]]);
            printf("\n");
        }
    }
}

/*=========================================================*/
/*           c f k p e r m                                 */
/*---------------------------------------------------------*/
/* Find next one for p[1]..p[n] in dictionary order.       */
/* Return -1, if next ordering no more exists.             */
/* Watch n, as no check is done for the size of array.     */
/* p[0] is used as a working area.                         */
/*---------------------------------------------------------*/
Private int cfkperm( n , p )
int n;
int p[];
{
	int i,j,x ;

	x = 0 ;
	for (i=1 ; i<=n ; i++)
		if (x>p[i]) x=p[i] ;

	p[0]= --x ; /* p[0] is smaller than any p[ ] */
	i=n-1 ;
	while (p[i]>=p[i+1]) i--;

	if (i>0)
	{	j=n;
		while (p[i]>=p[j]) j--;
		x=p[i] ; p[i]=p[j] ; p[j]=x ;
		i++;
		j=n;
		while (i<j)
		{	x=p[i] ; p[i]=p[j] ; p[j]=x ;
			i++;
			j--;
		}
		return 0 ;
	}
	else return -1;
}

/*----------------------------------------*/
/* First argument is replaced by calling  */
/* (cffradd)   frct <-- frct + frc        */
/* (cffrsub)   frct <-- frct - frc        */
/* (cffrmul)   frct <-- frct * frc        */
/* (cffrdiv)   frct <-- frct / frc        */
/* (cffrsml)   frct <-- frct * c          */
/*----------------------------------------*/

Public Void cffradd( frct , frc )
cftfrac *frct, *frc;
{
	frct->num = (frct->num * frc->den)
                  + (frct->den * frc->num) ;
	frct->den =  frct->den * frc->den  ;
	cffrrdc( frct );
}

Public Void cffrsub( frct , frc  )
cftfrac *frct, *frc;
{
	frct->num = (frct->num * frc->den)
                  - (frct->den * frc->num) ;
	frct->den =  frct->den * frc->den  ;
	cffrrdc( frct );
}

Public Void cffrmul( frct , frc  )
cftfrac *frct, *frc;
{
	frct->num = frct->num * frc->num ;
	frct->den = frct->den * frc->den ;
	cffrrdc( frct );
}

Public Void cffrdiv( frct , frc  )
cftfrac *frct, *frc;
{
	if (frc->num==0)
		{ printf(" (cffrdiv) Err. \n"); return ;}
	frct->num = frct->num * frc->den ;
	frct->den = frct->den * frc->num ;
	cffrrdc( frct );
}

Public Void cffrsml( frct , c )
cftfrac  *frct;
int c;
{
	frct->num = c * frct->num ;
	cffrrdc( frct );
}

Public Void cffrset( frc, num, den )
cftfrac *frc;
int num,den;
{
	frc->num = num ;
	frc->den = den ;
	cffrrdc( frc );
}

Public Void cffrclr( frc )
cftfrac *frc;
{
	frc->num = 0 ;
	frc->den = 1 ;
}

Public Void cffrcpy( frct , frc )
cftfrac *frct, *frc;
{
	frct->num =  frc->num ;
	frct->den =  frc->den ;
}

Public Void cffrrdc(frc)
cftfrac *frc;
{
	int x,y,z;

	x = frc->num ;
	y = frc->den ;

	if (y==0) { printf(" TROUBLE: ZERO DENOMINATOR \n");
		    return ;
		  }

	if (x==0) { frc->den =1 ;
		    return ;
		  }

	if ( x<0 )  x=-x ;
	if ( y<0 )  y=-y ;

	if ( x>y ) {   z=x;
		       x=y;
                       y=z;  }
	while( x>0 ){
		z = y % x ;
		y = x ;
		x = z ;
	}
	if (y==0) { printf(" (cffrrdc) Err. \n"); return ;}

	frc->num = frc->num/y ;
	frc->den = frc->den/y ;

        if( frc->den<0 )
		{ frc->num = -frc->num ;
		  frc->den = -frc->den ; }
	
}

Public Void cffrprt(name,frc)
cftfrac *frc;
char *name;
{
	printf(" %s  %4d/%4d \n",name,frc->num,frc->den);
}

Public Void cffrpr0(frc)
cftfrac *frc;
{
	printf(" %4d/%4d",frc->num,frc->den);
}
