/*********************************************************************
 *  File : cfgrf.c
 *
 *  Calculate color factor
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "grdata.h"
#include "rational.h"

/*--------------------------------------------------------------------
 */
/*
#define DBGFREE
*/

/*--------------------------------------------------------------------
 * Private references
 */
Private Void    cfcilin      ARG((NoArg));
Private Void    cfdelnod     ARG((node,int));
Private Void    cfdelilin    ARG((int));
Private long    cfconc       ARG((int*));
Private long    cfalloc      ARG((NoArg));
Private Void    cffree       ARG((NoArg));
Private Void    cfdump       ARG((NoArg));
Private long    cfnewgr      ARG((NoArg));
Private Void    cfdelc       ARG((int));
Private Void    cfreset      ARG((int));
Private bool    cfsvl        ARG((rational));
Private Void    cfdelglp     ARG((rational));
Private Void    cfexgl3      ARG((rational));

/* local variables */
Local int  nilines;
Local int  nnodes, ncnodes;
Local int  qcd;

int      *cf3g;
int      *ndwork;
int      nstack;

#define CGDEL     -1
#define CNOFFS    -2

typedef struct {
    int vtype;    /* CGDEL or original node number */
    int next[3];
} tcgr;

tcgr *cgr = NULL;
int   ncg;

#define N3     3L
#define N8     (N3*N3-1L)

#define CFMAX  100
typedef struct {
    long num;
    long den;
    long count;
} tcftbl;

Local tcftbl cftable[CFMAX];
Local int    cfntable = 0;

#if DBG
Local int    cfdbg = 1;
#endif

/*********************************************************************
 * Calculation of color factor.
 * Gluon 4 point vertex should not appear in the current graph.
 * The current graph is modified and it should be reset before
 * use the current graph again.
 */

/*==================================================================*/
Public  Void    cfactor(num, den)
long    *num;
long    *den;
/*--------------------------------------------------------------------
 *    Calculate color factor of "cgraph".
 *    External particle should not appear in the cgraph.
 *    If qcd vertices do not appear, this function returns 1/1.
 *    As side-effect, "cgraph" is modified.
 *
 */
{
    long lcgl, l, pfglb;
    int  left;
    ratio    sum, term;
    rational psum=&sum, pterm=&term;

#if DBG
    if(dbg >= cfdbg) {
        printf("cfactor: begin\n");
    }
#endif
    nilines = cgraph->nilines;
    nnodes  = ndnodes();
    qcd     = cprocess->qcd;

    cfcilin();               /* delete color singlet propagators */
    pfglb = cfconc(&left);    /* delete 2-point vertices */

    if(left < 1) {
        *num = pfglb;
        *den = 1L;
#if DBG
        if(dbg >= cfdbg) {
            printf("cfactor: end(%ld/%ld) :no qcd vertices\n", *num, *den);
        }
#endif
        return;
    }

    lcgl = cfalloc();   /* allocate work are */

    rtset(psum,  0L, 1L);
    for(l = 0; l < lcgl; l++) {
        /* expand 3 gluon vertices and calculate color factor */
        cfexgl3(pterm);

        rtadd(psum, pterm);
    }

    rtmuls(psum, pfglb, 1L);

    *num = psum->num;
    *den = psum->den;

    cffree();

    return;
}

/*==================================================================*/
Private Void    cfcilin(NoArg)
/*--------------------------------------------------------------------
 * Delete i-lines which have no color.
 */
{
    node nd;
    int  il, lg;

    for(il = 1; il < nilines; il++) {
        nd = iltable[il].ilnodes[0];
        lg = iltable[il].illegs[0];
        if(ptable[abs(ndparticle(nd, lg))]->color <= 1) {
            cfdelilin(il);
        }
    }
}

/*==================================================================*/
Private Void    cfdelnod(nd, lg)
node    nd;
int     lg;
/*--------------------------------------------------------------------
 * Delete leg "lg" of node "nd".
 */
{
    int lh, il;

/*
    if(ndtype(nd) == NDEXTERN) {
        error("cfdelnod", "cannot delete external particle");
        erfail();
    }
*/
    if(ndtype(nd) == NDDEL) {
/* TK : called for both sides of a line ???
        error("cfdelnod", "node is already deleted");
        fprintf(Stderr, "node = %d\n", nd);
        erfail();
*/
        return;
    }

    /* delete the leg */
    ndlegs(nd) = ndlegs(nd) - 1;

    /* no more legs : delete the node */
    if(ndlegs(nd) < 1) {
        ndtype(nd) = NDDEL;

    /* delete and pack */
    } else if(ndlegs(nd) > lg) {

        /* move leg from "lh" to "lg" */
        lh = ndlegs(nd);

        /* change leg of iline */
        il = ndiline(nd, lh);
        if(iltable[il].ilnodes[0] == nd) {
            iltable[il].illegs[0] = lg;
        } else {
            iltable[il].illegs[1] = lg;
        }

        ndparticle(nd, lg) = ndparticle(nd, lh);
        ndnextn(nd, lg)    = ndnextn(nd, lh);
        ndiline(nd, lg)    = ndiline(nd, lh);
    }
}

/*==================================================================*/
Private Void    cfdelilin(il)
int     il;
/*--------------------------------------------------------------------
 * Delete i-line "il" from the graph.
 */
{
    int l;

    if(iltable[il].iltype == ILDEL) {
        error("cfdelilin", "iline is already deleted");
        fprintf(Stderr, "iline = %d\n", il);
        erfail();
    }

    for(l = 0; l < 2; l++) {
        cfdelnod(iltable[il].ilnodes[l], iltable[il].illegs[l]);
    }

    iltable[il].iltype = ILDEL;
}

/*==================================================================*/
Private long    cfconc(left)
int     *left;
/*--------------------------------------------------------------------
 * Delete vertices with two legs.
 */
{
    node nd, nex[2];
    int  lex[2], iex[2], gex[2];
    int  i;
    long cfglb = 1L;

    *left = 0;
    for(nd = 0; nd < nnodes; nd++) {
        if(ndtype(nd) == NDDEL) {
            continue;
        }

        if(ndlegs(nd) <= 1) {
            error("cfconc", "illegal vertex");
            fprintf(Stderr, "node = %d\n", nd);
            erfail();
        } else if(ndlegs(nd) != 2) {
            (*left)++;
            continue;
        }

        /* check color representation of ilines */
        if(ptable[abs(ndparticle(nd,0))]->color !=
           ptable[abs(ndparticle(nd,1))]->color) {
            error("cfconc", "inconsistent color representation\n");
            fprintf(Stderr, "node = %d\n", nd);
            erfail();
        }

        /* (nex[0],lex[0]) --iex[0],gex[0]-> nd */
        /* nd <-gex[0],iex[0]-- (nex[1],lex[1]) */
        for(i = 0; i < 2; i++) {
            nex[i] = ndnextn(nd, i);
            iex[i] = ndiline(nd, i);
            if(iltable[iex[i]].ilnodes[0] == nd) {
                gex[i] = 0;
            } else {
                gex[i] = 1;
            }
            lex[i] = iltable[iex[i]].illegs[1-gex[i]];
        }

        /* loop with only one node */
        if(nex[0] == nd) {
            cfglb *= ptable[abs(ndparticle(nd, 0))]->color;
            iltable[iex[i]].iltype = ILDEL;
        } else {
            ndnextn(nex[0], lex[0]) = nex[1];
            ndnextn(nex[1], lex[1]) = nex[0];
            ndiline(nex[1], lex[1]) = iex[0];
            iltable[iex[0]].ilnodes[gex[0]] = nex[1];
            iltable[iex[0]].illegs[gex[0]]  = lex[1];
            iltable[iex[1]].iltype = ILDEL;
        }
        ndtype(nd) = NDDEL;
    }
    return cfglb;
}

/*==================================================================*/
Private long    cfalloc(NoArg)
/*--------------------------------------------------------------------
 * Allocate work area.  Returns the number of 3 gluon vertices.
 *
 * Output:
 *      cf3g[node] = [qqg -> 0; ggg -> 1; gggg -> 4; deleted -> -1];
 *      ndilwk(nd, 0) = leg of node of incoming q of qqg
 *      ndilwk(nd, 1) = leg of node of outgoing q of qqg
 *      ndilwk(nd, 2) = leg of node of g
 */
{
    long  lcgl = 1;
    int   lg, col[3];
    node  nd;
    pindex p;

    cf3g = (int*) malloc(nnodes*sizeof(int));
#ifdef DBGFREE
    printf("cfalloc: malloc cf3g = %p\n", cf3g);
#endif
    if(cf3g == NULL) {
        fprintf(Stderr, "*** cfalloc : memory allocation failed\n");
        exit(1);
    }
    ndwork = (int*) malloc(nnodes*sizeof(int));
#ifdef DBGFREE
    printf("cfalloc: malloc ndwork = %p\n", ndwork);
#endif
    if(ndwork == NULL) {
        fprintf(Stderr, "*** cfalloc : memory allocation failed\n");
        exit(1);
    }
    ncnodes = 0;
    for(nd = 0; nd < nnodes; nd++) {
        cf3g[nd] = -1;
        if(ndtype(nd) == NDDEL) {
            continue;
        }
        if(ndlegs(nd) != 3) {
            error("cfalloc", "illegal vertex");
            fprintf(Stderr,"nd = %d, nlegs=%d\n", nd, ndlegs(nd));
            erfail();
        } else {
            for(lg = 0; lg < 3; lg++) {
                ndilwk(nd, lg) = -1;
            }

            for(lg = 0; lg < 3; lg++) {
                p = ndparticle(nd, lg);
                if(p < 0) {
                    col[lg] = - ptable[abs(p)]->color;
                } else {
                    col[lg] = ptable[abs(p)]->color;
                }
                if(col[lg] == -N3) {
                    ndilwk(nd, 0) = lg;
                } else if(col[lg] == N3) {
                    ndilwk(nd, 1) = lg;
                } else if(col[lg] == N8) {
                    ndilwk(nd, 2) = lg;
                } else {
                    error("cfalloc", "illegal colored vertex (a)");
                    fprintf(Stderr,"nd = %d, nlegs=%d\n", nd, ndlegs(nd));
                    erfail();
                }
            }

            if(col[0] == N8 && col[1] == N8 && col[2] == N8) {
                ndilwk(nd, 0) = 0;
                ndilwk(nd, 1) = 1;
                ndilwk(nd, 2) = 2;
                ncnodes += 3;
                lcgl    *= 2;
                cf3g[nd] = 1;
            } else if(ndilwk(nd, 0) < 0 || ndilwk(nd, 1) < 0 ||
                      ndilwk(nd, 2) < 0) {
                error("cfalloc", "illegal colored vertex (b)");
                fprintf(Stderr,"nd = %d, nlegs=%d\n", nd, ndlegs(nd));
                erfail();
            } else {
                ncnodes++;
                cf3g[nd] = 0;
            }
        }
    }
    cgr = (tcgr*) malloc(ncnodes*sizeof(tcgr));
#ifdef DBGFREE
    printf("cfalloc: malloc cgr = %p\n", cgr);
#endif
    if(cgr == NULL) {
        fprintf(Stderr, "*** cfalloc : memory allocation failed\n");
        exit(1);
    }

    return lcgl;
}


/*==================================================================*/
Private Void    cffree(NoArg)
/*--------------------------------------------------------------------
 * Free allocated area by cfalloc.
 */
{
#ifdef DBGFREE
    printf("cffree: free cf3g = %p\n", cf3g);
#endif
    free(cf3g);
    cf3g   = NULL;

#ifdef DBGFREE
    printf("cffree: free ndwork = %p\n", ndwork);
#endif
    free(ndwork);
    ndwork = NULL;

#ifdef DBGFREE
    printf("cffree: free cgr = %p\n", cgr);
#endif
    free(cgr);
    cgr    = NULL;
}

/*==================================================================*/
Private Void    cfdump(NoArg)
/*--------------------------------------------------------------------
 */
{
    int n;

    printf("\ncfdump: nnodes=%d, ncnodes=%d, ncg=%d\n",
           nnodes, ncnodes, ncg);
    printf("nod:cf3g: ndilwk\n");
    for(n = 0; n < nnodes; n++) {
        if(ndtype(n) == NDDEL) {
            printf("%3d: deleted\n", n);
        } else {
            printf("%3d: %3d: (%3d, %3d, %3d)\n",
                   n, cf3g[n], ndilwk(n,0), ndilwk(n,1), ndilwk(n,2));
        }
    }
    if(cgr == NULL) {
        return;
    }
    printf("crg:type: next\n");
    for(n = 0; n < ncnodes; n++) {
        printf("%3d: %3d: (%3d, %3d, %3d)\n",
               n, cgr[n].vtype,
               cgr[n].next[0], cgr[n].next[1], cgr[n].next[2]);
    }
    printf("end of cfdump\n\n");
}

/*==================================================================*/
Private Void    cfg3lp(nd, ncgl, dir, stl, set2l)
node    nd;
int     ncgl, dir, stl;
bool    set2l;
/*--------------------------------------------------------------------
 *    Create quark loop corresponding to a 3 gluon vertex
 *
 *    nd    : original node number.
 *    ncgl  : starting index of created vertex.
 *    dir   : direction of quark loop
 *    stl   : base of legs of original node.
 *    set2l : whether the leg of last vertex is set or not.
 */
{
    int l, d;

    d = (dir + 1)/2;
    for(l = 0; l < 3; l++) {
        cgr[ncgl+l].vtype = nd;         /* original node number */
        cgr[ncgl+l].next[0] = ncgl + (l + 1 + d)%3;
        cgr[ncgl+l].next[1] = ncgl + (l + 2 - d)%3;
        if(set2l || l != 2) {
            cgr[ncgl+l].next[2]
                = CNOFFS - ndiline(nd, ndilwk(nd, l+stl));
        }
    }
}

/*==================================================================*/
Private long    cfnewgr(NoArg)
/*--------------------------------------------------------------------
 *    Create new graph data structure.
 *    Within this function, ndwork is used as a mapping
 *    old node number -> new node number.
 */
{
    node nd;
    int  l, n, nn, i, dir = 0, dir1 = 0;
    int  il;
    bool chng = True;
    long ans = 1L;

#if DBG
    if(dbg >= cfdbg) {
        printf("cfnewgr: begin\n");
    }
#endif
    ncg = 0;
    for(nd = 0; nd < nnodes; nd++) {
        ndwork[nd] = -1;
    }

    for(nd = 0; nd < nnodes; nd++) {
        if(ndtype(nd) == NDDEL) {
            continue;
        }
        if(cf3g[nd] < 0) {
            continue;
        } else if(cf3g[nd] == 0) {   /* qqg */
#if DBG
            if(ncg+1 > ncnodes) {
                error("cfnewgr", "illegal ncg, ncnodes\n");
                fprintf(Stderr, "nd = %d, ncnodes = %d, ncg = %d\n",
                                nd, ncnodes, ncg);
                erfail();
            }
#endif
            ndwork[nd] = ncg;
            cgr[ncg].vtype = nd;
            for(l = 0; l < 3; l++) {
                cgr[ncg].next[l] = CNOFFS - ndiline(nd, ndilwk(nd, l));
#if DBG
                if(dbg >= cfdbg) {
                    printf("cgr[%2d].next[%3d]=%2d, ndilwk(%2d,%2d)=%2d\n",
                        ncg,l,cgr[ncg].next[l],nd,l,ndilwk(nd,l));
                }
#endif
            }
            ncg++;
        } else if(cf3g[nd] == 1 || cf3g[nd] == 2) {  /* ggg */
            if(cf3g[nd] == 1) {
                dir = 1;
            } else {
                dir = -1;
            }
#if DBG
            if(ncg+3 > ncnodes) {
                error("cfnewgr", "illegal ncg, ncnodes\n");
                fprintf(Stderr, "nd = %d, ncnodes = %d, ncg = %d\n",
                                nd, ncnodes, ncg);
                erfail();
            }
#endif
            ans *= 2*dir;   /* ??? */
            ndwork[nd] = ncg;
            cfg3lp(nd, ncg, dir, 0, True);
            ncg += 3;
#if DBG
            if(dbg >= cfdbg) {
                printf("cf3g[%2d]=%d\n", nd, cf3g[nd]);
            }
#endif
            /* generate the next configuration */
            if(chng) {
                cf3g[nd]++;
                if(cf3g[nd] == 3) {
                    cf3g[nd] = 1;
                } else {
                    chng = False;
                }
            }
        } else {
            if(cf3g[nd] == 4) {
                dir  =   1;
                dir1 =   1;
            } else if(cf3g[nd] == 5) {
                dir  = - 1;
                dir1 =   1;
            } else if(cf3g[nd] == 6) {
                dir  =   1;
                dir1 = - 1;
            } else if(cf3g[nd] == 7) {
                dir  = - 1;
                dir1 = - 1;
#if DBG
            } else {
                fprintf(Stderr, "cfnewgr : illegal cf3g\n");
                fprintf(Stderr, "cf3g[%d] = %d\n", nd, cf3g[nd]);
                exit(1);
#endif
            }
#if DBG
            if(ncg+6 > ncnodes) {
                error("cfnewgr", "illegal ncg, ncnodes\n");
                fprintf(Stderr, "nd = %d, ncnodes = %d, ncg = %d\n",
                                nd, ncnodes, ncg);
                erfail();
            }
#endif
            ans *= 4*dir*dir1;   /* ??? */
            ndwork[nd] = ncg;
            cfg3lp(nd, ncg,   dir,  0, False);
            cfg3lp(nd, ncg+3, dir1, 2, False);
            cgr[ncg+2].next[2] = ncg + 5;
            cgr[ncg+5].next[2] = ncg + 2;
            ncg += 6;
            /* generate the next configuration */
            if(chng) {
                cf3g[nd]++;
                if(cf3g[nd] == 8) {
                    cf3g[nd] = 4;
                } else {
                    chng = False;
                }
            }
        }
    }
#if DBG
    if(ncg != ncnodes) {
        fprintf(Stderr, "*** cfnewgr: ncg = %d != ncnodes = %d\n",
                ncg, ncnodes);
        cfdump();
        exit(1);
    }
/*
    if(chng) {
        fprintf(Stderr, "*** cfnewgr: no more configuration\n");
        cfdump();
        exit(1);
    }
*/
#endif

#if DBG
    if(dbg >= cfdbg) {
        printf("cfnewgr: end of phase1\n");
        cfdump();
    }
#endif

    /* renumber node */
    for(n = 0; n < ncnodes; n++) {
        for(l = 0; l < 3; l++) {
            if(cgr[n].next[l] >= 0) {
                continue;
            }
            il  = CNOFFS - cgr[n].next[l]; /* original iline */
#if DBG
            if(dbg >= cfdbg) {
                printf("n(%d,%d)=il%d\n", n, l, il);
            }
#endif

            for(nn = n; nn < ncnodes; nn++) {
                for(i = 0; i < 3; i++) {
                    if(n == nn && l == i) {
                        continue;
                    }
                    if(CNOFFS - cgr[nn].next[i] == il) {
                        cgr[ n].next[l] = nn;
                        cgr[nn].next[i] = n;
#if DBG
                        if(dbg >= cfdbg) {
                            printf("\t===>n(%d,%d)\n", nn, i);
                        }
#endif
                        goto found;
                    }
                }
            }
            /* not found */
            {
                error("cfnewgr","inconsistent connection\n");
                cfdump();
                fprintf(Stderr, "n=(%d, %d)\n", n, l);
                exit(1);
            }
            found: ;
        }
    }

#if DBG
    if(dbg >= cfdbg) {
        printf("cfnewgr: end\n");
    }
#endif

    return ans;
}

/*==================================================================*/
Private Void    cfdelc(n)
int     n;
/*--------------------------------------------------------------------
 *    Delete vertex "n" from cgr-table
 */
{
    if(cgr[n].vtype == CGDEL){
        fprintf(Stderr, "*** cfdelc : node is already deleted\n");
        cfdump();
        fprintf(Stderr, "n = %d\n", n);
        exit(1);
    }

    /* push n to stack */
    ndwork[nstack++] = n;

    /* Is it necessary to save old value of vtype? */
    cgr[n].vtype = CGDEL;
    ncg--;
}

/*==================================================================*/
Private Void    cfreset(ps)
int     ps;
/*--------------------------------------------------------------------
 *    Pop stack until "ps"
 */
{
    int p, n;

    if(ps < 0 || nstack < ps){
        fprintf(Stderr, "*** cfreset : illegal pop of stack ps=%d\n", ps);
        exit(1);
    }

    for(p = nstack-1; p >= ps; p--) {
        n = ndwork[p];
        cgr[cgr[n].next[0]].next[1] = n;
        cgr[cgr[n].next[1]].next[0] = n;
        cgr[cgr[n].next[2]].next[2] = n;
        cgr[n].vtype = CGDEL + 1;
    }

    ncg += nstack - ps;
    nstack = ps;
}

/*==================================================================*/
Private bool    cfsvl(pterm)
rational pterm;
/*--------------------------------------------------------------------
 *    Eliminate self-energy part and bubbles
 */
{
    bool cont = True;
    int  n, nn, i, n1, n2;

#if DBG
    if(dbg >= cfdbg) {
        printf("cfsvl : begin\n");
    }
#endif

    for(i = 0; i < ncnodes; i++) {
        cont = False;
        for(n = 0; n < ncnodes; n++) {
            if(cgr[n].vtype == CGDEL) {
                continue;
            }
            nn = cgr[n].next[0];
            if(nn == cgr[n].next[1]) {
                if(nn == n) {

                    /*
                     * tadpole : Tr(T^a) = 0
                     *
                     *       ------
                     *   a   |    |
                     *   ....x    |
                     *       |    |
                     *       ------
                     */
                    rtset(pterm, 0L, 1L);
#if DBG
                    if(dbg >= cfdbg) {
                        printf("cfsvl : end : tadpole : n=%d\n", n);
                    }
#endif
                    return False;

                } else if(cgr[n].next[2] == nn) {
                    /*
                     * bubble :  Tr(T^a T^a) = N8/2
                     *
                     *       ------
                     *       |    |
                     *       x....x
                     *       |    |
                     *       ------
                     */
#if DBG
                    if(dbg >= cfdbg) {
                        printf("cfsvl : bubble : n=%d, nn=%d\n", n, nn);
                    }
#endif
                    rtmuls(pterm, N8, 2L);
                    cont = True;
                    cfdelc(n);
                    cfdelc(nn);

                } else {
                    /* gluon self-energy :  Tr(T^a T^b) = 1/2 \delta^{ab}
                     *
                     *       ------
                     *   a   |    |   b
                     *   ....x    x....
                     *       |    |
                     *       ------
                     */
#if DBG
                    if(dbg >= cfdbg) {
                        printf("cfsvl : gluon self-energy : n=%d, nn=%d\n",
                               n, nn);
                    }
#endif
                    n1 = cgr[ n].next[2];
                    n2 = cgr[nn].next[2];
                    cgr[n1].next[2] = n2;
                    cgr[n2].next[2] = n1;
                    rtmuls(pterm, 1L, 2L);
                    cont = True;
                    cfdelc(n);
                    cfdelc(nn);
                }

            } else if(nn == cgr[n].next[2]) {
                /* quark self-energy :  T^a T^a = N8/(2 N3)
                 *
                 *       ......
                 *       .    .
                 *       .    .
                 *   ----x----x----
                 */
#if DBG
                if(dbg >= cfdbg) {
                    printf("cfsvl : quark self-energy : n=%d, nn=%d\n", n, nn);
                }
#endif
                n1 = cgr[ n].next[1];
                n2 = cgr[nn].next[0];
                cgr[n1].next[0] = n2;
                cgr[n2].next[1] = n1;
                rtmuls(pterm, N8, 2L*N3);
                cont = True;
                cfdelc(n);
                cfdelc(nn);
            }
        }
        if(!cont) {
#if DBG
            if(dbg >= cfdbg) {
                printf("cfsvl : end : no more\n");
            }
#endif
            return True;
        }
    }
    fprintf(Stderr, "cfsvl: illegal control\n");
    cfdump();
    exit(1);

    /* dummy return */
    return False;
}

/*==================================================================*/
Private Void    cfdelglp(pterm)
rational pterm;
/*--------------------------------------------------------------------
 *    apply Fieltz transformation
 */
{
    int n, n0, n1, n00, n01, n10, n11;
    int ps = nstack;
    ratio    t1, t2;
    rational pt1 = &t1, pt2 = &t2;
#if DBG
    static int level = 0;
#endif

#if DBG
    if(dbg >= cfdbg) {
        printf("cfdelglp: begin %d\n", level++);
        cfdump();
    }
#endif

    if(!cfsvl(pterm)) {
        cfreset(ps);
#if DBG
        if(dbg >= cfdbg) {
            printf("cfdelglp: end   %d cfsvl=False\n", --level);
            cfdump();
        }
#endif
        return;
    }

    if(ncg < 1) {
        cfreset(ps);
#if DBG
        if(dbg >= cfdbg) {
            printf("cfdelglp: end   %d no more vertices\n", --level);
            cfdump();
        }
#endif
        return;
    }

    n0 = -1;
    for(n = 0; n < ncnodes; n++) {
        if(cgr[n].vtype == CGDEL) {
            continue;
        }
        n0 = n;
        break;
    }
    if(n0 < 0) {
        fprintf(Stderr, "*** cfdelglp: no gluon propagator\n");
        cfdump();
        exit(1);
    }

    /*
     *          n0
     *  n00 ->--+-->- n01
     *          |
     *  n10 ->--+-->- n10
     *          n1
     */
    n00 = cgr[n0].next[0];
    n01 = cgr[n0].next[1];
    n1  = cgr[n0].next[2];
    n10 = cgr[n1].next[0];
    n11 = cgr[n1].next[1];

#if DBG
    if(dbg >= cfdbg) {
        printf("cfdelglp: %d -- %d\n", n0, n1);
    }
#endif

    cfdelc(n0);
    cfdelc(n1);

    /*
     *  n00 ->----->- n01       -1
     *                         ----
     *  n10 ->----->- n10      2*N3
     */
    cgr[n00].next[1] = n01;
    cgr[n01].next[0] = n00;
    cgr[n10].next[1] = n11;
    cgr[n11].next[0] = n10;
    rtset(pt1, -1L, 2*N3);
    cfdelglp(pt1);

    /*
     *  n00 ->-\ /->- n01        1
     *          X              -----
     *  n10 ->-/ \->- n11        2
     */
    cgr[n00].next[1] = n11;
    cgr[n01].next[0] = n10;
    cgr[n10].next[1] = n01;
    cgr[n11].next[0] = n00;
    rtset(pt2, 1L, 2L);
    cfdelglp(pt2);

    rtadd(pt1, pt2);
    rtmul(pterm, pt1);

    cfreset(ps);

#if DBG
    if(dbg >= cfdbg) {
        printf("cfdelglp: end   %d normal end\n", --level);
        cfdump();
    }
#endif

}

/*==================================================================*/
Private Void    cfexgl3(pterm)
rational pterm;
/*--------------------------------------------------------------------
 *    Calculate color factor expanding 3 gluon vertices to
 *    a quark loop with the specified direction.
 */
{
    long glb;

    /* expand gluon vertex */
    glb = cfnewgr();

    /* initialize stack of deleted vertices */
    nstack = 0;

    rtset(pterm, 1L, 1L);

    cfdelglp(pterm);

    rtmuls(pterm, glb, 1L);
}

/*********************************************************************
 * Table of color factors
 */

/*==================================================================*/
Public  int     cfind(num, den)
long    num, den;
/*--------------------------------------------------------------------
 * Find color factor and returns its index in the table.
 * If the color factor is not found, it is added to the table.
 */
{
    int i;

    for(i = 0; i < cfntable; i++) {
        if(cftable[i].num == num && cftable[i].den == den) {
            cftable[i].count++;
            return i;
        }
    }
    if(cfntable+1 >= CFMAX) {
        fprintf(Stderr, "*** cfind : table of color factor overflow\n");
        exit(1);
    }
    cftable[cfntable].num = num;
    cftable[cfntable].den = den;
    cftable[cfntable].count = 1;

    cfntable++;

    return (cfntable - 1);
}

/*==================================================================*/
Public  int     cflist(ind, num, den)
int     ind;
long   *num, *den;
/*--------------------------------------------------------------------
 * Read color factor table.
 * This function set the value of num and den, and returns the
 * number of call of cfind for this value.
 * If ind is out of the table, this function returns 0.
 */
{
    if(ind >= cfntable) {
        return 0;
    }

    *num = cftable[ind].num;
    *den = cftable[ind].den;

    return cftable[ind].count;
}

#define CHECK_CFAMP
/*==================================================================*/
Public  Void    cfamp(num, den, con, col)
long    *num;
long    *den;
int     con[];
int     col[];
/*--------------------------------------------------------------------
 *    Calculate color factor of "cgraph".
 *    Qcd vertices should not appear
 *    con[n] : colored ext. fermion connected to "n", which
 *               is colored ext. fermion
 *               if "n" is not colored, the value is -1
 *
 */
{
    int c, c1, l, i0, p;
    node n, n0, n1, n2 = 0, nsrt, nend;
    static int ndflg[MAXNODES];
    int nextern = cprocess->nextern;
#ifdef CHECK_CFAMP
    int f;
#endif

    nnodes  = ndnodes();

    /* flag for visiting nodes */
    for(n = 0; n < nnodes; n++) {
        ndflg[n] = False;
    }

    qcd     = cprocess->qcd;
    if(qcd > 0) {
        error("cfamp", "called for QCD graph");
        erfail();
    }
    *num = 1;
    *den = 1;
    for(n = 0; n < nextern; n++) {
        con[n] = -1;
        col[n] = 1;
    }

    /* find path with color connecting two nodes */
    for(p = 0; p < 2; p++) {        /* extern, loop */
        nsrt = (p == 0) ? 0 : nextern;
        nend = (p == 0) ? nextern : nnodes;
        for(n0 = nsrt; n0 < nend; n0++) {
            if(ndflg[n0]) {
                continue;           /* already visited */
            }
            n = n0;
            ndflg[n] = True;

            if(p == 0) {
                if((c = ptable[abs(ndparticle(n, 0))]->color) == 1) {
                    continue;
                }
                n  = ndnextn(n, 0);
                i0 = ndiline(n, 0);
            } else {
                c = 1;
                for(l = 0; l < ndlegs(n); l++) {
                    n2 = ndnextn(n, l);
                    if((c = ptable[abs(ndparticle(n, l))]->color) != 1) {
                        break;
                    }
                }
                if(c == 1) {
                    continue;
                }
                if(ndflg[n2]) {  /* not possible : 4 point with color? */
                    error("cfamp", "Illegal color");
                    erfail();
                }
                if(n0 == n2) {   /* tadpole */
                    *num *= c;
                    continue;
                }
                i0 = ndiline(n, l);
                n = n2;
            }
            ndflg[n] = True;

            /* advance path until another external particle */
            do {
                n1 = -1;

                /* search nodes around "n" */
                for(l = 0; l < ndlegs(n); l++) {
                    n2 = ndnextn(n, l);
                    if(p == 1 && n2 == n0) {
                        if(ndiline(n, l) == i0) {
                            continue;
                        }
                    } else if(ndflg[n2]) {
                        continue;        /* already visited */
                    }
                    if((c1 = ptable[abs(ndparticle(n, l))]->color) != 1) {
                        if(n1 >= 0) {    /* not unique path */
                            error("cfamp",
                                  "More than 2 colored particles in a vertex.");
                            fprintf(Stderr, "n0 = %d, n = %d\n", n0, n);
                            erfail();
                        }
                        if(c != c1) {  /* check color representation */
                            error("cfamp","Color conservation?\n");
                            fprintf(Stderr, "n0 = %d, n = %d\n", n0, n);
                            erfail();
                        }
                        n1 = n2;
                    }
                }
                if(n1 < 0) {
                    error("cfamp", "Cannot advance path");
                    fprintf(Stderr, "n0 = %d, n = %d\n", n0, n);
                    erfail();
                }
                n = n1;
                ndflg[n] = True;    /* visit */
            } while(n != n0 && ndtype(n) != NDEXTERN);
            if(p == 0) {
                if(n == n0) {
                    error("cfamp","illegal open color line");
                    erfail();
                }
                con[n0] = n;
                con[n ] = n0;
                col[n0] = c;
                col[n ] = c;
            } else {
                if(n != n0) {
                    error("cfamp","illegal closed color line");
                    erfail();
                }
                *num *= c;
            }
        }
    }

#ifdef CHECK_CFAMP
    if(qcd > 0) {
        error("cfamp", "called for QCD graph");
        erfail();
    }
    for(f = 0; f < nflines; f++) {
        for(c = 0; c < fltable[f].length - 1; c++) {
            cfdelilin(ndiline(fltable[f].ndlist[c],
                              fltable[f].lglist[FL_OUT][c]));
        }
        if(fltable[f].type != FLOPEN) {
            cfdelilin(ndiline(fltable[f].ndlist[c],
                              fltable[f].lglist[FL_OUT][c]));
        }
    }
    for(f = 1; f < nilines; f++) {
        if(iltable[f].iltype == ILDEL) {
            continue;
        }
        for(c = 0; c < 2; c++) {
            n = iltable[f].ilnodes[c];
            if(ptable[abs(ndparticle(n, iltable[f].illegs[c])
                         )]->color != 1) {
                error("cfamp", "colored particle is left");
                fprintf(Stderr, "il=%d, leg=%d, nd=%d\n", f, c, n);
                erfail();
            }
        }
    }
#endif
}
