/*********************************************************************
 *  File : cfbase.c
 *                                95.11.17 (K.K.)
 *                                96.10.25 (K.K.)
 *                                03.10.25 (K.K.)
 *                                04.02.20 (K.K.)
 *  Calculate color factor
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "cfnbase.h"
#include "cfnconv.h"

/*--------------------------------------------------------------------
 * parameters
 */
/*
#define DBGFREE
*/
#define MODTK
#define MODKK

#undef CFKDBG          /* for Kato's debug */
#undef  CFKDBG1         /* for Kato's debug */
#define CFKMAXSTR  20   /* maximum number of strings */
#define CFKMAXGLUE  2   /* maximum number of gluons */
#define CFSHIFT   100  /* shift value for gluon reduction */
                       /* gluon=x --> quark=(x+CFSHIFT)         */
                       /*         and antiquark=(x+2*CFSHIFT)   */

#define CFVGGGP    2   /* temporal def. */
#define CFVGGGM    2   /* temporal def. */
#define CFVEXTQ   10   /* temporal def. */
#define CFVEXTAQ  11   /* temporal def. */
#define CFVEXTG   12   /* temporal def. */
#define CFVEXTGH  13   /* temporal def. */

#undef  CFGLUELMT

/*--------------------------------------------------------------------
 * Private functions
 */
Private int  cfbinit        ARG((cftbase*, int, int, int));
#ifdef MODTK
#else
Private int  cfcinit        ARG((cftbase*, int));
#endif

Private int  cfproji        ARG((cftbase*,cftgraph*,cftxgraph*));
Private Void cfprojx        ARG((cftbase*,cftgraph*,cftxgraph*));
Private int  cfprojw        ARG((cftbase*,cftxgraph*));
Private Void cfxgalloc      ARG((cftxgraph*,int,int));
Private Void cfxgfree       ARG((cftxgraph*));
Private Void cfxgcopy       ARG((cftxgraph*,cftxgraph*));
Private Void cfxgdump       ARG((cftxgraph*));
Private int cfkperm         ARG((int,int*));
#ifdef MODTK
Private cftxgraph *cfxhalloc      ARG((int,int));
Private Void       cfxhfree       ARG((NoArg));
#endif

/* local variables */
#ifdef MODTK
#else
Local int nnid[CFPKIND+1];
#endif
#define   MAXCFG 10000
Local int addrcfg = 0;
Local cftxgraph xcfg[MAXCFG];

/*------------------------------------------------------------------*/
typedef struct{ int snum ; int *slen ; int *sstr ;} cftbsbs ;
typedef struct{ int v ; int p ; } qstr;

Private int     cfbase00        ARG((cftgraph*, int, int, int, int));
Private int     cfbase01        ARG((int, int));
Private int     cfbase02        ARG((qstr*, int*, int, int, int, int));
Private int     cfbase04        ARG((int*, int, int*));
Private int     cfbase05        ARG((int, int*, int, int, int, int*, int));
Private int     cfbase07        ARG((cftbsbs*));
Private int     cfbase09        ARG((cftbsbs*, cftbsbs*));
Private Void    cfbase10        ARG((int, int));
Private int     cfbase11        ARG((int, int));
Private int     cfbase12        ARG((int, int));
Private int     cfbase15        ARG((cftbsbs*));
Private int     cfbase21        ARG((cftbsbs*, int));
Private int     cfbase31        ARG((cftbsbs*, int, int));
Private int     cfbase32        ARG((cftbsbs*));


/* working area for color bases*/
#define MAXCFGB 4000
Local   cftbsbs  cfbslst[MAXCFGB];
Local	int	cfnbase;
Local	int	cfnsize;

Local   cftbase   cbaseb0;
Local   cftgraph  colgrb0;
#ifdef MODTK
#else
Local	cftbsbs	  cfbslst0;
#endif

/*==================================================================*/
Public  int cfbase(cbase, ibase, iprnt)
cftbase *cbase;
int     ibase;
int     iprnt;
/*--------------------------------------------------------------------
 * cbase : pointer to color base data structure
 * ibase : input flag for the selection of bases
 * iprnt : input flag for printing information
 *
 *   template by T.Kaneko , July 1995
 *   coded by K.Kato , August 1995
 *   last update on 04/02/20
 *
 * returns status
 */
{
    int i, j, k;
    int k1;
#ifdef MODTK
    int ix;
#else
    int ix, ixx ,  ii;
    cftfrac w ;
#endif
    int next,nb;          /* next=number of external particles */
                          /* nb=number of color bases        */
    int nq,nqa,ng;        /* number of q, qbar, g            */
    int nqtot;  /* nq+ng */
    cftbase *cbase0=&cbaseb0 ;  /* working area for product of two bases */
    cftgraph *cfgrap0=&colgrb0 ;   /* working area for product of two bases */
    int	ne,nv,nb0;

	/* --- pickup colored particles(1) ---*/
	next = cbase->nex ;
	nq = nqa = ng = 0 ;
	for (i=1 ; i<=next ; i++ )
		{ j = cbase->cptype[i] ;
		  switch (j)
		   { case CFAQUARK  : nq++ ; break ;
		     case CFQUARK : nqa++ ; break ;
		     case CFGLUON  : ng++ ; break ;
		   } /* q, qbar, definition adjusted */
		}

	/* --- error check ---*/
	if (nq==0 && nqa==0 && ng==0)
	  { printf(" (cfbase) No colored external particles.  \n") ;
	  }

	if (nq != nqa)
	  { printf(" (cfbase) *BAD* Broken quark-antiquark pair %d:%d . \n",
			nq,nqa) ;
	    return 0;
	  }

	if (nq==0 && ng==1 )
	  { printf(" (cfbase) *BAD* Quark number is 0 and ng=%d.\n",ng) ;
	    return 0;
	  }

#ifdef CFGLUELMT
	if (ng>CFKMAXGLUE)
	  { printf(" (cfbase) *BAD* %d gluons not supported.\n",ng) ;
	    return 0;
	  }
#endif

	if ((nq+ng)>CFKMAXSTR)
	  { printf(" (cfbase) *BAD* Too many strings. nq,ng=%d,%d\n",nq,ng) ;
	    return 0;
	  }

#ifdef CFKDBG
printf("Before creation of color base ........\n");
#endif
	/* --- Number of color-bases ---*/
	nqtot = nq + ng ;

        if (nqtot>0)
	  { cfnsize = 2*nq+ng ;
	    cfbase01( nq, ng ) ;
	    for(i=0; i<cfnbase; i++ ) cfbase07( &cfbslst[i] );
	    cfbase10( 0, cfnbase-1 );
	    nb = cfnbase ; }
        else
        nb = 1 ;

        /* --- allocate cbase --- */
	cfbinit(cbase, nq, ng, nb);
        if ( cbase->nexcolor == 0 )
        {  cbase->constr[0]=0 ;
           cbase->nconid[0]=0 ;
           cffrset(&cbase->cwgt[0][0],1,1) ;
           return 1 ;
        }

	/* --- pickup colored particles(2) record into cbase  ---*/
#ifdef MODTK
	for( i = 1; i <= 2*nq + ng; i++ ) {
            cbase->expt[i] = 0;
        }
#else
	for( i=1 ; i<=cbase->nex ; i++ ) cbase->expt[i]=0;
#endif
	nq = nqa = ng = 0 ;
	for (i=1 ; i<=next ; i++ )
		{ j = cbase->cptype[i] ;
		  switch (j)
		   { case CFAQUARK  : cbase->quark[nq++]=i ; break ;
		     case CFQUARK : cbase->antiq[nqa++]=i ; break ;
		     case CFGLUON  : cbase->gluon[ng++]=i ; break ;
		   } /* q, qbar, definition adjusted */
		}
	if(nq>0)
	for(i=0;i<nq;i++) { k=cbase->quark[i] ; cbase->expt[k]= i+1; }
	if(nq>0)
	for(i=0;i<nq;i++) { k=cbase->antiq[i] ; cbase->expt[k]= nq+i+1; }
	if(ng>0)
	for(i=0;i<ng;i++) { k=cbase->gluon[i] ; cbase->expt[k]= 2*nq+i+1; }

   if (iprnt==1)
   {
        printf("(cfbase)  External=%d, Number of base=%d \n",
          cbase->nex,cbase->nbase);
	if(nq>0) { printf("quark: (%d)",cbase->nquarkp);
	for(i=0;i<nq;i++) printf(" %d",cbase->quark[i]);
	printf("\n"); }
	if(nq>0) { printf("antiq: (%d)",cbase->nquarkp);
	for(i=0;i<nq;i++) printf(" %d",cbase->antiq[i]);
	printf("\n"); }
	if(ng>0) { printf("gluon: (%d)",cbase->ngluon);
	for(i=0;i<ng;i++) printf(" %d",cbase->gluon[i]);
	printf("\n"); }
	{ printf(" expt:");
	for( i=1 ; i<=cbase->nex ; i++ ) printf(" %d",cbase->expt[i]);
	printf("\n"); }
        printf(" --- color bases --- \n");
	for(i=0; i<nb;i++) cfbase21(&cfbslst[i],i);
   }

	/* --- product of color bases --- */

	nb0 = 0 ;
	nv = (2*nq+ng) + 2*ng ;
	ne = 2*(nq+ng+ng) ;
/* printf("ENTER 1........\n");*/
        cfcbinit(cbase0, 0 );
/* printf("ENTER 1........\n");*/
	cfbinit(cbase0, 0, 0, nb0);
/* printf("ENTER 2........\n");*/
        cfcginit(cfgrap0,nv,ne,nb0);

#ifdef CFKDBG
printf("ENTER LOOP for base*base ........\n");
#endif

	for( k1=0 ; k1< cbase->nbase ; k1++)
	for( k =0 ; k <=k1           ; k++ )
	{
	  cfbase00( cfgrap0 , k, k1, nq, ng );
	  cfproj( cbase0, cfgrap0, ibase, 0 );
          cffrcpy( &cbase->cwgt[k1][k], &cfgrap0->wgt[0] ) ;
	}

	if ( cbase->nbase >1 )
	for( k1=0      ; k1< (cbase->nbase-1) ; k1++)
	for( k =(k1+1) ; k <  cbase->nbase    ; k++ )
	{
          cffrcpy( &cbase->cwgt[k1][k], &cbase->cwgt[k][k1] ) ;
	}

   if (iprnt==1)
   {
	printf("\n");
	printf(" (Base * Base) weight table ............\n");
	for (j=0 ; j<nb ; j++)
	{ printf(" cwgt[%d][ ] =",j);
	  for (i=0 ; i<nb ; i++) { cffrpr0(&cbase->cwgt[j][i]);
          if( i%5==4 || i==nb-1) printf("\n"); }
	}

   }

        /* ---- Average for initial state ---- */
        ix = 1 ;
        for (i=1 ; i<=next ; i++ )
	{
	    if ( cbase->initial[i] == CBINITIAL )
		{
		  if( cbase->cptype[i]==CFQUARK ) ix*=3 ;
                  if( cbase->cptype[i]==CFAQUARK) ix*=3 ;
                  if( cbase->cptype[i]==CFGLUON ) ix*=8 ;
                  if( cbase->cptype[i]==CFGHOST ) ix*=8 ;
                  if( cbase->cptype[i]==CFAGHOST) ix*=8 ;
		}
	}

   if (iprnt==1)
          printf(" .. Initial average :divided by  %d \n",ix);

#ifdef MODTK
        cbase->iniav = ix;    /* TK : added line */
#else
	cffrset( &w , 1, ix ) ;
	for( k1=0 ; k1< cbase->nbase ; k1++)
	for( k =0 ; k < cbase->nbase ; k++ )
	{
          cffrmul( &cbase->cwgt[k1][k], &w ) ;
	}
#endif

	return 0;
}


/*==================================================================*/
Public  int cfproj(cbase, cfgraph, ibase, iprnt)
cftbase  *cbase;
cftgraph *cfgraph;
int      ibase ;
int      iprnt;
/*--------------------------------------------------------------------
 * cbase   : pointer to color base data structure
 * cfgraph : information about a graph and its weight
 * ibase   : input flag for the selection of bases
 * iprnt   : input flag for printing information
 *
 *   template by T.Kaneko , July 1995
 *   coded by K.Kato , August 1995
 *   last update on 04/02/20
 *
 */
{
    int i;
    int cleg,cvtx;
#ifdef MODTK
    int nqqg, nggg, nhhg, nvqq, nvgg, next ;
#else
    int nqqg, nggg, nhhg, nvqq, next ;
#endif
    cftxgraph *cfg  ;

   /* count colored vertices */

#ifdef MODTK
   nqqg = nggg = nhhg = nvqq = nvgg = 0 ;
#else
   nqqg = nggg = nhhg = nvqq = 0 ;
#endif

#ifdef MODTK
   for (i=1 ; i<=cfgraph->nvtx ; i++)
     if ( cfgraph->nleg[i] >= 2 )
     {
#else
   for (i=1 ; i<=cfgraph->nvtx ; i++)
     if ( cfgraph->nleg[i] == 3 )
     {
#endif
       if ( cfgraph->lvtx[i] == CFVQQG )    nqqg ++ ;
       if ( cfgraph->lvtx[i] == CFVGGGP ||
            cfgraph->lvtx[i] == CFVGGGM   ) nggg ++ ;
       if ( cfgraph->lvtx[i] == CFVHHG )    nhhg ++ ;
#ifdef MODTK
       if ( cfgraph->lvtx[i] == CFVQQ )     nvqq ++ ;
       if ( cfgraph->lvtx[i] == CFVGG )     nvgg ++ ;
#endif
     }

#ifdef MODTK
   if ( nqqg+nggg+nhhg+nvqq+nvgg == 0 )
   { printf(" (cfproj) no colored vertex \n") ;
     cffrset( &cfgraph->wgt[0], 1, 1 ) ;
     return 0; }
#else
   if ( nqqg+nggg+nhhg+nvqq == 0 )
   { printf(" (cfproj) no colored vertex \n") ;
     cffrset( &cfgraph->wgt[0], 1, 1 ) ;
     return ; }
#endif

   next = 2*cbase->nquarkp + cbase->ngluon ; /* external colored p. */
   /* nqqg += cbase->ngluon ;     gluon reduction NOT included */
#ifdef MODTK
   cvtx = next + nqqg + nggg + nhhg + nvqq + nvgg  ;
   cleg = (next + 3*(nqqg+nggg+nhhg) + 2*(nvqq + nvgg))/2 ;
#else
   cvtx = next + nqqg + nggg + nhhg + nvqq   ;
   cleg = (next + 3*(nqqg+nggg+nhhg) + 2*nvqq)/2 ;
#endif

#ifdef CFKDBG
#ifdef MODTK
printf("cbase->nquarkp,  cbase->ngluon=%d, %d \n",
        cbase->nquarkp,  cbase->ngluon) ;
printf(" next, nqqg,  nggg,  nhhg,  nvqq, nvgg = %d: %d, %d, %d, %d, %d\n",
         next, nqqg,  nggg,  nhhg,  nvqq, nvgg )  ;
#else
printf("cbase->nquarkp,  cbase->ngluon=%d, %d \n",
        cbase->nquarkp,  cbase->ngluon) ;
printf(" next, nqqg,  nggg,  nhhg,  nvqq= %d: %d, %d, %d, %d\n",
         next, nqqg,  nggg,  nhhg,  nvqq )  ;
printf(" ........... cvtx,cleg = %d , %d  \n",cvtx,cleg);
#endif
#endif

   /* --- clear cfgraph->wgt ---*/
   for (i=0 ; i<cbase->nbase ; i++) cffrclr(&cfgraph->wgt[i]) ;

/* K.K. ??-->*/
	cleg = cfgraph->nedge;
	cvtx = cfgraph->nvtx + 2*cbase->nquarkp + cbase->ngluon;
/*<-- K.K. ??*/

   /* --- allocate cfg ---*/
#ifdef MODTK
   if(addrcfg != 0) {
       cfxhfree();
   }
   cfg = cfxhalloc( cvtx , cleg ) ;
#else
   addrcfg = 0 ;
   cfg = &xcfg[0] ;
   cfxgalloc( cfg , cvtx , cleg ) ;
#endif

   /* --- create cfg, reading cfgraph, referring cbase --- */
   cfproji(cbase,cfgraph,cfg) ;


   /*--- Now, cfg is ready. Do main job. ---*/
   cfprojx( cbase ,cfgraph, cfg ) ;


   /* --- result  ---*/
   if (iprnt==1)
   {
   printf(" ------ weights for color bases ------\n");
     for (i=0 ; i<cbase->nbase; i++)
     {  cffrpr0(&cfgraph->wgt[i]);
        if ( (i==(cbase->nbase-1)) || ((i%5)==4) )
              printf("\n");
     }
   }
#ifdef MODTK
   cfxhfree();
   return 1;
#endif
}
/*==================================================================*/
Private int cfproji(cbase, cfgraph, cfg)
cftbase  *cbase;
cftgraph *cfgraph;
cftxgraph *cfg;
/*--------------------------------------------------------------------
 * cbase  : pointer to color base data structure
 * cfgraph : information about a graph and its weight
 * cfg    : pointer to graph structure (working)
 *
 */
{
#ifdef MODTK
    int i, j, je, jv, ii;
#else
    int i, j, je, jv, ii, j1, jj;
#endif
    int leg,vtx ;

        /* Scan cfgraph, only colored particles and vertices  */
        /* are copied into cfg.                              */
        /* Now, external ends are also counted as vertices.  */
        /* Here, mvtx records internal vertex and (-1) times  */
        /* particle id number for external end.                */

   je=0 ;                                /* record colored edge */
   for (i=1 ; i<=cfgraph->nedge ; i++)
   {
       if ( cfgraph->tedge[i] == CFQUARK  ||
            cfgraph->tedge[i] == CFAQUARK ||
            cfgraph->tedge[i] == CFGLUON  ||
            cfgraph->tedge[i] == CFGHOST    )
       {  je++ ;
          cfg->tedge[je]   = cfgraph->tedge[i] ;
          if( cfg->tedge[je]==CFGHOST) cfg->tedge[je]=CFGLUON;
          if( cfg->tedge[je]!=CFQUARK && cfg->tedge[je]!=CFGLUON ) 
		{printf("(cfproji) BAD tedge code = %d at i=%d\n",
			cfg->tedge[je],i); return -1;}
          cfg->edge[0][je] = cfgraph->edge[0][i] ;
          cfg->edge[1][je] = cfgraph->edge[1][i] ;
       }
   }
	cfg->nedge = je;
#ifdef CFKDBG
   printf(" Pass cfproji-1\n");
#endif

   jv=0 ;                                /* record vertex */
   for (i=1 ; i<=cfgraph->nvtx ; i++)    /* internal vertex */
     {
        jv++ ;
        cfg->lvtx[jv] = cfgraph->lvtx[i] ;
	if(cfg->lvtx[jv]==CFVHHG) cfg->lvtx[jv]=CFVGGG;
        cfg->mvtx[jv] = i ;
     }
                                     /* external end */
   if (cbase->nquarkp >0 )
   {
     for (i=0 ; i<cbase->nquarkp ; i++)
     {
         jv++ ;
         cfg->lvtx[jv] = CFVEXTQ ;
         cfg->mvtx[jv] = -cbase->quark[i] ;
     }
     for (i=0 ; i<cbase->nquarkp ; i++)
     {
         jv++ ;
         cfg->lvtx[jv] = CFVEXTAQ ;
         cfg->mvtx[jv] = -cbase->antiq[i] ;
     }
   }
   if (cbase->ngluon >0 )
   {
     for (i=0 ; i<cbase->ngluon ; i++)
     {  
	 jv++ ;
         cfg->lvtx[jv] = CFVEXTG ;
         cfg->mvtx[jv] = -cbase->gluon[i] ;
     }
   }
#ifdef CFKDBG
   printf(" Pass cfproji-2\n");
#endif

/*---
   if( jv!= cfg->nvtx )
   { printf(" (cfproji) Err. for cfg->nvtx. %d:%d \n",jv,cfg->nvtx) ;
                     cfdump(cbase, cfgraph); return 0 ; }
   if( je!= cfg->nedge )
   { printf(" (cfproji) Err. for cfg->nedge. %d:%d \n",je,cfg->nedge) ;
                     cfdump(cbase, cfgraph); return 0 ; }
---*/

         /* Make rvtx */
   for (ii=0 ; ii<3 ; ii++)
   for (i=1; i<=cfg->nvtx; i++) cfg->rvtx[ii][i]=0 ;

   for (i =1 ; i<=cfg->nedge ; i++)
   for (ii=0 ; ii<=1   ; ii++)
   {
     vtx = cfg->edge[ii][i] ;
     leg = i ;
     if (vtx>0)
     {
	if(cfg->lvtx[vtx]==CFVQQ) cfg->rvtx[1-ii][vtx]=leg;
	if(cfg->lvtx[vtx]==CFVGG) 
	{ if(cfg->rvtx[0][vtx]==0) cfg->rvtx[0][vtx]=leg;
	  else cfg->rvtx[1][vtx]=leg;
	}
	if(cfg->lvtx[vtx]==CFVQQG) 
	{  if(cfg->tedge[leg]==CFQUARK) cfg->rvtx[1-ii][vtx]=leg;
	   if(cfg->tedge[leg]==CFGLUON) cfg->rvtx[2][vtx]=leg;    
	}
	if(cfg->lvtx[vtx]==CFVGGG) 
	{ if(cfg->rvtx[0][vtx]==0) cfg->rvtx[0][vtx]=leg;
	  else if(cfg->rvtx[1][vtx]==0) cfg->rvtx[1][vtx]=leg;
	  else cfg->rvtx[2][vtx]=leg;
	}
     }
     else
     {
	for(j=1;j<=cfg->nvtx;j++) if(cfg->mvtx[j]==vtx) jv=j;
	cfg->edge[ii][i]=jv;
	if(cfg->lvtx[jv]==CFVEXTQ)  cfg->rvtx[1][jv]=leg ;
	if(cfg->lvtx[jv]==CFVEXTAQ) cfg->rvtx[0][jv]=leg ;
	if(cfg->lvtx[jv]==CFVEXTG)  cfg->rvtx[0][jv]=leg ;
     }
   }
#ifdef CFKDBG
   printf(" Pass cfproji-3\n");
#endif

#ifdef CFKDBG
	printf("END of cfproji\n");
	cfxgdump(cfg);
#endif

#ifdef MODTK
    return 1;
#endif
}

/*==================================================================*/
Private Void cfprojx(cbase,cfgraph,cfg)
cftbase   *cbase;
cftgraph  *cfgraph;
cftxgraph *cfg;
/*--------------------------------------------------------------------
 * Graph reduction, called iteratively.
 *
 * cbase  : pointer to color base data structure
 * cfgraph : pointer to graph information, return with final weight.
 * cfg    : pointer to graph structure (working)
 */
{
    cftxgraph *cfg1,*cfg2;
#ifdef MODTK
    int i, j, ja;
#else
    int i, ix, j, jj, jx, k, ja;
#endif

    int v1,v2,v3,v4,v5,v6,v0;
    int e1,e2,e3,e4,e5,e6,e0;
    int id;
    int nocfg1,nocfg2;
    cftfrac cffr16,cffr12,cffr43;

    cffrset( &cffr16, -1, 6 ) ;
    cffrset( &cffr12,  1, 2 ) ;
    cffrset( &cffr43,  4, 3 ) ;

#ifdef CFKDBG
cffrprt( " >>>cfprojx: ", &cfg->wgtx );
/* printf("ENTER cfprojx ........\n");
cfxgdump(cfg) ; */
#endif

	if( cfg->wgtx.num == 0 ) return;

      /*---(1)--- quark-quark vertex reduction */
    for (i=1 ; i<=cfg->nvtx ; i++)
    if ( cfg->lvtx[i] == CFVQQ )
    {
	v2 = i;
	e1 = cfg->rvtx[0][v2];
	e2 = cfg->rvtx[1][v2];
	v1 = cfg->edge[0][e1];
	v3 = cfg->edge[1][e2];

	if( v1 != v3 )
	{
	  cfg->lvtx[v2] = 0 ;
	  cfg->rvtx[0][v3] = e1 ;
	  cfg->edge[1][e1] = v3 ;
	  cfg->tedge[e2] = 0 ;
	}
	else   /* quark loop */
	{
	  if( cfg->lvtx[v1]==CFVQQG )
	  {  cffrclr( &cfg->wgtx ); return; }
	  else
	  {  
		cfg->lvtx[v1] = 0 ;
		cfg->lvtx[v2] = 0 ;
		cfg->tedge[e1] = 0 ;
		cfg->tedge[e2] = 0 ;
		cffrsml( &cfg->wgtx , 3 ) ;
	  }
	}
    }

    for (i=1 ; i<=cfg->nedge ; i++)
    if ( cfg->tedge[i] == CFQUARK )
    {
	e1 = i;
	v1 = cfg->edge[0][e1];
	v2 = cfg->edge[1][e1];
	if( v1==v2 )
	{
	  if( cfg->lvtx[v1] == CFVQQG ) 
	    { cffrclr( &cfg->wgtx ); return ;}
	  else
	  {
	     cfg->tedge[e1] = 0 ;
	     cfg->lvtx[v1] = 0 ;
	     cffrsml( &cfg->wgtx , 3 ) ;
	  }
	}
    }

      /*---(2)--- gluon-gluon vertex reduction */
    for (i=1 ; i<=cfg->nvtx ; i++)
    if ( cfg->lvtx[i] == CFVGG )
    {
	v2 = i;
	e1 = cfg->rvtx[0][v2];
	e2 = cfg->rvtx[1][v2];
	if( cfg->edge[0][e1]==v2 ) 
           v1 = cfg->edge[1][e1];
	else
           v1 = cfg->edge[0][e1];
	if( cfg->edge[0][e2]==v2 ) 
           v3 = cfg->edge[1][e2]; 
	else
           v3 = cfg->edge[0][e2]; 

#ifdef CFKDBG
	printf(" gg e1 e2 v1 v2 v3 %d %d %d %d %d \n",e1,e2,v1,v2,v3);
#endif
	if( v1 != v3 )
	{
	  cfg->lvtx[v2] = 0 ;
	  for(j=0; j<3; j++) 
	    if( cfg->rvtx[j][v3] == e2 ) cfg->rvtx[j][v3] = e1 ;
	  for(j=0; j<2; j++)
	    if( cfg->edge[j][e1] == v2 ) cfg->edge[j][e1] = v3 ;
	  cfg->tedge[e2] = 0 ;
	}
	else   /* gluon loop */
	{
	  if( cfg->lvtx[v1]==CFVGGG )
	  {  cffrclr( &cfg->wgtx ); return; }
	  else
	  {  
		cfg->lvtx[v1] = 0 ;
		cfg->lvtx[v2] = 0 ;
		cfg->tedge[e1] = 0 ;
		cfg->tedge[e2] = 0 ;
		cffrsml( &cfg->wgtx , 8 ) ;
	  }
	}
    }

    for (i=1 ; i<=cfg->nedge ; i++)
    if ( cfg->tedge[i] == CFGLUON )
    {
	e1 = i;
	v1 = cfg->edge[0][e1];
	v2 = cfg->edge[1][e1];
	if( v1==v2 )
	{
	  if( cfg->lvtx[v1] == CFVGGG ) 
	    { cffrclr( &cfg->wgtx ); return ;}
	  else
	  {
	     cfg->tedge[e1] = 0 ;
	     cfg->lvtx[v1] = 0 ;
	     cffrsml( &cfg->wgtx , 8 ) ;
	  }
	}
    }

    /* 3g-check :::::: */
    v0 = -1 ;
    for (i=1 ; i<=cfg->nvtx ; i++)
    {
	if ( cfg->lvtx[i] == CFVGGGP ||
             cfg->lvtx[i] == CFVGGGM   )
        {  v0=i ; break ; }
    }

    /* g-bridge check :::::: */
    e0 = -1 ;
    for (j=1 ; j<=cfg->nedge ; j++)
    { if ( cfg->tedge[j] == CFGLUON )
	 {  
	   v1 = cfg->edge[0][j] ;
	   v2 = cfg->edge[1][j] ;
	   if( cfg->lvtx[v1]==CFVQQG && cfg->lvtx[v2]==CFVQQG )
		{ e0=j ; break; } 
         }
    }

    if( e0<0 && v0<0 )
    {
    /* Finished reduction, compute weight */


    if( cbase->nexcolor==0 )
      id = 0 ;
    else 
      id = cfprojw( cbase, cfg ) ;

#ifdef CFKDBG
printf("Here, just after the cfprojw ........\n");
    cffrprt( " *NOW=", &cfgraph->wgt[id]);
    cffrprt( " *ADD=", &cfg->wgtx ); 
/* cfxgdump(cfg) ;*/
#endif

    cffradd( &cfgraph->wgt[id], &cfg->wgtx ); 

    /*  free cfg  */
    cfxgfree(cfg) ;
    /* addrcfg-- ; */

	return ;
  }

      /*---(3)--- 3-gluon vertex reduction */

  if( v0 > 0 )
  {
	e1 = cfg->rvtx[0][v0];
	e2 = cfg->rvtx[1][v0];
	e3 = cfg->rvtx[2][v0];
	if( cfg->edge[0][e1]==v0 ) 
           v1 = cfg->edge[1][e1];
	else
           v1 = cfg->edge[0][e1];
	if( cfg->edge[0][e2]==v0 ) 
           v2 = cfg->edge[1][e2];
	else
           v2 = cfg->edge[0][e2];
	if( cfg->edge[0][e3]==v0 ) 
           v3 = cfg->edge[1][e3];
	else
           v3 = cfg->edge[0][e3];

#ifdef MODTK
      cfg1 = cfxhalloc(cfg->nvtx+2 , cfg->nedge+3 ) ;
      cfg2 = cfxhalloc(cfg->nvtx+2 , cfg->nedge+3 ) ;
#else
      addrcfg++  ;
      cfg1 = &xcfg[addrcfg] ;
      cfxgalloc(cfg1 , cfg->nvtx+2 , cfg->nedge+3 ) ;
      addrcfg++  ;
      cfg2 = &xcfg[addrcfg] ;
      cfxgalloc(cfg2 , cfg->nvtx+2 , cfg->nedge+3 ) ;
#endif

        cfxgcopy( cfg , cfg1 ) ;

	e4 = cfg->nedge + 1 ;
	e5 = cfg->nedge + 2 ;
	e6 = cfg->nedge + 3 ;
	v4 = cfg->nvtx + 1 ;
	v5 = cfg->nvtx + 2 ;
#ifdef CFKDBG
printf(" e1=%d, e2=%d, e3=%d, e4=%d, e5=%d, e6=%d\n",e1,e2,e3,e4,e5,e6);
printf(" v0=%d, v1=%d, v2=%d, v3=%d, v4=%d, v5=%d\n",v0,v1,v2,v3,v4,v5);
#endif

	cfg1->lvtx[v0] = CFVQQG ;
	cfg1->rvtx[0][v0] = e4 ;
	cfg1->rvtx[1][v0] = e6 ;
	cfg1->rvtx[2][v0] = e1 ;
	cfg1->lvtx[v4] = CFVQQG ;
	cfg1->rvtx[0][v4] = e5 ;
	cfg1->rvtx[1][v4] = e4 ;
	cfg1->rvtx[2][v4] = e2 ;
	cfg1->lvtx[v5] = CFVQQG ;
	cfg1->rvtx[0][v5] = e6 ;
	cfg1->rvtx[1][v5] = e5 ;
	cfg1->rvtx[2][v5] = e3 ;
	cfg1->mvtx[v4] = v4 ;
	cfg1->mvtx[v5] = v5 ;

	for(j=0; j<2; j++)
	  if( cfg1->edge[j][e2] == v0 ) cfg1->edge[j][e2] = v4 ;
	for(j=0; j<2; j++)
	  if( cfg1->edge[j][e3] == v0 ) cfg1->edge[j][e3] = v5 ;
	cfg1->tedge[e4] = CFQUARK ;
	cfg1->edge[0][e4] = v4 ;
	cfg1->edge[1][e4] = v0 ;
	cfg1->tedge[e5] = CFQUARK ;
	cfg1->edge[0][e5] = v5 ;
	cfg1->edge[1][e5] = v4 ;
	cfg1->tedge[e6] = CFQUARK ;
	cfg1->edge[0][e6] = v0 ;
	cfg1->edge[1][e6] = v5 ;


        cfxgcopy( cfg1 , cfg2 ) ;

        for (ja=1 ; ja<=3 ; ja++)
           {
              cfg2->edge[0][cfg->nedge + ja]=cfg1->edge[1][cfg->nedge + ja];
              cfg2->edge[1][cfg->nedge + ja]=cfg1->edge[0][cfg->nedge + ja];
           }
              cfg2->rvtx[0][v0]            =cfg1->rvtx[1][v0];
              cfg2->rvtx[1][v0]            =cfg1->rvtx[0][v0];
        for (ja=1 ; ja<=2 ; ja++)
           {
              cfg2->rvtx[0][cfg->nvtx + ja]=cfg1->rvtx[1][cfg->nvtx + ja];
              cfg2->rvtx[1][cfg->nvtx + ja]=cfg1->rvtx[0][cfg->nvtx + ja];
           }

           cffrsml( &cfg1->wgtx , -2 ) ;
           cffrsml( &cfg2->wgtx ,  2 ) ;

#ifdef CFKDBG
printf("Here, ggg-vertex processed ........\n");
printf(" ====> cfg1\n"); cfxgdump(cfg1);
printf(" ====> cfg2\n"); cfxgdump(cfg2);
#endif

           cfprojx(cbase,cfgraph,cfg2) ;
           cfprojx(cbase,cfgraph,cfg1) ;
  }


      /*---(4)--- gluon edge reduction */

  else if( e0>0 )
  {
	v1 = cfg->edge[0][e0] ;
	v2 = cfg->edge[1][e0] ;
	e1 = cfg->rvtx[0][v1] ;
	e3 = cfg->rvtx[1][v1] ;
	e2 = cfg->rvtx[0][v2] ;
	e4 = cfg->rvtx[1][v2] ;

	v3 = cfg->edge[0][e1] ;
	v5 = cfg->edge[1][e3] ;
	v4 = cfg->edge[0][e2] ;
	v6 = cfg->edge[1][e4] ;

	/* tadpole */
	if( e1==e3 || e2==e4 )
	{
	  cffrclr( &cfg->wgtx );
	  return ;
	}

	/* closed loop(s) created */
	if( e1==e4 && e2==e3 )
	{
	  cfg->lvtx[v1] = 0;
	  cfg->lvtx[v2] = 0;
          cfg->tedge[e0] = 0;  /*05.07.09 added */
	  cfg->tedge[e1] = 0;
	  cfg->tedge[e2] = 0;
	  cffrsml( &cfg->wgtx , 4 );
	  cfprojx(cbase,cfgraph,cfg);
	  return;
	}
	else
	{
	  if( e1==e4 ) 
	  {
	    cfg->lvtx[v1] = 0;
	    cfg->lvtx[v2] = 0;
	    cfg->rvtx[0][v5] = e2;
	    cfg->tedge[e0] = 0 ;
	    cfg->tedge[e1] = 0 ;
	    cfg->tedge[e3] = 0 ;
	    cfg->edge[1][e2] = v5 ;
            cffrmul( &cfg->wgtx , &cffr43 ) ;
	    cfprojx(cbase,cfgraph,cfg);
	    return;
	  }
	  if( e2==e3 ) 
	  {
	    cfg->lvtx[v1] = 0;
	    cfg->lvtx[v2] = 0;
	    cfg->rvtx[0][v6] = e1;
	    cfg->tedge[e0] = 0 ;
	    cfg->tedge[e2] = 0 ;
	    cfg->tedge[e4] = 0 ;
	    cfg->edge[1][e1] = v6 ;
            cffrmul( &cfg->wgtx , &cffr43 ) ;
	    cfprojx(cbase,cfgraph,cfg);
	    return;
	  }
	}

	   /* tadpole ? */
	nocfg1 = nocfg2 = 0;
	if( v1!=v3 && v2!=v3 && v3==v5 && cfg->lvtx[v3]==CFVQQG ) nocfg1=1 ; 
	if( v2!=v4 && v1!=v4 && v4==v6 && cfg->lvtx[v4]==CFVQQG ) nocfg1=1 ;
	if( v1!=v3 && v2!=v3 && v3==v6 && cfg->lvtx[v3]==CFVQQG ) nocfg2=1 ; 
	if( v2!=v4 && v1!=v4 && v4==v5 && cfg->lvtx[v4]==CFVQQG ) nocfg2=1 ; 

#ifdef CFKDBG
printf(" e0=%d, e1=%d, e2=%d, e3=%d, e4=%d\n",e0,e1,e2,e3,e4);
printf(" v1=%d, v2=%d, v3=%d, v4=%d, v5=%d, v6=%d\n",v1,v2,v3,v4,v5,v6);
printf(" nocfg1=%d, nocfg2=%d\n",nocfg1,nocfg2);
#endif
#ifdef MODTK
       cfg1 = cfxhalloc(cfg->nvtx , cfg->nedge ) ;
       cfg2 = cfxhalloc(cfg->nvtx , cfg->nedge ) ;
#else
       addrcfg++  ;
       cfg1 = &xcfg[addrcfg] ;
       cfxgalloc(cfg1 , cfg->nvtx , cfg->nedge ) ;
       addrcfg++  ;
       cfg2 = &xcfg[addrcfg] ;
       cfxgalloc(cfg2 , cfg->nvtx , cfg->nedge ) ;
#endif

           cfxgcopy(cfg,cfg1) ;
           cfxgcopy(cfg,cfg2) ;

           cfg1->lvtx[v1] = 0 ;
           cfg1->lvtx[v2] = 0 ;
	   cfg1->rvtx[0][v5] = e1 ; 
	   cfg1->rvtx[0][v6] = e2 ; 
           cfg2->lvtx[v1] = 0 ;
           cfg2->lvtx[v2] = 0 ;
	   cfg2->rvtx[0][v5] = e2 ; 
	   cfg2->rvtx[0][v6] = e1 ; 

           cfg1->tedge[e0] = 0 ;
           cfg1->edge[1][e1] = v5 ;
           cfg1->edge[1][e2] = v6 ;
           cfg1->tedge[e3] = 0 ;
           cfg1->tedge[e4] = 0 ;
           cfg2->tedge[e0] = 0 ;
           cfg2->edge[1][e1] = v6 ;
           cfg2->edge[1][e2] = v5 ;
           cfg2->tedge[e3] = 0 ;
           cfg2->tedge[e4] = 0 ;

           cffrmul( &cfg1->wgtx ,&cffr16 ) ;
           cffrmul( &cfg2->wgtx ,&cffr12 ) ;

#ifdef CFKDBG
printf("Here, gluon bridge processed ........\n");
    cffrprt( " *cfg1=", &cfg1->wgtx);
    cffrprt( " *cfg2=", &cfg2->wgtx); 
/*
printf(" ====> cfg1\n"); cfxgdump(cfg1);
printf(" ====> cfg2\n"); cfxgdump(cfg2);*/
#endif
           if(nocfg2==0) cfprojx(cbase,cfgraph,cfg2) ;
           if(nocfg1==0) cfprojx(cbase,cfgraph,cfg1) ;
  }

}

/*==================================================================*/
Private int cfprojw(cbase,cfg)
cftbase   *cbase;
cftxgraph *cfg;
/*--------------------------------------------------------------------
 * Find the color base ID
 *
 * cfg    : pointer to graph structure (working)
 */
{
    cftbsbs cfbs0; 
#ifdef MODTK
    int  i,k,n;
#else
    int  i,j,k,n;
#endif
    int   id;
    int  nq,ng,nlen,nstr;
#ifdef MODTK
    int  ns,np,npold,p,p1;
#else
    int  ns,np,npold,p,pp,p1,plast;
#endif
    int  e1,e2,v,v1,vlast;
    int  iend;
    int  sum,sum1;
    int  expvt[20]; /* external particle vertex */

	id = 0;
	nq = cbase->nquarkp;
	ng = cbase->ngluon;
	nlen = 5 ;
	nstr = nq*2+ng ;

	cfbase31( &cfbs0, nlen, nstr );
	cfbs0.snum=0 ;

	/* check cfg */
	sum =0;
	for(i=1; i<=cfg->nvtx; i++)
	  if( cfg->lvtx[i]== CFVQQG) sum++;
	if( sum != ng ) {
	printf("(cfprojw) ERROR=1, ng=%d, CFVQQG=%d, See following dump\n",
		ng,sum);
	cfxgdump(cfg);
	return 0;
	}
	p=0 ;
	for(i=1; i<=cfg->nvtx; i++)
	  if( cfg->lvtx[i]>0 && cfg->lvtx[i]!= CFVQQG && 
		cfg->lvtx[i]!=CFVNONQCD &&
		cfg->mvtx[i]>=0 ) { p= i; break ;}
	if( p>0 ) {
	printf("(cfprojw) ERROR=2, vertex-%d, See following dump\n",
		p);
	cfxgdump(cfg);
	return 0;
	}
	p=0 ;
	for(i=1; i<=cfg->nedge; i++)
	  if( cfg->tedge[i]>0 && cfg->tedge[i]!= CFQUARK &&
		cfg->tedge[i]!= CFGLUON ) { p= i; break ;}
	if( p>0 ) {
	printf("(cfprojw) ERROR=3, edge-%d, See following dump\n",
		p);
	cfxgdump(cfg);
	return 0;
	}
	sum = sum1=0;
	for(i=1; i<=cfg->nedge; i++)
	  { if( cfg->tedge[i]==CFQUARK) sum++;
	    if( cfg->tedge[i]==CFGLUON) sum1++;}
	if( sum != (nq+ng) || sum1 !=ng ) {
	printf("(cfprojw) ERROR=4, Q, G=%d, %d, See following dump\n",
		sum,sum1);
	cfxgdump(cfg);
	return 0;
	}
	
	for(n=1; n<=nstr; n++) expvt[n]=0; 

	for(i=1; i<=cfg->nvtx; i++)
	{
	  if( cfg->lvtx[i]>0 && cfg->mvtx[i] <0 )
	  {
	    k=-cfg->mvtx[i] ;
	    p=cbase->expt[k] ;
	    if( p<=0 || p>nstr) {
	    printf("(cfprojw) ERROR=5, Bad ext. vtx-%d, See following dump\n",
		i);
	    cfxgdump(cfg);
	    return 0;
	    }
	    expvt[p] = i;
	  }
	}
	for(n=1; n<=nstr; n++)
	  if( expvt[n] ==0 ) {
	  printf("(cfprojw) ERROR=6, missing ext-%d, See following dump\n",
		n);
	  cfxgdump(cfg);
	  return 0;
	  }
	
#ifdef CFKDBG
printf("cfprojw: enter ........nq=%d, ng=%d, nstr=%d\n",nq,ng,nstr);
printf(" ====> cfg\n"); cfxgdump(cfg);
#endif

	/* no external particles */
	if( nstr == 0 ) return 0;

	/* pack cfbs0 */
	ns = 0 ;
	np = 0 ;

	for(n=1; n<=nstr; n++)
	{
	    i = expvt[n] ;
	    cfg->mvtx[i] = -n;
	}

	if( nq>0 )
	{
	  for(n=1; n<=nq; n++)
	  {
	    p = 0;
	    for(i=1; i<=cfg->nvtx; i++) 
	    { if(cfg->lvtx[i]!=0 && cfg->mvtx[i]==-n ) 
		{ p = i ; break;   }
	    }
	    if( p>0 )
	    {
		cfbs0.sstr[np]=n;
		npold = np++;
		cfg->lvtx[p]=0;
		do
		{
		  e1=cfg->rvtx[1][p];
		  p1=cfg->edge[1][e1];

#ifdef CFKDBG
	printf("n, i, p, ns, np, e1, p1, %d %d %d %d %d %d %d\n",
	        n, i, p, ns, np, e1, p1); 
#endif
		  if( cfg->lvtx[p1] == CFVQQG )
		  {
		    e1 = cfg->rvtx[2][p1];
		    if( cfg->edge[0][e1] == p1)
		      v = cfg->edge[1][e1];
		    else
		      v = cfg->edge[0][e1];
		    cfbs0.sstr[np] = -cfg->mvtx[v] ;
		    cfg->lvtx[v]=0;
		    iend = 0;
		  }
		  else
		  {
		    cfbs0.sstr[np] = -cfg->mvtx[p1] ;
		    cfg->lvtx[p1]=0;
		    iend = 1;
		  } 
		  np++;
		  p = p1;
		} while( iend==0 ) ;
#ifdef CFKDBG
	printf("n, p, ns, np, npold, iend %d %d %d %d %d %d \n",
                n, p, ns, np, npold, iend); 
#endif
		cfbs0.snum++;
		cfbs0.slen[ns] = np-npold;
		ns++;
	    }

	  }
	}
	
	if( ng>0 )
	{
	  for(n=1; n<=ng; n++)
	  {
	    p = 0;
	    for(i=1; i<=cfg->nvtx; i++) 
	    { if(cfg->lvtx[i]!=0 && cfg->mvtx[i]==-(2*nq+n) ) 
		{ p = i ; break;   }
	    }
	    if( p>0 )
	    {
		if( cfg->rvtx[0][p] !=0 )
		  e1=cfg->rvtx[0][p];
		else
		  e1=cfg->rvtx[1][p];
		if( cfg->edge[0][e1] !=p )
		  v=cfg->edge[0][e1];
		else
		  v=cfg->edge[1][e1];
		npold = np;
		vlast = v ;
		do
		{
		  e1=cfg->rvtx[1][v];
		  v1=cfg->edge[1][e1];
		  e2=cfg->rvtx[2][v];
		  if( cfg->edge[0][e2] !=v )
		    p=cfg->edge[0][e2];
		  else
		    p=cfg->edge[1][e2];
		  cfbs0.sstr[np++]=-cfg->mvtx[p];
		  cfg->lvtx[p]=0;
		  cfg->lvtx[v]=0;
		  cfg->tedge[e1]=0;
		  cfg->tedge[e2]=0;
		  v = v1 ;

#ifdef CFKDBG
printf("n, i, p, ns, np, e1, v, e2, v1, %d %d %d %d %d %d %d %d %d\n",
        n, i, p, ns, np, e1, v, e2, v1); 
#endif
		} while( v!=vlast ) ;
		cfbs0.snum++;
		cfbs0.slen[ns] = np-npold;
		ns++;
	    }

	  }
	}

#ifdef CFKDBG
printf("cfprojw: process end ........\n");
printf(" ====> cfg\n"); cfxgdump(cfg);
printf(" ====> base\n"); cfbase21(&cfbs0,0);
#endif

	/* order simple bases */
	cfbase07( &cfbs0 );

	/* scan scbslst */
	id = cfbase15( &cfbs0 );

	return id;

}

/*==================================================================*/
Private  Void cfxgalloc(cfg,nv,ne)
cftxgraph *cfg   ;
int      nv,ne;
/*--------------------------------------------------------------------
 * Allocate memory for structure cftxgraph
 *
 * cfg    : pointer to graph structure
 * nv     : number of vertices
 * ne     : number of edges
 */
{
#ifdef DBGFREE
    printf("cfxgalloc: addrcfg = %d\n", addrcfg);
#endif
/*    cfg = ( struct cftxgraph * ) malloc( sizeof(struct  cftxgraph) ) ; */
/*    cfg = &xcfg ;*/
    if( cfg == NULL )
    {
        fprintf(Stderr, "*** cfxgalloc :0: memory allocation failed\n");
        exit(1);
    }
    cfg->nvtx  = nv ;
    cfg->nedge = ne ;

    cfg->mvtx = ( int * ) malloc( (nv+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->mvtx = %p->%p\n", cfg, cfg->mvtx);
#endif

    cfg->lvtx = ( int * ) malloc( (nv+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->lvtx = %p->%p\n", cfg, cfg->lvtx);
#endif
    if( cfg->mvtx == NULL || cfg->lvtx == NULL )
    {
        fprintf(Stderr, "*** cfxgalloc :1: memory allocation failed\n");
        exit(1);
    }
    cfg->rvtx[0] = ( int * ) malloc( (nv+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->rvtx[0] = %p->%p\n", cfg, cfg->rvtx[0]);
#endif
    cfg->rvtx[1] = ( int * ) malloc( (nv+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->rvtx[1] = %p->%p\n", cfg, cfg->rvtx[1]);
#endif
    cfg->rvtx[2] = ( int * ) malloc( (nv+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->rvtx[2] = %p->%p\n", cfg, cfg->rvtx[2]);
#endif

    cfg->tedge   = ( int * ) malloc( (ne+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->tedge = %p->%p\n", cfg, cfg->tedge);
#endif

    cfg->edge[0] = ( int * ) malloc( (ne+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->edge[0] = %p->%p\n", cfg, cfg->edge[0]);
#endif

    cfg->edge[1] = ( int * ) malloc( (ne+1)*sizeof(int) ) ;
#ifdef DBGFREE
    printf("cfxgalloc: malloc cfg->edge[1] = %p->%p\n", cfg, cfg->edge[1]);
#endif
    if( cfg->tedge == NULL    ||
        cfg->edge[0] == NULL  || cfg->edge[1] == NULL )
    {
        fprintf(Stderr, "*** cfxgalloc :2: memory allocation failed\n");
        exit(1);
    }

    cffrset(&cfg->wgtx, 1, 1) ;

#ifdef CFKDBG1
printf(" addrcfg=%d, nv,ne=%d,%d\n",addrcfg,nv,ne) ;
printf(" //cfg:%d  ->nvtx:%d  ->nedfe:%d\n",cfg,&cfg->nvtx,&cfg->nedge);
printf(" lvtx:%d -> %d ; mvtx:%d -> %d \n",
&cfg->lvtx,cfg->lvtx,&cfg->mvtx,cfg->mvtx);
printf(" ->tedge:%d  ->edge:%d , %d  ->wgtx:%d\n",
cfg->tedge,cfg->edge[0],cfg->edge[1],&cfg->wgtx);
#endif
}

/*==================================================================*/
Private Void cfxgfree(cfg)
cftxgraph *cfg   ;
/*--------------------------------------------------------------------
 * Dispose memory for structure cftxgraph
 *
 * cfg    : pointer to graph structure
 */
{
    int j;

    for (j=1 ; j>=0 ; j--) {
        if (cfg->edge[j] != NULL) {
#ifdef DBGFREE
            printf("cfxgfree:  free   cfg->edge[%d] = %p->%p\n",
                   j, cfg, cfg->edge[j]);
#endif
            free( cfg->edge[j] );
            cfg->edge[j] = NULL;
        }
    }

    if (cfg->tedge != NULL) {
#ifdef DBGFREE
        printf("cfxgfree:  free   cfg->tedge = %p->%p\n", cfg, cfg->tedge);
#endif
        free( cfg->tedge );
        cfg->tedge = NULL;
    }

    for (j=2 ; j>=0 ; j--) {
        if (cfg->rvtx[j] != NULL) {
#ifdef DBGFREE
            printf("cfxgfree:  free   cfg->rvtx[%d] = %p->%p\n",
                   j, cfg, cfg->rvtx[j]);
#endif
            free( cfg->rvtx[j] );
            cfg->rvtx[j] = NULL;
        }
    }

    if (cfg->lvtx != NULL) {
#ifdef DBGFREE
        printf("cfxgfree:  free   cfg->lvxt = %p->%p\n", cfg, cfg->lvtx);
#endif
        free(cfg->lvtx);
        cfg->lvtx = NULL;
    }

    if (cfg->mvtx != NULL) {
#ifdef DBGFREE
        printf("cfxgfree:  free   cfg->mvtx = %p->%p\n", cfg, cfg->mvtx);
#endif
        free( cfg->mvtx );
        cfg->mvtx = NULL;
    }

#ifdef DBGFREE
    printf("cfxgfree:  addrcfg = %d\n", addrcfg);
#endif
}

#ifdef MODTK
/*==================================================================*/
Private cftxgraph *cfxhalloc(nv,ne)
int      nv,ne;
/*--------------------------------------------------------------------
 */
{
    cftxgraph *cfg;

    if(addrcfg >= MAXCFG) {
        fprintf(Stderr, "*** cfxhalloc : addrcfg = %d >= %d\n",
                addrcfg, MAXCFG);
        exit(1);
    }
    cfg = xcfg + (addrcfg++);

    cfxgalloc(cfg, nv, ne);

    return cfg;
}

/*==================================================================*/
Private Void cfxhfree(NoArg)
/*--------------------------------------------------------------------
 * Dispose memory for structure cftxgraph
 *
 * cfg    : pointer to graph structure
 */
{
    while(addrcfg > 0) {
        --addrcfg;
        cfxgfree(xcfg + addrcfg);
    }
}
#endif

/*==================================================================*/
Private Void cfxgcopy(cfg,cfg1)
cftxgraph *cfg, *cfg1;
/*--------------------------------------------------------------------
 * copy structure cftxgraph
 *
 * cfg    : pointer to graph structure (source)
 * cfg1   : pointer to graph structure (target)
*/
{
   int i ;

   for (i=0 ; i<=cfg->nvtx ; i++ )
   {  cfg1->mvtx[i] = cfg->mvtx[i] ;
      cfg1->lvtx[i] = cfg->lvtx[i] ;  
      cfg1->rvtx[0][i] = cfg->rvtx[0][i] ;  
      cfg1->rvtx[1][i] = cfg->rvtx[1][i] ;  
      cfg1->rvtx[2][i] = cfg->rvtx[2][i] ;  }

   for (i=0 ; i<=cfg->nedge ; i++ )
   {  cfg1->tedge[i] = cfg->tedge[i] ;
      cfg1->edge[0][i] = cfg->edge[0][i] ;
      cfg1->edge[1][i] = cfg->edge[1][i] ;  }

   cffrcpy( &cfg1->wgtx, &cfg->wgtx) ;

}

/*==================================================================*/
Private int cfbinit(cbase, nq, ng, nb)
cftbase  *cbase;
int      nq,ng,nb;
/*--------------------------------------------------------------------
 * Allocate memory for color base
 *
 * cbase  : pointer to color base data structure
 * nq     : the number of quark pairs
 * ng     : the number of gluons
 * nb     : the number of color base
 */
{
    int i, j;

    cbase->nquarkp= nq;
    cbase->ngluon = ng;
    cbase->nbase  = nb;

    if (nb==0)
    {  cbase->nbase = 1 ;
       cbase->nexcolor = 0 ; }
    else
       cbase->nexcolor = 1 ;

    if ( cbase->nexcolor == 1 )
    {
        cbase->quark  = (int*) malloc(sizeof(int)*nq);
#ifdef DBGFREE
        printf("cfbinit:  malloc cbase->quark = %p->%p\n", cbase, cbase->quark);
#endif


        cbase->antiq  = (int*) malloc(sizeof(int)*nq);
#ifdef DBGFREE
        printf("cfbinit:  malloc cbase->antiq = %p->%p\n", cbase, cbase->antiq);
#endif

        cbase->gluon  = (int*) malloc(sizeof(int)*ng);
#ifdef DBGFREE
        printf("cfbinit:  malloc cbase->gluon = %p->%p\n", cbase, cbase->gluon);
#endif

#ifdef MODTK
        /* i in cbase->expt[i] may 2*nq+ng+ <= i <= cbase->nex */
        cbase->expt  = (int*) malloc(sizeof(int)*(cbase->nex+1));
#else
        cbase->expt  = (int*) malloc(sizeof(int)*(2*nq+ng+1));
#endif
#ifdef DBGFREE
        printf("cfbinit:  malloc cbase->expt = %p->%p\n", cbase, cbase->expt);
#endif
    }

    cbase->constr = (int*) malloc(sizeof(int)*cbase->nbase);
#ifdef DBGFREE
    printf("cfbinit:  malloc cbase->constr = %p->%p\n", cbase, cbase->constr);
#endif

    cbase->nconid = (int*) malloc(sizeof(int)*cbase->nbase);
#ifdef DBGFREE
    printf("cfbinit:  malloc cbase->nconid = %p->%p\n", cbase, cbase->nconid);
#endif

    cbase->conid  = (int**) malloc(sizeof(int*)*cbase->nbase);
#ifdef DBGFREE
    printf("cfbinit:  malloc cbase->conid = %p->%p\n", cbase, cbase->conid);
#endif

    cbase->cwgt   = (cftfrac**) malloc(sizeof(cftfrac**)*cbase->nbase);
#ifdef DBGFREE
    printf("cfbinit:  malloc cbase->cwgt = %p->%p\n", cbase, cbase->cwgt);
#endif

    if(cbase->constr == NULL ||
       cbase->nconid == NULL ||
       cbase->conid  == NULL ||
       cbase->cwgt   == NULL)
    {
        fprintf(Stderr, "*** cfbinit 1: memory allocation failed\n");
        exit(1);
    }

    for(i = 0; i < cbase->nbase; i++) {
        cbase->constr[i] = 0;
        cbase->nconid[i] = 0;
        cbase->conid[i]  = NULL;
        cbase->cwgt[i]   = (cftfrac*) malloc(sizeof(cftfrac)*cbase->nbase);
#ifdef DBGFREE
        printf("cfbinit:  malloc cbase->cwgt[%d] = %p->%p\n",
               i, cbase, cbase->cwgt[i]);
#endif
        if(cbase->cwgt[i] == NULL) {
            fprintf(Stderr, "*** cfbinit 2: memory allocation failed\n");
            exit(1);
        }
        for(j = 0; j < cbase->nbase; j++)
            cffrclr(&cbase->cwgt[i][j]);
    }
    return 1 ;
}

#ifdef MODTK
#else
/*==================================================================*/
Private int cfcinit(cbase, nc)
cftbase  *cbase;
int      nc;
/*--------------------------------------------------------------------
 * Allocate memory for color base
 *
 * cbase  : pointer to color base data structure
 * nc     : for which color base space of conid is allocated
 */
{
    int i;

    cbase->conid[nc] = (int*) malloc(sizeof(int)*cbase->nconid[nc]);
#ifdef DBGFREE
    printf("cfcinit:  malloc cbase->conid[%d] = %p->%p\n",
               nc, cbase, cbase->conid[nc]);
#endif
    if(cbase->conid[nc] == NULL) {
        fprintf(Stderr, "*** cfcinit : memory allocation failed\n");
        exit(1);
    }
    for(i = 0; i < cbase->nconid[nc]; i++)
        cbase->conid[nc][i] = 0;

#ifdef MODTK
    return 1;
#endif
}
#endif

/*********************************************************************
 * Dump data
 */
/*==================================================================*/
Public  Void cfdump(cbase, cfgraph)
cftbase  *cbase;
cftgraph *cfgraph;
/*--------------------------------------------------------------------
 * Dump data for debugging
 */
{
    cfbdump(cbase);
    cfgdump(cfgraph);
}

/*==================================================================*/
Public  Void cfbdump(cbase)
cftbase  *cbase;
/*--------------------------------------------------------------------
 * Dump data for debugging
 */
{
    int i, j;

    printf("*** dump of cbase ***\n");
    if(cbase == NULL) {
       printf("    cbase = NULL\n");
       return;
    }

    printf("    nex = %d\n", cbase->nex);
    if(cbase->cptype == NULL) {
        printf("    cptype = NULL\n");
    } else {
        for(i = 1; i <= cbase->nex; i++)
#ifdef MODTK
#ifdef MODKK
            printf("    cptype[%d] = %d (-:%s)\n", i, cbase->cptype[i],
                        cfparticle[cbase->cptype[i]]);
#else
            printf("    cptype[%d] = %d (%d:%s)\n", i, cbase->cptype[i],
                        cfnoind(-i), cfparticle[cbase->cptype[i]]);
#endif
#else
            printf("    cptype[%d] = %d (%s)\n", i, cbase->cptype[i],
                        cfparticle[cbase->cptype[i]]);

#endif
    }

    printf("\n    nbase  = %d\n", cbase->nbase);
    if(cbase->cwgt == NULL) {
        printf("    cwgt   = NULL\n");
    } else {
        for(i = 0; i < cbase->nbase; i++) {
            printf("    cwgt[%4d] = ", i);
            if(cbase->cwgt[i] == NULL) {
                printf("    NULL ");
            } else {
                for(j = 0; j < cbase->nbase; j++) {
                    cffrpr0(&cbase->cwgt[i][j]);
                    printf(",");
                    if(j % 5 == 4) {
                        printf("\n                 ");
                    }
                }
            }
            printf("\n");
        }
    }

    if(cbase->constr == NULL && cbase->nconid == NULL &&
       cbase->conid == NULL) {
        printf("    constr = NULL\n");
        printf("    nconid = NULL\n");
        printf("    conid  = NULL\n");
    } else {
        printf("\n       i:constr:nconid: conid ...\n");
        for(i = 0; i < cbase->nbase; i++) {
            printf("    %4d :", i);
            if(cbase->constr == NULL)
                printf("NULL :");
            else
                printf("%4d :", cbase->constr[i]);
            if(cbase->nconid == NULL)
                printf("NULL :");
            else
                printf("%4d :", cbase->nconid[i]);
            if(cbase->conid == NULL)
                printf("NULL :");
            else
                for(j = 0; j < cbase->nconid[i]; j++) {
#ifdef MODKK
                    printf("%4d(---)", cbase->conid[i][j]);
#else
                    printf("%4d(%4d)", cbase->conid[i][j],
                        cfnoind(-(cbase->conid[i][j])));
#endif
                    if(j % 5 == 4) {
                        printf("\n                      ");
                    }
                }
            printf("\n");
        }
    }
}

/*==================================================================*/
Public  Void cfgdump(cfgraph)
cftgraph *cfgraph;
/*--------------------------------------------------------------------
 * Dump data for debugging
*/
{
    int i;

    printf("*** dump of cfgraph ***\n");
    if(cfgraph == NULL) {
       printf("    cfgraph = NULL\n");
       return;
    }

    printf("    nvtx = %d,  nedge = %d\n",
           cfgraph->nvtx, cfgraph->nedge);

    if(cfgraph->nleg == NULL && cfgraph->lvtx == NULL) {
        printf("    nleg  = NULL, lvtx = NULL\n");
    } else {
#ifdef MODTK
       /*
        * Replaced by TK in order to display the original node number
        * defined in "out.grf".
        */
        printf("    i(org), nleg[i], lvtx[i]\n");
        for(i = 1; i <= cfgraph->nvtx; i++) {
#ifdef MODKK
#else
            printf(" %4d(%d)", i, cfnoind(i));
#endif
#else
        printf("    i, nleg[i], lvtx[i]\n");
        for(i = 1; i <= cfgraph->nvtx; i++) {
            printf(" %4d", i);
#endif
            if(cfgraph->nleg == NULL)
                printf(" NULL");
            else
                printf(" %8d", cfgraph->nleg[i]);
            if(cfgraph->lvtx == NULL)
                printf(" NULL");
            else
                printf(" %8d (%s)", cfgraph->lvtx[i],
                                   cfvertex[cfgraph->lvtx[i]]);
            printf("\n");
        }
    }
    printf("\n");

    if(cfgraph->tedge == NULL && cfgraph->edge == NULL) {
        printf("    tedge = NULL, edge = NULL\n");
    } else {
        printf("    i, edge[0][i], edge[1][i], tedge[i]\n");
        for(i = 1; i <= cfgraph->nedge; i++) {
                printf(" %4d", i);
#ifdef MODTK
            if(cfgraph->edge[0] == NULL)
                printf("     NULL");
            else
#ifdef MODKK
                printf(" %4d(---)", cfgraph->edge[0][i]);
#else
                printf(" %4d(%3d)", cfgraph->edge[0][i],
                                    cfnoind(cfgraph->edge[0][i]));
#endif
            if(cfgraph->edge[1] == NULL)
                printf("      NULL");
            else
#ifdef MODKK
                printf(" %4d(---)", cfgraph->edge[1][i]);
#else
                printf(" %4d(%3d)", cfgraph->edge[1][i],
                                    cfnoind(cfgraph->edge[1][i]));
#endif
#else
            if(cfgraph->edge[0] == NULL)
                printf("     NULL");
            else
                printf("  %8d", cfgraph->edge[0][i]);
            if(cfgraph->edge[1] == NULL)
                printf("      NULL");
            else
                printf("  %8d", cfgraph->edge[1][i]);
#endif
            if(cfgraph->tedge == NULL)
                printf(" NULL");
            else
                printf("  %8d (%s)", cfgraph->tedge[i],
                                    cfparticle[cfgraph->tedge[i]]);
            printf("\n");
        }
    }
}
/*==================================================================*/
Public  Void cfxgdump(cfg)
cftxgraph *cfg;
/*--------------------------------------------------------------------
 * Dump data for debugging
*/
{
    int i;

    printf("*** dump of cfg ***\n");
    if(cfg == NULL) {
       printf("    cfg = NULL\n");
       return;
    }

    printf("    nvtx = %d,  nedge = %d\n", cfg->nvtx, cfg->nedge);

    if(cfg->mvtx == NULL || cfg->lvtx == NULL || cfg->rvtx==NULL ) {
        printf("    either mvtx or lvtx or rvtx = NULL\n");
    } else {
        printf(" i,  mvtx,  lvtx,  rvtx[0,1,2]\n");
        for(i = 1; i <= cfg->nvtx; i++) 
	printf(" %d    %d    %d    %d    %d    %d\n", i,
         cfg->mvtx[i], cfg->lvtx[i], 
         cfg->rvtx[0][i],cfg->rvtx[1][i],cfg->rvtx[2][i]);
    }

    if(cfg->tedge == NULL && cfg->edge == NULL) {
        printf("    tedge = NULL, edge = NULL\n");
    } else {
        printf(" i, tedge,   edge[0,1]\n");
        for(i = 1; i <= cfg->nedge; i++) 
	printf(" %d    %d    %d    %d    \n", 
            i,cfg->tedge[i], cfg->edge[0][i], cfg->edge[1][i]);
    }
	cffrprt( "wgtx=", &cfg->wgtx );
}


/*==================================================================*/
/* #include "cfbase-x.c"*/
/* 03/10/25/ New cfbase package (color base) */
/* 04/02/21/ revised */

/*
typedef	struct{ int v ; int p ; } qstr;

typedef struct{ int snum ; int *slen ; int *sstr ;} cftbsbs ;

Private	int	cfbase01	ARG((int, int));
Private	int	cfbase02	ARG((qstr*, int*, int, int, int, int));
Private	int	cfbase04	ARG((int*, int, int*));
Private	int	cfbase05	ARG((int, int*, int, int, int, int*, int));
Private	int	cfbase07	ARG((cftbsbs*));
Private	int	cfbase09	ARG((cftbsbs*, cftbsbs*));
Private	Void	cfbase10	ARG((int, int));
Private	int	cfbase11	ARG((int, int));
Private	int	cfbase12	ARG((int, int));
Private	int	cfbase15	ARG((cftbsbs*));
Private	int	cfbase21	ARG((cftbsbs*, int));
Private	int	cfbase31	ARG((cftbsbs*, int, int));
Private	int	cfbase32	ARG((cftbsbs*));

*/

/* working area for glueballs */
#define	MAXGLB	400 
Local	cftbsbs	cfgblst[MAXGLB];

/*----------------------------------------------*/

/* create color base lists
   Input : External colored particles
           nq = number of quark pairs
           ng = number of gluons       */

Private	int	cfbase01(nq,ng)
int 	nq,ng;
{
#ifdef MODTK
int	n,k,j,j1,i;
#else
int	n,k,j,j1,j2,i;
#endif
int	n1;
int	nexcolor;
#ifdef MODTK
#else
int	z[10];
#endif
int	xq[6];  /* antiquarks */
int	xg[10];  /* gluons in strings */
int	xgg[10];  /* gluons in glueballs */
qstr	zz[10];
int	nz,m;
int	nn,nnn,nng;
int	nlen;

int	*nglb;
int	nbs1,nbs2,nb;
int	snum0,snum1;

	if (nq==0 && ng==0) {nexcolor=0; return 0;}
	if (nq<0 || ng<0) 
	{ printf("(cfbase01) *ERROR=1\n"); return -1;}
        if (nq*2+ng>10)
	{ printf("(cfbase01) *ERROR=2\n"); return -1;}
        if (nq==0 && ng<2)
	{ printf("(cfbase01) *ERROR=3\n"); return -1;}
        if (ng>6)
	{ printf("(cfbase01) *ERROR=4\n"); return -1;}

/* printf("cbase01: nq=%d, ng=%d \n",nq,ng);*/
cfnbase = 0;
nlen = 5;
nglb = (int*) malloc(sizeof(int));

/* ==== (1) no quarks, glueballs only */
if(nq==0)
{
	for(k=0;k<ng;k++) xgg[k]=k+1;
	*nglb = 0;
	cfbase04(xgg,ng,nglb);
	cfnbase = *nglb;
	for(n=0;n<cfnbase;n++)
	{
	   cfbase31( &cfbslst[n], nlen, ng);
	   cfbslst[n].snum = cfgblst[n].snum ;
	   for(i=0;i<cfbslst[n].snum;i++) 
		cfbslst[n].slen[i]=cfgblst[n].slen[i];
	   for(i=0;i<ng;i++)
		cfbslst[n].sstr[i]=cfgblst[n].sstr[i];
	}
	for(j=(*nglb)-1; j>=0; j--) cfbase32( &cfgblst[j] );
return 1;
}

/* ==== (2) quarks and gluons  
          ( nq-pairs + n-gluon)  and ( n1 gluons )  */

/* nng=pow(2,ng)-1;*/
nng=1; if(ng>0) for(j=0; j<ng; j++,nng*=2); nng--; 

for(nn=0;nn<=nng;nn++)
{
  n=0;
  n1=0;
  /*nnn=nn+pow(2,ng+1);printf("nnn, nn, ng, nng %d %d %d %d\n",nnn,nn,ng,nng);*/
  nnn=nn+1024;/*printf("nnn, nn, ng %d %d %d \n",nnn,nn,ng);*/
  for( k=0 ; k<ng ; nnn>>=1,k++)
  {
  if ( nnn & 01 )
	xg[n++]=2*nq+1+k;
  else
	xgg[n1++]=2*nq+1+k;
  }
  
#ifdef CFKDBG
	 printf("n=%d, n1=%d, (%d)\n",n,n1,nn);
#endif
  if(n1!=1)
  {
    /* quark strings */
    nbs1 = cfnbase ;
    if(nq>=0)
    {
	for(j=1;j<=nq;j++) xq[j]=nq+j; 
	xq[0]=-1 ;
	do
	{
	  for(j=0;j<nq;j++)
	  {
		zz[j*2].v=j+1 ;
		zz[j*2].p=j*2+1;
		zz[j*2+1].v=xq[j+1];
		zz[j*2+1].p=-1;
	  }
	  nz = nq*2-1;
	  m = 0;
          cfbase02(zz,xg,nz,m,n,nq);

	} while ( cfkperm( nq, xq ) !=-1) ;

    }
    nbs2 = cfnbase ;

#ifdef CFKDBG
	printf(" cfnbase %d - %d, n1=%d \n",nbs1,nbs2,n1);
#endif
    /* glueballs */
    if(n1!= 0) 
    {	
	*nglb = 0;
	cfbase04(xgg,n1,nglb);
#ifdef CFKDBG
	printf(" >>> %d \n",*nglb);
#endif

	for(j=0; j<*nglb; j++)
	for(j1=nbs1; j1<nbs2; j1++)
	{
	   nb = j1 + j*(nbs2-nbs1) ;
	   if( j>0 )
	   {
		cfbase31( &cfbslst[nb], nlen, cfnsize);
		cfnbase++;

		cfbslst[nb].snum = cfbslst[j1].snum-cfgblst[0].snum;
		for(i=0;i<cfbslst[nb].snum;i++)
			cfbslst[nb].slen[i] = cfbslst[j1].slen[i];
		for(i=0;i<2*nq+n;i++)
			cfbslst[nb].sstr[i] = cfbslst[j1].sstr[i];
	   }
	   snum0 = cfbslst[nb].snum ;
	   snum1 = cfgblst[j].snum ;
	   for(i=0;i<snum1;i++) 
		cfbslst[nb].slen[snum0+i]=cfgblst[j].slen[i];
	   cfbslst[nb].snum += snum1 ;
	   for(i=0;i<n1;i++)
		cfbslst[nb].sstr[2*nq+n+i]=cfgblst[j].sstr[i];
	}
#ifdef CFKDBG
	printf(" Now, cfnbase=%d, nglb=%d (n1=%d) nlen=%d, cfnsize=%d\n",cfnbase,*nglb,n1,nlen,cfnsize);
#endif
	for(j=(*nglb)-1;j>=0; j--) cfbase32( &cfgblst[j] );
#ifdef CFKDBG
	printf(" Now, cfnbase=%d, nglb=%d (n1=%d)\n",cfnbase,*nglb,n1);
#endif
    }
  }
}
return 1;
}

/*----------------------------------------------*/
Private	int	cfbase02( zz, xg, nz, m, n, nq)
qstr*	zz;
int*	xg;
int	nz,m,n,nq;
{
int	k,k1;
int	j,j1;
int	oldp;
int	nlen;
nlen = 5;
	/* printf("ENTER %d %d %d \n",nz,m,n);*/

	if(m==n)
	/* all gluons are inserted */
	{
	/* for(k=0;k<=nz;k++) printf(" (%d->%d)",zz[k].v,zz[k].p);
	printf("\n"); */

	cfbase31( &cfbslst[cfnbase], nlen,cfnsize );
	cfbslst[cfnbase].snum = nq ;

	j1=0;
	j=-1;
	
	k=0;
        k1=0;
	do
	{
		/*  if(k==0) printf("(");
		printf("%d ",zz[k1].v); */
		cfbslst[cfnbase].sstr[k] = zz[k1].v ;
		if(zz[k1].p==-1)
			/* {printf(") ("); k1++;} */
			{ cfbslst[cfnbase].slen[j1]=k-j; j=k ; j1++; k1++;}
		else
			k1=zz[k1].p;
		k++;
	} while( k<=nz );
	/* printf("\n"); */
	cfnbase++;
	return 0;
	}
	else
	/* insert a gluon */
	{
	for(k=0;k<=nz;k++)
	if( zz[k].p != -1)
		{
		nz++;
		oldp=zz[k].p;
		zz[nz].v=xg[m];
		zz[nz].p=oldp;
		zz[k].p=nz;
		m++;
		cfbase02( zz, xg, nz, m, n ,nq);
                m--;
		zz[k].p=oldp;
                nz--;
		}
	}
return 0;
}



/*----------------------------------------------*/
/* 03/10/06 glueballs */
/* restricted version, glues LE 6 */


/* glueballs ,
   input : ng= number of gluons */

Private	int	cfbase04(gg,ng,nglb)
int	gg[];
int	ng;
int	*nglb;
{
int	n,k,j,i;
int	n1,n2;
int	p,x;
int	g1[10],g2[10];
#ifdef MODTK
#else
int	g0;
#endif
int	ng1,ng2,nx;
int	nlen;

if(ng<2) return -1;

nlen=2;
  /* p=pow(2,ng-1)-1;*/
  p=1; if(ng>1) for(j=1; j<ng; j++,p*=2);p--; 
  for(k=0; k<=p; k++)
  { x=k ;
    n1=0;
    n2=0;
    for(j=0; j<ng; j++ , x>>=1)
    {
	if( x & 01 )
	  g2[n2++]=gg[j];
	else
	  g1[n1++]=gg[j];
    }
	if(n1!=1 && n2!=1)  
	{
	ng1 = 1 ; for(n=1;n<n1;n++) ng1*=n;
	ng2 = 1 ; for(n=1;n<n2;n++) ng2*=n;
	nx = ng1*ng2;
		/*printf("** %d %d %d %d %d\n",n1,n2,ng1,ng2,nx);*/
	for(n=0; n<nx; n++)
	{
		cfbase31( &cfgblst[n+(*nglb)], nlen, n1+n2);
		cfgblst[n+(*nglb)].snum = 0 ;
	}
	if(n1>0) { /* printf(" n1=%d: ",n1); 
	   for(i=0;i<n1;i++) printf(" %d", g1[i]); printf("\n");*/
		cfbase05(n1,g1,ng1,ng2,n1,nglb,0);
		}
	if(n2>0) { /* printf("n2=%d: ",n2); 
	   for(i=0;i<n2;i++) printf(" %d", g2[i]); printf("\n");*/
		cfbase05(n2,g2,ng1,ng2,n1,nglb,1);
		}
	*nglb += nx ;
		/*printf("*= %d %d \n",*nglb,nx);*/
	}
  }
  /* special case for 6-gluons (gg)(gg)(gg) */
  if( ng==6 )
  {
	for(i=0;i<ng;i++) g1[i]=gg[i];
	do
	{
	  if((g1[2]<g1[4]) && (g1[2]<g1[3]) && (g1[4]<g1[5]) )
	  {
		cfgblst[(*nglb)].snum = 3 ;
		cfbase31( &cfgblst[(*nglb)], 3, 6);
		for(i=0;i<3;i++) cfgblst[(*nglb)].slen[i]=2;
		for(i=0;i<ng;i++) cfgblst[(*nglb)].sstr[i]=g1[i];
		(*nglb)++;
	  }
	} while (cfkperm(ng-1,g1) != -1);
  }
return 0;
}

/*----------------------------------------------*/
Private	int	cfbase05(ng,xg,ng1,ng2,n1,nglb,ist)
int	ng;
int	xg[];
int	ng1,ng2,n1,ist;
int	*nglb;
{
int	nn,n,k,j;
#ifdef MODTK
int	x;
#else
int	x0,x;
#endif
int	flag;
int	nb;

do
{
	flag=0;
	for(n=1;n<ng;n++)
	if( xg[n-1]>xg[n] )
		{	flag=-1;
			x=xg[n-1];
			xg[n-1]=xg[n];
			xg[n]=x;
		}
}	while(flag == -1);

nn = 0 ;
do
{
	/* printf(" %d",x0);
	for(n=1;n<ng;n++) printf(" %d",xg[n]);
	printf("\n"); */

	if( ist == 0 )
	for(n=0;n<ng2;n++)
	{  nb = (*nglb) + n*ng1+nn ;
	   cfgblst[nb].snum++;
	   cfgblst[nb].slen[0]=ng;
	   for(j=0;j<ng;j++) cfgblst[nb].sstr[j]=xg[j];
	}
	if( ist == 1 )
	for(k=0;k<ng1;k++)
	{  nb = (*nglb) + nn*ng1+k ;
	   cfgblst[nb].snum++;
	   cfgblst[nb].slen[1]=ng;
	   for(j=0;j<ng;j++) cfgblst[nb].sstr[n1+j]=xg[j];
	}
	nn++ ;

} while( cfkperm(ng-1,xg) != -1);

return 0;
}

/*----------------------------------------------*/

/* Normalize ordering of a base, bubble sort */

Private	int	cfbase07( cbase )
cftbsbs	*cbase;
{
int	j,k;
int	s,x0,x1;
int	b0,buf[10];
int	flag;

	if (cbase->snum <= 1) return 0;

do
  {
  flag = 0;
  x0=0;
  s=0;
  for(j=0; j<cbase->snum-1 ; j++)
	{
	s = s+cbase->slen[j];
	x1 = s;
	if( cbase->sstr[x0] > cbase->sstr[x1] )
	  {
		flag = -1;
		for(k=0; k<cbase->slen[j]; k++) 
			buf[k]=cbase->sstr[x0+k];
		for(k=0; k<cbase->slen[j+1]; k++) 
			cbase->sstr[x0+k]=cbase->sstr[x1+k];
		for(k=0; k<cbase->slen[j]; k++) 
			cbase->sstr[x0+cbase->slen[j+1]+k]=buf[k];
		b0 = cbase->slen[j];
		cbase->slen[j] = cbase->slen[j+1];
		cbase->slen[j+1] = b0;
		x1 = x0+cbase->slen[j+1];

	  }
	x0 = x1;
	}
  } while( flag != 0 );

return 0;
}

/*----------------------------------------------*/

/* Compare two bases.                          */

Private	int	cfbase09( cbase1, cbase2 )
cftbsbs	*cbase1, *cbase2;
        /* cbase1 > cbase2 , return = +1
           cbase1 < cbase2 , return = -1
           cbase1 = cbase2 , return =  0  */
{
int	x,j,s;

	x = (cbase1->snum - cbase2->snum);
	if (x>0) return 1;
	if (x<0) return -1;
	s=0;
	for( j=0; j<cbase1->snum; j++)
	{
	  x = (cbase1->slen[j] - cbase2->slen[j]);
	  if (x>0) return 1;
	  if (x<0) return -1;
	  s += cbase1->slen[j];
	}
	for( j=0; j<s; j++)
	{
	  x = (cbase1->sstr[j] - cbase2->sstr[j]);
	  if (x>0) return 1;
	  if (x<0) return -1;
	}
	return 0;

}
/*----------------------------------------------*/

/* sort cfbslst, quick sort */
Private  Void    cfbase10(l,r)
int     l,r;
{
int     v;
if(l>=r)        return;
        v=cfbase11(l,r);
        cfbase10(l,v-1);
        cfbase10(v+1,r);
return;
}
/*----------------------------------------------*/
Private  int     cfbase11(l,r)
int     l,r;
{
int     i,j;
        i=l-1;
        j=r;
        for(;;)
        {
                while( cfbase09( &cfbslst[++i], &cfbslst[r] ) ==-1 )
                ;
                while(i<--j && ( cfbase09( &cfbslst[r], &cfbslst[j] )==-1 ) )
                ;
                if(i>=j)        break;
		cfbase12( i, j);
        }
		cfbase12( i, r);
return i;
}
/*----------------------------------------------*/
Private  int     cfbase12(x,y)
int     x,y;
{
int	i,sx,sy;
int	buf;
int	xbuf[20];

	buf = cfbslst[x].snum; 
	cfbslst[x].snum = cfbslst[y].snum;
	cfbslst[y].snum = buf;
	sx = 0;  sy = 0;
	for(i=0; i<cfbslst[y].snum; i++) 
	  { xbuf[i]=cfbslst[x].slen[i]; sx +=cfbslst[x].slen[i]; }
	for(i=0; i<cfbslst[x].snum; i++) 
	  { cfbslst[x].slen[i]=cfbslst[y].slen[i]; sy +=cfbslst[y].slen[i]; }
	for(i=0; i<cfbslst[y].snum; i++) 
	  cfbslst[y].slen[i]=xbuf[i];
	for(i=0; i<sx; i++) 
	  xbuf[i]=cfbslst[x].sstr[i]; 
	for(i=0; i<sy; i++) 
	  cfbslst[x].sstr[i]=cfbslst[y].sstr[i];
	for(i=0; i<sx; i++) 
	  cfbslst[y].sstr[i]=xbuf[i];
return 0;
}

/*----------------------------------------------*/

/* Get base id-number, binary search                  */

Private	int	cfbase15( cbase )
cftbsbs	*cbase;
{
int	x0,x1,x;
int	y;

	x0 = 0 ;
	x1 = cfnbase-1 ;
	y = cfbase09( cbase , &cfbslst[x0] );
	if( y==0 ) return x0;
	if( y==-1 ) {printf("(cfbase15) ERROR=1: INPUT ");
		cfbase21( cbase ,0 ) ; return -1; }
	y = cfbase09( cbase , &cfbslst[x1] );
	if( y==0 ) return x1;
	if( y==1 ) {printf("(cfbase15) ERROR=2: INPUT ");
		cfbase21( cbase, 0 ) ; return -1; }

	do
	{
	  x =(x0+x1)/2 ;
	  y = cfbase09( cbase , &cfbslst[x] );
	  if( y==0 ) return x;
	  if( y==1 ) x0=x ;
	  if( y==-1 ) x1=x ;
	} while((x1-x0)>1);

	printf("(cfbase15) ERROR=3: INPUT ");
	cfbase21( cbase, 0 ) ; return -1; 

}

/*----------------------------------------------*/

/* Print a base of type ctfbsbs.                */

Private	int	cfbase21( cbase ,id )
cftbsbs	*cbase;
int	id;
{
int	j,s;
int	cut[10];
	printf("(%d) snum=%d: ",id ,cbase->snum);
	for(j=0;j<10;j++) cut[j]=0;
	s=0 ;
	for(j=0;j<cbase->snum;j++) 
	{ printf("%d ",cbase->slen[j]); 
	  s+= cbase->slen[j]; 
	  cut[s-1]=1; }
	printf(": / ");
	for(j=0;j<s;j++) 
	if( cut[j]==1 )
	  printf("%d/ ",cbase->sstr[j]);
	else
	  printf("%d ",cbase->sstr[j]);
	printf("\n");
	return 0;
}

/*----------------------------------------------*/

/* Allocate a base of type cftbsbs.             */

Private	int	cfbase31( cbase, nlen, nstr )
cftbsbs	*cbase;
int	nlen,nstr;
{
	cbase->slen = (int*)malloc(sizeof(int)*nlen) ;
	cbase->sstr = (int*)malloc(sizeof(int)*nstr) ;
	return 0;
}

/*----------------------------------------------*/

/* Free a base of type ctfbsbs.                 */

Private	int	cfbase32( cbase )
cftbsbs	*cbase;
{
	free( cbase->sstr );
	cbase->sstr = NULL;
	free( cbase->slen );
	cbase->slen = NULL;
	return 0;
}
/*----------------------------------------------*/

/* Create bubble diagram for (base-x) * conj.(base-y)  */

Private int cfbase00( cfgraph , x , y , nq , ng )
cftgraph *cfgraph;
int      x,y,nq,ng ;
{
#ifdef MODTK
int	j,n;
#else
int	j,k,n;
#endif
int	p0,p,p1;
int	v,v1,vlast,vv;
int	nv,ne;

#ifdef CFKDBG
printf(" 1) x,y,nq,ng = %d %d %d %d \n",x,y,nq,ng);
#endif
  /* external ends of a base */
  if(nq>0) for(j=0;j<nq;j++)
  {
    cfgraph->nleg[j+1]=2 ;
    cfgraph->lvtx[j+1]= CFVQQ ;
    cfgraph->nleg[j+1+nq]=2 ;
    cfgraph->lvtx[j+1+nq]= CFVQQ ;
  }
  if(ng>0) for(j=0;j<ng;j++)
  {
    cfgraph->nleg[j+1+nq*2]=2 ;
    cfgraph->lvtx[j+1+nq*2]= CFVGG ;
  }
  ne = 0;
  nv = 2*nq+ng;

#ifdef CFKDBG
printf(" 2) x,y,nq,ng,ne,nv = %d %d %d %d %d %d\n",x,y,nq,ng,ne,nv);
#endif
  /* Base-x */
  for(n=0;n<cfbslst[x].snum;n++)
  {
    if(n==0) { p0=0 ; p1=cfbslst[x].slen[0]-1 ;}
    else     { p0+= cfbslst[x].slen[n-1] ; p1+=cfbslst[x].slen[n]; }

    vv = cfbslst[x].sstr[p0] ;
    if( vv < 2*nq )
	for(p=p0; p<p1; p++)
	{
	  v = cfbslst[x].sstr[p] ;
	  v1 = cfbslst[x].sstr[p+1] ;
	  if(p==p0) vlast=v; else vlast=nv;
	  if(p+1<p1)  
	  {
		nv++;
		cfgraph->nleg[nv] = 3;
		cfgraph->lvtx[nv] = CFVQQG;
		ne++;
		cfgraph->tedge[ne] = CFQUARK;
		cfgraph->edge[0][ne] = vlast;
		cfgraph->edge[1][ne] = nv;
		ne++;
		cfgraph->tedge[ne] = CFGLUON;
		cfgraph->edge[0][ne] = v1;
		cfgraph->edge[1][ne] = nv;
	  }
	  else
	  {
		ne++;
		cfgraph->tedge[ne] = CFQUARK;
		cfgraph->edge[0][ne] = vlast;
		cfgraph->edge[1][ne] = v1;
	  }
	}
    else
	for(p=p0; p<=p1; p++)
	{
	  v = cfbslst[x].sstr[p] ;
		nv++;
		cfgraph->nleg[nv] = 3;
		cfgraph->lvtx[nv] = CFVQQG;
		ne++;
		cfgraph->tedge[ne] = CFGLUON;
		cfgraph->edge[0][ne] = v;
		cfgraph->edge[1][ne] = nv;
		ne++;
		cfgraph->tedge[ne] = CFQUARK;
		cfgraph->edge[0][ne] = nv;
		cfgraph->edge[1][ne] = nv+1;
	  if( p==p0 ) vlast = nv;
	  if( p==p1 ) cfgraph->edge[1][ne] = vlast;
	}

  }

#ifdef CFKDBG
printf(" 3) x,y,nq,ng,ne,nv = %d %d %d %d %d %d\n",x,y,nq,ng,ne,nv);
#endif
  /* Base-y, conjugated */
  for(n=0;n<cfbslst[y].snum;n++)
  {
    if(n==0) { p0=0 ; p1=cfbslst[y].slen[0]-1 ;}
    else     { p0+= cfbslst[y].slen[n-1] ; p1+=cfbslst[y].slen[n]; }

    vv = cfbslst[y].sstr[p0] ;
    if( vv < 2*nq )
	for(p=p1; p>=p0+1; p--)
	{
	  v = cfbslst[y].sstr[p] ;
	  v1 = cfbslst[y].sstr[p-1] ;
	  if(p==p1) vlast=v; else vlast=nv;
	  if(p-1>p0)  
	  {
		nv++;
		cfgraph->nleg[nv] = 3;
		cfgraph->lvtx[nv] = CFVQQG;
		ne++;
		cfgraph->tedge[ne] = CFQUARK;
		cfgraph->edge[0][ne] = vlast;
		cfgraph->edge[1][ne] = nv;
		ne++;
		cfgraph->tedge[ne] = CFGLUON;
		cfgraph->edge[0][ne] = v1;
		cfgraph->edge[1][ne] = nv;
	  }
	  else
	  {
		ne++;
		cfgraph->tedge[ne] = CFQUARK;
		cfgraph->edge[0][ne] = vlast;
		cfgraph->edge[1][ne] = v1;
	  }
	}
    else
	for(p=p1; p>=p0; p--)
	{
	  v = cfbslst[y].sstr[p] ;
		nv++;
		cfgraph->nleg[nv] = 3;
		cfgraph->lvtx[nv] = CFVQQG;
		ne++;
		cfgraph->tedge[ne] = CFGLUON;
		cfgraph->edge[0][ne] = v;
		cfgraph->edge[1][ne] = nv;
		ne++;
		cfgraph->tedge[ne] = CFQUARK;
		cfgraph->edge[0][ne] = nv;
		cfgraph->edge[1][ne] = nv+1;
	  if( p==p1 ) vlast = nv;
	  if( p==p0 ) cfgraph->edge[1][ne] = vlast;
	}

  }
#ifdef CFKDBG
printf(" 4) x,y,nq,ng,ne,nv = %d %d %d %d %d %d\n",x,y,nq,ng,ne,nv);
	printf(" nq,ng=%d  %d .............\n",nq,ng);
	cfbase21( &cfbslst[x], x);
	cfbase21( &cfbslst[y], y);
	cfgdump(cfgraph);
#endif

return 0;
}
/*==================================================================*/
/*#include "cfb-sub.c"*/
/*=========================================================*/
/*           c f k p e r m  (rev.2)                        */
/*---------------------------------------------------------*/
/* Find next one for p[1]..p[n] in dictionary order.       */
/* Return -1, if next ordering no more exists.             */
/* Watch n, as no check is done for the size of array.     */
/* p[0] is smaller than any of p[1]..p[n].                 */
/*---------------------------------------------------------*/
Private int cfkperm( n , p )
int n;
int p[];
{
	int i,j,x ;

	i=n-1 ;
	while (p[i]>=p[i+1]) i--;

	if (i>0)
	{	j=n;
		while (p[i]>=p[j]) j--;
		x=p[i] ; p[i]=p[j] ; p[j]=x ;
		i++;
		j=n;
		while (i<j)
		{	x=p[i] ; p[i]=p[j] ; p[j]=x ;
			i++;
			j--;
		}
		return 0 ;
	}
	else return -1;
}

/*----------------------------------------*/
/* First argument is replaced by calling  */
/* (cffradd)   frct <-- frct + frc        */
/* (cffrsub)   frct <-- frct - frc        */
/* (cffrmul)   frct <-- frct * frc        */
/* (cffrdiv)   frct <-- frct / frc        */
/* (cffrsml)   frct <-- frct * c          */
/*----------------------------------------*/

Public Void cffradd( frct , frc )
cftfrac *frct, *frc;
{
	frct->num = (frct->num * frc->den)
                  + (frct->den * frc->num) ;
	frct->den =  frct->den * frc->den  ;
	cffrrdc( frct );
}

Public Void cffrsub( frct , frc  )
cftfrac *frct, *frc;
{
	frct->num = (frct->num * frc->den)
                  - (frct->den * frc->num) ;
	frct->den =  frct->den * frc->den  ;
	cffrrdc( frct );
}

Public Void cffrmul( frct , frc  )
cftfrac *frct, *frc;
{
	frct->num = frct->num * frc->num ;
	frct->den = frct->den * frc->den ;
	cffrrdc( frct );
}

Public Void cffrdiv( frct , frc  )
cftfrac *frct, *frc;
{
	if (frc->num==0)
		{ printf(" (cffrdiv) Err. \n"); return ;}
	frct->num = frct->num * frc->den ;
	frct->den = frct->den * frc->num ;
	cffrrdc( frct );
}

Public Void cffrsml( frct , c )
cftfrac  *frct;
int c;
{
	frct->num = c * frct->num ;
	cffrrdc( frct );
}

Public Void cffrset( frc, num, den )
cftfrac *frc;
int num,den;
{
	frc->num = num ;
	frc->den = den ;
	cffrrdc( frc );
}

Public Void cffrclr( frc )
cftfrac *frc;
{
	frc->num = 0 ;
	frc->den = 1 ;
}

Public Void cffrcpy( frct , frc )
cftfrac *frct, *frc;
{
	frct->num =  frc->num ;
	frct->den =  frc->den ;
}

Public Void cffrrdc(frc)
cftfrac *frc;
{
	int x,y,z;

	x = frc->num ;
	y = frc->den ;

	if (y==0) { printf(" TROUBLE: ZERO DENOMINATOR \n");
		    return ;
		  }

	if (x==0) { frc->den =1 ;
		    return ;
		  }

	if ( x<0 )  x=-x ;
	if ( y<0 )  y=-y ;

	if ( x>y ) {   z=x;
		       x=y;
                       y=z;  }
	while( x>0 ){
		z = y % x ;
		y = x ;
		x = z ;
	}
	if (y==0) { printf(" (cffrrdc) Err. \n"); return ;}

	frc->num = frc->num/y ;
	frc->den = frc->den/y ;

        if( frc->den<0 )
		{ frc->num = -frc->num ;
		  frc->den = -frc->den ; }
	
}

Public Void cffrprt(name,frc)
cftfrac *frc;
char *name;
{
	printf(" %s  %4d/%4d \n",name,frc->num,frc->den);
}

Public Void cffrpr0(frc)
cftfrac *frc;
{
	printf(" %4d/%4d",frc->num,frc->den);
}
