/*********************************************************************
 *    File : cfnbase.h
 *
 *    Include file for color base.
 *
 *    Add prefix "cf" for names which can be referred from outside.
 */

#define CFEXTEND

/* type of particles:
 *     these values are used in "cftbase.cptype" and "cftgraph.tedge".
 */
#define CFNPUNDEF  0      /* undefined particle */
#define CFNOCOLOR  1      /* not colored particle : 1 */
#define CFQUARK    2      /* quark      : 3  */
#define CFAQUARK   3      /* anti-quark : 3* */
#define CFGLUON    4      /* gluon      : 8, particle == anti-particle */
#define CFGHOST    5      /* ghost      : 8, particle */
#define CFAGHOST   6      /* anti-ghost : 8, anti-particle */
#define CFPKIND    7      /* the number of kinds of particles */

/* names of above values */
#ifdef DEFGLOBAL
Global char *cfparticle[] = {
    "undef",
    "no_color",
    "quark",
    "anti-quark",
    "gluon",
    "ghost",
    "anti-ghost",
};
#else
Global char *cfparticle[];
#endif

/* type of vertices :
 *     these values are used in "cftgraph.lvtx".
 */
#define CFVDEL     0      /* deleted vertex */
#define CFVQQG     1      /* (3  3* 8) */
#define CFVGGG     2      /* (8  8  8) */
#define CFVHHG     3      /* (8p 8a 8) */
#define CFVQQ      4      /* (3  3*) + singlets */
#define CFVGG      5      /* (8  8 ) + singlets  */
#define CFVHH      6      /* (8p 8a) + singlets */
#define CFVNONQCD  7      /* singlets */

#ifdef  CFEXTEND
#define CFVGGGG    8      /* (8  8  8  8 ) */
#define CFVHHGG    9      /* (8  8  8p 8a) */
#define CFVHHHH   10      /* (8p 8a 8p 8a) */
#define CFVQQGG   11      /* (3  3* 8  8 ) */
#define CFVQQHH   12      /* (3  3* 8p 8a) */
#define CFVQQQQ   13      /* (3  3* 3  3*) */
#define CFVKIND   14      /* the number of kinds of particles */
#endif

#define CFSHIFT   100     /* shift value for gluon reduction */
                          /* gluon=x --> quark=(x+CFSHIFT)         */
                          /*         and antiquark=(x+2*CFSHIFT)   */

#define CFVGGGP    2   /* temporal def. */

/* names of above values */
#ifdef DEFGLOBAL
Global char *cfvertex[] = {
    "deleted",
    "qqg",
    "ggg",
    "hhg",
    "qq",
    "gg",
    "hh",
    "non QCD",
#ifdef  CFEXTEND
    "gggg",
    "hhgg",
    "hhhh",
    "qqgg",
    "qqhh",
    "qqqq",
#endif
};
#else
Global char *cfvertex[];
#endif

#define CBINITIAL  (-1)   /* == ELINITIAL */
#define CBFINAL      1    /* == ELFINAL   */

/* interface of "cfbase" */

/* type for fraction */
typedef struct{
	int num;
	int den;
} cftfrac;

typedef struct {
    /* input for cbase */
    int     nex;          /* the number of external particles
                           */
    int    *cptype;       /* cptype[0..nex] : type of the particles.
                           * cptype[i] : for i-th external particle.
                           * 0th component is used as dummy.
                           */
    int    *initial;      /* initial[0..nex] : flag for initial particles
                           *   = CBINITIAL for initial particle
                           *   = CBFINAL   for final particle
                           * 0th component is used as dummy.
                           */

    /* output from cbase and input for cfproj */
    int      nbase;       /* the number of color base >0
                           * Even there is no colored parton
                           * it is 1. (see 'nexcolor')
                           */
    int      nexcolor;    /* If there is no external colored
                           * particle, it is 0 and 1 otherwise.
                           */
    int      nquarkp;     /* the number of quark pairs
                           */
    int      ngluon;      /* the number of gluons
                           */
    int     *quark;       /* quark[0..nquarkp-1] : address of quark 
                           */
    int     *antiq;       /* antiq[0..nquarkp-1] : address of antiquark 
                           */
    int     *gluon;       /* gluon[0..ngluon-1] : address of gluon
                           */
    int     *expt;        /* expt[1..nex] : address of external particle
                           */
    int     *constr;      /* constr[0..nbase-1] : number of strings
                           */
    int     *nconid;      /* nconid[0..nbase-1] : length of conid[i]
                           */
    int    **conid;       /* conid[0..nbase-1][0..nconid[i]-1] : 
                           * ID number particles 
                           */
    cftfrac **cwgt;       /* cwgt[0..nbase-1][0..nbase-1] : weight for 
                           * the color base
                           */

    long      iniav;       /* factor of initial average 
                            */
} cftbase;

/* interface to cfproj */
typedef struct {
    /* input for cfproj */
    int  nvtx;           /* the number of vertices
                          */ 
    int *nleg;           /* nleg[0..nvtx] : number of legs 
                          * when nleg[i] = 3, the vertex is 3 point vertex.
                          * 0th component is a dummy.
                          */
    int *lvtx;           /* lvtx[0..nvtx] : type of the vertices
                          * 0th component is a dummy.
                          */
    int  nedge;          /* the number of edges
                          */
    int *tedge;          /* nedge[0..nedge] : the type of edges
                          * 0th component is a dummy.
                          */
    int *edge[2];        /* starting    edge[0][0..nedge] and 
                          * terminating edge[1][0..nedge] vertices.
                          * 0th component is a dummy.
                          * Value:
                          *     -n < 0 : external particle corresponds to 
                          *              cftbase.cptype[n]
                          *      n > 0 : vertex corresponds to
                          *              cftgraph.nleg[n], lvtx[n]
                          *      n = 0 : not possible.
                          */

    /* output from cfproj */
    cftfrac *wgt;          /* wgt[0..nbase-1] : weight factor of the graph
                           */
} cftgraph;

/* working structure for a graph in cfproj */
typedef struct {
    /* input for cfproj */
                         /* 0th component is a dummy for all. */
    int  nvtx;           /* the number of vertices
                          */ 
    int  nedge;          /* the number of edges
                          */
    int *mvtx;           /* mvtx[0..nvtx] : address of vertex
                          *  negative = internal vertex
                          *  positive = external particle number
                          * 0th component is a dummy.
                          */
    int *lvtx;           /* lvtx[0..nvtx] : type of the vertex
                          * 0th component is a dummy.
                          */
    int *rvtx[3];        /* rvtx[0,1,2][0..nvtx] : edges connected to the 
                          * vertex 
                          * 0th component is a dummy.
                          */
    int *tedge;          /* nedge[0..nedge] : type of edge
                          * 0th component is a dummy.
                          */
    int *edge[2];        /* starting edge[0][0..nedge] and 
                          * terminating edge[1][0..nedge] (=vertex)
                          * 0th component is a dummy.
                          */

    /* output from cfproj */
    cftfrac wgtx;          /*  weight factor of the graph
                           */
} cftxgraph;

/* TI */
Global cftbase  *gcbase;
Global cftgraph *gcolgr;
/********************* entry points ******************************/

/* Calculate color base */
Public int cfbase        ARG((cftbase*, int, int));
/*
 * status = cfbase(cbase, ibase, iprnt);
 *
 * cbase : pointer to cftbase : color base data structure
 * ibase : input flag for the selection of bases
 * iprnt : input flag for printing information
 *
 * returns status
 */

/* Calculate weight to color base */
Public int cfproj       ARG((cftbase *, cftgraph *, int, int));
/*
 * status = cfproj(cbase, cfgraph, ibase, iprnt));
 *
 * cbase  : pointer to cftbase : color base data structure
 * cfgraph : pointer to cftgraph : information about a graph and its weight
 * ibase  : input flag for the selection of bases
 * iprnt  : input flag for printing information
 */
 
/* Dump data for debugging */
Public Void cfbdump      ARG((cftbase*));
Public Void cfgdump      ARG((cftgraph*));
Public Void cfdump       ARG((cftbase *, cftgraph *));

/* functions for handling fractions */
Public  Void cffradd        ARG((cftfrac* ,cftfrac*));
Public  Void cffrsub        ARG((cftfrac* ,cftfrac*));
Public  Void cffrmul        ARG((cftfrac* ,cftfrac*));
Public  Void cffrdiv        ARG((cftfrac* ,cftfrac*));
Public  Void cffrsml        ARG((cftfrac* ,int)); 
Public  Void cffrset        ARG((cftfrac* ,int, int));
Public  Void cffrclr        ARG((cftfrac*));
Public  Void cffrcpy        ARG((cftfrac* ,cftfrac*));
Public  Void cffrrdc        ARG((cftfrac*));
Public  Void cffrprt        ARG((char* ,cftfrac*));
Public  Void cffrpr0        ARG((cftfrac*));

