/********************************************************************
 *  File : cmfin.c
 *
 *  Process *.fin file for extended format.
 *
 *                                                 by Kaneko Toshiaki
 *  1995/05/30  Created.
 *      1. Function "cmamparm" is moved from "cmgrf.c".
 *      2. Function "cmamparm" should be called after
 *         "inclm.h/inclc.h/incomm.h" is created.
 *      3. Function "cmfobj" should be called after "cmamparm".
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "grdata.h"
#include "config.h"
#include "putid.h"
#include "gensc.h"
#include "lib.h"

/*--------------------------------------------------------------------
 * Local types
 */
typedef struct {
    string cmd;
    string fname;
    int    lang;
    int    defm;
    long   start;
    long   end;
} tcmd;

/*--------------------------------------------------------------------
 * Private functions
 */
Private Void    cmprline    ARG((string,FILE*,bool));
Private int     cmfcmd      ARG((FILE*,string,string*));
Private Void    cmdexec     ARG((int,int,FILE*));
Private string  cmgetline   ARG((FILE*, long*, long*, int));
Private Void    cmdmpficmd  ARG((NoArg));
Private Void    cmfile      ARG((string, int));
Private Void    cmfids      ARG((FILE*,string,int));

/*--------------------------------------------------------------------
 * Local variable
 */
#define COMMENT '%'

Local tcmd ficmd[] = {
     {"file",    NULL, -1, -1L, -1L},   /* this entry should be the first ? */
     {"fortran", NULL, -1, -1L, -1L},
     {"reduce",  NULL, -1, -1L, -1L},
     {"form",    NULL, -1, -1L, -1L},
     {"inclm",   NULL, -1, -1L, -1L},
     {"inclc",   NULL, -1, -1L, -1L},
     {"inclg",   NULL, -1, -1L, -1L},
     {"inclk",   NULL, -1, -1L, -1L},
     {"objs",    NULL, -1, -1L, -1L},
     {"libs",    NULL, -1, -1L, -1L},
     {"libdir",  NULL, -1, -1L, -1L},
};

#define MAXNFILE     150

Local tcmd optfile[MAXNFILE];
Local int  noptfile = 0;

#define CMD_NULL      -1
#define CMD_FILE       0
#define CMD_FORTRAN    1
#define CMD_REDUCE     2
#define CMD_FORM       3
#define CMD_INCLM      4
#define CMD_INCLC      5
#define CMD_INCLG      6
#define CMD_INCLK      7
#define CMD_OBJS       8
#define CMD_LIBS       9
#define CMD_LIBDIR    10
#define FICMD     (sizeof(ficmd)/sizeof(tcmd))

Local   FILE *fin = NULL;

#define MAXLINE     256

Local   char  buff[MAXLINE+3];

Local   bool  finset = False;

/*==================================================================*/
Public  Void    cmamparm(sdir)
string  sdir;
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    int    n;

    /* check whether 'amparm' is included in optfile */
    for(n = 0; n < noptfile; n++) {
        printf("cmamparm ....%s   [%s]\n",sdir,optfile[n].cmd);
        cmopend(optfile[n].lang, sdir, optfile[n].cmd);
        cmdexec(CMD_FILE, n, redfp);
    }
}

/*==================================================================*/
Public  Void    cmfin(NoArg)
/*------------------------------------------------------------------*
 * Process *.fin file.
 */
{
    string fname;

    pvfenam(momstr2, fnmdl, ".fin", MAXPATHLEN);
    fname = (string) strdup(momstr2);
    cmfile(fname, 1);

    if(caller == GRCRED || caller == GRCREDA || caller == GRCREDH ) {
        pvfenam(momstr2, fnmdl, ".rin", MAXPATHLEN);
        fname = (string) strdup(momstr2);
        cmfile(fname, 2);
    }

    if(caller == GRCFORM ) {
        pvfenam(momstr2, fnmdl, ".vin", MAXPATHLEN);
        fname = (string) strdup(momstr2);
        cmfile(fname, 3);
    }

    cmdmpficmd();

    finset = True;
}

/*==================================================================*/
Private Void    cmfile(fname, defm)
string  fname;
int     defm;
/*------------------------------------------------------------------*
 * Process *.fin file.
 */
{
    int    cmd, cmdold = CMD_NULL, ccmd, optf = -1, cmmnt;
    string line, arg;
    long   curr0 = 0L, curr = 0L, last = -1L;
    bool   ok = True;

    if((fin = fopen(fname, "r")) == NULL) {
        fprintf(Stderr, "*** cmfile: open error \"%s\"\n", fname);
        exit(1);
    }

    cmmnt = COMMENT;
    line = cmgetline(fin, &last, &curr, cmmnt);
    while(line != NULL) {
        if(line[0] == '+' || line[0] == '-') {
            ccmd = line[0];
            if((cmd = cmfcmd(fin, line, &arg)) == CMD_NULL) {
                continue;
            }

        /* skip : get next line */
        } else if(strlen(line) < 1) {
            curr = curr0;
            line = cmgetline(fin, &last, &curr, cmmnt);
            continue;

        /* something : get next line */
        } else {
            fprintf(Stderr, "*** cmfile : skip line in \"%s\"", fname);
            fprintf(Stderr, "\t%s\n", line);
            curr = curr0;
            line = cmgetline(fin, &last, &curr, cmmnt);
            continue;
        }

        /* start block */
        if(ccmd == '+') {
            if(cmdold != CMD_NULL) {
                fprintf(Stderr, "*** cmfile:%s: '- %s' is expected.\n",
                                fname, ficmd[cmdold].cmd);
                fprintf(Stderr, "in %s\n", line);
                if(optf < 0) {
                    exit(FAIL);
                }
                if(cmd == CMD_FILE   || cmd == CMD_FORTRAN ||
                   cmd == CMD_REDUCE || cmd == CMD_FORM)
                {
                    optfile[optf].end = last;
                } else {
                    ficmd[cmdold].end = last;
                }

            }
            cmdold = cmd;
            if(ficmd[cmdold].start >= 0 && ficmd[cmdold].defm >= defm 
               && ficmd[cmdold].fname == fname)
            {
                fprintf(Stderr, "*** cmfile:%s: ", fname);
                fprintf(Stderr, "'+%s' is defined twice[2]\n",
                        ficmd[cmdold].cmd);
                ok = False;
            }

            /* '+file <file name>' */
            if(cmd == CMD_FILE   || cmd == CMD_FORTRAN ||
               cmd == CMD_REDUCE || cmd == CMD_FORM)
            {
                if(strlen(arg) < 1) {
                    fprintf(Stderr, "*** cmfile:%s: ", fname);
                    fprintf(Stderr, "File name is not define\n");
                    ok = False;
                }
                for(optf = 0; optf < noptfile; optf++) {
                    if(strcmp(arg, optfile[optf].cmd) == 0) {
                        if(strcmp(fname, optfile[optf].fname) == 0) {
                            fprintf(Stderr, "*** cmfile:%s: ", fname);
                            fprintf(Stderr, "File %s is redefined\n", arg);
                        }
                        break;
                    }
                }
                if(optf >= noptfile) {
                    if(noptfile >= MAXNFILE) {
                        fprintf(Stderr, "*** cmfile:%s: ", fname);
                        fprintf(Stderr, "too many file definitions\n");
                        exit(1);
                    }
                    noptfile++;
                }
                optfile[optf].cmd   = (string) strdup(arg);
                optfile[optf].start = curr;
                optfile[optf].defm  = defm;
                optfile[optf].fname = fname;
                if(cmd == CMD_FILE) {
                    optfile[optf].lang = RONONE;
                } else if(cmd == CMD_FORTRAN) {
                    optfile[optf].lang = ROFORT;
                } else if(cmd == CMD_REDUCE) {
                    optfile[optf].lang = RORED;
                } else if(cmd == CMD_FORM) {
                    optfile[optf].lang = ROFORM;
                } else {
                    fprintf(Stderr, "*** cmfile:%s: ", fname);
                    fprintf(Stderr, "illegal control\n");
                    exit(1);
                }
                cmmnt = -1;
            } else {
                ficmd[cmdold].start = curr;
                ficmd[cmdold].defm  = defm;
                ficmd[cmdold].fname = fname;
            }
            do {
                curr0 = curr;
                line = cmgetline(fin, &last, &curr, cmmnt);
            } while(line != NULL && line[0] != '+' && line[0] != '-');
            cmmnt = COMMENT;

        /* end block */
        } else {
            if(cmd != cmdold) {
                if(cmdold != CMD_NULL) {
                    fprintf(Stderr, "*** cmfile:%s: ", fname);
                    fprintf(Stderr, "'-%s' is expected\n",
                                    ficmd[cmdold].cmd);
                    fprintf(Stderr, "in %s\n", line);
                    if(cmd == CMD_FILE   || cmd == CMD_FORTRAN ||
                       cmd == CMD_REDUCE || cmd == CMD_FORM)
                    {
                        optfile[optf].end = last;
                    } else {
                        ficmd[cmdold].end = last;
                    }
                } else {
                    fprintf(Stderr, "*** cmfile:%s: ", fname);
                    fprintf(Stderr, "%s without '+'\n", line);
                }
                cmprline(line, redfp, False);
            } else {
                if(cmd == CMD_FILE   || cmd == CMD_FORTRAN ||
                   cmd == CMD_REDUCE || cmd == CMD_FORM)
                {
                    optfile[optf].end = last;
                } else {
                    ficmd[cmdold].end = last;
                }

                cmdold = CMD_NULL;
            }
            curr0 = curr;
            line = cmgetline(fin, &last, &curr, cmmnt);
        }
    }

    if(!ok) {
        fprintf(Stderr, "*** cmfin : stop by error in *.fin\n");
        exit(FAIL);
    }

    fclose(fin);
    fin = NULL;
}

/*==================================================================*/
Private string  cmgetline(finf, last, curr, comment)
FILE    *finf;
long    *last;
long    *curr;
int     comment;
/*------------------------------------------------------------------*
 * Read a line
 */
{
    static int c = ' ';
    static char line[MAXLINE];
    int i, toolong = False;

    *last = *curr;

    if(c == EOF) {
        *curr = -2;
        c = ' ';
        return NULL;
    }

    do {
        i = 0;
        toolong = False;
        while((c = getc(finf)) != EOF && c != '\n') {
            if(i >= MAXLINE - 1) {
                toolong = True;
            } else {
                line[i++] = c;
            }
        }
        line[i] = '\0';
    } while (c != EOF && comment >= 0 && line[0] == comment);

    *curr = ftell(finf);

    if(toolong) {
        fprintf(Stderr, "*** too long line %s...\n", line);
    }

    return line;
}

/*==================================================================*/
Private Void    cmprline(line, rdfp, tolow)
string  line;
FILE   *rdfp;
bool    tolow;
/*------------------------------------------------------------------*
 * Output line to rdfp with tolower when tolow==True.
 */
{
    while(*line != '\0') {
        if(tolow) {
            putc(tolower(*line++), rdfp);
        } else {
            putc(*line++, rdfp);
        }
    }
    putc('\n', rdfp);
}

/*==================================================================*/
Private int  cmfcmd(finf, line, arg)
FILE    *finf;
string  line;
string  *arg;
/*-------------------------------------------------------------------
 * pickup the name of the command from line.
 * the first character is either '+' or '-', which is not checked here.
 */
{
    int    i, cmd;
    static char cmdstr[MAXLINE];
    string s;
    long   curr, last;

    s = line + 1;
    while(isspace(*s)) {
        s++;
    }
    for(i = 0; i < MAXLINE - 1; i++) {
        if(*s == '\0' || isspace(*s)) {
            break;
        } else {
            cmdstr[i] = tolower(*s++);
        }
    }
    cmdstr[i] = '\0';

    for(cmd = 0; cmd < FICMD; cmd++) {
        if(strcmp(cmdstr, ficmd[cmd].cmd) == 0) {
            break;
        }
    }
    if(cmd >= FICMD) {
        fprintf(Stderr, "*** command '%s' is not defined\n",
                cmdstr);
        cmprline(line, redfp, False);
        line = cmgetline(finf, &last, &curr, COMMENT);
        return CMD_NULL;
    }
    if(*s == '\0') {
        *arg = s;
    } else {
        s++;
        while(isspace(*s)) {
            s++;
        }
        *arg = s;
        for( ; *s != '\0'; s++) {
            if(isspace(*s)) {
                break;
            }
        }
        *s = '\0';
    }

    return cmd;
}

/*==================================================================*/
Private Void    cmdexec(cmd, optf, rdfp)
int     cmd;
int     optf;
FILE    *rdfp;
/*------------------------------------------------------------------*
 */
{
    string line;
    string fname;
    long   last, curr;
    long   fstart, fend;
    int    cmmnt = COMMENT;

    if(!finset) {
        cmfin();
    }

    switch(cmd) {
        case CMD_FILE:
        case CMD_FORTRAN:
        case CMD_REDUCE:
        case CMD_FORM:
            fstart = optfile[optf].start;
            fend   = optfile[optf].end;
            fname  = optfile[optf].fname;
            cmmnt = -1;
            break;

        case CMD_INCLM:
        case CMD_INCLC:
        case CMD_INCLG:
        case CMD_INCLK:
            fstart = ficmd[cmd].start;
            fend   = ficmd[cmd].end;
            fname  = ficmd[cmd].fname;
            break;


        default:
            fprintf(Stderr, "*** cmdexec : illegal control\n");
            exit(FAIL);
    }
    if(fstart < 0) {
        return;
    }
    if((fin = fopen(fname, "r")) == NULL) {
        fprintf(Stderr, "*** cmdexec : cannot open \"%s\"\n", fname);
        exit(FAIL);
    }
    fseek(fin, fstart, SEEK_SET);
    line = cmgetline(fin, &last, &curr, cmmnt);
    while(line != NULL && *line != '+' && *line != '-' && curr <= fend) {
/*         cmprline(line, rdfp, True);*/
         cmprline(line, rdfp, False);
         line = cmgetline(fin, &last, &curr, cmmnt);
    }
    fclose(fin);
    fin = NULL;
}

/*==================================================================*/
Public  Void    cmfobj(rdfp)
FILE    *rdfp;
/*------------------------------------------------------------------*
 * Add objects to Makefile
 */
{
    cmfids(rdfp, "", CMD_OBJS);
}

/*==================================================================*/
Public  Void    cmflib(rdfp)
FILE    *rdfp;
/*------------------------------------------------------------------*
 * Add objects to Makefile
 */
{
    cmfids(rdfp, "-l", CMD_LIBS);
}

/*==================================================================*/
Public  Void    cmflibdir(rdfp)
FILE    *rdfp;
/*------------------------------------------------------------------*
 * Add objects to Makefile
 */
{
    cmfids(rdfp, "-L", CMD_LIBDIR);
}

/*==================================================================*/
Private Void    cmfids(rdfp, pref, cmdtyp)
FILE    *rdfp;
string  pref;
int     cmdtyp;
/*------------------------------------------------------------------*
 * Add objects to Makefile
 */
{
    string line, s, t;
    long   curr, last;

    if(!finset) {
        cmfin();
    }

    if(ficmd[cmdtyp].start < 0) {
        return;
    }
    if((fin = fopen(ficmd[cmdtyp].fname, "r")) == NULL) {
        fprintf(Stderr, "*** cmdexec : cannot open \"%s\"\n",
                ficmd[cmdtyp].fname);
        exit(FAIL);
    }
    fseek(fin, ficmd[cmdtyp].start, SEEK_SET);

    line = s = cmgetline(fin, &last, &curr, COMMENT);
    while(line != NULL && *line != '+' && *line != '-' &&
          curr <= ficmd[cmdtyp].end)
    {
         while(True) {
             while(isspace(*s)) {
                 s++;
             }
             if(*s == '\0') {
                 break;
             }
             t = s;
             while(!isspace(*s) && *s != '\0') {
                 s++;
             }
             if(*s == '\0') {
                 sprintf(buff, "%s%s", pref, t);
                 putid(buff);
                 break;
             } else {
                 *s = '\0';
                 sprintf(buff, "%s%s", pref, t);
                 putid(buff);
                 s++;
             }
         }

         line = s = cmgetline(fin, &last, &curr, COMMENT);
    }

    putend();
    fprintf(rdfp, "\n");

    fclose(fin);
    fin = NULL;
}

/*==================================================================*/
Public Void    cmincl(NoArg)
/*------------------------------------------------------------------*
 */
{
    int i, j, lp, p, q, df, sizecp, maxkf, dimn;
    int *dimsz, *dimix, lo;

    lo   = lang;
    lang = ROFORT;
    cmopen(ROFORT, "inclm.h");

    fprintf(redfp,"* masses and width of particles\n");
    /* common /smmass/... */
    puinit(redfp,"      real*8         ",
                  "     &               ",
                  ",", ",", "", 72, 16);
    for(p = 0; p < nptable; p++) {
        if(isdigit(ptable[p]->mass[0])) {
            continue;
        }
        for(q = 0; q < p; q++) {
            if(strcmp(ptable[p]->mass, ptable[q]->mass) == 0) {
                break;
            }
        }
        if(p == q) {
            cmmass(p, 0, momstr);
            putid(momstr);
        }
    }
    putend();
    puinit(redfp,"      common /smmass/",
                  "     &               ",
                  ",", ",", "", 72, 16);
    for(p = 0; p < nptable; p++) {
        if(isdigit(ptable[p]->mass[0])) {
            continue;
        }
        for(q = 0; q < p; q++) {
            if(strcmp(ptable[p]->mass, ptable[q]->mass) == 0) {
                break;
            }
        }
        if(p == q) {
            cmmass(p, 2, momstr);
            putid(momstr);
        }
    }
    putend();
    putc(EOL, redfp);

    /* common /smmas2/... : mass**2 */
    puinit(redfp,"      real*8         ",
                  "     &               ",
                  ",", ",", "", 72, 16);
    for(p = 0; p < nptable; p++) {
        if(isdigit(ptable[p]->mass[0])) {
            continue;
        }
        for(q = 0; q < p; q++) {
            if(strcmp(ptable[p]->mass, ptable[q]->mass) == 0) {
                break;
            }
        }
        if(p == q) {
            cmmass(p, -10, momstr);
            cmtoright(momstr);
            putid(momstr);
        }
    }
    putend();

    puinit(redfp,"      common /smmas2/",
                  "     &               ",
                  ",", ",", "", 72, 16);
    for(p = 0; p < nptable; p++) {
        if(isdigit(ptable[p]->mass[0])) {
            continue;
        }
        for(q = 0; q < p; q++) {
            if(strcmp(ptable[p]->mass, ptable[q]->mass) == 0) {
                break;
            }
        }
        if(p == q) {
            cmmass(p, -2, momstr);
            cmtoright(momstr);
            putid(momstr);
        }
    }
    putend();
    putc(EOL, redfp);

    /* common /smgmma/... */
    puinit(redfp,"      real*8         ",
                  "     &               ",
                  ",", ",", "", 72, 16);
    for(p = 0; p < nptable; p++) {
        if(isdigit(ptable[p]->width[0])) {
            continue;
        }
        for(q = 0; q < p; q++) {
            if(strcmp(ptable[p]->width, ptable[q]->width) == 0) {
                break;
            }
        }
        if(p == q) {
            cmwidth(p, 0, momstr);
            cmtoright(momstr);
            putid(momstr);
        }
    }
    putend();
    puinit(redfp,"      common /smgmma/",
                  "     &               ",
                  ",", ",", "", 72, 16);
    for(p = 0; p < nptable; p++) {
        if(isdigit(ptable[p]->width[0])) {
            continue;
        }
        for(q = 0; q < p; q++) {
            if(strcmp(ptable[p]->width, ptable[q]->width) == 0) {
                break;
            }
        }
        if(p == q) {
            cmwidth(p, 2, momstr);
            cmtoright(momstr);
            putid(momstr);
        }
    }
    putend();
    putc(EOL, redfp);

    /* common /smcnst/ */
/* moved to .fin
    fprintf(redfp,"      common /smcnst/ pi, pi2, rad, gevpb, alpha, ");
    fprintf(redfp,"alphas, alpha0\n\n");
*/

    /* common /smkfms/ amprtl(kf-code) : value of mass parameter */
    maxkf = cmaxkf();
    fprintf(redfp,"* amprtl, agprtl : kfcode --> mass, width\n");
    fprintf(redfp,"      real*8          amprtl, agprtl\n");
    fprintf(redfp,"      common /smkfms/ amprtl(%d), agprtl(%d)\n",
            maxkf, maxkf);

    cmdexec(CMD_INCLM, 0, redfp);

    /*----------------------------------------------------------*/
    cmopen(ROFORT, "inclc.h");

    /*----------------------------------------------------------*/
    /* maxloop ---> 1   always give c1xxx                       */
    /*----------------------------------------------------------*/
    fprintf(redfp,"\n* phase of propagators\n");
    fprintf(redfp,"      common /smpphs/pphase,pphas1,pphas2,pphas3,");
    fprintf(redfp,"pphas4,pphas5,pphas6\n");
    fprintf(redfp,"      complex*16 pphase,pphas1,pphas2,pphas3,");
    fprintf(redfp,"pphas4,pphas5,pphas6\n");

    /* common /smcplc/... */
    for(i = 0; i < 2; i++) {
        if(i == 0) {
            puinit(redfp,"      common /smcplc/",
                         "     &               ",
                         ",", ",", "", 72, 16);
        } else {
            puinit(redfp,"      complex*16     ",
                         "     &               ",
                         ",", ",", "", 72, 16);
        }
        putid("     cqed");
        putid("     cqcd");
        for(lp = 0; lp <= 1; lp++) {
            for(p = nvstart; p < nvtable; p++) {
                if(vtable[p]->composit == V_AUX) {
                    continue;
                }
#ifdef OLD
                if(vtable[p]->loop > lp) {
                    continue;
                }
#else
                if(vtable[p]->loop != lp) {
                    continue;
                }
#endif
                for(j = 0; j < NFNAME; j++) {
                    if(vtable[p]->fname[j] == NULL) {
                        continue;
                    }
                    if(p != pvvnfinde(p, j)) {
                        continue;
                    }
                    Discard cmexcoupl(p, -i, lp, momstr, j);
                    sprintf(momstr3, "%9s", momstr);
                    cmtoright(momstr3);
                    putid(momstr3);
                }
            }
        }
        putend();
        putc(EOL, redfp);
    }

    /* TI : count size of amcplc */
    sizecp = 2;
    for(p = nvstart; p < nvtable; p++) {
        if(vtable[p]->composit == V_AUX) {
            continue;
        }
        if(vtable[p]->loop > 1) {
            continue;
        }
        if(p != pvvnfind(p)) {
            continue;
        }
        dimn = cmdimn(p, 0, &dimsz, &dimix);
        if(dimn < 1) {
            sizecp++;
        } else {
            for(i = 0, df = 1; i < dimn; i++) {
                df *= dimsz[i];
            }
            sizecp+=df;
        }
    }
    printf("---- size of smcplc   %d \n",sizecp);

    cmdexec(CMD_INCLC, 0, redfp);

    /*----------------------------------------------------------*/
    cmopen(ROFORT, "inclg.h");
    fprintf(redfp,"* Switch of gauge parameters\n");
    puinit(redfp,"      integer        ",
                  "     &               ",
                  ",", ",", "", 72, 16);
    putid("igaug00");
    for(i = 0; i < ngauge; i++) {
        sprintf(momstr, "igau%s", sgauge[i]);
        putid(momstr);
    }
    putend();
    puinit(redfp,"      common /smgaus/",
                  "     &               ",
                  ",", ",", "", 72, 16);
    putid("igaug00");
    for(i = 0; i < ngauge; i++) {
        sprintf(momstr, "igau%s", sgauge[i]);
        putid(momstr);
    }
    putend();

    fprintf(redfp,"      real*8         agauge\n");
    fprintf(redfp,"      common /smgaug/agauge(0:%d)\n", ngauge);

    cmdexec(CMD_INCLG, 0, redfp);

    lang = lo;
}

/*==================================================================*/
Public  Void    cmainclk(rdfp)
FILE    *rdfp;
/*------------------------------------------------------------------*
 */
{
    cmdexec(CMD_INCLK, 0, rdfp);
}

/*==================================================================*/
Public  Void    cmausrprm(rdfp)
FILE    *rdfp;
/*------------------------------------------------------------------*
 */
{
    printf("    +++ usrprm should be defined in .fin file\n");
}

/*==================================================================*/
Public  Void    cmtoright(str)
string  str;
/*------------------------------------------------------------------*
 *   move spaces to the left.
 */
{
    int  s, t, len;

    t = len = strlen(str);
    for(s = len-1; s >= 0; s--) {
        if(str[s] != ' ') {
            str[--t] = str[s];
        }
    }
    for(s = t-1; s >= 0; s--) {
        str[s] = ' ';
    }
}

/*==================================================================*/
Private Void cmdmpficmd(NoArg)
/*-------------------------------------------------------------------
 */
{
    int    cmd;
    string fname;

    printf("\n+++ Dump ficmd\n");
    for(cmd = 0; cmd < FICMD; cmd++) {
        if((fname = ficmd[cmd].fname) == NULL) {
            printf("%4d : %-16s %8ld %8ld Not defined\n",
                   cmd, ficmd[cmd].cmd,
                   ficmd[cmd].start, ficmd[cmd].end);
        } else {
            printf("%4d : %-16s %8ld %8ld \"%s\"\n",
                    cmd, ficmd[cmd].cmd,
                    ficmd[cmd].start, ficmd[cmd].end, fname);
        }
    }
    for(cmd = 0; cmd < noptfile; cmd++) {
        if((fname = optfile[cmd].fname) == NULL) {
            printf("%4d : %-16s %8ld %8ld No file\n",
                   cmd, optfile[cmd].cmd,
                   optfile[cmd].start, optfile[cmd].end);
        } else {
            printf("%4d : %-16s %8ld %8ld \"%s\"\n",
                   cmd, optfile[cmd].cmd,
                   optfile[cmd].start, optfile[cmd].end, fname);
        }
    }
}

/*==================================================================*/
Public  int     cmaxkf(NoArg)
/*------------------------------------------------------------------*
 */
{
    int maxkf = -1;
    int p, kf;

    for(p = 1; p < nptable; p++) {
        kf = ptable[p]->kfcode;
        if(kf > maxkf) {
            maxkf = kf;
        }
    }
    return maxkf;
}

/*==================================================================*/
Public  int     cmlang(fname)
string  fname;
/*------------------------------------------------------------------*
 */
{
    string suf = NULL;
    int    l, i, lng = -1;

    l = strlen(fname);
    suf = fname + l;

    for(i = l - 1; i >= 0; i--) {
        if(fname[i] == '.') {
            suf = fname + i + 1;
            l = l - i;
            break;
        }
    }

    if(suf == NULL) {
        if(strcmp(fname, "Makefile") == 0) {
            lng = ROMAKE;
        } else if(strcmp(fname, "makefile") == 0) {
            lng = ROMAKE;
        }
    } else if(l == 1) {
        if(strcmp(suf, "f") == 0) {
            lng = ROC;
        } else if(strcmp(suf, "c") == 0) {
            lng = ROC;
        } else if(strcmp(suf, "h") == 0) {
            lng = ROFORT;
        }
    } else if(l == 2) {
        if(strcmp(suf, "sh") == 0) {
            lng = ROSH;
        }
    } else if(l == 3) {
        if(strcmp(suf, "red") == 0) {
            lng = RORED;
        } else if(strcmp(suf, "frm") == 0) {
            lng = ROFORM;
        } else if(strcmp(suf, "csh") == 0) {
            lng = ROSH;
        }
    }
    return lng;
}
