/********************************************************************
 *  File : cmutl.c
 *
 *  Common part of source code generation.
 *                                                 by Kaneko Toshiaki
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <time.h>
#include <errno.h>
#include "grdata.h"
#include "config.h"
#include "gensc.h"
#include "version.h"

/*------------------------------------------------------------------*/
/* Private functions */
Private int     cmndf      ARG((int));
Private Void    cmiccmom   ARG((string,int,int,string,string,bool,bool));
Private bool    cmcoupl0   ARG((vindex,int,string,int,string,bool,bool,bool));
Private Void    cmzdnr     ARG((NoArg));

/*------------------------------------------------------------------*/
/* Local variable */
/* Local int    momoffs; */
Local char   tmpstr[MAXPATHLEN];

/*==================================================================*/
Public  Void    cmsepaf(fp, c)
FILE    *fp;
int     c;
/*------------------------------------------------------------------*
 */
{
    int l;

    putc('*', fp);
    for(l = 0; l < 71; l++) {
        putc(c, fp);
    }
    putc(EOL, fp);
}

/*==================================================================*/
Public  Void    cmsepar(c)
int     c;
/*------------------------------------------------------------------*
 */
{
    int l;

    putc('*', redfp);
    for(l = 0; l < 71; l++) {
        putc(c, redfp);
    }
    putc(EOL, redfp);
}

/*==================================================================*/
Public  Void    cmsepare(h,c)
int     h;
int     c;
/*------------------------------------------------------------------*
 *      Extended separator.
 */
{
    int l;

    putc(h, redfp);
    for(l = 0; l < 71; l++) {
        putc(c, redfp);
    }
    putc(EOL, redfp);
}

/*==================================================================*/
Public  string  cmtime(NoArg)
/*------------------------------------------------------------------*
 */
{
    time_t tloc;
    string sloc;

    tloc = time(&tloc);
    sloc = ctime(&tloc);

    return sloc;
}

/*==================================================================*/
Public  Void    cmopen(progt, fname)
int     progt;
string  fname;
/*------------------------------------------------------------------*
 * Open file
 */
{
    static bool ost=False;
    string cb, ce, ca = NULL, cc;
    int i, l;
    time_t tloc;
    string sloc;

#if 0  /* dummy output */
    fname = "/dev/null";
#endif
    switch(progt) {
        case ROFORM :
        case ROFORM1 :
            ca = "#-";
            cb = "*";
            cc = "*";
            ce = "*";
            break;
        case ROFORT :
            cb = "*";
            cc = "*";
            ce = "*";
            break;
        case ROC :
            cb = "/*";
            cc = " *";
            ce = " */";
            break;
        case ROMAKE :
            cb = "#";
            cc = "#";
            ce = "#";
            break;
        case ROSH :
            ca = "#!/bin/sh";
            cb = "#";
            cc = "#";
            ce = "#";
            break;
        case RONQS :
            ca = "##";
            cb = "##";
            cc = "##";
            ce = "##";
            break;
        case RORED :
            ca = "%";
            cb = "%";
            cc = "%";
            ce = "%";
            break;
        default :
            ca = "unknown language";
            cb = "< ?? ";
            cc = "  ?? ";
            ce = "  ?? >";
            break;
    }
    if(ost) {
        fclose(redfp);
        redfp = NULL;
    }

    if((redfp = fopen(fname, "w")) == NULL) {
        fprintf(Stderr, "cmopen : cannot open form file : %s\n",
                fname);
        exit(1);
    }

    /* without header */
    if(progt == RONONE) {
        return;
    }

    ost = True;

    /* the first line */
    if(ca != NULL) {
        fprintf(redfp, "%s\n", ca);
    }

    tloc = time(&tloc);
    sloc = ctime(&tloc);
    l    = strlen(sloc);
    if(sloc[l-1] == '\n') {
        sloc[l-1] = '\0';
    }
    fprintf(redfp, "%s", cb);
    fprintf(redfp, "     File %s : ", fname);
    fprintf(redfp, "%s\n", sloc);
    fprintf(redfp, "%s\n", cc);

    fprintf(redfp, "%s", cc);
    fprintf(redfp, "     Generated by GRACE (");
    if(caller == GRCFORT) {
        fprintf(redfp, "Fortran");
    } else if(caller == GRCFORM) {
        fprintf(redfp, "Form");
    } else if(caller == GRCRED ) {
        fprintf(redfp, "REDUCE");
    } else if(caller == GRCREDA) {
        fprintf(redfp, "REDUCE (for Amplitude)");
    } else if(caller == GRCREDH) {
        fprintf(redfp, "REDUCE (H.A.)");
    }
    fprintf(redfp, " source code generator)\n");

    fprintf(redfp, "%s", cc);
    fprintf(redfp, "         ");
    if(GRCVER2 >= 1000) {
        fprintf(redfp, "Ver.%d.%d",GRCVER0, GRCVER1);
    } else if(GRCVER3 >= 1000) {
        fprintf(redfp, "Ver.%d.%d.%d",GRCVER0, GRCVER1, GRCVER2);
    } else {
        fprintf(redfp, "Ver.%d.%d.%d(%d)",
                GRCVER0, GRCVER1, GRCVER2, GRCVER3);
    }
    fprintf(redfp, " %s\n", GRCDATE);
    fprintf(redfp, "%s\n", cc);

    fprintf(redfp, "%s", cc);
    fprintf(redfp, "     %s\n", CRIGHT);
    fprintf(redfp, "%s", cc);
    fprintf(redfp, "     F.Yuasa et al, Prog.Theor.Phys.Suppl.138(2000)18\n");

    fprintf(redfp, "%s", cc);
    for(i=1;i<72;i++) fprintf(redfp, "-"); fprintf(redfp, "\n");
    fprintf(redfp, "%s\n", ce);
}

/*==================================================================*/
Public  Void    cmclose(NoArg)
/*------------------------------------------------------------------*
 * Close file
 */
{
    if(redfp != NULL) {
        if(fclose(redfp) != 0) {
            perror("*** cmclose");
        }
        redfp = NULL;
    }
}

/*==================================================================*/
Public  string  cmpath(pathn, fname)
string  pathn;
string  fname;
/*------------------------------------------------------------------*
 * Open file
 */
{
    static char cmd[MAXPATHLEN];
    int  dlen;
    struct stat sb;
    struct stat *sbp = &sb;

    if(pathn == NULL || pathn[0] == '\0') {
        sprintf(cmd, "%s", fname);
        return cmd;
    }
    dlen = strlen(pathn);
    if(stat(pathn, sbp) == 0) {
        if(!S_ISDIR(sbp->st_mode)) {
            fprintf(Stderr, "*** file \"%s\" is not a directory\n", pathn);
            exit(FAIL);
        }
    } else {
        if(!isalnum(pathn[0])) {
            fprintf(Stderr, "*** strange path name '%s'\n", pathn);
            abort();
        }
        sprintf(cmd, "mkdir %s", pathn);
        if(system(cmd) != 0) {
            fprintf(Stderr, "*** cannot mkdir %s\n", pathn);
            exit(FAIL);
        }
    }
    sprintf(cmd, "%s/%s", pathn, fname);

    return cmd;
}

/*==================================================================*/
Public  Void    cmopend(progt, pathn, fname)
int     progt;
string  pathn;
string  fname;
/*------------------------------------------------------------------*
 * Open file
 */
{
    string cmd = cmpath(pathn, fname);
    cmopen(progt, cmd);
}

/*==================================================================*/
Public  Void    cmaopend(progt, pathn, fname, fp)
int     progt;
string  pathn;
string  fname;
FILE    **fp;
/*------------------------------------------------------------------*
 * Open file
 */
{
    string cmd = cmpath(pathn, fname);

    if((*fp = fopen(cmd, "a")) == NULL) {
        fprintf(Stderr, "*** cmaopend : file open \"%s\"\n", cmd);
        exit(FAIL);
    }
}

/*==================================================================*/
Public  int     cmdfcpl_del(v, nl, lp, ct)
int     v;
int     lp;
int     nl;
int     ct;
/*------------------------------------------------------------------*
 *  The number of array elements of coupling constant
 *
 *  Arguments
 *    v  : vertex in the vertex table.
 *    nl : the number of legs
 *    lp : loop
 *    ct : V_NOTYPE, V_V, V_A, V_VminusA or V_VplusA
 *------------------------------------------------------------------*/
{
    int sp = cmttype(v);

    /* old version */
    if(lp == 0) {
        if(sp == 1) {  /* fermion */
            if(ct == V_NOTYPE || lang == ROFORT) {
/*
            if(ct == V_NOTYPE) {
*/
                return 2;
            } else {
                return 0;
            }
        } else {
            return 0;
        }
    } else {
        if(sp == 1) {  /* fermion */
            if(nl == 2) {    /* ff */
                return 8;
            } else {
                return 2;
            }
        } else if(sp == 2) {  /* with ghost */
            if(nl == 2) {   /* self energy */
                return 4;
            } else {
                return 2;
            }
        } else if(sp == 3) {
            if(nl == 2) {   /* self energy */
                return 4;
            } else {
                return 2;
            }
        } else {
            if(nl == 2) {   /* self energy */
                return 2;
            } else {
                return 0;
            }
        }
    }
}


/*==================================================================*/
Public  int     cmtdfcpl_del(v, nl, lp, ct)
int     v;
int     lp;
int     nl;
int     ct;
/*------------------------------------------------------------------*
 *  The number of array elements of coupling constant
 *
 *  Arguments
 *    v  : vertex in the vertex table.
 *    nl : the number of legs
 *    lp : loop
 *    ct : V_NOTYPE, V_V, V_A, V_VminusA or V_VplusA
 *------------------------------------------------------------------*/
{
    int sp = cmttype(v);

    /* old version */
    if(lp == 0) {
        if(sp == 1) {  /* fermion */
            if(ct == V_NOTYPE) {
                return 2;
            } else {
                return 0;
            }
        } else {
            return 0;
        }
    } else {
        return cmdfcpl_del(v, nl, lp, ct);
    }
}

/*==================================================================*/
Public  int     cmfdimn(v, lp, dimsz, dimix, fn)
vindex  v;
int     lp;
int     **dimsz;
int     **dimix;
int     fn;
/*------------------------------------------------------------------*
 */
{
    int dimn;

    dimn   = vtable[v]->fndimn[fn];
    *dimsz = vtable[v]->fndim[fn];
    *dimix = vtable[v]->fnind[fn];

    return dimn;
}

/*==================================================================*/
Public  int     cmdimn(v, lp, dimsz, dimix)
vindex  v;
int     lp;
int     **dimsz;
int     **dimix;
/*------------------------------------------------------------------*
 */
{
    return cmfdimn(v, lp, dimsz, dimix, 0);
}

/*==================================================================*/
Public  int     cmcdimn(v, lp, dimsz, dimix)
vindex  v;
int     lp;
int     **dimsz;
int     **dimix;
/*------------------------------------------------------------------*
 */
{
    return cmfdimn(v, lp, dimsz, dimix, 1);
}

/*==================================================================*/
Public  int     cmtdimn(v, lp, dimsz, dimix)
vindex  v;
int     lp;
int     **dimsz;
int     **dimix;
/*------------------------------------------------------------------*
 *  Dimension of coupling with vtype
 */
{
    return cmfdimn(v, lp, dimsz, dimix, 0);
}

/*==================================================================*/
Public  int     cmctdimn(v, lp, dimsz, dimix)
vindex  v;
int     lp;
int     **dimsz;
int     **dimix;
/*------------------------------------------------------------------*
 *  Dimension of coupling with vtype
 */
{
    return cmfdimn(v, lp, dimsz, dimix, 1);
}

/*==================================================================*/
Private bool    cmcoupl0(v, ind, prefix, lp, str, mod, conj, ext)
vindex  v;
int     ind;
string  prefix;
int     lp;
string  str;
bool    mod;
bool    conj;
bool    ext;
/*------------------------------------------------------------------*
 *  Construct the name and arguments of a vertex.
 *
 *  Args:
 *    v    :  input : vertex in the vertex table.
 *    ind  :  input :
 *                    0  : coupling constant (simple name)
 *                   -1  : coupling constant with array size
 *                   -2  : phase of coupling constant (simple name)
 *                   -9  : coupling constant for fortran subroutine call
 *                   -10 : coupling constant with original indices
 *                   > 0 : coupling constant with indices.
 *                         The first index is the value of 'lp'.
 *    lp   :  input : loop of the vertex
 *    str  : output :
 *------------------------------------------------------------------*/
{
    int  ct, st, wind, indx, i, j, dimn, i0 = 0;
    char name[16];
    int  *dimsz, *dimix, fn;
    string s;
    string fname;
    pindex p0, p1;

/*
    printf(" --- cmcoupl0 ---- %d %d\n",ind,lp);
    if(lang == ROFORT) {
       printf(" ROFORT \n");
    } else if(lang == ROFORM) {
       printf(" ROFORM \n");
    } else if(lang == RORED) {
       printf(" RORED \n");
    } else  {
       printf(" others \n");
    }
*/

    st = True;
    ct = vtable[v]->ctype;

    if(ext && nextern > 2) {
        /* coupling of external self-energy */
        if(vtable[v]->nvlegs != 2) {
            fprintf(Stderr, "*** cmcoupl0: fname for %d point vertex for %s\n",
                        vtable[v]->nvlegs, vtable[v]->fname[0]);
            exit(FAIL);
        }
        p0 = abs(vtable[v]->vlegs[0]);
        p1 = abs(vtable[v]->vlegs[1]);

        /* ??? */
        if(!(ptable[p0]->extself || ptable[p1]->extself)) {
            fprintf(Stderr, "*** extself : ");
            fprintf(Stderr, "vertex for particle not external particle\n");
            pvprvertex(stderr, v);
            abort();
        }
        fn = 2;
    } else {
        fn = 0;
    }
    if(conj) {
        if(vtable[v]->fname[fn+1] == NULL) {
            if(fn == 0) {
                fprintf(Stderr, "*** cmcoupl0: cfname is not defined for %s\n",
                        vtable[v]->fname[0]);
                exit(FAIL);
            } else {
                fprintf(Stderr, "*** cmcoupl0: cefname is not defined for %s\n",
                        vtable[v]->fname[0]);
                exit(FAIL);
            }
        }
        fname = vtable[v]->fname[fn+1];
#if 0
        if(mod) {
            dimn  = cmctdimn(v, lp, &dimsz, &dimix);
        } else {
            dimn  = cmcdimn(v, lp, &dimsz, &dimix);
        }
#else
        dimn  = cmfdimn(v, lp, &dimsz, &dimix, fn+1);
#endif
    } else {
        if(vtable[v]->fname[fn] == NULL) {
            if(fn == 0) {
                fprintf(Stderr, "*** cmcoupl0: fname is not defined\n");
                exit(FAIL);
            } else {
                fprintf(Stderr, "*** cmcoupl0: efname is not defined for %s\n",
                        vtable[v]->fname[0]);
                exit(FAIL);
            }
        }
        fname = vtable[v]->fname[fn];
#if 0
        if(mod) {
            dimn  = cmtdimn(v, lp, &dimsz, &dimix);
        } else {
            dimn  = cmdimn(v, lp, &dimsz, &dimix);
        }
#else
        dimn  = cmfdimn(v, lp, &dimsz, &dimix, fn);
#endif
    }

    if(strlen(prefix) < 1) {
        if(mod) {
            sprintf(name, "t%s", fname);
        } else {
            sprintf(name, "%s", fname);
        }
    } else {
        if(mod) {
            sprintf(name, "t%s0%s", prefix, fname);
        } else {
            sprintf(name, "%s0%s", prefix, fname);
        }
    }

    /* as an argument of subroutine call */
    if(ind == -9) {
        wind = False;
        if(dimn > 0) {
            for(i = 0; i < dimn; i++) {
                if(dimix[i] > 0) {
                    wind = True;
                    break;
                }
            }
        }
        if(!wind) {
            sprintf(str, "%s", name);
        } else {
            s = str;
            if(lang == ROFORT) {
                sprintf(s, "%s(", name);
            } else if(lang == ROFORM) {
                printf("@ %s\n", name);
                sprintf(s, "%s(", name);
            } else if(lang == ROFORM1) {
                sprintf(s, "[%s(", name);
            } else {
                sprintf(s, "%s(", name);
            }
            s += strlen(s);
            for(i = 0; i < dimn; i++) {
                if(i != 0) {
                    sprintf(s, ",");
                    s++;
                }
                indx = max(1, dimix[i]);
                if(lang == ROFORT) {
                    sprintf(s, "%d", indx);
                } else if(lang == ROFORM) {
                    sprintf(s, "@@%d", indx);
                } else if(lang == ROFORM1) {
                    sprintf(s, "%d", indx);
                } else {
                    sprintf(s, "h%d", indx);
                }
                s += strlen(s);
            }
            if(lang == ROFORM) {
                printf("@\n");
                sprintf(s, "@");
            } else if(lang == ROFORM1) {
                sprintf(s, ")]");
            } else {
                sprintf(s, ")");
            }
        }

    /* as a simple name */
    } else if(ind == 0) {
        sprintf(str, "%s", name);

    /* ind = -2 : zxxx */
    } else if(ind == -2) {
        sprintf(str, "z%s", name+1);

    /* ind = -1 : cxxx(n) */
    } else if(ind == -1) {
        if(mod && lp < 1 && ct != V_NOTYPE) {
            i0 = 1;
        } else {
            i0 = 0;
        }
     
        if(i0 >= dimn) {
            sprintf(str, "%-6s   ", name);
        } else {
            s = str;
            sprintf(s, "%-6s(", name);
            s += strlen(s);
            for(i = i0, j = 0; i < dimn; i++, j++) {
                if(j != 0) {
                    sprintf(s, ",");
                    s++;
                }
                sprintf(s, "%d", dimsz[i]);
                s += strlen(s);
            }
            sprintf(s, ")");
        }

    /* with index */
    } else {
        s = str;
        if(mod) {
            indx = -1;
            if(lp > 0 || ct == V_NOTYPE) {
                indx = ind;
                i0 = 0;
     
            } else if(ct == V_V) {
                i0 = 1;
     
            } else if(ct == V_A) {
                i0 = 1;
                if(ind == 1) {
                    sprintf(s, "-");
                    s++;
                }
     
            } else if(ct == V_VminusA) {
                i0 = 1;
                if(ind != 1) {
                    sprintf(str, "0");
                    return False;
                }
            } else if(ct == V_VplusA) {
                i0 = 1;
                if(ind == 1) {
                    sprintf(str, "0");
                    return False;
                }
            }
        } else {
            i0 = 0;
        }
        if(lang == ROFORT) {
            sprintf(s, "%s", name);
        } else if(lang == ROFORM) {
            sprintf(s, "`%s", name);
        } else if(lang == ROFORM1) {
            sprintf(s, "[%s", name);
        } else {
            sprintf(s, "%s", name);
        }
        s += strlen(s);

        if(dimn < 1) {
            if(lang == ROFORM1) {
                sprintf(s, "]");
            }
            return True;
        } else {
            if(i0 < dimn) {
                if( lang != ROFORM ) {
                    sprintf(s, "(");
                    s++;
                }
            }
            for(i = i0, j = 0; i < dimn; i++) {
                    /* begin g980507 */
                if(ind != -10 && i == 0) {
                    /* end g980507 */
                    indx = ind;
                } else {
                    indx = max(1, dimix[i]);
                }
                if(j != 0) {
                    sprintf(s, ",");
                    s++;
                }
                j++;
                if(lang == ROFORT) {
                    sprintf(s, "%d", indx);
                } else if(lang == ROFORM) {
                    sprintf(s, "%d", indx);
                } else if(lang == ROFORM1) {
                    sprintf(s, "%d", indx);
                } else {
                    sprintf(s, "h%d", indx);
                }
                s += strlen(s);
/* 991117 TK 2000.05.18
                if( lang != ROFORM ) {
                    sprintf(s, ")");
                    s++;
                }
                if( lang == ROFORM ) {
                    sprintf(s, "'");
                    s++;
                }
*/
            }
            if(i0 < dimn) {
                if(lang == ROFORM) {
/*                    sprintf(s, ")");*/
                } else if(lang == ROFORM1) {
                    sprintf(s, ")]");
                } else {
                    sprintf(s, ")");
                }
            }
        }
    }

    return st;
}

/*==================================================================*/
Public  bool    cmcoupl(v, ind, lp, str)
vindex  v;
int     ind;
int     lp;
string  str;
/*------------------------------------------------------------------*
 *  Construct the name and arguments of a vertex.
 *
 *  Args:
 *    v    :  input : vertex in the vertex table.
 *    ind  :  input :
 *    lp   :  input : loop of the vertex
 *    str  : output :
 *------------------------------------------------------------------*/
{
    return cmcoupl0(v, ind, "", lp, str, False, False, False);
}

/*==================================================================*/
Public  bool    cmccoupl(v, ind, lp, str)
vindex  v;
int     ind;
int     lp;
string  str;
/*------------------------------------------------------------------*
 *  Construct the name and arguments of a vertex.
 *  Replace fname by cfname when possible.
 *
 *  Args:
 *    v    :  input : vertex in the vertex table.
 *    ind  :  input :
 *    lp   :  input : loop of the vertex
 *    str  : output :
 *------------------------------------------------------------------*/
{
    return cmcoupl0(v, ind, "", lp, str, False, True, False);
}

/*==================================================================*/
Public  bool    cmexcoupl(v, ind, lp, str, fn)
vindex  v;
int     ind;
int     lp;
string  str;
int     fn;
/*------------------------------------------------------------------*
 *  Construct the name and arguments of a vertex.
 *  Replace fname by cfname when possible.
 *
 *  Args:
 *    v    :  input : vertex in the vertex table.
 *    ind  :  input :
 *    lp   :  input : loop of the vertex
 *    str  : output :
 *    fn   :  input : same index as vtable->fname[fn]
 *------------------------------------------------------------------*/
{
    if(fn == 0) {
        return cmcoupl0(v, ind, "", lp, str, False, 0, 0);
    } else if(fn == 1) {
        return cmcoupl0(v, ind, "", lp, str, False, 1, 0);
    } else if(fn == 2) {
        return cmcoupl0(v, ind, "", lp, str, False, 0, 1);
    } else if(fn == 3) {
        return cmcoupl0(v, ind, "", lp, str, False, 1, 1);
    } else {
        fprintf(Stderr, "*** illegal fn = %d\n", fn);
        exit(FAIL);
    }
}

/*==================================================================*/
Public  bool    cmevcoupl(v, ind, lp, str)
vindex  v;
int     ind;
int     lp;
string  str;
/*------------------------------------------------------------------*
 *  Construct the name and arguments of a vertex.
 *
 *  Args:
 *    v    :  input : vertex in the vertex table.
 *    ind  :  input :
 *    lp   :  input : loop of the vertex
 *    str  : output :
 *------------------------------------------------------------------*/
{
    bool st;

    st = cmcoupl0(v, ind, "", lp, str, False, False, True);
    return st;
}

/*==================================================================*/
Public  bool    cmecoupl(n, ind, str)
node    n;
int     ind;
string  str;
/*------------------------------------------------------------------*
 *  Construct the name and arguments of a vertex.
 *
 *  Args:
 *    n    :  input : node
 *    ind  :  input :
 *    str  : output :
 *------------------------------------------------------------------*/
{
    vindex v;
    int    lp;
    bool   st;

    v  = ndvertex(n);
    lp = ndloop(n);
    if(ndlegs(n) == 2) {
#if 0
        int    t0, t1;
        t0 = ndtype(ndnextn(n,0));
        t1 = ndtype(ndnextn(n,1));
        if((t0 != NDTREEV && t0 != NDLOOPEDV) ||
           (t1 != NDTREEV && t1 != NDLOOPEDV)) {
            return cmevcoupl(v, ind, lp, str);
        }
#else
        st = cmevcoupl(v, ind, lp, str);
        return st;
#endif
    }
    return cmcoupl(v, ind, lp, str);
}

/*==================================================================*/
Public  bool    cmtcoupl(v, ind, lp, str)
vindex  v;
int     ind;
int     lp;
string  str;
/*------------------------------------------------------------------*
 *  Construct the name and arguments of a vertex.
 *
 *  Args:
 *    v    :  input : vertex in the vertex table.
 *    ind  :  input :
 *    lp   :  input : loop of the vertex
 *    str  : output :
 *------------------------------------------------------------------*/
{
    return cmcoupl0(v, ind, "", lp, str, True, False, False);
}

/*==================================================================*/
Public  bool    cmmass(p, ind, str)
pindex  p;
int     ind;
string  str;
/*------------------------------------------------------------------*
 *  Construct the name and arguments of a vertex.
 *
 *  Args:
 *    p    :  input : particle in the particle table.
 *    ind  :  input : 0: simple name, 1: with index, 2: with size
 *                    -10: simple name with '2'
 *    str  : output :
 *------------------------------------------------------------------*/
{
    int  st, cind, indx, i, dimn;
    string s;

    st = True;

    /* as a simple name */
    if(ind == 0) {
        sprintf(str, "%s", ptable[p]->mass);

    } else if(ind == -10) {
        sprintf(str, "%s2", ptable[p]->mass);

    /* mass with index */
    } else if(abs(ind) == 1) {
        dimn = ptable[p]->mdimn;
        cind = False;
        if(dimn > 0) {
            for(i = 0; i < dimn; i++) {
                if(ptable[p]->mind[i] > 0) {
                    cind = True;
                    break;
                }
            }
        }
        if(!cind) {
            if(ind > 0) {
                sprintf(str, "%s", ptable[p]->mass);
            } else {
                sprintf(str, "%s2", ptable[p]->mass);
            }
        } else {
            s = str;
            if(ind > 0) {
                sprintf(s, "%s(", ptable[p]->mass);
            } else {
                sprintf(s, "%s2(", ptable[p]->mass);
            }
            s += strlen(s);
            if(lang == RORED || lang == ROREDA) {
                *s = '!';
                s++;
            }
            for(i = 0; i < dimn; i++) {
                if(i != 0) {
                    sprintf(s, ",");
                    s++;
                    if(lang == RORED || lang == ROREDA) {
                        *s = '!';
                        s++;
                    }
                }
                indx = max(1, ptable[p]->mind[i]);
                sprintf(s, "%d", indx);
                s += strlen(s);
            }
            sprintf(s, ")");
        }

    /* ind = 1 : cxxx(n) */
    } else if(abs(ind) == 2) {
        dimn = ptable[p]->mdimn;
        if(dimn == 0) {
            if(ind > 0) {
                sprintf(str, "%-6s   ", ptable[p]->mass);
            } else {
                sprintf(str, "%-6s2  ", ptable[p]->mass);
            }
        } else {
            s = str;
            if(ind > 0) {
                 sprintf(s, "%-6s(", ptable[p]->mass);
            } else {
                 sprintf(s, "%-6s2(", ptable[p]->mass);
            }
            s += strlen(s);
            if(lang == RORED || lang == ROREDA) {
                *s = '!';
                s++;
            }
            for(i = 0; i < dimn; i++) {
                if(i != 0) {
                    sprintf(s, ",");
                    s++;
                    if(lang == RORED || lang == ROREDA) {
                        *s = '!';
                        s++;
                    }
                }
                sprintf(s, "%d", ptable[p]->mdim[i]);
                s += strlen(s);
            }
            sprintf(s, ")");
        }
    }

    return st;
}

/*==================================================================*/
Public  bool    cmwidth(p, ind, str)
pindex  p;
int     ind;
string  str;
/*------------------------------------------------------------------*
 *  Construct the name and arguments of a vertex.
 *
 *  Args:
 *    p    :  input : particle in the particle table.
 *    ind  :  input : 0: simple name, 1: with index, 2: with size
 *    str  : output :
 *------------------------------------------------------------------*/
{
    int  st, cind, indx, i, dimn;
    string s;

    st = True;

    /* as a simple name */
    if(ind == 0) {
        sprintf(str, "%s", ptable[p]->width);

    /* width with index */
    } else if(ind == 1) {
        dimn = ptable[p]->wdimn;
        cind = False;
        if(dimn > 0) {
            for(i = 0; i < dimn; i++) {
                if(ptable[p]->wind[i] > 0) {
                    cind = True;
                    break;
                }
            }
        }
        if(!cind) {
            sprintf(str, "%s", ptable[p]->width);
        } else {
            s = str;
            sprintf(s, "%s(", ptable[p]->width);
            s += strlen(s);
            for(i = 0; i < dimn; i++) {
                if(i != 0) {
                    sprintf(s, ",");
                    s++;
                }
                indx = max(1, ptable[p]->wind[i]);
                sprintf(s, "%d", indx);
                s += strlen(s);
            }
            sprintf(s, ")");
        }

    /* ind = 1 : cxxx(n) */
    } else if(ind == 2) {
        dimn = ptable[p]->wdimn;
        if(dimn == 0) {
            sprintf(str, "%-6s   ", ptable[p]->width);
        } else {
            s = str;
            sprintf(s, "%-6s(", ptable[p]->width);
            s += strlen(s);
            for(i = 0; i < dimn; i++) {
                if(i != 0) {
                    sprintf(s, ",");
                    s++;
                }
                sprintf(s, "%d", ptable[p]->wdim[i]);
                s += strlen(s);
            }
            sprintf(s, ")");
        }
    }

    return st;
}

/*=================================================================*/
Public  int     cmttype(n)
int     n;
/*------------------------------------------------------------------*
 * returns the type of vertex n of vertex table in the defined model.
 * return value
 *     -1 : particle with illegal spin
 *      0 : without fermion nor ghost
 *      1 : with fermion
 *      2 : with ghost
 *      3 : vector self-energy
 *------------------------------------------------------------------*/
{
    int  s, lg, ng, nf, p;

    if(vtable[n]->nvlegs == 2) { /* self-energy part */
        p = abs(vtable[n]->vlegs[0]);
        s = ptable[p]->ptype;
        if(s == P_VECTOR) {
            return 3;
        }
    }

    ng = nf = 0;
    for(lg = 0; lg < vtable[n]->nvlegs; lg++) {
        p = abs(vtable[n]->vlegs[lg]);
        s = ptable[p]->ptype;

        if(s == P_UNDEF) {
            return -1;
        } else if(s == P_DIRAC) {
            nf++;
        } else if(s == P_MAJORANA) {      /* ??? */
            nf++;
        } else if(s == P_GHOST) {
            ng++;
        }
    }

    if(ng > 0) {
        return 2;
    } else if(nf > 0) {
        return 1;
    } else {
        return 0;
    }
}

/*=================================================================*/
Public  Void    cmmom(str, n, leg, sgn)
string  str;
node    n;
int     leg;
bool    sgn;
/*------------------------------------------------------------------*
 * Construct momentum which comes to a node.
 *------------------------------------------------------------------*/
{
    int  il, s;

    if(ndtype(n) == NDONSHELL) {
        leg = 0;
    }

    il = ndiline(n, leg);

    if(sgn && iltable[il].ilnodes[0] == n) {
        s = -1;
    } else {
        s = 1;
    }

    cmimom(str, il, s);
}
/*=================================================================*/
Public  int     cminum(il)
int     il;
/*------------------------------------------------------------------*
 * Return il + momoffs;   (TI:2000/01/26)
 *------------------------------------------------------------------*/
{
        return il + momoffs;
}
/*=================================================================*/
Public  Void    cmimom(str, il, sgn)
string  str;
int     il;
int     sgn;
/*------------------------------------------------------------------*
 * consturct the name of an internal line : for looped graphs
 *------------------------------------------------------------------*/
{
    char c[3];
    int  itype;
    int  im, ii;

#if DBG
    if(dbg >= redbg) {
        printf("cmimom(%p, %d, %d)\n", str, il, sgn);
     }
#endif

    itype = iltable[il].iltype;

    switch (itype) {
        case ILEXTERN :
            if(lang == ROFORT || lang == ROREDA) {
                strcpy(c, "pe");
            } else {
                strcpy(c, "p");
            }
            im = il;
            for(ii = 1; ii < nilines; ii++) {
                if(iltable[il].ilmlist[ii] != 0) {
                    im = ii;
                    break;
                }
            }
            break;
        case ILLOOP :
            if(lang == ROFORT) {
                strcpy(c, "ke");
            } else {
                strcpy(c, "k");
            }
            im = il + momoffs;
            break;
        case ILDLOOP :
            if(lang == ROFORT) {
                strcpy(c, "le");
            } else {
                strcpy(c, "l");
            }
            im = il + momoffs;
            break;
        case ILECONJG :
            if(lang == ROFORT) {
                strcpy(c, "pc");
            } else {
                strcpy(c, "p");
            }
            im = il + momoffs;
            break;
        case ILTREE :
            if(lang == ROFORT) {
                strcpy(c, "pe");
            } else {
                strcpy(c, "q");
            }
            im = il + momoffs;
            break;
        default :
            if(lang == ROFORT) {
                strcpy(c, "??");
            } else {
                strcpy(c, "?");
            }
            im = il + momoffs;
            break;
    }

    if(sgn < 0) {
        sprintf(str, "-%s%d", c, im);
    } else {
        sprintf(str, "%s%d", c, im);
    }

    return;
}

/*=================================================================*/
Public  int     cmindx(str, typ, prt)
string  str;
int     typ;
pindex  prt;
/*------------------------------------------------------------------*
 * Construct momentum which comes to a node.
 *------------------------------------------------------------------*/
{
    int ptype;

    ptype = ptable[prt]->ptype;
    switch(ptype) {
        case P_SCALAR :
            strcpy(str, "lscalr");
            return 1;
        case P_MAJORANA :
            if(typ == ILEXTERN) {
                strcpy(str, "lextrn");
                return 2;
            } else {
                strcpy(str, "lintrn");
                return 4;
            }
        case P_DIRAC :
            if(typ == ILEXTERN) {
                strcpy(str, "lextrn");
                return 2;
            } else {
                strcpy(str, "lintrn");
                return 4;
            }
        case P_VECTOR :
            if(typ == ILEXTERN) {
                if(ptable[prt]->massless) {
                    sprintf(str, "lepexa");
                    return 2;
                } else {
                    sprintf(str, "lepexv");
                    return 3;
                }
            } else {
                if(ptable[prt]->massless) {
                    sprintf(str, "lepina");
                } else {
                    sprintf(str, "lepinv");
                }
                return 4;
            }
        default :
            strcpy(str, "l????");
            return 0;
    }
}


/*=================================================================*/
Public  Void    cmvcmom(str, n, leg, sgn, pref)
string  str;
node    n;
int     leg;
bool    sgn;
string  pref;
/*------------------------------------------------------------------*
 * Construct momentum which comes to a node.
 *------------------------------------------------------------------*/
{
    int  il, s;

    if(ndtype(n) == NDONSHELL) {
        leg = 0;
    }

    il = ndiline(n, leg);

    if(sgn && iltable[il].ilnodes[0] == n) {
        s = -1;
    } else {
        s = 1;
    }

    cmicmom(str, il, s, pref);
}

/*=================================================================*/
Public  Void    cmicmom(str, il, sgn, pref)
string  str;
int     il;
int     sgn;
string  pref;
/*------------------------------------------------------------------*
 * Construct momentum of iline : for tree graphs
 * Used for generated amplitudes.
 */
{
    cmiccmom(str, il, sgn, pref, "", ilenc, False);
}

/*=================================================================*/
Public  Void    cmicwgt(str, il, sgn, ptcl, pref)
string  str;
int     il;
int     sgn;
int     ptcl;
string  pref;
/*------------------------------------------------------------------*
 * Construct momentum of iline : for tree graphs
 * Used for generated amplitudes.
 * with suffix of the particle
 */
{
    int len, i;

    if(ptcl < 0) {
        ptcl = - ptcl;
    }

    len = 0;
    while(ptcl > 0) {
        i = ptcl%26;
        tmpstr[len++] = i + 'a';
        ptcl /= 26;
    }
    if(len == 0) {
        tmpstr[len++] = 'a';
    }
    tmpstr[len] = '\0';
    cmiccmom(str, il, sgn, pref, tmpstr, ilenc, False);
}

/*=================================================================*/
Public  Void    cmprfxini(NoArg)
/*------------------------------------------------------------------*
 */
{
    int pc, pd, cprc;
    bool ok = True;

    for(pc = 0; pc < nprocs; pc++) {
        cprc = reproct[pc];
        cprocess = dagetp(cprc);
        if(cprocess->pprefix == NULL || strlen(cprocess->pprefix) < 1) {
            if(nprocs <= 1) {
                reprocp[pc] = (string) strdup("");
            } else {
                sprintf(momstr, "%d", pc+1);
                reprocp[pc] = (string) strdup(momstr);
            }
        } else {
           printf("xxxxx %s\n",cprocess->pprefix);
            reprocp[pc] = (string) strdup(cprocess->pprefix);
            for(pd = 0; pd < pc - 1; pd++) {
                if(strcmp(reprocp[pc], reprocp[pd]) == 0) {
                    fprintf(Stderr, "*** duplicated prefix\n");
                    fprintf(Stderr, "%4d : \"%s\"\n", pc, reprocp[pc]);
                    fprintf(Stderr, "%4d : \"%s\"\n", pd, reprocp[pd]);
                    ok = False;
                }
            }
        }
    }
    if(!ok) {
        exit(FAIL);
    }
}

/*=================================================================*/
Public  Void    cmecmom(str, il, sgn, pref)
string  str;
int     il;
int     sgn;
string  pref;
/*------------------------------------------------------------------*
 * Construct momentum of iline : for tree graphs
 * Used for generating 'common /axet...' of not encoded ext. mom.
 */
{
    cmiccmom(str, il, sgn, pref, "", False, False);
}
/*=================================================================*/
Public  Void    cmenmom(str, il, sgn, pref)
string  str;
int     il;
int     sgn;
string  pref;
/*------------------------------------------------------------------*
 * Construct momentum of iline : for tree graphs
 * Used for generating 'common /axet...' of encoded ext. mom.
 */
{
    cmiccmom(str, il, sgn, pref, "", True, False);
}
/*=================================================================*/
Public  Void    cmccmom(str, il, sgn, pref)
string  str;
int     il;
int     sgn;
string  pref;
/*------------------------------------------------------------------*
 * Construct momentum of iline : for tree graphs
 * Used for the common block of encoded ilines
 */
{
    cmiccmom(str, il, sgn, pref, "", False, True);
}

/*=================================================================*/
Public  Void    cmpcmom(str, il, sgn, pref)
string  str;
int     il;
int     sgn;
string  pref;
/*------------------------------------------------------------------*
 */
{
    cmiccmom(str, il, sgn, pref, "", True, True);
}

/*=================================================================*/
Private Void    cmiccmom(str, il, sgn, pref, suffix, ilcd, mdpref)
string  str;
int     il;
int     sgn;
string  pref;
string  suffix;
bool    ilcd;
bool    mdpref;
/*------------------------------------------------------------------*
 * Construct momentum of iline : for tree graphs
 *
 * Generalized function of
 *     cmicmom(str, il, sgn, pref)
 *     cmccmom(str, il, sgn, pref)
 *
 * when il < 0, retuns the name without il-number.
 *
 * Flags :
 *     ilcd   : modify iline number to encoded one
 *     mdpref : modify prefix for internal lines
 */
{
    int itype, j;
    int c = pref[0], d = pref[1];

    if(ilcd) {
        mdpref = True;
    }

    if(il < 0) {
        sprintf(str, "%c%c%s%s", c, d, pref+2, cprefix);
        return;
    }

    if(mdpref) {
        d++;
    } else if(ilcd) {    /* includes when ilcd = True */
        if(il >= nilines) {
            fprintf(Stderr, "*** cmiccmom(str, %d, %d, %s, %s, %d, %d)\n",
                            il, sgn, pref, suffix, ilcd, mdpref);
            fprintf(Stderr, "    illegal i-line : out of the table\n");
            prdump();
            abort();
            exit(FAIL);
        }

        itype = iltable[il].iltype;

        /* only tree i-line */
        if(itype != ILEXTERN && itype != ILTREE) {
            fprintf(Stderr, "*** cmiccmom(str, %d, %d, %s, %s, %d, %d)\n",
                            il, sgn, pref, suffix, ilcd, mdpref);
            fprintf(Stderr, "    illegal i-line : itype = %d\n", itype);
            prdump();
            abort();
            exit(FAIL);
        }

        if(itype != ILEXTERN) {
            d++;
        }
    }

    if(ilcd) {
        j = iltable[il].ilmabs;
        if(sgn < 0) {
            if(j & 1) {
                j -= 1;
            } else {
                j += 1;
            }
            sgn = 1;
        }
    } else {
        j = il;
    }

    if(sgn > 0) {
        sprintf(str, "%c%c%s%s%d%s", c, d, pref+2, cprefix, j, suffix);
    } else {
        sprintf(str, "-%c%c%s%s%d%s", c, d, pref+2, cprefix, j, suffix);
    }
}

/*==================================================================*/
Public  bool    cmitype(il, fl1, fl2)
int     il;
int     *fl1, *fl2;
/*------------------------------------------------------------------*
 * Analyze type of an internal line.
 * It returns on a loop or not.
 *
 * Arguments:
 *    il  : internal line
 *    fl1, fl2 : output : fermion line number of nodes on both sides.
 *                        They are -1 for external particle,
 *                        and -2 for bosonic vertex.
 */
{
    int it;

    *fl1 = cmndf(iltable[il].ilnodes[0]);
    *fl2 = cmndf(iltable[il].ilnodes[1]);
    it = iltable[il].iltype;
    return (it == ILLOOP || it == ILDLOOP);
}

/*==================================================================*/
Private int     cmndf(nd)
int     nd;
/*------------------------------------------------------------------*
 * Returns the number of fermion line on which the node is located.
 * If the node is external particle, return value is -1 and
 * is -2 if the node is bosonic vertex.
 *
 * Arguments:
 *    nd : node.
 */
{
    int f, n;

    if(ndtype(nd) == NDEXTERN || ndtype(nd) == NDONSHELL) {
        return -1;
    }
    for(f = 0; f < nflines; f++) {
        for(n = 0; n < fltable[f].length; n++) {
            if(nd == fltable[f].ndlist[n]) {
                return f;
            }
        }
    }
    return -2;
}

/*=================================================================*/
Public  Uint    cmntype(n)
node    n;
/*------------------------------------------------------------------*
 * return value
 *     -1 : particle with illegal spin
 *      0 : without fermion nor ghost
 *      1 : with fermion
 *      2 : with ghost
 *------------------------------------------------------------------*/
{
    int  lg, s;
    Uint rt;

    rt = 0;
    for(lg = 0; lg < ndlegs(n); lg++) {
        s = ptable[abs(ndparticle(n, lg))]->ptype;
        if(s == P_UNDEF) {
            return 0;
        }
        rt |= pmask[s];
    }

    return rt;
}

/*==================================================================*/
Public  bool    cmphoton(n)
node    n;
/*------------------------------------------------------------------*
 * check external particle "n" is whether massless vector or not.
 */
{
    pindex p;

    p   = abs(ndparticle(n, 0));
    if(ptable[p]->ptype != P_VECTOR) {
        return False;
    }
    return ptable[p]->massless;
}
/*==================================================================*/
Public  Void    cmmkrnw(NoArg)
/*------------------------------------------------------------------*
 * generate entry "renew" to the Makefile.
 */
{
/* no good.
    fprintf(redfp,"\nrenew: clean\n");
    fprintf(redfp,"\trm -rf *.f *.red *.frm *.h *.c out2f ");
    fprintf(redfp,"*.mk *.sh *.dat form.set *.fout \n");
*/
}

/*==================================================================*/
Public  int     cmlpnum(NoArg)
/*------------------------------------------------------------------*
 * Count the maximum power of loop momentum of the numerator
 * The power of loop memontum of the denominator is 'ilxnumb'.
 *
 * This function works only for the case of one-loop.
 */
{
    int lnump;
    int y, nd, il, pt, lg, lgl, nlg, nfermi, nboson, nghost, iln = -1;
    int reptype[PTYPEMAX];

    lnump = 0;

    if(ilxnumb == 0) {
        return lnump;
    }
    /* 2002 Apr */
    lgl = -1;
    for(y = 0; y < ilxnumb; y++) {
        if(y == 0) {
            /* starting propagator */
            il = ilxtable[y];

            /* the next propagator */
            iln = ilxtable[(y + 1) % ilxnumb];

            /* find the node between il and iln */
            /* nd = iltable[iln].ilnodes[lgl] */

            /* try to go to direction 0 */
            nd = iltable[il].ilnodes[0];
            if(nd == iltable[iln].ilnodes[0]) {
                lgl = 0;
            } else if(nd == iltable[iln].ilnodes[1]) {
                lgl = 1;
            } else {
                lgl = -1;
            }

            /* try to go to direction 1 */
            if(lgl < 0) {
                nd = iltable[il].ilnodes[1];
                if(nd == iltable[iln].ilnodes[0]) {
                    lgl = 0;
                } else if(nd == iltable[iln].ilnodes[1]) {
                    lgl = 1;
                } else {
                    lgl = -1;
                }
            }
        } else {
            /* the current propagator */
            il = iln;

            /* the next propagator */
            iln = ilxtable[(y + 1) % ilxnumb];

            /* the next node : opposit side of lgl */
            nd  = iltable[il].ilnodes[1 - lgl];

            /* find lgl */
            if(nd == iltable[iln].ilnodes[0]) {
                lgl = 0;
            } else if(nd == iltable[iln].ilnodes[1]) {
                lgl = 1;
            } else {
                lgl = -1;
            }
        }
        if(lgl < 0) {
            fprintf(Stderr, "*** cmdiverge : illegal loop struct. ");
            fprintf(Stderr, "il=%d, iln=%d, nd=%d\n", il, iln, nd);
            exit(FAIL);
        }

        /* numerator of the propagator */
        pt = ptable[abs(ndparticle(iltable[il].ilnodes[0],
                                   iltable[il].illegs[0]))]->ptype;
        if(pt == P_DIRAC || pt == P_MAJORANA) {
            lnump++;
        }

        /* table of particle type connected to the vertex */
        nlg = ndlegs(nd);
        if(nlg != 3 && nlg != 4) {
            fprintf(Stderr, "*** cmdiverge : illegal vertex on a loop");
            fprintf(Stderr, "node = %d, legs = %d\n", nd, nlg);
            prdump();
            exit(FAIL);
        }

        /* classify particles by type */
        for(lg = 0; lg < PTYPEMAX; lg++) {
            reptype[lg] = 0;
        }
        for(lg = 0; lg < nlg; lg++) {
            reptype[ptable[abs(ndparticle(nd, lg))]->ptype]++;
        }
        if(reptype[P_UNDEF] != 0) {
            fprintf(Stderr, "*** cmdiverge : illegal vertex on a loop");
            fprintf(Stderr, "node = %d, ptype = P_UNDEF\n", nd);
            prdump();
            exit(FAIL);
        }

        nfermi = reptype[P_DIRAC] + reptype[P_MAJORANA];
        nboson = reptype[P_VECTOR ] + reptype[P_SCALAR  ];
        nghost = reptype[P_GHOST  ];

        if(reptype[P_AUX] != 0 || reptype[P_SET] != 0) {
            fprintf(Stderr, "*** cmlpnum : illgal particle : ");
            fprintf(Stderr, "aux = %d, set = %d\n",
                    reptype[P_AUX], reptype[P_AUX]);
            exit(0);
        }

        /* FFV, FFS */
        if(nfermi == 2 && nboson == 1 && nghost == 0) {

        /* VVV, VVS, VSS, SSS */
        } else if(nfermi == 0 && nboson == 3 && nghost == 0) {
            /* VVV */
            if(reptype[P_VECTOR] == 3) {
                lnump++;

            /* VSS */
            } else if(reptype[P_VECTOR] == 1) {
                lnump++;
            }

        /* VVVV, VVSS, VVVV */
        } else if(nfermi == 0 && nboson == 4 && nghost == 0) {

        /* GGV, GGS */
        } else if(nfermi == 0 && nboson == 1 && nghost == 2) {
            /* VVV */
            if(reptype[P_VECTOR] == 1) {
                lnump++;
            }

        /* VVGG *//* 980413 */
        } else if(nfermi == 0 && nboson == 2 && nghost == 2) {

        /* unknown */
        } else {
            fprintf(Stderr, "*** cmlpnum   : unknown vertex on a loop");
            fprintf(Stderr, "node = %d\n", nd);
            fprintf(Stderr, "DIRAC(%d),MAJOR(%d),VECTR(%d),SCALR(%d),GHOST(%d)\n",
                    reptype[P_DIRAC],reptype[P_MAJORANA],
                    reptype[P_VECTOR ],reptype[P_SCALAR  ],
                    reptype[P_GHOST  ]);
/*            prdump();*/
            exit(FAIL);
        }
    }
    return lnump;
}

/*==================================================================*/
Public  bool    cmdiverge(NoArg)
/*------------------------------------------------------------------*
 * returns
 *
 *    True  : divergent amplitude
 *    False : convergent amplitude
 */
{
    return (ilxnumb > 0) && ( (4 + cmlpnum() - 2*ilxnumb) >= 0 );
}

/*==================================================================*/
Public  Void    cmgauge(NoArg)
/*------------------------------------------------------------------*
 * Construct the table of gauge parameters.
 */
{
    int    p, i;
    Ulong  msk;

    ngauge = 0;
    msk    = 1L;
    for(p = 0; p < nptable; p++) {
        if(ptable[p]->gauge != NULL) {
            for(i = 0; i < ngauge; i++) {
                if(strcmp(ptable[p]->gauge, sgauge[i]) == 0) {
                    break;
                }
            }
            if(i == ngauge) {
                if(ngauge >= NGAUGE) {
                    fprintf(Stderr, "*** cmgauge : too many gauge param.\n");
                    exit(FAIL);
                }
                gmask[ngauge] = msk;
                msk <<= 1;
                sgauge[ngauge++] = ptable[p]->gauge;
            }
        }
    }
}

/*==================================================================*/
Public  Void    cmselect(NoArg)
/*------------------------------------------------------------------*
 * Construct the table of graph selection name
 */
{
    int    p, i;
    Ulong  msk;

    nselect = 0;
    msk    = 1L;
    for(p = 0; p < nptable; p++) {
        if(ptable[p]->pselect != NULL) {
            for(i = 0; i < nselect; i++) {
                if(strcmp(ptable[p]->pselect, sselect[i]) == 0) {
                    break;
                }
            }
            if(i == nselect) {
                if(nselect >= NSELECT) {
                    fprintf(Stderr, "*** cmselect : ");
                    fprintf(Stderr, "too many selection param.\n");
                    exit(FAIL);
                }
                sselmask[nselect] = msk;
                msk <<= 1;
                sselect[nselect++] = ptable[p]->pselect;
            }
        }
    }
    for(p = 0; p < nvtable; p++) {
        if(vtable[p]->vselect != NULL) {
            for(i = 0; i < nselect; i++) {
                if(strcmp(vtable[p]->vselect, sselect[i]) == 0) {
                    break;
                }
            }
            if(i == nselect) {
                if(nselect >= NSELECT) {
                    fprintf(Stderr, "*** cmselect : ");
                    fprintf(Stderr, "too many selection param.\n");
                    exit(FAIL);
                }
                sselmask[nselect] = msk;
                msk <<= 1;
                sselect[nselect++] = vtable[p]->vselect;
            }
        }
    }
}

/*==================================================================*/
Public  Void    cmachine(NoArg)
/*------------------------------------------------------------------*
 * Machine dependent part --> grclib/config.h (2.171)
 *------------------------------------------------------------------*/
{
    if(strcmp(machine, "vpp") == 0) {
        vpp     = True;
    }
/*
    if(strcmp(machine, "hpux") == 0) {
        fortcmp = "fort77";
        fflags  = "+ppu";
        cernlib = "/cern/pro/lib";
        reduce  = "reduce";

    } else if(strcmp(machine, "hiux") == 0) {
        fortcmp = "f90";
        fflags  = "-i,E,L,U,K,CL,CU,EU -W0,'LANGLVL(DARG(L))'";
        fflags  = "-W0,'langlvl(darg(l))'";
        cernlib = "/keklib/cern/lib";
        reduce  = "reduce";

    } else if(strcmp(machine, "sun") == 0) {
        fortcmp = "f77";
        fflags  = "-Nx800 -Nn1604";
        cernlib = "/cern/95a/lib";
        reduce  = "r34 -K16000";

    } else if(strcmp(machine, "dec") == 0) {
        fortcmp = "f77";
        fflags  = "";
        cernlib = "/cern/pro/lib";
        reduce  = "reduce";

    } else if(strcmp(machine, "sgi") == 0) {
        fortcmp = "f77";
        fflags  = "";
        cernlib = "/cern/pro/lib";
        reduce  = "reduce";

    } else if(strcmp(machine, "vpp") == 0) {
        fortcmp = "frtpx";
        fflags  = "-Ob -Wx -J -lm";
        cernlib = "/cern/pro/lib";
        vpp     = True;
        reduce  = "reduce";

    } else {
        fortcmp = "f77";
        fflags  = "-Nx800 -Nn1604";
        cernlib = "/cern/pro/lib";
        reduce  = "reduce";
    } 
*/
}

/*==================================================================*/
Public  Void    cmsnprpd(sdir)
string  sdir;
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    cmopend(ROFORT, sdir, "snprpd.f");
    fprintf(redfp,"*************************************************");
    fprintf(redfp,"***********************\n");
    if( vcode ) {
        fprintf(redfp,"      subroutine snprpd(pphase, nsampl, aprop, amomq, ");
    } else {
        fprintf(redfp,"      subroutine snprpd(pphase, aprop, amomq, ");
    }
    fprintf(redfp,"amassq, amag)\n");
    fprintf(redfp,"      implicit real*8(a-h,o-z)\n");
    if( vcode ) {
        fprintf(redfp,"      include 'inclvs.h'\n");
        fprintf(redfp,"      complex*16 pphase, aprop(n_vector)\n");
    } else {
        fprintf(redfp,"      complex*16 pphase, aprop\n");
    }
    if( vcode ) {
        fprintf(redfp,"      real*8     amomq(n_vector), amassq, amag\n");
    } else {
        fprintf(redfp,"      real*8     amomq, amassq, amag\n");
    }
    fprintf(redfp,"      common /chcntl/ jwidth, jtgamm\n");
    fprintf(redfp,"*\n");
    fprintf(redfp,"*    calculate denominator of propagator.\n");
    fprintf(redfp,"*\n");
    fprintf(redfp,"*     pphase : input  : phase of the ");
    fprintf(redfp,"propagator\n");
    fprintf(redfp,"*     aprop  : in/out : product of denominator ");
    fprintf(redfp,"of propagators.\n");
    fprintf(redfp,"*     amomp  : input  : square of mementum.\n");
    fprintf(redfp,"*     amassq : input  : square of mass.\n");
    fprintf(redfp,"*     amag   : input  : mass * width.\n");
    fprintf(redfp,"*------------------------------------------------");
    fprintf(redfp,"-----------------------\n");
    if( vcode ) {
        fprintf(redfp,"      if(jwidth .gt. 0) then\n");
        fprintf(redfp,"        do 1000 i = 1, nsampl\n");
        fprintf(redfp,"          if(amomq(i) .gt. 0) then\n");
        fprintf(redfp,"            aprop(i) = pphase*aprop(i)*");
        fprintf(redfp,"dcmplx(amomq(i) - amassq, amag)\n");
        fprintf(redfp,"          else\n");
        fprintf(redfp,"            aprop(i) = pphase*aprop(i)*");
        fprintf(redfp,"dcmplx(amomq(i) - amassq, 0.0d0)\n");
        fprintf(redfp,"          endif\n");
        fprintf(redfp," 1000   continue\n");
        fprintf(redfp,"      endif\n");
    } else {
    fprintf(redfp,"      if (amomq .gt. 0) then\n");
    fprintf(redfp,"          if(jwidth .eq. 0) then\n");
    fprintf(redfp,"              if(amassq .eq. 0.0d0) then\n");
    fprintf(redfp,"                  wid=0.d0\n");
    fprintf(redfp,"              else\n");
    fprintf(redfp,"                  wid=amag/amassq*amomq\n");
    fprintf(redfp,"              end if\n");
    fprintf(redfp,"              aprop = pphase*aprop*dcmplx(amomq -");
    fprintf(redfp," amassq, wid)\n");
    fprintf(redfp,"          else\n");
    fprintf(redfp,"              aprop = pphase*aprop*dcmplx(amomq -");
    fprintf(redfp," amassq, amag)\n");
    fprintf(redfp,"          endif\n");
    fprintf(redfp,"      else\n");
    fprintf(redfp,"        aprop = pphase*aprop*dcmplx(amomq - ");
    fprintf(redfp,"amassq, 0.0d0)\n");
    fprintf(redfp,"      endif\n");
    fprintf(redfp," \n");
    }
    fprintf(redfp,"*     call ctime('smprpd')\n");
    fprintf(redfp,"      return\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}
/*==================================================================*/
Public  Void    cmzdn(sdir)
string  sdir;
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    cmopend(ROFORT, sdir, "snprpdnc.f");
    cmsepar('*');
    fprintf(redfp,"      subroutine snprpdn(pphase, aprop, amomq, ");
    fprintf(redfp,"amassq, amag)\n");
/*    fprintf(redfp, "      function   zdnn(amomq,amassq,amag)\n");*/
    fprintf(redfp, "      implicit real*8(a-h,o-z)\n");
/*  fprintf(redfp, "      complex*16 zdnn\n");*/
    fprintf(redfp,"      complex*16 pphase, aprop\n");
    fprintf(redfp, "      real*8     amomq, amassq, amag\n");
    fprintf(redfp, "      common /chcntl/ jwidth, jtgamm\n");
    fprintf(redfp, "* phase of propagators\n");
#if 0
    fprintf(redfp, "      common /smpphs/pphase,pphas1,pphas2,pphas3,pphas4,");
    fprintf(redfp, "pphas5,pphas6\n");
    fprintf(redfp, "      complex*16 pphase,pphas1,pphas2,pphas3,pphas4,");
    fprintf(redfp, "pphas5,pphas6\n");
#endif
    fprintf(redfp, "*\n");
    fprintf(redfp, "*    calculate denominator of propagator.\n");
    fprintf(redfp, "*\n");
    fprintf(redfp, "*     pphase : input  : phase of the propagator\n");
    fprintf(redfp, "*     amomp  : input  : square of mementum.\n");
    fprintf(redfp, "*     amassq : input  : square of mass.\n");
    fprintf(redfp, "*     amag   : input  : mass * width.\n");
    cmsepar('-');
    fprintf(redfp, "      if (amomq .gt. 0) then\n");
    fprintf(redfp, "          if(jwidth .eq. 0) then\n");
    fprintf(redfp, "              if(amassq .eq. 0.0d0) then\n");
    fprintf(redfp, "                  wid=0.d0\n");
    fprintf(redfp, "              else\n");
    fprintf(redfp, "                  wid=amag/amassq*amomq\n");
    fprintf(redfp, "              end if\n");
    fprintf(redfp, "              aprop = pphase*aprop*dcmplx(amomq - amassq, wid)\n");
    fprintf(redfp, "          else\n");
    fprintf(redfp, "              aprop = pphase*aprop*dcmplx(amomq - amassq, amag)\n");
    fprintf(redfp, "          endif\n");
    fprintf(redfp, "      else\n");
    fprintf(redfp, "        aprop = pphase*aprop*dcmplx(amomq - amassq, 0.0d0)\n");
    fprintf(redfp, "      endif\n");
    fprintf(redfp, " \n");
    fprintf(redfp, "*     call ctime('snprpdn')\n");
    fprintf(redfp, "      return\n");
    fprintf(redfp, "      end\n");
    cmsepar('*');
    fprintf(redfp,"      subroutine snprpdc(pphase, aprop, amomq, ");
    fprintf(redfp,"amassq, amag)\n");
#if 0
    fprintf(redfp, "      function   zdnc(amomq,amassq,amag)\n");
#endif
    fprintf(redfp, "      implicit real*8(a-h,o-z)\n");
#if 0
    fprintf(redfp, "      complex*16 zdnc\n");
#endif
    fprintf(redfp,"      complex*16 pphase, pphasc, aprop\n");
    fprintf(redfp, "      real*8     amomq, amassq, amag\n");
    fprintf(redfp, "      common /chcntl/ jwidth, jtgamm\n");
/*
    fprintf(redfp, "* phase of propagators\n");
    fprintf(redfp, "      common /smpphs/pphase,pphas1,pphas2,pphas3,pphas4,pphs5,pphas6\n");
    fprintf(redfp, "      complex*16 pphase,pphas1,pphas2,pphas3,pphas4,pphs5,pphas6\n");
*/
    fprintf(redfp, "*\n");
    fprintf(redfp, "*    calculate denominator of propagator.\n");
    fprintf(redfp, "*\n");
    fprintf(redfp, "*     pphase : input  : phase of the propagator\n");
    fprintf(redfp, "*     amomp  : input  : square of mementum.\n");
    fprintf(redfp, "*     amassq : input  : square of mass.\n");
    fprintf(redfp, "*     amag   : input  : mass * width.\n");
    cmsepar('-');
    fprintf(redfp, "      amagc  = - amag\n");
    fprintf(redfp, "      pphasc = conjg(pphase)\n");
    cmsepar('-');
    fprintf(redfp, "      if (amomq .gt. 0) then\n");
    fprintf(redfp, "          if(jwidth .eq. 0) then\n");
    fprintf(redfp, "              if(amassq .eq. 0.0d0) then\n");
    fprintf(redfp, "                  wid=0.d0\n");
    fprintf(redfp, "              else\n");
    fprintf(redfp, "                  wid=amagc/amassq*amomq\n");
    fprintf(redfp, "              end if\n");
    fprintf(redfp, "              aprop = pphasc*aprop*dcmplx(amomq - amassq, wid)\n");
    fprintf(redfp, "          else\n");
    fprintf(redfp, "              aprop = pphasc*aprop*dcmplx(amomq - amassq, amagc)\n");
    fprintf(redfp, "          endif\n");
    fprintf(redfp, "      else\n");
    fprintf(redfp, "        aprop = pphasc*aprop*dcmplx(amomq - amassq, 0.0d0)\n");
    fprintf(redfp, "      endif\n");
    fprintf(redfp, " \n");
    fprintf(redfp, "*     call ctime('snprpdc')\n");
    fprintf(redfp, "      return\n");
    fprintf(redfp, "      end\n");

    if(rvcode) {
        cmzdnr();
    }
}
/*==================================================================*/
Private  Void    cmzdnr(NoArg)
/*------------------------------------------------------------------*
 *------------------------------------------------------------------*/
{
    cmopen(ROFORT, "snprpdncv.f");

    cmsepar('*');
    fprintf(redfp,"      subroutine snprpdnv");
    fprintf(redfp,"(pphase, aprop, amomq, amassq, amag)\n");
    fprintf(redfp,"*\n");
    fprintf(redfp,"* phase of propagators\n");
    fprintf(redfp,"*\n");
    fprintf(redfp,"*    calculate denominator of propagator.\n");
    fprintf(redfp,"*\n");
    fprintf(redfp,"*     pphase : input  : phase of the propagator\n");
    fprintf(redfp,"*     amomp  : input  : square of mementum.\n");
    fprintf(redfp,"*     amassq : input  : square of mass.\n");
    fprintf(redfp,"*     amag   : input  : mass * width.\n");
    cmsepar('-');
    fprintf(redfp,"      implicit none\n");
    fprintf(redfp,"      include 'inclvs.h'\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      complex*16 pphase, aprop(n_vector)\n");
    fprintf(redfp,"      real*8     amomq(n_vector), amassq, amag\n");
    fprintf(redfp,"      integer    jwidth, jtgamm\n");
    fprintf(redfp,"      common /chcntl/ jwidth, jtgamm\n");
    fprintf(redfp,"      real*8  wid\n");
    fprintf(redfp,"      integer iv\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      do iv = 1, n_vector\n");
    fprintf(redfp,"        if (amomq(iv) .gt. 0) then\n");
    fprintf(redfp,"          if(jwidth .eq. 0) then\n");
    fprintf(redfp,"            if(amassq .eq. 0.0d0) then\n");
    fprintf(redfp,"              wid=0.d0\n");
    fprintf(redfp,"            else\n");
    fprintf(redfp,"              wid=amag/amassq*amomq(iv)\n");
    fprintf(redfp,"            end if\n");
    fprintf(redfp,"            aprop(iv) = pphase*aprop(iv)*\n");
    fprintf(redfp,"     &          dcmplx(amomq(iv) - amassq, wid)\n");
    fprintf(redfp,"          else\n");
    fprintf(redfp,"            aprop(iv) = pphase*aprop(iv)*\n");
    fprintf(redfp,"     &          dcmplx(amomq(iv) - amassq, amag)\n");
    fprintf(redfp,"          endif\n");
    fprintf(redfp,"        else\n");
    fprintf(redfp,"          aprop(iv) = pphase*aprop(iv)*\n");
    fprintf(redfp,"     &        dcmplx(amomq(iv) - amassq, 0.0d0)\n");
    fprintf(redfp,"        endif\n");
    fprintf(redfp,"      enddo\n");
    fprintf(redfp," \n");
    fprintf(redfp,"*     call ctime('snprpdnv')\n");
    fprintf(redfp,"      return\n");
    fprintf(redfp,"      end\n");

    cmsepar('*');
    fprintf(redfp,"      subroutine snprpdcv");
    fprintf(redfp,"(pphase, aprop, amomq, amassq, amag)\n");
    fprintf(redfp,"*\n");
    fprintf(redfp,"* phase of propagators\n");
    fprintf(redfp,"*\n");
    fprintf(redfp,"*    calculate denominator of propagator.\n");
    fprintf(redfp,"*\n");
    fprintf(redfp,"*     pphase : input  : phase of the propagator\n");
    fprintf(redfp,"*     amomp  : input  : square of mementum.\n");
    fprintf(redfp,"*     amassq : input  : square of mass.\n");
    fprintf(redfp,"*     amag   : input  : mass * width.\n");
    cmsepar('-');
    fprintf(redfp,"      implicit none\n");
    fprintf(redfp,"      include 'inclvs.h'\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      complex*16 pphase, aprop(n_vector)\n");
    fprintf(redfp,"      real*8     amomq(n_vector), amassq, amag\n");
    fprintf(redfp,"      integer    jwidth, jtgamm\n");
    fprintf(redfp,"      common /chcntl/ jwidth, jtgamm\n");
    fprintf(redfp,"      real*8  wid\n");
    fprintf(redfp,"      integer iv\n");
    fprintf(redfp,"      real*8  amagc\n");
    fprintf(redfp,"      complex*16 pphasc\n");
    fprintf(redfp,"\n");
    fprintf(redfp,"      amagc  = - amag\n");
    fprintf(redfp,"      pphasc = conjg(pphase)\n");
    cmsepar('-');
    fprintf(redfp,"      do iv = 1, n_vector\n");
    fprintf(redfp,"        if (amomq(iv) .gt. 0) then\n");
    fprintf(redfp,"          if(jwidth .eq. 0) then\n");
    fprintf(redfp,"            if(amassq .eq. 0.0d0) then\n");
    fprintf(redfp,"              wid=0.d0\n");
    fprintf(redfp,"            else\n");
    fprintf(redfp,"              wid=amagc/amassq*amomq(iv)\n");
    fprintf(redfp,"            end if\n");
    fprintf(redfp,"            aprop(iv) = pphasc*aprop(iv)*\n");
    fprintf(redfp,"     &          dcmplx(amomq(iv) - amassq, wid)\n");
    fprintf(redfp,"          else\n");
    fprintf(redfp,"            aprop(iv) = pphasc*aprop(iv)*\n");
    fprintf(redfp,"     &          dcmplx(amomq(iv) - amassq, amagc)\n");
    fprintf(redfp,"          endif\n");
    fprintf(redfp,"        else\n");
    fprintf(redfp,"          aprop(iv) = pphasc*aprop(iv)*\n");
    fprintf(redfp,"     &        dcmplx(amomq(iv) - amassq, 0.0d0)\n");
    fprintf(redfp,"        endif\n");
    fprintf(redfp,"      enddo\n");
    fprintf(redfp," \n");
    fprintf(redfp,"*     call ctime('snprpdcv')\n");
    fprintf(redfp,"      return\n");
    fprintf(redfp,"      end\n");
    fprintf(redfp,"\n");
}

/*==================================================================*/
Public  int     cmisgn(il)
int     il;
/*------------------------------------------------------------------*
 * calculate the sign of the momentum of the propagator il
 */
{
    int p, t, t0, n, i, d, lg;

    p = abs(ndparticle(iltable[il].ilnodes[0],iltable[il].illegs[0]));

    t = ptable[abs(p)]->ptype;

    if(iltable[il].iltype == ILEXTERN) {
        return 1;
    } else if(t != P_MAJORANA && t != P_DIRAC) {
        return 1;
    }

    n = iltable[il].ilnodes[0];
    i = iltable[il].illegs[0];

    /* skip to fermion */
    for(lg = 0; lg < ndlegs(n); lg++) {
        t0 = ptable[abs(ndparticle(n, lg))]->ptype;
        if(t0 != P_MAJORANA && t0 != P_DIRAC) {
            i--;
        } else {
            lg++;
            if(i >= 2) {
                i -= 2;
            }
        }
        if(i < 2) {
            break;
        }
    }

    /* strange case, 4-fermi vertex ? */
    if(lg >= ndlegs(n) || i < 0 || i >= 2) {
        fprintf(Stderr, "*** cmisgn : illgal fermion vertex\n");
        fprintf(Stderr, "    il=%d, n=%d, i=%d, lg=%d, p=%d, t=%d\n",
                       il, n, i, lg, p, t);
        prdump();
        fprintf(Stderr, "*** cmisgn : illgal fermion vertex\n");
        fprintf(Stderr, "    il=%d, n=%d, i=%d, lg=%d, p=%d, t=%d\n",
                       il, n, i, lg, p, t);
        exit(FAIL);
    }

    d = - 2*i + 1;      /* normal direction of momentum */

    return d;
}

/*==================================================================*/
Public  bool    cmdic(mode, name)
int     mode;
string  name;
/*------------------------------------------------------------------*
 * mode : 0=initialize, 1=find
 *------------------------------------------------------------------*/
{
    static string stable[MAXDIR];
    static int    nstable=0;
    int i;

    if(mode == 0) {
        for(i = nstable-1; i >= 0; i--) {
            free(stable[i]);
            stable[i] = NULL;
        }
        nstable = 0;
        return False;
    }
    if(name == NULL || strlen(name) < 1) {
        return False;
    }
    for(i = 0; i < nstable; i++) {
        if(strcmp(stable[i], name) == 0) {
            return True;
        }
    }
    if(nstable >= MAXDIR) {
        fprintf(Stderr, "*** cmdic : too many names\n");
        exit(FAIL);
    }
    stable[nstable++] = (string) strdup(name);
    return False;
}
