/*********************************************************************
 * common.h - definitions of types and mocros for portable c
 *
 *   Input
 *     DEFGLOBAL  : if defined then define global variables.
 *
 *   1992/3/24    Kaneko T. based on
 *                "Portable C and UNIX System Programing"
 *                by J. E. Lapin, Prentic-Hall
 *
 *********************************************************************
 * Machin
 *   Define symbol "FACOM" for FACOM C-compiler.
 *   Define symbol "MINIX_ST" for minix on ATARI-ST.
 *   Turbo-C is automatically recognized.
 *
 */
#ifndef second

/* #define FACOM */
/* #define MINIX_ST */
/* #define SONY_NEWS */

#ifdef __TURBOC__
# define TURBOC
#endif

/*********************************************************************
 *  Symbols
 *
 *    ANSI     : if compiler is ANSI conpatible,
 *               then this symbol is defined.
 *    MINIX    : MINIX
 *    NO_CLOCK : if there are no clock function.
 */

#ifdef  __STDC__
# define ANSI
#endif

#ifdef TURBOC
# define ANSI
#endif

#ifdef  MINIX_ST
#define MINIX
#endif

#ifdef  MINIX
# define NO_CLOCK
# define float     int        /* no float in MINIX */
# define double    long       /* no long  in MINIX */
#endif

#ifdef  SONY_NEWS
# define NO_CLOCK
#endif

/*********************************************************************
 *  Standard types
 *
 *    bool   : integers as boolean. Value is either
 *             "True" or "False".
 *    flag   : integers as flags (>= 8bits).
 *    char   : unsigned 7 bit character.
 *             Use macro "CHAR(x)" to convert to this type.
 *    byte   : 8 bit unsigned integer.
 *             Use macro "BYTE(x)" to convert to this type.
 *    word   : unsigned integer (>=16bits)
 *             Use macro "WORD(x)" to convert to this type.
 *    path   : path of files.  Its maximum length is "MAXPATHLEN
 *    Void   : void finction type
 *    AnyP   : void pointer type
 *    Discard : discard return value of a function.
 */

typedef int             bool;
#ifndef True
#define True      1
#define False     0
#endif

typedef char            flag;

typedef char            byte;

typedef unsigned short  word;

typedef unsigned int    Uint;

typedef unsigned long   Ulong;

typedef char            *string;

#ifndef MAXPATHLEN
# define MAXPATHLEN     256
#endif
typedef char            path[MAXPATHLEN];

#ifdef ANSI
# define Void   void
#else
# define Void   int
#endif

#ifdef ANSI
# define AnyP  void*
#else
# define AnyP  void*
/* # define AnyP  char*  
 */
#endif

#ifdef ANSI
# define Discard   (void)
#else
# define Discard   (void)
#endif

/*********************************************************************
 * Declarations
 *
 *   Varaibles:
 *     Define with eigher "Global" or "Local".
 *
 *   Functions:
 *     Prototypes: Declare with either "Public", "Private" or "Extern".
 *                 Arguments are declared as the argument of a macro
 *                 "ARG(arg)".
 *                 For example;
 *
 *                 Public Void f  ARG((int *, char));
 *                 Extern int  g  ARG((NoArg));
 *
 *     Definition: Declare with either "Public" or "Private".
 *                 Arguments are declared as a list of variables.
 *                 Tyeps of arguments are declared in the succeeding
 *                 lines.
 *                 Use "NoArg" macro in the case of no arguments.
 *                 For example;
 *
 *                 Public Void f(p, c)
 *                 int *p;
 *                 char c;
 *                 {
 *                      ...
 *                 }
 *
 *     Discard the return value.
 *                 Extern int g  ARG((NoArg));
 *
 *                 Discard g();
 *
 *    Change type of function by casting
 *                 int cmp ARG((int, int));
 *
 *                 qsort(array, narray, sizeof(*array), 
 *                       FTYPE((int (*)(const void*,const void*)));
 */
#ifdef DEFGLOBAL
# define        Global
#else
# define        Global  extern
#endif
#define         Local   static

#define         Private static
#define         Public

/* void argument */
#ifdef ANSI
# define NoArg  void
#else
# define NoArg
#endif

#define Extern extern
#ifdef ANSI
# define ARG(x) x
#else
# define ARG(x) ()
#endif

#ifdef ANSI
# define FTYPE(x)  x
#else
# define FTYPE(x)
#endif

#ifdef ANSI
# define Const const
#else
# define Const 
#endif
/*********************************************************************
 *  I/O
 *
 *    Stderr  : use in stead of "stderr".
 */

/*
#define Stderr  stderr
*/
#define Stderr  stdout

/*********************************************************************
 *  Standard constant
 *
 *    SUCCEED : for "exit(SUCCEED)".
 *    FAIL    : for "exit(FAIL)".
 *    EOL     : end of line.
 *    EOS     : end of string.
 *    EOP     : end of page (Form Feed).
 */
#define SUCCEED   0
#define FAIL      (-1)

#define EOL       '\n'
#define EOS       '\0'
#define EOP       '\14'

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif

#ifndef R_OK    /* defined in POSIX standard */
# define R_OK    4
# define W_OK    2
# define X_OK    1
# define F_OK    0
#endif

/*********************************************************************
 *  Standard macros
 *
 *    CHAR(x)        : convert to char type with 7 bits.
 *    BYTE(x)        : convert to byte type.
 *    WORD(x)        : convert to word type.
 *
 *    HIBYTE(x)      : take an upper byte.
 *    LOWBYTE(x)     : take a lower byte.
 *
 *    HIWORD(x)      : take upper two bytes.
 *    LOWORD(x)      : take lower two bytes.
 *
 *    DECODE(x)      : value of a digit "x"
 *
 *    LURSHIFT(x, b) : logical right "b" bits shift of long "x"
 *    DIM(a)         : the number of elements of an array "a".
 */
#ifdef  FACOM
# define CHAR(x)    ((x) & 0xff)
#else
# define CHAR(x)    ((x) & 0x7f)
#endif
#define BYTE(x)    ((x) & 0xff)
#define WORD(x)    ((x) & 0xffffL)

#define HIBYTE(x)  (((x) >> 8) & 0xff)
#define LOBYTE(x)  ((x) & 0xff)
#define HIWORD(x)  (((x)>>16) & 0xffffL)
#define LOWORD(x)  ((x) & 0xffffL)
#define DECODE(x)  ((int) ((x)-'0'))

#define LURSHIFT(x, b) (((long)(x) >> (b)) & (0x7fffffff >> ((b)-1)))

#define DIM(a)  (sizeof(a)/sizeof(*(a)))

/*********************************************************************
 *  System dependent functions
 *
 *  double second(double *)
 *
 */

#ifdef NO_CLOCK

#define second(t) ((double) time(t))

#define second(t) ((double) time(t))

#else

#ifndef CLOCK_TCK
#ifdef  CLOCKS_PER_SEC
#define CLOCK_TCK CLOCKS_PER_SEC
#else
#ifdef  CLK_TCK
#define CLOCK_TCK CLK_TCK
#else
#define CLOCK_TCK 1000000
#endif
#endif
#endif

#define second(t) ((double)(*(t)=(double)(clock()/((double)CLOCK_TCK))))

#endif

/*********************************************************************
 *  Problems
 */
#ifdef MINIX
#define float     int        /* no float in MINIX */
#define double    long       /* no long  in MINIX */
#endif
/*
  end of common.h
*********************************************************************/
#endif
