/*********************************************************************
 *  File : conserv.c
 *
 *  Check conservation low in the defined model.
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grdata.h"
#include "conserv.h"

/*--------------------------------------------------------------------
 * Parameters
 */
#define MODCON

/*--------------------------------------------------------------------
 * Private functions
 */
Private Void    coinit         ARG((NoArg));
Private Void    covertex       ARG((NoArg));
Private Void    covadd         ARG((int));
Private Void    covadd1        ARG((int,int));
Private Void    covaddn        ARG((int,int));
Private int     cogcd          ARG((int,int));
Private Void    cosave         ARG((NoArg));
#ifdef CODBG
Private Void    covdump        ARG((int,int*,string));
#endif

/*--------------------------------------------------------------------
 * local variables
 */
#define COCDEL     -1
#define COCINF      0

Local int   *coctype=NULL;
Local int   *coppart=NULL;
Local int   *copvect=NULL;

/*==================================================================*/
Public  bool    constest(n, prt)
int     n;
int    *prt;
/*--------------------------------------------------------------------
 * check conservation low for a list "int prt[npptable]" of particles.
 */
{
    int p, c, sum;

    if(coctbl == NULL) {
        conserv();
    }

    for(p = 0; p < npptable; p++) {
        coppart[p] = 0;
    }
    for(c = 0; c < n; c++) {
        p = prt[c];
        if(p < 0) {
            coppart[-p]--;
        } else {
            coppart[p]++;
        }
    }

    if(coindep < 1) {
        return True;
    }

    for(c = 0; c < coindep; c++) {
        sum = 0;
        for(p = 0; p < npptable; p++) {
            sum += coctbl[c][p]*coppart[p];
        }
#ifdef MODCON
        if(coctype[c] > 1) {
            sum %= coctype[c];
        }
#endif
        if(sum != 0) {
            return False;
        }
    }
    return True;
}

/*==================================================================*/
Public  Void    conserv(NoArg)
/*--------------------------------------------------------------------
 * create table of conservation low "coctbl".
 */
{
    coinit();
#ifdef CODBG
    printf("after coinit\n"); coprint();printf("after coinit\n");
#endif
    covertex();
#ifdef CODBG
    printf("after covertex\n"); coprint();printf("after covertex\n");
#endif
    cosave();
#ifdef CODBG
    printf("after cosave\n"); coprint();printf("after cosave\n");
#endif
}

/*==================================================================*/
Private Void    coinit(NoArg)
/*--------------------------------------------------------------------
 */
{
    pindex p, c;

    if(coctbl != NULL) {
        return;
    }

    coindep = npptable;

    /* space of coctbl */
    coctbl = (int **) malloc(sizeof(int*)*npptable);
    if(coctbl == NULL) {
        fprintf(Stderr, "*** copopen : memory allocation error\n");
        exit(1);
    }
    for(p = 0; p < npptable; p++) {
        coctbl[p] = (int *) malloc(sizeof(int)*npptable);
        if(coctbl[p] == NULL) {
            fprintf(Stderr, "*** copopen : memory allocation error\n");
            exit(1);
        }
    }

    /* space of coppart and copvect */
    coctype = (int *) malloc(sizeof(int)*coindep);
    if(coctype == NULL) {
        fprintf(Stderr, "*** copopen : memory allocation error\n");
        exit(1);
    }
    coppart = (int *) malloc(sizeof(int)*npptable);
    if(coppart == NULL) {
        fprintf(Stderr, "*** copopen : memory allocation error\n");
        exit(1);
    }
    copvect = (int *) malloc(sizeof(int)*coindep);
    if(copvect == NULL) {
        fprintf(Stderr, "*** copopen : memory allocation error\n");
        exit(1);
    }

    for(p = 0; p < npptable; p++) {
        for(c = 0; c < coindep; c++) {
            coctbl[p][c] = 0;
        }
        coctbl[p][p] = 1;
    }
    for(c = 0; c < coindep; c++) {
        coctype[c] = COCINF;
    }
}


/*==================================================================*/
Private Void    covertex(NoArg)
/*--------------------------------------------------------------------
 * analysis of conservation of particle number
 */
{
    int v, p, l, c, mode;

#ifdef MODCON
    for(mode = 0; mode < 2; mode++) {
#else
    mode = 0; {
#endif
        /* initial data */
        if(mode == 0) {
            for(p = 0; p < npptable; p++) {
                for(c = 0; c < coindep; c++) {
                    coctbl[p][c] = 0;
                }
                coctbl[p][p] = 1;
            }
            for(c = 0; c < coindep; c++) {
                coctype[c] = COCINF;
            }
        } else {
            for(c = 0; c < coindep; c++) {
                if(coctype[c] != COCDEL) {
                    coctype[c] = 1;
                } else {
                    coctype[c] = COCINF;
                    coctbl[c][c] = 1;
                }
            }
        }

        for(v = nvstart; v < nvtable; v++) {
            if(vtable[v]->composit == V_COMPOSIT) {
                continue;
            }
            for(p = 0; p < npptable; p++) {
                coppart[p] = 0;
            }

            for(l = 0; l < vtable[v]->nvlegs; l++) {
                p = vtable[v]->vlegs[l];
                if(p > 0) {
                    coppart[p]++;
                } else {
                    coppart[-p]--;
                }
            }
            covadd(mode);
        }
    }
}

/*==================================================================*/
Private Void    covadd(mode)
int     mode;
/*--------------------------------------------------------------------
 */
{
    int p, c, c1, n;

#ifdef CODBG
if(mode != 0) covdump(npptable, coppart, "coppart");
#endif
    n  = 0;
    c1 = 0;
    for(c = coindep-1; c >= 0; c--) {
        copvect[c] = 0;
        for(p = 0; p < npptable; p++) {
            copvect[c] += coctbl[c][p]*coppart[p];
        }
        if(mode != 0) {
            if(coctype[c] > 1) {
                copvect[c] %= coctype[c];
                if(copvect[c] < 0) {
                    copvect[c] += coctype[c];
                }
            }
        }
        if(copvect[c] != 0) {
            n++;
#ifdef TEST0
            if(c1 == 0) {
                c1 = c;
            } else if(abs(copvect[c]) < abs(copvect[c1])) {
                c1 = c;
            }
#else
            if(c1 == 0) {
                c1 = c;
            } else if(coctype[c1] != COCINF) {
                if(coctype[c] == COCINF) {
                    c1 = c;
                } else if(coctype[c] > coctype[c1]) {
                    c1 = c;
                }
            }
#endif
        }
    }
    if(n == 1) {
        covadd1(c1, mode);
    } else if(n > 0) {
        covaddn(c1, mode);
    } else {
#ifdef CODBG
if(mode != 0) printf("no condition\n");
#endif
    }
    return;
}

/*==================================================================*/
Private Void    covadd1(c1, mode)
int     c1;
int     mode;
/*--------------------------------------------------------------------
 */
{
    int p;

    if(mode != 0) {
        if(coctype[c1] == COCINF) {
            coctype[c1] = abs(copvect[c1]);
        } else {
            coctype[c1] = cogcd(abs(copvect[c1]), coctype[c1]);
        }
        if(coctype[c1] > 1) {
            return;
        }
    }
    for(p = 0; p < npptable; p++) {
        coctbl[c1][p] = 0;
    }
    coctype[c1] = COCDEL;
}

/*==================================================================*/
Private Void    covaddn(c1, mode)
int     c1;
int     mode;
/*--------------------------------------------------------------------
 */
{
    int d, p, c, gcd;

    d = copvect[c1];
    for(c = 0; c < coindep; c++) {
        if(c == c1) {
            continue;
        }
        for(p = 0; p < npptable; p++) {
            coctbl[c][p] = d*coctbl[c][p] - coctbl[c1][p]*copvect[c];
        }
    }
    for(p = 0; p < npptable; p++) {
        coctbl[c1][p] = 0;
    }

    for(c = 0; c < coindep; c++) {
        if(mode != 0) {
            if(copvect[c] != 0 && coctype[c] != COCDEL && coctype[c1] > 1) {
                if(coctype[c] == COCINF) {
                    coctype[c] = coctype[c1];
                } else if(coctype[c1] > 1) {
                    coctype[c] = cogcd(coctype[c1], coctype[c]);
                }
            }
        }

        gcd = 0;
        for(p = 0; p < npptable; p++) {
            if(coctbl[c][p] != 0) {
                if(gcd == 0) {
                    gcd = abs(coctbl[c][p]);
                } else {
                    gcd = cogcd(gcd, abs(coctbl[c][p]));
                }
                if(gcd < 2) {
                    break;
                }
            }
        }
        if(gcd > 1) {
            for(p = 0; p < npptable; p++) {
                coctbl[c][p] /= gcd;
            }
        }
    }
    coctype[c1] = COCDEL;
}

/*==================================================================*/
Private int     cogcd(x, y)
int     x, y;
/*--------------------------------------------------------------------
 *    Calculate greatest common diviser of x and y.
 */
{
    int m;

    x = (x < 0) ? -x : x;
    y = (y < 0) ? -y : y;

    while(y != 0) {
        m = x % y;
        x = y;
        y = m;
    }
    return x;
}

/*==================================================================*/
Private Void    cosave(NoArg)
/*--------------------------------------------------------------------
 */
{
    int cl, ch, t;
    int *p;

    cl = 0;
    ch = coindep - 1;
    while(cl < ch) {
        while(cl < coindep && coctype[cl] != COCDEL) {
            cl++;
        }
        while(ch >= 0 && coctype[ch] == COCDEL) {
            ch--;
        }
        if(cl >= ch) {
            coindep = cl;
            return;
        }
        t = coctype[cl];
        coctype[cl] = coctype[ch];
        coctype[ch] = t;
        p = coctbl[cl];
        coctbl[cl] = coctbl[ch];
        free(p);
        cl++;
    }
    coindep = cl;
    return;
}

/*==================================================================*/
Public  Void    cocdump(NoArg)
/*--------------------------------------------------------------------
 */
{
    int p, c;

    fprintf(Stderr, "*** cocdump (npptable=%d, coindep=%d) ***\n",
                    npptable, coindep);

    if(coctbl == NULL) {
        fprintf(Stderr, "coctbl = NULL\n");
    } else {
        if(coctype != NULL) {
            fprintf(Stderr, "  :");
            for(c = 0; c < coindep; c++) {
                fprintf(Stderr, "%2d ", c);
            }
            putc('\n', Stderr);
        }
        if(coctype != NULL) {
            fprintf(Stderr, "tp:");
            for(c = 0; c < coindep; c++) {
                fprintf(Stderr, "%2d ", coctype[c]);
            }
            putc('\n', Stderr);
        }

        for(p = 0; p < npptable; p++) {
            fprintf(Stderr, "%2d:", p);
            for(c = 0; c < coindep; c++) {
                if(coctbl[c] == NULL) {
                    fprintf(Stderr, "NULL ");
                } else {
                    fprintf(Stderr, "%2d ", coctbl[c][p]);
                }
            }
            fprintf(Stderr, "%s\n", ntable[ptable[p]->pname[0]].name);
        }
    }
}

/*==================================================================*/
Public  Void    coprint(NoArg)
/*--------------------------------------------------------------------
 */
{
    int p, c;

    if(coctbl == NULL) {
        fprintf(Stderr, "coctbl = NULL\n");
        return;
    }

    for(p = 1; p < npptable; p++) {
        fprintf(Stderr, "%-16s : ", ntable[ptable[p]->pname[0]].name);
        for(c = 0; c < coindep; c++) {
            if(coctbl[c][p] == 0) {
                continue;
            }
            if(coctbl[c][p] > 0) {
                fprintf(Stderr," + %2d*E%d", abs(coctbl[c][p]), c);
            } else if(coctbl[c][p] < 0) {
                fprintf(Stderr," - %2d*E%d", abs(coctbl[c][p]), c);
            }
            if(coctype[c] > 0) {
                fprintf(Stderr,"(%d)", coctype[c]);
            }
        }
        fprintf(Stderr, "\n");
    }
}

#ifdef CODBG
/*==================================================================*/
Private Void    covdump(n, an, name)
int     n;
int    *an;
string  name;
/*--------------------------------------------------------------------
 */
{
    int i;

    fprintf(Stderr, "%s[0..%d-1] = ", name, n);
    for(i = 0; i < n; i++) {
        fprintf(Stderr, " %2d", an[i]);
    }
    putc('\n', Stderr);
}
#endif

/*==================================================================*/
Public Void    comdump(NoArg)
/*--------------------------------------------------------------------
 */
{
    int c, p;

    fprintf(Stderr, "comdump\n");
    fprintf(Stderr, "    %-4s : ", "ptcl");
    for(c = 0; c < coindep; c++) {
        fprintf(Stderr, "%3d ", c);
    }
    fprintf(Stderr, "\n");

    for(p = 1; p < npptable; p++) {
        fprintf(Stderr, "%2d: %-4s : ", p, ntable[ptable[p]->pname[0]].name);
        for(c = 0; c < coindep; c++) {
            fprintf(Stderr, "%3d ", coctbl[c][p]);
        }
        fprintf(Stderr, "\n");
    }
}
