/*******************************************************************
 *  File : dagrf.c
 *
 *    random access of graph data.
 *
 *                                               by Kaneko Toshiaki
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "grdata.h"

#define MAJORANA

/*--------------------------------------------------------------------
 * Global variables
 */
/* error of system calls */
/*
Extern int sys_nerr;
Extern char *sys_errlist[];
*/
Extern int errno;

/*--------------------------------------------------------------------
 * Private functions
 */
Private Void    daget           ARG((int prc, int gn, tgraph *grf));
Private int     dawrite         ARG((int, tgraph*));
Private bool    dafind          ARG((int,int,tgraph**));
Private bool    dachkext        ARG((int,int));
    /* for 4 gluon vertices */
Private int     dagl4ini        ARG((NoArg));
Private Void    dagl4nxtgr      ARG((NoArg));

/*--------------------------------------------------------------------
 * Local variable
 */
#ifndef MAXPROCS
#define MAXPROCS   5
#endif
#define MAXGRAPHS  50000
#define MAXBUFF    16     /* the number of graphs in the buffer */
#define INFOBUFFSIZE  64

Local int      nprocs = 0;
Local bool     expandgl4 = True;
Local FILE    *dafp = NULL;
Local tproc    pctbl[MAXPROCS];  /* table of processes */
Local long    *gptbl[MAXPROCS];  /* table of graph positions */
Local int      gntbl[MAXPROCS];  /* table of graph numbers */
Local char    *grtbl[MAXPROCS];  /* table of graph buffers */
Local int     *gbtbl[MAXPROCS];  /* table of graphs in the buffer */
Local int      gatbl[MAXPROCS];  /* table of acess # in the buffer */

Local int      gl4tbl[MAXNODES];       /* work space */
Local int      gl4pos[MAXNODES];       /* work space */
Local int      gl4mod[MAXNODES];       /* work space */

#ifdef SONY_NEWS
Local char     fname[32];
#endif

#define GBUFF(prc, n) ((tgraph *)((char *) \
                       grtbl[prc]+pctbl[prc].msize*(n)))
#define MBUFF(prc)     pctbl[prc].bgraph
#define PCONJG(ptcl)   ((ptable[abs(ptcl)]->antip)*ptcl)
#define NCONJG(v,o,e)  ((v)<(e)?(v):(v)+(o)-(e))
#define ICONJG(v,o)    ((v)+(o))

/*==================================================================*/
Public  int     daproc(proc)
tproc *proc;
/*--------------------------------------------------------------------
 *  returns the process id saved in the da-file
 */
{
    int n, i, msize, ngraphs;

    /* check the number of processes */
    if(nprocs >= MAXPROCS - 1) {
        error("daproc", "too many processes");
        dafail();
    }

    /* create graph table of graph positions */
    if((gptbl[nprocs] = (long*) malloc(MAXGRAPHS*sizeof(long))) == NULL) {
        error("daproc", "cannot allocate table of graphs");
        fprintf(Stderr, "for process = %d\n", nprocs);
        dafail();
    }

    /* create table of graphs */
    msize   = proc->msize;
    ngraphs = proc->bgraph;
    if((grtbl[nprocs] = (char*) malloc(ngraphs*msize)) == NULL) {
        error("daproc", "cannot allocate table of graphs (grtbl)");
        fprintf(Stderr, "for process = %d\n", nprocs);
        dafail();
    }

    /* create table of graphs in the buffer */
    if((gbtbl[nprocs] = (int*) malloc(ngraphs*sizeof(int))) == NULL) {
        error("daproc", "cannot allocate table of graphs (gbtbl)");
        fprintf(Stderr, "for process = %d\n", nprocs);
        dafail();
    }

    for(i = 0; i < ngraphs; i++) {
        gbtbl[nprocs][i] = -1;   /* flag for empty */
    }
    gatbl[nprocs] = 0;

    /* save next file position */
#if DBG
    if(dbg>=dadbg) {
        printf("daproc : ftell()=");
    }
#endif
    if((gptbl[nprocs][0] = ftell(dafp)) < 0) {
#if DBG
        if(dbg>=dadbg) {
            printf("%ld\n", gptbl[nprocs][0]);
        }
#endif
        error("daproc", "ftell failed.");
        fprintf(Stderr, "for process = %d\n", nprocs);
        dafail();
    }
#if DBG
    if(dbg>=dadbg) {
        printf("%ld\n", gptbl[nprocs][0]);
    }
#endif

    /* copy the process */
    memcpy(pctbl+nprocs, proc, sizeof(tproc));

    /* clear graph counter */
    gntbl[nprocs] = 0;

    n = nprocs++;

#if DBG
    if(dbg>=dadbg) {
        printf("daproc n=%d\n", n);
    }
#endif

    return n;
}

/*==================================================================*/
Public  tproc  *danproc(prc)
int     prc;
/*--------------------------------------------------------------------
 *  returns the process corresponding to the process id 'prc'
 */
{
    if(prc >= MAXPROCS) {
        error("danproc", "no more processes\n");
        dafail();
    }
    return (pctbl + prc);
}

/*==================================================================*/
Public  tgraph *danewg(prc, grb)
int     prc;
int     grb;
/*--------------------------------------------------------------------
 */
{
    tgraph *grf;

    if(prc < 0 || prc > nprocs) {
        error("danewg", "no such process\n");
        dafail();
    }
    if(grb >= MBUFF(prc)) {
        error("danewg", "out of buffer\n");
        dafail();
    }
    grf = GBUFF(prc,grb);
    rgigraph(danproc(prc), grf); /* set basen and offsl */

    return grf;
}

/*==================================================================*/
Public  tgraph *datempg(prc)
int     prc;
/*--------------------------------------------------------------------
 */
{
    tgraph *grf;

    if(prc < 0 || prc > nprocs) {
        error("datempg", "no such process\n");
        dafail();
    }
    if((grf = (tgraph*) malloc(pctbl[prc].msize)) == NULL) {
        error("datempg", "cannot allocate temporary graph");
        fprintf(Stderr, "for process = %d\n", prc);
        dafail();
    }
    rgigraph(danproc(prc), grf); /* set basen and offsl */

    return grf;
}

#if 0
/*==================================================================*/
Public  int     daput(prc, grb)
int      prc;
int      grb;
/*--------------------------------------------------------------------
 *  write a graph in the buffer at "grb" to a random access file
 *  return the graph id.
 */
{
    int gn, msize;
    tgraph *grf;

    /* check process number */
    if(prc < 0 || prc >= nprocs) {
        error("daput", "process number is out of range.");
        fprintf(Stderr, "%d should be in [0, %d]\n", prc, nprocs);
        dafail();
    }

    /* check graph number */
    gn = gntbl[prc];
    if(gn >= MAXGRAPHS-1) {
        error("daput", "too many graphs");
        fprintf(Stderr, "for process %d\n", prc);
        dafail();
    }

    /* check buffer */
    if(grb >= MBUFF(prc)) {
        error("daput", "out of buffer\n");
        dafail();
    }

    /* next queue pointer */
    gatbl[prc] = grb;
    gbtbl[prc][grb] = gn;
    grf = GBUFF(prc, grb);


    /* seek */
#if DBG
    if(dbg>=dadbg) {
        printf("daput : fseek(%p, %ld)\n",
               dafp, gptbl[prc][gn]);
    }
#endif
    if(fseek(dafp, gptbl[prc][gn], SEEK_SET) != 0) {
        error("daput", "fseek failed.");
        fprintf(Stderr, "for graph = (%d, %d)\n", prc, gn);
        dafail();
    }

    /* set flag which indicates iline table is not constructed */
    grf->nilines = -1;

    /* write graph to the file */
#if DBG
    if(dbg>=dadbg) {
        printf("daput : fwrite(%p, 1, %d, %p)\n",
               grf, msize, dafp);
    }
#endif
    if(fwrite(grf, 1, msize, dafp) != msize) {
        error("daput", "fwrite failed.");
        fprintf(Stderr, "for graph = (%d, %d)\n", prc, gn);
        dafail();
    }

    /* save next file position */
#if DBG
    if(dbg>=dadbg) {
        printf("daput : ftell(%p)\n", dafp);
     }
#endif
    if((gptbl[prc][gn+1] = ftell(dafp)) < 0) {
        error("daput", "ftell failed.");
        fprintf(Stderr, "for process = %d\n", nprocs);
        dafail();
    }

    /* count up the number of graphs */
    gntbl[prc]++;

    /* index of graph */
    return gn;
}
#endif

/*==================================================================*/
Public  Void    daclose(NoArg)
/*--------------------------------------------------------------------
 *  close all temporary file
 */
{
#ifdef SONY_NEWS
/*
    int f;
*/
#endif

    fclose(dafp);
    dafp = NULL;

#ifdef SONY_NEWS
/*
    sprintf(fname, "dagrf.%3.3d", f);
    if(remove(fname) != 0) {
        error("daclse : rm", sys_errlist[errno]);
        fprintf(Stderr, "fname = %d\n", fname);
        erfail();
    }
*/
#endif

    return;
}

/*==================================================================*/
Public  Void    dafail(NoArg)
/*--------------------------------------------------------------------
 *  close all temporary file
 */
{
    daclose();
    erfail();
}

/*==================================================================*/
Public  tproc *dagetp(proc)
int proc;
/*--------------------------------------------------------------------
 *  returns the process.
 */
{
    if(proc < 0 || proc > nprocs) {
        return NULL;
    }
    return (pctbl + proc);
}

/*==================================================================*/
Public  int     dangrf(proc)
int proc;
/*--------------------------------------------------------------------
 *  returns the number of the graphs of the process.
 */
{
    if(proc < 0 || proc > nprocs) {
        return 0;
    }
    return gntbl[proc];
}

/*==================================================================*/
Private bool   dafind(prc, gn, grfp)
int      prc;
int      gn;
tgraph **grfp;
/*--------------------------------------------------------------------
 *  find graph gn of process prc in the buffer.
 *  if not found, buffer of graph to be used.
 */
{
    int gp, i;
    bool st;

    gp = gatbl[prc];
    st = False;
    for(i = 0; i < MBUFF(prc); i++) {
        if(gbtbl[prc][i] == gn) {
            gp = i;
            st = True;
            break;
        }
    }
    *grfp = GBUFF(prc, gp);

    /* next queue pointer */
    gatbl[prc] = (gatbl[prc] == MBUFF(prc)-1) ? 1 : gatbl[prc] + 1;
    if(MBUFF(prc) == 0) {
        gatbl[prc] = 0;
    }
    gbtbl[prc][gp] = gn;

#if DBG
    if(dbg>=dadbg) {
        printf("dafind: prc=%d, gn=%d, gp=%d, *grfp=%p\n",
               prc,gn,gp,*grfp);
    }
#endif

    return st;
}

/*==================================================================*/
Private int     dawrite(prc, grf)
int      prc;
tgraph  *grf;
/*--------------------------------------------------------------------
 *  write a graph to a random access file
 *  return the graph id.
 */
{
    int gn, msize;
    tgraph *gb;

    /* check process number */
    if(prc < 0 || prc >= nprocs) {
        error("dawrite", "process number is out of range.");
        fprintf(Stderr, "%d should be in [0, %d]\n", prc, nprocs);
        dafail();
    }

    /* check graph number */
    gn = gntbl[prc];
    if(gn >= MAXGRAPHS-1) {
        error("dawrite", "too many graphs");
        fprintf(Stderr, "for process %d\n", prc);
        dafail();
    }

    /* buffer */
    if(dafind(prc, gn, &gb)) {
        return gn;
    }

    /* copy to the buffer */
    msize = pctbl[prc].msize;
    memcpy(gb, grf, msize);

    /* seek */
#if DBG
    if(dbg>=dadbg) {
        printf("dawrite : fseek(%p, %ld)\n",
               dafp, gptbl[prc][gn]);
    }
#endif
    if(fseek(dafp, gptbl[prc][gn], SEEK_SET) != 0) {
        error("dawrite", "fseek failed.");
        fprintf(Stderr, "for graph = (%d, %d)\n", prc, gn);
        dafail();
    }

    /* set flag which indicates iline table is not constructed */
    grf->nilines = -1;

    /* write graph to the file */
#if DBG
    if(dbg>=dadbg) {
        printf("dawrite : fwrite(%p, 1, %d, %p)\n",
               grf, msize, dafp);
    }
#endif
    if(fwrite(grf, 1, msize, dafp) != msize) {
        error("dawrite", "fwrite failed.");
        fprintf(Stderr, "for graph = (%d, %d)\n", prc, gn);
        dafail();
    }

    /* save next file position */
#if DBG
    if(dbg>=dadbg) {
        printf("dawrite : ftell(%p)\n", dafp);
    }
#endif
    if((gptbl[prc][gn+1] = ftell(dafp)) < 0) {
        error("dawrite", "ftell failed.");
        fprintf(Stderr, "for process = %d\n", nprocs);
        dafail();
    }

    /* count up the number of graphs */
    gntbl[prc]++;

    /* index of graph */
    return gn;
}

/*==================================================================*/
Public  tgraph *daread0(prc, gn)
int      prc;
int      gn;
/*--------------------------------------------------------------------
 *  read a graph from a random access file
 *  return the graph.
 */
{
    tgraph  *grf;

    if(!dafind(prc, gn, &grf)) {
        daget(prc, gn, grf);
    }

    return grf;
}

/*==================================================================*/
Public  tgraph *daread(prc, gn)
int      prc;
int      gn;
/*--------------------------------------------------------------------
 *  read a graph from a random access file
 *  Auxiliary vertices are connected.
 *  return the graph.
 */
{
    tgraph  *grf;
    tproc   *prcp;

    grf  = daread0(prc, gn);
    prcp = danproc(prc);
    rgcomposit(prcp, grf);

    return grf;
}

/*==================================================================*/
Public  tgraph *dareadb(prc, gn, gp)
int      prc;
int      gn;
int      gp;
/*--------------------------------------------------------------------
 *  read a graph from a random access file
 *  return the graph.
 */
{
    tgraph  *grf;

    if(gp >= MBUFF(prc)) {
        fprintf(Stderr, "*** dareadb(%d,%d,%d) : out of buffer\n",
                prc, gn, gp);
        exit(1);
    }

    grf = GBUFF(prc, gp);

    daget(prc, gn, grf);

    return grf;
}

/*==================================================================*/
Public  tgraph *dareadf(prc, gn)
int      prc;
int      gn;
/*--------------------------------------------------------------------
 *  read a graph from a random access file with refreshing buffer.
 *  return the graph.
 */
{
    tgraph  *grf;
    tproc   *prcp;

    Discard dafind(prc, gn, &grf);
    daget(prc, gn, grf);
    prcp = danproc(prc);
    rgcomposit(prcp, grf);

    return grf;
}

/*==================================================================*/
Private Void    daget(prc, gn, grf)
int      prc;
int      gn;
tgraph  *grf;   /* buffer to be read */
/*--------------------------------------------------------------------
 *  read a graph from a random access file
 *  return the graph.
 */
{
    int msize, ms;

    /* check process number */
    if(prc < 0 || prc >= nprocs) {
        error("daget", "process number is out of range.");
        fprintf(Stderr, "%d should be in [0, %d]\n", prc, nprocs);
        dafail();
    }

    /* check graph number */
    if(gntbl[prc] < 1) {
        fprintf(Stderr, "*err* daget : the process %d has no graphs: ",
                prc);
        erfail();
    } else if(gn >= gntbl[prc]) {
        error("daget", "the graph has not been saved");
        fprintf(Stderr, "for process %d\n", prc);
        dafail();
    }

    /* seek */
#if DBG
    if(dbg>=dadbg) {
        printf("daget : fseek(%p, %ld)\n", dafp, gptbl[prc][gn]);
    }
#endif
    if(fseek(dafp, gptbl[prc][gn], SEEK_SET) != 0) {
        error("daget", "fseek failed.");
        fprintf(Stderr, "for graph = (%d, %d)\n", prc, gn);
        dafail();
    }

    /* read graph to the file */
    msize = pctbl[prc].msize;

#if DBG
    if(dbg>=dadbg) {
        printf("daget : fread(%p, 1, %d, %p)\n",
               grf, msize, dafp);
    }
#endif
    if((ms = fread(grf, 1, msize, dafp)) != msize) {
        error("daget", "fread failed.");
        fprintf(Stderr, "for graph = (%d, %d)\n", prc, gn);
        fprintf(Stderr, "size      =  %d, %d \n", ms, msize);
        dafail();
    }
}

#define INFOPROCBEGIN  -1
#define INFOPROCEND    -2

/*==================================================================*/
Private int    dainfo(nprc, ngrf, proc, graph)
int     nprc;
int     ngrf;
tproc  *proc;
tgraph *graph;
/*--------------------------------------------------------------------
 */
{
    static int buffsize  = 0;
    static int used      = 0;
    static GrInfo *buff  = NULL;
    static int msg       = 0;
    int     obuffsize;
    GrInfo *obuff = NULL;
    int     lilines;
    int     llegs;
    int     lpow;
    int     i;

    /* only for 1 loop */
    if(proc->loop > 1) {
        if(msg == 0) {
            fprintf(Stderr, "+++ dainfo : loop = %d : skip\n", proc->loop);
            msg = 1;
        }
        grinfo[nprc] = NULL;
        return 0;
        /* error("dainfo", "higher loop");              */
        /* fprintf(Stderr, "loop = %d\n", proc->loop);  */
        /* dafail();                                    */
    }
    if(ngrf == INFOPROCBEGIN) {
        used = 0;
        return used;
    } else if(ngrf == INFOPROCEND) {
        grinfo[nprc] = (GrInfo*) malloc(used*sizeof(GrInfo));
        if(buff == NULL) {
            error("dainfo", "memory");
            fprintf(Stderr, "nprc = %d, ngrf = %d\n", nprc, ngrf);
            dafail();
        }
        memcpy((void*)grinfo[nprc], (void*)buff, used*sizeof(GrInfo));
        return used;
    }
    if(ngrf >= buffsize) {
        obuffsize = buffsize;
        if(buffsize <= 0) {
            buffsize = INFOBUFFSIZE;
        }
        while(ngrf >= buffsize) {
            buffsize *= 2;
        }
        obuff = buff;
        buff  = (GrInfo*) malloc(buffsize*sizeof(GrInfo));
        if(buff == NULL) {
            error("dainfo", "memory");
            fprintf(Stderr, "nprc = %d, ngrf = %d\n", nprc, ngrf);
            dafail();
        }
        memcpy((void*)buff, (void*)obuff, used*sizeof(GrInfo));
        free(obuff);
    }
    for (i = used; i < ngrf; i++) {
        buff[i].gid      = -1;
        buff[i].gilil    = -1;
        buff[i].gillegs  = -1;
        buff[i].gilpow   = -1;
    }
    illoop(&lilines, &llegs, &lpow);
    buff[ngrf].gid     = graph->grid;
    buff[ngrf].gilil   = lilines;
    buff[ngrf].gillegs = llegs;
    buff[ngrf].gilpow  = lpow;
    /* counter term */
    if(buff[ngrf].gilil < 1) {
        buff[ngrf].gidiv = 0;
    /* count power of loop momentum */
    } else if(4+buff[ngrf].gilpow-2*buff[ngrf].gilil < 0) {
        buff[ngrf].gidiv = 0;
    } else {
        buff[ngrf].gidiv = 1;
    }

    used = max(used, ngrf+1);
    return used;
}

/*==================================================================*/
Public  int    dasave(fngrf, proct, procg)
string  fngrf;
int     proct[], procg[];
/*--------------------------------------------------------------------
 *  Read grf file and save to ranodm access file.
 *  This returns the number of processes.
 *
 *  Arguments:
 *    fngrf : input  : File name of generated graphs.
 *    proct : output : The table of loops of processes.
 *    procg : output : The table of the numbers of graphs of
 *                     processes.
 */
{
    int       nprc = -1;
    int       ngrf, subgrid, nloop, subgr;
    tgraph   *graph;
    tproc     procd;
    tproc    *proc = &procd;

#if DBG
    int prp = 0;
#endif

    if(!rginit(fngrf)) {
        fprintf(Stderr, "rginit failed\n");
        exit(16);
    }

    if(dafp == NULL) {
#ifdef SONY_NEWS
        sprintf(fname, "dagrf.tmp");
        if((dafp = fopen(fname, "w+")) == NULL) {
            error("daproc", "temporary file cannot open");
            dafail();
        }
#else
        /* open temporary file */
        if((dafp = tmpfile()) == NULL) {
            error("daproc", "temporary file cannot open");
            dafail();
        }
#endif
    }

    while (rgproc(proc)) {
        proc->mult   = False;
        proc->bgraph = MAXBUFF;
        proc->proc1  = -1;
        proc->proc2  = -1;
        nprc  = daproc(proc);
        graph = datempg(nprc);
        dainfo(nprc, INFOPROCBEGIN, proc, NULL);
#if DBG
        prp = 0;
#endif

        ngrf = 0;
        while (rgraph(graph)) {
            if((nloop = dagl4ini()) > 0) {
                subgrid = 1;
            } else {
                subgrid = 0;
            }
            for(subgr = 0; subgr < nloop; subgr++) {
                graph->subid = subgrid;
                if(dawrite(nprc, graph) < 0) {
                    error("dasave", "negative graph number");
                    dafail();
                }
                if(subgr < nloop-1) {
                    dagl4nxtgr();
                    subgrid++;
                }

                ilmom();
                dainfo(nprc, ngrf, proc, graph);
                ngrf++;
            }
        }
        proc->graphs = ngrf;
        proct[nprc] = proc->loop;
        procg[nprc] = proc->graphs;

        prcgraphs[nprc] = dainfo(nprc, INFOPROCEND, proc, NULL);

        free(graph);
    }

    rgclose();

    return (nprc+1);
}

/*==================================================================*/
Public  int    dasave0(fngrf, proct, procg)
string  fngrf;
int     proct[], procg[];
/*--------------------------------------------------------------------
 *  Read grf file and save to ranodm access file.
 *  This returns the number of processes.
 *
 *  Arguments:
 *    fngrf : input  : File name of generated graphs.
 *    proct : output : The table of loops of processes.
 *    procg : output : The table of the numbers of graphs of
 *                     processes.
 */
{
    expandgl4 = False;

    return dasave(fngrf, proct, procg);
}

/*==================================================================*/
Public  int    damproc(prc1, prc2)
int prc1, prc2;
/*--------------------------------------------------------------------
 *  Create a process for multiplication off graphs.
 */
{
    tproc  procd;
    tproc *proc = &procd;
    int    next = pctbl[prc1].nextern, io;

    if(!dachkext(prc1, prc2)) {
        error("damproc", "Inconsistent process for multiplication");
        fprintf(Stderr, "processes : %d, %d\n", prc1, prc2);
        dafail();
    }

    proc->mult     = PC_MULT;
    proc->procid   = nprocs;
    proc->graphs   = 1;
    proc->bgraph   = 1;
    proc->proc1    = prc1;
    proc->proc2    = prc2;
    proc->nextern  = 0;

    /* total loop of multiplied graph */
    proc->loop     = pctbl[prc1].loop  + pctbl[prc2].loop + next - 1;

    /* consider two-point function of external particle has order 1.*/
    proc->order    = pctbl[prc1].order + pctbl[prc2].order + next;
    proc->maxnodes = pctbl[prc1].maxnodes + pctbl[prc2].maxnodes;
    proc->maxlegs  = pctbl[prc1].maxlegs  + pctbl[prc2].maxlegs ;
    proc->msize    = pctbl[prc1].msize    + pctbl[prc2].msize;
    proc->oextern  = pctbl[prc1].nextern;

    proc->expand   = pctbl[prc1].expand && pctbl[prc2].expand;
    proc->onepi    = pctbl[prc1].onepi  && pctbl[prc2].onepi;
    proc->gm2      = False;
    if(pctbl[prc1].kinem == NULL) {
        proc->kinem = pctbl[prc2].kinem;
    } else if(pctbl[prc2].kinem == NULL) {
        proc->kinem = pctbl[prc1].kinem;
    } else if(strcmp(pctbl[prc1].kinem, pctbl[prc2].kinem) == 0) {
        proc->kinem = pctbl[prc1].kinem;
    } else {
        fprintf(Stderr, "*** damproc : different kinematics\n");
        fprintf(Stderr, "    process 1 : \"%s\"\n",pctbl[prc1].kinem);
        fprintf(Stderr, "    process 2 : \"%s\"\n",pctbl[prc2].kinem);
        exit(1);
    }
    if(pctbl[prc1].pprefix == NULL) {
        proc->pprefix = pctbl[prc2].pprefix;
    } else if(pctbl[prc2].pprefix == NULL) {
        proc->pprefix = pctbl[prc1].pprefix;
    } else if(strcmp(pctbl[prc1].pprefix, pctbl[prc2].pprefix) == 0) {
        proc->pprefix = pctbl[prc1].pprefix;
    } else {
        fprintf(Stderr, "*** damproc : different prefix\n");
        fprintf(Stderr, "    process 1 : \"%s\"\n",pctbl[prc1].pprefix);
        fprintf(Stderr, "    process 2 : \"%s\"\n",pctbl[prc2].pprefix);
        exit(1);
    }

    for(io = 0; io < nsordtbl; io++) {
        proc->sorder[io] = pctbl[prc1].sorder[io]
                         + pctbl[prc2].sorder[io];
    }
    proc->qcd      = pctbl[prc1].qcd + pctbl[prc2].qcd;

    cprocess = proc;

    return daproc(proc);
}

/*==================================================================*/
Public  tgraph *damult0(prc, grf1, grf2, gid1, gid2)
int     prc;
int     grf1, grf2;
long    *gid1, *gid2;
/*--------------------------------------------------------------------
 *  Multiply two graphs : grfl1*conjg(grfl2).
 */
{
    tproc  *proc;
    tgraph *g1, *g2;
    tgraph *go;
    int     p1, p2, next;
    int     vi, vo, voffs, ioffs, vt, nl, lg, lgo;
    int     pt, pn, same, nlg;

    cprocess = proc = pctbl + prc;

    if(proc->mult != PC_MULT) {
        error("damult", "process is not for multiplication.");
        fprintf(Stderr, "process = %d\n", prc);
        dafail();
    }

    /* input graphs */
    p1 = proc->proc1;
    p2 = proc->proc2;
    g1 = dareadb(p1, grf1, 0);
    g2 = dareadb(p2, grf2, 1);
    *gid1 = g1->grid;
    *gid2 = g2->grid;

    next = pctbl[p1].nextern;

    /* output graph */
    cgraph = go = danewg(prc, 0);
    rgigraph(proc, go); /* set basen and offsl */
    go->grid   = 1;
    go->gtype  = 1;
    go->nnodes = nenodes(g1) + nenodes(g2) - next;
    go->ngl4   = nengl4(g1) + nengl4(g2);
    go->gbextp = 0L;
    go->gbcnsv = 0L;
    go->gbelem = 0L;
    go->sfactor = g1->sfactor * g2->sfactor;
    go->conjsrt = nenodes(g1);

    vo = 0;
    nl = 0;
    ioffs = 0;
    for(vi = 0; vi < nenodes(g1); vi++) {
        ioffs += nelegs(g1, vi);

        if(netype(g1, vi) == NDEXTERN) {
            rgsetv(proc, go, vo, 2, &nl);
            ndtype(vo)   = NDONSHELL;
            ndlegs(vo)   = 2;
            ndloop(vo)   = 0;
            ndltype(vo)  = NDLTREE;
            ndiorf(vo)   = neiorf(g1, vi);
            ndgl4(vo)    = 0;
            ndparticle(vo, 0) = neparticle(g1, vi, 0);
            ndnextn(vo, 0)    = nenextn(g1, vi, 0);
            ndiline(vo, 0)    = neiline(g1, vi, 0);
            vo++;
        } else {
            rgsetv(proc, go, vo, nelegs(g1, vi), &nl);
            ndtype(vo)   = netype(g1, vi);
            ndlegs(vo)   = nelegs(g1, vi);
            ndloop(vo)   = neloop(g1, vi);
            ndltype(vo)  = neltype(g1, vi);
            ndgl4(vo)    = negl4(g1, vi);
            ndvertex(vo) = nevertex(g1, vi);
            for(lg = 0; lg < nelegs(g1, vi); lg++) {
                ndparticle(vo, lg) = neparticle(g1, vi, lg);
                ndnextn(vo, lg)    = nenextn(g1, vi, lg);
                ndiline(vo, lg)    = neiline(g1, vi, lg);
            }
            vo++;
        }
    }

    voffs = vo;
    ioffs = ioffs / 2;
    for(vi = 0; vi < nenodes(g2); vi++) {
        if(netype(g2, vi) == NDEXTERN) {
            vt = vi;
            ndparticle(vt, 1)
                = PCONJG(neparticle(g2, vi, 0));
            ndnextn(vt, 1)
                = NCONJG(nenextn(g2, vi, 0),voffs,next);
            ndiline(vt, 1)
                = ICONJG(neiline(g2, vi, 0),ioffs);
        } else {
            rgsetv(proc, go, vo, nelegs(g2, vi), &nl);
            ndtype(vo)   = netype(g2, vi);
            ndlegs(vo)   = nelegs(g2, vi);
            ndloop(vo)   = neloop(g2, vi);
            ndltype(vo)  = neltype(g2, vi);
            ndgl4(vo)    = negl4(g2, vi);
            ndvertex(vo) = nevertex(g2, vi);

            /* exchange Dirac fermion */
            /* alos for Majorana */
            pt   = ptable[abs(neparticle(g2, vi, 0))]->ptype;
            same = False;
            nlg  = nelegs(g2, vi);
            for(lg = 0; lg < nlg; lg++) {
                if(lg < nlg - 1) {
                    pn = ptable[abs(neparticle(g2, vi, lg+1))]->ptype;
                } else {
                    pn = P_UNDEF - 1;
                }

#ifdef MAJORANA
                if(pt == P_DIRAC || pt == P_MAJORANA) {
                    if(same) {
                        lgo = lg - 1;
                        same = False;
                    } else {
                        lgo = lg + 1;
                        same = True;
                    }
                } else {
                    lgo = lg;
                    same = False;
                }
#else
                if(pt == P_DIRAC) {
                    if(same) {
                        lgo = lg - 1;
                        same = False;
                    } else if(pt == pn) {
                        lgo = lg + 1;
                        same = True;
                    } else {
                        error("damult","illegal conjugation of fermion vertex");
                        fprintf(Stderr,"node = %d, leg = %d\n", vi, lg);
                        dafail();
                    }
                } else {
                    lgo = lg;
                    same = False;
                }
#endif

                pt = pn;

                ndparticle(vo, lgo)
                    = PCONJG(neparticle(g2, vi, lg));
                ndnextn(vo, lgo)
                    = NCONJG(nenextn(g2, vi, lg),voffs,next);
                ndiline(vo, lgo)
                    = ICONJG(neiline(g2, vi, lg),ioffs);
            }
            if(same) {
                error("damult","illegal number of fermions in a vertex");
                fprintf(Stderr,"node = %d\n", vi);
                dafail();
            }
            vo++;
        }
    }

    cgraph->nilines = ilmake(True);
#ifdef MAJORANA
    go->sfactor = abs(go->sfactor)*flmfact();
#endif

    return go;
}

/*==================================================================*/
Public  tgraph *damult(prc, grf1, grf2, gid1, gid2)
int     prc;
int     grf1, grf2;
long    *gid1, *gid2;
/*--------------------------------------------------------------------
 *  read a graph from a random access file
 *  Auxiliary vertices are connected.
 *  return the graph.
 */
{
    tgraph  *grf;
    tproc   *prcp;

    grf  = damult0(prc, grf1, grf2, gid1, gid2);
    prcp = danproc(prc);
    rgcomposit(prcp, grf);

    return grf;
}

/*==================================================================*/
Public  tgraph *damultn(prc, grf1, grf2, gid1, gid2)
int     prc;
int     grf1, grf2;
long   *gid1, *gid2;
/*--------------------------------------------------------------------
 *  Multiply two graphs : grfl1*conjg(grfl2).
 */
{
    tproc  *proc;
    tgraph *g1, *g2;
    tgraph *go;
    int     p1, p2, next;
    int     vi, vo, voffs, ioffs, vt, nl, lg, lgo;
    int     pt, pn, same, nlg;

    cprocess = proc = pctbl + prc;

    if(proc->mult != PC_MULT) {
        error("damultn", "process is not for multiplication.");
        fprintf(Stderr, "process = %d\n", prc);
        dafail();
    }

    /* input graphs */
    p1 = proc->proc1;
    p2 = proc->proc2;
    g1 = dareadb(p1, grf1, 0);
    g2 = dareadb(p2, grf2, 1);
    *gid1 = g1->grid;
    *gid2 = g2->grid;

    next = pctbl[p1].nextern;

    /* output graph */
    cgraph = go = datempg(prc);
    rgigraph(proc, go); /* set basen and offsl */
    go->grid   = 1;
    go->gtype  = 1;
    go->nnodes = nenodes(g1) + nenodes(g2) - next;
    go->ngl4   = nengl4(g1) + nengl4(g2);
    go->gbextp = 0L;
    go->gbcnsv = 0L;
    go->gbelem = 0L;
    go->sfactor = g1->sfactor * g2->sfactor;
    go->conjsrt = nenodes(g1);

    vo = 0;
    nl = 0;
    ioffs = 0;
    for(vi = 0; vi < nenodes(g1); vi++) {
        ioffs += nelegs(g1, vi);

        if(netype(g1, vi) == NDEXTERN) {
            rgsetv(proc, go, vo, 2, &nl);
            ndtype(vo)   = NDONSHELL;
            ndlegs(vo)   = 2;
            ndloop(vo)   = 0;
            ndltype(vo)  = NDLTREE;
            ndiorf(vo)   = neiorf(g1, vi);
            ndgl4(vo)    = 0;
            ndparticle(vo, 0) = neparticle(g1, vi, 0);
            ndnextn(vo, 0)    = nenextn(g1, vi, 0);
            ndiline(vo, 0)    = neiline(g1, vi, 0);
            vo++;
        } else {
            rgsetv(proc, go, vo, nelegs(g1, vi), &nl);
            ndtype(vo)   = netype(g1, vi);
            ndlegs(vo)   = nelegs(g1, vi);
            ndloop(vo)   = neloop(g1, vi);
            ndltype(vo)  = neltype(g1, vi);
            ndgl4(vo)    = negl4(g1, vi);
            ndvertex(vo) = nevertex(g1, vi);
            for(lg = 0; lg < nelegs(g1, vi); lg++) {
                ndparticle(vo, lg) = neparticle(g1, vi, lg);
                ndnextn(vo, lg)    = nenextn(g1, vi, lg);
                ndiline(vo, lg)    = neiline(g1, vi, lg);
            }
            vo++;
        }
    }

    voffs = vo;
    ioffs = ioffs / 2;
    for(vi = 0; vi < nenodes(g2); vi++) {
        if(netype(g2, vi) == NDEXTERN) {
            vt = vi;
            ndparticle(vt, 1)
                = PCONJG(neparticle(g2, vi, 0));
            ndnextn(vt, 1)
                = NCONJG(nenextn(g2, vi, 0),voffs,next);
            ndiline(vt, 1)
                = ICONJG(neiline(g2, vi, 0),ioffs);
        } else {
            rgsetv(proc, go, vo, nelegs(g2, vi), &nl);
            ndtype(vo)   = netype(g2, vi);
            ndlegs(vo)   = nelegs(g2, vi);
            ndloop(vo)   = neloop(g2, vi);
            ndltype(vo)  = neltype(g2, vi);
            ndgl4(vo)    = negl4(g2, vi);
            ndvertex(vo) = nevertex(g2, vi);

            pt   = ptable[abs(neparticle(g2, vi, 0))]->ptype;
            same = False;
            nlg  = nelegs(g2, vi);
            for(lg = 0; lg < nlg; lg++) {
                if(lg < nlg - 1) {
                    pn = ptable[abs(neparticle(g2, vi, lg+1))]->ptype;
                } else {
                    pn = P_UNDEF - 1;
                }

                if(same) {
                    lgo = lg - 1;
                    same = False;
                    printf("damultn:same1: lg=%d, lgo=%d\n", lg, lgo);
                } else if(pt == pn) {
                    lgo = lg + 1;
                    same = True;
                    printf("damultn:same0: lg=%d, lgo=%d\n", lg, lgo);
                } else {
                    lgo = lg;
                    same = False;
                }
                pt = pn;

                ndparticle(vo, lgo)
                    = PCONJG(neparticle(g2, vi, lg));
                ndnextn(vo, lgo)
                    = NCONJG(nenextn(g2, vi, lg),voffs,next);
                ndiline(vo, lgo)
                    = ICONJG(neiline(g2, vi, lg),ioffs);
            }
            vo++;
        }
    }

    cgraph->nilines = ilmake(True);

#ifdef MAJORANA
    go->sfactor = abs(go->sfactor)*flmfact();
#endif

    return go;
}

/*==================================================================*/
Private bool    dachkext(prc1, prc2)
int     prc1, prc2;
/*--------------------------------------------------------------------
 *  Check if multiplication is possible or not.
 */
{
    tproc *p1, *p2;
    tgraph *g1, *g2;
    int   ext, n;

    p1 = &(pctbl[prc1]);
    p2 = &(pctbl[prc2]);

    ext = p1->nextern;
    if(ext != p2->nextern) {
        return False;
    }

    if(dangrf(prc1) < 1 || dangrf(prc2) < 1) {
        return False;
    }

    g1 = dareadb(prc1, 0, 0);
    g2 = dareadb(prc2, 0, 1);
    for(n = 0; n < ext; n++) {
        if(netype(g1, n) != NDEXTERN) {
            return False;
        } else if(netype(g2, n) != NDEXTERN) {
            return False;
        } else if(neiorf(g1, n) != neiorf(g2, n)) {
            return False;
        } else if(nename(g1, n) != nename(g2, n)) {
            return False;
        }
    }
    return True;
}

/*==================================================================*/
Private int     dagl4ini(NoArg)
/*--------------------------------------------------------------------
 *  Initialize for loop of expanding 4 gluon vertices.
 */
{
    node nd;
    int  ngl4 = 0;
    int  lpgl4 = 1;

    if(ndngl4() < 1) {
        return 1;
    }
    if(!expandgl4) {
        return 1;
    }

    for(nd = 0; nd < ndnodes(); nd++) {
        if(ndgl4(nd) > 0) {
            gl4tbl[ngl4] = nd;
            gl4pos[ngl4] = 0;
            ngl4++;
            ndgl4(nd) = 1;
            lpgl4 *= 3;
        }
    }

#if DBG
    if(ngl4 != ndngl4()) {
        fprintf(Stderr, "*** dagl4ini : ngl4=%d != ndngl4()=%d\n",
                ngl4, ndngl4());
        prdump();
        exit(1);
    }
#endif

    return lpgl4;
}

/*==================================================================*/
Private Void    dagl4nxtgr(NoArg)
/*--------------------------------------------------------------------
 * 2004/10/10 : TK rewritten
 *
 *  Get next configuration of expanding 4 gluon vertices.
 */
{
    node nd;
    int  k, n, lg, part, nextn, iline, nmod;

    nmod = 0;
    for(n = 0; n < ndngl4(); n++) {
        if(gl4pos[n] == 2) {
            gl4pos[n] = 0;
            gl4mod[nmod++] = n;
        } else {
            gl4pos[n]++;
            gl4mod[nmod++] = n;
            break;
        }
    }
    if(nmod == 0 || n >= ndngl4()) {
        fprintf(Stderr, "*** dalg4nxtgr : illegal contorl\n");
        exit(FAIL);
    }
    for(k = 0; k < nmod; k++) {
        n  = gl4mod[k];
        nd = gl4tbl[n];
        if(ndlegs(nd) != 4) {
            fprintf(Stderr, "*** dalg4nxtgr : illegal legs != 4, nd=%d\n", nd);
            exit(FAIL);
        }
        part  = ndparticle(nd, 1);
        nextn = ndnextn(nd, 1);
        iline = ndiline(nd, 1);
        for(lg = 1; lg < 3; lg++) {
            ndparticle(nd, lg) = ndparticle(nd, lg+1);
            ndnextn(nd, lg)    = ndnextn(nd, lg+1);
            ndiline(nd, lg)    = ndiline(nd, lg+1);
        }
        ndparticle(nd, 3) = part;
        ndnextn(nd, 3)    = nextn;
        ndiline(nd, 3)    = iline;
    }
}

/*==================================================================*/
Public  int    dapproc(prc)
int prc;
/*--------------------------------------------------------------------
 *  Create a process for projection.
 */
{
    tproc  procd;
    tproc *proc = &procd;
    int    next = pctbl[prc].nextern, io;

    proc->mult     = PC_PROJ;
    proc->procid   = nprocs;
    proc->graphs   = 1;
    proc->bgraph   = 1;
    proc->proc1    = prc;
    proc->proc2    = -1;
    proc->nextern  = 0;

    /* total loop of multiplied graph */
    proc->loop     = pctbl[prc].loop + next - 1;

    proc->order    = pctbl[prc].order;
    proc->maxnodes = pctbl[prc].maxnodes - next + 1;
    proc->maxlegs  = pctbl[prc].maxlegs;
    proc->msize    = pctbl[prc].msize;
    proc->oextern  = pctbl[prc].nextern;

    proc->expand   = pctbl[prc].expand;
    proc->onepi    = pctbl[prc].onepi;
    proc->gm2      = pctbl[prc].gm2;
    proc->kinem    = pctbl[prc].kinem;
    proc->pprefix  = pctbl[prc].pprefix;

    for(io = 0; io < nsordtbl; io++) {
        proc->sorder[io] = pctbl[prc].sorder[io] ;
    }
    proc->qcd      = pctbl[prc].qcd;

    cprocess = proc;

    return daproc(proc);
}

/*==================================================================*/
Public  tgraph *daproj(prc, grf)
int     prc;
int     grf;
/*--------------------------------------------------------------------
 *  Add projection operator.
 */
{
    tproc  *proc;
    tgraph *gr;
    tgraph *go;
    int     pc, next;
    int     vi, vo, nl, lg;
    node    fp, fa;
    int     pa, pt;

    cprocess = proc = pctbl + prc;

    if(proc->mult != PC_PROJ) {
        error("daproj", "process is not for projection.");
        fprintf(Stderr, "process = %d\n", prc);
        dafail();
    }

    /* input graphs */
    pc = proc->proc1;
    gr = daread0(pc, grf);

    next = pctbl[pc].nextern;

    /* output graph */
    cgraph = go = datempg(prc);
    rgigraph(proc, go); /* set basen and offsl */
    go->grid    = gr->grid;
    go->gtype   = gr->gtype;
    go->nnodes  = nenodes(gr);
    go->ngl4    = nengl4(gr);
    go->gbextp  = gr->gbextp;
    go->gbcnsv  = gr->gbcnsv;
    go->gbelem  = gr->gbelem;
    go->sfactor = gr->sfactor;
    go->conjsrt = nenodes(gr);

    fp = fa = -1;
    for(vi = 0; vi < next; vi++) {
        if(netype(gr, vi) != NDEXTERN) {
            error("daproj", "not external line.");
            fprintf(Stderr, "node = %d\n", vi);
            dafail();
        }
        pa = neparticle(gr,vi,0);
        pt = ptable[abs(pa)]->ptype;
        if(pt == P_DIRAC) {
            if(pa > 0) {
                if(fp != -1) {
                    error("daproj", "too many dirac/majorana.");
                    fprintf(Stderr, "node = %d\n", vi);
                    dafail();
                }
                fp = vi;
            } else {
                if(fa != -1) {
                    error("daproj", "too many anti-dirac/majorana.");
                    fprintf(Stderr, "node = %d\n", vi);
                    dafail();
                }
                fa = vi;
            }
        } else if(pt == P_MAJORANA) {
            if(fp == -1) {
                fp = vi;
            } else if(fa != -1) {
                error("daproj", "too many anti-dirac/majorana.");
                fprintf(Stderr, "node = %d\n", vi);
                dafail();
            }
            fa = vi;
        }
    }
    if((fp == -1 && fa != -1) || (fp != -1 && fa == -1)) {
        error("daproj", "illegal external dirac/majorana.");
        fprintf(Stderr, "fp = %d, fa = %d\n", fp, fa);
        dafail();
    }

    /* create a node for the projection operator */
    vo = 0;
    nl = 0;
    rgsetv(proc, go, vo, next, &nl);
    ndtype(vo)        = NDPROJ;
    ndlegs(vo)        = next;
    ndloop(vo)        = 0;
    ndltype(vo)       = NDLTREE;
    ndvertex(vo)      = VNULL;
    ndgl4(vo)         = 0;

    if(fp != -1) {
        ndparticle(vo, 0) = neparticle(gr, fa, 0);
        ndnextn(vo, 0)    = nenextn(gr, fa, 0);
        ndiline(vo, 0)    = neiline(gr, fa, 0);

        ndparticle(vo, 1) = neparticle(gr, fp, 0);
        ndnextn(vo, 1)    = nenextn(gr, fp, 0);
        ndiline(vo, 1)    = neiline(gr, fp, 0);

        lg = 2;
    } else {
        lg = 0;
    }
    for(vi = 0; vi < next; vi++) {
        if(vi == fp || vi == fa) {
            continue;
        }
        ndparticle(vo, lg) = neparticle(gr, vi, 0);
        ndnextn(vo, lg)    = nenextn(gr, vi, 0);
        ndiline(vo, lg)    = neiline(gr, vi, 0);
        lg++;
    }

    /* other vertices */
    vo++;
    for(vi = next; vi < nenodes(gr); vi++) {

        if(netype(gr, vi) != NDEXTERN) {
            rgsetv(proc, go, vo, nelegs(gr, vi), &nl);
            ndtype(vo)   = netype(gr, vi);
            ndlegs(vo)   = nelegs(gr, vi);
            ndloop(vo)   = neloop(gr, vi);
            ndltype(vo)  = neltype(gr, vi);
            ndgl4(vo)    = negl4(gr, vi);
            ndvertex(vo) = nevertex(gr, vi);
            for(lg = 0; lg < nelegs(gr, vi); lg++) {
                ndparticle(vo, lg) = neparticle(gr, vi, lg);
                if(nenextn(gr, vi, lg) < next) {
                    ndnextn(vo, lg)    = 0;
                } else {
                    ndnextn(vo, lg)    = nenextn(gr, vi, lg);
                }
                ndiline(vo, lg)    = neiline(gr, vi, lg);
            }
            vo++;
        }
    }
    ndnodes() = vo;

    cgraph->nilines = ilmake(True);

    return go;
}

/*==================================================================*/
Public  int     dan888c(sgn)
int     *sgn;
/*--------------------------------------------------------------------
 *  Return the number of vertices of color 8-8-8-[1-...] coupling
 *  of the conjugated part of a multiplied graph.
 *
 *  It is used to calculate the sign produced in
 *
 *      conjg(-i f_^{abc}) = (-1)*(-i f_^{abc})
 *
 *  where f is the structure constant of color SU(3).
 */
{
    int n, lg;
    int n8leg, n888 = 0;

    *sgn = 1;
    if(cprocess->mult != PC_MULT) {
        return n888;
    }
    for(n = cgraph->conjsrt; n < cgraph->nnodes; n++) {
        if(ndlegs(n) < 3) {
            continue;
        }
        n8leg = 0;
        for(lg = 0; lg < ndlegs(n); lg++) {
            if(ptable[abs(ndparticle(n, lg))]->color == 8) {
                n8leg++;
            }
        }
        if(n8leg == 3) {
            n888++;
            *sgn *= -1;
        }
    }
    return n888;
}

/*==================================================================*/
Public  Void    dainfodmp(FILE *fp)
/*--------------------------------------------------------------------
 *  */
{
    int p, g;

    for(p = 0; p < nprocs; p++) {
        fprintf(fp, "process %d,  %ld graphs\n", p, prcgraphs[p]);
        fprintf(fp, "  grid, liline,  llegs,   lpow,    div\n");
        for(g = 0; g < prcgraphs[p]; g++) {
            fprintf(fp, "%6d, %6d, %6d, %6d, %6d\n",
                    grinfo[p][g].gid, grinfo[p][g].gilil,
                    grinfo[p][g].gillegs, grinfo[p][g].gilpow,
                    grinfo[p][g].gidiv);
        }
    }
}
