/*********************************************************************
 *  File : divide.c
 *
 *  Take "k" stones from "n" boxes, where i-th box keeps "p[i]" stones.
 *  This package lists up all possible configurations.
 *
 *  An sample main function is as the following:
 *
 *    int main(void)
 *    {
 *    #define N 4
 *
 *        int n=N, k=4;
 *        static int p[N]={3,1,2,4}, q[N];
 *
 *        printf("*** recursive version ***\n");
 *        ndvconf = 0;
 *        dvdump(n, p);
 *
 *        dvinitr(n, p, k, q);
 *
 *        printf("*** non-recursive version ***\n");
 *        ndvconf = 0;
 *        dvdump(n, p);
 *
 *        dvinit(n, p, k, q);
 *        dvdump(n, q);
 *        while(dvnext(n, p, q)) {
 *            dvdump(n, q);
 *        }
 *
 *        return 0;
 *    }
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdio.h>
#include "common.h"
#include "divide.h"

/*--------------------------------------------------------------------
 * local variables
 */
#define MAXN 64
Local   int  ndvconf = 0;

/*==================================================================*/
Public  Void  dvinit(n, p, k, q)
int n;
int p[];
int k;
int q[];
/*--------------------------------------------------------------------
 * generate configuration taking k elements from p[].
 */
{
    int i, j;

    j = k;
    for(i = 0; i < n; i++) {
        q[i] = (j < p[i]) ? j : p[i];
        j   -= q[i];
    }
}

/*==================================================================*/
Public  bool  dvnext(n, p, q)
int n;
int p[];
int q[];
/*--------------------------------------------------------------------
 * generate configuration taking k elements from p[].
 */
{
    int i, imin, j, ips, iqs;

    ips = p[n-1];
    iqs = q[n-1];
    for(i = n-2; i >= 0; i--) {
        imin = iqs + q[i] - ips;
        if(imin < 0) {
            imin = 0;
        }
        if(q[i] > imin) {
            q[i]--;
            iqs++;
            for(j = i+1; j < n; j++) {
                q[j] = (iqs < p[j]) ? iqs : p[j];
                iqs -= q[j];
            }
            return True;
        }
        ips  += p[i];
        iqs  += q[i];
    }
    return False;
}

/*==================================================================*/
Public  Void  dvinitr(n, p, k, q)
int n;
int p[];
int k;
int q[];
/*--------------------------------------------------------------------
 * generate configuration taking k elements from p[].
 */
{
    dvnextr(n, p, q, 0, k);
}

/*==================================================================*/
Public  Void  dvnextr(n, p, q, l, k)
int n;
int p[];
int q[];
int l;
int k;
/*--------------------------------------------------------------------
 * generate configuration taking k elements from p[].
 */
{
    int i, imin, imax;

    if(k < 1) {
        for(i = l; i < n; i++) {
            q[i] = 0;
        }
        dvdump(n, q);
        return;
    }
    if(l >= n) {
        printf("****");
        dvdump(n, q);
        return;
    }

    imax = (k < p[l]) ? k : p[l];
    imin = k;
    for(i = l+1; i < n; i++) {
        imin -= p[i];
    }
    if(imin < 0) {
        imin = 0;
    }

    for(i = imax; i >= imin; i--) {
        q[l] = i;
        k -= i;
        dvnextr(n, p, q, l+1, k);
        k += i;
    }
}

/*==================================================================*/
Public  Void  dvdump(n, p)
int n;
int p[];
/*--------------------------------------------------------------------
 * generate configuration taking k elements from p[].
 */
{
    int i;

    printf("%6d:",ndvconf);
    for(i = 0; i < n; i++) {
       printf("%4d", p[i]);
    }
    printf("\n");
    ndvconf++;
}
