/*********************************************************************
 *  File : ergrf.c
 *
 *  Error routines.
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "common.h"
#include "ergrf.h"
#include "dbg.h"

Extern  Void    prdump     ARG((NoArg));

/*
#define ABORT
*/

/*--------------------------------------------------------------------
 * local functions
 */
#if DBG
Private int     erfnfind   ARG((string));
#endif

/*--------------------------------------------------------------------
 * local variables
 */
Local   string erfunc, ermsg;

#if DBG
#define MAXTRCFUNCS 10
Local   int    erfnln = 0;
Local   string erfns[MAXTRCFUNCS];
Local   int    erfnl[MAXTRCFUNCS];
#endif

/*==================================================================*/
Public  Void    error(func, msg)
string func;
string msg;
/*--------------------------------------------------------------------
    Print error message.
*/
{
    erfunc = func;
    ermsg  = msg;
    fprintf(Stderr, " *err* %s : %s\n", func, msg);
    prdump();
}


/*==================================================================*/
Public  Void    erdbg(func, msg, ldbg)
string func;
string msg;
int    ldbg;
/*--------------------------------------------------------------------

    Print error message.
*/
{
#if DBG
    if(dbg==0) {
        return;
    } else if(dbg < ldbg) {
        return;
    }

    erfunc = func;
    ermsg  = msg;
    fprintf(Stderr, " *dbg* %s : %s\n", func, msg);
#endif
}

/*==================================================================*/
Public  Void    ermon(func, msg)
string func;
string msg;
/*--------------------------------------------------------------------
    Print error message.
*/
{
    erfunc = func;
    ermsg  = msg;
    fprintf(Stderr, " *mon* %s : %s\n", func, msg);
    prdump();
}

/*==================================================================*/
Public  Void    erfail(NoArg)
/*--------------------------------------------------------------------
    Print error message and exit(FAIL).
*/
{
    fprintf(Stderr, " *end* %s : %s\n", erfunc, ermsg);
/*
    {int i = 0; i = 1/i;}
*/
#ifdef ABORT
    abort();
#endif
    exit(FAIL);
}

/*==================================================================*/
Public  Void    ercont(NoArg)
/*--------------------------------------------------------------------
    Print error message.
*/
{
    fprintf(Stderr, " *end* %s : %s\n", erfunc, ermsg);
}

/*==================================================================*/
Public  Void    erenter(func)
string func;
/*--------------------------------------------------------------------
    Print trace massage for entering a function.
*/
{
#if DBG
    int fn;

    if(!trace) {
        return;
    }

    fn  = erfnfind(func);

    fprintf(Stderr, " *trc* %s : enter  : %d\n", func, erfnl[fn]++);

#endif
}

/*==================================================================*/
Public  Void    erreturn(func)
string func;
/*--------------------------------------------------------------------
    Print trace massage in returning from a function.
*/
{
#if DBG
    int fn;

    if(!trace) {
        return;
    }

    fn  = erfnfind(func);

    fprintf(Stderr, " *trc* %s : exit   : %d\n", func, --erfnl[fn]);

#endif
}

#if DBG
/*==================================================================*/
Private int     erfnfind(func)
string func;
/*--------------------------------------------------------------------
    Find trace function name.
*/
{
    int i, fn;

    for(i = 0, fn = -1; i < erfnln; i++) {
        if(strcmp(func, erfns[i]) == 0) {
            fn = i;
            break;
        }
    }
    if(fn < 0) {
        fn++;
        if(fn >= MAXTRCFUNCS) {
            error("erfnind", "too many trace functions");
            erfail();
        }
        erfns[fn] = (string) erstrdup(func);
        erfnl[fn] = 0;
    }

    return fn;
}
#endif

/*==================================================================*/
Public string  erstrdup(str)
string   str;
/*--------------------------------------------------------------------
 *   duplicate string.
 */
{
    string s;

    s = (string) malloc((strlen(str)+1)*sizeof(char));
    if(s == NULL) {
        fprintf(Stderr, "*err* pvstrcup: memory allocail error\n");
        fprintf(Stderr, "for %s\n", str);
        exit(1);
    }

    return strcpy(s, str);
}
