/*********************************************************************
 *  File : ilgrf.c
 *
 *  Tools for internal lines.
 *
 *      Analysis and alocation of momentum conservation.
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grdata.h"


/*--------------------------------------------------------------------
 * Parameter
 */
#ifdef OLD
#undef OLD
#endif

#define MOMCHK

/*--------------------------------------------------------------------
 * Local functions
 */
Private Void    ilalloc      ARG((int));
Private Void    ilndswap     ARG((node, int, int));
Private Void    ilbicon      ARG((NoArg));
Private int     ilbinit      ARG((NoArg));
Private Void    ilbend       ARG((NoArg));
Private Void    ilbisrch     ARG((node,node,int,int));
Private Void    ilexchg      ARG((NoArg));
Private Void    ilswap       ARG((int,bool));
Private Void    ildmpabs     ARG((FILE*,Ulong,string));
Private Void    ilencall     ARG((NoArg));

/*--------------------------------------------------------------------
 * local variables
 */
Local int  siltbl  = 0;
Local int  nilines = 0;
Local int  bicount;
Local int  bivdef[MAXNODES];
Local int  bivlow[MAXNODES];
Local int  *mconsv = NULL;
Local int  *loopm  = NULL;    /* list of loop momenta */
Local int  nloopm  = 0;

#define bidef(v)   bivdef[v]
#define bilow(v)   bivlow[v]
#define BINEW      (MAXNODES + 1)

/*==================================================================*/
Public  int     ilmake(mode)
bool    mode;
/*--------------------------------------------------------------------
 *    Construct internal line table for the current graph.
 *      items "ilnodes" and "illegs" are set here.
 *
 *    if(mode) {
 *        (set ndnextn(*,*) )
 *    }
 *------------------------------------------------------------------*/
{
    node nd;
    int  lg, il, ip, niln, n, f, l, tp, lo, li;

#if DBG
    if(dbg >= ildbg) {
        printf("ilmake(%d)\n", mode);
    }
#endif

    /* size of the table */
    niln = 0;
    for(nd = 0; nd < ndnodes(); nd++) {
        for(lg = 0; lg < ndlegs(nd); lg++) {
            niln = max(niln, ndiline(nd, lg));
        }
    }
    niln++;

    ilalloc(niln);
    /* clear the internal line table */
    for(il = 0; il < niln; il++) {
        iltable[il].ilnodes[0] = -1;
        iltable[il].ilnodes[1] = -1;
        iltable[il].iltype   = ILTREE;
        iltable[il].ilname   = 0;
        iltable[il].ilfline  = -1;
        iltable[il].ilfldir  = 0;
    }

    /* set ilnodes and illegs */
    for(nd = 0; nd < ndnodes(); nd++) {
        for(lg = 0; lg < ndlegs(nd); lg++) {
            if(mode) {
                ndnextn(nd, lg) = 0;
            }

            il = ndiline(nd, lg);
            /* 0 -->(particle)--> 1
             * this is modified later 
             */
            if(!cprocess->gm2 && ndparticle(nd, lg) < 0) {
                ip = 0;
                if(iltable[il].ilnodes[0] >= 0) {
                    error("ilmake", "too many legs for an i-line (1)");
                    fprintf(Stderr,"il = %d\n", il);
                    erfail();
                }
            } else {
                ip = 1;
                if(iltable[il].ilnodes[1] >= 0) {
                    if(iltable[il].ilnodes[0] >= 0) {
                        error("ilmake", "too many legs for an i-line (2)");
                        fprintf(Stderr,"il = %d\n", il);
                        erfail();
                    }
                    ip = 0;
                }
            }
            iltable[il].ilnodes[ip] = nd;
            iltable[il].illegs[ip]  = lg;

            if(ndtype(nd) == NDEXTERN) {
                iltable[il].iltype   =  ILEXTERN;
                iltable[il].ilname   =  0;
            } else if(ndtype(nd) == NDONSHELL) {
                iltable[il].iltype   =  ILEXTERN;
                iltable[il].ilname   =  0;
            } else if(ndtype(nd) == NDPROJ) {
                iltable[il].iltype   =  ILEXTERN;
                iltable[il].ilname   =  0;
            }
        }
    }

    /* check table.  iltbl[0] is not used.*/
    for(il = 1; il < niln; il++) {
        for(ip = 0; ip < 2; ip++) {
            if(iltable[il].ilnodes[ip] < 0) {
                error("ilmake", "too few legs for an i-line");
                fprintf(Stderr,"il = %d\n", il);
                erfail();
            }

            if(mode) {
                ndnextn(iltable[il].ilnodes[ip],iltable[il].illegs[ip])
                    = iltable[il].ilnodes[1-ip];
            }
        }
    }

    cgraph->nilines = niln;

    /* construct fermion line */
    flget(cgraph);

    for(f = 0; f < nflines; f++) {
        for(n = 0; n < fltable[f].length; n++) {
            nd = fltable[f].ndlist[n];
            lo = fltable[f].lglist[FL_OUT][n];
            li = fltable[f].lglist[FL_IN][n];
            tp = ndtype(nd);
            if(tp != NDPROJ && tp != NDONSHELL &&
               li >= 0 && lo >= 0 && lo < li)
            {
#if 0 /* ??? */
                if(!majop) {
                    error("ilmake", "illegal in/ount of f-line");
                    prdump();
                    fldump();
                    fprintf(Stderr, "f = %d, n = %d, nd = %d, (%d,%d)\n",
                            f, n, nd, li, lo);
                    erfail();
                }
#endif
                ilndswap(nd, li, lo);
                l  = li;
                li = lo;
                lo = l;
                /* TK 2000.05.18 the following two lines are added */
                fltable[f].lglist[FL_OUT][n] = lo;
                fltable[f].lglist[FL_IN][n]  = li;
            }
            tp = ptable[abs(ndparticle(n,lo))]->ptype;
            if(lo >= 0) {
                il = ndiline(nd, lo);
                iltable[il].ilfline = f;
                if(iltable[il].illegs[0] == lo) {
                    iltable[il].ilfldir  = -1;
                    if(tp == P_MAJORANA) {
                        ilswap(il, False);
                    }
                } else if(iltable[il].illegs[1] == lo) {
                    iltable[il].ilfldir  = 1;
                } else {
                    error("ilmake", "illegal f-line");
                    prdump();
                    fprintf(Stderr, "f = %d, n = %d, il = %d\n",
                            f, n, il);
                    erfail();
                }
            }
        }
    }

#if DBG
    if(dbg >= ildbg) {
        printf("ilmake(%d) : end\n", mode);
    }
#endif

    /* check for debugging */
    flcheck();

    return niln;
}
/*==================================================================*/
Private Void    ilalloc(niln)
int     niln;
/*--------------------------------------------------------------------
 * allocate memory for the internal line table
 *------------------------------------------------------------------*/
{
    int il, jl;

#if DBG
    if(dbg >= ildbg) {
        printf("ilalloc(%d)\n", niln);
    }
#endif

    if(niln < 1) {
        error("ilalloc", "0 i-lines");
        erfail();
    }

    /* allocate memory for the internal line table */
    if(siltbl < niln) {
        if(siltbl > 0) {
            free((AnyP)mconsv);
            free((AnyP)loopm);
            free((AnyP)ilxtable);
            free((AnyP)ndxtable);
            for(il = 0; il < siltbl; il++) {
                free((AnyP)iltable[il].ilmlist);
            }
            free((AnyP)iltable);
        }

        if((iltable = (tiline*) malloc(niln*sizeof(tiline)))
             == NULL) {
             error("ilalloc", "memory allocation failed.");
             erfail();
        }
        for(il = 0; il < niln; il++) {
            if((iltable[il].ilmlist
                 = (tmom*) malloc(niln*sizeof(tmom))) == NULL) {
                 error("ilalloc", "memory allocation failed.");
                 erfail();
            }
            for(jl = 0; jl < niln; jl++) {
                iltable[il].ilmlist[jl] = 0;
            }
        }
        if((mconsv = (int*) malloc(niln*sizeof(int))) == NULL) {
             error("ilalloc", "memory allocation failed.");
             erfail();
        }
        if((loopm = (tmom*) malloc(niln*sizeof(tmom))) == NULL) {
             error("ilalloc", "memory allocation failed.");
             erfail();
        }
        if((ilxtable = (int*) malloc(niln*sizeof(int))) == NULL) {
             error("ilalloc", "memory allocation failed.");
             erfail();
        }
        if((ndxtable = (int*) malloc(niln*sizeof(int))) == NULL) {
             error("ilalloc", "memory allocation failed.");
             erfail();
        }
        siltbl = niln;
    }
#if DBG
    if(dbg >= ildbg) {
        printf("ilalloc(%d): end\n", niln);
    }
#endif
}

/*==================================================================*/
Public  Void    ildump(NoArg)
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    int il, ip, ii, ce, nd, lg, p;
    tiline *ti;

    /* TK 2004.03.14 */
    if(cgraph == NULL) {
        printf("\n Null graph\n");
        return;
    }
    
    nilines = cgraph->nilines;

    printf("\n nilines = %d, siltbl = %d\n", nilines, siltbl);
    ildmpabs(stdout, cgraph->gbextp, " gbextp:");
    ildmpabs(stdout, cgraph->gbcnsv, ", gbcnsv:");
    printf("\n");

    if(nilines < 0) {
        return;
    }

    for(il = 1; il < min(nilines, siltbl); il++) {
        ti = iltable + il;
        printf("%2d  t:%d x:%d f:%d, d:%2d",
               il, ti->iltype, ti->ilname, ti->ilfline,
               ti->ilfldir);
        printf(" [");
        for(ip = 0; ip < 2; ip++) {
            nd = ti->ilnodes[ip];
            lg = ti->illegs[ip];
            p  = ndparticle(nd, lg);
            printf(" %2d(%d:%d)", nd, lg, p);
        }
        printf("]: ");
        for(ii = 1; ii < min(nilines, siltbl); ii++) {
            if((ce = ti->ilmlist[ii]) == -1) {
                printf("-%d", ii);
            } else if(ce == 1) {
                printf("+%d", ii);
            } else if(ce > 0) {
                printf("+%d*%d", ce, ii);
            } else if(ce < 0) {
                printf("%d*%d", ce, ii);
            }
        }
        printf("\n");
    }

    printf("ilxnumb = %d\n", ilxnumb);
    for(il = 0; il < ilxnumb;  il++) {
        printf(" il:%d [nd:%d]", ilxtable[il], ndxtable[il]);
    }
    printf("\n");
}

/*==================================================================*/
Public  int     ilnextl(n, l)
node    n;
int     l;
/*--------------------------------------------------------------------
 * determine momentum of internal lines.
 *------------------------------------------------------------------*/
{
    int il, j;
    node nv;

    nv = ndnextn(n, l);
    il = ndiline(n, l);
    j = 0;
    if(iltable[il].ilnodes[j] != nv) {
        j = 1;
    }
    return iltable[il].illegs[j];
}

/*==================================================================*/
Private Void    ilndswap(nd, l1, l2)
node    nd;
int     l1, l2;
/*--------------------------------------------------------------------
 *   Swap two legs l1 and l2 of node nd.
 */
{
    int t, i1, i2;

    t = ndnextn(nd, l1);
    ndnextn(nd, l1) = ndnextn(nd, l2);
    ndnextn(nd, l2) = t;

    t = ndparticle(nd, l1);
    ndparticle(nd, l1) = ndparticle(nd, l2);
    ndparticle(nd, l2) = t;

    i1 = ndiline(nd, l1);
    i2 = ndiline(nd, l2);
    ndiline(nd, l1) = i2;
    ndiline(nd, l2) = i1;

    t = ndilwk(nd, l1);
    ndilwk(nd, l1) = ndilwk(nd, l2);
    ndilwk(nd, l2) = t;

    if(iltable[i1].ilnodes[0] == nd &&
       iltable[i1].illegs[0] == l1) {
        iltable[i1].illegs[0] = l2;
    } else if(iltable[i1].ilnodes[1] == nd &&
               iltable[i1].illegs[1] == l1) {
        iltable[i1].illegs[1] = l2;
    } else {
        error("ilndswap", "illegal connection (i1)");
        fprintf(Stderr,"nd = %d, l1 = %d, l2 = %d, i1 = %d\n",
                nd, l1, l2, i1);
        erfail();
    }

    if(iltable[i2].ilnodes[0] == nd &&
       iltable[i2].illegs[0] == l2) {
        iltable[i2].illegs[0] = l1;
    } else if(iltable[i2].ilnodes[1] == nd &&
               iltable[i2].illegs[1] == l2) {
        iltable[i2].illegs[1] = l1;
    } else {
        error("ilndswap", "illegal connection (i2)");
        fprintf(Stderr,"nd = %d, l1 = %d, l2 = %d, i2 = %d\n",
                nd, l1, l2, i2);
        erfail();
    }
}

/*==================================================================*/
Public  Void    ilmom(NoArg)
/*--------------------------------------------------------------------
 * determine momentum of internal lines.
 *------------------------------------------------------------------*/
{
#if DBG
    if(dbg>=ildbg) {
        printf("ilmom: begin\n");
    }
#endif

    if(cprocess->loop < 0) {
        error("ilmom", "illegal number of loops.");
        erfail();
    }

    if(!cprocess->expand) {
        error("ilmom", "not expanded graph.");
        erfail();
    }

    /* determine loop momenta */
    cgraph->nilines = ilmake(True);

    nilines = cgraph->nilines;

    ilbicon();

    ilexchg();

    ilencall();

#if DBG
    if(dbg>=ildbg) {
        printf("ilmom: end\n");
    }
#endif
}

/*==================================================================*/
Private Void    ilbicon(NoArg)
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    int nd;

#if DBG
    if(dbg>=ildbg) {
        printf("ilbicon: begin (loop=%d, expand=%d)\n",
                cprocess->loop, cprocess->expand);
    }
#endif

    nd = ilbinit();

    ilbisrch(nd, -1, -1, 0);

    ilbend();

#if DBG
    if(dbg>=ildbg) {
        printf("ilbicon: end\n");
    }
#endif

    /* No check of the anoswer in the test version */
    return;
}

/*==================================================================*/
Private int     ilbinit(NoArg)
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    node v, eo, ee;
    int il, im, il0, il1, dir, l;

    bicount = 0;
    nloopm  = 0;
    ilxnumb = 0;

    for(v = 0; v < ndnodes(); v++) {
        bidef(v) = BINEW;
        bilow(v) = BINEW;
    }

#if DBG
    if(dbg >= ildbg) {
        printf("ilbinit: nilines = %d\n", nilines);
    }
#endif
    cgraph->gbextp = 0L;
    cgraph->gbcnsv = 0L;
    for(il = 0; il < nilines; il++) {
        iltable[il].iltype   = ILTREE;
        iltable[il].ilname   = 0;
        iltable[il].ilmabs   = 0L;
        iltable[il].ilmsgn   = 0L;
        for(im = 0; im < nilines; im++) {
            iltable[il].ilmlist[im] = 0;
        }
        loopm[il] = 0;
    }

    for(il = 0; il < nilines; il++) {
        mconsv[il] = 0;
    }

    ee = eo = -1;
    for(v = ndnodes() - 1; v >= 0; v--) {
        if(ndtype(v) == NDEXTERN) {
            eo = v;
            il0 = ndiline(v, 0);
            iltable[il0].iltype   =  ILEXTERN;
            iltable[il0].ilname   =  0;
            dir = - ndiorf(v);
            if(iltable[il0].ilnodes[1] == v) {
                dir = - dir;
            }
            if(dir < 0) {
                ilswap(il0, False);
            }
            iltable[il0].ilmlist[il0] = 1;

            if(ndiorf(v) == ELINITIAL) {
                mconsv[il0] = 1;
            } else {
                mconsv[il0] = -1;
            }

        } else if(ndtype(v) == NDONSHELL) {
            eo = v;

            il0 = ndiline(v, 0);
            iltable[il0].iltype   = ILEXTERN;
            iltable[il0].ilname   = 0;

            il1 = ndiline(v, 1);
            iltable[il1].iltype  = ILEXTERN;
            iltable[il1].ilname  = 0;

            dir = - ndiorf(v);
            if(iltable[il0].ilnodes[1] == v) {
                dir = - dir;
            }
            if(dir < 0) {
                ilswap(il0, False);
            }
            dir = - ndiorf(v);
            if(iltable[il1].ilnodes[1] == v) {
                dir = - dir;
            }
            if(dir < 0) {
                ilswap(il1, False);
            }

            iltable[il0].ilmlist[il0] = 1;
            iltable[il1].ilmlist[il0] = 1;

            if(ndiorf(v) == ELINITIAL) { /* ??? */
                mconsv[il0] = 1;
                mconsv[il1] = 1;
            } else {
                mconsv[il0] = -1;
                mconsv[il1] = -1;
            }

        } else if(ndtype(v) == NDPROJ) {
            eo = v;
            for(l = 0; l < ndlegs(v); l++) {
                il0 = ndiline(v, l);
                iltable[il0].iltype  = ILEXTERN;
                iltable[il0].ilname  = 0;
                dir = - ndiorf(v);
                if(iltable[il0].ilnodes[1] == v) {
                    dir = - dir;
                }
                if(dir < 0) {
                    ilswap(il0, False);
                }
                iltable[il0].ilmlist[il0] = 1;

                if(ndiorf(v) == ELINITIAL) {
                    mconsv[il0] = 1;
                } else {
                    mconsv[il0] = -1;
                }
            }
        }
    }

#if DBG
    if(dbg >= ildbg) {
        for(il = 0; il < nilines; il++) {
            printf("monsv[%d]=%d\n", il, mconsv[il]);
        }

        printf("ilbinit: end\n");
    }
#endif

    if(ee != -1) {
        return ee;
    } else if(eo != -1) {
        return eo;
    }

    fprintf(Stderr, "*** ilbinit : no external particles\n");
    return 0;
}

/*==================================================================*/
Private Void    ilbend(NoArg)
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    node nd;
    int  il;

    for(nd = 0; nd < cgraph->nnodes; nd++) {
        ndltype(nd) = 0;
    }

    for(il = 0; il < cgraph->nilines; il++) {
        if(iltable[il].iltype == ILLOOP ||
           iltable[il].iltype == ILDLOOP) {
            nd = iltable[il].ilnodes[0];
            ndltype(nd) = ndltype(nd) + 1;
            nd = iltable[il].ilnodes[1];
            ndltype(nd) = ndltype(nd) + 1;
        }
    }

    for(nd = 0; nd < cgraph->nnodes; nd++) {
        if(ndltype(nd) == ndlegs(nd)) {
            ndltype(nd) = NDLINSIDE;
        } else if(ndltype(nd) > 0) {
            ndltype(nd) = NDLTOUCH;
        } else {
            ndltype(nd) = NDLTREE;
        }
    }
}

/*==================================================================*/
Private Void    ilbisrch(v, ov, il, ilg)
node    v, ov;
int     il, ilg;
/*--------------------------------------------------------------------
   Search graph to fined bridge.

   Arguments:
       v   : input  : node to visit.
       ov  : input  : the last node.
       il  : input  : i-line number connecting ov and v
       ilg : input  : iltable[il].ilnodes[ilg] = v

   Returns
       whether the node v is in a loop or not.
 *------------------------------------------------------------------*/
{
    node w;
    int lg, dir, dirn, iln, ilgn, i;
    Ulong ib;

#if DBG
    if(dbg >= ildbg) {
        printf("*dbg* ilbisrch : begin %d -> %d\n", ov, v);
    }
#endif

    dir = ildir(ilg);   /* incoming to the node v */

    /* visit node "v" */
    bicount++;
    bidef(v)  = bicount;
    bilow(v)  = bicount;

    /* visit sons */
    for(lg = 0; lg < ndlegs(v); lg++) {
        if(ndiline(v, lg) == il) /* back to ov */
            continue;

        w = ndnextn(v, lg);

        iln = ndiline(v, lg);
        if(iltable[iln].ilnodes[0] == v) {
            ilgn = 1;
        } else {
            ilgn = 0;
        }
        dirn = ildir(ilgn);  /* incoming to the node w */

        /* self loop */
        if(w == v) {
            if(iltable[iln].illegs[0] != lg) {
                continue;
            }
            iltable[iln].iltype = ILLOOP;
            iltable[iln].ilmlist[iln] = 1;

            /* Feynman parameter */
            ilxtable[ilxnumb] = iln;
            ndxtable[ilxnumb] = v;
            iltable[iln].ilname = ++ilxnumb;
            loopm[nloopm++] = iln;
            continue;

        /* external particle */
        } else if(ndtype(w) == NDEXTERN) {

        /* squared external particle */
        } else if(ndtype(w) == NDONSHELL) {

        /* projection operator */
        } else if(ndtype(w) == NDPROJ) {

        /* edge (v, w) is a back edge */
        } else if(bidef(w) != BINEW) {
            if(bidef(v) > bidef(w)) {
                iltable[iln].iltype = ILLOOP;

                /* incoming to v */
                iltable[iln].ilmlist[iln] = dirn;

                ib = 1L << iln;

                /* Feynman parameter */
                ilxtable[ilxnumb] = iln;
                ndxtable[ilxnumb] = v;
                iltable[iln].ilname = ++ilxnumb;
                loopm[nloopm++] = iln;
            }
            bilow(v)  = min(bilow(v), bidef(w));

        } else {    /* new vertex */
            ilbisrch(w, v, iln, ilgn);
            bilow(v)  = min(bilow(v), bilow(w));
        }

        if(il > 0 && iltable[il].iltype != ILEXTERN) {
            for(i = 1; i < nilines; i++) {
                /* direction = - (vo->v)*(w->v) */
                iltable[il].ilmlist[i]
                    += dir*dirn * iltable[iln].ilmlist[i];
#if DBG
                if(dbg>=ildbg) {
                    printf("%d->%d: %d * %d(%d,%d)\n",
                           v,w,dir*dirn,iltable[iln].ilmlist[i],iln,i);
                }
#endif
            }
        }
#if DBG
        if(dbg >= ildbg) {
            printf("%d -> %d -> %d : %d\n", ov, v, w,
                   iltable[iln].iltype);
        }
#endif

    } /* of for */

#if DBG
        if(dbg >= ildbg) {
            printf("ilbisrch(%d, %d, %d, %d): bilow(v)=%d, bidef(ov)=%d\n",
                   v, ov, il, ilg,bilow(v),bidef(ov));
        }
#endif
    /* Feynman parameter */
    if(ov >= 0 && bilow(v) <= bidef(ov)) {
        ilxtable[ilxnumb]  = il;
        ndxtable[ilxnumb]  = ov;
        iltable[il].iltype = ILDLOOP;
        iltable[il].ilname = ++ilxnumb;
    }

#if DBG
    if(dbg >= ildbg) {
        if(ov >= 0) {
            printf("%d(%d) ->(%d)-> %d(%d)\n",
                   ov, bidef(ov), il, v, bilow(v));
        }
    }
#endif
}

/*==================================================================*/
Public  Void    iliconsv(NoArg)
/*--------------------------------------------------------------------
 * Momentum conservation of tree internal line.
 *------------------------------------------------------------------*/
{
    int il, ii, ij, ce, cs;
    int nextern = cprocess->nextern;
    bool cont;
#if DBG
    int tst = 0;
#endif

    cont = True;
    while(cont) {
#if DBG
        if(++tst > max(nilines*nilines, 2)) {
            fprintf(Stderr, "*** program loops at iliconsv\n");
            ildump();
            exit(1);
        }
#endif
        cont = False;
        for(il = 0; il < nilines; il++) {
            if(iltable[il].iltype == ILEXTERN) {
                continue;
            } else if(iltable[il].iltype == ILLOOP) {
                continue;
            }

            for(ii = 0; ii < nilines; ii++) {
                if(iltable[ii].iltype == ILEXTERN) {
                    continue;
                } else if(iltable[ii].iltype == ILLOOP) {
                    continue;
                } else if((ce = iltable[ii].ilmlist[il]) == 0) {
                    continue;
                }
#if DBG
                printf("iliconsv: il = %d, ii = %d\n", il, ii);
                ildump();
#endif
                iltable[ii].ilmlist[il] = 0;
                for(ij = 0; ij < nilines; ij++) {
                    iltable[ii].ilmlist[ij] +=
                        ce*iltable[il].ilmlist[ij];
                }
                cont = True;
            }
        }
    }

    for(il = 1; il < nilines; il++) {
        if(iltable[il].iltype == ILEXTERN) {
            continue;
        } else if(iltable[il].iltype == ILLOOP) {
            continue;
        }
        ce = 0;
        cs = 0;
        for(ii = 0; ii < nilines; ii++) {
            if(iltable[il].ilmlist[ii] != 0) {
                cs++;
                ce = ii;
            }
        }
        if(cs > nextern/2) {
            cs = iltable[il].ilmlist[ce]*mconsv[ce];
            for(ii = 0; ii < nilines; ii++) {
                iltable[il].ilmlist[ii] -= cs*mconsv[ii];
            }
        }
    }
}

/*==================================================================*/
Private Void    ilexchg(NoArg)
/*--------------------------------------------------------------------
 * Change direction of ilines.
 *------------------------------------------------------------------*/
{
    int il, is, i, p;
#ifdef OLD
    int iv;
#endif

#if DBG
    if(dbg >= ildbg) {
        printf("ilexchg\n");
    }
#endif
    for(i = 0; i < nloopm; i++) {
        il = loopm[i];
        if(iltable[il].ilmlist[il] < 0) {
            ilswap(il, True);
        }
    }
    for(il = 1; il < nilines; il++) {
#ifdef OLD
        if(iltable[il].iltype == ILLOOP) {
        } else if(iltable[il].iltype == ILDLOOP) {
            for(i = 0; i < nloopm; i++) {
                if((is = iltable[il].ilmlist[loopm[i]]) < 0) {
                    ilswap(il, True);
                    break;
                } else if(is > 0) {
                    break;
                }
            }
        } else if(iltable[il].iltype != ILEXTERN) {
            is = 0;
            for(i = 1; i < nilines; i++) {
                if((iv = iltable[il].ilmlist[i]) > 0) {
                    is++;
                } else if(iv < 0) {
                    is--;
                }
            }
            if(is < 0) {
                ilswap(il, True);
            }
        }
#else
        if(iltable[il].iltype == ILLOOP) {
        } else if(iltable[il].iltype == ILDLOOP) {
            for(i = 0; i < nloopm; i++) {
                if((is = iltable[il].ilmlist[loopm[i]]) < 0) {
                    ilswap(il, True);
                    break;
                } else if(is > 0) {
                    break;
                }
            }
        } else if(iltable[il].iltype == ILTREE) {
            p = ndparticle(iltable[il].ilnodes[1],
                           iltable[il].illegs[1]);
            if(p < 0) {
                ilswap(il, True);
            }
        }
#endif
    }
#if DBG
    if(dbg >= ildbg) {
        printf("ilexchg: end\n");
    }
#endif
}

/*==================================================================*/
Private Void    ilswap(il, mode)
int     il;
bool    mode;
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    int tn, tl, i;

#if DBG
    if(dbg >= ildbg) {
        printf("ilswap(%d, %d)\n", il, mode);
    }
#endif
    if(cprocess->gm2) {
        return;
    }

    tn = iltable[il].ilnodes[0];
    iltable[il].ilnodes[0] = iltable[il].ilnodes[1];
    iltable[il].ilnodes[1] = tn;

    tl = iltable[il].illegs[0];
    iltable[il].illegs[0] = iltable[il].illegs[1];
    iltable[il].illegs[1] = tl;

    iltable[il].ilfldir  *= -1;

    if(!mode) {
        return;
    }
    for(i = 1; i < nilines; i++) {
        iltable[il].ilmlist[ i] *= -1;
    }

#if DBG
    if(dbg >= ildbg) {
        printf("ilswap: end\n");
    }
#endif
}

/*==================================================================*/
Private Void    ildmpabs(fp, labm, prefix)
FILE    *fp;
Ulong   labm;
string  prefix;
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    int i;
    Ulong lbit;

    fprintf(fp, "%s", prefix);
    lbit = 1L << cgraph->nnodes;
    for(i = cgraph->nnodes; i >= 0; i--) {
        if(labm & lbit) {
            fprintf(fp, "1");
        } else {
            fprintf(fp, "0");
        }
        lbit >>= 1;
    }
}

/*==================================================================*/
Private Void    ilencall(NoArg)
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    int il, v;
    int ielm = -1;
    Ulong kabs, ksgn, mbit, kcns, kelm;
    Ulong labm, lsgn;

#ifdef DBGENCODE
    printf("ilencall: graph : %ld\n", cgraph->grid);
#endif
    /* kabs : mask of external particles
     * ksgn : mask of final particles (- momentum)
     */
    kabs = ksgn = 0L;
    mbit = 2L;
    for(v = 0; v < ndnodes(); v++) {
        if((ndtype(v) == NDEXTERN) || (ndtype(v) == NDONSHELL)) {
            if(ndiorf(v) == ELINITIAL) {
                kabs |= mbit;
                ielm  = v;
            } else if(ndiorf(v) == ELFINAL) {
                kabs |= mbit;
                ksgn |= mbit;
                ielm   = v;
            }
        }
        mbit <<= 1;
    }
    kelm = 2 << ielm;
    kcns = kabs | 1;

    cgraph->gbextp = kabs;
    cgraph->gbcnsv = ksgn;
    cgraph->gbelem = kelm;

    /* labm : bit-string corresponding to the list of momenta
     * lsgn : bit-string corresponding to the list signs of momenta
     * lbit : labm without loop momenta.
     */
    for(il = 1; il < min(nilines, siltbl); il++) {
        labm = ilencode(siltbl, iltable[il].ilmlist, &lsgn);
        iltable[il].ilmabs = labm;
        iltable[il].ilmsgn = lsgn;
    }
#ifdef DBGENCODE
    ildump();
#endif
}

/*==================================================================*/
Public  Void    ilprmom(fp, il, pref)
FILE    *fp;
int     il;
string  pref;
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    int i;
    Ulong labm, lsgn;

    labm = iltable[il].ilmabs >> 1;
    lsgn = iltable[il].ilmsgn >> 1;

    for(i = 1; i < min(nilines, siltbl); i++) {
        if(labm & 1) {
            if(lsgn & 1) {
                fprintf(fp, "-");
            } else {
                fprintf(fp, "+");
            }
            fprintf(fp, " %s%d ", pref, i);
        }
        labm >>= 1;
        lsgn >>= 1;
    }
}

/*==================================================================*/
Public  Ulong   ilencode(nmvect, mvect, csgn)
int     nmvect;
int     *mvect;
Ulong   *csgn;
/*--------------------------------------------------------------------
 * usable only for tree graphs.
 *
 * This function should be called after initialization of ilencall
 */
{
    int ii;
    int ml;
    Ulong kabs, ksgn, mbit, kcns, kelm;
    Ulong labm, lsgn, lbit, ans;

    kabs = cgraph->gbextp;
    ksgn = cgraph->gbcnsv;
    kelm = cgraph->gbelem;
    kcns = kabs | 1;

    ans  = 0L;
    labm = 0L;
    lsgn = 0L;
    mbit = 2L;
    for(ii = 1; ii < min(nilines, nmvect); ii++) {
        ml    = mvect[ii];
        if(ml > 0) {
            labm |= mbit;
        } else if(ml < 0) {
            labm |= mbit;
            lsgn |= mbit;
        }
        mbit <<= 1;
    }
    lbit = kabs & labm;

    if((lbit & lsgn) == (lbit & ksgn)) {
        /* positive global sign */
    } else if((lbit & lsgn) == (lbit & ~ksgn)) {
        labm |= 1;
    } else {
        fprintf(Stderr, "*** ilencode : illegal combination of sign\n");
        ildmpabs(Stderr, kabs, " kabs = ");
        ildmpabs(Stderr, ksgn, " ksgn = ");
        ildmpabs(Stderr, kelm, " kelm = ");
        fprintf(Stderr, "\n");
        ildmpabs(Stderr, labm, " labm = ");
        ildmpabs(Stderr, lsgn, " lsgn = ");
        ildmpabs(Stderr, lbit, " lbit = ");
        fprintf(Stderr, "\n");
        printf("nilines = %d, nmvect = %d\n", nilines, nmvect);
        for(ii = 1; ii < min(nilines, nmvect); ii++) {
            printf("mvect[%d] = %d\n", ii, mvect[ii]);
        }
        ildump();
        exit(FAIL);
    }

    /* use momentum conservation : eliminate the last momentum */
    if(labm & kelm) {            /* includes the last momentum */
        labm  ^= kcns;
        if(lsgn & kelm) {
            lsgn ^= ~ksgn;
        } else {
            lsgn ^= ksgn;
        }
        lsgn &= labm & ~(1L);
    }

    ans   = labm;
    *csgn = lsgn;

    return ans;
}

/*==================================================================*/
Public  Void    ildecode(labm, nmvect, mvect)
Ulong   labm;
int     nmvect;
int     *mvect;
/*--------------------------------------------------------------------
 * usable only for tree graphs.
 */
{
    int i, sgn;
    Ulong kabs, ksgn, kelm, kcns, mabs, lbit, mbit, kbit, jabs;

    kabs = cgraph->gbextp;
    kelm = cgraph->gbelem;
    ksgn = cgraph->gbcnsv;
    kcns = kabs | 1;

    /* use momentum conservation */
    mabs  = labm ^ kcns;
    kbit  = 2L;
    lbit  = mbit = 0L;
    for(i = 1; i < min(nilines, nmvect); i++) {
        if(labm & kbit) {
            lbit++;
        }
        if(mabs & kbit) {
            mbit++;
        }
        kbit <<= 1;
    }
    if(lbit > mbit) {
        jabs = mabs;
    } else {
        jabs = labm;
    }
/*
    printf("labm=%8.8lx(%ld),mabs=%8.8lx(%ld),jabs=%8.8lx,kcns=%8.8lx\n",
           labm,lbit,mabs,mbit,jabs,kcns);
*/

    sgn = 1 - 2*(jabs & 1);
    jabs >>= 1;
    ksgn >>= 1;
    mvect[0] = 0;
    for(i = 1; i < min(nilines, nmvect); i++) {
        if(jabs & 1) {
            mvect[i] = sgn*(1 - 2*(ksgn & 1));
        } else {
            mvect[i] = 0;
        }
        jabs >>= 1;
        ksgn >>= 1;
    }
    for( ; i < nmvect; i++) {
        mvect[i] = 0;
    }
}

/*==================================================================*/
Public  Void    illoop(lilines, llegs, lpow)
int     *lilines;
int     *llegs;
int     *lpow;
/*--------------------------------------------------------------------
 * returns the information of the current graph
 */
{
    int i;
    node nd;
    int  il, ip, vt;

    if(cprocess->loop > 1) {
        fprintf(Stderr, "*** %d loop is not supported\n", cprocess->loop);
        exit(FAIL);
    }

    /* counter term */
    if(cprocess->loop != 0 && ilxnumb == 0) {
        *lilines = 0;
        for(nd = 0; nd < ndnodes(); nd++) {
            if(ndtype(nd) != NDLOOPEDV) {
                continue;
            }
            vt = ndvertex(nd);
            if(vt < 0) {
                fprintf(Stderr, "*** illegal looped vertex : ");
                fprintf(Stderr, "grid = %ld\n", cgraph->grid);
                fprintf(Stderr, "vt = %d, nd = %d/%d\n", vt, nd, ndnodes());
                exit(FAIL);
            } else if(vtable[vt]->loop != 0) {
                *llegs = vtable[vt]->nvlegs;
                break;
            }
        }
        *lpow    = 0;
        return;
    }

    /* ordinary case */
    *lilines = ilxnumb;
    *llegs   = 0;
    *lpow    = 0;
    for(i = 0; i < ilxnumb; i++) {
        nd = ndxtable[i];
        il = ilxtable[i];
        ip = ndparticle(iltable[il].ilnodes[1], iltable[il].illegs[1]);
        *llegs += ndlegs(nd) - 2;
        *lpow  += pvnumpowr(ip);
        *lpow  += pvpowr(ndvertex(nd));
    }
}
