/*********************************************************************
 * File : mpart.c (prefix = mp)
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "common.h"
#include "mpart.h"

/*------------------------------------------------------------------*
 * parameters
 */
#define MPCUT   1
#define PRESULT 0
#define TRANS   0
#define MAXM    10
#define MAXN    20
#define DEBUG

/*------------------------------------------------------------------*
 * Local types
 */
typedef struct {
    int m, n;
} tstack;

/*------------------------------------------------------------------*
 * Local functions
 */
Private bool  mpput  ARG((int,int,int**,int,int,int));
Private bool  mpgen0  ARG((int,int,int**));
Private bool  mpgen1  ARG((int,int,int**));
Private bool  mpgen2  ARG((int,int,int**,int,int,bool));

#define max(x, y)     (((x) > (y)) ? (x) : (y))
#define min(x, y)     (((x) < (y)) ? (x) : (y))
/*------------------------------------------------------------------*
 * Local variables
 */
#define MAXSK  (MAXM*MAXN)
Local tstack mpstack[MAXSK];
Local int    mpskpt = 0;

#ifdef DEBUG
Local int    maxcall = 0;
Local bool   dbg = False;
#endif

/*********************************************************************
*/
/*==================================================================*/
Public  int   **mpnew(m, n)
int     m, n;
/*--------------------------------------------------------------------
 *  Allocate memory for int matrix[m][n]
 */
{
    int i;
    int **mat;

    if((mat = (int **) malloc(sizeof(int*)*m)) == NULL) {
        fprintf(Stderr,"*** mpnew : memory (m,n)=(%d,%d)\n", m, n);
        exit(FAIL);
    }
    for(i = 0; i < m; i++) {
        if((mat[i] = (int *) malloc(sizeof(int)*n)) == NULL) {
            fprintf(Stderr,"*** mpnew : memory (m,n)=(%d,%d)\n", m, n);
            exit(FAIL);
        }
    }
    return mat;
}

/*==================================================================*/
Public  int   mpchkpt(sp)
int     sp;
/*--------------------------------------------------------------------
 *  Return the current stack pointer
 */
{
    if(sp >= 0) {
        mpskpt = sp;
    }
    return mpskpt;
}

/*==================================================================*/
Public  int   mpush(s, c)
int     s, c;
/*--------------------------------------------------------------------
 *  Save to stack
 */
{
#ifdef DEBUG
    int i;

    for(i = 0; i < mpskpt; i++) {
        if(mpstack[i].m == s && mpstack[i].n == c) {
            fprintf(Stderr, "*** mpush : (%d,%d) already appeared\n",
                    s, c);
            exit(FAIL);
        }
    }
#endif

    if(mpskpt >= MAXSK) {
        fprintf(Stderr, "*** mpush : stack overflow\n");
        exit(FAIL);
    }
    mpstack[mpskpt].m = s;
    mpstack[mpskpt].n = c;

    return mpskpt++;
}

/*==================================================================*/
Public  Void  mpull(m, n, g, pt)
int     m, n;
int     **g;
int     pt;
/*--------------------------------------------------------------------
 *  Restore from stack
 */
{
    int s, c, v;

    if(mpskpt < pt || pt < 0) {
        fprintf(Stderr, "*** mpull : illegal arg. %d > %d\n", pt, mpskpt);
        exit(FAIL);
    }
    while(pt < mpskpt) {
        --mpskpt;
        s = mpstack[mpskpt].m;
        c = mpstack[mpskpt].n;
/*
printf("pull:%d:(%d,%d)\n", mpskpt, s, c);
*/
        v = g[s][c];
        g[m][c] += v;
        g[s][n] += v;
        g[m+1][c]++;
        g[s][n+1]++;
        g[s][c] = -1;
    }
}

/*==================================================================*/
Public  Void  mpdumpm(m, n, g, add)
int     m, n;
int     **g;
int     add;
/*--------------------------------------------------------------------
 *  Print matrix g[m+2][n+2]
 */
{
    int s, c;

    printf("   :");
    for(c = 0; c < n + add; c++) {
        if(c >= n) {
            printf(" :");
        }
        printf(" %2d", c);
    }
    printf("\n");

    for(s = 0; s < m + add; s++) {
        if(s == 0 || s >= m) {
            printf("---:");
            for(c = 0; c < n + add; c++) {
                if(c >= n) {
                    printf("-:");
                }
                printf("---");
            }
            printf("\n");
        }
        printf("%2d :", s);
        for(c = 0; c < n + add; c++) {
            if(c >= n) {
                printf(" :");
            }
            printf(" %2d", g[s][c]);
        }
        printf("\n");
    }
}

/*==================================================================*/
Private bool  mpput(m, n, g, s, c, v)
int     m, n;
int   **g;
int     s, c, v;
/*--------------------------------------------------------------------
 */
{
    if(g[s][c] >= 0) {
        return False;
    }
    if(g[m][c] < v || g[s][n] < v) {
        return False;
    }
    if(g[m+1][c] < 1 || g[s][n+1] < 1) {
        return False;
    }
#if TRANS
#else
    mpush(s, c);
#endif
    g[s][c]   = v;
    g[s][n]  -= v;
    g[m][c]  -= v;
    g[s][n+1]--;
    g[m+1][c]--;

    return True;
}

/*==================================================================*/
Private bool  mpresult(m, n, g)
int     m, n;
int   **g; 
/*--------------------------------------------------------------------
 */
{
#if PRESULT
    static long nconf = 0L;

    printf("\n+++ New Config %ld\n", ++nconf);
    mpdumpm(m, n, g, 2);
#endif

    return True;
}

/*==================================================================*/
Private bool  mpgen0(m, n, g)
int     m, n;
int   **g; 
/*--------------------------------------------------------------------
 */
{
    int s, c, v, ls, lc, ns, nc;
    bool mod = True;

    while(mod) {
        mod = False;
        for(c = 0; c < n; c++) {
            if(g[m+1][c] < 1) {
                if(g[m][c] > 0) {
                    goto bad;
                }
            } else if(g[m+1][c] == 1) {
                ns = 0; ls = -1;
                for(s = 0; s < m; s++) {
                    if(g[s][c] < 0) {
                        ns++;
                        ls = s;
                    }
                }
                if(ns != 1) {
                    fprintf(Stderr, "*** mpgen0 : illegal control, ");
                    fprintf(Stderr, " ns=%d,c=%d\n", ns, c);
                    mpdumpm(m, n, g, 2);
                    exit(FAIL);
                }
                v = g[m][c];
                if(!mpput(m, n, g, ls, c, v)) {
                    goto bad;
                }
/*
printf("mgen0:1;s=%d,c=%d\n", ls, c); mpdumpm(m,n,g,2);
*/
                mod = True;
            }
        }
        for(s = 0; s < m; s++) {
            if(g[s][n+1] < 1) {
                if(g[s][n] > 0) {
                    goto bad;
                }
            } else if(g[s][n+1] == 1) {
                nc = 0; lc = -1;
                for(c = 0; c < n; c++) {
                    if(g[s][c] < 0) {
                        nc++;
                        lc = c;
                    }
                }
                if(nc != 1) {
                    fprintf(Stderr, "*** mpgen0 : illegal control, ");
                    fprintf(Stderr, " nc=%d,s=%d\n", nc, s);
                    mpdumpm(m, n, g, 2);
                    exit(FAIL);
                }
                v = g[s][n];
                if(!mpput(m, n, g, s, lc, v)) {
                    goto bad;
                }
/*
printf("mgen0:2;s=%d,c=%d\n", s, lc); mpdumpm(m,n,g,2);
*/
                mod = True;
            }
        }
    }
/*
printf("mgen0:OK;\n"); mpdumpm(m,n,g,2);
*/
    return True;

    bad:;
/*
printf("mgen0:NO;\n"); mpdumpm(m,n,g,2);
*/
    return False;
}

/*==================================================================*/
Private bool  mpgen2(m0, n0, g, t0, d0, tr)
int     m0, n0;
int   **g;
int     t0, d0;
bool    tr;
/*--------------------------------------------------------------------
 * Fixed s0, starting from c0
 */
{
    int **g1;
    int m, n, s0, c0;
    int s, c, u, v, xv, um1, umin, umax;
    bool st = False;
#if TRANS
#else
    int sp;
#endif

#ifdef DEBUG
    static int level = 0;

    maxcall++;
    level++;
    if(dbg) {
        printf("begin:%2d:mpgen2(%d, %d, g, %d, %d, %d)\n",
               level, m0, n0, t0, d0, tr);
    }
#endif


#if TRANS
    if(tr) {
        m  = n0;
        n  = m0;
        s0 = d0;
        c0 = t0;
    } else {
        m  = m0;
        n  = n0;
        s0 = t0;
        c0 = d0;
    }
#else
    if(tr) {
        fprintf(Stderr, "*** mpgen2 : cannot transpose\n");
        exit(FAIL);
    }
    m  = m0;
    n  = n0;
    s0 = t0;
    c0 = d0;
#endif

#if TRANS
    g1 = mpnew(m+2, n+2);
    for(s = 0; s < m+2; s++) {
        for(c = 0; c < n+2; c++) {
            if(tr) {
                g1[s][c] = g[c][s];
            } else {
                g1[s][c] = g[s][c];
            }
        }
    }
#else
    g1 = g;
    sp = mpchkpt(-1);
#endif

    s = s0;
    for(c = c0; c < n; c++) {
        if(g1[s0][c] < 0) {
            c0 = c;
            break;
        }
    }
    if(c >= n) {
        fprintf(Stderr, "*** mpgen2 : illegal control\n");
        exit(FAIL);
    }

    if(g1[s0][n+1] == 1) {
        if(!mpput(m, n, g1, s0, c0, g1[s0][n])) {
            st = False;
        } else if(!mpgen0(m, n, g1)) {
            st = False;
        } else {
            st = mpgen1(m, n, g1);
        }
/*
printf("mpgen2:1:\n"); mpdumpm(m, n, g1, 2);
*/
    } else {
        umax = min(g1[m][c0], g1[s0][n]);

        for(c = c0+1, xv = 0; c < n; c++) {
            if(g1[s0][c] < 0) {
                xv += g1[m][c];
            }
        }
        v    = g1[s0][n] - xv;
        umin = max(0, v);

/*
        for(s = s0+1, xv = 0; s < m; s++) {
            if(g1[s][c0] < 0) {
                xv += g1[s][n];
            }
        }
        v    = g1[m][c0] - xv;
        um1  = max(0, v);
        umin = max(umin, um1);
*/

        st  = False;
        for(u = umin; u <= umax; u++) {
            if(dbg) {
                printf("     :%2d:mpgen2(%d, %d, g, %d, %d, %d) ",
                       level, m0, n0, t0, d0, tr);
                printf(" %d <= %d <= %d\n", umin, u, umax);
            }
            if(!mpput(m, n, g1, s0, c0, u)) {
                st = False;
            } else if(!mpgen0(m, n, g1)) {
                st = False;
            } else if(g1[s0][n+1] > 0) {
                st = mpgen2(m, n, g1, s0, c0+1, False);
            } else {
                st = mpgen1(m, n, g1);
            }
/*
printf("mpgen2:2:\n"); mpdumpm(m, n, g1, 2);
*/
#if TRANS
            for(s = 0; s < m+2; s++) {
                for(c = 0; c < n+2; c++) {
                    if(tr) {
                        g1[s][c] = g[c][s];
                    } else {
                        g1[s][c] = g[s][c];
                    }
                }
            }
#else
            mpull(m, n, g1, sp);
#endif
#if MPCUT
            if(st) {
                break;
            }
#endif
        }
    }

#if TRANS
    free(g1);
#else
    mpull(m, n, g1, sp);
#endif

#ifdef DEBUG
    if(dbg) {
        printf("end:  %2d:mpgen2(%d, %d, g, %d, %d, %d) : st = %d\n",
               level, m, n, s0, c0, tr, st);
    }
    level--;
#endif
    return st;
}

/*==================================================================*/
Private bool  mpgen1(m, n, g)
int     m, n;
int   **g;
/*--------------------------------------------------------------------
 */
{
#if TRANS
    int c;
#endif
    int s;
    int minrc, mini, minp, mins;
    bool st = False;
#ifdef DEBUG
    static int level=0;

    level++;
    if(dbg) {
        printf("begin:%2d:mpgen1(%d, %d, g)\n", level, m, n);
    }
#endif
    minrc = 0;
    mini  = -1;
    minp  = 0;
    mins  = 0;
    for(s = 0; s < m; s++) {
        if((g[s][n+1] > 0) &&
           (minp == 0 || minp > g[s][n+1] || mins > g[s][n])) {
            minp = g[s][n+1];
            mins = g[s][n];
            mini = s;
            minrc = 1;
        }
    }
#if TRANS
    for(c = 0; c < n; c++) {
        if((g[m+1][c] > 0) &&
           (minp == 0 || minp > g[m+1][c] || mins > g[m][c])) {
            minp = g[m+1][c];
            mins = g[m][c];
            mini = c;
            minrc = 2;
        }
    }
#endif
    if(minrc == 0) {
        st = mpresult(m, n, g);
    } else if(minrc == 1) {
        st = mpgen2(m, n, g, mini, 0, False);
    } else {
        st = mpgen2(m, n, g, 0, mini, True);
    }
#ifdef DEBUG
    if(dbg) {
        printf("end:  %2d:mpgen1(%d, %d, g) : st = %d\n", level, m, n, st);
    }
    level--;
#endif

    return st;
}

/*==================================================================*/
Public  int   mpgen(m, n, f, q, r)
int     m, n;
int   **f;
int    *q;
int    *r;
/*--------------------------------------------------------------------
 *  Generate or test a matrix g[m][n] which satisfiles the following
 *  conditions.
 *
 *  1. sum_c g[s][c] = q[s],  for all s
 *  2. sum_s g[s][c] = r[s],  for all c
 *  3. g[s][c] = 0,  if f[s][c] = 0,   fro all s, c
 */
{
    int **g;
    int s, c, v, lc, ns, nc, ic;
    bool st;

    g = mpnew(m+2, n+2);

    ns = 0;
    for(s = 0; s < m; s++) {
        ns       += q[s];
        g[s][n]   = q[s];    /* sum g[s][] */
        g[s][n+1] = 0;       /* # assignable to g[s][] */
    }
    nc = 0;
    for(c = 0; c < n; c++) {
        nc       += r[c];
        g[m][c]   = r[c];    /* sum g[][c] */
        g[m+1][c] = 0;       /* # assignable to g[][c] */
    }
    if(ns != nc) {
        goto bad;
    }
    g[m][n]   = ns;
    g[m][n+1] = 0;
    g[m+1][n] = 0;

    ic = 0;
    for(s = 0; s < m; s++) {
        if(g[s][n] == 0) {
            for(c = 0; c < n; c++) {
                g[s][c] = 0;
            }
        } else {
            lc = -1;
            for(c = 0; c < n; c++) {
                if(g[m][c] == 0 || f[s][c] == 0) {
                    g[s][c] = 0;  /* fix to 0 */
                } else {
                    g[s][c] = -1;
                    g[m+1][c]++;
                    g[s][n+1]++;
                    ic++;
                    lc = c;
                }
            }
            if(g[s][n+1] < 1) {
                if(g[s][n] > 0) {
                    goto bad;
                }
            } else if(g[s][n+1] == 1) {
                v = g[s][n];
                if(!mpput(m, n, g, s, lc, v)) {
                    goto bad;
                }
            }
        }
    }
    g[m+1][n+1] = ic;
    
if(m==8 && n==18) {
printf("mpgen:1:\n");mpdumpm(m, n, g, 2);
}
    if(!mpgen0(m, n, g)) {
        goto bad;
    }
if(m==8 && n==18) {
printf("mpgen:2:\n");mpdumpm(m, n, g, 2);
}
    st = mpgen1(m, n, g);

#if TRANS
#else
    Discard mpchkpt(0);
#endif

#ifdef DEBUG
    if(dbg) {
        printf("mpgen : OK, st = %d\n", st);
/*
        mpdumpm(m, n, g, 2);
*/
    }
#endif
    return st;

    bad:;
#if TRANS
#else
    Discard mpchkpt(0);
#endif

#ifdef DEBUG
    if(dbg) {
        printf("mpgen : Bad\n");
/*
        mpdumpm(m, n, g, 2);
*/
    }
#endif
    return 0;
}

/*********************************************************************
*/
#ifdef TESTMAIN
#define M MAXM
#define N MAXN
/*==================================================================*/
int main()
/*--------------------------------------------------------------------
 */
{
    int **f, **g, *r, *q;
    int v, s, c;
    int m, n;
    
    srand(12345);
    printf("test mpart ; M=%d, N=%d, TRANS=%d\n", M, N, TRANS);

    if((q = (int *) malloc(sizeof(int*)*M)) == NULL) {
        fprintf(Stderr,"*** main : memory\n");
        exit(FAIL);
    }
    if((r = (int *) malloc(sizeof(int*)*N)) == NULL) {
        fprintf(Stderr,"*** main : memory\n");
        exit(FAIL);
    }
    f = mpnew(M, N);
    g = mpnew(M, N);

    for(m = 2; m < M; m++) {
        for(n = 2; n < N; n++) {
#ifdef DEBUG
            maxcall = 0;
#endif
            for(s = 0; s < m; s++) {
                q[s] = 0;
            }
            for(c = 0; c < n; c++) {
                r[c] = 0;
            }
            for(s = 0; s < m; s++) {
                for(c = 0; c < n; c++) {
                    if((rand() & 0x040) == 0) {
                        f[s][c] = 1;
                        v = ((rand() >> 4) & 0x0f);
                        g[s][c] = v;
                        q[s]   += v;
                        r[c]   += v;
                    } else {
                        f[s][c] = 0;
                        g[s][c] = 0;
                    }
                }
            }
if(m==8 && n==18) {
dbg = True;
}
            if(dbg || PRESULT) {
                printf("\n f = \n"); mpdumpm(m, n, f, 0);
                printf("\n g = \n"); mpdumpm(m, n, g, 0);
                printf("\n q =");
                for(s = 0; s < m; s++) {
                    printf(" %2d", q[s]);
                }
                printf("\n r =");
                for(c = 0; c < n; c++) {
                    printf(" %2d", r[c]);
                }
                printf("\n");
            }
            mpgen(m, n, f, q, r);

#ifdef DEBUG
            fprintf(stderr, "m=%d, n=%d, maxcall=%d\n", m, n, maxcall);
#endif
        }
    }
    return 0;
}
#endif
