/*********************************************************************
 *  File : mvert.c (prefix = mv)
 *
 *  match vertex
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "grdata.h"

/*--------------------------------------------------------------------
 * Local functions
 */
Private int     mvwand       ARG((int,vindex[],int,vindex[]));
Private int     mvmat        ARG((int,vindex[],int*,pindex[],bool));
Private vindex  mvmatcl      ARG((int,int[]));

/*--------------------------------------------------------------------
 * Local variables
 */
Local   vindex  mvwlist[LVTABLE];
Local   pindex  plist[MAXTL];
Local   int     mlist[MAXTL];

/* Include or not include composit vertex 
 * graph generation : not include
 * read out.grf     : include
 */
Local   bool    inclcomp = False;

/*********************************************************************
 *  For assignment
 */

/*==================================================================*/
Private  int  mvwand(nvt, mvt, nvs, mvs)
int      nvt;
vindex   mvt[];
int      nvs;
vindex   mvs[];
/*--------------------------------------------------------------------
 * Construct and list of two list of vertices "mvs[nvs]"
 * and "mvt[nvt]".
 * The resulting list is saved to "mvt[return_value]"
 * These list are assumed to be sorted in ascending order.
 *
 * Returns:
 *     the number of elements of the resulting list.
 *
 * Arguments:
 *     nvt : size of a list
 *     mvt : a list.  The resulting list is saved in there.
 *     nvs : size of another list
 *     mvs : another list.
 */
{
    int ls = 0, lt = 0, lr = 0;

#if DBG
    if(dbg >= pmdbg) {
        printf("mvwand : s%d :", nvs);
        pvldump(stdout, nvs, mvs);
        printf("mvwand : t%d :", nvt);
        pvldump(stdout, nvt, mvt);
    }
#endif

    /* enpty set */
    if(nvt < 1) {
        return nvt;
    }

    while(True) {
        while(ls < nvs && mvs[ls] < mvt[lt]) {
            ls++;
        }
        if(ls >= nvs) {
            break;
        }

        while(lt < nvt && mvt[lt] < mvs[ls]) {
            lt++;
        }
        if(lt >= nvt) {
            break;
        }

        if(mvs[ls] == mvt[lt]) {
            mvt[lr++] = mvt[lt];
            ls++;
            lt++;
        }
    }

#if DBG
    if(dbg >= pmdbg) {
        printf("mvwand : r%d :", nvt);
        pvldump(stdout, nvt, mvt);
    }
#endif

    return lr;
}

/*==================================================================*/
Private int  mvlist(p, a, nl, pl)
pindex  p;
int     a;
int     nl;
vindex  pl[];
/*--------------------------------------------------------------------
 *  construct list of vertices which includes
 *  particle "p"(particle a=0 or anti-particle a=1) in its "nl" legs.
 */
{
    int i, nvl, j;
    vindex v;

    j = 0;
    nvl = ptable[p]->nvlist[a];
    for(i = 0; i < nvl; i++) {
        v = ptable[p]->vlist[a][i];
        if(!inclcomp && vtable[v]->composit == V_COMPOSIT) {
            continue;
        }
        if(vtable[v]->nvlegs == nl) {
            if(j < LVTABLE) {
                pl[j] = v;
            }
            j++;
        }
    }
    p = (a == 0) ? p : -p;
    for(v = nvmodel; v < nvtable; v++) {
        if(!inclcomp && vtable[v]->composit == V_COMPOSIT) {
            continue;
        }
        if(vtable[v]->nvlegs != nl) {
            continue;
        }
        for(i = 0; i < nl; i++) {
            if(vtable[v]->vlegs[i] == p) {
                break;
            }
        }
        if(i < nl) {
            if(j < LVTABLE) {
                pl[j] = v;
            }
            j++;
        }
    }
    if(j >= LVTABLE){
        fprintf(Stderr, "too many candidate vertices %d(%d)\n",
                j, LVTABLE);
        exit(1);
    }
    return j;
}

/*==================================================================*/
Private int  mvmat(nd, pl, nt, pt, final)
int    nd;
vindex  pl[];
int    *nt;
pindex  pt[];
bool    final;
/*--------------------------------------------------------------------
 *  Construct the list of vertices which are assignable to the node
 *  "nd" ignoring multiplicity of the connection.
 *  This function return the length of the list.
 *
 *  input
 *    nd : node to be matched.
 *    nt : if(nt == NULL) then not use "nt" and "pt" for output.
 *    final : final check
 *
 *  output
 *    pl : list of possible vertices.
 *    pt : list of particles assignable to undefined legs.
 *         This data is prepared when input value of "nt" is not NULL.
 *    nt : length of "pt".
 *         This data is prepared when input value of "nt" is not NULL.
 */
{
#define MAXLGS 10

    bool match;
    int  l, a, nvl, nvl1, i, j, np=0, nl, lp;
    int  lt, v, vv, il;
    pindex p, vl[MAXLGS];
    pindex pl1[LVTABLE];

#if DBG
    if(dbg >= pmdbg) {
        printf("*dbg* mvmat(%d,pl,%p,pt,final)\n",nd,nt);
    }
#endif

    if(nt != NULL) {
        *nt = 0;
    }

    nl = ndlegs(nd);
    lp = ndloop(nd);

    if(nl > MAXLGS) {
        fprintf(Stderr, "*** mvmat : too small size of nl=%d\n", nl);
        exit(16);
    }

    /* construct list of assigned particles for "nd"
     * plist : list of particles,
     * mlist : list of multiplicity
     * np    : the length of the list
     */
    for(l = 0; l < ndlegs(nd); l++) {
        p = ndparticle(nd, l);
        if(p == PNULL) {
            if(final) {
                /* final with not assigned leg */
                return 0;
            }
            continue;
        }

        /* multiple connection ? */
        for(i = 0; i < np; i++) {
            if(plist[i] == p) {
                mlist[i]++;
                goto end_save;
            }
        }

        /* new particle */
        plist[np] = p;
        mlist[np] = 1;
        np++;

        end_save:;
    }

#if DBG
    if(dbg >= pmdbg) {
        printf("mvmat : np = %d, ", np);
        for(l = 0; l < np; l++) {
            printf("%d[%d] ", plist[l], mlist[l]);
        }
        printf("\n");
    }
#endif

    /* construct list of candidate vertices
     * pl  : list of vertices
     * nvl : length of "pl".
     */
    nvl = 0;
    for(l = 0; l < np; l++) {
        p = plist[l];
        if(p == PNULL)  /* cannot occur ?? */
            continue;

        /* a = 0 for particle, 1 for anti-particle */
        a = (p < 0) ? 1 : 0;
        p = abs(p);

        /* construct initial list */
        if(l == 0) {

            nvl = mvlist(p, a, nl, pl);

        /* and with another list */
        } else {
#if DBG
            if(dbg >= pmdbg) {
                printf("mvmat : p=%d, a=%d :", p, a);
                pvldump(stdout, ptable[p]->nvlist[a],
                                ptable[p]->vlist[a]);
            }
#endif
            nvl1 = mvlist(p, a, nl, pl1);
            nvl = mvwand(nvl, pl, nvl1, pl1);

            /* empty */
            if(nvl < 1) {
                return 0;
            }
        }
#if DBG
        if(dbg >= pmdbg) {
            printf("mvmat : l=%d, nvl=%d :", l, nvl);
            pvldump(stdout, nvl, pl);
        }
#endif
    }

    /* check each vertex
     * pl : list of vertices possible to assign
     * lt : length of the list.
     */
    lt = 0;

    /* for vv = candidat_vertex */
    for(v = 0; v < nvl; v++) {
        vv = pl[v];

        if(nl != vtable[vv]->nvlegs) {
            continue;
        }

        /* for counter term,  lp > vtable[vv]->loop */
        if(lp <  vtable[vv]->loop) {
            continue;
        }

        for(i = 0; i < nl; i++) {
            vl[i] = vtable[vv]->vlegs[i];
        }

        il = nl;       /* # of partices not in plist, mlist */

        /* for plist and multiplicity
         * vl : list of particle possible to assign.
         * il : length of vl.
         */
        for(l = 0; l < np; l++) {
            for(j = 0; j < mlist[l]; j++) {

                /* delete particle from vl */
                match = False;
                for(i = 0; i < il; i++) {
                    if(vl[i] == plist[l]) {
                        vl[i] = vl[--il];
                        match = True;
                        break;
                    }
                }
                if(!match) {
                    goto not_match;  /* not accept */
                }
            }
        }

        /* accept the vertex vv */
        pl[lt++] = vv;

        /* save possible particles which are assignable to
         * the undefined legs
         */
        if(nt != NULL) {
            for(i = 0; i < il; i++) {
                for(j = 0; j < *nt; j++) {
                    if(pt[j] == vl[i]) {
                        break;
                    }
                }
                if(j >= *nt) {
                    pt[(*nt)++] = vl[i];
                }
            }
        }
        not_match:;
    }

#if DBG
    if(dbg >= pmdbg) {
        printf("mvmat nd = %d, nvl=%d, mvwlist=", nd, nvl);
        pvldump(stdout, nvl, pl);
    }
#endif
    return lt;
}

/*==================================================================*/
Public  int     mvable(nd)
int    nd;
/*--------------------------------------------------------------------
 *  Check if this node is possible to assign a vertex.
 *  The node may have undetermined legs.
 *
 *  returns :
 *     the length of list of candidate vertices.
 *
 *  Argument:
 *     nd : node
 */
{
    int nvl;

    /* not construct the list of assignable particles */
    nvl = mvmat(nd, mvwlist, (int*)NULL, (pindex*)NULL, False);

#if DBG
    if(dbg >= pmdbg) {
        printf("mvable: nd = %d, nvl = %d\n", nd, nvl);
    }
#endif

    return nvl;
}

/*==================================================================*/
Public  int     mvmlist(nd, nl)
int    nd;
int    nl[];
/*--------------------------------------------------------------------
 * Find vertex in the model which is in the same structure of
 * node "nd".
 * This function should be called for a node whithou undetermined
 * legs.
 * We assume only one vertex is possible to assing (bat condition???)
 *
 * Returns :
 *     the number of candidate vertices.
 *
 * Argument :
 *     nd : node.
 *     nl : output : list of vertices
 */
{
    int lnl;

#if DBG
    if(dbg >= pmdbg) {
        printf("mvmlist: nd = %d\n", nd);
    }
#endif

    /* final check withou constructing the list of assignable
     * particles
     */
    lnl = mvmat(nd, nl, (int*)NULL, (pindex*)NULL, True);

#if DBG
    if(dbg >= pmdbg) {
        printf("mvmlist: nd = %d, lnl=%d\n", nd, lnl);
    }
#endif
    return lnl;
}

/*==================================================================*/
Public  int mvcandp(nd, prtcl)
int    nd;
pindex *prtcl;
/*--------------------------------------------------------------------
 *  Construct the list of particles assignable to
 *  free legs of node "nd".
 *  This function returns the length of "prtcl".
 *
 *  Returns:
 *      length of the list.
 *
 *  input
 *    nd : node
 *
 *  output
 *    prtcl : list of particles assignable to the undefined legs.
 */
{
    int np;

    Discard mvmat(nd, mvwlist, &np, prtcl, False);

    return np;
}

/*==================================================================*/
Public vindex mvmatc(nd, mord)
int    nd;
int    mord[];
/*--------------------------------------------------------------------
 *  Find vertex which match to node "nd" in the generated graph.
 *  free legs of node "nd".
 *  This function returns the length of "prtcl".
 *
 *  Returns:
 *      index of the vertex defined in the table
 *
 *  input
 *    nd : node
 *    mord : list of order of coupling constants.
 *
 *  output
 *    prtcl : list of particles assignable to the undefined legs.
 */
{
    inclcomp = True;
    return mvmatcl(nd, mord);
}

/*==================================================================*/
Public vindex mvmatcg(nd, mord)
int    nd;
int    mord[];
/*--------------------------------------------------------------------
 *  Find vertex which match to node "nd" for graph generation.
 *  free legs of node "nd".
 *  This function returns the length of "prtcl".
 *
 *  Returns:
 *      index of the vertex defined in the table
 *
 *  input
 *    nd : node
 *    mord : list of order of coupling constants.
 *
 *  output
 *    prtcl : list of particles assignable to the undefined legs.
 */
{
    inclcomp = False;
    genct    = True;
    return mvmatcl(nd, mord);
}

/*==================================================================*/
Private vindex mvmatcl(nd, mord)
int    nd;
int    mord[];
/*--------------------------------------------------------------------
 *  Find vertex which match to node "nd".
 *  free legs of node "nd".
 *  This function returns the length of "prtcl".
 *
 *  Returns:
 *      index of the vertex defined in the table
 *
 *  input
 *    nd : node
 *    mord : list of order of coupling constants.
 *
 *  output
 *    prtcl : list of particles assignable to the undefined legs.
 */
{
    int lnl, io, v, l, lg, lp;
    node nl[LVTABLE];
    pindex vl[MAXLGS];

    lnl = mvmat(nd, nl, (int*)NULL, (pindex*)NULL, True);

    if(lnl < 1) {
        if(cprocess->expand) {
            fprintf(Stderr, "mvmatc : cannot find counter term : ");
            fprintf(Stderr, "node=%d\n", nd);
            prdump();
            exit(FAIL);
        } else {
            lg = ndlegs(nd);
            lp = ndloop(nd);
            for(l = 0; l < lg; l++) {
                vl[l] = ndparticle(nd, l);
            }
            mord[0] = 2*lp + lg - 2;
            for(io = 1; io < nmordtbl; io++) {
                mord[io] = 0;
            }
            if(genct) {
                return pvaddctn(lg, vl, mord);
            }
        }
    }

    for(l = 0; l < lnl; l++) {
        v = nl[l];
        for(io = 0; io < nmordtbl; io++) {
            if(mord[io] != vtable[v]->morder[io]) {
                break;
            }
        }
        if(io >= nmordtbl) {
            return v;
        }
    }
    
    if(genct) {
        v = pvaddct(nl[0], mord);
    
        return v;
    }
    fprintf(Stderr, "*** mvmatc : no vertex is defined for node %d\n", nd);
    fprintf(Stderr, "    in graph=%ld\n", cgraph->grid);
    exit(FAIL);
}
