/*********************************************************************
 * File : ordset.c (prefix = os)
 *
 * functions for set of orders.
 *                                                 by Kaneko Toshiaki
 *
 * 1998/06/25
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#ifdef TESTMAIN
#define DEFGLOBAL
#endif

#include "grdata.h"

/*------------------------------------------------------------------*
 * parameters
 */
#define TRACE 0

/*------------------------------------------------------------------*
 * Local types
 */
typedef enum {OSDISJ, OSSAME, OSLESS, OSGREATER,
              OSDESC, OSANSC, OSOVLP, OSZG, OSZL, OSZ} tinc;

/*------------------------------------------------------------------*
 * Local functions
 */
Private tinc    osinc        ARG((int,int));
Private int     ostree       ARG((int*,int*,int,int));
Private int     osatree      ARG((int*,int*,int,int));

/*------------------------------------------------------------------*
 * Local variables
 */
/* osroot is an array of the following masks */
#define OSEMPTY  0
#define OSROOT   1
#define OSBRANCH 2
#define OSLEAF   4
#define OSROOTB  3  /* (OSROOT | OSBRANCH) */
#define OSROOTL  5  /* (OSROOT | OSLEAF)   */

Local int  osincl[MAXSORD][MAXSORD];  /* relation two sets of orders */
Local int  osroot[MAXSORD];           /* the set is root etc. */
Local int  osconj[OSZ+1];             /* conjugate relation */
Local char osincc[OSZ+2] = "!=<>v^Xglz";
Local char osnodc[OSROOTL+2] = " RbBlL";
Local int  sordr[MAXSORD];
Local int  mleft[MAXSORD];

/*------------------------------------------------------------------*
 * Global variables
 *
 * char *mordtbl[MAXMORD];          : name of coupling const
 * int   mordopt[MAXMORD];          : option flags
 * int   nmordtbl;                  : # of data in the table
 *
 * char *sordtbl[MAXSORD];          : name of set of coupling const
 * int   sordset[MAXSORD][MAXMORD]; : bit flags of set of coupling const
 *                                    value may be -1(original coupl.const),
 *                                    0 or 1
 * int   sordelm[MAXSORD];          : # of elements
 * int   nsordtbl;                  : # of data in the table
 */

/*********************************************************************
  Read model definition file.
*/
/*==================================================================*/
Private tinc    osinc(i, j)
int     i, j;
/*--------------------------------------------------------------------
 * Calculate the relation between set-i and set-j.
 *
 * Input
 *     int i, j : index of two sets.
 *     Global sordset
 *
 * Ouptut
 *     tinc return_value :
 *         OSDISJ    : set-i \cap set-j = \emptyset.
 *         OSSAME    : set-i == set-j.
 *         OSLESS    : set-i \subset set-j && set-i != set-j.
 *         OSGREATER : set-i \supset set-j && set-i != set-j.
 *         OSOVLP    : set-i \cap set-j != \emptyset &&
 *                     set-i \not\subset set-j &&
 *                     set-i \not\supset set-j &&
 */
{
    int  k;
    tinc r;

    osconj[OSDISJ]    = OSDISJ;
    osconj[OSSAME]    = OSSAME;
    osconj[OSLESS]    = OSGREATER;
    osconj[OSGREATER] = OSLESS;
    osconj[OSOVLP]    = OSOVLP;
    osconj[OSZ]       = OSZ;
    osconj[OSZG]      = OSZL;
    osconj[OSZL]      = OSZG;

    r = OSZ;
    for(k = 0; k < nmordtbl; k++) {
        if(sordset[i][k] != 0) {
            if(sordset[j][k] != 0) {   /* i <= j or i >= j */
                switch(r) {
                    case OSZ:
                        r = OSSAME;
                        break;

                    case OSZL:
                        r = OSLESS;
                        break;

                    case OSZG:
                        r = OSGREATER;
                        break;

                    case OSDISJ:
                        r = OSOVLP;
                        goto ret;

                    case OSSAME:
                    case OSLESS:
                    case OSGREATER:
                        break;

                    case OSOVLP:
                    case OSANSC:
                    case OSDESC:
                        goto err;
                }
            } else {                      /* i >= j or disjoint */
                switch(r) {
                    case OSZ:
                    case OSZG:
                        r = OSZG;
                        break;

                    case OSZL:
                        r = OSDISJ;
                        break;

                    case OSLESS:
                        r = OSOVLP;
                        goto ret;
                
                    case OSSAME:
                    case OSGREATER:
                        r = OSGREATER;
                        break;

                    case OSDISJ:
                        break;

                    case OSOVLP:
                    case OSANSC:
                    case OSDESC:
                        goto err;
                }
            }
        } else {
            if(sordset[j][k] != 0) {      /* i <= j or disjoint */
                switch(r) {
                    case OSZ:
                    case OSZL:
                        r = OSZL;
                        break;

                    case OSZG:
                        r = OSDISJ;
                        break;

                    case OSGREATER:
                        r = OSOVLP;
                        goto ret;

                    case OSSAME:
                    case OSLESS:
                        r = OSLESS;
                        break;

                    case OSDISJ:
                        break;

                    case OSOVLP:
                    case OSANSC:
                    case OSDESC:
                        goto err;
                }
            } else {                     /* k \notin i,  k \notin j */
                switch(r) {
                    case OSOVLP:
                    case OSANSC:
                    case OSDESC:
                        goto err;

                    default:
                        break;
                }
            }
        }
    }
    switch(r) {
        case OSZ:
        case OSZL:
        case OSZG:
            fprintf(Stderr, "*** osinc : emplty set of orders\n");
            goto err;

        case OSDESC:
        case OSANSC:
            goto err;

        default:
            break;
    }

    ret:;
    return r;

    err:;
    fprintf(Stderr, "*** osinc(%d,%d) : illegal control\n", i, j);
    osdump(NULL);
    exit(FAIL);
}

/*==================================================================*/
Public  bool    osinit(sord)
int    *sord;
/*--------------------------------------------------------------------
 * Input:
 *     int sord[nsordtbl] : orders in sets.
 *
 * Output:
 *     bool return_value : 
 */
{
    int i, j, k;
    bool st = True;

    osconj[OSDISJ]    = OSDISJ;
    osconj[OSSAME]    = OSSAME;
    osconj[OSLESS]    = OSGREATER;
    osconj[OSGREATER] = OSLESS;
    osconj[OSOVLP]    = OSOVLP;
    osconj[OSZ]       = OSZ;
    osconj[OSZG]      = OSZL;
    osconj[OSZL]      = OSZG;

    for(i = 0; i < nsordtbl; i++) {
        if(i < nmordtbl) {
            osroot[i] = OSROOTL;
        } else if(sord[i] <= 0) {
            osroot[i] = OSEMPTY;
        } else {
            osroot[i] = OSROOTB;
        }
        sordelm[i] = 0;
        for(j = 0; j < nsordtbl; j++) {
            osincl[i][j] = OSZ;
        }
        for(j = 0; j < nmordtbl; j++) {
            if(sordset[i][j] != 0) {
                sordelm[i]++;
            }
        }
        if(i < nmordtbl) {
            if(sordelm[i] != 1) {
                fprintf(Stderr, "*** osinit: |set[%d]| = %d should be 1\n",
                        i, sordelm[i]);
                st = False;
            }
        } else {
            if(sordelm[i] <= 1) {
                fprintf(Stderr, "*** osinit: |set[%d]| = %d should be ",
                        i, sordelm[i]);
                fprintf(Stderr, "greater than 1\n");
                st = False;
            }
        }
    }
    if(!st) {
        osdump(sord);
        exit(FAIL);
    }

    /* Phase 1. */
    for(i = 0; i < nsordtbl; i++) {
        if(osroot[i] == OSEMPTY) {
            continue;
        }
        osincl[i][i] = OSSAME;
        for(j = i+1; j < nsordtbl; j++) {
            if(osroot[j] == OSEMPTY) {
                continue;
            }
            osincl[i][j] = osinc(i, j);
            osincl[j][i] = osconj[osincl[i][j]];
        }
    }

    /* Phase 2. */
    for(i = 0; i < nsordtbl; i++) {
        if(osroot[i] == OSEMPTY) {
            continue;
        }
        for(j = i+1; j < nsordtbl; j++) {
            switch(osincl[i][j]) {
                case OSOVLP:
                    st = False;
                    break;

                case OSDISJ:
                case OSDESC:
                case OSANSC:
                    break;

                case OSLESS:   /* set-i < set-j */
                    osroot[i] &= ~OSROOT;
                    for(k = 0; k < nsordtbl; k++) {
                        if(osincl[i][k] == OSLESS) {
                            if(osincl[j][k] == OSGREATER) {
                                /* set-i < set-k < set-j */
                                osincl[j][i] = OSANSC;
                                osincl[i][j] = OSDESC;
                            }
                        }
                    }
                    break;

                case OSGREATER:  /* set-i > set-j */
                    osroot[i] &= ~OSROOT;
                    for(k = 0; k < nsordtbl; k++) {
                        if(osincl[j][k] == OSLESS) {
                            if(osincl[i][k] == OSGREATER) {
                                /* set-i > set-k > set-j */
                                osincl[i][j] = OSANSC;
                                osincl[j][i] = OSDESC;
                            }
                        }
                    }
                    break;

                case OSSAME:   /* set-i < set-j, for i < j */
                    osroot[i] &= ~OSROOT;
                    osincl[i][j] = OSLESS;
                    osincl[j][i] = OSGREATER;
                    for(k = i+1; k < j; k++) {
                        if(osincl[i][k] == OSSAME) {
                            osincl[j][i] = OSANSC;
                            osincl[i][j] = OSDESC;
                        }
                    }
                    break;
            }
        }
    }
    return st;
}

/*==================================================================*/
Public  Void    osmessage(NoArg)
/*--------------------------------------------------------------------
 */
{
    int i, j, k;
    bool msg = False;
    int  msgk;

    for(i = 0; i < nsordtbl-1; i++) {
        for(j = i + 1; j < nsordtbl; j++) {
            if(osincl[i][j] == OSOVLP) {
                fprintf(Stderr, "*** OrdSet : %s and %s ",
                        sordtbl[i], sordtbl[j]);
                fprintf(Stderr, "cannt be specified ");
                fprintf(Stderr, "at the same time\n");
            }
        }
    }
    for(i = 0; i < nsordtbl; i++) {
        sordr[i] = 1;
    }
    Discard osinit(sordr);
    for(i = 0; i < nsordtbl; i++) {
        msg  = False;
        for(k = sordelm[i] + 1; k <= nmordtbl; k++) {
            msgk = 0;
            for(j = 0; j < nsordtbl; j++) {
                if(sordelm[j] != k) {
                    continue;
                }
                if(osincl[i][j] == OSLESS ||
                   osincl[i][j] == OSDESC) {
                    if(!msg) {
                        fprintf(Stderr, "When %s is specified, ", sordtbl[i]);
                        fprintf(Stderr, "one can specify\n");
                        msg = True;
                    }
                    if(msgk == 0) {
                        fprintf(Stderr, "\tat most one of {");
                        msgk = 1;
                    }
                    if(msgk == 2) {
                        fprintf(Stderr, ",");
                    }
                    msgk = 2;
                    fprintf(Stderr, "%s", sordtbl[j]);
                }
            }
            if(msgk != 0) {
                fprintf(Stderr, "}\n");
            }
        }
    }
}

/*==================================================================*/
Public  Void    osdmpa(msg, a, n)
string  msg;
int     *a;
int     n;
/*--------------------------------------------------------------------
 */
{
    int i;

    printf("%s : %4d :",  msg, n);
    for(i = 0; i < n; i++) {
        printf(" %2d,", i);
    }
    printf("\n");
    printf("%s :      :", msg);
    for(i = 0; i < n; i++) {
        printf(" %2d,", a[i]);
    }
    printf("\n\n");
}

/*==================================================================*/
Public  Void    osdump(sord)
int    *sord;
/*--------------------------------------------------------------------
 */
{
    int i, j;

    printf("+++ dump OrdSet: nmordtbl=%d, nsordtbl=%d\n",
           nmordtbl, nsordtbl);
    printf("--- id, sordtbl, sordelm, sordset\n");
    for(i = 0; i < nsordtbl; i++) {
        printf("%2d:%-8s:%3d:", i, sordtbl[i], sordelm[i]);
        for(j = 0; j < nmordtbl; j++) {
            printf(" %2d", sordset[i][j]);
        }
        printf("\n");
    }
    printf("\n--- id, sordtbl, order, osroot, osincl\n");
    for(j = 0; j < 17; j++) {
        putchar(' ');
    }
    for(j = 0; j < nsordtbl; j++) {
        printf(" %1d", (j%10));
    }
    printf("\n");
    for(i = 0; i < nsordtbl; i++) {
        printf("%2d:%-8s:", i, sordtbl[i]);
        if(sord == NULL) {
            printf("--:");
        } else {
            printf("%2d:", sord[i]);
        }
        printf("%c:", osnodc[osroot[i]]);
        for(j = 0; j < nsordtbl; j++) {
            printf(" %c", osincc[osincl[i][j]]);
        }
        printf("\n");
    }
}

/*==================================================================*/
Private int     ostree(sord, gord, s, r0)
int    *sord;
int    *gord;
int     s;
int     r0;
/*--------------------------------------------------------------------
 * input :
 *     int sord[nsordtbl]  : list of available orders of sets.
 *     int gord[nmordtbl]  : list of used orders of coupling const.
 *     int s               : the set to be visited.
 *     int r0              : additional order to sord to be assigned
 *                           to the descenderes of 's'.
 * ouput :
 *     int mleft[nsordtbl] : order of coupling constants not used.
 *     int return_value    : additional order to sord to be assigned
 *                           to the other sets.
 */
{
    int r1, r, t, u, w;

#if(TRACE >= 1)
    static int level = 0;

    printf("begin %3d: ostree(sord, gord, %d, %d)\n", ++level, s, r0);
# if(TRACE >= 2)
    osdmpa("sord ", sord,  nsordtbl);
    osdmpa("gord ", gord,  nsordtbl);
    osdmpa("mleft", mleft, nsordtbl);
# endif
#endif

    /* the order of coupling constant to be allocated */
    r = r1 = r0 + sord[s];

    /* leaf */
    if(s < nmordtbl) {
        r -= gord[s];
        if(r < 0) {
#if(TRACE >= 1)
            printf("end L %3d: ostree(sord, gord, %d, %d)\n", level--, s, r0);
# if(TRACE >= 2)
            osdmpa("sord ", sord,  nsordtbl);
            osdmpa("gord ", gord,  nsordtbl);
            osdmpa("mleft", mleft, nsordtbl);
# endif
#endif
            return r;
        }
    } else {

        /* branch : 1. consume order in the amount of gord */
        u = 0;
        for(t = 0; t < nsordtbl; t++) {
            if(osincl[s][t] != OSGREATER) {
                continue;
            }
            r = ostree(sord, gord, t, r);
            if(r < 0) {
#if(TRACE >= 1)
                printf("end B %3d: ostree(sord, gord, %d, %d)\n", level--, s, r0);
# if(TRACE >= 2)
                osdmpa("sord ", sord,  nsordtbl);
                osdmpa("gord ", gord,  nsordtbl);
                osdmpa("mleft", mleft, nsordtbl);
# endif
#endif
                return r;
            }
        }
    }

    u = r1 - r;    /* used order */
    if(u < 0) {
        fprintf(Stderr, "*** ostree : illegal control\n");
        osdump(sord);
        osdmpa("mleft ", mleft, nsordtbl);
        fprintf(Stderr, "r0 = %d, r1 = %d, r = %d, s = %d, u = %d\n",
                r0, r1, r, s, u);
        osdmpa("sord ", sord,  nsordtbl);
        osdmpa("gord ", gord,  nsordtbl);
        osdmpa("mleft", mleft, nsordtbl);
        fprintf(Stderr, "*** ostree : illegal control\n");
        exit(FAIL);
    }
    /* w <= r0, w + mleft[s] = r */
    if(sord[s] < u) {
        mleft[s] = 0;   /* all of sord[s] is used */
        w = r;
    } else {
        mleft[s] = sord[s] - u;   /* not all of sord[s] is used */
        w = r0;                   /* rest of sord[s] is moved to mleft[s] */
    }

#if(TRACE >= 1)
    printf("end   %3d: ostree(sord, gord, %d, %d)\n", level--, s, r0);
    printf("r0 = %d, r1 = %d, r = %d, sord[%d] = %d, mleft[%d]=%d, w =%d\n",
            r0, r1, r, s, sord[s], s, mleft[s], w);
# if(TRACE >= 2)
    osdmpa("sord ", sord,  nsordtbl);
    osdmpa("gord ", gord,  nsordtbl);
    osdmpa("mleft", mleft, nsordtbl);
# endif
#endif

    return w;
}

/*==================================================================*/
Private int     osatree(sord, gord, s, r0)
int    *sord;
int    *gord;
int     s;
int     r0;
/*--------------------------------------------------------------------
 * input :
 *     int sord[nsordtbl]  : list of orders of sets.
 *     int gord[nmordtbl]  : list of orders of coupling const.
 *     int s               : the set to be visited.
 *     int r0              : additional order to sord to be assigned
 *                           to the descenderes of 's'.
 * ouput :
 *     int mused           : used order
 *     int mleft[nsordtbl] : order of coupling constants not used.
 *     int return_value    : additional order to sord to be assigned
 *                           to the other sets.
 */
{
    int r, t;

#if(TRACE >= 1)
    static int level = 0;

    printf("begin %3d: osatree(sord, gord, %d, %d)\n", ++level, s, r0);
# if(TRACE >= 2)
    osdmpa("sord ", sord,  nsordtbl);
    osdmpa("gord ", gord,  nsordtbl);
    osdmpa("mleft", mleft, nsordtbl);
# endif
#endif

    /* the order of coupling constant to be allocated */
    r = r0 + mleft[s];

    /* leaf */
    if(s < nmordtbl) {
        mleft[s] = r;
        r = 0;
    } else {

        /* branch : 1. consume order in the amount of gord */
        for(t = 0; t < nsordtbl; t++) {
            if(osincl[s][t] != OSGREATER) {
                continue;
            }
            r = osatree(sord, gord, t, r);
            mleft[s] = 0;
        }
    }
#if(TRACE >= 1)
    printf("end   %3d: osatree(sord, gord, %d, %d)\n", ++level, s, r0);
# if(TRACE >= 2)
    osdmpa("sord ", sord,  nsordtbl);
    osdmpa("gord ", gord,  nsordtbl);
    osdmpa("mleft", mleft, nsordtbl);
# endif
#endif

    return r;
}

/*==================================================================*/
Public  int     oscheck(sord, gord, cut)
int    *sord;
int    *gord;
bool    cut;
/*--------------------------------------------------------------------
 * input :
 *     int sord[nsordtbl]  : list of available orders of sets.
 *     int gord[nmordtbl]  : list of used orders of coupling const.
 *     bool cut            : if True then stop when all order is
 *                           found not be used completely.
 * ouput :
 *     int return_value    : remaining order of sets.
 */
{
    int s, r, t = 0, c = 0;

#if(TRACE >= 1)
    printf("oscheck : 0: begin\n");
# if(TRACE >= 2)
    osdump(sord);
# endif
#endif

    /* for roots */
    for(s = 0; s < nsordtbl; s++) {
        if((osroot[s]&OSROOT) == 0) {
            continue;
        }
        c++;
        r = ostree(sord, gord, s, 0);
        if(r < 0) {
#if(TRACE >= 1)
            printf("oscheck : 1: return %d\n", 0);
#endif
            return r;
        } else if(cut && r != 0) {
#if(TRACE >= 1)
            printf("oscheck : 2: return %d\n", r);
#endif
            return r;
        }
        t += r;
    }
    if(c < 1) {
        printf("*** oscheck : no root!!!\n");
        exit(FAIL);
    }
#if(TRACE >= 1)
    printf("oscheck : 3: return %d\n", t);
#endif
    return t;
}

/*==================================================================*/
Public  int     osleft(cleft, sord, gord)
int    *cleft;
int    *sord;
int    *gord;
/*--------------------------------------------------------------------
 * Assume \sum gord[] <= \sum sord[]
 *
 * input :
 *     int sord[nsordtbl]  : list of orders of sets.
 *     int gord[nmordtbl]  : list of orders of coupling const.
 *                           to the descenderes of 's'.
 * ouput :
 *     int cleft[nmordtbl] : order of coupling constants not used.
 *     int return_value    : additional order to sord to be assigned
 *                           to the other sets, which should be 0.
 */
{
    int s, r = -1, c = 0;

    for(s = 0; s < nsordtbl; s++) {
        mleft[s] = 0;
    }
    for(s = 0; s < nsordtbl; s++) {
        if((osroot[s]&OSROOT) == 0) {
            continue;
        }
        c++;
        r = ostree(sord, gord, s, 0);
        if(r < 0) {
#if(TRACE >= 1)
            printf("osleft : 1: return %d\n", -1);
#endif
            return -1;
        }
    }
    if(c < 1) {
        printf("*** osleft : no root!!!\n");
        exit(FAIL);
    }

    for(s = 0; s < nsordtbl; s++) {
        if((osroot[s]&OSROOT) == 0) {
            continue;
        }
        r = osatree(sord, gord, s, 0);
        if(r != 0) {
#if(TRACE >= 1)
            printf("osleft : 2: return %d\n", -1);
#endif
            return -1;
        }
    }
    for(s = 0; s < nmordtbl; s++) {
        cleft[s] = mleft[s];
    }
#if(TRACE >= 1)
    printf("osleft : 3: return %d\n", 0);
#endif
    return 0;
}

#ifdef TESTMAIN

#define NTEST  6
#define NMORDTBL  4
#define NSORDTBL  ((1 << NMORDTBL) -1)
int gord[MAXMORD];

/*==================================================================*/
Void setup(mode)
int  mode;
/*--------------------------------------------------------------------
 */
{
    char str[100];
    int  i, k, l, m, c, n;

    nmordtbl = NMORDTBL;
    nsordtbl = NSORDTBL;

    for(i = 0; i < nmordtbl; i++) {
        for(l = 0; l < nmordtbl; l++) {
            sordset[i][l] = 0;
        }
        sordset[i][i] = 1;
    }
    n = nmordtbl;
    for(i = 2; i <= nsordtbl; i++) {
        c = i;
        k = 0;
        for(l = 0; l < nmordtbl; l++) {
            m = c % 2;
            c = c / 2;
            sordset[n][l] = m;
            k += m;
        }
        if(k > 1) {
            n++;
        }
        if(n >= nsordtbl) {
            break;
        }
    }

    for(i = 0; i < nsordtbl; i++) {
        sprintf(str, "s%2.2d", i);
        sordtbl[i] = strdup(str);
    }

    for(i = 0; i < nmordtbl; i++) {
        gord[i] = 2;
    }
    for(i = 0; i < nsordtbl; i++) {
        sord[i] = 0;
    }
    if(mode == 0) {
        sord[ 0] = 2;
        sord[ 3] = 2;
        sord[11] = 2;
        sord[13] = 2;
    } else if(mode == 1) {
        sord[ 0] = 1;
        sord[ 3] = 3;
        sord[11] = 2;
        sord[13] = 2;
    } else if(mode == 2) {
        sord[ 0] = 2;
        sord[ 1] = 1;
        sord[ 2] = 2;
        sord[ 3] = 1;
    } else if(mode == 3) {
        sord[ 4] = 4;
        sord[ 6] = 6;
    } else if(mode == 4) {
        sord[14] = 10;
    } else {
        sord[ 0] = 2;
        sord[ 3] = 2;
        sord[11] = 4;
        sord[13] = 2;
        sord[14] = 2;
    }
}

/*==================================================================*/
int main()
/*--------------------------------------------------------------------
 */
{
    int i, r;

    for(i = 0; i < NTEST; i++) {
/*
    i = 5; {
*/
        printf("*** Test No. %d\n", i);
        setup(i);
        if(osinit(sord)) {
            r = oscheck(sord, gord, False);
            printf("oscheck(False) = %d\n", r);

            r = oscheck(sord, gord, True);
            printf("oscheck(True) = %d\n", r);
            osdmpa("sord ", sord, nsordtbl);
            osdmpa("mleft", mleft, nsordtbl);
            osdmpa("gord ", gord, nmordtbl);

            r = osleft(mleft, sord, gord);
            printf("osleft()      = %d\n", r);
            osdmpa("sord ", sord, nsordtbl);
            osdmpa("mleft", mleft, nsordtbl);
            osdmpa("gord ", gord, nmordtbl);
        } else {
            printf("osinit : Failed\n");
            osmessage();
        }
    }
    return 0;
}
#endif
