/*********************************************************************
 * File : pset.c (prefix = ps)
 *
 * read particle table
 *                                                 by Kaneko Toshiaki
 *
 * 1995/09/08
 *    Ordering conventison is changed in the definiton of vertices.
 *    See History.
 *
 *    Functions:
 *        "pvnenter" : modified
 *        "pvertex"  : modified
 *        "pvncheck" : created
 *        "pvlord"   : modified
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "grdata.h"
#include "pset.h"


/*--------------------------------------------------------------------
 * parameters
 */
#define PVNEWN  0
#define PVOLDN  1
Extern  nindex  pvfindnam       ARG((string,int,int));
Extern  string  pvname          ARG((pindex));


/*--------------------------------------------------------------------
 * Local functions
 */
#define Pred    FTYPE((int (*)(int)))
Private int    psnot           ARG((int));
Private int    psand           ARG((int, int));
Private int    psor            ARG((int, int));
Private int    psdup           ARG((string, int));
Private int    pscond          ARG((int, int(*)(int), PS_OPR, int));

Private int    psptype         ARG((int));
Private int    pscolor         ARG((int));
Private int    pscharge3       ARG((int));
Private int    psmassless      ARG((int));
Private int    psantip         ARG((int));
Private int    psparticle      ARG((int));

#ifdef PSETDBG
Private int    pstoantip       ARG((int));
Private int    psexista        ARG((int));
#endif

/*********************************************************************
*/
/*==================================================================*/
Public  Void    psinit(NoArg)
/*--------------------------------------------------------------------
 * Add sets of particle to the particle-table.
 */
{
    ps_all   = psnew("all", 1);
    ps_empty = psnot(psdup("empty", ps_all));

    ps_color    = pscond(psnew("colored",  1), pscolor,    PS_NE, 1);
    ps_nocolor  = pscond(psnew("nocolor",  1), pscolor,    PS_EQ, 1);

    ps_charge   = pscond(psnew("charged",  1), pscharge3,  PS_NE,  0);
    ps_nocharge = pscond(psnew("nocharge", 1), pscharge3,  PS_EQ,  0);
    ps_charge3  = pscond(psnew("charge1",  1), pscharge3,  PS_EQ,  3);
    ps_charge_3 = pscond(psnew("charge-1",  1),pscharge3,  PS_EQ, -3);
    ps_charge2  = pscond(psnew("charge23", 1), pscharge3,  PS_EQ,  2);
    ps_charge_2 = pscond(psnew("charge-23", 1),pscharge3,  PS_EQ, -2);
    ps_charge1  = pscond(psnew("charge13",1),  pscharge3,  PS_EQ,  1);
    ps_charge_1 = pscond(psnew("charge-13",1), pscharge3,  PS_EQ, -1);

    ps_massive  = pscond(psnew("massive",  1), psmassless, PS_EQ, 0);
    ps_massless = pscond(psnew("massless", 1), psmassless, PS_NE, 0);

    ps_particle = pscond(psnew("particle", 1),      psparticle, PS_EQ, 1);
    ps_antip    = pscond(psnew("anti-particle", 1), psantip,    PS_EQ, 1);

    ps_scalar    = pscond(psnew("scalar", 1),   psptype, PS_EQ, P_SCALAR);
    ps_majorana  = pscond(psnew("majorana", 1), psptype, PS_EQ, P_MAJORANA);
    ps_dirac_all = pscond(psnew("dirac-all", 1),psptype, PS_EQ, P_DIRAC);
    ps_dirac     = psand( psdup("dirac",        ps_dirac_all), ps_particle);
    ps_adirac    = psand( psdup("anti-dirac",   ps_dirac_all), ps_antip);
    ps_ghost     = pscond(psnew("ghost", 1),    psptype, PS_EQ, P_GHOST);
    ps_vector    = pscond(psnew("vector", 1),   psptype, PS_EQ, P_VECTOR);
    ps_not_ghost = psnot(psdup("not-ghost",     ps_ghost));

    ps_boson      = psor(psdup( "boson",        ps_scalar),    ps_vector);
    ps_fermion    = psor(psdup( "fermion",      ps_majorana),  ps_dirac_all);

    ps_lepton_all = psand(psdup("lepton-all",   ps_fermion),   ps_nocolor);
    ps_lepton     = psand(psdup("lepton",       ps_lepton_all),ps_particle);
    ps_alepton    = psand(psdup("anti-lepton",  ps_lepton_all),ps_antip);

    ps_quark_all  = psand(psdup("quark-all",    ps_fermion),   ps_color);
    ps_quark      = psand(psdup("quark",        ps_quark_all), ps_particle);
    ps_aquark     = psand(psdup("anti-quark",   ps_quark_all), ps_antip);

    ps_hadron     = psand(psdup("hadron",       ps_color),     ps_not_ghost);
}

/*==================================================================*/
Public  int     psnew(name, val)
string  name;
int     val;
/*--------------------------------------------------------------------
 *  Create a new set.  each bit is set to val (1 or 0)
 */
{
    int  i, np = nptable;

    if(nppword < 1) {
        fprintf(Stderr, "*** psnew: Illegal nppword = %d\n", nppword);
        erfail();
    }

    /* check table size */
    if(nptable >= LPTABLE) {
        fprintf(Stderr, "\n*err* psnew: particle table overflow\n");
        erfail();
    }

    /* create space for new particle */
    nptable++;
    if((ptable[np] = (tptable*) malloc(sizeof(tptable))) == NULL) {
        fprintf(Stderr, "\n*err* psnew: memory allocation error\n");
        erfail();
    }

    /* register particle name */
    ptable[np]->pname[0] = pvfindnam(name,  np, PVNEWN);
    ptable[np]->pname[1] = ptable[np]->pname[0];

    ptable[np]->psym[0] = ntable[ptable[np]->pname[0]].name;
    ptable[np]->psym[1] = ptable[np]->psym[0];

    ptable[np]->antip = 1;

    ptable[np]->mass     = "0";
    ptable[np]->mdimn    = 0;
    ptable[np]->width    = "0";
    ptable[np]->wdimn    = 0;
    ptable[np]->ptype    = P_SET;
    ptable[np]->charge3  = 0;
    ptable[np]->color    = 1;
    ptable[np]->pcode    = 0;
    ptable[np]->kfcode   = 0;
    ptable[np]->furry    = False;
    ptable[np]->massless = False;
    ptable[np]->chane    = NULL;
    ptable[np]->chani    = NULL;
    ptable[np]->sname    = NULL;
    ptable[np]->snami    = NULL;
    ptable[np]->rname    = NULL;
    ptable[np]->rnami    = NULL;
    ptable[np]->gauge    = NULL;
    ptable[np]->pselect  = NULL;

    ptable[np]->pset = (Ulong*) malloc(sizeof(Ulong)*nppword);
    if(ptable[np]->pset == NULL) {
        fprintf(Stderr, "*** psnew : Memory\n");
        erfail();
    }
    for(i = - npptable + 1; i < npptable; i++) {
        psput(np, i, val);
    }

    return np;
}

/*==================================================================*/
Private int     psnot(arg)
int     arg;
/*--------------------------------------------------------------------
 *  arg = NOT arg
 */
{
    int i;

    for(i = - npptable + 1; i < npptable; i++) {
        if(psget(arg, i)) {
            psput(arg, i, 0);
        } else {
            psput(arg, i, 1);
        }
    }

    return arg;
}

/*==================================================================*/
Private int     psand(arg0, arg1)
int     arg0;
int     arg1;
/*--------------------------------------------------------------------
 *  arg0 = arg0 AND arg1
 */
{
    int i;

    for(i = - npptable + 1; i < npptable; i++) {
        if(psget(arg0, i)) {
            if(!psget(arg1, i)) {
                psput(arg0, i, 0);
            }
        }
    }

    return arg0;
}

/*==================================================================*/
Private int     psor(arg0, arg1)
int     arg0;
int     arg1;
/*--------------------------------------------------------------------
 *  arg0 = arg0 OR arg1
 */
{
    int i;

    for(i = - npptable + 1; i < npptable; i++) {
        if(!psget(arg0, i)) {
            if(psget(arg1, i)) {
                psput(arg0, i, 1);
            }
        }
    }

    return arg0;
}

/*==================================================================*/
Private int     psdup(name, set)
string  name;
int     set;
/*--------------------------------------------------------------------
 *  Duplicate a set with name
 */
{
    int ans;

    ans = psnew(name, 1);
    ans = psand(ans, set);

    return ans;
}

/*==================================================================*/
Private int     pscond(set, fget, opr, val)
int     set;
int     (*fget) ARG((int));
PS_OPR  opr;
int     val;
/*--------------------------------------------------------------------
 *  make a subset of set with a condition.
 *  set = {x in set | fget(x) opr val}
 */
{
    int  i, pval;
    bool cond = False, ocond;

    for(i = - npptable + 1; i < npptable; i++) {
        ocond = psget(set, i);
        if(!ocond) {
            continue;
        }
        pval = fget(i);
        switch (opr) {
            case PS_EQ:
                cond = (pval == val);
                break;

            case PS_NE:
                cond = (pval != val);
                break;

            case PS_LT:
                cond = (pval < val);
                break;

            case PS_GT:
                cond = (pval > val);
                break;

            case PS_LE:
                cond = (pval <= val);
                break;

            case PS_GE:
                cond = (pval >= val);
                break;

            default:
                fprintf(Stderr, "*** pscond: illegal operator %d\n", opr);
                erfail();
                break;
        }
        if(ocond && !cond) {
            psput(set, i, cond);
        } else if(!ocond && cond) {
            psput(set, i, cond);
        }
    }
    return set;
}

#ifdef PSETDBG
/*==================================================================*/
Private int     pstoantip(arg)
int     arg;
/*--------------------------------------------------------------------
 *  arg = { -p | p in arg, psexita(p) != 0}
 */
{
    int i;

    for(i = - npptable + 1; i < npptable; i++) {
        if(psget(arg, i) && psexista(i) != 0) {
            psput(arg, -i, 1);
        }
    }

    return arg;
}
#endif

/*==================================================================*/
Private int    psptype(p)
int     p;
/*--------------------------------------------------------------------
 *  returns the value of ptype of paritcle p.
 */
{
    return ptable[abs(p)]->ptype;
}

/*==================================================================*/
Private int    pscolor(p)
int     p;
/*--------------------------------------------------------------------
 *  returns the value of color of paritcle p.
 */
{
    return ptable[abs(p)]->color;
}

/*==================================================================*/
Private int    pscharge3(p)
int     p;
/*--------------------------------------------------------------------
 *  returns the value of charg3 of paritcle p.
 */
{
    return ptable[abs(p)]->charge3;
}

/*==================================================================*/
Private int    psmassless(p)
int     p;
/*--------------------------------------------------------------------
 *  returns the value of massless of paritcle p.
 */
{
    return ptable[abs(p)]->massless;
}

/*==================================================================*/
Private int    psparticle(p)
int     p;
/*--------------------------------------------------------------------
 *  returns the value of massless of paritcle p.
 */
{
    return (p > 0);
}

/*==================================================================*/
Private int    psantip(p)
int     p;
/*--------------------------------------------------------------------
 *  returns the value of massless of paritcle p.
 */
{
    return (p < 0 && ptable[-p]->antip < 0);
}

#ifdef PSETDBG
/*==================================================================*/
Private int    psexista(p)
int     p;
/*--------------------------------------------------------------------
 *  returns 0 if the particle p has not have the anti-particle,
 *  otherwise -p (p may be anti-particle)
 */
{
    if(ptable[-p]->antip < 0) {
        return -p;
    } else {
        return 0;
    }
}
#endif

/*==================================================================*/
Public  bool psget(set, pos)
int     set;
int     pos;
/*--------------------------------------------------------------------
 *  returns the value of massless of paritcle p.
 */
{
    int wrd, offs, spos;
    Ulong mask;

    if(set < 0 || set >= nptable || ptable[set]->ptype != P_SET) {
        fprintf(Stderr, "*** psget: illegal set %d\n", set);
        erfail();
    }

    spos = pos + npptable - 1;
    wrd  = spos / nppbits;
    offs = spos - nppbits*wrd;
    if(wrd < 0 || wrd >= nppword || offs < 0) {
        fprintf(Stderr, "*** psget: illegal position %d\n", pos);
        erfail();
    }
    mask = 1 << offs;
    mask &= ptable[set]->pset[wrd];

    if(mask == 0) {
        return False;
    } else {
        return True;
    }
}

/*==================================================================*/
Public  Void  psput(set, pos, val)
int     set;
int     pos;
int     val;
/*--------------------------------------------------------------------
 *  returns the value of massless of paritcle p.
 */
{
    int   wrd, offs, p, s, spos;
    Ulong mask;

    if(set < 0 || set >= nptable || ptable[set]->ptype != P_SET) {
        fprintf(Stderr, "*** psput: illegal set %d\n", set);
        erfail();
    }

    if(pos < 0) {
        p = - pos;
        s = - 1;
    } else {
        p = pos;
        s = 1;
    }
    if(pos < 0 && !psantip(pos)) {
        val = False;
    }
    if(ptable[p]->ptype == P_UNDEF || ptable[p]->ptype == P_SET) {
        val = False;
    }

    spos = pos + npptable - 1;
    wrd  = spos / nppbits;
    offs = spos - nppbits*wrd;
    if(wrd < 0 || wrd >= nppword || offs < 0) {
        fprintf(Stderr, "*** psput: illegal position %d/%d\n", pos, npptable);
        erfail();
    }

    mask = 1 << offs;
    if(val) {
        ptable[set]->pset[wrd] |= mask;
    } else {
        ptable[set]->pset[wrd] &= ~mask;
    }
}

/*==================================================================*/
Public  Void  psdump(NoArg)
/*--------------------------------------------------------------------
 *  Dump set of particles
 */
{
    int p, i, nset=0, len;
    string s;

    printf("\n+++ Dump sets of particles +++\n");
    for(p = 0; p < nptable; p++) {
        if(ptable[p]->ptype != P_SET) {
            continue;
        }

        s = pvname(p);
        printf("%6d: %s={", p, s);
        len  = strlen(s) + 6 + 4;
        nset = 0;
        for(i = - npptable + 1; i < npptable; i++) {
            if(psget(p, i)) {
                if(nset != 0) {
                    printf(", ");
                    len += 2;
                }
                s = pvname(i);
                len += strlen(s);
                if(len > 64) {
                    printf("\n            ");
                    len = 12 + strlen(s);
                }
                printf("%s", s);
                nset++;
            }
        }
        printf("};\n");
    }
}
