/*********************************************************************
 *  File :  ptable.h
 *
 *  Table of particles
 *
 *                                              by Kaneko Toshiaki
 */

#ifndef PNULL

/*
#ifndef DBG
#define DBG 2
Global int dbg;
Global int pvdbg;
Global int rddbg;
#endif
*/
/*===================================================================
 *  Indices of tables for model definition.
 */
typedef int     pindex;     /* index in the particle table */
                            /* +(particle) or -(antiparticle) */
#define PNULL   0           /* undefined partile as a value of pindex */

typedef int     nindex;     /* index in the particle name table */

typedef int     vindex;         /* index in the vertex table */
#define VNULL   0               /* undefined vertex */
#define VSELF   -1              /* undefined vertex */


/*===================================================================
 *  Table of particle names.
 */

/* #define LNTABLE (2*LPTABLE) */ /* size of table of particle name */

/*===================================================================
 *  Table of definitions of coupling constants.
 */

#define MAXMORD  50      /* number of kinds of order */
#define MAXSORD  100     /* number of set of orders */

/*===================================================================
 *  Maximum number of indices of dimension.
 */
#define MAXFDIM   7

/*===================================================================
 *  Table of particle definitions.
 */
/* The value of ptype */
#define P_UNDEF    0
#define P_SCALAR   1
#define P_MAJORANA 2
#define P_DIRAC    3
#define P_GHOST    4
#define P_VECTOR   5
#define P_AUX      6
#define P_SET      7          /* set of particles */
#define PTYPEMAX   8

#define M_UNDEF    1
#define M_SCALAR   (1 << P_SCALAR)
#define M_MAJORANA (1 << P_MAJORANA)
#define M_DIRAC    (1 << P_DIRAC)
#define M_GHOST    (1 << P_GHOST)
#define M_VECTOR   (1 << P_VECTOR)
#define M_AUX      (1 << P_AUX)
#define M_SET      (1 << P_SET)
#define M_BOSON    (M_SCALAR   | M_VECTOR | M_AUX)
#define M_DFERMION (M_DIRAC    | M_GHOST )   /* fermion with direction */
#define M_FERMION  (M_DFERMION | M_MAJORANA)

#ifdef DEFGLOBAL
 Global char   *ptypesym  = "?smfxv";
 Global string  pvtnam[PTYPEMAX] =
       {"undef", "scalar", "majorana", "fermion", "ghost", "vector", "aux",
        "set"};
 Global Uint    pmask[PTYPEMAX] =
       {M_UNDEF, M_SCALAR, M_MAJORANA, M_DIRAC, M_GHOST, M_VECTOR, M_AUX,
        M_SET};
#else
 Global char    *ptypesym;
 Global string  pvtnam[PTYPEMAX];
 Global Uint    pmask[PTYPEMAX];
#endif

typedef struct {            /* particle table */
        string  psym[2];    /* print name */
        string  mass;       /* fortran name for mass */
        string  width;      /* fortran name for width */
        string  gname[3];   /* name for graphic output */
        string  chane;      /* name of chanel for ext. part. */
        string  chani;      /* name of chanel for propagator */
        string  sname;      /* name of subroutine ext. part. */
        string  snami;      /* name of subroutine of propagator */
        string  rname;      /* optional name for reduce proj. op. */
        string  rnami;      /* optional name for reduce propagator */
        string  gauge;      /* name of gauge parameter */
        string  pselect;    /* name of graph selection */
        nindex  pname[2];   /* name of the [particle, anti-p] */
        int     mdimn;          /* dimension of mass */
        int     mdim[MAXFDIM];  /* dimension of mass */
        int     mind[MAXFDIM];  /* index of mass */
        int     wdimn;          /* dimension of width */
        int     wdim[MAXFDIM];  /* dimension of width */
        int     wind[MAXFDIM];  /* index of width */
        int     pcode;      /* particle code */
        int     kfcode;     /* KF particle code */
        int     nvlist[2];  /* length of vlist */
        int     antip;      /* (antiparticle == particle) ? 1 : -1 */
        int     charge3;    /* 3*spin */
        int     ptype;      /* type of the particle */
        int     color;      /* color multiplet */
        bool    massless;   /* massless or massive */
        bool    furry;      /* flag for Furry's theorem */
        bool    extself;    /* massless or massive */
        vindex *vlist[2];   /* list of vertices */
        Ulong  *pset;       /* set of particles */
}       tptable;
Global tptable *ptable[LPTABLE]; /* table of particles */
Global int      nptable;         /* the number of particles in ptable */
Global int      npptable;        /* the number of particles without sets */
Global int      nppbits;         /* the number of bits in Ulong */
Global int      nppword;         /* the number of words in "pset" */

/*===================================================================
 *  Table of particle names.
 */
typedef struct {
        char   *name;           /* name */
        pindex  pindx;          /* position in the ptable ??? */
}       tntable;

#define LNTABLE (2*LPTABLE)     /* size of table of particle name */
Global tntable  ntable[LNTABLE];  /* table of particle names */
Global int      nntable;          /* the number of data in the table */

#define N_MASS      0
#define N_WIDTH     1
#define N_SIZE      2

Global string   mtable[N_SIZE][LPTABLE];
Global int      nmtable[N_SIZE];

Global bool     defaux;

/*===================================================================
 *  Table of definitions of coupling constants.
 */

/*  MAXMORD : number of kinds of order, defined in "grdata.h" */

Global char    *mordtbl[MAXMORD];       /* name of coupling const */
Global int      mordopt[MAXMORD];       /* option flags. */
Global int      nmordtbl;               /* # of data in the table */
Global char    *sordtbl[MAXSORD];       /* name of set of coupling const */
Global int      sordset[MAXSORD][MAXMORD];    /* bits of set of coupling const */
Global int      sordelm[MAXSORD];       /* # of elements */
Global int      nsordtbl;               /* # of data in the table */

#define MOPT_QCD 1    /* flag in mordopt[] */

typedef struct tstitle {
    string tstr;
    struct tstitle *next;
} ttitle;

#ifdef DEFGLOBAL
Global ttitle *title = NULL;
#else
Global ttitle *title;
#endif

/* version and sub-version numbers */
#define MAXVER 4
Global int      version[MAXVER];

Global char    *mprefix;

/* phase of propagators
 * (i)**pphase/(p.p - m**2)
 *      pphase = 1 : standard
 *      pphase = 2 : Kyoto
 */
Global int      pphase;                 

/*-------------------------------------------------------------------
 *  Table of vertix definitions.
 */

#define MAXTL     4 /* The maximum number of legs of tree vertex.*/
#define MINTL     3 /* The minimum number of legs of tree vertex.*/

Global int maxtl, mintl;  /* Defined in model file */

#define VTTREE    0    /* tree vertex */
#define VTLOOP    1    /* vertex with loop */
#define VTGENCT   2    /* generated counter term */

#define V_NOTYPE    0
#define V_V         1
#define V_A         2
#define V_VminusA   3
#define V_VplusA    4

#define V_NORMAL    0
#define V_AUX       1
#define V_COMPOSIT  2

/* type of vertex : to be consistent with the ordering of ptype */
#define VT_DEF(u,v,w,x) ((((u)*PTYPEMAX+(v))*PTYPEMAX+(w))*PTYPEMAX+(x))

#define VT_SSS     VT_DEF(0, P_SCALAR,   P_SCALAR,   P_SCALAR)
#define VT_SSV     VT_DEF(0, P_SCALAR,   P_SCALAR,   P_VECTOR)
#define VT_SVV     VT_DEF(0, P_SCALAR,   P_VECTOR,   P_VECTOR)
#define VT_VVV     VT_DEF(0, P_VECTOR,   P_VECTOR,   P_VECTOR)

#define VT_SSA     VT_DEF(0, P_SCALAR,   P_SCALAR,   P_AUX)
#define VT_SVA     VT_DEF(0, P_SCALAR,   P_VECTOR,   P_AUX)
#define VT_VVA     VT_DEF(0, P_VECTOR,   P_VECTOR,   P_AUX)

#define VT_SMM     VT_DEF(0, P_SCALAR,   P_MAJORANA, P_MAJORANA)
#define VT_SMD     VT_DEF(0, P_SCALAR,   P_MAJORANA, P_DIRAC)
#define VT_SDM     VT_DEF(0, P_SCALAR,   P_DIRAC,    P_MAJORANA)
#define VT_SDD     VT_DEF(0, P_SCALAR,   P_DIRAC,    P_DIRAC)
#define VT_SGG     VT_DEF(0, P_SCALAR,   P_GHOST,    P_GHOST)

#define VT_MMV     VT_DEF(0, P_MAJORANA, P_MAJORANA, P_VECTOR)
#define VT_MDV     VT_DEF(0, P_MAJORANA, P_DIRAC,    P_VECTOR)
#define VT_DMV     VT_DEF(0, P_DIRAC,    P_MAJORANA, P_VECTOR)
#define VT_DDV     VT_DEF(0, P_DIRAC,    P_DIRAC,    P_VECTOR)
#define VT_GGV     VT_DEF(0, P_GHOST,    P_GHOST,    P_VECTOR)

#define VT_SSSS    VT_DEF(P_SCALAR, P_SCALAR, P_SCALAR, P_SCALAR)
#define VT_SSVV    VT_DEF(P_SCALAR, P_SCALAR, P_VECTOR, P_VECTOR)
#define VT_VVVV    VT_DEF(P_VECTOR, P_VECTOR, P_VECTOR, P_VECTOR)
#define VT_SSGG    VT_DEF(P_SCALAR, P_SCALAR, P_GHOST,  P_GHOST)
#define VT_GGVV    VT_DEF(P_GHOST,  P_GHOST,  P_VECTOR, P_VECTOR)

#define NFNAME     4

/* Vertex table */
typedef struct {
        int     vtype;              /* type of the vertex */
        int     nvlegs;             /* the number of legs */
        int     morder[MAXMORD];    /* table of coupling constants */
        int     loop;               /* loop in the vertex */
        int     ctype;              /* type of coupling */
        int     composit;           /* normal, aux, composit */
        int     fndimn[NFNAME];          /* dimension of coupling */
        int     fndim[NFNAME][MAXFDIM];  /* dimension of coupling */
        int     fnind[NFNAME][MAXFDIM];  /* index of coupling */
        int     auxp;               /* auxiliary p. for connection */
        int    *auxv;               /* auxiliary v. for connection */
        string  fname[NFNAME];      /* FName, CFName, EFName, CEFName */
        string  rname;              /* optional name for reduce */
        string  sname;              /* name of subroutine */
        string  chanel;             /* name of chanel subroutine */
        string  vselect;            /* name of graph selection */
        string  vid;                /* vertex identifier */
        pindex *vlegs;              /* list of particles */
}       tvtable;


Global int      nvtable;            /* the number of defined vertices */
Global int      nvmodel;            /* vetices defined in the model */
Global tvtable *vtable[LVTABLE];    /* table of vertices */
Global int      nvstart;            /* star of the table */
Global bool     majop;              /* majorana is defined or not */

Global string   ctable[LVTABLE];
Global int      vnamdef[LVTABLE];
Global int      nctable;

#endif
