/*********************************************************************
 *  File : putid.c        Prifix = "pu".
 *
 *  print id's in the following format
 *
 *      hdr1 id_1       sep1 id_2       sep1 ... sep1 id_{k_1} sep2
 *      hdr2 id_{k_1+1} sep1 id_{k_1+2) sep1 ... sep1 id_{k_2} sep2
 *             ...
 *      hdr2 id_{k_{n-1}+1} sep1 ... sep1 id_{k_n}} sep3
 *
 *      llength  = length of a line
 *      nlines   = maximum number of lines ("n" in the above example).
 *                 If there are more id's, data will be generated with
 *                 hdr1 again.
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "common.h"
#include "putid.h"

/*--------------------------------------------------------------------
 * Private functions
 */
Private Void    putid0   ARG((string,int));
Private Void    putline  ARG((NoArg));

/*--------------------------------------------------------------------
 * Local variable
 */
#define MAXSTR  64
#define MAXSTR1 (MAXSTR+1)

Local   FILE    *pufp;
Local   char     hdr1[MAXSTR1], hdr2[MAXSTR1];
Local   char     sep1[MAXSTR1], sep2[MAXSTR1], sep3[MAXSTR1];
Local   int      llength, nlines;
Local   string   line=NULL;
Local   int      lpos, npos, ipos;
Local   int      lhdr1, lhdr2, lsep1, lsep2, lsep3, msep2, msep3;

/*==================================================================*/
Public  Void    puinit(fp,hd1,hd2,sp1,sp2,sp3,llen,nlin)
FILE    *fp;
string  hd1, hd2;
string  sp1, sp2, sp3;
int     llen, nlin;
/*--------------------------------------------------------------------
 *  Initialization
 *------------------------------------------------------------------*/
{
    pufp = fp;

    if((lhdr1 = strlen(hd1)) > MAXSTR) {
        fprintf(Stderr,"hdr1 is too long %d > %d\n", lhdr1, MAXSTR);
        exit(1);
    }
    strncpy(hdr1, hd1, MAXSTR);

    if((lhdr2 = strlen(hd2)) > MAXSTR) {
        fprintf(Stderr,"hdr2 is too long %d > %d\n", lhdr2, MAXSTR);
        exit(1);
    }
    strncpy(hdr2, hd2, MAXSTR);

    if((lsep1 = strlen(sp1)) > MAXSTR) {
        fprintf(Stderr,"sep1 is too long %d > %d\n", lsep1, MAXSTR);
        exit(1);
    }
    strncpy(sep1, sp1, MAXSTR);

    if((lsep2 = strlen(sp2)) > MAXSTR) {
        fprintf(Stderr,"sep2 is too long %d > %d\n", lsep2, MAXSTR);
        exit(1);
    }
    strncpy(sep2, sp2, MAXSTR);

    if((lsep3 = strlen(sp3)) > MAXSTR) {
        fprintf(Stderr,"sep3 is too long %d > %d\n", lsep3, MAXSTR);
        exit(1);
    }
    strncpy(sep3, sp3, MAXSTR);

    llength = llen;
    nlines  = nlin;

    if(llength < lhdr1 || llength < lhdr2 ||
       llength < lsep1 || llength < lsep2 || llength < lsep3) {
        fprintf(Stderr, "puinit: too short line %d : \n",
                llength);
        fprintf(Stderr, "    (hdr) %d, %d\n",
                lhdr1, lhdr2);
        fprintf(Stderr, "    (sep) %d, %d, %d\n",
                lsep1, lsep2, lsep3);
        exit(1);
    }

    if(line != NULL) {
        free(line);
    }

    if((line = (string) malloc(sizeof(char)*(llength+1)))
       == NULL) {
        fprintf(Stderr, "puinit: memory allocation failed\n");
        exit(1);
    }

    msep2 = (lsep1 > lsep2) ? lsep1 : lsep2;
    msep3 = (lsep1 > lsep3) ? lsep1 : lsep3;

    lpos = 0;
    npos = 1;
    ipos = 0;
}

/*==================================================================*/
Public  Void    putid(id)
string  id;
/*--------------------------------------------------------------------
 *  put id
 *------------------------------------------------------------------*/
{
    int lid;

    lid = strlen(id);

    if(ipos == 0) {
        putid0(hdr1, lhdr1);
    }

    if(npos == nlines) {
        if(lpos+lid+msep3 >= llength) {
            putid0(sep3, lsep3);
            putline();
            putid0(hdr1, lhdr1);
            npos = 1;
            ipos = 0;
        } else if(ipos != 0) {
            putid0(sep1, lsep1);
        }
    } else if(lpos+lid+msep2 >= llength) {
        putid0(sep2, lsep2);
        putline();
        putid0(hdr2, lhdr2);
    } else if(ipos != 0) {
        putid0(sep1, lsep1);
    }
    putid0(id, lid);
    ipos++;
}

/*==================================================================*/
Public  Void    putend(NoArg)
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    if(ipos > 0) {
        putid0(sep3, lsep3);
        putline();
    }
    ipos = 0;
    lpos = 0;
    npos = 1;
}

/*==================================================================*/
Private Void    putid0(id, lid)
string  id;
int     lid;
/*--------------------------------------------------------------------
 *  put id
 *------------------------------------------------------------------*/
{
    int sp = llength - lpos;
    int c;

    if(sp < lid) {
        if(sp < 1) {
            putline();
            putid0(hdr2, lhdr2);
            putid0(id, lid);
            return;
        }
        fprintf(Stderr, "putid0 : too long id = %s (sp=%d,lid=%d)\n",
               id, sp, lid);
        c = id[sp];
        id[sp] = '\0';
        putid0(id, sp);
        putline();
        putid0(hdr2, lhdr2);
        id[sp] = c;
        putid0(id+sp, lid - sp);

    } else {
        strcpy(line+lpos, id);
        lpos += lid;
    }
}

/*==================================================================*/
Private Void    putline(NoArg)
/*--------------------------------------------------------------------
 *------------------------------------------------------------------*/
{
    fprintf(pufp, "%s\n", line);
    lpos = 0;
    npos++;
}
