/*********************************************************************
 * File : rdptbl.c (prefix = pv)
 *
 * read particle table
 *                                                 by Kaneko Toshiaki
 *
 * 1995/09/08
 *    Ordering conventison is changed in the definiton of vertices.
 *    See History.
 *
 *    Functions:
 *        "pvnenter" : modified
 *        "pvertex"  : modified
 *        "pvncheck" : created
 *        "pvlord"   : modified
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "grdata.h"
#include "version.h"
#include "config.h"
#include "path.h"
#include "token.h"

/*------------------------------------------------------------------*
 * parameters
 */
#define SHOWTITLE
#define CHANEL
#define OLDNAME
#define WARNNONSTAND
/*
#define OLDORD
#define VDEFCHECK
*/

#define MAXFNAM   10   /* maximum length of id names in fortran */
#define N3        3    /* fundamental representation of color group */
#define N8 (N3*N3 - 1) /* adjoint representation of color group */

/*------------------------------------------------------------------*
 * Local functions
 */
Private int     pvpenter     ARG((string,string,string,string));
Private string  pvmname      ARG((string,int));
Private string  pvvname      ARG((string,int,int*));
Private int     pvnenter     ARG((int,string*));
Private int     pcvnenter    ARG((int,string*));
Private Void    pvendtbl     ARG((NoArg));
Private Void    pvscale      ARG((int));
Private bool    pvparticle   ARG((NoArg));
Private bool    pvpnam       ARG((int, char*, char*));
Private bool    pvertex      ARG((NoArg));
Private bool    pcvertex     ARG((NoArg));
Private bool    pvfort       ARG((string));
Private string  pvstrdup     ARG((bool, string));
Private int     pvprtxp      ARG((NoArg));
Private int     pvprtv1      ARG((NoArg));
Private Void    pvprtv2      ARG((int));
Private string  pvctname     ARG((string,int*,int));
Private int     pvlord       ARG((pindex,pindex));
Private bool    pvfermiv     ARG((int));
Private bool    pvncheck     ARG((int));
Private bool    pvfnchk      ARG((NoArg));
/* change to Public 1998/11/5 TK
Private Void    pvpdump      ARG((int));
Private Void    pvndump      ARG((int));
*/
Private bool    pvcompgl4    ARG((NoArg));
#ifdef DEFSNAME
Private string  pvrname      ARG((int));
Private string  pvsname      ARG((int));
Private string  pvchanel     ARG((int));
#endif
#ifdef VDEFCHECK
Private bool    pvcpcpl      ARG((int,int));
#endif

#ifndef sign
#define sign(x)      (((x) <  0 ) ? (-1) : (1))
#endif

/*------------------------------------------------------------------*
 * Local variables
 */
#define PVNEWN  0
#define PVOLDN  1

Local   FILE    *prpf;
Local   FILE    *pvpf;
Local   int     pvind = 0;
Local   char    pvfnam[MAXPATHLEN];

#define MAXVAL 64

#define MAXCMD 32
Local char   cmd[MAXCMD], argstr[MAXCMD];

Local int    nonstand = 0;
Local int    *pvrdsln = NULL;

/* help message */
Local string syntax[] = {
"Model definition file",
"  Syntax :",
"",
"\t file   ::= [ order-definition | particle-definition ",
"\t\t | vertex-definition ]* \"Mend\" \";\".",
"",
"\t order-definition ::=",
"\t\t \"Order={\" coupling-name [ \",\" coupling-name ]* \"};\".",
"",
"\t particle-definition ::= particle [ anti-particle | mass |",
"\t\t width | ptype | charge | color | pcode | kfcode | massless | furry ]*",
"\t\t \"Pend;\".",
"\t particle ::= \"Particle=\" particle-name ",
"\t\t [ \"[\" print-name \"]\" ] \";\".",
"\t anti-particle ::= \"Antiparticle=\" particle-name ",
"\t\t [ \"[\" print-name \"]\" ] \";\".",
"\t mass   ::= \"Mass=\"   ( mass-name  | \"0\" ) \";\".",
"\t width  ::= \"Width=\"  ( width-name | \"0\" ) \";\".",
"\t ptype  ::= \"PType=\"  ( Scalar | Ferion | Ghost | Vector) \";\".",
"\t charge ::= \"Charge=\" rational-number \";\".",
"\t color  ::= \"Color=\"  integer-number  \";\".",
"\t pcode  ::= \"PCode=\"  integer-number  \";\".",
"\t kfcode ::= \"KFCode=\"  integer-number  \";\".",
"\t massless ::= \"Massless \";\".",
"\t furry  ::= \"Furry=\"  ( \"Yes\" | \"No\" )  \";\".",
"",
"\t vertex-definition ::= p-list [ v-order | v-name ]* \"Vend;\".",
"\t p-list  ::= \"Vertex={\" [particle-name]+ \"};\".",
"\t v-order ::= coupling-name \"=\" integer-number \";\".",
"\t v-name  ::= \"FName=\" vertex-name [ (rational-number, ... ) ] \";\".",
"\t vc-name ::= \"CFName=\" vertex-name [ (rational-number, ... ) ] \";\"." };

/*********************************************************************
  Read model definition file.
*/
/*==================================================================*/
Public  Void    pvfname(fname, fin, len)
string  fname;             /* normalized name */
string  fin;               /* name of input file */
int     len;
/*--------------------------------------------------------------------
 *  Normalize input model file name
 */
{
    string fullname;

    if((fullname = pnfname("GRCMODEL", modelpath, "", fin)) == NULL) {
        fprintf(Stderr, "*** pvfname : file not found %s\n", fin);
        exit(1);
    }
    if(strlen(fullname) >= len) {
        fprintf(Stderr, "*** pvfname : too long path name %d\n", len);
        exit(1);
    }
    strcpy(fname, fullname);
}

/*==================================================================*/
Public  Void    pvfenam(fname, fin, ext, maxlen)
string  fname;             /* normalized name */
string  fin;               /* name of input file */
string  ext;               /* name of input file */
int     maxlen;
/*--------------------------------------------------------------------
 *  Change extension part ".mdl" to ext.
 */
{
    int ln;

    ln = strlen(fin);
    if(ln >= 5 && strcmp(fin+ln-4, ".mdl") == 0) {
        ln -= 4;
    }
    if(ln+4 >= maxlen) {
        fprintf(Stderr, "*** pvfenam : length of fname is too small\n");
        exit(1);
    }
    strncpy(fname, fin, ln);
    fname[ln] = '\0';
    strcat(fname, ext);
}

/*==================================================================*/
Public  Void    pvread(fname)
string  fname;             /* name of input file */
/*--------------------------------------------------------------------
 *  Read particle and vertex table.
 *
 *  In the current version, this is test one.
 */
{
    int arg, opt, nopt;
    int i, j, n, nv;
    string vp[MAXTL];
    string str;
    bool err;
    ttitle *ttl = NULL;
    ttitle **ptitle;

#if DBG
    erdbg("pvread", "begin", pvdbg);
#if 1
    printf("VT_SSS  = %o\n", VT_SSS);
    printf("VT_SSV  = %o\n", VT_SSV);
    printf("VT_SVV  = %o\n", VT_SVV);
    printf("VT_VVV  = %o\n", VT_VVV);
    printf("VT_SSA  = %o\n", VT_SSA);
    printf("VT_SVA  = %o\n", VT_SVA);
    printf("VT_VVA  = %o\n", VT_VVA);
    printf("VT_SMM  = %o\n", VT_SMM);
    printf("VT_SMD  = %o\n", VT_SMD);
    printf("VT_SDD  = %o\n", VT_SDD);
    printf("VT_SGG  = %o\n", VT_SGG);
    printf("VT_MMV  = %o\n", VT_MMV);
    printf("VT_MDV  = %o\n", VT_MDV);
    printf("VT_DMV  = %o\n", VT_DMV);
    printf("VT_DDV  = %o\n", VT_DDV);
    printf("VT_GGV  = %o\n", VT_GGV);
    printf("VT_SSSS = %o\n", VT_SSSS);
    printf("VT_SSVV = %o\n", VT_SSVV);
    printf("VT_VVVV = %o\n", VT_VVVV);
    printf("VT_SSGG = %o\n", VT_SSGG);
    printf("VT_GGVV = %o\n", VT_GGVV);
#endif
#endif

    /* alloc workspace */
    if(pvrdsln == NULL) {
        if((pvrdsln = (int*) malloc(LVTABLE*sizeof(int))) == NULL) {
            fprintf(Stderr, "*** pvread : memory\n");
            exit(FAIL);
        }
    }

    strcpy(pvfnam, fname);

    if((prpf = fopen(pvfnam, "r")) == NULL) {
        fprintf(Stderr, "\n*err* pvread: Model file open err,\n");
        fprintf(Stderr, "    File name = %s\n", pvfnam);
        exit(1);
    }

#if DBG
    if(dbg >= pvdbg) {
        printf("pvread: model = \"%s\"\n", pvfnam);
        rdprint(pvind);
    }
#endif

    /* initialize token package */
#if DBG
    if(dbg >= pvdbg) {
        rdinit(prpf, RDNOCHANGE, stdout);
    } else {
#endif
        rdinit(prpf, RDNOCHANGE, (FILE*) NULL);
#if DBG
    }
#endif

    title    = NULL; /* titel */
    nntable  = 0; /* table of particle names */
    nptable  = 0; /* table of particles */
    nvtable  = 0; /* table of vertices */
    nmordtbl = 0; /* table of coupling constants */
    nctable  = 0; /* table of names of coupling const of vertex */
    defaux   = False;  /* AUX particles are included */
    for(i = 0; i < MAXVER; i++) {
        version[i] = -1;
    }
    pphase = -3;         /* -3: not defined, 1: standard, 2: Kyoto */
    mprefix  = NULL;     /* prefix of the model */
    for(i = 0; i < N_SIZE; i++) {
        nmtable[i] = 0;  /* table of names of mass and width */
    }
    majop    = False;
    maxtl    = MINTL;
    mintl    = MAXTL;

    /* dummy particle */
    if(pvpenter("undef", "undef", "?", "?") != PNULL) {
        fprintf(Stderr, "\n*err* pvread: illegal particle table.\n");
        exit(1);
    }
    ptable[0]->mass     = "0";
    ptable[0]->mdimn    = 0;
    ptable[0]->width    = "0";
    ptable[0]->wdimn    = 0;
    ptable[0]->ptype    = P_UNDEF;
    ptable[0]->charge3  = 0;
    ptable[0]->color    = 1;
    ptable[0]->pcode    = 0;
    ptable[0]->kfcode   = 0;
    ptable[0]->furry    = False;
    ptable[0]->massless = False;
    ptable[0]->extself  = False;
    ptable[0]->rname    = NULL;
    ptable[0]->rnami    = NULL;
    ptable[0]->sname    = NULL;
    ptable[0]->snami    = NULL;
    ptable[0]->chane    = NULL;
    ptable[0]->chani    = NULL;
    ptable[0]->gauge    = NULL;
    ptable[0]->pselect  = NULL;


    /* dummy vertices */
    for(i = 0; i < MAXTL; i++) {
        vp[i] = "undef";
    }
    for(n = MINTL; n <= MAXTL; n++) {
        nv = pvnenter(n, vp);
        vtable[nv]->vtype     = VTTREE;
        vtable[nv]->morder[0] = n - 2;
        vtable[nv]->loop      = 0;
        for(i = 1; i < nmordtbl; i++) {
            vtable[nv]->morder[i] = 0;
        }
        for(i = 1; i < NFNAME; i++) {
            vtable[nv]->fname[i]  = NULL;
            vtable[nv]->fndimn[i] = 0;
        }
        vtable[nv]->fname[0]  = "?";

        vtable[nv]->rname     = NULL;
        vtable[nv]->sname     = NULL;
        vtable[nv]->chanel    = NULL;
        vtable[nv]->vselect   = NULL;
        vtable[nv]->vid       = NULL;
        vtable[nv]->ctype     = V_NOTYPE;
        vtable[nv]->composit  = V_NORMAL;
        vtable[nv]->auxp      = 0;
        vtable[nv]->auxv      = NULL;
    }
    nvstart = nv + 1; /* read definition is >= nvstart */

    /* read "order= ... ", "particle= ...", "vertex= ..." or
            "medn;".
    */
    err = False;
    while (True) {
        if((pvind = rdstat()) < 0) {
            fprintf(Stderr, "\n*err* pvread : read error\n");
            if(pvind == RDEOF) {
                break;
            } else {
                continue;
            }
        }

        for(i = 0; i < MAXCMD - 1; i++) {
            cmd[i] = tolower(rdid(pvind)[i]);
        }
#if DBG
       if(dbg >= pvdbg) {
            erdbg("pvread", "statement :", pvdbg);
            rdprint(pvind);
        }
#endif

        if(rdtype(pvind) == RDEOF) {
            fprintf(Stderr, "\n*err* pvread : EOF before Mend;\n");
            err = True;
            break;

        }else if(rdtype(pvind) != RDID) {
            fprintf(Stderr, "\n*err* pvread : Id is expected\n");
            rddump(pvind);
            err = True;

        /* end of the file */
        } else if(strcmp(cmd, "mend") == 0) {
#if DBG
            erdbg("pvread", "id = mend", pvdbg);
#endif
            break;

        /* definition of coupling constants */
        } else if(strcmp(cmd, "title") == 0) {
#if DBG
            erdbg("pvread", "id = title", pvdbg);
#endif
            arg = rdarg(pvind);
            if((rdtype(arg) != RDSTRING)) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "Title=<string>\n");
                err = True;
            } else {
                if((ttl = (ttitle*) malloc(sizeof(ttitle))) == NULL) {
                    fprintf(Stderr, "\n*err* pvread : memory\n");
                    exit(FAIL);
                }
                ttl->tstr = pvstrdup(False, rdid(arg));
                ttl->next = NULL;
#ifdef SHOWTITLE
                printf("%s\n", ttl->tstr);
#endif
            }
            ptitle = &title;
            while(*ptitle != NULL) {
                ptitle = &((*ptitle)->next);
            }
            *ptitle = ttl;

        /* definition of coupling constants */
        } else if(strcmp(cmd, "order") == 0) {
#if DBG
            erdbg("pvread", "id = order", pvdbg);
#endif

            /* do not redefine */
            if(nmordtbl != 0) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "Order cannot be redefined.\n");
                rddump(pvind);
                err = True;

            /* definition */
            } else {
                for(arg = rdarg(pvind); rdtype(arg) == RDID;
                                         arg = rdarg(arg)) {
                    if(nmordtbl < MAXMORD-1) {
                        mordtbl[nmordtbl] = pvstrdup(True, rdid(arg));
                        mordopt[nmordtbl] = 0;
                        nmordtbl++;
                    } else {
                        fprintf(Stderr, "\n*err* pvread : ");
                        fprintf(Stderr, "order table ovberflow.\n");
                        err = True;
                    }
                    /* options : "qcd" */
                    for(opt = rdopt(arg), nopt = 0; opt >= 0;
                        opt = rdbrothr(opt), nopt++) {
                        if(rdtype(opt) != RDID) {
                            fprintf(Stderr, "\n*err* pvread : ");
                            fprintf(Stderr, "Illegal option to Order\n");
                            err = True;
                            continue;
                        }
                        if(strcmp(rdid(opt), "qcd") == 0) {
                           mordopt[nmordtbl] = MOPT_QCD;
                        }
                    }
                } /* for arg */
            }


        /* definition of coupling constants */
        } else if(strcmp(cmd, "ordset") == 0) {
#if DBG
            erdbg("pvread", "id = ordset", pvdbg);
#endif

            /* do not redefine */
            if(nmordtbl == 0) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "OrdSet cannot be defined before Order.\n");
                rddump(pvind);
                err = True;

            /* definition */
            } else {
                /* first call : set table for original coupling constants */
                if(nsordtbl == 0) {
                    for(i = 0; i < nmordtbl; i++) {
                        for(j = 0; j < MAXMORD; j++) {
                            sordset[i][j] = 0;
                        }
                        sordset[i][i] = -1;   /* leaf */
                        sordelm[i] = 1;
                        sordtbl[i] = mordtbl[i];
                    }
                    nsordtbl = nmordtbl;
                }

                /* new ordset */
                for(arg = rdarg(pvind); arg > 0; arg = rdarg(arg)) {

                    /* name of set of orders */
                    if(rdtype(arg) != RDID) {
                        fprintf(Stderr, "\n*err* pvread : ");
                        fprintf(Stderr, "Illegal OrdSet\n");
                        err = True;
                        continue;
                    }
                    str = rdid(arg);
                    for(i = 0; i < nsordtbl; i++) {
                        if(strcmp(str, sordtbl[i]) == 0) {
                            fprintf(Stderr, "\n*err* pvread : ");
                            fprintf(Stderr, "OrdSet \"%s\" is duplicated\n",
                                    str);
                            err = True;
                            continue;
                        }
                    }
                    if(nsordtbl >= MAXSORD - 1) {
                        fprintf(Stderr, "\n*err* pvread : ");
                        fprintf(Stderr, "OrdSet table ovberflow.\n");
                        err = True;
                        continue;
                    }

                    /* elements in the set */
                    n = 0;
                    for(j = 0; j < MAXMORD; j++) {
                        sordset[nsordtbl][j] = 0;
                    }
                    for(opt = rdopt(arg), nopt = 0; opt >= 0;
                        opt = rdbrothr(opt), nopt++) {
                        if(rdtype(opt) != RDID) {
                            fprintf(Stderr, "\n*err* pvread : ");
                            fprintf(Stderr, "Illegal element of OrdSet\n");
                            err = True;
                            continue;
                        }
                        for(i = 0; i < nsordtbl; i++) {
                            if(strcmp(rdid(opt), sordtbl[i]) == 0) {
                                for(j = 0; j < nmordtbl; j++) {
                                    if(sordset[i][j] != 0) {
                                        sordset[nsordtbl][j] = 1;
                                    }
                                }
                            }
                        }
                    }
                    if(nopt == 0) {
                        fprintf(Stderr, "\n*err* pvread : ");
                        fprintf(Stderr, "Empty set in of OrdSet\n");
                        fprintf(Stderr, "name=\"%s\"\n", str);
                        err = True;
                        continue;
                    }

                    sordelm[nsordtbl] = n;
                    sordtbl[nsordtbl] = pvstrdup(True, str);
                    nsordtbl++;

                } /* for arg */
            }

        /* version number */
        } else if(strcmp(cmd, "version") == 0) {
#if DBG
            erdbg("pvread", "id = version", pvdbg);
#endif

            /* do not redefine */
            if(version[0] >= 0) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "Order cannot be redefined.\n");
                rddump(pvind);
                err = True;

            /* definition */
            } else {
                for(arg = rdarg(pvind), i = 0; rdtype(arg) == RDINT;
                        arg = rdarg(arg), i++) {
                    if(i < MAXVER) {
                        version[i] = (int)rdint(arg);
                    } else {
                        fprintf(Stderr, "\n*err* pvread : ");
                        fprintf(Stderr, "too deep sub-version level\n");
                        err = True;
                    }
                } /* for arg */
            }


        /* definition of prefix to the model */
        } else if(strcmp(cmd, "prefix") == 0) {
#if DBG
            erdbg("pvread", "id = prefix", pvdbg);
#endif

            /* do not redefine */
            if(mprefix != NULL) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "MPrefix cannot be redefined.\n");
                rddump(pvind);
                err = True;

            /* definition */
            } else {
                arg = rdarg(pvind);
                if((rdtype(arg) != RDSTRING)) {
                    fprintf(Stderr, "\n*err* pvread : ");
                    fprintf(Stderr, "Prefix=<string>\n");
                    err = True;
                } else {
                    mprefix = pvstrdup(False, rdid(arg));
                }
            }
        /* phase of propagators */
        } else if(strcmp(cmd, "pphase") == 0) {

            /* do not redefine */
            if(pphase >= 0) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "PPhase cannot be redefined.\n");
                rddump(pvind);
                err = True;

            /* definition */
            } else {
                arg = rdarg(pvind);
                if((rdtype(arg) != RDINT)) {
                    fprintf(Stderr, "\n*err* pvparticle: illgal pphase.\n");
                    rddump(pvind);
                    err = True;
                } else {
                    pphase = (int)rdint(arg);
                }
            }

        /* definition of particles */
        } else if(strcmp(cmd, "particle") == 0) {
#if DBG
            erdbg("pvread", "id = particel", pvdbg);
#endif
            err = !pvparticle() || err;

        /* definition of vertices */
        } else if(strcmp(cmd, "vertex") == 0) {
            /* vertex should be defined after particles */
            if(nptable < 2) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "vertex without particle\n");
                rddump(pvind);
                err = True;
            }

            /* vertex should be defined after order */
            if(nmordtbl < 1) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "vertex without ");
                fprintf(Stderr, "definition of coupling consitants\n");
                rddump(pvind);
                err = True;
            }
            err = !pvertex() || err;

            if((nvtable-nvstart) % 500 == 0) {
                printf("+++ %d vertices\n", nvtable-nvstart);
            }


        /* definition of composit vertices */
        } else if(strcmp(cmd, "cvertex") == 0) {
            /* composit vertex should be defined after particles */
            if(nptable < 2) {
                fprintf(Stderr, "\n*err* pvread : ");
                fprintf(Stderr, "composit vertex without particle\n");
                rddump(pvind);
                err = True;
            }

            err = !pcvertex() || err;

            if((nvtable-nvstart) % 500 == 0) {
                printf("+++ %d vertices\n", nvtable-nvstart);
            }

        } else {
            fprintf(Stderr, "\n*err* pvread : syntax\n");
            rddump(pvind);
            err = True;
        }
    }
    if(nsordtbl == 0) {
        for(i = 0; i < nmordtbl; i++) {
            for(j = 0; j < MAXMORD; j++) {
                sordset[i][j] = 0;
            }
            sordset[i][i] = -1;   /* leaf */
            sordelm[i] = 1;
            sordtbl[i] = mordtbl[i];
        }
        nsordtbl = nmordtbl;
    }

    rdend();
    fclose(prpf);
    prpf = NULL;

    compgl4 = pvcompgl4();

    
    if(GRCVER0 != version[0] || GRCVER1 != version[1] ||
       GRCVER2 != version[2]) {
        fprintf(Stderr, "\n*err* pvread: inconsistent version\n");
        fprintf(Stderr, "    model %d.%d.%d\n",
                version[0], version[1], version[2]);
        fprintf(Stderr, "    grc   %d.%d.%d\n",
                GRCVER0, GRCVER1, GRCVER2);
        err = True;
    }

    err |= !pvfnchk();

#if DBG
    erdbg("pvread", "end", pvdbg);
#endif

    if(err) {
        rdprint(pvind);
        fprintf(Stderr, "\n*err* pvread: stop by error.\n");
        exit(1);
    }

    pvendtbl();

    nvmodel  = nvtable;
    npptable = nptable;
    nppbits  = (sizeof(Ulong)*8)/sizeof(char);  /* assume char = 8bits */
    nppword  = (2*npptable + nppbits - 2)/nppbits;

    printf("+++ Defined : %d particles, %d vertices (%d non-standard)\n", 
           npptable-1, nvmodel-nvstart, nonstand);

#if DBG
    if(dbg >= pvdbg) {
        pvdump();
    }
#endif

    free(pvrdsln);
    pvrdsln = NULL;
}

/*==================================================================*/
Private bool    pvparticle(NoArg)
/*--------------------------------------------------------------------
    read particle upto "pend;"
--------------------------------------------------------------------*/
{
    int arg, np, ngn;
    char pnam[MAXLID], apnam[MAXLID], psym[MAXLID], apsym[MAXLID];
    char mass[7], width[7];
    char *gname[3], *gauge, *pselect;
    char *rname, *rnami, *sname, *snami, *chane, *chani;
    int  charge[2], color, value, pcode, kfcode, ptype, massless;
    int  extself;
    int  i;
    bool furry, err;
    int  mdimn, mdim[MAXFDIM], mind[MAXFDIM];
    int  wdimn, wdim[MAXFDIM], wind[MAXFDIM];

#if DBG
    erdbg("pvparticle", "begin", pvdbg);
#endif

    /* get particle name and print-name */
    arg = rdarg(pvind);
    err = !pvpnam(arg, pnam, psym);

    /* variables to be set */
    apnam[0]  = '\0';      /* mark for undef */
    mass[0]   = '\0';      /* mark for undef */
    rname     = NULL;      /* mark for undef */
    rnami     = NULL;      /* mark for undef */
    sname     = NULL;      /* mark for undef */
    snami     = NULL;      /* mark for undef */
    chane     = NULL;      /* mark for undef */
    chani     = NULL;      /* mark for undef */
    gauge     = NULL;      /* mark for undef */
    pselect   = NULL;      /* mark for undef */
    mdimn     = 0;         /* mark for undef */
    width[0]  = '\0';      /* mark for undef */
    wdimn     = 0;         /* mark for undef */
    ptype     = P_UNDEF;   /* mark for undef */
    charge[1] = 0;         /* mark for undef */
    color     = 1;         /* default */
    pcode     = 0;         /* default */
    kfcode    = 0;         /* default */
    furry     = False;     /* default */
    massless  = False;     /* default */
    extself   = False;     /* default */
    gname[0]  = NULL;
    gname[1]  = NULL;
    gname[2]  = NULL;

    /* read "antiparticle=...", "charge=...", "color=..." or "pend;"
    */
    while (True) {
        pvind = rdstat();

        for(i = 0; i < MAXCMD - 1; i++) {
            cmd[i] = tolower(rdid(pvind)[i]);
        }

        if(rdtype(pvind) == RDEOF) {
            fprintf(Stderr, "\n*err* pvparticle : EOF\n");
            err = True;
            break;

        } else if(rdtype(pvind) != RDID) {
            fprintf(Stderr, "\n*err* pvparticle : illegal data\n");
            rddump(pvind);
            err = True;

        /* end of particle definition */
        } else if(strcmp(cmd, "pend") == 0) {
            break;

        /* antiparticle = name [ print-name ] */
        } else if(strcmp(cmd, "antiparticle") == 0) {
            arg = rdarg(pvind);
            err = !pvpnam(arg, apnam, apsym) || err;

        /* gname = {common-name [, particle-name, anti-particle-name]}; */
        } else if(strcmp(cmd, "gname") == 0) {
            for(arg = rdarg(pvind), ngn = 0; arg >= 0;
                     arg = rdarg(arg), ngn++) {
                if(ngn < 3) {
                    gname[ngn] = pvstrdup(False, rdid(arg));
                }
            }
            if(ngn > 3) {
                fprintf(Stderr, "\n*err* pvparticle : illgal gname\n");
                rddump(pvind);
                err = True;
            }

        /* ptype = name; */
        } else if(strcmp(cmd, "ptype") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvparticle : illgal ptype\n");
                rddump(pvind);
                err = True;
            } else {
                for(ptype = P_AUX; ptype > P_UNDEF; ptype--) {
                    for(i = 0; i < MAXCMD - 1; i++) {
                        argstr[i] = tolower(rdid(arg)[i]);
                    }
                    if(strcmp(argstr, pvtnam[ptype]) == 0) {
                        break;
                    }
                }
                if(ptype <= P_UNDEF) {
                    fprintf(Stderr, "\n*err* pvparticle :");
                    fprintf(Stderr, " illgal ptype (not defined)\n");
                    rddump(pvind);
                    err = True;
                }
            }

        /* charge = rational-number; */
        } else if(strcmp(cmd, "charge") == 0) {
            arg = rdarg(pvind);
            if((rdtype(arg) != RDRNUM && rdtype(arg) != RDINT)) {
                fprintf(Stderr, "\n*err* pvparticle: illgal charge.\n");
                rddump(pvind);
                err = True;
            } else {
                charge[0] = (int)rdrnum(arg, 0);
                charge[1] = (int)rdrnum(arg, 1);
            }

        /* pcode = integer; */
        } else if(strcmp(cmd, "pcode") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDINT) {
                fprintf(Stderr, "\n*err* pvparticle :");
                fprintf(Stderr, " illgal pcode (not int)\n");
                rddump(pvind);
                err = True;
            } else {
                pcode = (int)rdint(arg);
            }

        /* kfcode = integer; */
        } else if(strcmp(cmd, "kfcode") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDINT) {
                fprintf(Stderr, "\n*err* pvparticle :");
                fprintf(Stderr, " illgal kfcode (not int)\n");
                rddump(pvind);
                err = True;
            } else {
                kfcode = (int)rdint(arg);
            }

        /* color = integer; */
        } else if(strcmp(cmd, "color") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDINT) {
                fprintf(Stderr, "\n*err* pvparticle :");
                fprintf(Stderr, " illgal color (not int)\n");
                rddump(pvind);
                err = True;
            } else {
                color = (int)rdint(arg);
            }

        /* furry = bool; */
        } else if(strcmp(cmd, "furry") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* pvparticle :");
                fprintf(Stderr, " furry=yes/no\n");
                rddump(pvind);
                err = True;
            } else {
                furry = value;
            }

        /* massive; */
        } else if(strcmp(cmd, "massive") == 0) {
            massless = False;

        /* massless; */
        } else if(strcmp(cmd, "massless") == 0) {
            massless = True;

        /* off shell renormalization; */
        } else if(strcmp(cmd, "extself") == 0) {
            if(!rdbool(arg, &value)) {
                fprintf(Stderr, "\n*err* pvparticle :");
                fprintf(Stderr, " extself=yes/no\n");
                rddump(pvind);
                err = True;
            } else {
                extself = value;
            }

        /* off shell renormalization : diagonal factor */

        /* off shell renormalization : off-diagonal factor */

        /* mass = fortran-name; */
        } else if(strcmp(cmd, "mass") == 0) {
            arg = rdarg(pvind);

            /* name */
            if(rdtype(arg) == RDID) {
                strcpy(mass, rdid(arg));
                if(!pvfort(mass)) {
                    fprintf(Stderr, "\n*err* pvparticle: mass should ");
                    fprintf(Stderr, " be fortran id.\n");
                    rddump(pvind);
                    err = True;
                }
                mdimn = 0;
                for(arg = rdarg(arg); arg >= 0; arg = rdarg(arg)) {
                    if(mdimn >= MAXFDIM) {
                        fprintf(Stderr, "\n*err* pvparticle: too many");
                        fprintf(Stderr, "indices for Mass\n");
                        rddump(pvind);
                        err = True;
                    } else if(rdtype(arg) == RDINT) {
                        mind[mdimn] = 0;
                        mdim[mdimn] = (int)rdrnum(arg, 0);
                    } else if(rdtype(arg) == RDRNUM) {
                        mind[mdimn] = (int)rdrnum(arg, 0);
                        mdim[mdimn] = (int)rdrnum(arg, 1);
                    } else {
                        fprintf(Stderr, "\n*err* pvparticle: array for mass");
                        fprintf(Stderr, "should be integer or rational\n");
                        rddump(pvind);
                        err = True;
                    }
                    mdimn++;
                }

            /* 0 mass */
            } else if(rdtype(arg) == RDINT &&
                      ((int)rdint(arg)) == 0) {
                mass[0] = '0';
                mass[1] = '\0';

            } else {
                fprintf(Stderr, "\n*err* pvparticle: mass should ");
                fprintf(Stderr, " be fortran id.\n");
                rddump(pvind);
                err = True;
            }

        /* width = fortran-name; */
        } else if(strcmp(cmd, "width") == 0) {
            arg = rdarg(pvind);

            /* fortran name */
            if(rdtype(arg) == RDID) {
                strcpy(width, rdid(arg));
                if(!pvfort(width)) {
                    fprintf(Stderr, "\n*err* pvparticle: width should ");
                    fprintf(Stderr, " be fortran id.\n");
                    rddump(pvind);
                    err = True;
                }
                wdimn = 0;
                for(arg = rdarg(arg); arg >= 0; arg = rdarg(arg)) {
                    if(wdimn >= MAXFDIM) {
                        fprintf(Stderr, "\n*err* pvparticle: too many");
                        fprintf(Stderr, "indices for Width\n");
                        rddump(pvind);
                        err = True;
                    } else if(rdtype(arg) == RDINT) {
                        wind[wdimn] = 0;
                        wdim[wdimn] = (int)rdrnum(arg, 0);
                    } else if(rdtype(arg) == RDRNUM) {
                        wind[wdimn] = (int)rdrnum(arg, 0);
                        wdim[wdimn] = (int)rdrnum(arg, 1);
                    } else {
                        fprintf(Stderr, "\n*err* pvparticle: array for Width");
                        fprintf(Stderr, "should be integer or rational\n");
                        rddump(pvind);
                        err = True;
                    }
                    wdimn++;
                }

            /* 0 width */
            } else if(rdtype(arg) == RDINT &&
                      ((int)rdint(arg)) == 0) {
                width[0] = '0';
                width[1] = '\0';
            } else {
                fprintf(Stderr, "\n*err* pvparticle: width should ");
                fprintf(Stderr, " be fortran id.\n");
                rddump(pvind);
                err = True;
            }

        } else if(strcmp(cmd, "rname") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvparticle : illegal rname-1\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                rname  = pvstrdup(True, rdid(arg));
            }
            arg = rdarg(arg);
            if(arg >= 0) {
                if(rdtype(arg) != RDID) {
                    fprintf(Stderr, "\n*err* pvparticle : illegal rname-2\n");
                    rddump(pvind);
                    err = True;
                } else if(!err) {
                    rnami  = pvstrdup(True, rdid(arg));
                }
            }

        } else if(strcmp(cmd, "sname") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvparticle : illegal sname-1\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                sname  = pvstrdup(True, rdid(arg));
            }
            arg = rdarg(arg);
            if(arg >= 0) {
                if(rdtype(arg) != RDID) {
                    fprintf(Stderr, "\n*err* pvparticle : illegal sname-2\n");
                    rddump(pvind);
                    err = True;
                } else if(!err) {
                    snami  = pvstrdup(True, rdid(arg));
                }
            }

        } else if(strcmp(cmd, "chanel") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvparticle : illegal chanel\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                chane  = pvstrdup(True, rdid(arg));
            }
            arg = rdarg(arg);
            if(arg >= 0) {
                if(rdtype(arg) != RDID) {
                    fprintf(Stderr, "\n*err* pvparticle : illegal chanel-2\n");
                    rddump(pvind);
                    err = True;
                } else if(!err) {
                    chani  = pvstrdup(True, rdid(arg));
                }
            }

        } else if(strcmp(cmd, "gauge") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDSTRING) {
                fprintf(Stderr, "\n*err* pvparticle : illegal gauge\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                gauge = pvstrdup(True, rdid(arg));
            }

        } else if(strcmp(cmd, "pselect") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDSTRING) {
                fprintf(Stderr, "\n*err* pvparticle : illegal pselect\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                pselect = pvstrdup(True, rdid(arg));
            }

        /* unknown statement */
        } else {
            fprintf(Stderr, "\n*err* pvparticle : illegal data\n");
            rddump(pvind);
            err = True;
        }
    }

    if(apnam[0] == '\0') {
        fprintf(Stderr, "\n*err* pvparticle : undefined anti-particle");
        fprintf(Stderr, " of %s\n", pnam);
        err = True;
    }
    if(mass[0] == '\0') {
        fprintf(Stderr, "\n*err* pvparticle : undefined mass");
        fprintf(Stderr, " of %s\n", pnam);
        err = True;
    }
    if(width[0] == '\0') {
        fprintf(Stderr, "\n*err* pvparticle : undefined width");
        fprintf(Stderr, " of %s\n", pnam);
        err = True;
    }
    if(charge[1] == 0) {
        fprintf(Stderr, "\n*err* pvparticle : undefined charge");
        fprintf(Stderr, " of %s\n", pnam);
        err = True;
    }
#if DBG
    erdbg("pvparticle", "end", pvdbg);
#endif

    if(err) {
        rdprint(pvind);
        return False;
    }

    /* save particle and antiparticle */
    np = pvpenter(pnam, apnam, psym, apsym);

    if(ptype == P_MAJORANA) {
        if(ptable[np]->antip != 1) {
            fprintf(Stderr, "\n*err* pvparticle : illegal majorana :");
            fprintf(Stderr, " %s\n", pnam);
            rdprint(pvind);

            ptable[np]->pname[1] = ptable[np]->pname[0];
            ptable[np]->psym[1]  = ptable[np]->psym[0];
            ptable[np]->antip = 1;

            err = True;
        }
        majop = True;
    }
    if(ptype == P_AUX) {
        defaux = True;
    }

    ptable[np]->mass     = pvmname(mass, N_MASS);
    ptable[np]->mdimn    = mdimn;
    for(i = 0; i < mdimn; i++) {
        ptable[np]->mdim[i] = mdim[i];
        ptable[np]->mind[i] = mind[i];
    }
    ptable[np]->width    = pvmname(width, N_WIDTH);
    ptable[np]->wdimn    = wdimn;
    for(i = 0; i < wdimn; i++) {
        ptable[np]->wdim[i] = wdim[i];
        ptable[np]->wind[i] = wind[i];
    }
    ptable[np]->charge3  = ((int) 3*charge[0])/charge[1];
    ptable[np]->pcode    = pcode;
    ptable[np]->kfcode   = kfcode;
    ptable[np]->ptype    = ptype;
    ptable[np]->color    = color;
    ptable[np]->furry    = furry;
    ptable[np]->massless = massless;
    ptable[np]->extself  = extself;
    ptable[np]->rname    = rname;
    ptable[np]->rnami    = rnami;
    ptable[np]->sname    = sname;
    ptable[np]->snami    = snami;
    ptable[np]->chane    = chane;
    ptable[np]->chani    = chani;
    ptable[np]->gauge    = gauge;
    ptable[np]->pselect  = pselect;

    if(gname[0] == NULL) {
        gname[0] = psym;
        gname[1] = psym;
        gname[2] = apsym;
    } else {
        if(gname[1] == NULL) {
            gname[1] = gname[0];
        }
        if(gname[2] == NULL) {
            gname[2] = gname[0];
        }
    }
    ptable[np]->gname[0] = gname[0];
    ptable[np]->gname[1] = gname[1];
    ptable[np]->gname[2] = gname[2];

    return (!err);
}

/*==================================================================*/
Private bool    pvpnam(arg, pnam, psym)
int     arg;
char    *pnam;
char    *psym;
/*--------------------------------------------------------------------
    read particle name and abbraviation.

    particle-name [ "[" print-name "]" ]
--------------------------------------------------------------------*/
{
    int opt;

    if(rdtype(arg) != RDID) {
        fprintf(Stderr, "\n*err* pvparticle : illegal particle name\n");
        rddump(arg);
        rdprint(pvind);
        return False;

    } else if(strlen(rdid(arg)) >= MAXLID - 1) {
        fprintf(Stderr, "\n*err* pvpnam : too long name\n");
        rddump(arg);
        rdprint(pvind);
        return False;

    } else {
        strcpy(pnam, rdid(arg));
        if(rdtype(opt = rdopt(arg)) == RDSTRING) {
            if(strlen(rdid(opt)) >= MAXLID - 1) {
                fprintf(Stderr, "\n*err* pvpnam : too long option\n");
                rddump(arg);
                rdprint(pvind);
                return False;
            } else {
                strcpy(psym, rdid(opt));
            }
        } else if(opt < 0) {
            psym[0] = '\0';
        } else {
            fprintf(Stderr, "\n*err* pvpnam: illegal print name ");
            fprintf(Stderr, "(should be string).\n");
            rddump(arg);
            rdprint(pvind);
            return False;
        }
    }
    return True;
}
/*==================================================================*/
Private int     pvpenter(pnam, anam, psym, asym)
string   pnam;
string   anam;
string   psym;
string   asym;
/*--------------------------------------------------------------------
    Enter particle to the table.
    In the ptable, items pname, aname, antip are stored.
--------------------------------------------------------------------*/
{
    int np = nptable;
    int i;

#if DBG
   if(dbg >= pvdbg) {
        printf("pvpenter: nptable = %d, (%s, %s, %s, %s)\n",
                nptable, pnam, anam, psym, asym);
        rdprint(pvind);
    }
#endif

    /* check table size */
    if(nptable >= LPTABLE) {
        fprintf(Stderr, 
          "\n*err* pvpenter: particle table overflow(LPTABLE=%d)\n",LPTABLE);
        exit(1);
    }

    /* create space for new particle */
    nptable++;
    if((ptable[np] = (tptable*) malloc(sizeof(tptable))) == NULL) {
        fprintf(Stderr, "\n*err* pvpenter: memory allocation error\n");
        exit(1);
    }

    /* register particle name */
    ptable[np]->pname[0] = pvfindnam(pnam, np, PVNEWN);
    if(psym[0] == '\0') {
        ptable[np]->psym[0] = ntable[ptable[np]->pname[0]].name;
    } else {
        ptable[np]->psym[0] = pvstrdup(False, psym);
    }

    for(i = 0; i < MAXCMD - 1; i++) {
        argstr[i] = tolower(anam[i]);
    }

    if(strcmp(pnam, anam) == 0 || strcmp("particle", argstr) == 0) {
        ptable[np]->pname[1] = ptable[np]->pname[0];
        ptable[np]->psym[1]  = ptable[np]->psym[0];
        ptable[np]->antip = 1;
    } else {
        ptable[np]->pname[1] = pvfindnam(anam, -np, PVNEWN);
        if(asym[1] == '\0') {
            ptable[np]->psym[1] = ntable[ptable[np]->pname[1]].name;
        } else {
            ptable[np]->psym[1] = pvstrdup(False, asym);
        }
        ptable[np]->antip = -1;
    }

    return np;
}

/*==================================================================*/
Public  string   pvename(iorf, pc)
int     iorf;
pindex  pc;
/*--------------------------------------------------------------------
    returns name of pindex pc.
--------------------------------------------------------------------*/
{
    pindex pca;

    if(pc == PNULL) {
        return "undef";
    }

    pca = pc;
    if(iorf == ELINITIAL) {
        pca *= ptable[abs(pc)]->antip;
    }

    return pvname(pca);
}

/*==================================================================*/
Public  string   pvname(pc)
pindex  pc;
/*--------------------------------------------------------------------
    returns name of pindex pc.
--------------------------------------------------------------------*/
{
    nindex n;
    if(pc == PNULL) {
        return "undef";
    }
    n = (pc < 0) ? ptable[-pc]->pname[1] : ptable[pc]->pname[0];
    return ntable[n].name;
}

/*==================================================================*/
Public  string   pvgname(pc, gpa)
pindex  pc;
int     gpa;
/*--------------------------------------------------------------------
    returns gname of pindex pc.

    gpa = 0 : generic name
          1 : particle (pc > 0),      anti-aprticle (pc < 0)
          2 : anti-particle (pc > 0), particle (pc < 0)
--------------------------------------------------------------------*/
{
    if(pc == PNULL) {
        return "undef";
    }
    if(pc < 0) {
        pc = - pc;
        if(gpa != 0) {
            gpa = 3 - gpa;
        }
    }
    return ptable[pc]->gname[gpa];
}

/*==================================================================*/
Public pindex   pvfind(pnam)
string   pnam;
/*--------------------------------------------------------------------
    find pname from name table (inverse of pvname)
--------------------------------------------------------------------*/
{
    if(strcmp(pnam, "undef") == 0) {
        return PNULL;
    } else {
        return ntable[pvfindnam(pnam, 0, PVOLDN)].pindx;
    }
}

/*==================================================================*/
Public  int     pvfindnam(pnam, pind, mode)
string   pnam;
int     pind;
int     mode;
/*--------------------------------------------------------------------
    find pname from name table and add it.

    pind : particle index to be added.

    if mode >= 0 : pname should be new one.
            <    : pname should be old one.
--------------------------------------------------------------------*/
{
    int p;

    /* check the same name in the table */
    for(p = 0; p < nntable; p++) {
        if(strcmp(ntable[p].name, pnam) == 0) {
            if(mode == PVNEWN) {
                fprintf(Stderr, "\n*err* pvfindnam: particle is ");
                fprintf(Stderr," already defined. name=%s.\n", pnam);
                exit(1);
            } else {
                return p;
            }
        }
    }

    /* not found */
    if(mode != PVNEWN) {
        fprintf(Stderr, "\n*err* pvfindnam: particle is not found\n");
        fprintf(Stderr,"name=%s.\n", pnam);
        exit(1);
    }

    /* check size to save */
    if(nntable >= LNTABLE) {
        fprintf(Stderr, "\n*err* pvfindnam: name table overflow\n");
        exit(1);
    }

    p = nntable++;
    if((ntable[p].name = pvstrdup(False, pnam)) == NULL) {
        fprintf(Stderr, "\n*err* pvfindnam: memory allocation error\n");
        exit(1);
    }
    ntable[p].pindx = pind;

    return p;
}

/*==================================================================*/
Public vindex   pvidfind(vid)
string   vid;
/*--------------------------------------------------------------------
    find pname from name table (inverse of pvname)
--------------------------------------------------------------------*/
{
    int nv;

    for(nv = nvstart; nv < nvtable; nv++) {
        if(vtable[nv]->vid == NULL) {
            continue;
        }
        if(strcmp(vtable[nv]->vid, vid) == 0) {
            return nv;
        }
    }
    return -1;
}

/*==================================================================*/
Private string  pvmname(nam, ntype)
string  nam;
int     ntype;
/*--------------------------------------------------------------------
    find pname from name table and add it.
--------------------------------------------------------------------*/
{
    int p;

    /* check the same name in the table */
    for(p = 0; p < nmtable[ntype]; p++) {
        if(strcmp(mtable[ntype][p], nam) == 0) {
            return mtable[ntype][p];
        }
    }

    /* check size to save */
    if(nmtable[ntype] >= LPTABLE) {
        fprintf(Stderr, "\n*err* pvmname: name table overflow\n");
        exit(1);
    }

    p = nmtable[ntype]++;
    if((mtable[ntype][p] = pvstrdup(True, nam)) == NULL) {
        fprintf(Stderr, "\n*err* pvmname: memory allocation error\n");
        exit(1);
    }

    return mtable[ntype][p];
}

/*==================================================================*/
Private string  pvvname(nam, vnew, vold)
string  nam;
int     vnew;
int    *vold ;
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    int p;

    /* check the same name in the table */
    for(p = 0; p < nctable; p++) {
        if(strcmp(ctable[p], nam) == 0) {
            *vold = vnamdef[p];
            return ctable[p];
        }
    }

    /* check size to save */
    if(nctable >= LVTABLE) {
        fprintf(Stderr, "\n*err* pvvname: name table overflow\n");
        exit(1);
    }

    p = nctable++;
    if((ctable[p] = pvstrdup(False, nam)) == NULL) {
        fprintf(Stderr, "\n*err* pvvname: memory allocation error\n");
        exit(1);
    }

    vnamdef[p] = *vold = vnew;

    return ctable[p];
}

/*==================================================================*/
Public  int     pvvnfind(vn)
int     vn;
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    int p;
    string nam = vtable[vn]->fname[0];

    /* check the same name in the table */
    for(p = 0; p < nctable; p++) {
        if(strcmp(ctable[p], nam) == 0) {
            return vnamdef[p];
        }
    }
    return -1;
}

/*==================================================================*/
Public  int     pvvnfinde(vn, fn)
int     vn;
int     fn;
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    int p;
    string nam = vtable[vn]->fname[fn];

    if(nam == NULL) {
        return -1;
    }

    /* check the same name in the table */
    for(p = 0; p < nctable; p++) {
        if(strcmp(ctable[p], nam) == 0) {
            return vnamdef[p];
        }
    }
    return -1;
}

/*==================================================================*/
Private bool    pvertex(NoArg)
/*--------------------------------------------------------------------
    read vertex.
--------------------------------------------------------------------*/
{
    string vp[MAXTL+1];
    string id;
    int    arg, nvp, nv = 0, nc, i, ord, cg, p, c, lp, nvdef, fn;
    int    p0, p1;
    bool   err = False;

#if DBG
    erdbg("pvertex", "begin", pvdbg);
#endif

    /* for list of particles */
    for(arg = rdarg(pvind), nvp = 0; arg >= 0;
             arg = rdarg(arg), nvp++)
    {

        /* too many particles in the list */
        if(nvp > MAXTL) {
            fprintf(Stderr, "\n*err* pvertex : too many legs\n");
            rddump(pvind);
            err = True;
            break;

        } else if(rdtype(arg) != RDID) {
            fprintf(Stderr, "\n*err* pvertex : illegal leg\n");
            rddump(pvind);
            err = True;
            break;

        /* save temporary */
        } else {
            vp[nvp] = rdid(arg);
        }
    }

    if(!err) {
        nv = pvnenter(nvp, vp);

        pvrdsln[nv] = rdsln;

        cg = 0;
        for(i = 0; i < nvp; i++) {
            p   = vtable[nv]->vlegs[i];
            c   = sign(p)*ptable[abs(p)]->charge3;
            cg += c;
        }

        if(cg != 0) {
            fprintf(Stderr, "\n*err* pvertex : charge not conserved\n");
            rddump(pvind);
            err = True;
        }

        for(i = 0; i < nmordtbl; i++) {
            vtable[nv]->morder[i] = 0;
        }
        for(i = 0; i < NFNAME; i++) {
            vtable[nv]->fname[i]  = NULL;
            vtable[nv]->fndimn[i] = 0;
        }
        vtable[nv]->rname     = NULL;
        vtable[nv]->sname     = NULL;
        vtable[nv]->chanel    = NULL;
        vtable[nv]->vselect   = NULL;
        vtable[nv]->vid       = NULL;
        vtable[nv]->ctype     = V_NOTYPE;
        vtable[nv]->composit  = V_NORMAL;
        vtable[nv]->auxp      = 0;
        vtable[nv]->auxv      = NULL;
    }

    while (True) {
        pvind = rdstat();

        for(i = 0; i < MAXCMD - 1; i++) {
            cmd[i] = tolower(rdid(pvind)[i]);
        }

        if(rdtype(pvind) == RDEOF) {
            fprintf(Stderr, "\n*err* pvertex : EOF\n");
            rddump(pvind);
            err = True;
            break;

        } else if(rdtype(pvind) != RDID) {
            fprintf(Stderr, "\n*err* pvertex : illegal data\n");
            rddump(pvind);
            err = True;

        } else if(strcmp(cmd, "vend") == 0) {
            break;

        } else if(strcmp(cmd, "cvend") == 0) {
            fprintf(Stderr, "\n*err* pvertex :");
            fprintf(Stderr, "vertex : cannot use 'CVend' for 'Vertex\n");
            rddump(pvind);
            err = True;
            break;

        } else if(strcmp(cmd, "fname") == 0 || strcmp(cmd, "cfname") == 0 ||
                  strcmp(cmd, "efname") == 0 || strcmp(cmd, "cefname") == 0) {
            if(strcmp(cmd, "fname") == 0) {
                fn = 0;
            } else if(strcmp(cmd, "cfname") == 0) {
                fn = 1;
            } else if(strcmp(cmd, "efname") == 0) {
                fn = 2;
            } else {
                fn = 3;
            }
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvertex : ");
                fprintf(Stderr, "illegal fname/cfname/efname\n");
                rddump(pvind);
                err = True;
            } else if(!pvfort(rdid(arg))) {
                fprintf(Stderr, "\n*err* pvertex: name should be ");
                fprintf(Stderr, "fortran name\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->fname[fn]  = pvvname(rdid(arg), nv, &nvdef);
                vtable[nv]->fndimn[fn] = 0;
if(strcmp(vtable[nv]->fname[fn], "ce20sw") == 0) { printf("pvertex:%d:%d:%s\n", nv, fn, vtable[nv]->fname[fn]); }
                for(arg = rdarg(arg); arg >= 0; arg = rdarg(arg)) {
                    if(vtable[nv]->fndimn[fn] >= MAXFDIM) {
                        fprintf(Stderr, "\n*err* pvertex: too many");
                        fprintf(Stderr, "indices for CFName\n");
                        rddump(pvind);
                        err = True;
                    } else if(rdtype(arg) == RDINT) {
                        vtable[nv]->fnind[fn][vtable[nv]->fndimn[fn]] = 0;
                        vtable[nv]->fndim[fn][vtable[nv]->fndimn[fn]]
                            = (int)rdrnum(arg, 0);
                    } else if(rdtype(arg) == RDRNUM) {
                        vtable[nv]->fnind[fn][vtable[nv]->fndimn[fn]]
                            = (int)rdrnum(arg, 0);
                        vtable[nv]->fndim[fn][vtable[nv]->fndimn[fn]]
                            = (int)rdrnum(arg, 1);
                    } else {
                        fprintf(Stderr, "\n*err* pvertex: ");
                        fprintf(Stderr, "array for fname/cfname");
                        fprintf(Stderr, "should be integer or rational\n");
                        rddump(pvind);
                        err = True;
                    }
                    vtable[nv]->fndimn[fn]++;
                }
            }

        } else if(strcmp(cmd, "rname") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvertex : illegal rname\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->rname = pvstrdup(True, rdid(arg));
            }

        } else if(strcmp(cmd, "sname") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvertex : illegal sname\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->sname = pvstrdup(True, rdid(arg));
            }

        } else if(strcmp(cmd, "chanel") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvertex : illegal chanel\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->chanel = pvstrdup(True, rdid(arg));
            }

        } else if(strcmp(cmd, "vselect") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvertex : illegal vselect\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->vselect = pvstrdup(True, rdid(arg));
            }

        } else if(strcmp(cmd, "vid") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pvertex : illegal vid\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->vid = pvstrdup(True, rdid(arg));
            }

        } else if(strcmp(cmd, "ftype") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDSTRING) {
                fprintf(Stderr, "\n*err* pvertex : illegal ftype\n");
                rddump(pvind);
                err = True;
            } else {
                id = rdid(arg);
                for(i = 0; i < strlen(id); i++) {
                    id[i] = toupper(id[i]);
                }
                if(strcmp(id, "NONE") == 0) {
                    vtable[nv]->ctype = V_NOTYPE;
                } else if(strcmp(id, "V") == 0) {
                    vtable[nv]->ctype = V_V;
                } else if(strcmp(id, "A") == 0) {
                    vtable[nv]->ctype = V_A;
                } else if(strcmp(id, "V-A") == 0) {
                    vtable[nv]->ctype = V_VminusA;
                } else if(strcmp(id, "V+A") == 0) {
                    vtable[nv]->ctype = V_VplusA;
                } else if(strcmp(id, "S") == 0) {
                    vtable[nv]->ctype = V_V;
                } else if(strcmp(id, "P") == 0) {
                    vtable[nv]->ctype = V_A;
                } else if(strcmp(id, "S-P") == 0) {
                    vtable[nv]->ctype = V_VminusA;
                } else if(strcmp(id, "S+P") == 0) {
                    vtable[nv]->ctype = V_VplusA;
                } else {
                    fprintf(Stderr, "\n*err* pvertex: illegal FType");
                    fprintf(Stderr, "%s\n", id);
                    rddump(pvind);
                    err = True;
                }
            }

        /* coupling constants */
        } else {
            arg = -1;
            for(i = 0; i < nmordtbl; i++) {
                if(strcmp(mordtbl[i], cmd) == 0) {
                    arg = rdarg(pvind);
                    if(rdtype(arg) != RDINT) {
                        fprintf(Stderr, "\n*err* pvertex : ");
                        fprintf(Stderr, "illegal fname\n");
                        rddump(pvind);
                        err = True;
                    } else if(err) {
                        ;
                    } else if(vtable[nv]->morder[i] != 0) {
                        fprintf(Stderr, "\n*err* pvertex : ");
                        fprintf(Stderr, "duplicate coupling\n");
                        rddump(pvind);
                        err = True;
                    } else {
                        vtable[nv]->morder[i] = (int)rdint(arg);
                    }
                }
            }
            if(arg < 0) {
                fprintf(Stderr, "\n*err* pvertex : illegal data\n");
                rddump(arg);
                err = True;
            }
        }
    }
#if DBG
    erdbg("pvertex", "test", pvdbg);
#endif

    for(i = 0; i < nvp; i++) {
        p = vtable[nv]->vlegs[i];
        if(ptable[abs(p)]->ptype == P_AUX) {
            vtable[nv]->composit = V_AUX;
            continue;
        }
    }

    if(vtable[nv]->fname[0] == NULL) {
        vtable[nv]->fname[0] = "";
    }

    while(pphase < 0) {
        pphase += 4;
    }

    ord = 0;
    for(i = 0; i < nmordtbl; i++) {
        ord += vtable[nv]->morder[i];
    }

    lp = ord - nvp + 2;
    if(lp % 2 != 0 || lp < 0) {  /* change for counter terms */
        fprintf(Stderr, "\n*err* pvertex : illegal order\n");
        fprintf(Stderr, "loop=%d.5\n", lp/2);
        err = True;
    } else {
        vtable[nv]->loop = lp / 2;
    }

#ifdef DEFSNAME
    if(vtable[nv]->rname == NULL) {
        vtable[nv]->rname = pvrname(nv);
    }
    if(vtable[nv]->sname == NULL) {
        vtable[nv]->sname = pvsname(nv);
    }
    if(vtable[nv]->chanel == NULL) {
        vtable[nv]->chanel = pvchanel(nv);
    }
#endif

    if(lp == 0) {
        maxtl = max(maxtl, nvp);
        mintl = min(mintl, nvp);
        vtable[nv]->vtype = VTTREE;
    } else {
        vtable[nv]->vtype = VTLOOP;
    }

#ifdef VDEFCHECK
    if(!err && nv != nvdef) {
        err = !pvcpcpl(nv, nvdef);
        if(err) {
            fprintf(Stderr,"\n*err* pvertex : Fortran name %s ",
                    vtable[nv]->fname[0]);
            fprintf(Stderr,"defined for vertex %d is \n", nv);
            fprintf(Stderr,"inconsistent to the definition for %d.\n",
                    nvdef);
        }
    }
#endif

    if(pvfermiv(nv)) {
        if(vtable[nv]->ctype == V_NOTYPE) {
            nc = 2;
        } else {
            nc = 1;
        }
    } else if(vtable[nv]->ctype != V_NOTYPE) {
        fprintf(Stderr, "\n*err* pvertex : FType should not be ");
        fprintf(Stderr, "defined for boson vertex\n");
        err = True;
    } else {
        nc = 1;
    }

    /* check when efname or cefname is defined */
    if(vtable[nv]->fname[2] != NULL || vtable[nv]->fname[3] != NULL) {
        if(vtable[nv]->nvlegs != 2) {
            fprintf(Stderr, "\n*err* pvertex : EFName or CEFName is defined ");
            fprintf(Stderr, "for %d point vertex\n", vtable[nv]->nvlegs);
            err = True;
        }
        p0 = abs(vtable[nv]->vlegs[0]);
        p1 = abs(vtable[nv]->vlegs[1]);
        if(!ptable[p0]->extself && !ptable[p1]->extself) {
            fprintf(Stderr, "\n*err* pvertex : EFName or CEFName is defined ");
            fprintf(Stderr, "for a vertex without 'Extself' particles\n");
            err = True;
        }
    }

    if(!err) {
        err = !pvncheck(nv);
    }
    if(err) {
        rdprint(pvind);
    }
#if DBG
    erdbg("pvertex", "end", pvdbg);
#endif
    return ! err;
}


/*==================================================================*/
Private bool    pcvertex(NoArg)
/*--------------------------------------------------------------------
 *  read composit vertex.
 */
{
    string vp[MAXTL+1];
    int    arg, nvp, nv = 0, i, ord, cg, p, c, lp, nvdef;
    int    npp, paux, pna, pnv, nlgs, v, j, fn;
    int    mord[MAXMORD];
    bool   err = False;
    bool   orddef = False;

#if DBG
    erdbg("pcvertex", "begin", pvdbg);
#endif

    /* for list of vid's */
    for(arg = rdarg(pvind), nvp = 0; arg >= 0;
             arg = rdarg(arg), nvp++) {

        /* too many vid's in the list */
        if(nvp > MAXTL) {
            fprintf(Stderr, "\n*err* pcvertex : too many legs\n");
            rddump(pvind);
            err = True;
            break;

        } else if(rdtype(arg) != RDID) {
            fprintf(Stderr, "\n*err* pcvertex : illegal leg\n");
            rddump(pvind);
            err = True;
            break;

        /* save temporary */
        } else {
            vp[nvp] = rdid(arg);
        }
    }

    if(!err) {
        nv = pcvnenter(nvp, vp);
        for(i = 0; i < NFNAME; i++) {
            vtable[nv]->fname[i]  = NULL;
            vtable[nv]->fndimn[i] = 0;
        }
        vtable[nv]->rname     = NULL;
        vtable[nv]->sname     = NULL;
        vtable[nv]->chanel    = NULL;
        vtable[nv]->vselect   = NULL;
        vtable[nv]->vid       = NULL;
        vtable[nv]->ctype     = V_NOTYPE;
        vtable[nv]->composit  = V_COMPOSIT;
    }

    if(!err) {
        pvrdsln[nv] = rdsln;

        paux = abs(vtable[nv]->auxp);
        cg   = 0;
        pna  = 0;
        nlgs = 0;
        for(i = 0; i < nmordtbl; i++) {
            mord[i] = 0;
        }
        for(i = 0; i < 2; i++) {
            v = vtable[nv]->auxv[i];
            if(v < 0) {
                fprintf(Stderr, "\n*err* pcvertex :");
                fprintf(Stderr, "illegal auxv\n");
                fprintf(Stderr, "nv = %d, i = %d\n", nv, i);
                rddump(pvind);
                err = True;
                continue;
            }
            
            for(j = 0; j < nmordtbl; j++) {
                mord[j] += vtable[v]->morder[j];
            }
            pnv = 0;
            for(j = 0; j < vtable[v]->nvlegs; j++) {
                p   = vtable[v]->vlegs[j];
                if(paux == p) {
                    pna++;
                    pnv++;
                    continue;
                } else if(paux == -p) {
                    pna--;
                    pnv++;
                    continue;
                } else {
                    nlgs++;
                    c   = sign(p)*ptable[abs(p)]->charge3;
                    cg += c;
                }
            }
            if(pnv != 1) {
                fprintf(Stderr, "\n*err* pcvertex :");
                fprintf(Stderr, "illegal number of aux-p. pnv=%d\n", pnv);
                fprintf(Stderr, "nv = %d, v = %d\n", nv, v);
                rddump(pvind);
                err = True;
            }
        }

        if(ptable[paux]->antip == -1) {
            if(pna != 0) {
                fprintf(Stderr, "\n*err* pcvertex: ");
                fprintf(Stderr, "illegal number of aux-p., %d!=0\n", pna);
                fprintf(Stderr, "nv = %d, v = %d\n", nv, v);
                rddump(pvind);
                err = True;
            }
        } else {
            if(pna != 2) {
                fprintf(Stderr, "\n*err* pcvertex: ");
                fprintf(Stderr, "illegal number of aux-p., %d!=2\n", pna);
                fprintf(Stderr, "nv = %d, v = %d\n", nv, v);
                rddump(pvind);
                err = True;
            }
        }

        if(cg != 0) {
            fprintf(Stderr, "\n*err* pcvertex : charge not conserved\n");
            fprintf(Stderr, "nv = %d, v = %d\n", nv, v);
            rddump(pvind);
            err = True;
        }

        for(i = 0; i < nmordtbl; i++) {
            vtable[nv]->morder[i] = 0;
        }

        if(err) {
            fprintf(Stderr, "nv = %d, aux = %s\n",
                    nv, pvname(vtable[nv]->auxp));
            for(i = 0; i < 2; i++) {
                v = vtable[nv]->auxv[i];
                if(v < 0) {
                    continue;
                }
                fprintf(Stderr, "%4d: v=%d, {", i, v);
                for(j = 0; j < vtable[v]->nvlegs; j++) {
                    fprintf(Stderr, "%s ", pvname(vtable[v]->vlegs[j]));
                }
                fprintf(Stderr, "}\n");
            }
        }

        vtable[nv]->nvlegs = nlgs;
        if((vtable[nv]->vlegs = (pindex*)malloc(sizeof(pindex)*nlgs))
            == NULL) {
            fprintf(Stderr, "\n*err* pcvertex: memory allocation ");
            fprintf(Stderr, "error(1)\n");
            exit(1);
        }
        for(i = 0, npp = 0; i < 2; i++) {
            v = vtable[nv]->auxv[i];
            for(j = 0; j < vtable[v]->nvlegs; j++) {
                p = vtable[v]->vlegs[j];
                if(abs(p) != abs(vtable[nv]->auxp)) {
                    vtable[nv]->vlegs[npp++] = p;
                }
            }
        }
        if(vtable[nv]->nvlegs != npp) {
            fprintf(Stderr, "\n*err* pcvertex: illgal control\n");
            fprintf(Stderr, "vtable[%d]->nvlegs = %d, npp = %d\n",
                    nv, vtable[nv]->nvlegs, npp);
            exit(1);
        }
    }

    while (True) {
        pvind = rdstat();

        for(i = 0; i < MAXCMD - 1; i++) {
            cmd[i] = tolower(rdid(pvind)[i]);
        }

        if(rdtype(pvind) == RDEOF) {
            fprintf(Stderr, "\n*err* pcvertex : EOF\n");
            rddump(pvind);
            err = True;
            break;

        } else if(rdtype(pvind) != RDID) {
            fprintf(Stderr, "\n*err* pcvertex : illegal data\n");
            rddump(pvind);
            err = True;
            break;

        } else if(strcmp(cmd, "vend") == 0) {
            fprintf(Stderr, "\n*err* pcvertex :");
            fprintf(Stderr, "vertex : cannot use 'Vend' for 'CVertex\n");
            rddump(pvind);
            err = True;
            break;

        } else if(strcmp(cmd, "cvend") == 0) {
            break;

        } else if(strcmp(cmd, "fname") == 0 || strcmp(cmd, "cfname") == 0 ||
                  strcmp(cmd, "efname") == 0 || strcmp(cmd, "cefname") == 0) {
            if(strcmp(cmd, "fname") == 0) {
                fn = 0;
            } else if(strcmp(cmd, "cfname") == 0) {
                fn = 1;
            } else if(strcmp(cmd, "efname") == 0) {
                fn = 2;
            } else {
                fn = 3;
            }
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pcvertex : ");
                fprintf(Stderr, "illegal fname/cfname/efname\n");
                rddump(pvind);
                err = True;
            } else if(!pvfort(rdid(arg))) {
                fprintf(Stderr, "\n*err* pcvertex: name should be ");
                fprintf(Stderr, "fortran name\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->fname[fn]  = pvvname(rdid(arg), nv, &nvdef);
                vtable[nv]->fndimn[fn] = 0;
if(strcmp(vtable[nv]->fname[fn], "ce20sw") == 0) { printf("pcvertex:%d:%d:%s\n", nv, fn, vtable[nv]->fname[fn]); }
                for(arg = rdarg(arg); arg >= 0; arg = rdarg(arg)) {
                    if(vtable[nv]->fndimn[fn] >= MAXFDIM) {
                        fprintf(Stderr, "\n*err* pcvertex: too many");
                        fprintf(Stderr, "indices for CFName\n");
                        rddump(pvind);
                        err = True;
                    } else if(rdtype(arg) == RDINT) {
                        vtable[nv]->fnind[fn][vtable[nv]->fndimn[fn]] = 0;
                        vtable[nv]->fndim[fn][vtable[nv]->fndimn[fn]]
                            = (int)rdrnum(arg, 0);
                    } else if(rdtype(arg) == RDRNUM) {
                        vtable[nv]->fnind[fn][vtable[nv]->fndimn[fn]]
                            = (int)rdrnum(arg, 0);
                        vtable[nv]->fndim[fn][vtable[nv]->fndimn[fn]]
                            = (int)rdrnum(arg, 1);
                    } else {
                        fprintf(Stderr, "\n*err* pcvertex: ");
                        fprintf(Stderr, "array for fname/cfname");
                        fprintf(Stderr, "should be integer or rational\n");
                        rddump(pvind);
                        err = True;
                    }
                    vtable[nv]->fndimn[fn]++;
                }
            }

        } else if(strcmp(cmd, "rname") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pcvertex : illegal rname\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->rname = pvstrdup(True, rdid(arg));
            }

        } else if(strcmp(cmd, "sname") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pcvertex : illegal sname\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->sname = pvstrdup(True, rdid(arg));
            }

        } else if(strcmp(cmd, "chanel") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pcvertex : illegal chanel\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->chanel = pvstrdup(True, rdid(arg));
            }

        } else if(strcmp(cmd, "vselect") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pcvertex : illegal vselect\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->vselect = pvstrdup(True, rdid(arg));
            }

        } else if(strcmp(cmd, "vid") == 0) {
            arg = rdarg(pvind);
            if(rdtype(arg) != RDID) {
                fprintf(Stderr, "\n*err* pcvertex : illegal vid\n");
                rddump(pvind);
                err = True;
            } else if(!err) {
                vtable[nv]->vid = pvstrdup(True, rdid(arg));
            }

        } else if(strcmp(cmd, "ftype") == 0) {
            fprintf(Stderr, "\n*err* pcvertex : cannot define ftype ");
            fprintf(Stderr, "for a composit vertex.");
            err = True;

        /* coupling constants */
        } else {
            arg = -1;
            for(i = 0; i < nmordtbl; i++) {
                if(strcmp(mordtbl[i], cmd) == 0) {
                    arg = rdarg(pvind);
                    if(rdtype(arg) != RDINT) {
                        fprintf(Stderr, "\n*err* pcvertex : ");
                        fprintf(Stderr, "illegal fname\n");
                        rddump(pvind);
                        err = True;
                    } else if(err) {
                        ;
                    } else if(vtable[nv]->morder[i] != 0) {
                        fprintf(Stderr, "\n*err* pcvertex : ");
                        fprintf(Stderr, "duplicate coupling\n");
                        rddump(pvind);
                        err = True;
                    } else {
                        vtable[nv]->morder[i] = (int)rdint(arg);
                        orddef = True;
                    }
                }
            }
            if(arg < 0) {
                fprintf(Stderr, "\n*err* pcvertex : illegal data\n");
                rddump(arg);
                err = True;
            }
        }
    }
#if DBG
    erdbg("pcvertex", "test", pvdbg);
#endif

    while(pphase < 0) {
        pphase += 4;
    }

    if(orddef) {
        for(i = 0; i < nmordtbl; i++) {
            if(mord[i] != vtable[nv]->morder[i]) {
                fprintf(Stderr, "\n*err* pcvertex : illegal order\n");
                fprintf(Stderr, "mord[%d]=%d != %d\n", i, mord[i],
                                vtable[nv]->morder[i]);
                err = True;
            }
        }
    } else {
        for(i = 0; i < nmordtbl; i++) {
            vtable[nv]->morder[i] = mord[i];
        }
    }
    ord = 0;
    for(i = 0; i < nmordtbl; i++) {
        ord += vtable[nv]->morder[i];
    }

    lp = ord - vtable[nv]->nvlegs + 2;
    if(lp % 2 != 0 || lp < 0) {  /* change for counter terms */
        fprintf(Stderr, "\n*err* pcvertex : illegal order\n");
        fprintf(Stderr, "loop=%d.5\n", lp/2);
        err = True;
    } else {
        vtable[nv]->loop = lp / 2;
    }

#ifdef DEFSNAME
    if(vtable[nv]->rname == NULL) {
        vtable[nv]->rname = pvrname(nv);
    }
    if(vtable[nv]->sname == NULL) {
        vtable[nv]->sname = pvsname(nv);
    }
    if(vtable[nv]->chanel == NULL) {
        vtable[nv]->chanel = pvchanel(nv);
    }
#endif

    if(lp == 0) {
        maxtl = max(maxtl, vtable[nv]->nvlegs);
        mintl = min(mintl, vtable[nv]->nvlegs);
        vtable[nv]->vtype = VTTREE;
    } else {
        vtable[nv]->vtype = VTLOOP;
    }

#ifdef VDEFCHECK
    if(!err && nv != nvdef) {
        err = !pvcpcpl(nv, nvdef);
        if(err) {
            fprintf(Stderr,"\n*err* pcvertex : Fortran name %s ",
                    vtable[nv]->fname[0]);
            fprintf(Stderr,"defined for composit vertex %d is \n", nv);
            fprintf(Stderr,"inconsistent to the definition for %d.\n",
                    nvdef);
        }
    }
#endif

    if(err) {
        rdprint(pvind);
        pvdump();
    }
#if DBG
    erdbg("pcvertex", "end", pvdbg);
#endif
    return ! err;
}

/*==================================================================*/
Private bool    pvfnchk(NoArg)
/*------------------------------------------------------------------*
 * Check name of coupling constant for Fortran code.
 */
{
    int nv, iv, j, l, fn1, fn2;
    bool status = True;

    for(nv = nvstart; nv < nvtable; nv++) {
        if(vtable[nv]->composit == V_AUX) {
            continue;
        }
        if(vtable[nv]->fname[0] == NULL) {
            fprintf(Stderr,"\n*err* pvfnchk : Fortran name is not defined\n");
            exit(FAIL);
        }
        for(fn1 = 0; fn1 < NFNAME; fn1++) {
            if(vtable[nv]->fname[fn1] == NULL) {
                continue;
            }
            l = strlen(vtable[nv]->fname[fn1]);
            for(iv = nvstart; iv <= nv; iv++) {
                for(fn2 = 0; fn2 < NFNAME; fn2++) {
                    if(nv == iv && fn1 == fn2) {
                        continue;
                    }
                    if(vtable[iv]->fname[fn2] == NULL) {
                        continue;
                    }
                    if(strcmp(vtable[nv]->fname[fn1],
                              vtable[iv]->fname[fn2]) == 0) {
                        if(vtable[nv]->fndimn[fn1] != vtable[iv]->fndimn[fn2]) {
                            fprintf(Stderr, "\n*err* pvfnchk : ");
                            fprintf(Stderr, "different dimension\n");
                            fprintf(Stderr, "nodes %d and %d, fname = %s, ",
                                         iv, nv, vtable[nv]->fname[0]);
                            if(pvrdsln != NULL) {
                                fprintf(Stderr, "line no. %d and %d",
                                        pvrdsln[iv], pvrdsln[nv]);
                            }
                            fprintf(Stderr, "\n");
                            status = False;
                            break;
                        }
                        for(j = 0; j < vtable[nv]->fndimn[fn1]; j++) {
                            if(vtable[nv]->fndim[fn1][j] != 
                               vtable[iv]->fndim[fn2][j]) {
                                fprintf(Stderr, "\n*err* pvfnchk : ");
                                fprintf(Stderr, "different dimension\n");
                                fprintf(Stderr, "nv = %d, iv = %d, fname = %s",
                                                 nv,iv,vtable[nv]->fname[fn1]);
                                fprintf(Stderr, "\n");
                                status = False;
                                break;
                            } else if(vtable[nv]->fnind[fn1][j] != 
                                      vtable[iv]->fnind[fn2][j]) {
                                break;
                            }
                        }
        
                        /* the following condition include the case of
                         * vtable[nv]->fndimn == 0
                         */
                        if(j >= vtable[nv]->fndimn[fn1]) {
                            fprintf(Stderr, "\n*warn* pvfnchk : ");
                            fprintf(Stderr, "same coupling\n");
                            fprintf(Stderr, "nodes %d and %d, fname = %s, ",
                                             iv, nv, vtable[nv]->fname[fn1]);
                            if(pvrdsln != NULL) {
                                fprintf(Stderr, "line no. %d, %d",
                                        pvrdsln[iv], pvrdsln[nv]);
                            }
                            fprintf(Stderr, "\n");
                        }
                        break;
                    }
                }
            }
        }
    }
    return status;
}

#ifdef VDEFCHECK
/*==================================================================*/
Private bool    pvcpcpl(nn, no)
int     nn, no;
/*------------------------------------------------------------------*
 * Check consistency of coupling constant.
 */
{
    bool    err = False;
    tvtable *vn, *vo;
    int     i, t1, t2;

    if(nn == no) {
        return True;
    }

    vn = vtable[nn];
    vo = vtable[no];

    err |= (vn->vtype  != vo->vtype);
    err |= (vn->nvlegs != vo->nvlegs);
    err |= (vn->loop   != vo->loop);
    for(i = 0; i < vn->nvlegs; i++) {
        t1 = ptable[abs(vn->vlegs[i])]->ptype;
        t2 = ptable[abs(vo->vlegs[i])]->ptype;
        if(t1 == P_MAJORANA) {
            t1 = P_DIRAC;
        }
        if(t2 == P_MAJORANA) {
            t2 = P_DIRAC;
        }
        err |= (t1 != t2);
    }

    if(vn->ctype == V_NOTYPE) {
        vn->ctype = vo->ctype;
    } else {
        err |= (vn->ctype != vo->ctype);
    }

    return !err;
}
#endif

/*==================================================================*/
Private int     pvlord(p1, p2)
pindex  p1, p2;
/*------------------------------------------------------------------*
 * compare type of two particles.
 * scalar < {majorana, anti-fermion, fermion} < anti-ghost < ghost
 * < vector
 *
 * OLD scalar < majorana < anti-fermion < fermion < anti-ghost
 * OLD        < ghost < vector
 */
{
    int    s1, s2, b1, b2;
    int    a1, a2, d;

    if(p1 >= 0) {
        a1  = 1;
    } else {
        p1  = -p1;
        a1  = -1;
    }
    s1  = ptable[p1]->ptype;
#ifdef OLDORD
    if(s1 == P_DIRAC || s1 == P_GHOST) {
        b1 = False;
    } else {
        b1 = True;
    }
#else
    if(s1 == P_MAJORANA) {
        s1 = P_DIRAC;
    }
    if(s1 == P_GHOST) {
        b1 = False;
    } else {
        b1 = True;
    }
#endif

    if(p2 >= 0) {
        a2  = 1;
    } else {
        p2  = -p2;
        a2  = -1;
    }
    s2  = ptable[p2]->ptype;
#ifdef OLDORD
    if(s2 == P_DIRAC || s2 == P_GHOST) {
        b2 = False;
    } else {
        b2 = True;
    }
#else
    if(s2 == P_MAJORANA) {
        s2 = P_DIRAC;
    }
    if(s2 == P_GHOST) {
        b2 = False;
    } else {
        b2 = True;
    }
#endif

    d   = s1 - s2;
    if(d != 0 || b1 || b2) {
        return d;
    }

    /* check particle/anti-paritlce */
    return (a1 - a2);
}

/*==================================================================*/
Private int     pvnenter(n, vp)
int     n;
string  *vp;
/*--------------------------------------------------------------------
 *  Enter vertex.
 */
{
    int    i, nv;

#if DBG
    erdbg("pvnenter", "begin", pvdbg);
#endif

    if(nvtable >= LVTABLE) {
        fprintf(Stderr, 
          "\n*err* pvnenter: vertex table overflow(LVTABLE=%d)\n",LVTABLE);
        exit(1);
    }

    /* create space for new vertex */
    nv = nvtable++;
    if((vtable[nv] = (tvtable*) malloc(sizeof(tvtable))) == NULL) {
        fprintf(Stderr, "\n*err* pvnenter: memory allocation ");
        fprintf(Stderr, "error(1)\n");
        exit(1);
    }
    if((vtable[nv]->vlegs = (pindex*)malloc(sizeof(pindex)*n))
        == NULL) {
        fprintf(Stderr, "\n*err* pvnenter: memory allocation ");
        fprintf(Stderr, "error(2)\n");
        exit(1);
    }

    vtable[nv]->nvlegs = n;

    if(vp == NULL) {
        return nv;
    }

    for(i = 0; i < n; i++) {
        vtable[nv]->vlegs[i] = pvfind(vp[i]);
    }

#if DBG
    erdbg("pvnenter", "end", pvdbg);
#endif

    return nv;
}

/*==================================================================*/
Private int     pcvnenter(n, vp)
int     n;
string  *vp;
/*--------------------------------------------------------------------
 *  Enter vertex.
 */
{
    int    i, nv, vaux;
    bool   err = False;

#if DBG
    erdbg("pcvnenter", "begin", pvdbg);
#endif

    if(n != 3) {
        fprintf(Stderr, "\n*err* pcvnenter: composit vertex:\n");
        fprintf(Stderr, "CVertex={aux-particle, vertex1, vertex2}; ...;");
        fprintf(Stderr, "CVend;\n");
        exit(1);
    }

    if(nvtable >= LVTABLE) {
        fprintf(Stderr, "\n*err* pcvnenter: vertex table overflow\n");
        exit(1);
    }

    /* create space for new vertex */
    nv = nvtable++;
    if((vtable[nv] = (tvtable*) malloc(sizeof(tvtable))) == NULL) {
        fprintf(Stderr, "\n*err* pcvnenter: memory allocation ");
        fprintf(Stderr, "error(1)\n");
        exit(1);
    }
    if((vtable[nv]->auxv = (vindex*)malloc(sizeof(vindex)*2))
        == NULL) {
        fprintf(Stderr, "\n*err* pcvnenter: memory allocation ");
        fprintf(Stderr, "error(2)\n");
        exit(1);
    }

    if(vp == NULL) {
        return nv;
    }

    /* aux particle to be connected */
    vtable[nv]->auxp = pvfind(vp[0]);

    /* vertices to be connected */
    for(i = 1; i < n; i++) {
        vaux = pvidfind(vp[i]);
        if(vaux < 0 || vtable[vaux]->composit != V_AUX) {
            fprintf(Stderr, "\n*err* pcvnenter: there appears ");
            fprintf(Stderr, "vertex which is not auxiliary.");
            fprintf(Stderr, " v = %s, vaux = %d\n", vp[i], vaux);
            err = True;
        } else {
            vtable[nv]->auxv[i-1] = vaux;
        }
    }
    if(err) {
        pvdump();
    }

#if DBG
    erdbg("pcvnenter", "end", pvdbg);
#endif

    return nv;
}

/*==================================================================*/
Private bool    pvncheck(nv)
int     nv;
/*--------------------------------------------------------------------
 *  Check consistency of a vertex "nv".
 *------------------------------------------------------------------*/
{
#ifdef CHANEL
    int i, n, p, t;
    int ft[2], fp[2], nf;

    n = vtable[nv]->nvlegs;

    /* ordering of particle */
    for(i = 0; i < n; i++) {
        if(i > 0 && pvlord(vtable[nv]->vlegs[i-1],
                           vtable[nv]->vlegs[i]) > 0) {
            fprintf(Stderr, "\n*err* pvncheck : wrong order of");
            fprintf(Stderr, " legs %s > %s\n",
                            pvname(vtable[nv]->vlegs[i-1]),
                            pvname(vtable[nv]->vlegs[i]));

            fprintf(Stderr, "Particles in a vertex should be");
            fprintf(Stderr, "arranged in the order of\n");
            fprintf(Stderr, "\tscalar, {majorana, anti-fermion, ");
            fprintf(Stderr, "fermion}, anti-ghost, ghost, vector\n");
            pvndump(nv);
            return False;
        }
    }

#ifndef OLDORD
    /* check Fermion number conservation */
    nf = 0;
    for(i = 0; i < n; i++) {
        p = vtable[nv]->vlegs[i];
        t = ptable[abs(p)]->ptype;
        if(t == P_MAJORANA || t == P_DIRAC) {
            if(nf >= 2) {
                fprintf(Stderr, "*** Vertex with more than two fermions");
                fprintf(Stderr, "is not supported.\n");
                pvndump(nv);
                return False;
            }
            fp[nf]   = p;
            ft[nf++] = t;
        }
    }
    if(nf > 0) {
        if(nf == 1) {
            fprintf(Stderr, "*** Vertex with only one fermion");
            fprintf(Stderr, "is not supported.\n");
            pvndump(nv);
            return False;
        }

        /* standard Fermion vertex is (anti-fermion, fermion) */
        if((ft[0] == P_DIRAC && fp[0] > 0) ||
           (ft[1] == P_DIRAC && fp[1] < 0)) {
            nonstand++;
#ifdef WARNNONSTAND
            fprintf(Stderr, "+++ Non standard Fermion vertex.\n");
            if((ft[0] == P_DIRAC && fp[0] > 0)) {
                fprintf(Stderr,"\t{Charg_conjugate(%s)",
                               pvname(fp[0]));
            } else {
                fprintf(Stderr,"\t{%s", pvname(fp[0]));
            }
            if((ft[1] == P_DIRAC && fp[1] < 0)) {
                fprintf(Stderr,", Charg_conjugate(%s)",
                               pvname(fp[1]));
            } else {
                fprintf(Stderr,", %s", pvname(fp[1]));
            }
            for(i = 0; i < n; i++) {
                p = vtable[nv]->vlegs[i];
                t = ptable[abs(p)]->ptype;
                if(t == P_MAJORANA || t == P_DIRAC) {
                    continue;
                }
                fprintf(Stderr,", %s", pvname(p));
            }
            fprintf(Stderr,"}\n");
#endif
        }
    }
#endif
#endif
    return True;
}

/*==================================================================*/
Private Void    pvendtbl(NoArg)
/*--------------------------------------------------------------------
    Now entering particle and vertices has finished.
    Set up other items of the table here.
--------------------------------------------------------------------*/
{
    int         p, v, l, vc[2], vl[2];
    vindex     *vt[2];

#if DBG
    erdbg("pvendtbl", "begin", pvdbg);
#endif

    /* work space */
    for(l = 0; l < 2; l++) {
        if((vt[l] = (vindex*)malloc(nvtable*sizeof(vindex))) == NULL) {
            fprintf(Stderr, "\n*err* pvendtbl: memory allocation ");
            fprintf(Stderr, "error(1)\n");
            exit(1);
        }
    }

    /* list of vertices using particle "p" */
    for(p = 0; p < nptable; p++) {
        vc[0] = vc[1] = 0;
        vl[0] = vl[1] = VNULL;
        for(v = 0; v < nvtable; v++) {
            for(l = 0; l < vtable[v]->nvlegs; l++) {
                if(v != vl[0] && vtable[v]->vlegs[l] == p) {
                    vt[0][vc[0]++] = v;
                    vl[0] = v;
                }
                if(v != vl[1] && vtable[v]->vlegs[l] == -p) {
                    vt[1][vc[1]++] = v;
                    vl[1] = v;
                }
            }
        }
#if DBG
        if(dbg >= pvdbg) {
            erdbg("pvendtbl", "1", pvdbg);
            printf(" p = %d vc[0] = %d, vc[1] = %d\n", p, vc[0], vc[1]);
        }
#endif
        ptable[p]->nvlist[0] = vc[0];
        if(vc[0] >= 1 &&
           (ptable[p]->vlist[0] = (vindex*)malloc(vc[0]*sizeof(vindex)))
            == NULL) {
            fprintf(Stderr, "\n*err* pvendtbl: memory allocation ");
            fprintf(Stderr, "error(2)\n");
            fprintf(Stderr, "size = %ld\n", ((long)vc[0]*sizeof(vindex)));
            exit(1);
        }
        for(v = 0; v < vc[0]; v++) {
            ptable[p]->vlist[0][v] = vt[0][v];
        }

        ptable[p]->nvlist[1] = vc[1];
        if(vc[1] >= 1 &&
           (ptable[p]->vlist[1] = (vindex*)malloc(vc[1]*sizeof(vindex)))
            == NULL) {
            fprintf(Stderr, "\n*err* pvendtbl: memory allocation ");
            fprintf(Stderr, "error(3)\n");
            fprintf(Stderr, "size = %ld\n", ((long)vc[1]*sizeof(vindex)));
            exit(1);
        }
        for(v = 0; v < vc[1]; v++) {
            ptable[p]->vlist[1][v] = vt[1][v];
        }
    }
    for(l = 0; l < 2; l++) {
        free((AnyP) vt[l]);
    }
#if 0
    /* test type of vertex */
    printf("Test type of vertex\n");
    for(v = 2; v < nvtable; v++) {
        if(vtable[v]->loop == 0) {
            pvpowr(v);
        }
    }
    printf("Test type of vertex : end nvtable=%d\n", nvtable);
#endif

#if DBG
    erdbg("pvendtbl", "end", pvdbg);
#endif
}

/*==================================================================*/
Private bool    pvfort(name)
string  name;
/*--------------------------------------------------------------------
    Check name is fortran name.
--------------------------------------------------------------------*/
{
    int i;

    if(strlen(name) > MAXFNAM) {
        fprintf(Stderr, "\n*err* pvfort: name excees %d characters.\n",
                MAXFNAM);
        fprintf(Stderr, "      name='%s'\n", name);
        return False;
    }
    if(MAXFNAM >= 6 && !isalpha(name[0])) {
        fprintf(Stderr, "\n*err* pvfort: first char of name should ");
        fprintf(Stderr, "be alphabet.\n");
        return False;
    }
    for(i = 1; i < MAXFNAM; i++) {
        if(name[i] == '\0') {
            break;
        }
        if(!isalnum(name[i])) {
            fprintf(Stderr,
                    "\n*err* pvfort : %s contains illegal char",
                     name);
            fprintf(Stderr, "(not alphabet nor decimal digit\n");
            return False;
        }
    }
    for(i = 0; i < strlen(name); i++) {
        name[i] = tolower(name[i]);
    }
    return True;
}
/*********************************************************************
  Output
*/
/*==================================================================*/
Public  Void    pvprint(fn)
string  fn;
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int    p, v;
    ttitle *ttl;

    if(strcmp(fn, "stdout") == 0) {
        pvpf = stdout;
    } else if((pvpf = fopen(fn, "w")) == NULL) {
        fprintf(Stderr, "\n*err* pvprint : open err %s\n", fn);
        return;
    }

    /* table of coupling constatns */
    fprintf(pvpf, "Model=\"%s\";\n\n", fnmdl);
    pvscale('%');
    for(ttl = title; ttl != NULL; ttl = ttl->next) {
        printf("Titel=%s\n", ttl->tstr);
    }
    fprintf(pvpf, "Order={%s", mordtbl[0]);
    for(v = 1; v < nmordtbl; v++) {
        fprintf(pvpf, ",%s", mordtbl[v]);
    }
    fprintf(pvpf, "};\n");

    fprintf(pvpf, "Version={%d", version[0]);
    for(v = 1; v < MAXVER; v++) {
        fprintf(pvpf, ",%d", version[v]);
    }
    fprintf(pvpf, "};\n");

    fprintf(pvpf, "PPhase=%d;", pphase);

    fprintf(pvpf, "Prefix=\"%s\";", mprefix);

    fprintf(pvpf, "%% maxtl = %d, mintl = %d\n", maxtl, mintl);

    pvscale('-');

    /* table of particles */
    for(p = 1; p < nptable; p++) {
        pvprparticle(pvpf, p);
    }

    /* vertices */
    pvscale('-');
    for(v = nvstart; v < nvmodel; v++) {
        pvprvertex(pvpf, v);
    }

    /* end of the file */
    fprintf(pvpf,"Mend;\n");

    if(pvpf != stdout && pvpf != NULL) {
        fclose(pvpf);
        pvpf = NULL;
     }
}

/*==================================================================*/
Public  Void    pvprparticle(fp, p)
FILE    *fp;
int     p;
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int    i;

    if(fp == stdout) {
            fprintf(fp, "%d ", p);
    }

    /* names and print-names of particle and antiparticle */
    fprintf(fp, "Particle=%s[\"%s\"]; ",
            ntable[ptable[p]->pname[0]].name,
            ptable[p]->psym[0]);
    fprintf(fp, "Antiparticle=%s[\"%s\"];",
            ntable[ptable[p]->pname[1]].name,
            ptable[p]->psym[1]);
    fprintf(fp, "\n  ");

    /* mass */
    if(ptable[p]->mdimn < 1) {
        fprintf(fp, "Mass=%s; ", ptable[p]->mass);
    } else {
        fprintf(fp, "Mass=%s(", ptable[p]->mass);
        for(i = 0; i < ptable[p]->mdimn; i++) {
            if(i != 0) {
                fprintf(fp, ",");
            }
            if(ptable[p]->mind[i] > 0) {
                fprintf(fp, "%d/%d",
                    ptable[p]->mind[i], ptable[p]->mdim[i]);
            } else {
                fprintf(fp, "%d", ptable[p]->mdim[i]);
            }
        }
        fprintf(fp, "); ");
    }

    /* width */
    if(ptable[p]->wdimn < 1) {
        fprintf(fp, "  Width=%s; ", ptable[p]->width);
    } else {
        fprintf(fp, "  Width=%s(", ptable[p]->width);
        for(i = 0; i < ptable[p]->wdimn; i++) {
            if(i != 0) {
                fprintf(fp, ",");
            }
            if(ptable[p]->wind[i] > 0) {
                fprintf(fp, "%d/%d",
                    ptable[p]->wind[i], ptable[p]->wdim[i]);
            } else {
                fprintf(fp, "%d", ptable[p]->wdim[i]);
            }
        }
        fprintf(fp, "); ");
    }

    /* ptype */
    fprintf(fp, "PType=%s; ", pvnptype(p));

    /* electric charge */
    if(ptable[p]->charge3%3 == 0) {
        fprintf(fp, "Charge=%d; ", (int) ptable[p]->charge3/3);
    } else {
        fprintf(fp, "Charge=%d/3; ", (int) ptable[p]->charge3);
    }

    fprintf(fp, "\n  ");

    /* color */
    fprintf(fp, "Color=%d; ", ptable[p]->color);

    /* pcode */
    fprintf(fp, "PCode=%d; ", ptable[p]->pcode);

    /* kfcode */
    fprintf(fp, "KFCode=%d; ", ptable[p]->kfcode);

    /* massless */
    if(ptable[p]->massless) {
        fprintf(fp, "Massless; ");
    } else {
        fprintf(fp, "Massive; ");
    }

    /* off shell renormalization */
    if(ptable[p]->extself) {
        fprintf(fp, "Extself=yes; ");
    } else {
        fprintf(fp, "Extself=no; ");
    }

    fprintf(fp, "\n  ");

    /* rname */
    if(ptable[p]->rnami != NULL) {
        if(ptable[p]->rname != NULL) {
            fprintf(fp, "RName={\"%s\", ", ptable[p]->rname);
            fprintf(fp, "\"%s\"}; ",       ptable[p]->rnami);
        } else {
            fprintf(fp, "RName={?,\"%s\"};", ptable[p]->rnami);
        }
    } else {
        if(ptable[p]->rname != NULL) {
            fprintf(fp, "RName=\"%s\"; ", ptable[p]->rname);
        }
    }

    /* sname */
    if(ptable[p]->snami != NULL) {
        if(ptable[p]->sname != NULL) {
            fprintf(fp, "SName={\"%s\", ", ptable[p]->sname);
            fprintf(fp, "\"%s\"}; ",       ptable[p]->snami);
        } else {
            fprintf(fp, "SName={?,\"%s\"};", ptable[p]->snami);
        }
    } else {
        if(ptable[p]->sname != NULL) {
            fprintf(fp, "SName=\"%s\"; ", ptable[p]->sname);
        }
    }

    /* chanel */
    if(ptable[p]->chani != NULL) {
        if(ptable[p]->chane != NULL) {
            fprintf(fp, "Chanel={\"%s\", ", ptable[p]->chane);
            fprintf(fp, "\"%s\"}; ",        ptable[p]->chani);
        } else {
            fprintf(fp, "Chanel={?,\"%s\"};", ptable[p]->chani);
        }
    } else {
        if(ptable[p]->chane != NULL) {
            fprintf(fp, "Chanel=\"%s\"; ", ptable[p]->chane);
        }
    }

    /* gauge */
    if(ptable[p]->gauge != NULL) {
        fprintf(fp, "Gauge=\"%s\"; ", ptable[p]->gauge);
    }

    /* pselect */
    if(ptable[p]->pselect != NULL) {
        fprintf(fp, "PSelect=\"%s\"; ", ptable[p]->pselect);
    }
    fprintf(fp, "\n");

    /* end of the particle */
    fprintf(fp,"Pend;\n");
}

/*==================================================================*/
Public  Void    pvprvertex(fp, v)
FILE    *fp;
int     v;
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int    p, i, j;

    if(fp == stdout) {
        fprintf(fp, "%d ", v);
    }

    /* list of particle */
    if(vtable[v]->composit == V_COMPOSIT) {
        fprintf(fp, "CVertex={%s", pvname(vtable[v]->auxp));
        for(p = 0; p < 2; p++) {
            fprintf(fp, ", %s", vtable[vtable[v]->auxv[p]]->vid);
        }
        fprintf(fp,"}; ");
    } else {
        fprintf(fp, "Vertex={%s", pvname(vtable[v]->vlegs[0]));
        for(p = 1; p < vtable[v]->nvlegs; p++) {
            fprintf(fp, ", %s", pvname(vtable[v]->vlegs[p]));
        }
        fprintf(fp,"}; ");
    }

    /* coupling constants */
    for(p = 0; p < nmordtbl; p++) {
        fprintf(fp, "%s=%d; ",
                mordtbl[p], vtable[v]->morder[p]);
    }
    fprintf(fp,"\n  ");

    /* fortran name */
    for(j = 0; j < NFNAME; j++) {
        if(vtable[v]->fndimn[j] > 0) {
            if(j == 0) {
                fprintf(fp, "FName");
            } else if(j == 1) {
                fprintf(fp, "CFName");
            } else if(j == 2) {
                fprintf(fp, "EFName");
            } else {
                fprintf(fp, "CEFName");
            }
            fprintf(fp, "=%s(", vtable[v]->fname[j]);
            for(i = 0; i < vtable[v]->fndimn[j]; i++) {
                if(i > 0) {
                    fprintf(fp, ",");
                }
                if(vtable[v]->fnind[j][i] > 0) {
                    fprintf(fp, "%d/%d",
                            vtable[v]->fnind[j][i], vtable[v]->fndim[j][i]);
                } else {
                    fprintf(fp, "%d", vtable[v]->fndim[j][i]);
                }
            }
            fprintf(fp, "); ");
        } else if(vtable[v]->fname[j] != NULL) {
            if(j == 0) {
                fprintf(fp, "FName");
            } else if(j == 1) {
                fprintf(fp, "CFName");
            } else if(j == 2) {
                fprintf(fp, "EFName");
            } else {
                fprintf(fp, "CEFName");
            }
            fprintf(fp, "=%s; ", vtable[v]->fname[j]);
        }
    }

    if(vtable[v]->rname != NULL) {
        fprintf(fp, "RName=%s; ", vtable[v]->rname);
    }

    if(vtable[v]->sname != NULL) {
        fprintf(fp, "SName=%s; ", vtable[v]->sname);
    }

    if(vtable[v]->chanel != NULL) {
        fprintf(fp, "Chanel=%s; ", vtable[v]->chanel);
    }

    if(vtable[v]->vselect != NULL) {
        fprintf(fp, "VSelect=%s; ", vtable[v]->vselect);
    }

    if(vtable[v]->vid != NULL) {
        fprintf(fp, "VId=%s; ", vtable[v]->vid);
    }

    /* type of fermion coupling */
    if(vtable[v]->ctype == V_V) {
        fprintf(fp, "FType=\"V\"; ");
    } else if(vtable[v]->ctype == V_A) {
        fprintf(fp, "FType=\"A\"; ");
    } else if(vtable[v]->ctype == V_VminusA) {
        fprintf(fp, "FType=\"V-A\"; ");
    } else if(vtable[v]->ctype == V_VplusA) {
        fprintf(fp, "FType=\"V+A\"; ");
    } else {
    }
    fprintf(fp,"\n");

    /* end of the vertex */
    if(vtable[v]->composit == V_COMPOSIT) {
        fprintf(fp,"CVend;\n");
        fprintf(fp,"%%vtype=%d;\n",vtable[v]->vtype);
    } else {
        fprintf(fp,"Vend;\n");
    }
}

/*==================================================================*/
Public  Void    pvprnam(fn)
string  fn;
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int    p, v, i, j;
    ttitle *ttl;

    if(strcmp(fn, "stdout") == 0) {
        pvpf = stdout;
    } else if((pvpf = fopen(fn, "w")) == NULL) {
        fprintf(Stderr, "\n*err* pvprint : open err %s\n", fn);
        return;
    }

    fprintf(pvpf, "Model=\"%s\";\n\n", fnmdl);

    for(ttl = title; ttl != NULL; ttl = ttl->next) {
        fprintf(pvpf, "Title=%s\n", ttl->tstr);
    }

    /* table of coupling constatns */
    fprintf(pvpf, "Order={%s", mordtbl[0]);
    for(v = 1; v < nmordtbl; v++) {
        fprintf(pvpf, ",%s", mordtbl[v]);
    }
    fprintf(pvpf, "};\n");

    fprintf(pvpf, "Version={%d", version[0]);
    for(v = 1; v < MAXVER; v++) {
        fprintf(pvpf, ",%d", version[v]);
    }
    fprintf(pvpf, "};\n");

    fprintf(pvpf, "PPhase=%d;\n", pphase);

    if(mprefix == NULL) {
        fprintf(pvpf, "Prefix=NULL;\n");
    } else {
        fprintf(pvpf, "Prefix=\"%s\";\n", mprefix);
    }

    /* table of coupling constatns */
    pvscale('%');

    /* table of particles */
    for(p = 1; p < nptable; p++) {
        fprintf(pvpf, "No=%d; ", p);

        /* mass */
        if(ptable[p]->mdimn < 1) {
            fprintf(pvpf, "  Mass=%-8s; ", ptable[p]->mass);
        } else {
            fprintf(pvpf, "  Mass=%s(", ptable[p]->mass);
            for(i = 0; i < ptable[p]->mdimn; i++) {
                if(i != 0) {
                    fprintf(pvpf, ",");
                }
                if(ptable[p]->mind[i] > 0) {
                    fprintf(pvpf, "%d/%d",
                        ptable[p]->mind[i], ptable[p]->mdim[i]);
                } else {
                    fprintf(pvpf, "%d", ptable[p]->mdim[i]);
                }
            }
            fprintf(pvpf, "); ");
        }

        /* width */
        if(ptable[p]->wdimn < 1) {
            fprintf(pvpf, "  Width=%-8s; ", ptable[p]->width);
        } else {
            fprintf(pvpf, "  Width=%s(", ptable[p]->width);
            for(i = 0; i < ptable[p]->wdimn; i++) {
                if(i != 0) {
                    fprintf(pvpf, ",");
                }
                if(ptable[p]->wind[i] > 0) {
                    fprintf(pvpf, "%d/%d",
                        ptable[p]->wind[i], ptable[p]->wdim[i]);
                } else {
                    fprintf(pvpf, "%d", ptable[p]->wdim[i]);
                }
            }
            fprintf(pvpf, "); ");
        }

        if(ptable[p]->ptype == P_AUX) {
            fprintf(pvpf, "Aux; ");
        }

        /* names and print-names of particle and antiparticle */
        fprintf(pvpf, "Particle=%s; \n",
                ntable[ptable[p]->pname[0]].name);

    }

    /* vertices */
    pvscale('-');
    for(v = nvstart; v < nvtable; v++) {
        fprintf(pvpf, "No=%d;", v);

        /* fortran name */
        for(j = 0; j < NFNAME; j++) {
            if(vtable[v]->fndimn[j] > 0) {
                if(j == 0) {
                    fprintf(pvpf, "FName");
                } else if(j == 1) {
                    fprintf(pvpf, "CFName");
                } else if(j == 2) {
                    fprintf(pvpf, "EFName");
                } else {
                    fprintf(pvpf, "CEFName");
                }
                fprintf(pvpf, "=%-s(", vtable[v]->fname[j]);
                for(i = 0; i < vtable[v]->fndimn[j]; i++) {
                    if(i > 0) {
                        fprintf(pvpf, ",");
                    }
                    if(vtable[v]->fnind[j][i] > 0) {
                        fprintf(pvpf, "%d/%d",
                                vtable[v]->fnind[j][i], vtable[v]->fndim[j][i]);
                    } else {
                        fprintf(pvpf, "%d", vtable[v]->fndim[j][i]);
                    }
                }
                fprintf(pvpf, "); ");
            } else if(vtable[v]->fname[j] != NULL) {
                if(j == 0) {
                    fprintf(pvpf, "FName");
                } else if(j == 1) {
                    fprintf(pvpf, "CFName");
                } else if(j == 2) {
                    fprintf(pvpf, "EFName");
                } else {
                    fprintf(pvpf, "CEFName");
                }
                fprintf(pvpf, "=%-11s; ", vtable[v]->fname[j]);
            }
        }

        if(vtable[v]->rname != NULL) {
            fprintf(pvpf, "RName=%-8s; ", vtable[v]->rname);
        }

        if(vtable[v]->sname != NULL) {
            fprintf(pvpf, "SName=%-8s; ", vtable[v]->sname);
        }

        if(vtable[v]->chanel != NULL) {
            fprintf(pvpf, "Chanel=%-8s; ", vtable[v]->chanel);
        }

        if(vtable[v]->vselect != NULL) {
            fprintf(pvpf, "VSelect=%-8s; ", vtable[v]->vselect);
        }

        if(vtable[v]->vid != NULL) {
            fprintf(pvpf, "VId=%s; ", vtable[v]->vid);
        }

        if(vtable[v]->composit == V_COMPOSIT) {
            fprintf(pvpf, "Composit; ");
        } else if(vtable[v]->composit == V_AUX) {
            fprintf(pvpf, "Aux; ");
        }

        /* list of particle */
        if(vtable[v]->composit == V_COMPOSIT) {
            fprintf(pvpf, "CVertex={%s", pvname(vtable[v]->auxp));
            for(p = 0; p < 2; p++) {
                fprintf(pvpf, ", %s", vtable[vtable[v]->auxv[p]]->vid);
            }
            fprintf(pvpf,"}; \n\t");
        }
        fprintf(pvpf, "Vertex={%s", pvname(vtable[v]->vlegs[0]));
        for(p = 1; p < vtable[v]->nvlegs; p++) {
            fprintf(pvpf, ", %s", pvname(vtable[v]->vlegs[p]));
        }
        fprintf(pvpf,"}; \n");
    }

    if(pvpf != stdout && pvpf != NULL) {
        fclose(pvpf);
        pvpf = NULL;
    }
}

/*==================================================================*/
Public  Void    pvprgnam(fn)
string  fn;
/*--------------------------------------------------------------------
    Dump gname table.
--------------------------------------------------------------------*/
{
    int    p;

    if(strcmp(fn, "stdout") == 0) {
        pvpf = stdout;
    } else if((pvpf = fopen(fn, "w")) == NULL) {
        fprintf(Stderr, "\n*err* pvprint : open err %s\n", fn);
        return;
    }

    /* table of particles */
    for(p = 1; p < nptable; p++) {

        /* mass and width */
        fprintf(pvpf, "%-20s %-20s %-20s %-20s\n",
                ntable[ptable[p]->pname[0]].name, ptable[p]->gname[0],
                ptable[p]->gname[1], ptable[p]->gname[2]);
        if(ptable[p]->pname[0] != ptable[p]->pname[1]) {
            fprintf(pvpf, "%-20s %-20s %-20s %-20s\n",
                    ntable[ptable[p]->pname[1]].name, ptable[p]->gname[0],
                    ptable[p]->gname[2], ptable[p]->gname[1]);
        }
    }

    if(pvpf != stdout && pvpf != NULL) {
        fclose(pvpf);
        pvpf = NULL;
    }
}

/*==================================================================*/
Public  Void    pvprtex(fn)
string  fn;
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int n;

    if(strcmp(fn, "stdout") == 0) {
        pvpf = stdout;
    } else if((pvpf = fopen(fn, "w")) == NULL) {
        fprintf(Stderr, "\n*err* pvprtex : open err %s\n", fn);
        return;
    }

    fprintf(pvpf, "\\documentstyle[12pt]{article}\n");
    fprintf(pvpf, "\\begin{document}\n {\\small\n");

    n = pvprtxp();
    n = pvprtv1();
    pvprtv2(n);

    fprintf(pvpf, "}\\end{document}\n");

    if(pvpf != stdout && pvpf != NULL) {
        fclose(pvpf);
        pvpf = NULL;
    }
}

/*==================================================================*/
Private int     pvprtxp(NoArg)
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    int    p, v, i;
    ttitle *ttl;

    /* table of coupling constatns */
    pvscale('%');

    fprintf(pvpf, "{\\bf Filename : }\n");
    fprintf(pvpf, "\"%s\"\n\n", fnmdl);

    if(title != NULL) {
        fprintf(pvpf, "\\begin{verbatim}\n");
        for(ttl = title; ttl != NULL; ttl = ttl->next) {
            fprintf(pvpf, "%s\n", ttl->tstr);
        }
        fprintf(pvpf, "\\end{verbatim}\n");
    }

    fprintf(pvpf, "{\\bf Order : }\n");
    for(v = 0; v < nmordtbl; v++) {
        fprintf(pvpf, "%s ", mordtbl[v]);
    }
    fprintf(pvpf, "\n\n");

    fprintf(pvpf, "{\\bf Version : }\n");
    for(v = 0; v < MAXVER; v++) {
        fprintf(pvpf, "%d ", version[v]);
    }
    fprintf(pvpf, "\n\n");

    fprintf(pvpf, "{\\bf PPhase : } %d\n\n", pphase);

    fprintf(pvpf, "{\\bf Prefix : } \"%s\"\n\n", mprefix);

    /* table of particles */
    pvscale('%');
    fprintf(pvpf, "{\\bf Particles}\n\n");
    fprintf(pvpf, "{\\footnotesize \n");

    /* page 1: item 1--8 */
    fprintf(pvpf, "\\begin{tabular}{|r");
    for(p = 2; p <= 8; p++) {
        if(p >= 4) {
            fprintf(pvpf, "|r");
        } else {
            fprintf(pvpf, "|l");
        }
    }
    fprintf(pvpf, "|}\n\\hline\\hline\n");

    fprintf(pvpf, "1");
    for(p = 2; p <= 8; p++) {
        fprintf(pvpf, "& %d", p);
    }
    fprintf(pvpf, "\\\\ \n\\hline\n");

    for(p = 1; p < nptable; p++) {

        /* 1 : seq. num. */
        fprintf(pvpf, " %d ", p);  /* 1 */

        /* 2 : name of particle */
        fprintf(pvpf, " & %s ", ntable[ptable[p]->pname[0]].name);

        /* 3 : name of anti-particle */
        fprintf(pvpf, " & %s ", ntable[ptable[p]->pname[1]].name);

        /* 4 : ptype */
        fprintf(pvpf, " & %s ", pvnptype(p));

        /* 5 : electric charge */
        if(ptable[p]->charge3%3 == 0) {
            fprintf(pvpf, " & %d ", (int) ptable[p]->charge3/3);
        } else {
            fprintf(pvpf, " & %d/3 ", (int) ptable[p]->charge3);
        }

        /* 6 : color */
        fprintf(pvpf, " & %d ", ptable[p]->color);

        /* 7 : pcode */
        fprintf(pvpf, " & %d ", ptable[p]->pcode);

        /* 8 : kfcode */
        fprintf(pvpf, " & %d ", ptable[p]->kfcode);

        /* end of the particle */
        fprintf(pvpf,"\\\\ \n");
    }
    fprintf(pvpf, "\\hline\\hline\n");
    fprintf(pvpf,"\\end{tabular}\n\n");

    /* page 2: 1, 2, 9--11 */
    fprintf(pvpf, "\\begin{tabular}{|r|l");
    for(p = 9; p <= 11; p++) {
        fprintf(pvpf, "|l");
    }
    fprintf(pvpf, "|}\n\\hline\\hline\n");

    fprintf(pvpf, "1 & 2");
    for(p = 9; p <= 11; p++) {
        fprintf(pvpf, "& %d", p);
    }
    fprintf(pvpf, "\\\\ \n\\hline\n");

    for(p = 1; p < nptable; p++) {

        /* 1 : seq. num. */
        fprintf(pvpf, " %d ", p);  /* 1 */

        /* 2 : name of particle */
        fprintf(pvpf, " & %s ", ntable[ptable[p]->pname[0]].name);

        /* 9 : massless */
        if(ptable[p]->massless) {
            fprintf(pvpf, " & massless ");
        } else {
            fprintf(pvpf, " & massive  ");
        }

        /* 10 : mass */
        if(ptable[p]->mdimn < 1) {
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->mass);
        } else {
            fprintf(pvpf, " & {\\tt %s(", ptable[p]->mass);
            for(i = 0; i < ptable[p]->mdimn; i++) {
                if(i != 0) {
                    fprintf(pvpf, ",");
                }
                if(ptable[p]->mind[i] > 0) {
                    fprintf(pvpf, "%d/%d",
                        ptable[p]->mind[i], ptable[p]->mdim[i]);
                } else {
                    fprintf(pvpf, "%d", ptable[p]->mdim[i]);
                }
            }
            fprintf(pvpf, ")} ");
        }

        /* 11 : width */
        if(ptable[p]->wdimn < 1) {
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->width);
        } else {
            fprintf(pvpf, " & {\\tt %s(", ptable[p]->width);
            for(i = 0; i < ptable[p]->wdimn; i++) {
                if(i != 0) {
                    fprintf(pvpf, ",");
                }
                if(ptable[p]->wind[i] > 0) {
                    fprintf(pvpf, "%d/%d",
                        ptable[p]->wind[i], ptable[p]->wdim[i]);
                } else {
                    fprintf(pvpf, "%d", ptable[p]->wdim[i]);
                }
            }
            fprintf(pvpf, ")} ");
        }

        /* end of the particle */
        fprintf(pvpf,"\\\\ \n");
    }
    fprintf(pvpf, "\\hline\\hline\n");
    fprintf(pvpf,"\\end{tabular}\n\n");

    /* page 3. item 1, 2, 12-- */
    fprintf(pvpf, "\\begin{tabular}{|r|l");
    for(p = 12; p <= 22; p++) {
        fprintf(pvpf, "|l");
    }
    fprintf(pvpf, "|}\n\\hline\\hline\n");

    fprintf(pvpf, "1 & 2 ");
    for(p = 12; p <= 22; p++) {
        fprintf(pvpf, "& %d", p);
    }
    fprintf(pvpf, "\\\\ \n\\hline\n");

    for(p = 1; p < nptable; p++) {

        /* 1 : seq. num. */
        fprintf(pvpf, " %d ", p);  /* 1 */

        /* 2 : name of particle */
        fprintf(pvpf, " & %s ", ntable[ptable[p]->pname[0]].name);

        /* 12 : chane */
        if(ptable[p]->chane == NULL) {
            fprintf(pvpf, " & --- ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->chane);
        }

        /* 13 : chani */
        if(ptable[p]->chani == NULL) {
            fprintf(pvpf, " & --- ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->chani);
        }

        /* 14 : sname */
        if(ptable[p]->sname == NULL) {
            fprintf(pvpf, " & --- ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->sname);
        }

        /* 15 : snami */
        if(ptable[p]->snami == NULL) {
            fprintf(pvpf, " & --- ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->snami);
        }

        /* 16 : rname */
        if(ptable[p]->rname == NULL) {
            fprintf(pvpf, " & --- ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->rname);
        }

        /* 17 : rnami */
        if(ptable[p]->rnami == NULL) {
            fprintf(pvpf, " & --- ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->rnami);
        }

        /* 18 : gauge */
        if(ptable[p]->gauge == NULL) {
            fprintf(pvpf, " & --- ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->gauge);
        }

        /* 19 : pselect */
        if(ptable[p]->pselect == NULL) {
            fprintf(pvpf, " & --- ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", ptable[p]->pselect);
        }

        /* 20 : extself */
        if(ptable[p]->extself) {
            fprintf(pvpf, " & Off-shell ");
        } else {
            fprintf(pvpf, " & On-shell ");
        }

        /* end of the particle */
        fprintf(pvpf,"\\\\ \n");
    }
    fprintf(pvpf, "\\hline\\hline\n");
    fprintf(pvpf,"\\end{tabular}\n\n");
    fprintf(pvpf,"\\newpage\n\n");
    fprintf(pvpf,"\\begin{itemize}\n");
    fprintf(pvpf,"\\item[ 1] No.\n");
    fprintf(pvpf,"\\item[ 2] Name of particle\n");
    fprintf(pvpf,"\\item[ 3] Name of anti-particle\n");
    fprintf(pvpf,"\\item[ 4] Type\n");
    fprintf(pvpf,"\\item[ 5] Charge\n");
    fprintf(pvpf,"\\item[ 6] Color\n");
    fprintf(pvpf,"\\item[ 7] Particle code\n");
    fprintf(pvpf,"\\item[ 8] KF Particle code\n");
    fprintf(pvpf,"\\item[ 9] Massive or massless\n");
    fprintf(pvpf,"\\item[10] Mass\n");
    fprintf(pvpf,"\\item[11] Width\n");
    fprintf(pvpf,"\\item[12] Name of the chanel (ext. part.)\n");
    fprintf(pvpf,"\\item[13] Name of the chanel (int. part.)\n");
    fprintf(pvpf,"\\item[14] Name of the external particle\n");
    fprintf(pvpf,"\\item[15] Name of the propagator\n");
    fprintf(pvpf,"\\item[16] Name for reduce (proj. op.)\n");
    fprintf(pvpf,"\\item[17] Name for reduce (propagator)\n");
    fprintf(pvpf,"\\item[18] Name of gauge parameter\n");
    fprintf(pvpf,"\\item[19] Name of graph selection\n");
    fprintf(pvpf,"\\item[20] Off-shell/on-shell renormalization\n");
    fprintf(pvpf,"\\item[21] Factor for diagonal external selfenergy\n");
    fprintf(pvpf,"\\item[22] Factor for off-diagonal external selfenergy\n");
    fprintf(pvpf,"\\end{itemize}\n");
    fprintf(pvpf,"} %% footnotsize\n");
    fprintf(pvpf,"\\newpage\n");

    return 16;
}

/*==================================================================*/
Private int     pvprtv1(NoArg)
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    int    p, v, i, j;

    /* vertices */
    pvscale('%');
    fprintf(pvpf, "{\\bf Vertices}\n\n");
    fprintf(pvpf, "{\\footnotesize \n");

    for(v = nvstart; v < nvtable; v++) {

        if((v - nvstart) % 40 == 0) {
            if(v != nvstart) {
                fprintf(pvpf, "\\hline\\hline \n");
                fprintf(pvpf,"\\end{tabular}\n\n");
            }

            fprintf(pvpf, "\\begin{tabular}{|r");
            for(p = 1; p < 5+NFNAME+nmordtbl; p++) {
                fprintf(pvpf, "|l");
            }
            fprintf(pvpf, "|}\n\\hline\\hline\n");

            fprintf(pvpf, "1");
            for(p = 1; p < 5+NFNAME+nmordtbl; p++) {
                fprintf(pvpf, "& %d", p+1);
            }
            fprintf(pvpf, "\\\\ \n\\hline\n");
        }

        /* 1 : No. */
        fprintf(pvpf, "%d ", v);

        /* 2 - 5 : list of particle */
        if(vtable[v]->composit == V_COMPOSIT) {
            fprintf(pvpf, " & %s ", pvname(vtable[v]->auxp));
            for(p = 0; p < 2; p++) {
                fprintf(pvpf, " & %s ", vtable[vtable[v]->auxv[p]]->vid);
            }
            fprintf(pvpf, " & -- ");
        } else {
            for(p = 0; p < 4; p++) {
                if(p < vtable[v]->nvlegs) {
                    fprintf(pvpf, " & %s ", pvname(vtable[v]->vlegs[p]));
                } else {
                    fprintf(pvpf, " & -- ");
                }
            }
        }

        /* 6 - (5+nmordtbl) : coupling constants*/
        for(p = 0; p < nmordtbl; p++) {
            fprintf(pvpf, " & %d ", vtable[v]->morder[p]);
        }

        /* (6) + (NFNAME-1) +nmordtbl : fortran name */
        for(j = 0; j < NFNAME; j++) {
            if(vtable[v]->fndimn[j] > 0) {
                fprintf(pvpf, " & {\\tt %s(", vtable[v]->fname[j]);
                for(i = 0; i < vtable[v]->fndimn[j]; i++) {
                    if(i > 0) {
                        fprintf(pvpf, ",");
                    }
                    if(vtable[v]->fnind[j][i] > 0) {
                        fprintf(pvpf, "%d/%d",
                                vtable[v]->fnind[j][i], vtable[v]->fndim[j][i]);
                    } else {
                        fprintf(pvpf, "%d", vtable[v]->fndim[j][i]);
                    }
                }
                fprintf(pvpf, ")} ");
            } else if(vtable[v]->fname[j] != NULL) {
                fprintf(pvpf, " & {\\tt %s} ", vtable[v]->fname[j]);
            } else {
                fprintf(pvpf, " & ");
            }
        }

        /* end of the vertex */
        fprintf(pvpf,"\\\\ \n");
    }
    fprintf(pvpf, "\\hline\\hline \n");
    fprintf(pvpf,"\\end{tabular}\n\n");
    fprintf(pvpf,"\\begin{itemize}\n");

    fprintf(pvpf,"\\item[ 1] No.\n");
    fprintf(pvpf,"\\item[ 2 -- 5] list of particle \n");
    for(p = 0; p < nmordtbl; p++) {
        fprintf(pvpf,"\\item[ %d] ", p+6);
        fprintf(pvpf," order of coupling constnat : {\\tt %s}\n",
                mordtbl[p]);
    }
    fprintf(pvpf,"\\item[%d] Name of vertex\n", p+6);

    fprintf(pvpf,"\\end{itemize}\n");
    fprintf(pvpf,"} %% footnotsize\n");

    fprintf(pvpf,"\\newpage\n");

    return p+6;
}


/*==================================================================*/
Private Void    pvprtv2(offs)
int     offs;
/*--------------------------------------------------------------------
--------------------------------------------------------------------*/
{
    int    p = 0, v, i, j;

    /* vertices */
    pvscale('%');
    fprintf(pvpf, "{\\bf Values of coupling const}\n\n");
    fprintf(pvpf, "{\\footnotesize \n");

    for(v = nvstart; v < nvtable; v++) {

        if((v - nvstart) % 40 == 0) {
            if(v != nvstart) {
                fprintf(pvpf, "\\hline\\hline \n");
                fprintf(pvpf,"\\end{tabular}\n\n");
            }

            fprintf(pvpf, "\\begin{tabular}{|r");
            for(p = offs; p < offs+7+NFNAME; p++) {
                fprintf(pvpf, "|l");
            }
            fprintf(pvpf, "|}\n\\hline\\hline\n");

            fprintf(pvpf, "1");
            for(p = offs; p < offs+7+NFNAME; p++) {
                fprintf(pvpf, "& %d", p+1);
            }
            fprintf(pvpf, "\\\\ \n\\hline\n");
        }

        /* 0 : No. */
        fprintf(pvpf, "%d ", v);

        /* 1, 2 : fortran name */
        for(j = 0; j < NFNAME; j++) {
            if(vtable[v]->fndimn[j] > 0) {
                fprintf(pvpf, " & {\\tt %s(", vtable[v]->fname[j]);
                for(i = 0; i < vtable[v]->fndimn[j]; i++) {
                    if(i > 0) {
                        fprintf(pvpf, ",");
                    }
                    if(vtable[v]->fnind[j][i] > 0) {
                        fprintf(pvpf, "%d/%d",
                                vtable[v]->fnind[j][i], vtable[v]->fndim[j][i]);
                    } else {
                        fprintf(pvpf, "%d", vtable[v]->fndim[j][i]);
                    }
                }
                fprintf(pvpf, ")} ");
            } else if(vtable[v]->fname[j] != NULL) {
                fprintf(pvpf, " & {\\tt %s} ", vtable[v]->fname[j]);
            } else {
                fprintf(pvpf, " & ");
            }
        }

        /* 3 : fortran subroutine name */
        if(vtable[v]->chanel == NULL) {
            fprintf(pvpf, " & {\\tt --} ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", vtable[v]->chanel);
        }

        /* 4 : fortran subroutine name */
        if(vtable[v]->sname == NULL) {
            fprintf(pvpf, " & {\\tt --} ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", vtable[v]->sname);
        }

        /* 5 : reduce name */
        if(vtable[v]->rname == NULL) {
            fprintf(pvpf, " & {\\tt --} ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", vtable[v]->rname);
        }

        /* 6 : graph selection */
        if(vtable[v]->vselect == NULL) {
            fprintf(pvpf, " & {\\tt --} ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", vtable[v]->vselect);
        }

        /* 7 : graph selection */
        if(vtable[v]->vid == NULL) {
            fprintf(pvpf, " & {\\tt --} ");
        } else {
            fprintf(pvpf, " & {\\tt %s} ", vtable[v]->vid);
        }

        /* 8 : type of fermion coupling */
        if(vtable[v]->ctype == V_V) {
            fprintf(pvpf, " & V ");
        } else if(vtable[v]->ctype == V_A) {
            fprintf(pvpf, " & A ");
        } else if(vtable[v]->ctype == V_VminusA) {
            fprintf(pvpf, " & V-A ");
        } else if(vtable[v]->ctype == V_VplusA) {
            fprintf(pvpf, " & V+A ");
        } else {
            fprintf(pvpf, " & -- ");
        }

        /* 9 : type of fermion coupling */
        if(vtable[v]->composit == V_NORMAL) {
            fprintf(pvpf, " & N ");
        } else if(vtable[v]->composit == V_AUX) {
            fprintf(pvpf, " & A ");
        } else if(vtable[v]->composit == V_COMPOSIT) {
            fprintf(pvpf, " & C ");
        } else {
            fprintf(pvpf, " & ? ");
        }

        /* end of the vertex */
        fprintf(pvpf,"\\\\ \n");
    }
    fprintf(pvpf, "\\hline\\hline \n");
    fprintf(pvpf,"\\end{tabular}\n\n");
    fprintf(pvpf,"\\begin{itemize}\n");
    fprintf(pvpf,"\\item[ 1] No.\n");
    fprintf(pvpf,"\\item[ %d] Name of vertex\n", offs+1);
    fprintf(pvpf,"\\item[ %d] Name of chanel\n", offs+2);
    fprintf(pvpf,"\\item[ %d] Name of subroutine\n", offs+3);
    fprintf(pvpf,"\\item[ %d] Name for reduce\n", offs+4);
    fprintf(pvpf,"\\item[ %d] Name of graph selection\n", offs+5);
    fprintf(pvpf,"\\item[ %d] Vertex identifier\n", offs+6);
    fprintf(pvpf,"\\item[ %d] Type of fermion vertex\n", offs+7);
    fprintf(pvpf,"\\item[ %d] Normal, Aux, Composit\n", offs+8);
    fprintf(pvpf,"\\end{itemize}\n");
    fprintf(pvpf,"} %% footnotsize\n");
}

/*==================================================================*/
Public   Void    pvldump(out, len, lst)
FILE    *out;
int      len;
vindex   lst[];
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int i;

    fprintf(out, "{");
    for(i = 0; i < len; i++) {
        fprintf(out, " %d", lst[i]);
    }
    fprintf(out,"}\n");
}

/*==================================================================*/
Public  Void    pvpdump(p)
int     p;
/*--------------------------------------------------------------------
 *  Dump a particle.
 *------------------------------------------------------------------*/
{
    int    l, i;

    fprintf(Stderr, "%d ", p);

    /* names and print-names of particle and antiparticle */
    fprintf(Stderr, "Particle=%s[\"%s\"]; ",
            ntable[ptable[p]->pname[0]].name,
            ptable[p]->psym[0]);
    fprintf(Stderr, "Antiparticle=%s[\"%s\"];\n",
            ntable[ptable[p]->pname[1]].name,
            ptable[p]->psym[1]);

    /* mass */
    if(ptable[p]->mdimn < 1) {
        fprintf(Stderr, "  Mass=%s; ", ptable[p]->mass);
    } else {
        fprintf(Stderr, "  Mass=%s(", ptable[p]->mass);
        for(i = 0; i < ptable[p]->mdimn; i++) {
            if(i != 0) {
                fprintf(Stderr, ",");
            }
            if(ptable[p]->mind[i] > 0) {
                fprintf(Stderr, "%d/%d",
                    ptable[p]->mind[i], ptable[p]->mdim[i]);
            } else {
                fprintf(Stderr, "%d", ptable[p]->mdim[i]);
            }
        }
        fprintf(Stderr, "); ");
    }

    /* width */
    if(ptable[p]->wdimn < 1) {
        fprintf(Stderr, "  Width=%s; ", ptable[p]->width);
    } else {
        fprintf(Stderr, "  Width=%s(", ptable[p]->width);
        for(i = 0; i < ptable[p]->wdimn; i++) {
            if(i != 0) {
                fprintf(Stderr, ",");
            }
            if(ptable[p]->wind[i] > 0) {
                fprintf(Stderr, "%d/%d",
                    ptable[p]->wind[i], ptable[p]->wdim[i]);
            } else {
                fprintf(Stderr, "%d", ptable[p]->wdim[i]);
            }
        }
        fprintf(Stderr, "); ");
    }

    /* ptype */
    fprintf(Stderr, "PType=%s; ", pvnptype(p));

    /* electric charge */
    fprintf(Stderr, "Charge3=%d; ", (int) ptable[p]->charge3);

    /* color */
    fprintf(Stderr, "Color=%d; ", ptable[p]->color);

    /* pcode */
    fprintf(Stderr, "PCode=%d; ", ptable[p]->pcode);

    /* kfcode */
    fprintf(Stderr, "KFCode=%d; ", ptable[p]->kfcode);

    /* massless */
    if(ptable[p]->massless) {
        fprintf(Stderr, "massless; ");
    } else {
        fprintf(Stderr, "massive; ");
    }

    /* chanel */
    if(ptable[p]->chane == NULL) {
        if(ptable[p]->chani == NULL) {
            fprintf(Stderr, "Chanel=No; ");
        } else {
            fprintf(Stderr, "Chanel={?, \"%s\"}; ", ptable[p]->chani);
        }
    } else if(ptable[p]->chani == NULL) {
        fprintf(Stderr, "Chanel=\"%s\"; ", ptable[p]->chane);
    } else {
        fprintf(Stderr, "Chanel={\"%s\", ", ptable[p]->chane);
        fprintf(Stderr, "\"%s\"}; ", ptable[p]->chani);
    }

    /* rname */
    if(ptable[p]->rname == NULL) {
        if(ptable[p]->rnami == NULL) {
            fprintf(Stderr, "RName=No; ");
        } else {
            fprintf(Stderr, "RName={?, \"%s\"}; ", ptable[p]->rnami);
        }
    } else if(ptable[p]->rnami == NULL) {
        fprintf(Stderr, "RName=\"%s\"; ", ptable[p]->rname);
    } else {
        fprintf(Stderr, "RName={\"%s\", ", ptable[p]->rname);
        fprintf(Stderr, "\"%s\"}; ", ptable[p]->rnami);
    }

    /* sname */
    if(ptable[p]->sname == NULL) {
        if(ptable[p]->snami == NULL) {
            fprintf(Stderr, "SName=No; ");
        } else {
            fprintf(Stderr, "SName={?, \"%s\"}; ", ptable[p]->snami);
        }
    } else if(ptable[p]->snami == NULL) {
        fprintf(Stderr, "SName=\"%s\"; ", ptable[p]->sname);
    } else {
        fprintf(Stderr, "SName={\"%s\", ", ptable[p]->sname);
        fprintf(Stderr, "\"%s\"}; ", ptable[p]->snami);
    }

    /* gauge */
    if(ptable[p]->gauge == NULL) {
        fprintf(Stderr, "Gauge=No; ");
    } else {
        fprintf(Stderr, "Gauge=\"%s\"; ", ptable[p]->gauge);
    }

    /* pselect */
    if(ptable[p]->pselect == NULL) {
        fprintf(Stderr, "PSelect=No; ");
    } else {
        fprintf(Stderr, "PSelect=\"%s\"; ", ptable[p]->pselect);
    }

    /* furry */
    fprintf(Stderr, "Furry=%d; ", ptable[p]->furry);

    fprintf(Stderr, "\n");

    /* nvlist and vlist */
    for(l = 0; l < 2; l++) {
        fprintf(Stderr, "  nvl[%d] = %d ", l, ptable[p]->nvlist[l]);
        pvldump(Stderr, ptable[p]->nvlist[l],
                        ptable[p]->vlist[l]);
    }
}


/*==================================================================*/
Public  Void    pvndump(v)
int     v;
/*--------------------------------------------------------------------
 *  Dump a vertex.
 *
 * changed to Public 1998/11/5 TK
 *------------------------------------------------------------------*/
{
    int    p, i, j;

    fprintf(Stderr, "%d ", v);

    /* list of particle */
    if(vtable[v]->composit == V_COMPOSIT) {
        fprintf(Stderr, "CVertex={%s", pvname(vtable[v]->auxp));
        if(vtable[v]->auxv == NULL) {
            fprintf(Stderr, ", NULL");
        } else {
            for(p = 0; p < 2; p++) {
                fprintf(Stderr, ", %s", vtable[vtable[v]->auxv[p]]->vid);
            }
        }
    } else {
        if(vtable[v]->vlegs == NULL) {
            fprintf(Stderr, "Vertex={NULL");
        } else {
            fprintf(Stderr, "Vertex={%s", pvname(vtable[v]->vlegs[0]));
            for(p = 1; p < vtable[v]->nvlegs; p++) {
                fprintf(Stderr, ", %s", pvname(vtable[v]->vlegs[p]));
            }
        }
    }
    fprintf(Stderr,"}; ");

    /* coupling constants */
    for(p = 0; p < nmordtbl; p++) {
        if(vtable[v]->morder[p] != 0) {
            fprintf(Stderr, "%s=%d; ", mordtbl[p],
                    vtable[v]->morder[p]);
        }
    }

    /* fortran name */
    for(j = 0; j < NFNAME; j++) {
        if(vtable[v]->fndimn[j] > 0) {
            if(j == 0) {
                fprintf(Stderr, "FName");
            } else {
                fprintf(Stderr, "CFName");
            }
            fprintf(Stderr, "=%s(", vtable[v]->fname[j]);
            for(i = 0; i < vtable[v]->fndimn[j]; i++) {
                if(i > 0) {
                    fprintf(Stderr, ",");
                }
                if(vtable[v]->fnind[j][i] > 0) {
                    fprintf(Stderr, "%d/%d",
                            vtable[v]->fnind[j][i], vtable[v]->fndim[j][i]);
                } else {
                    fprintf(Stderr, "%d", vtable[v]->fndim[j][i]);
                }
            }
            fprintf(Stderr, "); ");
        } else if(vtable[v]->fname[j] != NULL) {
            if(j == 0) {
                fprintf(Stderr, "FName");
            } else {
                fprintf(Stderr, "CFName");
            }
            fprintf(Stderr, "=%s; ", vtable[v]->fname[j]);
        }
    }

    /* type of fermion coupling */
    if(vtable[v]->ctype == V_V) {
        fprintf(Stderr, "FType=\"V\"; ");
    } else if(vtable[v]->ctype == V_A) {
        fprintf(Stderr, "FType=\"A\"; ");
    } else if(vtable[v]->ctype == V_VminusA) {
        fprintf(Stderr, "FType=\"V-A\"; ");
    } else if(vtable[v]->ctype == V_VplusA) {
        fprintf(Stderr, "FType=\"V+A\"; ");
    } else {
        fprintf(Stderr, "FType=\"---\"; ");
    }

    fprintf(Stderr, "\n");
    fprintf(Stderr, "%%vtype=%d; ",vtable[v]->vtype);
    if(vtable[v]->vid != NULL) {
        fprintf(Stderr, " Vid=%s; ", vtable[v]->vid);
    }
    if(vtable[v]->composit == V_NORMAL) {
        fprintf(Stderr, " composit=Normal;");
    } else if(vtable[v]->composit == V_AUX) {
        fprintf(Stderr, " composit=Aux;");
    } else if(vtable[v]->composit == V_COMPOSIT) {
        fprintf(Stderr, " composit=Aux;");
    } else {
        fprintf(Stderr, " composit=???;");
    }
    fprintf(Stderr, "\n");
}

/*==================================================================*/
Public  Void    pvdump(NoArg)
/*--------------------------------------------------------------------
    Dump tables.
--------------------------------------------------------------------*/
{
    int    p, v, n;
    ttitle *ttl;

    fprintf(Stderr, "----- pvdump ----- \n");

    fprintf(Stderr, "Model=\"%s\";\n\n", fnmdl);

    /* title */
    for(ttl = title; ttl != NULL; ttl = ttl->next) {
        printf("Titel=%s\n", ttl->tstr);
    }

    /* table of coupling constatns */
    fprintf(Stderr, "Order={%s", mordtbl[0]);
    for(v = 1; v < nmordtbl; v++) {
        fprintf(Stderr, ",%s", mordtbl[v]);
    }
    fprintf(Stderr, "};\n");

    /* table of set of coupling constatns */
    fprintf(Stderr, "OrdSet={");
    for(v = nmordtbl; v < nsordtbl; v++) {
        if(v != nmordtbl) {
            fprintf(Stderr, ",\n        ");
        }
        fprintf(Stderr, "%s[", sordtbl[v]);
        n = 0;
        for(p = 0; p < nmordtbl; p++) {
            if(sordset[v][p] != 0) {
                if(n != 0) {
                    fprintf(Stderr, ",");
                }
                fprintf(Stderr, "%s", mordtbl[p]);
                n++;
            }
        }
        fprintf(Stderr, "]");
    }
    fprintf(Stderr, "\n       };\n");

    fprintf(Stderr, "Version={%d", version[0]);
    for(v = 1; v < MAXVER; v++) {
        fprintf(Stderr, ",%d", version[v]);
    }
    fprintf(Stderr, "};\n");

    fprintf(Stderr, "PPhase=%d;\n", pphase);

    fprintf(Stderr, "Prefix=\"%s\";\n", mprefix);

    fprintf(Stderr, "%% maxtl = %d, mintl = %d\n", maxtl, mintl);

    compgl4 = pvcompgl4();
    fprintf(Stderr, "%% compgl4 = %d\n", compgl4);

    /* table of particles */
    for(p = 1; p < nptable; p++) {
        pvpdump(p);
    }
    for(v = 1; v < nvtable; v++) {
        pvndump(v);
    }

    /* end of the file */
    fprintf(Stderr,"---- end of pvdump ---\n");
}

/*==================================================================*/
Private Void    pvscale(ch)
int     ch;
/*--------------------------------------------------------------------
   print "%-------- ... \n" when ch = '-'
--------------------------------------------------------------------*/
{
    int i;

    putc('%', pvpf);
    for(i = 1; i < 40; i++) {
        putc(ch, pvpf);
    }
    putc(EOL, pvpf);
}

/*==================================================================*/
Public  Void    pvhelp(fp)
FILE    *fp;
/*--------------------------------------------------------------------
    Print help
--------------------------------------------------------------------*/
{
    int i, synlen;

    synlen = sizeof(syntax)/sizeof(syntax[0]);
    for(i = 0; i < synlen; i++) {
        fprintf(fp, "%s\n", syntax[i]);
    }
}

/*==================================================================*/
Private string  pvstrdup(conv, str)
bool    conv;
string  str;
/*--------------------------------------------------------------------
    duplicate string.
--------------------------------------------------------------------*/
{
    string s, p;

    s = (string) malloc((strlen(str)+1)*sizeof(char));
    if(s == NULL) {
        fprintf(Stderr, "\n*err* pvstrcup: memory allocation error\n");
        fprintf(Stderr, "for %s\n", str);
        exit(1);
    }

    /* convert to lower case characters */
    /* strcpy(s, str); */
    p = s;
    while(*str != '\0') {
        if(conv) {
            *p++ = tolower(*str++);
        } else {
            *p++ = *str++;
        }
    }
    *p = *str;


    return s;
}

/*==================================================================*/
Public  int     pvboson(p)
pindex  p;
/*--------------------------------------------------------------------
 * returns spin*2 of the particle
--------------------------------------------------------------------*/
{
    return (pmask[ptable[p]->ptype] & M_BOSON);
}

/*==================================================================*/
Public  int     pvdfermi(p)
pindex  p;
/*--------------------------------------------------------------------
 * returns spin*2 of the particle
--------------------------------------------------------------------*/
{
    return (pmask[ptable[p]->ptype] & M_DFERMION);
}

/*==================================================================*/
Public  int     pvfermi(p)
pindex  p;
/*--------------------------------------------------------------------
 * returns spin*2 of the particle
--------------------------------------------------------------------*/
{
    return (pmask[ptable[p]->ptype] & M_FERMION);
}

/*==================================================================*/
Public  int     pvsdeg(p)
pindex  p;
/*--------------------------------------------------------------------
 * returns spin*2 of the particle
--------------------------------------------------------------------*/
{
    bool m;
    m = ptable[p]->massless;

    switch (ptable[p]->ptype) {
    case P_SCALAR  : return 1;
    case P_MAJORANA: return 1;
    case P_DIRAC : if(m) {
                         return 1; 
                     } else {
                         return 2;
                     }
    case P_GHOST   : return 1;
    case P_VECTOR  : if(m) {
                         return 2; 
                     } else {
                         return 3;
                     }
    case P_AUX     : return 1;
    default   : return -1;
    }
}

/*==================================================================*/
Public  int     pvnumpowr(p)
pindex  p;
/*--------------------------------------------------------------------
 * returns the power of the numerator of the propagator
--------------------------------------------------------------------*/
{
    switch (ptable[abs(p)]->ptype) {
    case P_SCALAR  : return 0;
    case P_MAJORANA: return 1;
    case P_DIRAC :   return 1;
    case P_GHOST   : return 1;
    case P_VECTOR  : return 0;
    case P_AUX     : return 0;
    default   : return -1;
    }
}

/*==================================================================*/
Public  string   pvnptype(p)
pindex  p;
/*--------------------------------------------------------------------
 * returns the name of particle type
--------------------------------------------------------------------*/
{
    return pvtnam[ptable[p]->ptype];
}

/*==================================================================*/
Public  vindex   pvaddct(v, mord)
vindex  v;
int     mord[];
/*--------------------------------------------------------------------
 * add counter term
 * The vertex is the higher order of vertex "v" with coupling
 * constants "mord".
--------------------------------------------------------------------*/
{
    vindex vct;
    int    nl, to, l, lp, j;

    /* check the number of loops */
    nl = vtable[v]->nvlegs;
    to = 0;
    for(l = 0; l < nmordtbl; l++) {
        to += mord[l];
    }
    lp = to - nl + 2;
    if(lp%2 != 0 || lp < 1) {
        fprintf(Stderr, "\n*err*pvaddct : invalid counter term\n");
        fprintf(Stderr, "   v = %d, order = %d, nleg = %d\n", v, to, nl);
        prdump();
        exit(1);
    }
    lp /= 2;

    vct = pvnenter(nl, ((string*)NULL));

    vtable[vct]->vtype  = VTGENCT;
    vtable[vct]->nvlegs = nl;
    vtable[vct]->vlegs = vtable[v]->vlegs;

    for(l = 0; l < nmordtbl; l++) {
        vtable[vct]->morder[l] = mord[l];
    }
    vtable[vct]->loop     = lp;
    vtable[vct]->ctype    = vtable[v]->ctype;
    for(j = 0; j < NFNAME; j++) {
        vtable[vct]->fname[j]    = pvctname(vtable[v]->fname[j], mord, lp);
        vtable[vct]->fndimn[j]   = vtable[v]->fndimn[j];
        for(l = 0; l < vtable[v]->fndimn[j]; l++) {
            vtable[vct]->fndim[j][l] = vtable[v]->fndim[j][l];
            vtable[vct]->fnind[j][l] = vtable[v]->fnind[j][l];
        }
    }
    vtable[vct]->rname    = vtable[v]->rname;
    vtable[vct]->sname    = vtable[v]->sname;
    vtable[vct]->chanel   = vtable[v]->chanel;
    vtable[vct]->vselect  = vtable[v]->vselect;
    vtable[vct]->vid      = vtable[v]->vid;
    vtable[vct]->composit = vtable[v]->composit;
    vtable[vct]->auxp     = vtable[v]->auxp;
    vtable[vct]->auxv     = vtable[v]->auxv;

    return vct;
}

/*==================================================================*/
Public  vindex   pvaddctn(nlegs, legs, mord)
int     nlegs;
int     *legs;
int     mord[];
/*--------------------------------------------------------------------
 * add counter term
 * The vertex is the higher order of vertex "v" with coupling
 * constants "mord".
--------------------------------------------------------------------*/
{
    vindex vct;
    int    to, l, lp;

    /* check the number of loops */
    to = 0;
    for(l = 0; l < nmordtbl; l++) {
        to += mord[l];
    }
    lp = to - nlegs + 2;
    if(lp%2 != 0 || lp < 1) {
        fprintf(Stderr, "\n*err*pvaddct : invalid counter term\n");
        fprintf(Stderr, "   order = %d, nleg = %d\n", to, nlegs);
        prdump();
        exit(1);
    }
    lp /= 2;

    vct = pvnenter(nlegs, ((string*)NULL));

    vtable[vct]->vtype  = VTGENCT;
    vtable[vct]->nvlegs = nlegs;
    for(l = 0; l < nlegs; l++) {
        vtable[vct]->vlegs[l] = legs[l];
    }

    for(l = 0; l < nmordtbl; l++) {
        vtable[vct]->morder[l] = mord[l];
    }
    vtable[vct]->loop      = lp;
    vtable[vct]->ctype     = V_NOTYPE;
    vtable[vct]->composit  = V_NORMAL;
    vtable[vct]->fname[0]  = "gnrtd";
    vtable[vct]->fname[1]  = "cgnrtd";
    vtable[vct]->fname[2]  = "egnrtd";
    vtable[vct]->fname[3]  = "cegnrtd";
    for(l = 0; l < NFNAME; l++) {
        vtable[vct]->fndimn[l] = 0;
    }
    vtable[vct]->auxp      = 0;
    vtable[vct]->auxv       = NULL;

    return vct;
}

/*==================================================================*/
Private string   pvctname(name, mord, lp)
string  name;
int    *mord;
int     lp;
/*--------------------------------------------------------------------
 * construct name of counter term from tree vertex name.
 */
{
    static char nstr[512];
    string str;
#ifdef OLDNAME

    sprintf(nstr, "%s", name);
#else
    int    i;

    sprintf(nstr, "%s", name);

    str = nstr + strlen(nstr);
    for(i = 0; i < nmordtbl; i++) {
        sprintf(str, "%1d", mord[i]);
        str += strlen(str);
    }
#endif
    str = pvstrdup(False, nstr);

    return str;
}

/*==================================================================*/
Private bool    pvfermiv(nv)
int     nv;
{
    int i, tp;

    for(i = 0; i < vtable[nv]->nvlegs; i++) {
        tp = ptable[abs(vtable[nv]->vlegs[i])]->ptype;
        if(tp == P_DIRAC || tp == P_MAJORANA) {
            return True;
        }
    }
    return False;
}

/*==================================================================*/
Private bool    pvcompgl4(NoArg)
/*--------------------------------------------------------------------
 *
 */
{
    int v, i;

    for(v = nvstart; v < nvtable; v++) {
        if(vtable[v]->composit != V_COMPOSIT) {
            continue;
        }
        if(vtable[v]->nvlegs != 4) {
            continue;
        }
        for(i = 0; i < 4; i++) {
            if(ptable[abs(vtable[v]->vlegs[i])]->color != N8) {
                break;
            }
        }
        if(i >= 4) {
            return True;;
        }
    }
    return False;
}


/*==================================================================*/
Public  bool    pvcgl4(nv)
int     nv;
/*--------------------------------------------------------------------
 *
 */
{
    int i, v;

    v = ndvertex(nv);

    if(vtable[v]->composit != V_COMPOSIT) {
        return False;
    }

    if(ndlegs(nv) != 4) {
        return False;
    }

    for(i = 0; i < ndlegs(nv); i++) {
        if(ptable[abs(ndparticle(nv, i))]->color != N8) {
            return False;
        }
    }
    return True;
}

/*==================================================================*/
Public  int     pvgl3(nv)
int     nv;
/*--------------------------------------------------------------------
 *
 */
{
    int i;

    if(ndlegs(nv) != 3) {
        return 0;
    }

    for(i = 0; i < ndlegs(nv); i++) {
        if(ptable[abs(ndparticle(nv, i))]->color != N8) {
            return 0;
        }
    }
    return 1;
}

/*==================================================================*/
Public  int     pvgl4(nv)
int     nv;
/*--------------------------------------------------------------------
 *
 */
{
    int i, v;

    v = ndvertex(nv);

    if(vtable[v]->composit == V_COMPOSIT) {
        return 0;
    }

    if(ndlegs(nv) != 4) {
        return 0;
    }

    for(i = 0; i < ndlegs(nv); i++) {
        if(ptable[abs(ndparticle(nv, i))]->color != N8) {
            return 0;
        }
    }
    return 4;
}

#ifdef DEFSNAME
/*=================================================================*/
Private string pvrname(nv)
int     nv;
/*------------------------------------------------------------------*
 * Construct the name of vertex.
 *------------------------------------------------------------------*/
{
    static char str[128];
    static int  ptype[PTYPEMAX];

    string s = str;
    int  t, i, l = 0, lg, p;

    for(lg = 0; lg < PTYPEMAX; lg++) {
        ptype[lg] = 0;
    }
    for(lg = 0; lg < vtable[nv]->nvlegs; lg++) {
        p = ptable[abs(vtable[nv]->vlegs[lg])]->ptype;
        ptype[p]++;
    }

    for(t = 0; t < PTYPEMAX; t++) {
        for(i = 0; i < ptype[t]; i++) {
            *s++ = ptypesym[t];
            l++;
            if(l >= 128) {
                fprintf(Stderr, "*** pvrname : too long name\n");
                exit(1);
            }
        }
    }

    return pvstrdup(True, str);
}

/*=================================================================*/
Private string pvsname(nv)
int     nv;
/*------------------------------------------------------------------*
 * Construct the name of vertex.
 *------------------------------------------------------------------*/
{
    static char str[128];
    static int  ptype[PTYPEMAX];

    string s = str;
    int  t, i, l = 0, lg, p;

    for(lg = 0; lg < PTYPEMAX; lg++) {
        ptype[lg] = 0;
    }
    for(lg = 0; lg < vtable[nv]->nvlegs; lg++) {
        p = ptable[abs(vtable[nv]->vlegs[lg])]->ptype;
        ptype[p]++;
    }

    for(t = 0; t < PTYPEMAX; t++) {
        for(i = 0; i < ptype[t]; i++) {
            *s++ = ptypesym[t];
            l++;
            if(l >= 128) {
                fprintf(Stderr, "*** pvsname : too long name\n");
                exit(1);
            }
        }
    }

    return pvstrdup(True, str);
}

/*=================================================================*/
Private string pvchanel(nv)
int     nv;
/*------------------------------------------------------------------*
 * Construct the name of vertex.
 *------------------------------------------------------------------*/
{
    static char str[128];
    static int  ptype[PTYPEMAX];

    string s = str;
    int  t, i, l = 0, lg, p;

    for(lg = 0; lg < PTYPEMAX; lg++) {
        ptype[lg] = 0;
    }
    for(lg = 0; lg < vtable[nv]->nvlegs; lg++) {
        p = ptable[abs(vtable[nv]->vlegs[lg])]->ptype;
        ptype[p]++;
    }

    *s++ = 's';
    *s++ = 'm';
    for(t = 0; t < PTYPEMAX; t++) {
        for(i = 0; i < ptype[t]; i++) {
            *s++ = ptypesym[t];
            l++;
            if(l >= 128) {
                fprintf(Stderr, "*** pvchanel : too long name\n");
                exit(1);
            }
        }
    }

    return pvstrdup(True, str);
}
#endif

/*==================================================================*/
Private Void pvprcode(NoArg)
/*--------------------------------------------------------------------
 */
{
    printf("VT_SSS  = %o %d\n", VT_SSS, VT_SSS);
    printf("VT_SSV  = %o %d\n", VT_SSV, VT_SSV);
    printf("VT_SVV  = %o %d\n", VT_SVV, VT_SVV);
    printf("VT_VVV  = %o %d\n", VT_VVV, VT_VVV);
    printf("VT_SSA  = %o %d\n", VT_SSA, VT_SSA);
    printf("VT_SVA  = %o %d\n", VT_SVA, VT_SVA);
    printf("VT_VVA  = %o %d\n", VT_VVA, VT_VVA);
    printf("VT_SMM  = %o %d\n", VT_SMM, VT_SMM);
    printf("VT_SMD  = %o %d\n", VT_SMD, VT_SMD);
    printf("VT_SDM  = %o %d\n", VT_SDM, VT_SDM);
    printf("VT_SDD  = %o %d\n", VT_SDD, VT_SDD);
    printf("VT_SGG  = %o %d\n", VT_SGG, VT_SGG);
    printf("VT_MMV  = %o %d\n", VT_MMV, VT_MMV);
    printf("VT_MDV  = %o %d\n", VT_MDV, VT_MDV);
    printf("VT_DDV  = %o %d\n", VT_DDV, VT_DDV);
    printf("VT_GGV  = %o %d\n", VT_GGV, VT_GGV);
    printf("VT_SSSS = %o %d\n", VT_SSSS, VT_SSSS);
    printf("VT_SSVV = %o %d\n", VT_SSVV, VT_SSVV);
    printf("VT_VVVV = %o %d\n", VT_VVVV, VT_VVVV);
    printf("VT_SSGG = %o %d\n", VT_SSGG, VT_SSGG);
    printf("VT_GGVV = %o %d\n", VT_GGVV, VT_GGVV);
}

/*=================================================================*/
Public  Ulong  pvcode(nv)
int     nv;
/*------------------------------------------------------------------*
 * return the encoded type of the vertex
 *------------------------------------------------------------------*/
{
    Ulong vtype;
    int   lg;

    if(vtable[nv]->nvlegs > MAXTL || vtable[nv]->nvlegs < MINTL) {
        fprintf(Stderr, "pvcode: illegal vertex, nlegs = %d",
                vtable[nv]->nvlegs);
        fprintf(Stderr, "nv = %d\n", nv);
        pvndump(nv);
        exit(1);
    }
    vtype = 0;
    for(lg = 0; lg < vtable[nv]->nvlegs; lg++) {
        vtype = vtype*PTYPEMAX + ptable[abs(vtable[nv]->vlegs[lg])]->ptype;
    }
    return vtype;
}

/*=================================================================*/
Public  int    pvpowr(nv)
int     nv;
/*------------------------------------------------------------------*
 * Construct the power of the momentum of the vertex.
 *------------------------------------------------------------------*/
{
    Ulong  vtcode;

    vtcode = pvcode(nv);
    switch (vtcode) {
        case VT_SSS:
            return 0;

        case VT_SSV:
            return 1;

        case VT_SVV:
            return 0;

        case VT_VVV:
            return 1;

        case VT_SSA:
        case VT_SVA:
        case VT_VVA:
            return 0;

        case VT_SMM:
        case VT_SMD:
        case VT_SDM:
        case VT_SDD:
            return 0;

        case VT_SGG:
            return 0;

        case VT_MMV:
        case VT_MDV:
        case VT_DMV:
        case VT_DDV:
            return 0;

        case VT_GGV:
            return 1;

        case VT_SSSS:
            return 0;

        case VT_SSVV:
            return 0;

        case VT_VVVV:
            return 0;

        case VT_SSGG:
            return 0;

        case VT_GGVV:
            return 0;

        default:
            fprintf(Stderr, "pvpowr: illegal vertex nlegs = %d, ",
                    vtable[nv]->nvlegs);
            fprintf(Stderr, "nv = %d, vtcode = %ld\n", nv, vtcode);
            pvprcode();
            pvndump(nv);
            exit(1);
    }
}

/*==================================================================*/
Public  Void pvreord(nd)
int    nd;
/*--------------------------------------------------------------------
 * Reorder the legs of node "nd" in accordance with the definition
 * in the model specification.
 * We assume vertex is already determined.
 *
 * Argument:
 *     nd : node.
 */
{
    int    iv, ig, nl;
    pindex vp, np;
    vindex v;

    v  = ndvertex(nd);
    nl = ndlegs(nd);
    for(iv = 0; iv < nl-1; iv++) {
        vp = vtable[v]->vlegs[iv];
        np = ndparticle(nd, iv);
        if(vp == np) {
            continue;
        }

        /* find "vp" from the latter part of the list and swap */
        for(ig = iv+1; ig < nl; ig++) {
            if(vp == ndparticle(nd, ig)) {
                pvlswap(nd, ig, iv);
                break;
            }
        }

        /* "vp" is not found */
        if(ig >= nl) {
            fprintf(Stderr, "pvreord: graph %ld : ", cgraph->grid);
            fprintf(Stderr, "particle %d of ", vp);
            fprintf(Stderr, "node %d is not found ", nd);
            fprintf(Stderr, "in the vertex %d\n", v);
            exit(1);
        }
    }
}

/*==================================================================*/
Public  Void pvlswap(nd, lg1, lg2)
int    nd;
int    lg1, lg2;
/*--------------------------------------------------------------------
 * swap two legs "lg1" and "lg2" of the node "nd".
 *
 * Arguments:
 *     nd : node.
 *     lg1, lg2 : legs to be swaped.
 */
{
    int lgn, lgp, lgi;

    lgn = ndnextn(nd, lg1);
    lgp = ndparticle(nd, lg1);
    lgi = ndiline(nd, lg1);

    ndnextn(nd, lg1)    = ndnextn(nd, lg2);
    ndparticle(nd, lg1) = ndparticle(nd, lg2);
    ndiline(nd, lg1)    = ndiline(nd, lg2);

    ndnextn(nd, lg2)    = lgn;
    ndparticle(nd, lg2) = lgp;
    ndiline(nd, lg2)    = lgi;
}
