/*********************************************************************
 * $id$
 *
 *  File : genprc.c
 *
 *  Generate "in.prc"
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#define  DEFGLOBAL
#define  MAIN_GRC
#include "grc.h"
#include "config.h"
#include "token.h"
#include "args.h"
#include "conserv.h"
#include "pset.h"
#include "version.h"

/* debug
#define  GPRCDBG
*/

/* Entry points */
Public  int     main         ARG((int, char**));

/* Local functions */
Private int     grcarg        ARG((int, char**));
Private Void    gropth        ARG((string));
Private Void    prproc        ARG((int,int));

Private Void    grfproc       ARG((int));
Private Void    grfprc        ARG((int,int,int));

Private bool    prdup         ARG((int*));
Private int     prcmpint      ARG((Const Void *, Const Void *));
Private Void    prcstr        ARG((FILE*, int*));

/* Local variables */
#define pvyesno(x)      ((x)? "Yes" : "No")

Local   int     nprc = 0;
Local   char    finame[] = "in.prs";
Local   char    foname[] = "out.grf";
Local   int    *cpext;
Local   int     nsproc;
Local   bool    merge = True;

Local   int     ninitial = 0;

Local   int     *prctbl[MAXPROCS];
Local   int      prnext[MAXPROCS];
Local   int     *prcmp;
Local   int     nprocs = 0;

Local   int     deach   = 0;
Local   int     dhlp    = 0;
Local   bool    tishika = False;
Local   int     dti   = 0;
Local   bool    printprc;
Local   int     dprint  = 0;
Local   FILE *tifp = NULL;
#if DBG
Local   int     ddbg    = 0;
#endif

Local   topt    options[] = {
    {"e", ARFLAG, (int*) &deach},
    {"h", ARFLAG, (int*) &dhlp},
    {"?", ARFLAG, (int*) &dhlp},
    {"t", ARFLAG, (int*) &dti},
    {"n", ARFLAG, (int*) &dprint},
#if DBG
    {"d", ARINT,  (int*) &ddbg},
#endif
};

Local int noptions = sizeof(options)/sizeof(topt);
#define YorN(x)   ((x) ? "Yes": "No")

/*==================================================================*/
Public  int    main(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
    Main routine for graph drawin
--------------------------------------------------------------------*/
{
    grcmsg("Feynman graph generator");          /* First message */

    argc = grcarg(argc, argv);   /* analyze arguments */
    if(argc != 1) {
        gropth(argv[0]);
        exit(1);
    }

#if DBG
    /* debugging flags for modules.
       if(dbg >= xxdbg) then enable debugging
    */
    gcdbg = 1;  /* debug flag for grc.c */
    pvdbg = 2;  /* debug flag for rdptbl.c */
    rddbg = 2;  /* debug flag for token.c */
    rgdbg = 2;  /* debug flag for rdgraph.c */
    rpdbg = 2;  /* debug flag for rdproces.c */
    sfdbg = 2;  /* debug flag for sfgrf.c */
#endif

    /* read process */
    if(!rpinit(fnprc, True)) {
        exit(16);
    }

    conserv();

    if(tishika) {
        if((tifp = fopen("tishika.h", "w")) == NULL) {
            fprintf(Stderr, "Cannot open \"tishika.h\"\n");
            exit(FAIL);
        }
    }
#if DBG
    if(dbgst > 0) {
        pvprint("out.mdl");  /* normalized model file */
    }
    if(dbg >= gcdbg)
        pvdump();  /* normalized model file */

#endif

    prinit();           /* initialize print routine */

    /* generate graphs for processes */
    /* nprocess = 0; */

    /* multi processes */
    while((nprc= rpread()) > 0) {
        nprocess++;
        sfnum   = 0;
        sfden   = 1;

        prfirst();        /* print header */

        grfproc(nprc);

        prlast();         /* end of printing */
    }

    if(tishika) {
        fclose(tifp);
    }
    return 0;
}

/*==================================================================*/
Private int    grcarg(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
*/
{
    int    pos;
    string cmd;

    cmd     = argv[0];
    skipasp = False;
    fngrf   = NULL;
    tishika = False;
    printprc = True;

#if DBG
    ddbg    = 0;
#endif

    argc = aropt(argc, argv, noptions, options);

    if(dhlp != 0) {
        gropth(cmd);
        exit(0);
    }

    if(fngrf == NULL)
        fngrf = foname;

    switch (deach) {
        case ARPLUS:
            merge = False;
            break;
        case ARMINUS:
            merge = False;
            break;
    }

    switch (dti) {
        case ARPLUS:
            tishika = True;
            break;
        case ARMINUS:
            tishika = True;
            break;
    }

    switch (dprint) {
        case ARPLUS:
            printprc = False;
            break;
        case ARMINUS:
            printprc = False;
            break;
    }

    /* input process file name */
    pos = 1;
    if((fnprc = arnext(argc, argv, &pos)) != NULL) {
        pos--;
        argc = ardel(argc, argv, pos, 1);
    } else {
        fnprc = finame;
    }

#if DBG
    /* debugging flags */
    dbg   = 0;
    dbgst = -1;

    if(ddbg > 0) {
        dbg   = ddbg;
    }
    if(ddbg < 0) {
        dbgst = -ddbg;
    }
#endif

    return argc;
}

/*==================================================================*/
Private Void   gropth(cmd)
string  cmd;
/*--------------------------------------------------------------------
*/
{
    printf("Syntax %s [options] [Input_File_name]\n", cmd);
    printf("\nOption\n");
    printf("    -e ");
    printf(": output into separate files\n");
    printf("    -n ");
    printf(": No output process file\n");
    printf("    -t ");
    printf(": output Information file\n");
#if DBG
    printf("    -p int              ");
    printf(": print level\n");
    printf("    -d int              ");
    printf(": debug level\n");
#endif
    printf("    -h                  ");
    printf(": print this message\n");
    printf("\nDefault file names:\n");
    printf("     Input_File_name = %s\n", finame);
}

/*==================================================================*/
Private Void    grfproc(proc)
int proc;
/*--------------------------------------------------------------------
 * Construct all possible process
 *------------------------------------------------------------------*/
{
    int i;

    if((cpext = (int *) malloc(sizeof(int)*nextern)) == NULL) {
        fprintf(Stderr, "*** grfproc: memory allocation error\n");
        exit(1);
    }
    if((prcmp = (int *) malloc(sizeof(int)*nextern)) == NULL) {
        fprintf(Stderr, "*** grfproc: memory allocation error\n");
        exit(1);
    }

    ninitial = 0;
    for(i = 0; i < nextern; i++) {
        cpext[i] = ndname(i);
        if(ndiorf(i) == ELINITIAL) {
            if(i != ninitial) {
                fprintf(Stderr, "*** grfproc: initial particles ");
                fprintf(Stderr, "should be defined before final ones\n");
                fprintf(Stderr, "\ti=%d, ninitial=%d\n", i, ninitial);
                exit(1);
            }
            ninitial++;
        }
    }

    nsproc = 0;

    grfprc(proc, 0, -npptable);

    free(prcmp);
    free(cpext);
}

/*==================================================================*/
Private Void    grfprc(proc, start, sp)
int     proc;
int     start;   /* the external particle number to be assigned */
int     sp;      /* last particle */
/*--------------------------------------------------------------------
 * Construct all possible process
 *------------------------------------------------------------------*/
{
    int i, p, pp, porg;

#ifdef GPRCDBG
#endif
    /* initial particles */
    for(i = start; i < nextern; i++) {
        porg = abs(cpext[i]);
        if(ptable[porg]->ptype == P_SET) {
            for(p = - npptable + 1; p < npptable; p++) {
                if(psget(porg, p)) {
                    if(ndiorf(i) == ELINITIAL) {
                        pp = p * ptable[abs(p)]->antip;
                        cpext[i] = pp;
                    } else {
                        cpext[i] = p;
                    }
                    grfprc(proc, i+1, p);
                }
            }
            cpext[i] = porg;
            return;
        }
    }

#ifdef GPRCDBG
    printf("grfprc(%d) (nextern=%d):test:    ", start, nextern);
    for(i = 0; i < nextern; i++) {
        printf("%s ", pvname(cpext[i]));
    }
    printf("\n");
#endif

    if(constest(nextern, cpext)) {
        if(prdup(cpext)) {

            nsproc++;
            if(printprc) {
                prproc(proc, nsproc);
            }
        }
    }
}

/*==================================================================*/
Private int    prcmpint(pi0, pi1)
Const Void *pi0;
Const Void *pi1;
/*--------------------------------------------------------------------
 */
{
    return *((int*)pi0) - *((int*)pi1);
}

/*==================================================================*/
Private bool   prdup(pext)
int     *pext;
/*--------------------------------------------------------------------
 *  Check duplication of processes
 */
{
    int i, j;
#ifdef GPRCDBG
    int k;
#endif

    for(j = 0; j < nextern; j++) {
        prcmp[j] = pext[j];
    }
    qsort((void*)prcmp, ninitial, sizeof(int), prcmpint);
    qsort((void*)(prcmp+ninitial), nextern-ninitial, sizeof(int), prcmpint);

    for(i = 0; i < nprocs; i++) {
        if(prnext[i] != nextern) {
            continue;
        }
        for(j = 0; j < nextern; j++) {
            if(prctbl[i][j] != prcmp[j]) {
                break;
            }
        }
        if(j >= nextern) {
#ifdef GPRCDBG
            printf(" %5d-", i+1);
            prcstr(stdout, pext);
            printf("prdup : False : nprocs = %d : same as %d\n", nprocs, i+1);

            for(k = 0; k < nprocs; k++) {
                printf("prctbl[%2d] = ", k);
                for(j = 0; j < nextern; j++) {
                    printf(" %4d", prctbl[k][j]);
                }
                printf("\n");
            }
            printf("prcmp      = ");
            for(j = 0; j < nextern; j++) {
                printf(" %4d", prcmp[j]);
            }
            printf("\n");
#endif
            return False;
        }
    }
    if(nprocs >= MAXPROCS) {
        fprintf(Stderr, "*** prdup : Too many processes\n");
        exit(FAIL);
    }
    if((prctbl[nprocs] = (int *) malloc(sizeof(int)*nextern)) == NULL) {
        fprintf(Stderr, "*** prdup : Memory\n");
        exit(FAIL);
    }
    prnext[i] = nextern;
    for(j = 0; j < nextern; j++) {
        prctbl[nprocs][j] = prcmp[j];
    }
    nprocs++;
#ifdef GPRCDBG
    printf("prdup : Ture  : nprocs = %d\n", nprocs);
#endif

    return True;
}

/*==================================================================*/
Public  Void   prinit(NoArg)
/*--------------------------------------------------------------------
 *  Print initial message
 */
{
    printf("\n");
    printf("    Model  file name = \"%s\"\n", fnmdl);
    printf("    Input  file name = \"%s\"\n", fnprc);
}

/*==================================================================*/
Public  Void   prfirst(NoArg)
/*--------------------------------------------------------------------
 *  Print the first message.
 */
{
    node n;

    printf("*************************************************\n");
    printf("Order = %d (", order);
    for(n = 0; n < nsordtbl; n++) {
        if(n > 0)
            printf(", ");
        printf("%s = %d", sordtbl[n], sorder[n]);
    }
    printf(") loop = %d\n", loop);
    printf("\n%4d external lines\n\n", nextern);

    for(n = 0; n < nextern; n++) {
        printf("%4d %s %s\n",
                tovert(n), eliorf(ndiorf(n)),
                pvename(ndiorf(n), ndname(n)));
    }
    printf("=================================================\n");
}

/*==================================================================*/
Public  Void   prlast(NoArg)
/*--------------------------------------------------------------------
 *  Print the last message.
 *------------------------------------------------------------------*/
{
    printf("*************************************************\n");
}

/*==================================================================*/
Public  Void    prdump(NoArg)
/*--------------------------------------------------------------------
 */
{
}

/*==================================================================*/
Private Void    prcstr(fp, cpext)
FILE    *fp;
int     *cpext;
/*--------------------------------------------------------------------
 */
{
    int i;
    int k = 0;

    for(i = 0; i < nextern; i++) {
        if(ndiorf(i) == ELINITIAL) {
            fprintf(fp, " ");
            if(ptable[abs(cpext[i])]->antip >= 0) {
                fprintf(fp, "%s", pvname(cpext[i]));
            } else {
                fprintf(fp, "%s", pvname(-cpext[i]));
            }
        }
    }
    fprintf(fp, " -->");
    if(tishika) {
        k = 0;
        fprintf(tifp, "{");
    }
    for(i = 0; i < nextern; i++) {
        if(tishika) {
            fprintf(fp, " ");
            fprintf(fp, "%s", pvname(cpext[i]));
            fprintf(tifp,"%c\"%s\"",(k==0 ?' ':','),pvname(cpext[i]));
            k++;
        } else {
            if(ndiorf(i) == ELFINAL) {
                fprintf(fp, " ");
                fprintf(fp, "%s", pvname(cpext[i]));
            }
        }
    }
    fprintf(fp, "\n");
    if(tishika) {
        fprintf(tifp, "},\n");
    }
}

/*==================================================================*/
Private Void    prproc(proc, sub)
int proc;
int sub;
/*--------------------------------------------------------------------
    Print the processes.
--------------------------------------------------------------------*/
{
    int i, j, op, aop;
    static char str[256];
    static FILE   *fp;
    bool pr;

    if(merge) {
        sprintf(str, "in.prc");
    } else {
        sprintf(str, "p%ds%d.prc", proc, sub);
    }

    if(!merge || nprocs <= 1) {
        if((fp = fopen(str, "w")) == NULL) {
            fprintf(Stderr, "\n*err* prproc: open error to write ");
            fprintf(Stderr,"filename = %s\n", str);
            exit(1);
        }
    }

    if(merge) {
        printf("%6d:", nprocs);
    } else {
        printf("\"%s\":", str);
    }
    prcstr(stdout, cpext);

    /* name of model file */
    if(!merge || nprocs <= 1) {
        fprintf(fp, "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
        fprintf(fp, "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
        fprintf(fp, "Model=\"%s\";\n", fnmdl);
    }

    /* process number */
    fprintf(fp, "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
    fprintf(fp, "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
    fprintf(fp, "Process; %% %d(%d.%d)\n", nprocs, proc, sub);

    /* order of coupling constants */
    for(i = 0; i < nsordtbl; i++)
        fprintf(fp, "  %s=%d; ", sordtbl[i], sorder[i]);
    fprintf(fp,"\n");

    /* initial particles */
    j = 0;
    fprintf(fp, "  Initial={");
    for(i = 0; i < nextern; i++) {
        if(ndiorf(i) == ELINITIAL) {
            if(j > 0)
                fprintf(fp, ", ");
            j++;
            if(ptable[abs(cpext[i])]->antip >= 0) {
                fprintf(fp, "%s", pvname(cpext[i]));
            } else {
                fprintf(fp, "%s", pvname(-cpext[i]));
            }
            op = ndname(i);
            if(cpext[i] != op) {
                aop = abs(op);
                pr = False;
                if(ptable[aop]->pcode != 0) {
                    fprintf(fp, "[");
                    if(op < 0) {
                        fprintf(fp, "PCode=%d", -ptable[aop]->pcode);
                    } else {
                        fprintf(fp, "PCode=%d", ptable[aop]->pcode);
                    }
                    pr = True;
                }
                if(ptable[aop]->kfcode != 0) {
                    if(pr) {
                        fprintf(fp, ";");
                    } else {
                        fprintf(fp, "[");
                    }
                     
                    if(op < 0) {
                        fprintf(fp, "KFCode=%d", -ptable[aop]->kfcode);
                    } else {
                        fprintf(fp, "KFCode=%d", ptable[aop]->kfcode);
                    }
                    pr = True;
                }
                if(pr) {
                    fprintf(fp, "]");
                }
            }
        }
    }
    fprintf(fp,"};\n");

    /* final particles */
    j = 0;
    fprintf(fp, "  Final  ={");
    for(i = 0; i < nextern; i++) {
        if(ndiorf(i) == ELFINAL) {
            if(j > 0)
                fprintf(fp, ", ");
            j++;
            fprintf(fp, "%s", pvname(cpext[i]));
        }
    }
    fprintf(fp,"};\n");

    /* flags */
    fprintf(fp,"  Expand=%s; Opi=%s; Block=%s; Gm2=%s;\n",
            YorN(expand),
            YorN(onepi),
            YorN(block),
            YorN(gm2));
    fprintf(fp,"  Tadpole=%s; Extself=%s; Hep=%s; Selfe=%s;\n",
            YorN(tadpole),
            mdlbools[extself],
            YorN(hep),
            YorN(selfe));
    fprintf(fp,"  Countert=%s;\n",
            YorN(countert));
    fprintf(fp,"  AnyCT=%s;\n",
            YorN(anyct));
    if(kinem != NULL)
        fprintf(fp,"  kinem=\"%s\";\n", kinem);
    if(prefix != NULL)
        fprintf(fp,"  prefix=\"%s%d\";\n", prefix, sub);

    /* end */
    fprintf(fp, "Pend;\n\n");
    if(!merge) {
        fclose(fp);
    }
}
