/*********************************************************************
 * $id$
 *
 *  File : grc.c
 *
 *  Entry for graph generation
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#define  DEFGLOBAL
#define  MAIN_GRC
#include "grc.h"
#include "config.h"
#include "token.h"
#include "pset.h"
#include "args.h"
#include "version.h"

/*--------------------------------------------------------------------
 * Entry points
 */
Public  int     main         ARG((int, char**));

/*--------------------------------------------------------------------
 * Local functions
 */
Private int     grcarg       ARG((int, char**));
Private Void    gropth       ARG((string));

/*--------------------------------------------------------------------
 * Local variables
 */
Local   int     nprc = 0;
Local   char    finame[] = "in.prc";
Local   char    foname[] = "out.grf";

Local   int     dskasp  = 0;
Local   int     doutf   = 0;
Local   int     dtype   = 0;
Local   int     dtone   = 0;
Local   int     dgone   = 0;
Local   int     dhlp    = 0;
#if DBG
Local   int     dprt    = 0;
Local   int     ddbg    = 0;
Local   int     dmon    = 0;
Local   int     dtrc    = 0;
Local   int     dsel    = -1;
Local   int     dcpt    = 0;
Local   int     dbit    = -1;
Local   int     dogt    = 0;
#endif

Local   topt    options[] = {
    {"o", ARSTR,  (int*) &fngrf},
    {"a", ARFLAG, (int*) &dskasp},
    {"c", ARFLAG, (int*) &doutf},
    {"g", ARINT,  (int*) &dgone},
    {"t", ARINT,  (int*) &dtype},
    {"T", ARFLAG, (int*) &dtone},
    {"h", ARFLAG, (int*) &dhlp},
    {"?", ARFLAG, (int*) &dhlp},
#if DBG
    {"p", ARINT,  (int*) &dprt},
    {"d", ARINT,  (int*) &ddbg},
    {"m", ARFLAG, (int*) &dmon},
    {"r", ARFLAG, (int*) &dtrc},
    {"s", ARINT,  (int*) &dsel},
    {"i", ARINT,  (int*) &dcpt},
    {"b", ARINT,  (int*) &dbit},
    {"z", ARFLAG, (int*) &dogt},
#endif
};

Local int noptions = sizeof(options)/sizeof(topt);

/*==================================================================*/
Public  int    main(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
    Main routine for graph drawin
--------------------------------------------------------------------*/
{
#if DBG
    int np;
#endif

    grcmsg("Feynman graph generator");          /* First message */

    argc = grcarg(argc, argv);   /* analyze arguments */
    if(argc != 1) {
        gropth(argv[0]);
        exit(1);
    }

#if DBG
    /* debugging flags for modules.
       if(dbg >= xxdbg) then enable debugging
    */
    aldbg = 3;  /* debug flag for alist.c */
    axdbg = 1;  /* debug flag for axconn.c */
    bidbg = 1;  /* debug flag for biconn.c */
    eldbg = 3;  /* debug flag for elem.c */
    fldbg = 3;  /* debug flag for fline.c */
    fodbg = 3;  /* debug flag for out.c */
    gcdbg = 3;  /* debug flag for grc.c */
    gndbg = 3;  /* debug flag for gnode.c */
    gsdbg = 3;  /* debug flag for scalar.c */
    ildbg = 3;  /* debug flag for iline.c */
    lcdbg = 3;  /* debug flag for lclass.c */
    padbg = 3;  /* debug flag for assign.c */
    pmdbg = 3;  /* debug flag for matchv.c */
    prdbg = 3;  /* debug flag for print.c */
    pvdbg = 3;  /* debug flag for rdptbl.c */
    rddbg = 3;  /* debug flag for token.c */
    rgdbg = 3;  /* debug flag for rdgraph.c */
    rpdbg = 3;  /* debug flag for rdproces.c */
    sfdbg = 3;  /* debug flag for sfgrf.c */
#endif

    /* read process */
    if(!rpinit(fnprc, False)) {
        exit(16);
    }
#if DBG
    if(dbit >= 0) {
        for(np = 0; np < rproc; np++) {
            rpextn[np]->countert = (dbit & 0x00001);
            rpextn[np]->selfe    = (dbit & 0x00002) >> 1;
            rpextn[np]->extself  = (dbit & 0x00004) >> 2;
            rpextn[np]->tadpole  = (dbit & 0x00008) >> 3;
            rpextn[np]->onepi    = (dbit & 0x00010) >> 4;
            rpextn[np]->expand   = (dbit & 0x00020) >> 5;
            rpextn[np]->block    = (dbit & 0x00040) >> 6;
            rpextn[np]->anyct    = (dbit & 0x00080) >> 7;
        }
    }
#endif

    /* open output file of graph structure */
    if(outf) {
        Discard foinit(fngrf);
    }

#if DBG
    if(dbgst > 0) {
        pvprint("out.mdl");  /* normalized model file */
        rprint("out.prc");  /* normalized process file */
        if(!outf)
            Discard foinit(fngrf);   /* open output file of graph structure */
    }
    if(dbg >= gcdbg)
        pvdump();  /* normalized model file */

#endif

    prinit();           /* initialize print routine */

    /* generate graphs for processes */
    nprocess = 0;

    if(outf) {
        prgrsp = fopen("out.grs", "w");
    } else {
        prgrsp = NULL;
    }

    /* multi processes */
    while((nprc = rpread()) > 0) {

        if(undefp) {
            fprintf(Stderr, "*** Undefined particles are included ");
            fprintf(Stderr, "in the external particles\n");
            exit(1);
        }
#if DBG
        if(dcpt != 0)
            cptest = dcpt;
#endif
        nprocess++;
        duptest = True;
        sfnum   = 0;
        sfden   = 1;

        nopitbl = 0;

#if DBG
        if(dbg >= gcdbg) {
            rprin1(stdout, nprc-1, True);
        }
#endif

        prfirst();        /* print header */
        if(outf) {
            foproces(rpropt);
        }

        prtype(-1);       /* initialize type counter */

        bfginit();

        if(!osinit(sorder)) {
            osmessage();
            exit(FAIL);
        }

#if DBG
        if(dbg >= gcdbg)
            padump();
#endif

        gvertex();        /* gnerate graph */

        prtype(0);        /* print the last type */

        if(outf)          /* print ending message */
            fopend();     /* end of graph structure output */
        if(!outf)
            fofpend2();   /* output graph number */

#if DEBG
        /* Print classification of graphs */
        bfgend();
#endif

        prlast();         /* end of printing */

#if DBG
        if(dbit >= 0) {
            printf("DBit = %4d[%d:%d:%d:%d:%d:%d:%d:%d], ",
                    dbit, countert, selfe, extself, tadpole,
                    onepi, expand, block, anyct);
            printf("Ngraph = %8ld, Ntype = %6d\n", nconfig, gntype);
        }
#endif
    }
    if(outf) {
        foclose();        /* close file of graph structure */
    }

    if(prgrsp != NULL) {
        fclose(prgrsp);
    }
    return 0;
}

/*==================================================================*/
Private int    grcarg(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
*/
{
    int    pos;
    string cmd;

    cmd     = argv[0];
    skipasp = False;
    outf    = True;
    fngrf   = NULL;
    display = False;

#if DBG
    dispno  = -1;
    ddbg    = 0;
    dprt    = 0;
    trace   = False;
    mon     = False;
    onegrtype = False;
#endif

    argc = aropt(argc, argv, noptions, options);

    if(dhlp != 0) {
        gropth(cmd);
        exit(0);
    }

    if(fngrf == NULL)
        fngrf = foname;

    if(dskasp != 0) {
        skipasp = True;
    }

    switch (doutf) {
        case ARPLUS:
            outf = True;
            break;
        case ARMINUS:
            outf = False;
            break;
    }

    seltype = dtype;

    switch (dtone) {
        case ARPLUS:
            onetype = True;
            break;
        case ARMINUS:
            onetype = True;
            break;
    }

    onegraph = dgone;

    /* input process file name */
    pos = 1;
    if((fnprc = arnext(argc, argv, &pos)) != NULL) {
        pos--;
        argc = ardel(argc, argv, pos, 1);
    } else {
        fnprc = finame;
    }

#if DBG
    /* debugging flags */
    dbg   = 0;
    dbgst = -1;
    selprop  = 0;

    if(ddbg > 0) {
        dbg   = ddbg;
    }
    if(ddbg < 0) {
        dbgst = -ddbg;
    }

    if(dprt > 0) {
        display = dprt;
        dispno  = -1;
    } else {
        display = False;
        dispno  = -dprt;
    }

    if(dtrc != 0) {
        trace = True;
    }

    if(dogt != 0) {
        onegrtype = True;
    }

    if(dmon != 0) {
        mon = True;
    }

    if(dsel > 0) {
        selprop = dsel;
    } else {
        selprop = 0;
    }
#endif

    return argc;
}

/*==================================================================*/
Private Void   gropth(cmd)
string  cmd;
/*--------------------------------------------------------------------
*/
{
    printf("Comman syntax\n");
    printf("\n");
    printf("  %s [options] [Input_File_name]\n", cmd);
    printf("\n");
    printf("Option\n");
    printf("\t-a                  ");
    printf(": skip particle assignment\n");
    printf("\t-l int              ");
    printf(": with detailed message \n");
    printf("\t-o Output_File_name ");
    printf(": specify output file\n");
    printf("\t-t int              ");
    printf(": output for specified type\n");
    printf("\t-T                  ");
    printf(": output one graph from each type\n");
    printf("\t-c                  ");
    printf(": only count\n");
#if DBG
    printf("\t-p int              ");
    printf(": print level\n");
    printf("\t-d int              ");
    printf(": debug level\n");
    printf("\t-m                  ");
    printf(": enable monitor\n");
    printf("\t-i int              ");
    printf(": cptest flag\n");
    printf("\t-r                  ");
    printf(": enable trace\n");
#endif
    printf("\t-h                  ");
    printf(": print this message\n");
    printf("\nDefault file names:\n");
    printf("\t Input_File_name = %s\n", finame);
    printf("\tOutput_File_name = %s\n", foname);

    printf("\n\n");
    rphelp(stdout);
}
