/*********************************************************************
 * $id$
 *
 *  File :  grc.h
 *
 *  Global variables.
 *
 *                                               by Kaneko Toshiaki
 */

#include "common.h"
#include "dbg.h"
#include "grdata.h"

/*===================================================================
 *   File names
 */
Global char    *fnprc;  /* process */
Global char    *fngrf;  /* graph */

/*===================================================================
 *   Particle table
 */
#include "ptable.h"

/*-------------------------------------------------------------------
 *  Table of process
 */

typedef struct  {
        int     nextern;
        int     particle[MAXELINE];
        int     iorf[MAXELINE];
        int     sorder[MAXSORD];
        int     spcode[MAXELINE];      /* pcode of the set */
        int     skfcode[MAXELINE];     /* kfcode of the set */

        bool    expand;
        int     expmin;
        int     expmax;
        bool    onepi;
        bool    block;
        bool    tadpole;
        bool    selfe;     /* exclude self-energy part */
        int     extself;

        bool    shrinkaux;  /* shrink propagator of AUX. in graph selection */
        bool    countert;   /* include counter term */
        bool    anyct;      /* include not defined counter term */
        bool    hep;
        bool    gm2;
        bool    undefp;    /* include undefined particle */
        bool    skipasp;
        string  kinem;
        int     cptest;    /* test by orderly algorithm only */
        string  pprefix;   /* prefix of the generated code */

    }   rpextt;

Global  rpextt  *rpextn[MAXPROCS]; /* the table of processes */

Global  int     rproc;          /* the total number of processes */

/*-------------------------------------------------------------------
 *  The parameters of the current graph
 */

Global  int     nextern;    /* The number of external particles.  */
Global  int     nnodes;     /* the number of nodes */
Global  int     loop;       /* The number of loops.  */
Global  int     order;      /* The order of coupling constants.  */
Global  int     rprocc;     /* The current processes in process tbl. */
Global  long    nconfig;    /* The number of generated configurations.*/
Global  long    ndump;      /* The number of dumps */
Global  int     graphn;     /* The current graph number in the file */
Global  int     nlgtbl;     /* The number of legs */
Global  int     nilines;    /* The number of internal lines */
Global  int     morder[MAXMORD];   /*   The table of order */
Global  int     sorder[MAXSORD];   /*   The table of order */
Global  bool    duptest;    /* Test duplication */
Global  int     cptest;     /* use cpdtest instead of lcftest */
Global  long    sfnum;      /* numerator of weighted sum of graphs */
Global  long    sfden;      /* denominator of w. sum of graphs */

/*-------------------------------------------------------------------
 *  Table of nodes.
 */

/*  MAXNODES**MAXTLEGS can be represented by "long" type data.
 *  This condition is used in "elmult" and array "lcdeg".
 */
/* type of node index */

/* value of tnode afford (MAXTL+1)**(MAXLTL) of connection.
 * as "elmult" codes is connection in (MAXTL+1)**multiplicity.
 * In the non-expand mode, multiplicity = 1 and #legs may > MAXTL+1.
 * In the expand node, multiplicity <= MAXTL
 */

/* value of tnode[NDTYPE] */
#define NDEXTERN     0        /* external line */
#define NDTREEV      1        /* tree vertex */
#define NDLOOPEDV    2        /* looped vertex */
#define NDNULL    9999        /* null node index = nptable */

/*-------------------------------------------------------------------
 * Table of class of nodes.
 */
Global  int     clsstbl[MAXNODES];
#define ndclass(n)      clsstbl[n]

/*-------------------------------------------------------------------
 * table of class and degree 
 */
typedef long   tdnode[MAXNODES];   /* table of nodes */
Global  tdnode *nodetbl[MAXNODES];
#define nddeg(n, c)     (*(nodetbl[n]))[c]

/*-------------------------------------------------------------------
 *  Table of node level.
 */
Global  int     ndlvl[MAXNODES];
#define ndlevel(n)      ndlvl[n]
#define NDINF           MAXNODES + 1    /* inifinite level */

/*-------------------------------------------------------------------
 *  Table of # free legs 
 */
Global  int     ndfrl[MAXNODES];
#define ndfree(n)       ndfrl[n]

/*-------------------------------------------------------------------
 *  Table of connections.
 */
Global int      edgetbl[MAXNODES][MAXNODES];

/*-------------------------------------------------------------------
 *  Table of l-classes.
 */
#define LCSTACK  MAXNODES*MAXNODES  /* L-class stack size */

/* type of l-class index */
typedef int lclass;

/* l-class
 *  class table
 *    lcpton[lcctop[c] .. lcctop[c+1]-1] : list of nodes in
 *                                          the class "c".
 */
Global  int     nlclass;            /* the number of l-classes */
Global  int     lcctop[MAXNODES+1];         /* class to pointer */
Global  node    lcpton[MAXNODES];           /* pointer to node */
Global  int     lcstack[LCSTACK];           /* stack of l-class */
/*-------------------------------------------------------------------
 *  Table of particle assignment.
 */
/* type of save area index of particle assignment
 */
typedef int alist;

/* type of particle assignment table
 */
typedef struct {
        int     nundef;         /* # undefined particle */
        alist   asave;          /* save area of old assignment */
}       tpntbl;

/* table for particle assignment to node
 */
Global  tpntbl  *pntbl[MAXNODES];
#define ndundef(n)      pntbl[n]->nundef
#define ndasave(n)      pntbl[n]->asave

/* particle of external line. particle/unti-particle
 * is defined as incoming particle.
 * for outgoing particle multiply "ptable[abs(ndiorf(e))]->antip"
 */

/*-------------------------------------------------------------------
 *  Table of a-list used in particle assignment.
 */
/* Save area of particle assignment
 */
#define LALIST  100

/* type of save area of particle assignment
 */
typedef struct {
        int     level;
        int     aleg;
        alist   next;
}       talist;
#define allevel(l)      alcell[l].level
#define alleg(l)        alcell[l].aleg
#define alnext(l)       alcell[l].next

/* Save area for particle assignment
 */
Global  alist   alcfree;
Global  talist  alcell[LALIST];
#define ALNULL  LALIST
 
/*-------------------------------------------------------------------
 * Work space.
 */
Global  int     ndwork[MAXNODES];       /* work space */
Global  int     ndwork2[MAXNODES];       /* work space */

/*-------------------------------------------------------------------
 *  Flags
 */
/* Flags for conditions of generated graphs
 */
/* glogal flag */
Global  bool    shrinkaux;      /* shrink AUX propagator in graph selection */

Global  bool    expand;         /* expand looped parts  */
Global  int     expmin;         /* expand looped parts  */
Global  int     expmax;         /* expand looped parts  */
Global  bool    tadpole;        /* include tadpoles */
Global  int     extself;        /* self-eng. part at e-line */
Global  bool    onepi;          /* one paritcle irreducibility */
Global  bool    block;          /* block */
Global  bool    selfe;          /* self-entery part */

/* flags for scalar graphs */
Global  bool    sexpand;         /* expand looped parts  */
Global  int     sexpmin;         /* expand looped parts  */
Global  int     sexpmax;         /* expand looped parts  */
Global  bool    stadpole;        /* include tadpoles */
Global  int     sextself;        /* self-eng. part at e-line */
Global  bool    sonepi;          /* one paritcle irreducibility */
Global  bool    sblock;          /* block */
Global  bool    sselfe;          /* self-entery part */

/* flags for graphs with particle assignment */
Global  bool    aexpand;         /* expand looped parts  */
Global  int     aexpmin;         /* expand looped parts  */
Global  int     aexpmax;         /* expand looped parts  */
Global  bool    atadpole;        /* include tadpoles */
Global  int     aextself;        /* self-eng. part at e-line */
Global  bool    aonepi;          /* one paritcle irreducibility */
Global  bool    ablock;          /* block */
Global  bool    aselfe;          /* self-entery part */

Global  bool    hep;            /* HEP */
Global  bool    countert;       /* counter term */
Global  bool    anyct;          /* not defined counter term */
Global  bool    gm2;            /* g-2 */
Global  int     selprop;        /* counter term */
Global  bool    undefp;         /* include undefined particles */
Global  string  kinem;          /* path of kinematics */
Global  string  prefix;         /* prefix of the generated code */

/* value of extself : False, True and the following */
#define INMODEL (True + 1)
#ifdef DEFGLOBAL
Global  string  mdlbools[] = {"no", "yes",  "mdl"};
Global  int     mdlboolv[] = {False,  True, INMODEL};
Global  int     nmdlbool   = sizeof(mdlboolv)/sizeof(mdlboolv[0]);
#else
Global  string   mdlbools[];
Global  int      mdlboolv[];
Global  int      nmdlbool;
#endif

/* Flags set by biconn
 */
Global  bool    bitadp;
Global  bool    biexts;
Global  bool    bicountert;
Global  bool    bi1pi;
Global  bool    biblock;

/* Flags set by axconn
 */
Global  bool    axtadp;
Global  bool    axexts;
Global  bool    axcountert;
Global  bool    ax1pi;
Global  bool    axblock;

/* Output flags
 */
Global  bool    display;        /* display configurations.  */
Global  bool    outf;           /* output to a file */
Global  int     seltype;        /* select type */
Global  bool    onetype;        /* one form each type */
Global  bool    onegraph;       /* output one graph */
Global  bool    passg;          /* padump or not */

/* Flag for assign
 */
Global  bool    skipasp;        /* skip assigning particle */

#ifdef DEFGLOBAL
Global  long    orda0=0L, orda1=0L;  /* graph number in ord. algor. */
#else
Global  long    orda0, orda1;   /* graph number in ord. algor. */
#endif

/*-------------------------------------------------------------------
 * table of one-particle irreducible parts
 */
typedef struct {
    int nlegs;
    int next;
    int loop;
    int nedges;
    int extself;
} topip;

Global topip opitbl[MAXLOOP];
Global int   nopitbl;
Global FILE  *prgrsp;

/*-------------------------------------------------------------------
 * for external self-energy part.
 */
Global bool extsgraph;
Global int  extbrv[MAXELINE];
Global int  extbrw[MAXELINE];

/*-------------------------------------------------------------------
 */
#include "proto.h"
