/*********************************************************************
 * $id$
 *
 *  File : hepend.c (prefix = he)
 *
 *  Assign particle to internal lines.
 *  Looped vertices are ignored for the assignment.
 *  Duplication of graphs in terms of automorphism groups of graphs
 *  is not removed.
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include "grc.h"
#include "divide.h"

/*------------------------------------------------------------------*/
/* Local functions */
Private Void    heset1ct   ARG((node));
Private Void    heset2ct   ARG((node,int*,int*));

/* Local variables */
Local   int *mord;

/*==================================================================*/
Public  Void    hetest(NoArg)
/*--------------------------------------------------------------------
 * The last part of particle assignment.
 * This function checks duplication of graphs and calculate
 * symmetry factor.
 */
{
    long   nsym;

    /* discard duplicated graphs : nsym > 0 : the symmetry factor */
    if((nsym = lcatest()) > 0) {

        if(expand || countert) {
#if DBG
            if(dbg >= padbg) {
                printf("hetest: check f-line\n");
            }
#endif
            /* construct fermion line */
            flget(cgraph);

            /* check Furry's theorem */
            if(flfurry()) {
#if DBG
                if(dbg >= padbg) {
                    printf("hetest: pass furry\n");
                }
#endif

                /* symmetry factor */
                cgraph->sfactor = nsym*flfact();

                /* output the resulting graph */
                presult();
            }
        } else {
            cgraph->sfactor = 1;
            presult();
        }
#if DBG
    } else {
        if(dbg >= padbg) {
            printf("hetest: del by lcatest\n");
        }
#endif
    }
    lcreset();
}

/*==================================================================*/
Public  Void    hesetct(tord)
int     *tord;
/*--------------------------------------------------------------------
 *  Assign order to the counter terms
 */
{
    mord = tord;

    if(!countert || !anyct) {
        hetest();
        return;
    }

    heset1ct(0);
}

/*==================================================================*/
Private Void    heset1ct(n)
node    n;
/*--------------------------------------------------------------------
 *  Assign order to a counter term
 */
{
    int no, to, io;
    int q[MAXMORD];

#if DBG
    if(dbg >= padbg) {
        printf("heset1ct:n=%d->", n);
    }
#endif

    /* find not assigned node */
    while (n < nnodes && (ndvertex(n) != VNULL || ndloop(n) < 1
          || ndtype(n) == NDEXTERN)) {
        n++;
    }
#if DBG
    if(dbg >= padbg) {
        printf("%d/%d\n", n, nnodes);
    }
#endif

    /* no more nodes */
    if(n >= nnodes) {
        hetest();
        return;
    }

    no = 2*ndloop(n) + ndlegs(n) - 2;
    to = 0;
    for(io = 0; io < nmordtbl; io++) {
        to += mord[io];
    }
    if(no > to) {
        return;
    }

    dvinit(nmordtbl, mord, no, q);
    heset2ct(n, mord, q);

    while(dvnext(nmordtbl, mord, q)) {
        heset2ct(n, mord, q);
    }
}


/*==================================================================*/
Private Void    heset2ct(n, kord, q)
node    n;
int     kord[], q[];
/*--------------------------------------------------------------------
 */
{
    int io;

    ndvertex(n) = mvmatcg(n, q);
    for(io = 0; io < nmordtbl; io++) {
        kord[io] -= q[io];
    }
    heset1ct(n+1);
    for(io = 0; io < nmordtbl; io++) {
        kord[io] += q[io];
    }
    ndvertex(n) = VNULL;
}
