/*********************************************************************
 * $id$
 *
 *  File : print.c
 *
 *  Print graphs
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "grc.h"

/*--------------------------------------------------------------------
 * Local functions
 */
Private Void    prdinc       ARG((long));
Private long    prgcd        ARG((long,long));

Private bool    prsel        ARG((NoArg));
Private Void    prgrsmmry    ARG((NoArg));

/*--------------------------------------------------------------------
 * Local variables
 */
#define pvyesno(x)      ((x)? "Yes" : "No")
Local  double   prclk1, prclk2;
Local  long     pr1pi;
Local  long     lastn = -1;

/*==================================================================*/
Public  Void    prinit(NoArg)
/*--------------------------------------------------------------------
    Print initial message
*/
{
    printf("\n");
    printf("    Model  file name = \"%s\" (DefAux = %d)\n", fnmdl, defaux);
    printf("    Input  file name = \"%s\"\n", fnprc);

    if(outf)
    printf("    Output file name = \"%s\"\n", fngrf);
}

/*==================================================================*/
Public  Void    presult(NoArg)
/*--------------------------------------------------------------------
    Print resulting graph.
*/
{
    int l;
    int sel;

    nconfig++;
    if(nconfig < 2) {
        pr1pi = 0;
    }
    if(shrinkaux && ax1pi) {
        pr1pi++;
    } else if(!shrinkaux && bi1pi) {
        pr1pi++;
    }

    cprocess->graphs = nconfig;
    cgraph->grid     = nconfig;

    if(!skipasp) {
        prdinc(cgraph->sfactor);
    }

#if DBG
    if(nconfig == dbgst) {
        dbg     = 2;
        display = 15;
        trace   = True;
        outf    = True;
    }
#endif

    l = bfconn();

    if(display > 0) {
        printf("\n Graph No. %8ld (%d)", nconfig, l);
        prgtype();
    }

    if(display > 1) {
        prgraph();
#if DBG
    } else if(dispno == nconfig) {
        prgraph();
        error("presult", "dispno == nconfig, stop by option");
        erfail();
#endif
    }

    prgrsmmry();

    if(outf && (sel = prsel())) {
        if(skipasp)
            Discard painit();
        if((skipasp || sel) && cptest < 100) {
            fograph();
        }
    }
}

/*==================================================================*/
Public  Void    prgraph(NoArg)
/*--------------------------------------------------------------------
    Print the current graph.
*/
{
    printf("Graph No. %ld (%ld)\n", nconfig, ++ndump);
    prparm();
    eldump();
    lcdump(0);
    padump();
    illdump();
    putchar(EOL);

    if(outf) {
        if(skipasp)
            Discard painit();
        fograph();
    }
}


/*==================================================================*/
Public  Void    prdump(NoArg)
/*--------------------------------------------------------------------
    Dump current graph.
*/
{
    printf("Graph No. %ld (%ld)\n", nconfig, ++ndump);
    prparm();
    eldump();
    lcdump(1);
    padump();
    illdump();

    if(outf) {
        if(skipasp)
            Discard painit();
        fograph();
    }
}


/*==================================================================*/
Public  Void    prparm(NoArg)
/*--------------------------------------------------------------------
    Print parameters
*/
{
    printf("Parameter\n");
    printf("nextern = %4d, loop = %4d, order = %4d, nnodes = %4d\n",
           nextern, loop, order, nnodes);
    printf("nlclass = %4d\n", nlclass);
}

/*==================================================================*/
Public  Void    prtype(mode)
int     mode;
/*--------------------------------------------------------------------
    Print type
        mode < 0 : initialization
             = 0 : ending message
             > 0 : next type
*/
{
    long  genn;

    if(mode < 0) {
        gntype  = 0;
        ndump = 0;
        return;
    }
    if(gntype > 0 && display >= 0) {
        genn = nconfig-lastn;
        if(skipasp || genn > 0)
            printf("*** Graph type %4d Graphs = %6ld  Total = %6ld\n",
                   gntype, genn, nconfig);
    }
    if(mode > 0) {
        gntype++;
        if(skipasp) {
            printf("\n*** Graph type %d:  nodes = %d\n\n", 
                   gntype, nnodes);
            prnodes();
            putchar(EOL);
        }
        lastn = nconfig;
    }
}

/*==================================================================*/
Public  Void    prnodes(NoArg)
/*--------------------------------------------------------------------
    Print parameters
*/
{
    node        v;

    printf("node :  typ : legs : loop \n");
    for(v = 0; v < nnodes; v++) {
        printf("%4d : %4d : %4d : %4d\n",
               v, ndtype(v), ndlegs(v), ndloop(v));
    }
}

/*==================================================================*/
Public  Void    prfirst(NoArg)
/*--------------------------------------------------------------------
    Print the first message.

--------------------------------------------------------------------*/
{
    node n;

    printf("*************************************************\n");
    printf("Order = %d (", order);
    for(n = 0; n < nsordtbl; n++) {
        if(n > 0)
            printf(", ");
        printf("%s = %d", sordtbl[n], sorder[n]);
    }
    printf(") loop = %d\n", loop);
    printf("\n%4d external lines\n\n", nextern);

    for(n = 0; n < nextern; n++) {
        printf("%4d %s %s\n",
                tovert(n), eliorf(ndiorf(n)),
                pvename(ndiorf(n), ndname(n)));
    }

    printf("\nConditions of generationg graphs:\n");
    if(expand) {
        printf("    Gnenerate expanded graph = Yes ");
        if(expmax < 0) {
            printf("[%d ... \\infty)\n", expmin);
        } else {
            printf("[%d ... %d]\n", expmin, expmax);
        }
    } else {
        printf("    Gnenerate expanded graph = No\n");
    }
    printf("    Shrink AUX propagators   = %s\n", pvyesno(shrinkaux));
    printf("    Only 1PI graph           = %s\n", pvyesno(onepi));
    printf("    Only one block           = %s\n", pvyesno(block));
    printf("    Gnenerate tadpole        = %s\n", pvyesno(tadpole));
    printf("    Self energy at ext.      = %s\n", mdlbools[extself]);
    printf("    Counter terms            = %s\n", pvyesno(countert));
    printf("    Undefined counter terms  = %s\n", pvyesno(anyct));
    printf("    Expanded self energy     = %s\n", pvyesno(selfe));
    printf("    Assign particle          = %s\n", pvyesno(!skipasp));
    printf("    HEP                      = %s\n", pvyesno(hep));

    printf("\nConditions for outputs:\n");
    printf("    Display graph data       = %s\n", pvyesno(display));
    printf("    Write graph data to file = %s\n", pvyesno(outf));

#if DBG
    printf("\nConditions for debugging:\n");
    printf("    Debug                    = %s\n", pvyesno(dbg));
    printf("    Monitor performance      = %s\n", pvyesno(mon));
    printf("    Trace functions          = %s\n", pvyesno(trace));
    printf("    Check the result         = %s\n", pvyesno(check));

    if(selprop > 0) {
        printf("\nSelect graph with propagator of %s\n",
               ntable[ptable[selprop]->pname[0]].name);
    }
#endif
    printf("=================================================\n");
    second(&prclk1);
}

/*==================================================================*/
Public  Void    prlast(NoArg)
/*--------------------------------------------------------------------
    Print the last message.
--------------------------------------------------------------------*/
{
    double sec;

    second(&prclk2);
    sec = prclk2 - prclk1;
    printf("\nExt = %d, Loop = %d : Total = %ld",
           nextern, loop, nconfig);
    if(!skipasp) {
        printf(", Weighted sum = ");
        prprn(sfnum, sfden);
    }
    printf("(%8.2f sec)\n", sec);
    putchar(EOL);
    printf("type = %d, 1PI = %ld\n", gntype, pr1pi);
    if(!outf) {
        printf("\nE = %d, L = %d : T = %ld ", nextern, loop, nconfig);
        printf("orda1 = %ld, orda0 = %ld ", orda1, orda0);
        printf("(%8.2f sec)\n", sec);
    }
    printf("*************************************************\n");
}

/*==================================================================*/
Public  Void    prgtype(NoArg)
/*--------------------------------------------------------------------
    Print type of the graph.
--------------------------------------------------------------------*/
{
    if(loop < 1) {
        printf(" tree   ");
    } else if(!expand) {
        printf(" skelton");
    } else if(shrinkaux && axblock) {
        printf(" block  ");
    } else if(!shrinkaux && biblock) {
        printf(" block  ");
    } else if(shrinkaux && ax1pi) {
        printf(" 1PI    ");
    } else if(!shrinkaux && bi1pi) {
        printf(" 1PI    ");
    } else {
        printf("        ");
    }

    if(extself && shrinkaux && axexts) {
        printf(" e-self ");
    } else if(extself && !shrinkaux && biexts) {
        printf(" e-self ");
    } else {
        printf("        ");
    }

    if(tadpole && shrinkaux && axtadp) {
        printf(" tadpole");
    } else if(tadpole && !shrinkaux && bitadp) {
        printf(" tadpole");
    } else {
        printf("        ");
    }

    putchar(EOL);
}

/*==================================================================*/
Public  Void   prprn(num, den)
long    num, den;
/*--------------------------------------------------------------------
*/
{
    int sgn = 1;

    if(num < 0) {
        sgn = - 1;
        num = - num;
    }
    if(den < 0) {
        sgn = - sgn;
        den = - den;
    }
    if(sgn < 0)
        putchar('-');
    if(num == 0 || den == 1)
        printf("%ld\n", num);
    else
        printf("%ld/%ld\n", num, den);

}

/*==================================================================*/
Private Void   prdinc(den)
long    den;
/*--------------------------------------------------------------------
*/
{
    long  gcd;
    long  den1;
    int   sgn;

    if(den > 0) {
        den1 = den;
        sgn = 1;
    } else {
        den1 = - den;
        sgn = - 1;
    }
    gcd    = prgcd(sfden, den1);
    den1  /= gcd;
    sfnum  = sfnum * den1 +  sgn*(sfden / gcd);
    sfden *= den1;
    gcd    = prgcd(sfnum, sfden);
    sfnum /= gcd;
    sfden /= gcd;
}

/*==================================================================*/
Private long   prgcd(x, y)
long    x, y;
/*--------------------------------------------------------------------
*/
{
    long t;

    while (y != 0) {
        t = x % y;
        x = y;
        y = t;
    }

    return x;
}

/*==================================================================*/
Private bool   prsel(NoArg)
/*--------------------------------------------------------------------
*/
{
    int  l;
    node n;
    static int ogntype = -1;     

    if(onegraph && onegraph != nconfig)
        return False;

    if(onetype && lastn + 1 != nconfig)
        return False;

    if(seltype != 0 && gntype != seltype)
        return False;

#if DBG 
    if(onegrtype) {
        if(gntype == ogntype) {
            return False;
        }
    }
#endif
    
    if(selprop <= 0) {
        ogntype = gntype;
        return True;    
    }
    for(n = nextern; n < nnodes; n++) {
        for(l = 0; l < ndlegs(n); l++) {
            if(ndnextn(n,l) < nextern)
                continue;
            if(selprop == abs(ndparticle(n,l))) {
                ogntype = gntype;
                return True;
            }
        }
    }
    return False;
}

/*==================================================================*/
Private Void    prgrsmmry(NoArg)
/*--------------------------------------------------------------------
 */
{
    int v, lg = 0, lp = 0, nv = 0;
    static bool first = True;

    if(!outf) {
        return;
    }
    if(first) {
        fprintf(prgrsp, "#format\n");
        fprintf(prgrsp, "#  line  ::= nprc '| ' nconf '|' vtx '|' 1PI ");
        fprintf(prgrsp, "|' '1' .\n");
        fprintf(prgrsp, "#  vtx   ::= lp lg nv (':' lp lg nv)*          ");
        fprintf(prgrsp, "info of vertices\n");
        fprintf(prgrsp, "#  1PI   ::= lp lg eg es (':' lp lg eg es)*    ");
        fprintf(prgrsp, "info of one-particle irreducible parts.\n");
        fprintf(prgrsp, "#  nprc  ::= integer    process number\n");
        fprintf(prgrsp, "#  nconf ::= integer    graph number\n");
        fprintf(prgrsp, "#  lp    ::= integer    number of loops\n");
        fprintf(prgrsp, "#  lg    ::= integer    number of legs\n");
        fprintf(prgrsp, "#  le    ::= integer    number of edges\n");
        fprintf(prgrsp, "#  eg    ::= integer    number of edges\n");
        fprintf(prgrsp, "#  es    ::= bool       external self or not\n");
        first = False;
    }
    if(prgrsp == NULL) {
        return;
    }
    fprintf(prgrsp, "%d|", nprocess);
    fprintf(prgrsp, "%ld|", nconfig);
    nv = 0;
    for(v = nextern; v < nnodes; v++) {
        if(lg != ndlegs(v) || lp != ndloop(v)) {
            if(nv > 0) {
                fprintf(prgrsp, "%d %d %d:", lp, lg, nv);
            }
            lg = ndlegs(v);
            lp = ndloop(v);
            nv = 0;
        }
        nv++;
    }
    if(nv > 0) {
        fprintf(prgrsp, "%d %d %d|", lp, lg, nv);
    }
    for(v = 0; v < nopitbl; v++) {
        if(v > 0) {
            fprintf(prgrsp, ":");
        }
        fprintf(prgrsp, "%d %d %d %d", opitbl[v].loop,
                opitbl[v].nlegs, opitbl[v].nedges, opitbl[v].extself);
    }
    fprintf(prgrsp, "|1\n");
}
