/*********************************************************************
 *  File : dcx.c
 *
 *  Window controller for X
 *                                                 by Kaneko Toshiaki
 */

#ifdef  TOOLKIT
# include <X11/Intrinsic.h> 
# include <X11/StringDefs.h>
# include <X11/Xaw/Form.h>
# include <X11/Xaw/Paned.h>
# include <X11/Xaw/Label.h>
# include <X11/Xaw/Command.h>
#else
# ifdef X11
#  include <X11/Xlib.h>
#  include <X11/Xutil.h>
# endif
#endif

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include "common.h"
#include "dcx.h"
#include "vr.h"
#include "vrx.h"
#include "draw.h"
#include "dbg.h"

/*--------------------------------------------------------------------
 * Parameters
 */
#ifdef X11
# undef   Global
# define  Global
# include "dcx.h"
# undef   Global
# define  Global extern
#endif

#ifndef TOOLKIT
#define Widget      int
#define XtPointer   int* 
#endif

/*--------------------------------------------------------------------
 * Private functions
 */
Private Void setappdat    ARG((int, char**));
Private Void dccreate     ARG((NoArg));
Private Void cv_redraw    ARG(());
Private Void cv_credraw   ARG(());
Private Void cv_quit      ARG(());
Private Void cv_next      ARG(());
Private Void cv_prev      ARG(());
Private Void cv_forward   ARG(());
Private Void cv_back      ARG(());
Private Void cv_greater   ARG(());
Private Void cv_smaller   ARG(());
Private Void cv_label     ARG(());
Private Void cv_text      ARG(());
Private Void cv_vtext     ARG(());
Private Void cv_itext     ARG(());
Private Void cv_ps        ARG(());
Private Void cv_eps       ARG(());
Private Void cv_jump      ARG(());
Private Void cv_digit0    ARG((int));
#ifdef TOOLKIT
Private Void cv_digit     ARG((Widget, XtPointer, XtPointer));
#endif

/*--------------------------------------------------------------------
 * Local variable
 */
#ifdef TOOLKIT
Local   Widget toplevel;
Local   XtAppContext app_con;

Local   char *APP_CLASS="GrcDraw";     /* the name of application class */

static XtActionsRec actions[] = {
    {"cv_quit",    (XtActionProc)cv_quit},
    {"cv_redraw",  (XtActionProc)cv_redraw},
    {"cv_credraw", (XtActionProc)cv_credraw},
    {"cv_next",    (XtActionProc)cv_next},
    {"cv_prev",    (XtActionProc)cv_prev},
    {"cv_forward", (XtActionProc)cv_forward},
    {"cv_back",    (XtActionProc)cv_back},
    {"cv_greater", (XtActionProc)cv_greater},
    {"cv_smaller", (XtActionProc)cv_smaller},
    {"cv_label",   (XtActionProc)cv_label},
    {"cv_text",    (XtActionProc)cv_text},
    {"cv_vtext",   (XtActionProc)cv_vtext},
    {"cv_itext",   (XtActionProc)cv_itext},
    {"cv_ps",      (XtActionProc)cv_ps},
    {"cv_eps",     (XtActionProc)cv_eps},
    {"cv_digit",   (XtActionProc)cv_digit},
    {"cv_jump",    (XtActionProc)cv_jump},
};

Local   String fallback_resources[] = {
    "*translations: #override\\n\
        <Key>0:   cv_digit(0)\\n\
        <Key>1:   cv_digit(1)\\n\
        <Key>2:   cv_digit(2)\\n\
        <Key>3:   cv_digit(3)\\n\
        <Key>4:   cv_digit(4)\\n\
        <Key>5:   cv_digit(5)\\n\
        <Key>6:   cv_digit(6)\\n\
        <Key>7:   cv_digit(7)\\n\
        <Key>8:   cv_digit(8)\\n\
        <Key>9:   cv_digit(9)\\n\
        <Key>Return:  cv_jump()\\n\
        <Key>KP_Enter:  cv_jump()\\n\
        <Key>B:   cv_back()\\n\
        <Key>E:   cv_eps()\\n\
        <Key>F:   cv_forward()\\n\
        <Key>G:   cv_greater()\\n\
        <Key>I:   cv_itext()\\n\
        <Key>J:   cv_jump()\\n\
        <Key>L:   cv_label()\\n\
        <Key>N:   cv_next()\\n\
        <Key>O:   cv_ps()\\n\
        <Key>P:   cv_prev()\\n\
        <Key>Q:   cv_quit()\\n\
        <Key>R:   cv_credraw()\\n\
        <Key>S:   cv_smaller()\\n\
        <Key>T:   cv_text()\\n\
        <Key>V:   cv_vtext()",
    "*graphs.translations: #override\\n\
        <Configure>: cv_redraw()\\n\
        <Expose>: cv_redraw()",
    "*graphs.width: 600",
    "*graphs.height: 600",
    "*cform.fromHoriz: graphs",
    "*cform.borderWidth: 0",
    "*showGrip: off",
    "*Command.width: 100",
    "*Command.height: 20",
    "*proc_label.borderWidth: 0",
    "*proc_label.fromVert: quit_button",
    "*forp_button.fromVert: proc_label",
    "*backp_button.fromVert: forp_button",
    "*grph_label.borderWidth: 0",
    "*grph_label.fromVert: backp_button",
    "*forward_button.fromVert: grph_label",
    "*back_button.fromVert: forward_button",
    "*greater_button.fromVert: back_button",
    "*smaller_button.fromVert: greater_button",
    "*label_button.fromVert: smaller_button",
    "*text_button.fromVert: label_button",
    "*ps_button.fromVert: text_button",
    "*Form.background: white",
    "*Label.background: white",
    "*Label.foreground: black",
    "*Command.background: white",
    "*Command.foreground: black",
    NULL
};

Local Widget form;
Local Widget graphs;
Local Widget cform;
Local Widget quit_button;
Local Widget proc_label, forp_button, backp_button;
Local Widget grph_label, forward_button, back_button;
Local Widget greater_button, smaller_button;
Local Widget label_button, text_button;
Local Widget ps_button;
#else  /* !TOOLKIT */
# ifdef X11
Local string     dispnam = NULL;
Local Uint       width = 600, height = 600;
Local char       prog_name[8] = "grcdraw";
Local XSizeHints size_hints;
Local XWMHints   wmhints;
# endif
#endif

Local Ulong  forecolr, backcolr;
Local int    ngrph = 0, jgrph = 0;
Local bool   dcdbg = False;

#ifdef TURBOC
Local int    backcolr = VR_WHITE;
#endif

/*==================================================================*/
Public  Void   dcinit(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("dcinit\n");

#ifdef TOOLKIT
    toplevel = XtVaAppInitialize(&app_con, APP_CLASS,
                NULL, 0,  /* format of options */
                &argc, argv,
                fallback_resources,
                NULL);
    disp    = XtDisplay(toplevel);
    screen  = DefaultScreen(disp);
    root    = XtWindow(toplevel);
    
    forecolr = BlackPixel(disp, screen);
    backcolr = WhitePixel(disp, screen);
#else
# ifdef X11
/*
    disp    = XOpenDisplay(getenv("DISPLAY"));
 */
    if((disp = XOpenDisplay((char*)NULL)) == NULL) {
        Discard fprintf(Stderr, "Cannot connect to X server %s\n",
                        XDisplayName(dispnam));
        exit(1);
    }
    screen  = DefaultScreen(disp);
    root    = RootWindow(disp, 0);

    size_hints.flags = 0;
    size_hints.min_width = size_hints.min_height = 50;
    size_hints.x = size_hints.y = 0;
    size_hints.width  = width;
    size_hints.height = height;

    win     = XCreateSimpleWindow(disp, root, size_hints.x, size_hints.y,
              width, height, ((Ulong) 4), 
              BlackPixel(disp, screen), WhitePixel(disp, screen));

    wmhints.flags = InputHint | StateHint;
    wmhints.input = True;   /* window manager must direct input */
    wmhints.initial_state = NormalState;
    XSetWMHints(disp, win, &wmhints);

# endif
#endif


    setappdat(argc, argv);

    dccreate();
}

/*====================================================================*/
Private Void setappdat(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("setappdat\n");

    if(argc != 1) {
        fprintf(Stderr, "unknown option %s\n", argv[1]);
        exit(1);
    }

#ifdef TOOLKIT
    /* get resource with XtVaGetApplicationResources */

    /* actions */
    XtAppAddActions(app_con, actions, XtNumber(actions));
#else
# ifdef X11
    XSetStandardProperties(disp, win, prog_name, prog_name,
            (Pixmap) 0,
            argv, argc, &size_hints);
# endif
#endif

}

/*==================================================================*/
Private Void   dccreate(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("dccreate\n");

#ifdef TOOLKIT
    /* form including buttons and drawing plane of the function
     * value.
     */
    form        = XtVaCreateManagedWidget("form",
                      formWidgetClass, toplevel, NULL);

    /* graphic plane
     */
    graphs      = XtVaCreateManagedWidget("graphs",
                      widgetClass, form,
                      XtNbackground, backcolr,
                      NULL);

    /* cform for buttons
     */
    cform       = XtVaCreateManagedWidget("cform",
                      formWidgetClass, form,
                      NULL);

    /* quit button
     */
    quit_button = XtVaCreateManagedWidget("quit_button",
                      commandWidgetClass, cform, 
                      XtNlabel, "quit",
                      NULL);
    XtAddCallback(quit_button, XtNcallback, cv_quit, NULL);

    /* label for process number
     */
    proc_label  = XtVaCreateManagedWidget("proc_label",
                      labelWidgetClass, cform, 
                      XtNlabel, " ", NULL);

    /* next button
     */
    forp_button = XtVaCreateManagedWidget("forp_button",
                      commandWidgetClass, cform, 
                      XtNlabel, "next proc", NULL);
    XtAddCallback(forp_button, XtNcallback, cv_next, NULL);

    /* prev button
     */
    backp_button = XtVaCreateManagedWidget("backp_button",
                      commandWidgetClass, cform, 
                      XtNlabel, "prev proc", NULL);
    XtAddCallback(backp_button, XtNcallback, cv_prev, NULL);
 
    /* label for graph number
     */
    grph_label  = XtVaCreateManagedWidget("grph_label",
                      labelWidgetClass, cform, 
                      XtNlabel, " ", NULL);

    /* forward button
     */
    forward_button = XtVaCreateManagedWidget("forward_button",
                      commandWidgetClass, cform, 
                      XtNlabel, "forward", NULL);
    XtAddCallback(forward_button, XtNcallback, cv_forward, NULL);

    /* back button
     */
    back_button = XtVaCreateManagedWidget("back_button",
                      commandWidgetClass, cform, 
                      XtNlabel, "back", NULL);
    XtAddCallback(back_button, XtNcallback, cv_back, NULL);

    /* greater button
     */
    greater_button = XtVaCreateManagedWidget("greater_button",
                      commandWidgetClass, cform, 
                      XtNlabel, "greater", NULL);
    XtAddCallback(greater_button, XtNcallback, cv_greater, NULL);

    /* smaller button
     */
    smaller_button = XtVaCreateManagedWidget("smaller_button",
                      commandWidgetClass, cform, 
                      XtNlabel, "smaller", NULL);
    XtAddCallback(smaller_button, XtNcallback, cv_smaller, NULL);

    /* label button
     */
    label_button = XtVaCreateManagedWidget("label_button",
                      commandWidgetClass, cform, 
                      XtNlabel, "label", NULL);
    XtAddCallback(label_button, XtNcallback, cv_label, NULL);

    /* text button
     */
    text_button = XtVaCreateManagedWidget("text_button",
                      commandWidgetClass, cform, 
                      XtNlabel, "text", NULL);
    XtAddCallback(text_button, XtNcallback, cv_text, NULL);

    /* ps button
     */
    ps_button = XtVaCreateManagedWidget("ps_button",
                      commandWidgetClass, cform, 
                      XtNlabel, "ps file", NULL);
    XtAddCallback(ps_button, XtNcallback, cv_ps, NULL);

    /* initialize drawing package
     */
    wid = &graphs;
#else
# ifdef X11

    XSelectInput(disp, win,
                 ExposureMask | ButtonPressMask | KeyPressMask);

    XMapWindow(disp, win);
/*
    XNextEvent(disp, &event);
*/

# endif
#endif

    vxrinit();
}

/*==================================================================*/
Public  Void   dcmain(NoArg)
/*--------------------------------------------------------------------
*/
{
#ifndef TOOLKIT
# ifdef X11
    XEvent event;
    int   index=0;
# endif
    int   key;
#endif

    if(dcdbg)
        printf("dcmain\n");

#ifdef TOOLKIT
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_con);
#else
# ifdef X11
    while(True) {
        XNextEvent(disp, &event);

        if(dcdbg) {
            printf("event type = %d(%d, %d)\n",
                event.type, ConfigureNotify, Expose);
        }
    
        switch (event.type)
        {
          case ConfigureNotify :
               cv_redraw();
               break;
          case Expose :
               cv_redraw();
               break;
          case ButtonPress :
               break;
          case KeyPress :
               key = XKeycodeToKeysym(disp, 
                         event.xkey.keycode, index);
               if(dcdbg)
                   printf("key = %d\n", key);
               if(key > 255 || key < 0)
                   key = 0;
# else
    backcolr = vrback(backcolr);
    cv_redraw();
    while(True) {
               key = getchar();
# endif
               key = tolower(key);
               switch (key)
               {
               case '0':
               case '1':
               case '2':
               case '3':
               case '4':
               case '5':
               case '6':
               case '7':
               case '8':
               case '9':
                   cv_digit0(key);
                   break;
               case '\r':
                   cv_jump();
                   break;
               case 'j':
                   cv_jump();
                   break;
               case 'r':
                   cv_credraw();
                   break;
               case 'q':
                   cv_quit();
                   break;
               case 'n':
                   cv_next();
                   break;
               case 'p':
                   cv_prev();
                   break;
               case 'f':
                   cv_forward();
                   break;
               case 'b':
                   cv_back();
                   break;
               case 'g':
                   cv_greater();
                   break;
               case 's':
                   cv_smaller();
                   break;
               case 'l':
                   cv_label();
                   break;
               case 't':
                   cv_text();
                   break;
               case 'i':
                   cv_itext();
                   break;
               case 'v':
                   cv_vtext();
                   break;
               case 'o':
                   cv_ps();
                   break;
               case 'e':
                   cv_eps();
                   break;
# ifdef TURBOC
               case 'w':
                   backcolr = (backcolr == VR_WHITE) ? VR_BLACK : VR_WHITE;
                   backcolr = vrback(backcolr);
                   cv_credraw();
                   break;
# endif
               }
# ifdef X11
               break;
          default :
               if(dcdbg)
                   printf("other type %d\n", event.type);
               break;
        }
# endif
    }
#endif
}

/*====================================================================*/
Private Void cv_quit()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_quit\n");

    drquit();
}

/*====================================================================*/
Private Void cv_redraw()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_redraw\n");

    drredraw(False);

    if(dcdbg)
        printf("cv_redraw: end\n");
}

/*====================================================================*/
Private Void cv_credraw()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_credraw\n");

    drredraw(True);

    if(dcdbg)
        printf("cv_credraw: end\n");
}

/*====================================================================*/
Private Void cv_next()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_next\n");

    drproc(1);
}

/*====================================================================*/
Private Void cv_prev()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_prev\n");

    drproc(-1);
}

/*====================================================================*/
Private Void cv_forward()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_forward\n");

    drforward(1);
}

/*====================================================================*/
Private Void cv_back()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_back\n");

    drforward(-1);
}

/*====================================================================*/
Private Void cv_greater()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_greater\n");

    drresize(-1);
}

/*====================================================================*/
Private Void cv_smaller()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_smaller\n");

    drresize(1);
}

/*====================================================================*/
Private Void cv_label()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_label\n");

    drslabel();
}

/*====================================================================*/
Private Void cv_text()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_text\n");

    drstext();
}

/*====================================================================*/
Private Void cv_vtext()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_vtext\n");

    drsvtext();
}

/*====================================================================*/
Private Void cv_itext()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_itext\n");

    drsitext();
}

/*====================================================================*/
Private Void cv_ps()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_ps\n");

    drps(0);
}

/*====================================================================*/
Private Void cv_eps()
/*--------------------------------------------------------------------
*/
{
    if(dcdbg)
        printf("cv_ps\n");

    drps(1);
}

#ifdef TOOLKIT
/*====================================================================*/
Private Void cv_digit(w, d1, d2)
Widget w;
XtPointer d1;
XtPointer d2;
/*--------------------------------------------------------------------
*/
{
    char *str;
    str = *((char**) d2);

    if(dcdbg)
        printf("cv_ps\n");

    cv_digit0(str[0]);
}
#endif

/*====================================================================*/
Private Void cv_digit0(c)
int     c;
/*--------------------------------------------------------------------
*/
{
    if(isdigit(c)) {
        jgrph = jgrph * 10 + c - '0';
    }
    if(jgrph > ngrph) 
        jgrph = ngrph;

    dcgrphl(jgrph-1, ngrph);
}

/*====================================================================*/
Private Void cv_jump()
/*--------------------------------------------------------------------
*/
{
    drjump(jgrph-1);
    jgrph = 0;
}

/*====================================================================*/
Public  Void dcprocl(cv, total)
int cv, total;
/*--------------------------------------------------------------------
*/
{
#ifdef  TOOLKIT
    static char str[256];

    sprintf(str, "%4d/%-4d\n", cv+1, total);
    XtVaSetValues(proc_label, XtNlabel, str, NULL);
#endif
}

/*====================================================================*/
Public  Void dcgrphl(cv, total)
int cv, total;
/*--------------------------------------------------------------------
*/
{
#ifdef  TOOLKIT
    static char str[256];

    ngrph = total;
    sprintf(str, "%4d/%-4d\n", cv+1, total);
    XtVaSetValues(grph_label, XtNlabel, str, NULL);
#else
    ngrph = total;
#endif
}
