/*********************************************************************
 *  File : draw.c
 *
 *  Main program for graph drawing
 *
 *  Entry points:
 *    1. main   : entry of the program called by OS.
 *    2. prdump : dump the current graph.
 *
 *                                                 by Kaneko Toshiaki
 */

#define  DEFGLOBAL

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "grdata.h"
#include "vr.h"
#include "args.h"
#include "config.h"
#include "token.h"
#include "version.h"
#include "window.h"
#include "drgrf.h"
#include "dcx.h"
#include "draw.h"

/*--------------------------------------------------------------------
 * Private functions
 */
Private int    drcarg        ARG((int, char**));
Private Void   drhelp        ARG((string));

/*--------------------------------------------------------------------
 * Global variables
 */
        int     drlabel = 0;
        int     drcomp  = 0;
        int     drtext  = 1;
        int     drvtext = 0;
        int     dritext = 0;

/*--------------------------------------------------------------------
 * Local variable
 */
#ifdef FACOM
static  char    finame[]={"grf"};
#else
static  char    finame[]={"out.grf"};   /* default input file */
#endif

Local   string  fngrf = NULL;

Local   int     drstart = 0;
Local   int     drprocs = 0;
Local   int     drwinds = 0;
Local   int     drhlp   = 0;
Local   int     dlabel  = 0;
Local   int     dcomp   = 0;
Local   int     dtext   = 0;
#if DBG
Local   int     ddbg    = 0;
#endif

Local   topt    options[] = {
    {"s", ARINT,  (int*) &drstart},
    {"p", ARINT,  (int*) &drprocs},
    {"w", ARINT,  (int*) &drwinds},
    {"c", ARFLAG, (int*) &dcomp},
    {"l", ARFLAG, (int*) &dlabel},
    {"t", ARFLAG, (int*) &dtext},
    {"h", ARFLAG, (int*) &drhlp},
    {"?", ARFLAG, (int*) &drhlp},
#if DBG
    {"d", ARINT, (int*) &ddbg},
#endif
};

Local int noptions = sizeof(options)/sizeof(topt);

Local   int     proct[MAXPROCS], procg[MAXPROCS];

Local   int         nprc, ngrf, ng, nplanes;
Local   int         nprocesses;
Local   tproc       *proc;
Local   tgraph      *graph;
Local   int         prmode = 0;

/*==================================================================*/
Public  int    main(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
 *  Main routine for graph drawin
 */
{

#if DBG
    dadbg = 3;  /* debug flag for dagrf.c */
    drdbg = 1;  /* debug flag for drgrf.c */
    fldbg = 2;  /* debug flag for flgrf.c */
    ildbg = 2;  /* debug flag for ilgrf.c */
    pmdbg = 2;  /* debug flag for matchv.c */
    pvdbg = 2;  /* debug flag for rdptbl.c */
    rddbg = 2;  /* debug flag for token.c */
    rgdbg = 2;  /* debug flag for rdgrf.c */
    tldbg = 2;  /* debug flag for tlgrf.c */

    dbg = 1;
#endif

    argc = drcarg(argc, argv);
    dcinit(argc, argv);  /* initialize drawing package */

    if(drstart > 1)
        printf("skip to Graph %d\n", drstart);

    /* initalization */
    genct      = True;
    nprocesses = dasave(fngrf, proct, procg);

    if(nprocesses < drprocs) {
        fprintf(Stderr, "draw : total %d processes\n", nprocesses);
        daclose();
        exit(0);
    }
    nprc = drprocs;
    proc = dagetp(nprc);  /* get process */
    ngrf = dangrf(nprc);  /* the number of graphs */
    cprocess = proc;

    drwinds = drinit(OUTX, drwinds,&prmode);
    nplanes = drwinds*drwinds;

    ng  = drstart;      /* the current graph */

    dcprocl(nprc, nprocesses);
    dcgrphl(ng, ngrf); 
    dcmain();

    return 0;
}

/*==================================================================*/
Public Void drredraw(clr)
bool clr;
/*--------------------------------------------------------------------
 */
{
    static int lng = -1, lmaxl = -1;
    long l, maxl;
    bool st;

    maxl = ng + nplanes;
    maxl = (maxl > ngrf) ? ngrf : maxl;

    dcgrphl(ng, ngrf);
    st = drclr(clr || lng != ng || lmaxl != maxl);
    lng   = ng;
    lmaxl = maxl;
/*
    if(!st) {
        return;
    }
*/
    if(ngrf < 1) {
        drsetn(2);
        drctext("No Graph!!");
        drwinds = drsetn(drwinds);
        return;
    }
    for(l = ng; l < maxl; l++) {
        if(drcomp) {
            graph = daread(nprc, l);
        } else {
            graph = daread0(nprc, l);
        }
        cgraph = graph;
        if(cprocess->expand) {
            ilmom();
        }

        drgraph(proc, graph, l);      /* draw the graph */
    }
}

/*==================================================================*/
Public Void drforward(n)
int    n;
/*--------------------------------------------------------------------
 */
{
    int ngo = ng, nng = ng + n*nplanes;

    if(ngrf > 0) {
        if(nng < ngrf)
            ng = nng;

        if(ng < 0)
            ng = 0;

        if(ngo == ng)
            return;
    }

    drredraw(True);
}

/*==================================================================*/
Public Void drps(mode)
int    mode;
/*--------------------------------------------------------------------
 */
{
    prmode = mode;

    wdopen(OUTFIG, &prmode);
    drredraw(True);
    wdcls(OUTFIG);
}
        
/*==================================================================*/
Public Void drproc(n)
int    n;
/*--------------------------------------------------------------------
 */
{
    n += nprc;
    if(n >= nprocesses)
        return;

    if(n < 0)
        return;

    nprc = n;
    proc = dagetp(nprc);  /* get process */
    ngrf = dangrf(nprc);  /* the number of graphs */
    cprocess = proc;
    ng   = 0;
    dcprocl(nprc, nprocesses);
    drredraw(True);
}

/*==================================================================*/
Public Void drslabel(NoArg)
/*--------------------------------------------------------------------
 */
{
    drlabel = ! drlabel;
    drredraw(True);
}

/*==================================================================*/
Public Void drsvtext(NoArg)
/*--------------------------------------------------------------------
 */
{
    drvtext = ! drvtext;
    drredraw(True);
}

/*==================================================================*/
Public Void drsitext(NoArg)
/*--------------------------------------------------------------------
 */
{
    dritext = ! dritext;
    drredraw(True);
}

/*==================================================================*/
Public Void drstext(NoArg)
/*--------------------------------------------------------------------
 */
{
    drtext  = ! drtext;
    drredraw(True);
}

/*==================================================================*/
Public Void drresize(n)
int    n;
/*--------------------------------------------------------------------
 */
{
    int owinds = drwinds;

    if(n > 0) {
        drwinds *= 2;
    } else if(n < 1) {
        drwinds /= 2;
        if(drwinds < 1)
            drwinds = 1;
    }
    if(drwinds == owinds)
        return;

    drwinds = drsetn(drwinds);
    nplanes = drwinds*drwinds;
    drredraw(True);
}

/*==================================================================*/
Public Void drjump(n)
int    n;
/*--------------------------------------------------------------------
 */
{
    if(ngrf > 0) {
        if(n > ngrf)
            n = ngrf;
    
        if(n < 0)
           n = 0;
    
        if(n == ng)
           return;
    
        ng = n;
    }
    drredraw(True);
}

/*==================================================================*/
Public Void drquit(NoArg)
/*--------------------------------------------------------------------
 */
{
    drclose(OUTX);

    daclose();        /* close read package */
    exit(0);
}

/*==================================================================*/
Private int    drcarg(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
 */
{
    int    pos;

    argc = aropt(argc, argv, noptions, options);

    if(drhlp != 0) {
        drhelp(argv[0]);
        exit(0);
    }

    switch (dlabel) {
        case ARPLUS:
            drlabel = True;
            break;
        case ARMINUS:
            drlabel = False;
            break;
    }

    switch (dcomp) {
        case ARPLUS:
            drcomp = True;
            break;
        case ARMINUS:
            drcomp = False;
            break;
    }

    switch (dtext) {
        case ARPLUS:
            drtext = True;
            break;
        case ARMINUS:
            drtext = False;
            break;
    }

    /* input process file name */
    pos = 1;
    if((fngrf = arnext(argc, argv, &pos)) != NULL) {
        pos--;
        argc = ardel(argc, argv, pos, 1);
    } else {
        fngrf = finame;
    }

#if DBG
    /* debugging flags */
    dbg   = 0;

    if(ddbg > 0) {
        dbg = ddbg;
    } else if(ddbg < 0) {
        dbg = -ddbg;
    }
#endif

    return argc;
}

/*==================================================================*/
Private Void   drhelp(cmd)
string  cmd;
/*--------------------------------------------------------------------
 */
{
    grcmsg("Feynman graph drawer");

    printf("\n\n");
    printf("Syntax %s [Graph_File_name]\n",
           cmd);
    printf("\nOption\n");
    printf("    -w int              ");
    printf(": change #graphs/page\n");
    printf("    -t                  ");
    printf(": without text\n");
    printf("    -l                  ");
    printf(": with particle name of propagators\n");
    printf("    -s int              ");
    printf(": starting graph number\n");
    printf("    -p int              ");
    printf(": starting process number\n");
#if DBG
    printf("    -d int              ");
    printf(": set debug level\n");
#endif
    printf("    -h                  ");
    printf("\n");

    printf("\nDefault file name:\n");
    printf("     Graph_File_name = %s\n", finame);
    printf("\n\nCommands\n");
    printf("     n      : next process\n");
    printf("     p      : previous process\n");
    printf("     f      : move forward\n");
    printf("     b      : move backward\n");
    printf("     <int>j : jump\n");
    printf("     s : smaller graphs\n");
    printf("     g : larger graphs\n");
    printf("     l : with/without labels\n");
    printf("     t : with/without text\n");
    printf("     o : output PostScript file\n");
    printf("     e : output EPS file\n");
    printf("     q : quit the program\n");
    printf("\n");
}

/*==================================================================*/
Public  Void   prdump(NoArg)
/*--------------------------------------------------------------------
 *  Dummy function of dump graph
 */
{
    rgdump(cprocess, cgraph);
    ildump();
    fldump();
}
