/*********************************************************************
 *  File : drgrf.c
 *
 *  Draw graphs
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "grdata.h"
#include "vrcolor.h"
#include "window.h"
#include "drgrf.h"

/*--------------------------------------------------------------------
 * Parameters
 */
#define  DREPS  0.0001

/*--------------------------------------------------------------------
 * Private functions
 */
Private Void    drarc    
        ARG((coord,coord,coord,coord,coord,coord,pindex,node,int));
Private Void    drcircle
        ARG((coord,coord,coord,coord,coord,pindex,node,int));
Private int     drcmp           ARG((node*,node*));
Private int     drcolor         ARG((pindex));
Private Void    dredges         ARG((NoArg));
Private Void    drelpos         ARG((int));
Private Void    drline          ARG((coord,coord,coord,coord,pindex,node,int));
Private Void    drndpos         ARG((NoArg));
Private Void    drtrpos         ARG((NoArg));
Private Void    drnodes         ARG((NoArg));
Private Void    drtick          ARG((coord,coord,coord,coord,pindex,node,int));
Private Void    drvisit         ARG((node,int*));
Private Void    drvst           ARG((node,int,int*));
Private Void    drvxpos         ARG((NoArg));
Private bool    dravoid         ARG((node,node,coord*,coord*));
Private int     drstree         ARG((node,node,coord[2][3],coord*));
Private Void    drvaccum        ARG((NoArg));

/*--------------------------------------------------------------------
 * Local variable
 */
Local   coord   drlm, drrm, drdm, drum;
Local   coord   drxn, dryn, drmltw, drself, drmind, dravd;
Local   coord   drx[MAXNODES], dry[MAXNODES];
Local   coord   drxo[MAXNODES], dryo[MAXNODES];
Local   coord   drlbl, drlbr, drldd, drldl, drchrh, drtlx, drtly;
Local   coord   drarr;
Local   int     drfont;
Local   int     drlvl[5][MAXNODES];        /* level of a node */
Local   int     drsrt[5][MAXNODES];        /* sort key of a node */
Local   int     dreln[3][MAXNODES];        /* list of e-lines */
Local   int     drvtx[MAXNODES];        /* list of vertices*/
Local   int     drrefs[MAXNODES];        /* list of vertices*/
Local   int     drenu[3];          /* # initial, final */ 
Local   int     drcmode = 1;       /* mode in drcmp */ 
Local   int     drblack;

Local   int     drnwind;
Local   int     nnodes;
Local   int     nextern;
Local   int     loop;

extern  int     drlabel;
extern  int     drtext;
extern  int     drvtext;
extern  int     dritext;

#define QCMP FTYPE((int (*)(Const void *, Const void *)))

/*==================================================================*/
Public  int     drinit(dest, wn, opt)
int     dest;
int     wn;
int    *opt;
/*--------------------------------------------------------------------
 *    Initialize drawing out package.
 *
 *    Argument
 *        wn : input : number of windows.
 *                    If wn <= 0 then take default value.
 */
{
    drfont = 2;

    /* margin : drlm=left, drrm=right, drdm=down, drum=up */
/*
    if(font <= 0) {
*/
        drlm = 0.2;
        drrm = 0.8;
        drdm = 0.3;
        drum = 0.7;
/*
    } else {
        drlm = 0.3;
        drrm = 0.7;
        drdm = 0.3;
        drum = 0.7;
    }
*/

    drxn = 0.015;        /* tick size for nodes */
    dryn = 0.015;

    drlbl = 0.05;       /* left of label for initial */
    drlbr = drrm + drlbl;    /* right of label for final */
    drldd = 0.020;      /* increment of external label */
    drldl = 0.010;      /* increment of edge label */

    drchrh = 0.05;      /* character height */

    drarr  = 2*drxn;    /* size of arrow */

    drmind = 2*drxn;
    dravd  = 4*drxn;

    drself = 0.10; /* size of self-loop */
    drmltw = drself;  /* width of multiple connection */

    wdinit();
    wdopen(dest, opt);

    drnwind = drsetn(wn);

    drtlx = drlm;               /* title position */
    drtly = 0.95 - drchrh;

    drblack = wdblack();

    return drnwind;
}

/*==================================================================*/
Public  int     drsetn(n)
int     n;
/*--------------------------------------------------------------------
 *  set number of graphs one the screen and draw graphs.
 */
{
    drnwind = wdsetn(n);
    return drnwind;
}

/*==================================================================*/
Public  bool    drclr(bool clr)
/*--------------------------------------------------------------------
 */
{
    return wdclr(clr);
}

/*==================================================================*/
Public  bool    drclose(dest)
int     dest;
/*--------------------------------------------------------------------
 */
{
    wdcls(dest);
    return True;
}

/*==================================================================*/
Public  int     drgraph(proc, graph, ng)
tproc   *proc;
tgraph  *graph;
long    ng;
/*--------------------------------------------------------------------
 *    Draw a graph
 */
{
    char        str[32];
    coord       chrh;

    cprocess  = proc;
    cgraph    = graph;
    nnodes    = ndnodes();
    nextern   = cprocess->nextern;
    loop      = cprocess->loop;

    wdnew();

    /* draw flame */
    /*
    wdline((coord) 0.0, (coord) 0.0, (coord) 0.0, (coord) 1.0, 
           drblack);
    wdline((coord) 0.0, (coord) 1.0, (coord) 1.0, (coord) 1.0, 
           drblack);
    wdline((coord) 1.0, (coord) 1.0, (coord) 1.0, (coord) 0.0, 
           drblack);
    wdline((coord) 1.0, (coord) 0.0, (coord) 0.0, (coord) 0.0, 
           drblack);
    */

    if(drtext > 0) {
        wdchrh(drchrh, &chrh);
        if(ng+1 == cgraph->grid)
            sprintf(str, "Graph %ld", cgraph->grid);
        else
            sprintf(str, "Graph %ld(%ld)", cgraph->grid, ng+1);
        wdtext(drtlx, drtly, str, drblack);
    }

    if(loop < 1)
        drtrpos();  /* position of tree vertieces */
    else
        drndpos();  /* position of looped vertieces */

    drnodes();  /* draw nodes */
    dredges();  /* draw edges */

    return wdend();
}

/*==================================================================*/
Private int    drcmp(nd1, nd2)
node *nd1, *nd2;
/*--------------------------------------------------------------------
 *    compare
 */
{
    int  r = 0, i;
    node n1, n2;

    for(i = 0; i < drcmode; i++) {
        n1 = *nd1;
        n2 = *nd2;
        r  = drsrt[i][n1] - drsrt[i][n2];
        if(r != 0)
            return r;
    }
    return r;
}

/*==================================================================*/
Private Void    drtrpos(NoArg)
/*--------------------------------------------------------------------
 *    Determine the position of nodes of a tree graph.
 */
{
    node  n, ul, dl, n0, n1;
    int   sroot, i, j, srl, slvl, slvl1;
    coord i1, i2, jl1, jl2, jl3;
    coord x, y, dy;
    coord lx1, lx2, ly1, ly2, ly3, di, dj, dij;
    coord cm[2][3];

    ul = dl = -1;

    for(n = 0; n < nnodes; n++)
        drvtx[n] = -1;           /* flag of visiting */

    /* determine ul(up left) e-line */
    drenu[0] = drenu[1] = drenu[2] = 0;
    for(n = 0; n < nextern; n++) {
        if(ndiorf(n) == ELINITIAL) {
            dreln[0][drenu[0]++] = n;   /* list of initial particles */
            drvtx[n] = n;
            if(ul < 0)
                ul = n;       /* node at the upper left corner */
        }
    }

    /* no initial particle. */
    if(drenu[0] == 0) {
        dreln[0][drenu[0]++] = 0;
        ul = 0;
    }

    /* distance from ul */
    if(ul >= 0) {
        drvisit(ul, drlvl[0]);
    } else {
        for(n = 0; n < nnodes; n++)
            drlvl[0][n] = 0;
    }

    /* determine position of initial e-lines */
    for(n = 0; n < nnodes; n++)
        drsrt[0][n] = drlvl[0][n];   /* sort key */
    if(drenu[0] > 1) {
        drcmode = 1;                 /* the number of sort key */
        qsort(dreln[0], drenu[0], sizeof(*(dreln[0])), QCMP drcmp);
        dl = dreln[0][drenu[0]-1];   /* node at the lower left corner */
    }

    /* list of nodes between ul and dl */
    if(dl >= 0) {
        drvisit(dl, drlvl[1]);
        sroot = 2;
        drenu[sroot] = 0;
#if DBG
        for(n = 0; n < nnodes; n++) {
            dreln[sroot][n] = -1;
        }
#endif
        srl = drlvl[0][dl]+drlvl[1][dl];

        for(n = 0; n < nnodes; n++) {
            if(ndiorf(n) == ELINITIAL)
                continue;
            if(drlvl[0][n]+drlvl[1][n] == srl) {
                i = drlvl[0][n];           /* distance from ul */
                dreln[sroot][i-1] = n;
                drenu[sroot] = max(drenu[sroot], i);
                drvtx[n] = n;
            }
        }
#if DBG
        for(i = 0; i < drenu[sroot]; i++) {
            if(drenu[sroot] < 0) {
                fprintf(Stderr,"drtrpos : illegal tree graph\n");
                for(n = 0; n < nnodes; n++)
                    fprintf(Stderr,"drlvl = (%d, %d)\n",
                            drlvl[0][n], drlvl[1][n]);
                exit(1);
            }
        }
#endif
    } else {   /* no dl */
        for(n = 0; n < nnodes; n++)
            drlvl[1][n] = drlvl[0][n];
        sroot = 0;
    }

    /* determine sub-tree */
    slvl = 0;
    Discard drstree(-1, -1, NULL, &x);
    for(i = 0; i < drenu[sroot]; i++) {
        drrefs[i] = drenu[1];  /* start of finlal particle */
        n = dreln[sroot][i];
        slvl1 = drstree(n, n, NULL, &y);
        slvl  = max(slvl, slvl1);
    }
    slvl++;

    /* determine position of the external particles */
    drelpos(0);
    drelpos(1);

    /* determine position of sub-root */
    if(sroot == 2) {
        x  = drlm + (drrm - drlm) / slvl;
        dy = (drum - drdm)/drenu[sroot];
        y  = drum - 0.5*dy;
        for(i = 0; i < drenu[sroot]; i++) {
            n = dreln[sroot][i];
            drx[n] = x;
            dry[n] = y;
            y -= dy;
        }
    }

    for(n = 0; n < nnodes; n++)
        drvtx[n] = -1;
    for(j = 0; j < drenu[0]; j++) {
        n = dreln[0][j];
        drvtx[n] = n;
    }
    if(sroot == 2) {
        for(j = 0; j < drenu[sroot]; j++) {
            n = dreln[2][j];
            drvtx[n] = n;
        }
    }
    for(i = 0; i < drenu[sroot]; i++) {
        n = dreln[sroot][i];
        n0 = dreln[1][drrefs[i]];
        if(i == drenu[sroot] - 1)
            n1 = dreln[1][drenu[1]-1];
        else
            n1 = dreln[1][drrefs[i+1]-1];

        i1 = drxo[n];
        i2 = drxo[n0];
        jl1 = dryo[n];
        jl2 = dryo[n0];
        jl3 = dryo[n1];
        lx1 = drx[n];
        lx2 = drx[n0];
        ly1 = dry[n];
        ly2 = dry[n0];
        ly3 = dry[n1];

        if(i2 == i1) {
            di = 1;
            cm[0][0] = 0;
            cm[0][1] = 0;
            cm[0][2] = lx1;
        } else {
            di = i2 - i1;
            cm[0][0] = (- lx1 + lx2)/di;
            cm[0][1] = 0;
            cm[0][2] = (- i1*lx2 + i2*lx1)/di;
        }

        if(n0 == n1) {
            di = i2 - i1;
            cm[1][0] = (- ly1 + ly2)/di;
            cm[1][1] = 0;
            cm[1][2] = (- i1*ly2 + i2*ly1)/di;
        } else {
            dj = jl3 - jl2;
            dij = di*dj;
            cm[1][0] = (- jl1*ly2 + jl1*ly3 + jl2*ly1 
                        - jl2*ly3 - jl3*ly1 + jl3*ly2)/dij;
            cm[1][1] = (- ly2 + ly3)/dj;
            cm[1][2] = (  i1*jl2*ly3 - i1*jl3*ly2 + i2*jl1*ly2
                        - i2*jl1*ly3 - i2*jl2*ly1 + i2*jl3*ly1)/dij;
        }

#if DBG
        if(fabs(-lx1+cm[0][0]*i1+cm[0][1]*jl1+cm[0][2]) > 1.0e-4)
            printf("lx1 = %f = %f\n",
                   lx1, cm[0][0]*i1+cm[0][1]*jl1+cm[0][2]);
        if(fabs(-lx2+cm[0][0]*i2+cm[0][1]*jl2+cm[0][2]) > 1.0e-4)
            printf("lx2 = %f = %f\n",
                   lx2, cm[0][0]*i2+cm[0][1]*jl2+cm[0][2]);
        if(fabs(-ly1+cm[1][0]*i1+cm[1][1]*jl1+cm[1][2]) > 1.0e-4)
            printf("ly1 = %f = %f\n",
                   ly1, cm[1][0]*i1+cm[1][1]*jl1+cm[1][2]);
        if(fabs(-ly2+cm[1][0]*i2+cm[1][1]*jl2+cm[1][2]) > 1.0e-4)
            printf("ly2 = %f = %f\n",
                   ly2, cm[1][0]*i2+cm[1][1]*jl2+cm[1][2]);
        if(fabs(-ly3+cm[1][0]*i2+cm[1][1]*jl3+cm[1][2]) > 1.0e-4)
            printf("ly3 = %f = %f\n",
                   ly3, cm[1][0]*i2+cm[1][1]*jl3+cm[1][2]);
#endif
        drstree(n, n, cm, &y);
    }

}

/*==================================================================*/
Private Void    drndpos(NoArg)
/*--------------------------------------------------------------------
 *    Determine the position of nodes.
 */
{
    node n, ul, ur, dl, dr;
    int  l, s, t, u, drvnu;

    ul = ur = dl = dr = -1;

    if(nextern < 1) {
        drvaccum();
        return;
    }

    /* determine ul(up left) e-line */
    drenu[0] = drenu[1] = 0;
    for(n = 0; n < nextern; n++) {
        if(ndiorf(n) == ELINITIAL) {
            dreln[0][drenu[0]++] = n;
            if(ul < 0)
                ul = n;
        } else {
            dreln[1][drenu[1]++] = n;
            if(ur < 0)
                ur = n;
        }
    }

    if(drenu[0] == 0 && drenu[1] == 0) {
        dreln[0][drenu[0]++] = 0;
        ul = 0;
    /*
        error("drndpos", "no external particle.\n");
        erfail();
    */
    }

    /* distance from ul */
    if(ul >= 0) {
        drvisit(ul, drlvl[0]);
    } else {
        for(n = 0; n < nnodes; n++)
            drlvl[0][n] = 0;
    }

    /* determine position of initial e-lines */
    for(n = 0; n < nnodes; n++)
        drsrt[0][n] = drlvl[0][n];
    if(drenu[0] > 1) {
        drcmode = 1;
        qsort(dreln[0], drenu[0], sizeof(*(dreln[0])), QCMP drcmp);
        dl = dreln[0][drenu[0]-1];
    }

    /* distance from dl */
    if(dl >= 0) {
        drvisit(dl, drlvl[1]);
    } else {
        for(n = 0; n < nnodes; n++)
            drlvl[1][n] = drlvl[0][n];
    }

    /* sort key of final e-line */
    for(l = 0; l < drenu[1]; l++) {
        n = dreln[1][l];
        s           = drlvl[0][n];
        t           = drlvl[1][n];
        drsrt[0][n] = s - t;
        drsrt[1][n] = s;
    }

    /* determine position of final e-lines */
    if(drenu[1] == 1) {
        ur = dreln[1][0];
    } else if(drenu[1] > 1) {
        drcmode = 2;
        qsort(dreln[1], drenu[1], sizeof(*(dreln[1])), QCMP drcmp);
        ur = dreln[1][0];
    }

    /* distance from ur */
    if(ur >= 0) {
        drvisit(ur, drlvl[2]);
    } else {
        for(n = 0; n < nnodes; n++)
            drlvl[2][n] = 0;
    }

    /* sort key of vertices */
    drvnu = nnodes - nextern;
    for(n = nextern; n < nnodes; n++) {
        drvtx[n-nextern] = n;
        s = drlvl[0][n];
        t = drlvl[1][n];
        u = drlvl[2][n];
        drsrt[0][n] = s + t;
        drsrt[1][n] = s - t;
        drsrt[2][n] = u;
    }
        
    drcmode = 3;
    qsort(drvtx, drvnu, sizeof(*drvtx), QCMP drcmp);

    if(loop > 0 && drvnu > 2 && 
       drsrt[0][drvtx[0]] == drsrt[0][drvtx[drvnu-1]]) {
        /* fine tuning of positions of final external particles */
        if(drenu[1] > 1)
            dr = dreln[1][drenu[1]-1];

        /* distance from dr */
        if(dr >= 0) {
            drvisit(ur, drlvl[4]);
        } else {
            for(n = 0; n < nnodes; n++)
                drlvl[4][n] = 0;
        }
        /* sort key of vertices */
        for(n = 0; n < nnodes; n++)
            drsrt[4][n] = drlvl[4][n];
        for(n = nextern; n < nnodes; n++) {
            drvtx[n-nextern] = n;
            drsrt[0][n] += - drsrt[3][n] - drsrt[4][n];
        }
        drcmode = 3;
        qsort(drvtx, drvnu, sizeof(*drvtx), QCMP drcmp);
    }

    /* fine tuning of positions of final external particles */
    drelpos(0);
    drelpos(1);
    drvxpos();
}

/*==================================================================*/
Private int     drstree(root, sroot, cm, ylvl)
node    root;
node    sroot;
coord   cm[2][3];
coord  *ylvl;
/*--------------------------------------------------------------------
 *    drvtx : flag of visit
 */
{
    node        i, j;
    int         lvl, lvl1, level, ny;
    static int  ylevel=0;
    coord       yl;

    if(root < 0) {
        ylevel = 0;
        return -1;
    }

    drvtx[root] = 0;   /* temporal visit flag */
    if(root < nextern && root != sroot) {    /* external particle */
        if(ndiorf(root) == ELINITIAL && root != sroot) {
            fprintf(Stderr, "drstree: unexpected initial particle\n");
            fprintf(Stderr, "node = %d\n", root);
        }
        if(cm == NULL && ndiorf(root) != ELINITIAL) {
            dreln[1][drenu[1]++] = root;
        }
        level = 0;
        *ylvl  = ylevel--;
    } else {
        lvl = 0;
        *ylvl = 0;
        ny    = 0;
        for(j = 0; j < ndlegs(root); j++) {
            i = ndnextn(root, j);
            if(drvtx[i] < 0) {
                lvl1 = drstree(i, sroot, cm, &yl);
                lvl  = max(lvl, lvl1);
                *ylvl += yl;
                ny++;
            }
        }
        level = lvl + 1;
        if(ny > 0)
            *ylvl /= ny;
    }
    if(cm == NULL) {
        drxo[root] = - level;
        dryo[root] = *ylvl;
    } else {
        drx[root] = cm[0][0]*drxo[root] 
                  + cm[0][1]*dryo[root]
                  + cm[0][2];
        dry[root] = cm[1][0]*drxo[root] 
                  + cm[1][1]*dryo[root]
                  + cm[1][2];
    }
    drvtx[root] = level;   /* visit */
    return level;
}

/*==================================================================*/
Private Void    drvst(root, level, nlvl)
node root;
int  level;
int  nlvl[];
/*--------------------------------------------------------------------
 *   main part of drvisit
 */
{
    node        i, j;
    int         lvl = level+1;

    for(j = 0; j < ndlegs(root); j++) {
        i = ndnextn(root, j);
        if(nlvl[i] < 0 || nlvl[i] > level) {
            nlvl[i] = level;   /* visit */
            drvst(i, lvl, nlvl);
        }
    }
}

/*==================================================================*/
Private Void    drvisit(root, nlvl)
node root;
int  nlvl[];
/*--------------------------------------------------------------------
 */
{
    node        n;

    for(n = 0; n < nnodes; n++)
        nlvl[n] = -1;

    nlvl[root] = 0;

    drvst(root, 1, nlvl);
}

/*==================================================================*/
Private Void    drelpos(iorf)
int     iorf;
/*--------------------------------------------------------------------
 *  Determine position of e-lines.
 */
{
    int  i;
    node n;
    coord y, dy;
    coord xpos;

    if(drenu[iorf] < 1)
        return;

    if(iorf == 0)
        xpos = drlm;
    else
        xpos = drrm;

    /* dry[i] = (i - base)*dy + drum */
    if(drenu[iorf] == 1) {
        y  = (drum + drdm) * 0.5;
        dy = 1.0;
    } else {
        y  = drum;
        dy = (drum - drdm) / (drenu[iorf] - 1);
    }

    for(i = 0; i < drenu[iorf]; i++) {
        n = dreln[iorf][i];
        drx[n] = xpos;
        dry[n] = y;
        y -= dy;
    }
}

/*==================================================================*/
Private Void    drvxpos(NoArg)
/*--------------------------------------------------------------------
 *  Determine position of vertices.
 */
{
    int  numb, l;
    int  i, lvlo, lvln, ny;
    node n;
    coord y, dy;
    coord x, dx;
    int   drlvc[MAXNODES];        /* # nodes of a level */

    if(nextern == 0) {
        drvaccum();
        return;
    }
    if((numb = nnodes - nextern) < 1)
        return;

    lvlo = -9999;   /* the last level */
    lvln = 0;       /* # of x-position */
    ny = 0;
    for(i = 0; i < numb; i++) {
        n  = drvtx[i];
        if(drsrt[0][n] != lvlo) {
            drlvc[lvln++] = i;
            lvlo = drsrt[0][n];
            ny = 0;
        }
        dry[n] = ny++;
    }
    drlvc[lvln] = numb;

    dx = (drrm - drlm) / (lvln + 1);
    x  = drlm + dx;
    for(l = 0; l < lvln; l++) { 
        dy = (drum - drdm)/(drlvc[l+1] - drlvc[l]);
        y  = drum - 0.5*dy;
        for(i = drlvc[l]; i < drlvc[l+1]; i++) {
            n = drvtx[i];
            drx[n] = x;
            dry[n] = y;
            y -= dy;
        }
        x += dx;
    }
}

/*==================================================================*/
Private Void    drvaccum(NoArg)
/*--------------------------------------------------------------------
 *  Determine position of vertices.
 */
{
    node n;
    coord dx, y1, t1, th, t0;
    coord dy, x1;
    coord pi;

    if(nnodes == 1) {
        drx[0] = 0.5;
        dry[0] = 0.5;
        return;
    }
    pi = 4.0*atan(((double) 1.0));
    x1 =   0.5;
    y1 =   0.5;
    t1 = 2*pi/((coord) nnodes);
    if(nnodes%2 == 0) {
        t0 = pi - 0.5*t1;
    } else {
        t0 = pi - 0.0;
    }
    th = 0.0;
    for(n = 0; n < nnodes; n++) { 
        if(n%2 == 0)
            th = t0 - t1 * ((coord) ((int)(n/2)));
        else
            th = t0 + t1 * ((coord) ((int)((n+1)/2)));
        dx = 0.25*cos(th);
        dy = 0.25*sin(th);
        drx[n] = x1 + dx;
        dry[n] = y1 + dy;
    }
}

/*==================================================================*/
Private Void    drnodes(NoArg)
/*--------------------------------------------------------------------
 *  Draw nodes.
 */
{
    int         i, l;
    coord       symx, symy;
    pindex      ptcl;
    string      sym;
    char        str[32];
    coord       dr;
    coord       chrh;
/*
    string      iorf;
 */

    for(i = 0; i < nnodes; i++) {
        if(ndloop(i) < 1) {
            wdline(drx[i],      dry[i]+dryn, drx[i]+drxn, dry[i],
                   VR_RED);
            wdline(drx[i]+drxn, dry[i],      drx[i],      dry[i]-dryn,
                   VR_RED);
            wdline(drx[i],      dry[i]-dryn, drx[i]-drxn, dry[i],
                   VR_RED);
            wdline(drx[i]-drxn, dry[i],      drx[i],      dry[i]+dryn,
                   VR_RED);
        } else {
            dr = 2*dryn;
            for(l = 0; l < ndloop(i); l++) {
                wdcircle(drx[i], dry[i], dr, VR_RED);
                dr += dryn;
            }
        }

        if(i < nextern) {
            ptcl = ndname(i);
            symy = dry[i] - drldl;
            if(drx[i] < 0.5) {
                symx = drlbl;
/*
                iorf = "in";
*/
            } else {
                symx = drlbr;
/*
                iorf = "fi";
*/
            }
            if(ndiorf(i) == ELINITIAL)
                ptcl *= -1;
            if(ptcl >= 0) {
                sym = ptable[ptcl]->psym[0];
            } else {
                sym = ptable[-ptcl]->psym[1];
            }
            if(drtext > 0) {
/*
                sprintf(str, "%d %s", i, iorf);
*/
                wdchrh(drchrh, &chrh);
                sprintf(str, "%d", i);
                wdtext(symx, (coord)(symy + chrh*0.5), str, drblack);
                sprintf(str, "%.16s", sym);
                wdtext(symx, (coord)(symy - chrh*0.6), str, drblack);
            }
        } else {
            if(drtext > 0 && drvtext > 0) {
                wdchrh(((coord) 0.5*drchrh), &chrh);
                sprintf(str, "%d", i);
                wdtext(drx[i]+drxn, dry[i]+dryn, str, drblack);
            }
        }
    }
}

/*==================================================================*/
Private Void    dredges(NoArg)
/*--------------------------------------------------------------------
 *  Draw edges.
 */
{
    int         i, j, m, k, l, il, kk;
    coord       y, x, r, dx, dy, rx, ry, dr, lx, ly, rw;
    pindex      ptcl[MAXNODES];
    int         jj[MAXNODES];

    /* not self loop */
    for(i = 0; i < nnodes-1; i++) {
        for(l = 0; l < ndlegs(i); l++) {
            il = ndiline(i, l);
            j  = ndnextn(i, l);
            jj[l] = j;
         /* ???
            if(i == j)
                continue;
         */
            if(i > j)
                continue;

            if((m = tledge(i, j, ptcl)) < 1) {
                ;
            } else if(m == 1) {
                if(dravoid(i, j, &x, &y)) {
                    drarc(drx[i], dry[i], x, y, drx[j], dry[j],
                          ptcl[0], i, il);
                } else {
                    drline(drx[i], dry[i], drx[j], dry[j],
                           ptcl[0], i, il);
                }
            } else {
                kk = 0;
                for(k = 0; k < l; k++)
                    if(jj[k] == j)
                        kk++;

                /* vector i --> j */
                rx = drx[j] - drx[i];
                ry = dry[j] - dry[i];
                r  = sqrt(rx*rx + ry*ry);
                if(r < 1.0e-4)
                    r = 1.0e-4;
                rx /= r;
                ry /= r;
    
                /* width between edges */
                rw = min(drmltw, r);
    
                dx = - ry*rw/(m-1);
                dy =   rx*rw/(m-1);
                x  = 0.5*(drx[j] + drx[i] + ry*rw);
                y  = 0.5*(dry[j] + dry[i] - rx*rw);
    
                /* arc passes (x + k*dx, y + k*dy) */
                x += kk*dx;
                y += kk*dy;
                drarc(drx[i], dry[i], x, y, drx[j], dry[j], ptcl[kk], i, il);
            }
        }
    }

    /* self loop */
    for(i = 0; i < nnodes; i++) {
        if((m = tledge(i, i, ptcl)) < 1) {
            continue;
        }
        m /= 2;  /* the number of internal lines */

        x = y = lx = ly = 0.0;
        for(j = 0; j < nnodes; j++) {
            if(j != i &&  tlconn(i, j)> 0) {
                lx = drx[j] - drx[i];
                ly = dry[j] - dry[i];
                x += lx;
                y += ly;
            }
        }
        r  = sqrt(x*x + y*y);
        if(r < 0.00001) {
            r = sqrt(lx*lx + ly*ly);
            if(r < 0.00001) {
                dx = 0;
                dy = 1;
            } else {
                dx = -ly/r;
                dy =  lx/r;
            }
        } else {
            dx = - x/r;
            dy = - y/r;
        }

        r  = 0.5*drself;
        dr = r/(m + 1);
        x  = drx[i] + dx*r;
        y  = dry[i] + dy*r;
        /* arc passes (x + k*dx, y + k*dy) */
        for(k = 0; k < m; k++) {
            /* circle is better than arc */
            drcircle(drx[i], dry[i], x, y, r, ptcl[k], i, -1);
            x -= dx*dr;
            y -= dy*dr;
            r -= dr;
        }
    }
}

/*==================================================================*/
Private bool    dravoid(n0, n1, ax, ay)
node    n0, n1;
coord   *ax, *ay;
/*--------------------------------------------------------------------
 *  Test if line (r0 - r1) is overlapping with another node.
 *  If so, (*ax, *ay) is a point, on which avoiding arc should pass.
 */
{
    node        n;
    coord       mx, my, r, minx, maxx, miny, maxy;

#if DBG
    if(dbg > drdbg)
        printf("dravoid = [%d(%f, %f), %d(%f, %f)] \n",
               n0, drx[n0], dry[n0], n1, drx[n1], dry[n1]);
#endif

    /* (mx, my) is a normalized normal vector to (r1 - r0) */
    my =   drx[n1] - drx[n0];
    mx = -(dry[n1] - dry[n0]);
    r  = sqrt(mx*mx + my*my);
    if(r < DREPS)                /* very near */
        return False;
    mx /= r;
    my /= r;

    /* size of box including r0 and r1 */
    if(drx[n0] < drx[n1]) {
        minx = drx[n0] - drmind;
        maxx = drx[n1] + drmind;
    } else {
        minx = drx[n1] - drmind;
        maxx = drx[n0] + drmind;
    }
    if(dry[n0] < dry[n1]) {
        miny = dry[n0] - drmind;
        maxy = dry[n1] + drmind;
    } else {
        miny = dry[n1] - drmind;
        maxy = dry[n0] + drmind;
    }

#if DBG
    if(dbg > drdbg)
        printf(" box : (%f, %f) - (%f, %f)\n", minx, miny, maxx, maxy);
#endif

    /* check overlapping */
    for(n = nextern; n < nnodes; n++) {

#if DBG
        if(dbg > drdbg)
            printf("\n n = %d ", n);
#endif
        if(n == n0 || n == n1)
            continue;

        /* out of box */
        if(drx[n] < minx || drx[n] > maxx)
            continue;
        if(dry[n] < miny || dry[n] > maxy)
            continue;

#if DBG
        if(dbg > drdbg)
            printf(" distance ");
#endif

        /* distance form the line (r0 - r1) */
        r = mx*(drx[n] - drx[n0]) + my*(dry[n] - dry[n0]);
        if(r < 0)
            r = -r;

        if(r > drmind)
            continue;

        /* avoiding displacement */
        mx *= dravd;
        my *= dravd;

#if DBG
        if(dbg > drdbg)
            printf(" m = (%f, %f) ", mx, my);
#endif
        /* direction */
        if(mx > - DREPS && mx < DREPS) {
            if(my*(dry[n] - 0.5) < 0) {
                mx = - mx;
                my = - my;
            }
        } else {
            if(mx*(drx[n] - 0.5) < 0) {
                mx = - mx;
                my = - my;
            }
        }
#if DBG
        if(dbg > drdbg)
            printf(" => (%f, %f) \n", mx, my);
#endif

        *ax = drx[n] + mx;
        *ay = dry[n] + my;

#if DBG
        if(dbg > drdbg)
            printf(" YES, n = %d (%f, %f)\n", n, *ax, *ay);
#endif

        return True;

    }
#if DBG
    if(dbg > drdbg)
        printf(" NO\n");
#endif
    return False;
}

/*==================================================================*/
Private Void    drline(dtx, dty, dsx, dsy, ptcl, nd, il)
coord dtx, dty;
coord dsx, dsy;
pindex ptcl;
node   nd;
int    il;
/*--------------------------------------------------------------------
 *  Draw edge connectiong from (dtx, dty) to (dsx, dsy) with particle
 *  ptcl.
 */
{
    coord cx, cy;

    wdline(dtx, dty, dsx, dsy, drcolor(ptcl));

    cx = (dtx + dsx) * 0.5;
    cy = (dty + dsy) * 0.5;

    drtick(cx, cy, dsx-dtx, dsy-dty, ptcl, nd, il);
}

/*==================================================================*/
Private Void    drcircle(dtx, dty, cx, cy, r, ptcl, nd, il)
coord dtx, dty;      /* node */
coord  cx,  cy;      /* coneter of the circle */
coord  r;            /* radius */
pindex ptcl;
node   nd;
int    il;
/*--------------------------------------------------------------------
 */
{
    coord tx, ty;
    coord dx, dy;

    wdcircle(cx, cy, r, drcolor(ptcl));

    tx = 2 * cx - dtx;
    ty = 2 * cy - dty;
    dx = - cy + dty;
    dy =   cx - dtx;

    drtick(tx, ty, dx, dy, ptcl, nd, il);
}

/*==================================================================*/
Private Void    drarc(lx0, ly0, lx1, ly1, lx2, ly2, ptcl, nd, il)
coord lx0, ly0;
coord lx1, ly1;
coord lx2, ly2;
pindex ptcl;
node   nd;
int    il;
/*--------------------------------------------------------------------
 */
{
    coord d, cx, cy, r, r1, ddx, ddy;
    coord rs0, rs1, rs2;
    coord dx, dy;
    int   color;

    color = drcolor(ptcl);

    d = ((lx0 - lx1)*(ly2 - ly1) - (lx2 - lx1)*(ly0 - ly1));
    if (fabs(d) < 0.000001) {

        wdline(lx0, ly0, lx1, ly1, color);
        wdline(lx1, ly1, lx2, ly2, color);
        ddx = lx2 - lx0;
        ddy = ly2 - ly0;
        if(ddx*ddx + ddy*ddy < 0.000001) {
            ddx = ly1 - ly0;
            ddy = lx0 - lx1;
        }
        if(ddx*ddx + ddy*ddy < 0.000001) {
            return;
        }
        drtick(lx1, ly1, ddx, ddy, ptcl, nd, il);

    } else {

        wdarc(lx0, ly0, lx1, ly1, lx2, ly2, color);

        rs0 = lx0*lx0 + ly0*ly0;
        rs1 = lx1*lx1 + ly1*ly1;
        rs2 = lx2*lx2 + ly2*ly2;
        cx  = (  (ly2 - ly1)*(rs0 - rs1) - (ly0 - ly1)*(rs2 - rs1))/(2.0*d);
        cy  = (- (lx2 - lx1)*(rs0 - rs1) + (lx0 - lx1)*(rs2 - rs1))/(2.0*d);
        r   = sqrt((lx0 - cx)*(lx0 - cx) + (ly0 - cy)*(ly0 - cy));

        dx =   ly0 - ly2;
        dy = - lx0 + lx2;
        r1 = sqrt(dx*dx + dy*dy);

        if ((lx1 - cx)*dx + (ly1 - cy)*dy > 0) {
            dx = cx + dx*r/r1;
            dy = cy + dy*r/r1;
        } else {
            dx = cx - dx*r/r1;
            dy = cy - dy*r/r1;
        }

        ddx = lx2 - lx0;
        ddy = ly2 - ly0;
        if(ddx*ddx + ddy*ddy < 0.000001) {
            ddx = ly1 - ly0;
            ddy = lx0 - lx1;
        }
        if(ddx*ddx + ddy*ddy < 0.000001) {
            return;
        }
        drtick(dx, dy, ddx, ddy, ptcl, nd, il);
    }
}

/*==================================================================*/
Private Void    drtick(x, y, dx, dy, ptcl, nd, il)
coord   x,   y;      /* position */
coord  dx,  dy;      /* direction */
pindex ptcl;
node   nd;
int    il;
/*--------------------------------------------------------------------
 */
{
    coord  r, nx, ny, chrh;
    coord  xl, yl;
    int    color;
    int    dir;
    string sym;
    char   str[32];

    if(dritext > 0) {
        if(iltable[il].ilnodes[0] == 0)
            dir = -1;
        else
            dir =  1;

        if(ptcl < 0)
            ptcl = -ptcl;
    } else {
        if(ptcl == PNULL) {
            dir = 0;
        } else if(ptcl > 0) {
            if(ptable[ptcl]->antip > 0 && ptable[ptcl]->ptype != P_MAJORANA)
                dir = 0;
            else
                dir = 1;
        } else {
            dir = -1;
            ptcl  = -ptcl;
        }
    }

    if(dir < 0) {
        dx    = -dx;
        dy    = -dy;
    }

    r = sqrt(dx*dx + dy*dy);
    dx = drarr*dx/r;
    dy = drarr*dy/r;

    if(dx > 0) {
        nx = - 0.5 * dy;
        ny =   0.5 * dx;
    } else {
        nx =   0.5 * dy;
        ny = - 0.5 * dx;
    }

    /* draw arrow */
    if(dir != 0) {

        color = VR_BLACK;

        wdline(x, y, x + dx + nx, y + dy + ny, color);
        wdline(x, y, x + dx - nx, y + dy - ny, color);
    }

    /* with particle name */
    if(drtext > 0 && drlabel > 0) {
        wdchrh(((coord) 0.5*drchrh), &chrh);
        sym = ptable[ptcl]->psym[0];
        xl  = x + 2*chrh*nx/drarr;
        yl  = y + 2*chrh*ny/drarr;
        
        wdtext(xl, yl, sym, drblack);
    }

    /* with internal line number */
    if(dritext > 0) {
        wdchrh(((coord) 0.5*drchrh), &chrh);
        sprintf(str,"%d", il);
        xl  = x - 2*chrh*nx/drarr;
        yl  = y - 2*chrh*ny/drarr;
        
        wdtext(xl, yl, str, drblack);
    }
}

/*==================================================================*/
Private int     drcolor(ptcl)
pindex  ptcl;
/*--------------------------------------------------------------------
 */
{
    int pt;

    if(ptcl == PNULL)
        return drblack;

    ptcl = (ptcl < 0) ? -ptcl : ptcl;
    pt = ptable[ptcl]->ptype;

    if(pt == P_SCALAR)
        return VR_GREEN;
    else if(pt == P_DIRAC)
        return VR_BLUE;
    else if(pt == P_VECTOR)
        return VR_MAGENTA;
    else
        return VR_CYAN;
}

/*==================================================================*/
Public  Void    drctext(str)
string  str;
/*--------------------------------------------------------------------
 */
{
    wdtext(0.05, 0.5, str, drblack);
}
