/*********************************************************************
  File : tread.c

  Entry for testing particle table
*/

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

#define  DEBUG
#define  DEFGLOBAL
#include "grdata.h"
#include "config.h"
#include "token.h"
#include "cfnbase.h"
#include "cfnconv.h"
#include "gensc.h"

/*--------------------------------------------------------------------
 * Local types
 */
#define SL_PLUS    (1)
#define SL_MINUS   (-1)
#define SL_BEGIN   (2)
#define SL_NEXT    (3)
#define SL_END     (-2)

typedef struct {
    int sgn;
    int arg[2];
} tsel;

/*--------------------------------------------------------------------
 * Local functions
 */
Private Void   grcarg           ARG((int, char**));
Private string grdword          ARG((NoArg));
Private bool   grdline          ARG((tsel*));
Private Void   grdenter         ARG((int,int,int));
Private bool   grdnext          ARG((NoArg));
Private bool   grdtst           ARG((tgraph*));
Private Void   grddump          ARG((NoArg));

/*--------------------------------------------------------------------
 * Local variables
 */
Local   char    fgname[] = "out.grf";
Local   char    fngrf[MAXPATHLEN];

Local   char    foname[] = "out1.grf";
Local   char    fogrf[MAXPATHLEN];
FILE    *fps;
Local   char    fsname[] = "stdin";
Local   char    fsgrf[MAXPATHLEN];

Local   int     proct[MAXPROCS];
Local   int     procl[MAXPROCS];
Local   int     procg[MAXPROCS];

/* initial size of selections */
#define INITSEL 100

Local tsel *sels;
Local int   nsels;

#define MAXSTRLEN 256

#define ilpartl(l,lg)     \
        ndparticle(iltable[l].ilnodes[lg],iltable[l].illegs[lg])
#define ilpart(l)     abs(ilpartl(l, 0))

/*==================================================================*/
Public  int    main(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
    Main routine for graph drawin
--------------------------------------------------------------------*/
{
    FILE     *fpout;
    int       nprocs, ngrf;
    int       gr;
    int       pc, cnprc, nextern, nilines, cngraph;
    tgraph   *graph = NULL;
    tproc    *proc;

    colorv = 1;

#if DBG
    trace   = 0;
#endif

#if DBG
    /* debugging flags for modules.
     * if(dbg >= xxdbg) then enable debugging
     */
    dbg = 0;
    rgdbg = 2;  /* debug flag for rdgrf.c */
    pvdbg = 2;  /* debug flag for rdptbl.c */
    fldbg = 2;  /* debug flag for flgrf.c */
    ildbg = 2;  /* debug flag for ilgrf.c */
    pmdbg = 2;  /* debug flag for matchv.c */
    rgdbg = 2;  /* debug flag for rdgrf.c */
    pvdbg = 2;  /* debug flag for rdptbl.c */
    rddbg = 2;  /* debug flag for token.c */
#endif

    grcarg(argc, argv);   /* analyze arguments */

    if(strcmp(fsgrf, "stdin") == 0) {
        fps = stdin;
    } else if((fps = fopen(fsgrf, "r")) == NULL) {
        fprintf(Stderr, "*** input file open error \"%s\"\n", fsgrf);
        exit(FAIL);
    }

    if((fpout = fopen(fogrf, "w")) == NULL) {
        fprintf(Stderr, "*** cannot open \"%s\"\n", fogrf);
        exit(FAIL);
    }

    /* CF */   
    if((gcolgr = (cftgraph *) malloc(sizeof(cftgraph))) == NULL) {    
         fprintf(Stderr, "*** gcolgr: memory allocation failed\n");
         exit(1);
    }
    if((gcbase = (cftbase *) malloc(sizeof(cftbase))) == NULL) {
         fprintf(Stderr, "*** gcbase: memory allocation failed\n");
         exit(1);
    }

    genct = True;
    nprocs = dasave(fngrf, procl, procg);
    for(pc = 0; pc < nprocs; pc++) {
        proct[pc] = pc;
    }

    fofinit(fpout);

    for(pc = 0; pc < nprocs; pc++) {
        fprintf(stderr, "%% +++ process = %d/%d +++\n", pc+1, nprocs);

        cnprc    = proct[pc];       /* process number */
        proc     = dagetp(cnprc);     /* process data */
        cprocess = proc;
        nextern  = cprocess->oextern;

        if(!grdnext()) {
            fprintf(stderr, "more than %d processes\n", nprocs+1);
        }

        ngrf = 0;
        for(gr = 0; gr < procg[pc]; gr++) {
            fprintf(stderr, "%% +++ graph = %d/%d (%d/%d) +++\n",
                    gr+1, procg[pc], pc+1, nprocs);

            cngraph = gr;
            graph   = daread(cnprc, cngraph);
            cgraph  = graph;
            ilmom();
            nilines = cgraph->nilines;
            flget(cgraph);

            if(grdtst(graph)) {
                if(ngrf == 0) {
                    cngraph = 0;
                    cgraph  = daread0(cnprc, cngraph);
                    cgraph  = graph;
                    ilmom();
                    nilines = cgraph->nilines;
                    flget(cgraph);

                    fofproc(fpout, proc, cgraph, NULL);

                    cngraph = gr;
                    cgraph  = daread0(cnprc, cngraph);
                    cgraph  = graph;
                    ilmom();
                    nilines = cgraph->nilines;
                    flget(cgraph);
                }
                ngrf++;
                fofgraph(fpout, proc, graph);
                foflines(fpout);
                fprintf(stderr, "grid=%ld : included\n", graph->grid);
            } else {
                fprintf(stderr, "grid=%ld : excluded\n", graph->grid);
            }
        }
        if(ngrf != 0) {
            fofpend(fpout, procg[pc]);
        }
    }
    fofend(fpout, nprocs+1);

    daclose();

    return 0;
}

/*==================================================================*/
Private Void   grcarg(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
*/
{
    int    c;
    string cmd;

    cmd     = argv[0];

    strcpy(fogrf, foname);
    strcpy(fsgrf, fsname);

    while(--argc > 0 && (*++argv)[0] == '-') {
        c = (*argv)[1];
        if(c == 'i') {
            --argc;
            ++argv;
            strcpy(fsgrf, *argv);
        } else {
            while ((c = *++argv[0]) != 0) {
                switch (c) {
                case 'h':
                case '?':
                    fprintf(Stderr, "\n\n");
                    fprintf(Stderr, "Syntax %s ", cmd);
                    fprintf(Stderr, "[-i sel_file] [[input_file] ");
                    fprintf(Stderr, "ouput_file]\n");
                    fprintf(Stderr, "\nOption\n");
                    fprintf(Stderr, "    -h                  ");
                    fprintf(Stderr, ": more info\n");
                    fprintf(Stderr, "\nDefault file name is:\n");
                    fprintf(Stderr, "     Graph_File_name = %s\n", fgname);
                    exit(0);
                }
            }
        }
    }

    /* input process file name */
    if(argc == 0) {
        strcpy(fngrf, fgname);
        strcpy(fogrf, foname);
    } else if(argc == 1) {
        strcpy(fngrf, fgname);
        strcpy(fogrf, argv[0]);
    } else if(argc == 2) {
        strcpy(fngrf, argv[0]);
        strcpy(fogrf, argv[1]);
    } else {
        fprintf(Stderr, "Too many arguments\n");
        exit(FAIL);
    }
}

/*==================================================================*/
Public  Void   prdump(NoArg)
/*--------------------------------------------------------------------
    Dump.
*/
{
    rgdump(cprocess, cgraph);
}

/*==================================================================*/
Private string  grdword(NoArg)
/*--------------------------------------------------------------------
*/
{
    static int  ch = ' ';
    static char s[MAXSTRLEN];
    static int  left = False;
    int i;

    if(ch == EOF) {
        ch = ' ';
        return NULL;
    }

    /* skip space */
    while((left || (ch = getc(fps)) != EOF) && isspace(ch) && ch != '\n')
        left = False;

    left = False;

    /* comment --> '\n' */
    if(ch == '%') {
        while((ch = getc(fps)) != EOF && ch != '\n')
            ;
    }

    if(ch == EOF) {
        ch = ' ';
        return NULL;
    }

    if(isdigit(ch)) {
        i = 0;
        do {
            s[i++] = ch;
        } while(isdigit(ch = getc(fps)));
        s[i] = '\0';
        left = True;
        return s;
    } else if(isalpha(ch)) {
        i = 0;
        do {
            s[i++] = tolower(ch);
        } while(isalpha(ch = getc(fps)));
        s[i] = '\0';
        left = True;
        return s;
    } else {
        s[0] = ch;
        s[1] = '\0';
        return s;
    }
}

/*==================================================================*/
Private bool   grdline(sl)
tsel    *sl;
/*--------------------------------------------------------------------
*/
{
    string s;

    sl->sgn   = SL_PLUS;
    sl->arg[0] = 0;
    sl->arg[1] = 0;

    while((s = grdword()) != NULL) {
        if(s[0] == '\n') {
            continue;
        } else if(isalpha(s[0])) {
            if(strcmp(s, "end") == 0) {
                sl->sgn = SL_END;
            } else if(strcmp(s, "next") == 0) {
                sl->sgn = SL_NEXT;
            }

            /* skip until eof */
            while((s = grdword()) != NULL) {
                if(s[0] == '\n')
                    break;
            }
            if(sl->sgn == SL_PLUS) {
                fprintf(Stderr, "*** Illegal word \"%s\"\n", s);
                return False;
            }
            return True;
        } else if(isdigit(s[0])) {
            sl->arg[0] = atol(s);
            if((s = grdword()) == NULL) {
                fprintf(Stderr, "*** Unexpected EOF\n");
                return False;
            } else if(s[0] == '\n') {
                sl->arg[1] = sl->arg[0];
            } else {
                sl->arg[1] = atol(s);
                while((s = grdword()) != NULL) {
                    if(s[0] == '\n')
                        break;
                }
            }
            return True;
        } else if(s[0] == '+') {
        } else if(s[0] == '-') {
            sl->sgn = SL_MINUS;
        } else {
            while((s = grdword()) != NULL) {
                if(s[0] == '\n')
                    break;
            }
            fprintf(Stderr, "*** Illegal symbol \"%s\"(%d)\n", s, s[0]);
            return False;
        }
    }
    return False;
}

/*==================================================================*/
Private Void   grdenter(sgn, arg0, arg1)
int     sgn;
int     arg0;
int     arg1;
/*--------------------------------------------------------------------
*/
{
    static int maxnsels = 0;
    tsel   *ts;
    int    i;

    switch (sgn) {
        case SL_BEGIN :
            nsels = 0;
            break;

        case SL_PLUS :
        case SL_MINUS :
            if(nsels >= maxnsels) {
                if(maxnsels < 1) {
                    maxnsels = INITSEL;
                    ts = NULL;
                    if(nsels > 0) {
                        fprintf(Stderr, "*** grdenter : nsels = %d\n", nsels);
                        exit(FAIL);
                    }
                } else {
                    maxnsels *= 2;
                    ts = sels;
                }
                sels = (tsel*) malloc(maxnsels*sizeof(tsel));
                if(sels == NULL) {
                    fprintf(Stderr, "*** grdenter : memory %d\n", maxnsels);
                }
                for(i = 0; i < nsels; i++) {
                    sels[i].sgn   = ts[i].sgn;
                    sels[i].arg[0] = ts[i].arg[0];
                    sels[i].arg[1] = ts[i].arg[1];
                }
                if(ts != NULL)
                    free(ts);
            }
            sels[nsels].sgn   = sgn;
            sels[nsels].arg[0] = arg0;
            sels[nsels].arg[1] = arg1;
            nsels++;
            break;

        default :
            fprintf(Stderr, "*** illegal option to grdenter(%d,%d,%d)\n",
                    sgn, arg0, arg1);
    }
}

/*==================================================================*/
Private Void   grddump(NoArg)
/*--------------------------------------------------------------------
*/
{
    int i;

    fprintf(Stderr, "+++ grddump :  nsels=%d\n", nsels);
    for(i = 0; i < nsels; i++) {
        fprintf(Stderr, "%d: %d %d %d\n",
               i, sels[i].sgn, sels[i].arg[0], sels[i].arg[1]);
    }
}

/*==================================================================*/
Private bool   grdnext(NoArg)
/*--------------------------------------------------------------------
*/
{
    int first = True;
    static int next = SL_PLUS;
    tsel sl;

    if(next == SL_END)
        return False;

    grdenter(SL_BEGIN, 0, 0);

    while(grdline(&sl)) {
        next = SL_PLUS;
        if(sl.sgn == SL_END || sl.sgn == SL_NEXT) {
            next = sl.sgn;
            sl.sgn = SL_MINUS;
            sl.arg[0] = 1;
        }
        if(first) {
            first = False;
            if(sl.arg[0] != 0) {
                if(sl.sgn == SL_PLUS)
                    grdenter(SL_MINUS, 0, 0);
                else
                    grdenter(SL_PLUS, 0, 0);
            }
        }
        if(next != SL_PLUS)
            break;
        grdenter(sl.sgn, sl.arg[0], sl.arg[1]);
    }

    grddump();

    return True;
}

/*==================================================================*/
Private bool   grdtst(graph)
tgraph  *graph;
/*--------------------------------------------------------------------
*/
{
    int grid = graph->grid;
    int st   = True;
    int i;

    for(i = 0; i < nsels; i++) {
        if(sels[i].sgn == SL_PLUS) {
            if(sels[i].arg[0] == 0)
                st = True;
            else if(sels[i].arg[0] <= grid && grid <= sels[i].arg[1])
                st |= True;
        } else if(sels[i].sgn == SL_MINUS) {
            if(sels[i].arg[0] == 0)
                st = False;
            else if(sels[i].arg[0] <= grid && grid <= sels[i].arg[1])
                st &= False;
        } else {
            fprintf(Stderr, "*** grdtst : illegal sgn %d\n", sels[i].sgn);
            exit(FAIL);
        }
    }
    return st;
}
