/*********************************************************************
 *  File : plot.c
 *
 *  Main program for graph plotting
 *
 *  Entry points:
 *    1. main   : entry of the program called by OS.
 *    2. prdump : dump the current graph.
 *
 *                                                 by Kaneko Toshiaki
 */

#define  DEFGLOBAL
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "grdata.h"
#include "args.h"
#include "vr.h"
#include "config.h"
#include "token.h"
#include "pset.h"
#include "window.h"
#include "version.h"
#include "drgrf.h"

/*--------------------------------------------------------------------
 * Global variables
 */
        int     drlabel = 1;
        int     drcomp  = 0;
        int     drtext  = 1;
        int     drvtext = 0;
        int     dritext = 0;

/*--------------------------------------------------------------------
 * Private functions
 */
Private int    drcarg        ARG((int, char**));
Private Void   drhelp        ARG((string));

/*--------------------------------------------------------------------
 * Local variable
 */
#ifdef FACOM
static  char    finame[]={"grf"};
#else
static  char    finame[]={"out.grf"};   /* default input file */
#endif

Local   string  fngrf = NULL;

Local   int     drwinds = 0;
Local   int     drhlp   = 0;
Local   int     dlabel  = 0;
Local   int     dcomp   = 0;
Local   int     dtext   = 0;
Local   int     ditext  = 0;
Local   int     dvtext  = 0;
#if DBG
Local   int     ddbg    = 0;
#endif

Local   topt    options[] = {
    {"c", ARFLAG, (int*) &dcomp},
    {"h", ARFLAG, (int*) &drhlp},
    {"l", ARFLAG, (int*) &dlabel},
    {"i", ARFLAG, (int*) &ditext},
    {"v", ARFLAG, (int*) &dvtext},
    {"t", ARFLAG, (int*) &dtext},
    {"w", ARINT,  (int*) &drwinds},
    {"?", ARFLAG, (int*) &drhlp},
#if DBG
    {"d", ARINT, (int*) &ddbg},
#endif
};

Local int noptions = sizeof(options)/sizeof(topt);

Local   int     proct[MAXPROCS], procg[MAXPROCS];

Local   int         prmode = 0;

/*==================================================================*/
Public  int    main(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
    Main routine for graph plotting
--------------------------------------------------------------------*/
{
    int         nprc, ngrf, ng, nplanes;
    int         nprocesses;
    tproc       *proc;
    tgraph      *graph;

#if DBG
    dadbg = 3;  /* debug flag for dagrf.c */
    drdbg = 2;  /* debug flag for drgrf.c */
    fldbg = 2;  /* debug flag for flgrf.c */
    ildbg = 2;  /* debug flag for ilgrf.c */
    pmdbg = 2;  /* debug flag for matchv.c */
    pvdbg = 2;  /* debug flag for rdptbl.c */
    rddbg = 2;  /* debug flag for token.c */
    rgdbg = 1;  /* debug flag for rdgrf.c */
    tldbg = 2;  /* debug flag for tlgrf.c */

    dbg = 0;
#endif

    argc = drcarg(argc, argv);

    /* initalization */
    genct      = True;
    nprocesses = dasave(fngrf, proct, procg);

    /* initialize drawing package */
    drwinds = drinit(OUTFIG, drwinds, &prmode);
    nplanes = drwinds*drwinds;

    /* repeat for graphs */
    
    printf("%d processes\n", nprocesses);
    for(nprc = 0; nprc < nprocesses; nprc++) {

        proc = dagetp(nprc);  /* get process */
        ngrf = dangrf(nprc);  /* the number of graphs */

        if(nprc > 0) {
            wdopen(OUTFIG, &prmode);
        }
        for(ng = 0; ng < ngrf; ng++) {
            if((ng % nplanes) == 0 && ng > 0) {
                wdcls(OUTFIG);
                wdopen(OUTFIG, &prmode);
            }
            if(drcomp) {
                graph = daread(nprc, ng);
            } else {
                graph = daread0(nprc, ng);
            }
            Discard drgraph(proc, graph, ng);    /* draw the graph */
        }
        wdcls(OUTFIG);
        drclr(True);
        printf("Process %d: %d graphs are plotted\n", nprc+1, ngrf);
    }

    drclose(OUTFIG);          /* close drawing package */

    daclose();        /* close read package */

    return 0;
}

/*==================================================================*/
Private int    drcarg(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
*/
{
    int    pos;

    argc = aropt(argc, argv, noptions, options);

    if(drhlp != 0) {
        drhelp(argv[0]);
        exit(0);
    }

    switch (dlabel) {
        case ARPLUS:
            drlabel = True;
            break;
        case ARMINUS:
            drlabel = False;
            break;
    }

    switch (dcomp) {
        case ARPLUS:
            drcomp = True;
            break;
        case ARMINUS:
            drcomp = True;
            break;
    }

    switch (dtext) {
        case ARPLUS:
            drtext = True;
            break;
        case ARMINUS:
            drtext = False;
            break;
    }

    switch (ditext) {
        case ARPLUS:
            dritext = True;
            break;
        case ARMINUS:
            dritext = True;
            break;
    }

    switch (dvtext) {
        case ARPLUS:
            drvtext = True;
            break;
        case ARMINUS:
            drvtext = True;
            break;
    }

    /* input process file name */
    pos = 1;
    if((fngrf = arnext(argc, argv, &pos)) != NULL) {
        pos--;
        argc = ardel(argc, argv, pos, 1);
    } else {
        fngrf = finame;
    }

#if DBG
    /* debugging flags */
    dbg   = 0;

    if(ddbg > 0) {
        dbg = ddbg;
    } else if(ddbg < 0) {
        dbg = -ddbg;
    }
#endif

    return argc;
}

/*==================================================================*/
Private Void   drhelp(cmd)
string  cmd;
/*--------------------------------------------------------------------
 */
{
    grcmsg("Feynman graph plotter");

    printf("\n\n");
    printf("Syntax %s [Graph_File_name]\n",
           cmd);
    printf("\nOption\n");
    printf("    -w int              ");
    printf(": change #graphs/page.\n");
    printf("    -t                  ");
    printf(": without text\n");
    printf("    -l                  ");
    printf(": without label\n");
#if DBG
    printf("    -d int              ");
    printf(": set debug level\n");
#endif
    printf("    -h                  ");
    printf(": print this message\n");
    printf("\n");
}

/*==================================================================*/
Public  Void   prdump(NoArg)
/*--------------------------------------------------------------------
    Dump.
*/
{
}
