/*********************************************************************
 *  File : pread.c (prefix = pv)
 *
 *  read particle table
 *                                                 by Kaneko Toshiaki
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#define  DEFGLOBAL
#include "grdata.h"
#include "args.h"
#include "config.h"
#include "token.h"
#include "version.h"

/*--------------------------------------------------------------------
 * Local functions
 */
Private int     grcarga      ARG((int, char**));
Private Void    gropth       ARG((string));
Private Void    genconj      ARG((NoArg));

#ifndef sign
#define sign(x)      (((x) <  0 ) ? (-1) : (1))
#endif

/*--------------------------------------------------------------------
 * Local variables
 */
Local   char    finame[] = "";
Local   string  fname = NULL;
Local   char    firsp[MAXPATHLEN];

Local   int     dhlp    = 0;
#if DBG
Local   int     ddbg    = 0;
Local   int     dmon    = 0;
Local   int     dtrc    = 0;
#endif

Local   topt    options[] = {
    {"h", ARFLAG, (int*) &dhlp},
    {"?", ARFLAG, (int*) &dhlp},
#if DBG
    {"d", ARINT,  (int*) &ddbg},
    {"m", ARFLAG, (int*) &dmon},
    {"t", ARFLAG, (int*) &dtrc},
#endif
};

Local int noptions = sizeof(options)/sizeof(topt);

/*==================================================================*/
Public  int    main(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
    Main routine for graph drawin
--------------------------------------------------------------------*/
{

    grcmsg("Particle table");             /* First message */

    argc = grcarga(argc, argv);   /* analyze arguments */
    if(argc != 1) {
        gropth(argv[0]);
        exit(1);
    }

    if(strlen(fname) < 1) {
        printf("*** Please specify the name of model file.\n");
        exit(1);
    }

#if DBG
    /* debugging flags for modules.
       if(dbg >= xxdbg) then enable debugging
    */
    pvdbg = 2;  /* debug flag for rdptbl.c */
    rddbg = 2;  /* debug flag for token.c */
#endif

    dbg = 0;

    /* read process */
    fnmdl = firsp;
    pvfname(fnmdl, fname, MAXPATHLEN); /* model file name */
    pvread(fnmdl);

    pvprtex("out.tex");

    pvprnam("out.nam");

    pvprgnam("gname.dat");

    genconj();

    return 0;
}

/*==================================================================*/
Private int    grcarga(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
*/
{
    int    pos;
    string cmd;

    cmd     = argv[0];

#if DBG
    ddbg    = 0;
    trace   = False;
    mon     = False;
#endif

    argc = aropt(argc, argv, noptions, options);

    if(dhlp != 0) {
        gropth(cmd);
        exit(0);
    }

    /* input process file name */
    pos = 1;
    if((fname = arnext(argc, argv, &pos)) != NULL) {
        pos--;
        argc = ardel(argc, argv, pos, 1);
    } else {
        fname = finame;
    }

#if DBG
    /* debugging flags */
    dbg   = 0;
    dbgst = -1;

    if(ddbg > 0) {
        dbg   = ddbg;
    }
    if(ddbg < 0) {
        dbgst = -ddbg;
    }

    switch (dtrc) {
        case ARPLUS:
            trace = True;
            break;
        case ARMINUS:
            trace = False;
            break;
    }

    switch (dmon) {
        case ARPLUS:
            mon = True;
            break;
        case ARMINUS:
            mon = False;
            break;
    }
#endif

    return argc;
}

/*==================================================================*/
Private Void   gropth(cmd)
string  cmd;
/*--------------------------------------------------------------------
*/
{
    printf("Syntax %s [options] [Input_File_name]\n", cmd);
    printf("\nOption\n");
#if DBG
    printf("    -d int              ");
    printf(": debug leve \n");
    printf("    -m                  ");
    printf(": enable monitor\n");
    printf("    -t                  ");
    printf(": enable trace\n");
#endif
    printf("    -h                  ");
    printf(": more info\n");
    printf("\nDefault file name is:\n");
    printf("     Input_File_name = %s\n", finame);
}

/*==================================================================*/
Public Void   prdump(NoArg)
/*--------------------------------------------------------------------
*/
{
}

/*==================================================================*/
Private Void   pfadef(FILE *pvfp, string cnam, int dimn, int *dsize)
/*--------------------------------------------------------------------
*/
{
    int i;

    fprintf(pvfp, "%s", cnam);
    if(dimn > 0) {
        fprintf(pvfp, "(");
        for(i = 0; i < dimn; i++) {
             if(i != 0) {
                 fprintf(pvfp, ",");
             }
             fprintf(pvfp, "%d", dsize[i]);
        }
        fprintf(pvfp, ")");
    }
    
}

/*==================================================================*/
Private Void   pfaindx(FILE *pvfp, string cnam, int dimn, string n)
/*--------------------------------------------------------------------
*/
{
    int i;

    fprintf(pvfp, "%s", cnam);
    if(dimn > 0) {
        fprintf(pvfp, "(");
        for(i = 0; i < dimn; i++) {
             if(i != 0) {
                 fprintf(pvfp, ",");
             }
             fprintf(pvfp, "%s%d", n, i);
        }
        fprintf(pvfp, ")");
    }
    
}

/*==================================================================*/
Private Void   genconj(NoArg)
/*--------------------------------------------------------------------
*/
{
    int v, i;
    FILE *pvfp;

    for(v = 0; v < nvtable; v++) {
        if(vtable[v]->fname[1] == NULL) {
            vtable[v]->fname[1] = strdup(vtable[v]->fname[0]);
            vtable[v]->fname[1][0]++;
            vtable[v]->fndimn[1] = vtable[v]->fndimn[0];
            for(i = 0; i < vtable[v]->fndimn[0]; i++) {
                vtable[v]->fndim[1][i] = vtable[v]->fndim[0][i];
                vtable[v]->fnind[1][i] = vtable[v]->fnind[0][i];
            }
        }
    }
    pvprint("mod.mdl");
    pvfp = fopen("cconj.f", "w");
    if(pvfp == NULL) {
        fprintf(Stderr, "*** cannot open file\n");
        exit(FAIL);
    }
    for(v = 0; v < nvtable; v++) {
        if(strcmp(vtable[v]->fname[0], "?") == 0) {
            continue;
        }
        fprintf(pvfp, "      complex*16 ");
        pfadef(pvfp, vtable[v]->fname[1], vtable[v]->fndimn[0],
                     vtable[v]->fndim[0]);
        fprintf(pvfp, "\n");
    }
    fprintf(pvfp, "\n");
    for(v = 0; v < nvtable; v++) {
        if(strcmp(vtable[v]->fname[0], "?") == 0) {
            continue;
        }
        for(i = vtable[v]->fndimn[0]-1; i >= 0; i--) {
            fprintf(pvfp, "      do i%d = 1, %d\n",
                    i, vtable[v]->fndim[0][i]);
        }

        fprintf(pvfp, "        ");
        pfaindx(pvfp, vtable[v]->fname[1], vtable[v]->fndimn[0], "i");
        fprintf(pvfp, " = conjg(");
        pfaindx(pvfp, vtable[v]->fname[0], vtable[v]->fndimn[0], "i");
        fprintf(pvfp, ")\n");

        for(i = 0; i < vtable[v]->fndimn[0]; i++) {
            fprintf(pvfp, "      enddo\n");
        }
    }
    fclose(pvfp);
}
