/*********************************************************************
  File : tread.c

  Entry for testing particle table
*/

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#define  DEFGLOBAL
#include "grdata.h"
#include "config.h"
#include "token.h"
#include "cfnbase.h"
#include "cfnconv.h"
#include "args.h"

/*--------------------------------------------------------------------
 * Local functions
 */
Private int    grcarg           ARG((int, char**));
Private Void   grhelp           ARG((string));

/*--------------------------------------------------------------------
 * Local variables
 */
Local   char    finame[] = "out.grf";
Local   char    foname[] = "out1.grf";
Local   char    *fgname  = NULL;
Local   char    *fgoutn  = NULL;

Local   int     proct[MAXPROCS];
Local   int     procl[MAXPROCS];
Local   int     procg[MAXPROCS];

Local   int     drnum = 0;
Local   int     drhlp, dnum;
#if DBG
Local   int     ddbg;
#endif

Local   topt    options[] = {
    {"n", ARFLAG, (int*) &dnum},
    {"h", ARFLAG, (int*) &drhlp},
    {"?", ARFLAG, (int*) &drhlp},
#if DBG
    {"d", ARINT, (int*) &ddbg},
#endif
};

Local int noptions = sizeof(options)/sizeof(topt);

/*==================================================================*/
Public  int    main(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
    Main routine for graph drawin
--------------------------------------------------------------------*/
{
    FILE     *fpout;
    int       nprocs;
    int       gr;
    int       pc, cnprc, nextern, nilines, cngraph, c;
    tgraph   *graph;
    tproc    *proc;
    int       lastgr;
    bool      grput;
    /* CF */
    int       csgn = 1, grn;
    int       ibase = 0, iprt = 0;
    bool      skip;

#if DBG
    trace   = 0;
#endif

#if DBG
    /* debugging flags for modules.
     * if(dbg >= xxdbg) then enable debugging
     */
    dbg = 0;
    dadbg = 1;  /* debug flag for dagrf.c */
    fldbg = 2;  /* debug flag for flgrf.c */
    ildbg = 2;  /* debug flag for ilgrf.c */
    pmdbg = 2;  /* debug flag for matchv.c */
    rddbg = 2;  /* debug flag for token.c */
    rgdbg = 2;  /* debug flag for rdgrf.c */
    pvdbg = 2;  /* debug flag for rdptbl.c */
#endif

    argc = grcarg(argc, argv);   /* analyze arguments */
    printf("\"%s\" -> \"%s\"\n", fgname, fgoutn);

    if((fpout = fopen(fgoutn, "w")) == NULL) {
        fprintf(Stderr, "*** cannot open \"%s\"\n", fgoutn);
        exit(FAIL);
    }

    if((gcolgr = (cftgraph *) malloc(sizeof(cftgraph))) == NULL) {
         fprintf(Stderr, "*** gcolgr: memory allocation failed\n");
         exit(1);
    }
    if((gcbase = (cftbase *) malloc(sizeof(cftbase))) == NULL) {
         fprintf(Stderr, "*** gcbase: memory allocation failed\n");
         exit(1);
    }

    genct  = True;
    nprocs = dasave(fgname, procl, procg);
    for(pc = 0; pc < nprocs; pc++) {
        proct[pc] = pc;
    }

    fofinit(fpout);

    cnprc    = proct[0];       /* process number */
    cprocess = dagetp(cnprc);   /* process data */
    cngraph  = 0;
    cgraph   = daread0(cnprc, cngraph);
    ilmom();
    nilines = cgraph->nilines;
    flget(cgraph);

    /* gcbase is used in ftinit */
    cfcbase(gcbase);
    cfbase(gcbase, ibase, iprt);

    for(pc = 0; pc < nprocs; pc++) {
        printf("%% +++ process = %d/%d +++\n", pc+1, nprocs);

        cnprc    = proct[pc];       /* process number */
        proc     = dagetp(cnprc);     /* process data */
        cprocess = proc;
        nextern  = cprocess->oextern;

        grn = 0;
        lastgr = -1;
        grput  = False;
        for(gr = 0; gr < procg[pc]; gr++) {
            printf("%% +++ graph = %d/%d (%d/%d) +++\n",
                   gr+1, procg[pc], pc+1, nprocs);

            if(!(gcbase->nbase == 1 && gcbase->nexcolor == 0)) {
                cngraph = gr;
                graph   = daread0(cnprc, cngraph);
                cgraph  = graph;

                if(lastgr == cgraph->grid) {
                    if(grput) {
                        continue;
                    }
                } else {
                    lastgr = cgraph->grid;
                    grput  = False;
                }

                ilmom();
                nilines = cgraph->nilines;
                flget(cgraph); 

                csgn = cfcproj(gcbase,gcolgr);
                cfproj(gcbase, gcolgr, ibase, iprt);

                skip = False;
                for(c = 0; c < gcbase->nbase; c++) {
                    if(gcolgr->wgt[c].num != 0) {
                        skip = False;
                        break;
                    }
                    skip = True;
                }
                cfcgfree(gcolgr);

                if(skip) {
                    printf("Graph %ld.%ld is deleted\n",
                           cgraph->grid, cgraph->subid);
                    continue;
                }

                grput = True;
                cgraph->subid = 0;
            }
            grn++;

            cngraph = gr;
            graph   = daread(cnprc, cngraph);
            if(drnum) {
                graph->grid = grn;
            }
            cgraph  = graph;

            ilmom();
            nilines = cgraph->nilines;
            flget(cgraph); 

            if(grn == 1) {
                fofproc(fpout, proc, graph, NULL);
            }
            fofgraph(fpout, proc, graph);
            foflines(fpout);
        }
        if(grn > 0) {
            procg[pc] = grn;
            fofpend(fpout, procg[pc]);
        }
    }

    cfcbfree(gcbase);

    fofend(fpout, nprocs+1);

    daclose();

/*  the following lines are dangerous because it is not checked
 *  wheter "out.grf.orig" exists or not.
 */
/*
    system("mv out.grf out.grf.orig");
    system("mv out1.grf out.grf");
*/
    return 0;
}

/*==================================================================*/
Private int    grcarg(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
 */
{
    int    pos;
    char   *f1, *f2;

    argc = aropt(argc, argv, noptions, options);

    if(drhlp != 0) {
        grhelp(argv[0]);
        exit(0);
    }
    switch (dnum) {
        case ARPLUS:
            drnum = True;
            break;
        case ARMINUS:
            drnum = False;
            break;
    }


    /* input process file name */
    f1  = f2 = NULL;
    pos = 1;
    if((f1 = arnext(argc, argv, &pos)) != NULL) {
        pos--;
        argc = ardel(argc, argv, pos, 1);
        if((f2 = arnext(argc, argv, &pos)) != NULL) {
            pos--;
            argc = ardel(argc, argv, pos, 1);
        }
    }
    if(f1 != NULL) {
        if(f2 != NULL) {
            fgname = f1;
            fgoutn = f2;
        } else {
            fgname = finame;
            fgoutn = f1;
        }
    } else {
        fgname = finame;
        fgoutn = foname;
    }
    if(strcmp(fgname, fgoutn) == 0) {
        fprintf(Stderr, "*** input and output file name is the same\n");
        fprintf(Stderr, "    \"%s\"\n", fgname);
        exit(1);
    }

#if DBG
    /* debugging flags */
    dbg   = 0;

    if(ddbg > 0) {
        dbg = ddbg;
    } else if(ddbg < 0) {
        dbg = -ddbg;
    }
#endif

    return argc;
}

/*==================================================================*/
Private Void   grhelp(cmd)
string  cmd;
/*--------------------------------------------------------------------
 */
{
    grcmsg("Eliminate graphs whose color factor is 0.");

    printf("\n\n");
    printf("Syntax %s [[Graph_File_name] New_Graph_File_name]\n",
           cmd);
    printf("\nOption\n");
    printf("    -n                  ");
    printf(": rename graph number\n");
#if DBG
    printf("    -d int              ");
    printf(": set debug level\n");
#endif
    printf("    -h                  ");
    printf(": print this message\n");
    printf("\n");

    printf("\nDefault file name:\n");
    printf("     Graph_File_name = %s\n", finame);
    printf("\n");
}

/*==================================================================*/
Public  Void   prdump(NoArg)
/*--------------------------------------------------------------------
 *  Dummy function of dump graph
 */
{
    rgdump(cprocess, cgraph);
    ildump();
    fldump();
}
