/*********************************************************************
 *  File : testcns.c
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdio.h>

#define  DEFGLOBAL
#include "grdata.h"
#include "config.h"
#include "token.h"
#include "args.h"
#include "conserv.h"
#include "version.h"

/*--------------------------------------------------------------------
 * Local functions
 */
Private int     grcarg       ARG((int, char**));
Private Void    gropth       ARG((string));
Private Void    prfirst      ARG((NoArg));
Private Void    prinit       ARG((NoArg));
Private Void    prlast       ARG((NoArg));
Private Void    prproc       ARG((int,int));

Private Void    grfproc      ARG((int));
Private Void    grfprc        ARG((int,int,int,int));

#define YorN(x)   ((x) ? "Yes": "No")

/*--------------------------------------------------------------------
 * Local variables
 */
int     start;

Local   int     nprc = 0;
Local   char    finame[] = "in.prs";
Local   char    foname[] = "out.grf";
Local   int    *cpext;
Local   int     nsproc;

Local   int     dskasp  = 0;
Local   int     doutf   = 0;
Local   int     dtype   = 0;
Local   int     dtone   = 0;
Local   int     dgone   = 0;
Local   int     dhlp    = 0;
#if DBG
Local   int     dprt    = 0;
Local   int     ddbg    = 0;
Local   int     dmon    = 0;
Local   int     dtrc    = 0;
Local   int     dsel    = -1;
Local   int     dcpt    = 0;
#endif

Local   topt    options[] = {
    {"o", ARSTR,  (int*) &fngrf},
    {"a", ARFLAG, (int*) &dskasp},
    {"c", ARFLAG, (int*) &doutf},
    {"g", ARINT,  (int*) &dgone},
    {"t", ARINT,  (int*) &dtype},
    {"T", ARFLAG, (int*) &dtone},
    {"h", ARFLAG, (int*) &dhlp},
    {"?", ARFLAG, (int*) &dhlp},
#if DBG
    {"p", ARINT,  (int*) &dprt},
    {"d", ARINT,  (int*) &ddbg},
    {"m", ARFLAG, (int*) &dmon},
    {"r", ARFLAG, (int*) &dtrc},
    {"s", ARINT,  (int*) &dsel},
    {"i", ARINT,  (int*) &dcpt},
#endif
};

Local int noptions = sizeof(options)/sizeof(topt);

/*==================================================================*/
Public  int    main(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
    Main routine for graph drawin
--------------------------------------------------------------------*/
{
    bool mode;

    grcmsg("Feynman graph generator");          /* First message */

    argc = grcarg(argc, argv);   /* analyze arguments */
    if(argc != 1) {
        gropth(argv[0]);
        exit(1);
    }

#if DBG
    /* debugging flags for modules.
       if(dbg >= xxdbg) then enable debugging
    */
    gcdbg = 1;  /* debug flag for grc.c */
    pvdbg = 2;  /* debug flag for rdptbl.c */
    rddbg = 2;  /* debug flag for token.c */
    rgdbg = 2;  /* debug flag for rdgraph.c */
    rpdbg = 2;  /* debug flag for rdproces.c */
    sfdbg = 2;  /* debug flag for sfgrf.c */
#endif

    /* read process */
    if(!rpinit(fnprc, False))
        exit(16);

    conserv();
    cocdump();
    coprint();

    printf("nptable = %d, nvtable = %d\n", nptable, nvtable);
    return 0;
}

/*==================================================================*/
Private int    grcarg(argc, argv)
int     argc;
char    **argv;
/*--------------------------------------------------------------------
*/
{
    string s;
    int    pos;
    string cmd;

    cmd     = argv[0];
    skipasp = False;
    outf    = True;
    fngrf   = NULL;
    display = False;

#if DBG
    dispno  = -1;
    ddbg    = 0;
    dprt    = 0;
    trace   = False;
    mon     = False;
#endif

    argc = aropt(argc, argv, noptions, options);

    if(dhlp != 0) {
        gropth(cmd);
        exit(0);
    }

    if(fngrf == NULL)
        fngrf = foname;

    if(dskasp != 0) {
        skipasp = True;
    }

    switch (doutf) {
        case ARPLUS:
            outf = True;
            break;
        case ARMINUS:
            outf = False;
            break;
    }

    seltype = dtype;

    switch (dtone) {
        case ARPLUS:
            onetype = True;
            break;
        case ARMINUS:
            onetype = True;
            break;
    }

    onegraph = dgone;

    /* input process file name */
    pos = 1;
    if((fnprc = arnext(argc, argv, &pos)) != NULL) {
        pos--;
        argc = ardel(argc, argv, pos, 1);
    } else {
        fnprc = finame;
    }

#if DBG
    /* debugging flags */
    dbg   = 0;
    dbgst = -1;
    selprop  = 0;

    if(ddbg > 0) {
        dbg   = ddbg;
    }
    if(ddbg < 0) {
        dbgst = -ddbg;
    }

    if(dprt > 0) {
        display = dprt;
        dispno  = -1;
    } else {
        display = False;
        dispno  = -dprt;
    }

    if(dtrc != 0) {
        trace = True;
    }

    if(dmon != 0) {
        mon = True;
    }

    if(dsel > 0) {
        selprop = dsel;
    } else {
        selprop = 0;
    }
#endif

    return argc;
}

/*==================================================================*/
Private Void   gropth(cmd)
string  cmd;
/*--------------------------------------------------------------------
*/
{
    printf("Syntax %s [options] [Input_File_name]\n", cmd);
    printf("\nOption\n");
    printf("    -a                  ");
    printf(": skip particle assignment\n");
    printf("    -l int              ");
    printf(": with detailed message \n");
    printf("    -o Output_File_name ");
    printf(": specify output file\n");
    printf("    -t int              ");
    printf(": output for specified type\n");
    printf("    -T                  ");
    printf(": output one graph from each type\n");
    printf("    -c                  ");
    printf(": only count\n");
#if DBG
    printf("    -p int              ");
    printf(": print level\n");
    printf("    -d int              ");
    printf(": debug level\n");
    printf("    -m                  ");
    printf(": enable monitor\n");
    printf("    -i int              ");
    printf(": cptest flag\n");
    printf("    -r                  ");
    printf(": enable trace\n");
#endif
    printf("    -h                  ");
    printf(": print this message\n");
    printf("\nDefault file names:\n");
    printf("     Input_File_name = %s\n", finame);
    printf("    Output_File_name = %s\n", foname);
}

/*==================================================================*/
Private Void    grfproc(proc)
int proc;
/*--------------------------------------------------------------------
 * Construct all possible process
 *------------------------------------------------------------------*/
{
    int i;

    if((cpext = (int *) malloc(sizeof(int)*nextern)) == NULL) {
        fprintf(Stderr, "*** grfproc: memory allocation error\n");
        exit(1);
    }

    for(i = 0; i < nextern; i++)
        cpext[i] = ndname(i);

    nsproc = 0;

    grfprc(proc, 0, 1, 0);

    free(cpext);
}

/*==================================================================*/
Private Void    grfprc(proc, start, sp, ss)
int     proc;
int     start;   /* the external particle number to be assigned */
int     sp;      /* last particle */
int     ss;      /* last sing (0, 1) */
/*--------------------------------------------------------------------
 * Construct all possible process
 *------------------------------------------------------------------*/
{
    int i, p, s, sign;

    if(sp < 1)
        sp = 1;

/*
printf("grfprc(%d, %d, %d, %d)", proc, start, sp, ss);
for(i = 0; i < nextern; i++) printf("%d ", cpext[i]); printf("\n");
*/

    /* initail particles */
    for(i = start; i < nextern; i++) {
        if(cpext[i] == 0) {
            for(p = sp; p < nptable; p++) {
                if(p != sp)
                    ss = 0;
                for(s = ss; s <= 1; s++) {
                    if(s == 0 && ptable[p]->antip >= 0)
                        continue;
                    cpext[i] = (2*s - 1)*p;
/*
                    grfprc(proc, i+1, sp, ss);
*/
                    grfprc(proc, i+1, p, s);
                }
            }
            cpext[i] = 0;
            return;
        }
    }
/*
printf("grfprc(%d) (nextern=%d):", start, nextern);
for(i = 0; i < nextern; i++) printf("%d ", cpext[i]); printf("\n");
*/
    if(constest(nextern, cpext)) {
/*
printf("grfprc(%d) (nextern=%d):generate:", start, nextern);
for(i = 0; i < nextern; i++) printf("%d ", cpext[i]); printf("\n");
*/

        nsproc++;
        prproc(proc, nsproc);
    }
}

/*==================================================================*/
Private Void   prinit(NoArg)
/*--------------------------------------------------------------------
    Print initial message
*/
{
    printf("\n");
    printf("    Model  file name = \"%s\"\n", fnmdl);
    printf("    Input  file name = \"%s\"\n", fnprc);

    if(outf)
    printf("    Output file name = \"%s\"\n", fngrf);
}

/*==================================================================*/
Private Void   prfirst(NoArg)
/*--------------------------------------------------------------------
    Print the first message.

--------------------------------------------------------------------*/
{
    node n;

    printf("*************************************************\n");
    printf("Order = %d (", order);
    for(n = 0; n < nmordtbl; n++) {
        if(n > 0)
            printf(", ");
        printf("%s = %d", mordtbl[n], morder[n]);
    }
    printf(") loop = %d\n", loop);
    printf("\n%4d external lines\n\n", nextern);

    for(n = 0; n < nextern; n++) {
        printf("%4d %s %s\n",
                tovert(n), eliorf(ndiorf(n)),
                pvename(ndiorf(n), ndname(n)));
    }

    printf("\nConditions of generationg graphs:\n");
    printf("    Gnenerate expanded graph = %s\n", YorN(expand));
    printf("    Only 1PI graph           = %s\n", YorN(onepi));
    printf("    Gnenerate tadpole        = %s\n", YorN(tadpole));
    printf("    Assign particle          = %s\n", YorN(!skipasp));
    printf("    HEP                      = %s\n", YorN(hep));
    printf("    Block                    = %s\n", YorN(block));
    printf("    Self energy              = %s\n", YorN(selfe));
    printf("    Counter terms            = %s\n", YorN(countert));

    printf("\nConditions for outputs:\n");
    printf("    Display graph data       = %s\n", YorN(display));
    printf("    Write graph data to file = %s\n", YorN(outf));

#if DBG
    printf("\nConditions for debugging:\n");
    printf("    Debug                    = %s\n", YorN(dbg));
    printf("    Monitor performance      = %s\n", YorN(mon));
    printf("    Trace functions          = %s\n", YorN(trace));
    printf("    Check the result         = %s\n", YorN(check));

    if(selprop > 0) {
        printf("\nSelect graph with propagator of %s\n",
               ntable[ptable[selprop]->pname[0]].name);
    }
#endif
    printf("=================================================\n");
}

/*==================================================================*/
Private Void   prlast(NoArg)
/*--------------------------------------------------------------------
 *  Print the last message.
 *------------------------------------------------------------------*/
{
    printf("*************************************************\n");
}

/*==================================================================*/
Public  Void    prdump(NoArg)
/*--------------------------------------------------------------------
 */
{
}

/*==================================================================*/
Private Void    prproc(proc, sub)
int proc;
int sub;
/*--------------------------------------------------------------------
    Print the processes.
--------------------------------------------------------------------*/
{
    int i, j;
    static mode = True;
    static char str[256];
    FILE   *fp;

    sprintf(str, "p%ds%d.prc", proc, sub);
    printf("Create \"%s\"\n", str);
    if((fp = fopen(str, "w")) == NULL) {
        fprintf(Stderr, "\n*err* prproc: open error to write ");
        fprintf(Stderr,"filename = %s\n", str);
        exit(1);
    }

    /* name of model file */
    fprintf(fp, "%%\n");
    fprintf(fp, "Model=\"%s\";\n", fnmdl);
    mode = False;

    /* process number */
    fprintf(fp, "%%\n");
    fprintf(fp, "Process; %% %d.%d\n", proc, sub);

    /* order of coupling constants */
    for(i = 0; i < nmordtbl; i++)
        fprintf(fp, "%s=%d; ", mordtbl[i], morder[i]);
    fprintf(fp,"\n");

    /* initail particles */
    j = 0;
    fprintf(fp, "Initial={");
    for(i = 0; i < nextern; i++)
        if(ndiorf(i) == ELINITIAL) {
            if(j > 0)
                fprintf(fp, ", ");
            j++;
            if(ptable[abs(cpext[i])]->antip >= 0)
                fprintf(fp, "%s", pvname(cpext[i]));
            else
                fprintf(fp, "%s", pvname(-cpext[i]));
        }
    fprintf(fp,"};\n");

    /* final particles */
    j = 0;
    fprintf(fp, "Final  ={");
    for(i = 0; i < nextern; i++)
        if(ndiorf(i) == ELFINAL) {
            if(j > 0)
                fprintf(fp, ", ");
            j++;
            fprintf(fp, "%s", pvname(cpext[i]));
        }
    fprintf(fp,"};\n");

    /* flags */
    fprintf(fp,"Expand=%s; Opi=%s; Block=%s; HEP=%s; Gm2=%s;\n",
            YorN(expand),
            YorN(onepi),
            YorN(block),
            YorN(hep),
            YorN(gm2));
    fprintf(fp,"Tadpole=%s; Extself=%s; Selfe=%s; Countert=%s;\n",
            YorN(tadpole),
            YorN(extself),
            YorN(selfe),
            YorN(countert));
    if(kinem != NULL)
        fprintf(fp,"kinem=\"%s\";\n", kinem);

    /* end */
    fprintf(fp, "Pend;\n\n");
    fclose(fp);
}
