/*********************************************************************
 *  File : vr.c
 *
 *  Graphic Interface package
 *
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include "vr.h"
#include "vrx.h"
#include "vrf.h"

/*---------------------------------------------------------------------
    Local variables
*/

/*---------------------------------------------------------------------
    Local functions
*/
Private Void vrdest    ARG((int));

/*---------------------------------------------------------------------
    Local variables
*/

Local   coord wwidth;
Local   int outdest=OUTX;

/*********************************************************************
 * system control
 */
/*==================================================================*/
Public  Void vrinit(NoArg)
/*--------------------------------------------------------------------
 * initialize device
 */
{
    vxrinit();
    vfrinit();
}

/*==================================================================*/
Public  Void vropen(dest, opt)
int     dest;
int    *opt;
/*--------------------------------------------------------------------
 * this function is called when new bwindow is opened
 */
{
    vrdest(dest);

    if(outdest == OUTX)
        vxropen();
    else if(outdest == OUTFIG)
        vfropen(opt);
    else
        fprintf(Stderr, "vropen : unknown destination.\n");
}

/*==================================================================*/
Public  Void vrcls(dest)
int     dest;
/*--------------------------------------------------------------------
 * this function is called when window is closed
 */
{

    vrdest(dest);

    if(outdest == OUTX) {
        vxrcls();
    } else if(outdest == OUTFIG) {
        vfrcls();
        outdest = OUTX;
    } else {
        fprintf(Stderr, "vrcls : unknown destination.\n");
    }
}

/*==================================================================*/
Public  Void vrdisp(NoArg)
/*--------------------------------------------------------------------
 * display graphic objects
 */
{
    if(outdest == OUTX)
        vxrdisp();
    else if(outdest == OUTFIG)
        vfrdisp();
    else
        fprintf(Stderr, "vrdisp : unknown destination.\n");
}

/*==================================================================*/
Public  bool vrclr(clr)
/*--------------------------------------------------------------------
 * clear the current graphic display
 */
{
    bool st = False;

    if(outdest == OUTX)
        st = vxrclr(clr);
    else if(outdest == OUTFIG)
        st = vfrclr(clr);
    else
        fprintf(Stderr, "vrclr : unknown destination.\n");
    return st;
}

/*==================================================================*/
Public  Void vrnew(NoArg)
/*--------------------------------------------------------------------
 * this function is called when new subwindow is opened
 */
{
    if(outdest == OUTX)
        vxrnew();
    else if(outdest == OUTFIG)
        vfrnew();
    else
        fprintf(Stderr, "vropen : unknown destination.\n");
}

/*==================================================================*/
Public  Void vrend(NoArg)
/*--------------------------------------------------------------------
 * this function is called when new subwindow is closed
 */
{
    if(outdest == OUTX)
        vxrend();
    else if(outdest == OUTFIG)
        vfrend();
    else
        fprintf(Stderr, "vrend : unknown destination.\n");
}

/*==================================================================*/
Private Void vrdest(dest)
int     dest;
/*--------------------------------------------------------------------
*/
{
    int dest0;

    if(dest < OUTX || dest > OUTFIG) {
        fprintf(Stderr, "vrdest : illegal destination %d\n", dest);
        exit(1);
    }
    dest0   = outdest;
    outdest = dest;

    if(dest0 == OUTFIG && outdest != OUTFIG)
        vfrcls();
}
/*********************************************************************
 * Graphic objects
 *
 */
/*==================================================================*/
Public  Void vrline(lx0, ly0, lx1, ly1, color)
coord lx0, ly0;
coord lx1, ly1;
int color;
/*--------------------------------------------------------------------
*/
{

    if(outdest == OUTX)
        vxrline(lx0, ly0, lx1, ly1, color);
    else if(outdest == OUTFIG)
        vfrline(lx0, ly0, lx1, ly1, color);
    else
        fprintf(Stderr, "vrline : unknown destination.\n");
}

/*==================================================================*/
Public  Void vrcrcl(cx, cy, r, color)
coord cx, cy;
coord r;
int color;
/*--------------------------------------------------------------------
*/
{
    if(outdest == OUTX)
        vxrcrcl(cx, cy, r, color);
    else if(outdest == OUTFIG)
        vfrcrcl(cx, cy, r, color);
    else
        fprintf(Stderr, "vrcrcl : unknown destination.\n");
}

/*==================================================================*/
Public  Void vrarc(lx0, ly0, lx1, ly1, lx2, ly2, color)
coord lx0, ly0;
coord lx1, ly1;
coord lx2, ly2;
int color;
/*--------------------------------------------------------------------
*/
{
    if(outdest == OUTX)
        vxrarc(lx0, ly0, lx1, ly1, lx2, ly2, color);
    else if(outdest == OUTFIG)
        vfrarc(lx0, ly0, lx1, ly1, lx2, ly2, color);
    else
        fprintf(Stderr, "vrarc : unknown destination.\n");
}

/*==================================================================*/
Public  Void vrtext(x, y, str, color)
coord x, y;
string str;
int color;
/*--------------------------------------------------------------------
*/
{
    if(outdest == OUTX)
        vxrtext(x, y, str, color);
    else if(outdest == OUTFIG)
        vfrtext(x, y, str, color);
    else
        fprintf(Stderr, "vrtext : unknown destination.\n");
}

/*********************************************************************
 * Utility funcitons
 *
 */
/*==================================================================*/
Public  int vrcolr(NoArg)
/*--------------------------------------------------------------------
 * Return the number of colors.
*/
{
    if(outdest == OUTX)
        return vxrcolr();
    else if(outdest == OUTFIG)
        return 2;
    else
        fprintf(Stderr, "vrtext : unknown destination.\n");
    return 2;
}

/*==================================================================*/
Public  int vrback(color)
int     color;
/*--------------------------------------------------------------------
 * Return the background color
*/
{
    if(outdest == OUTX)
        return vxrback(color);
    else if(outdest == OUTFIG)
        return VR_WHITE;
    else
        fprintf(Stderr, "vrback : unknown destination.\n");
    return VR_WHITE;
}

/*==================================================================*/
Public  int vrgnum(NoArg)
/*--------------------------------------------------------------------
*/
{
    return vxrgnum();
}

/*==================================================================*/
Public  Void  vrchrw(str, ans)
string str;
coord  *ans;
/*--------------------------------------------------------------------
*/
{
    vxrchrw(str, ans);
}

/*==================================================================*/
Public  Void  vrchrh(size, ans)
coord size;
coord *ans;
/*--------------------------------------------------------------------
*/
{
    vxrchrh(size, ans);
}

/*==================================================================*/
Public  Void  vrwvw(size)
coord size;
/*--------------------------------------------------------------------
 * set width of wave line
 */
{
    wwidth = size;
    vfrwvw(wwidth);
}

