/*********************************************************************
 *  File : vrps.c
 *
 *  Graphic Interface package
 *
 *   Version for xlib.
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include <sys/types.h>
#ifndef __TURBOC__
#include <sys/time.h>
#endif
#include <time.h>
#include "vr.h"
#include "vrf.h"

/*---------------------------------------------------------------------
 *  Local functions
 */
Private Void vfarc
        ARG((coord,coord,coord,coord,coord,coord,coord,int));
Private Void vfrprolog  ARG((NoArg));
Private Void vfrsetcol  ARG((int));

/*---------------------------------------------------------------------
    Local variables
*/
Local   char *verstr = "0.0 (1994/6/9)";

#define ARCEPS  1.0e-8
Local   char figname[128];
Local   FILE *figfp=NULL;
Local   int outcount=0;
Local   double pi, rd;
Local   int   llx, lly, urx, ury;
Local   coord wwidth = 0.001;
Local   bool  opened=False;

#define XSIZE 600
#define YSIZE 664
#define XOFFS  32
#define FSIZE (XSIZE-2*XOFFS)
#define YOFFS (YSIZE - FSIZE)
#define LWIDTH 1.0
Local   int xsize = XSIZE;
Local   int ysize = YSIZE;
Local   int fsize = FSIZE;
Local   int xoffs = XOFFS;
Local   int yoffs = YOFFS;

#define maxd(x,y) ((int)((x) > (y) ? (x) : (y)))
#define vfrrc(x) maxd(((int) fsize*(x)),0)
#define vfrxc(x) (vfrrc(x)+ xoffs)
#define vfryc(x) (vfrrc(x)+ yoffs)

Local   bool dbg=False;

#define PS  0
#define EPS 1
Local int prmode = PS;
#define SCALE1 1000
Local long scale = SCALE1;

/*==================================================================*/
Public Void vfrinit(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(dbg)
        printf("vfrinit\n");

    if(opened) {
        if(figfp != NULL)
            vfrcls();
        opened = False;
    }
        
    figfp = NULL;

    pi   = 4.0*atan(1.0);
    rd   = 180.0*64.0/pi;

    xsize = XSIZE;
    ysize = YSIZE;
    xoffs = XOFFS;
    scale = SCALE1;
    if(prmode == EPS) {
        scale = SCALE1/2;
    }
    fsize = xsize - 2*xoffs;
    yoffs = ysize - fsize;
    llx  = xoffs;
    lly  = yoffs;
    urx  = llx + fsize;
    ury  = lly + fsize;
}

/*==================================================================*/
Public Void vfropen(mode)
int    *mode;
/*--------------------------------------------------------------------
*/
{
    time_t  tloc;
    if(dbg)
        printf("vfropen(%d)\n", *mode);

    prmode = *mode;
    vfrinit();

    if(opened)
        return;

    if(prmode == EPS)
        sprintf(figname, "f%3.3d.eps", ++outcount);
    else
        sprintf(figname, "f%3.3d.ps", ++outcount);
    if((figfp = fopen(figname, "w")) == NULL) {
        fprintf(Stderr, "vfropen: open error %s\n", figname);
        exit(1);
    }
    if(dbg)
        printf("vfropen : %s opened\n", figname);

    opened = True;
    fprintf(figfp, "%%!\n");
    fprintf(figfp, "%%%%Title: %s\n",figname);
    fprintf(figfp, "%%%%Creator: grcdraw %s by Kaneko Toshiaki\n",
                   verstr);
    time(&tloc);
    fprintf(figfp, "%%%%CreationDate: %s\n", ctime(&tloc));
    fprintf(figfp, "%%%%For: \n");
    fprintf(figfp, "%%%%Pages: 1\n");
    fprintf(figfp, "%%%%BoundingBox: %1d %1d %1d %1d\n",
                   ((int) scale*llx/SCALE1-1),
                   ((int) scale*lly/SCALE1-1),
                   ((int) scale*urx/SCALE1+1),
                   ((int) scale*ury/SCALE1+1));
    fprintf(figfp, "%%%%DocumentFonts: Times-Roman Symbol\n");
    fprintf(figfp, "%%%%EndComments\n");

    vfrprolog();

    fprintf(figfp, "%%%%Page 1 1\n\n");
    fprintf(figfp, "grcdict begin\n");
    fprintf(figfp, "/Times-Roman findfont 10 scalefont setfont\n");
    if(scale != SCALE1)
        fprintf(figfp, "%f %f scale\n",
            ((float)scale/SCALE1), ((float)scale/SCALE1));
    if(dbg)
        printf("vfropen(%d): end\n", *mode);
}

/*==================================================================*/
Private Void vfrprolog(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(dbg)
        printf("vfrprolog\n");

    fprintf(figfp,"\n");
    fprintf(figfp,"/grcdict 4 dict def\n");
    fprintf(figfp," grcdict begin\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"/grcwavetodict 16 dict def\n");
    fprintf(figfp," grcwavetodict /mtrx matrix put\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"%%  x y n grcwaveto -\n");
    fprintf(figfp,"/grcwaveto\n");
    fprintf(figfp,"{ grcwavetodict begin\n");
    fprintf(figfp,"    /n exch def\n");
    fprintf(figfp,"    /y exch def\n");
    fprintf(figfp,"    /x exch def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"    currentpoint\n");
    fprintf(figfp,"    /cy exch def\n");
    fprintf(figfp,"    /cx exch def\n");
    fprintf(figfp,"    /dx x cx sub def\n");
    fprintf(figfp,"    /dy y cy sub def\n");
    fprintf(figfp,"    /r  dx dx mul dy dy mul add sqrt n div def\n");
    fprintf(figfp,"    /th dy dx atan def\n");
    fprintf(figfp,"    /w currentlinewidth def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"    /savematrix mtrx currentmatrix def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"        cx cy translate\n");
    fprintf(figfp,"        r r scale\n");
    fprintf(figfp,"        th rotate\n");
    fprintf(figfp,"        /lx 0 def\n");
    fprintf(figfp,"        /dir 1 def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"        w r div setlinewidth\n");
    fprintf(figfp,"        n {\n");
    fprintf(figfp,"            0.5 lx add dir 1 lx add 0 0.3 ");
    fprintf(figfp,"arcto\n");
    fprintf(figfp,"            pop pop pop pop\n");
    fprintf(figfp,"            /lx lx 1 add def\n");
    fprintf(figfp,"            lx 0 lineto\n");
    fprintf(figfp,"            /dir dir neg def\n");
    fprintf(figfp,"        } repeat\n");
    fprintf(figfp,"        stroke\n");
    fprintf(figfp,"    savematrix setmatrix\n");
    fprintf(figfp,"    w setlinewidth\n");
    fprintf(figfp,"  end\n");
    fprintf(figfp,"} def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"/grcwarcdict 36 dict def\n");
    fprintf(figfp," grcwarcdict /mtrx matrix put\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"%%  x y r ang1 ang2 n grcwarc -\n");
    fprintf(figfp,"/grcwarc\n");
    fprintf(figfp,"{ grcwarcdict begin\n");
    fprintf(figfp,"    /n  exch def\n");
    fprintf(figfp,"    /ae exch def\n");
    fprintf(figfp,"    /as exch def\n");
    fprintf(figfp,"    /r  exch def\n");
    fprintf(figfp,"    /y  exch def\n");
    fprintf(figfp,"    /x  exch def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"    /lt  ae as sub abs def\n");
    fprintf(figfp,"    /thh ae as sub n div 2 div def\n");
    fprintf(figfp,"    thh abs 20 gt\n");
    fprintf(figfp,"    {\n");
    fprintf(figfp,"        /n lt 20 div ceiling def\n");
    fprintf(figfp,"        /thh ae as sub n div 2 div def\n");
    fprintf(figfp,"    } if\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"    /ch thh cos def\n");
    fprintf(figfp,"    /w  thh abs r mul 3.1416 mul 180 div def\n");
    fprintf(figfp,"    /h  w 2 mul def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"    /s  r h add def\n");
    fprintf(figfp,"    /t  r h sub def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"    /th   thh 2 mul def\n");
    fprintf(figfp,"    /cth  th  cos def\n");
    fprintf(figfp,"    /sth  th  sin def\n");
    fprintf(figfp,"    /thd  thh 4 mul def\n");
    fprintf(figfp,"    /cthd thd cos def\n");
    fprintf(figfp,"    /sthd thd sin def\n");
    fprintf(figfp,"    /tht  thh 3 mul def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"    /lx as cos r mul def\n");
    fprintf(figfp,"    /ly as sin r mul def\n");
    fprintf(figfp,"    /lhx as tht sub cos s mul def\n");
    fprintf(figfp,"    /lhy as tht sub sin s mul def\n");
    fprintf(figfp,"    /llx as thh sub cos t mul def\n");
    fprintf(figfp,"    /lly as thh sub sin t mul def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"    /cr w 0.6 mul def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"    /savematrix mtrx currentmatrix def\n");
    fprintf(figfp,"        x y translate\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"        lx ly lineto\n");
    fprintf(figfp,"        /dir 1 def\n");
    fprintf(figfp,"        n {\n");
    fprintf(figfp,"            dir 0 ge\n");
    fprintf(figfp,"            {\n");
    fprintf(figfp,"                /lt  lhx def\n");
    fprintf(figfp,"                /lhx cthd lt mul sthd lhy mul ");
    fprintf(figfp,"sub def\n");
    fprintf(figfp,"                /lhy sthd lt mul cthd lhy mul ");
    fprintf(figfp,"add def\n");
    fprintf(figfp,"                /ltx lhx def\n");
    fprintf(figfp,"                /lty lhy def\n");
    fprintf(figfp,"            } {\n");
    fprintf(figfp,"                /lt  llx def\n");
    fprintf(figfp,"                /llx cthd lt mul sthd lly mul ");
    fprintf(figfp,"sub def\n");
    fprintf(figfp,"                /lly sthd lt mul cthd lly mul ");
    fprintf(figfp,"add def\n");
    fprintf(figfp,"                /ltx llx def\n");
    fprintf(figfp,"                /lty lly def\n");
    fprintf(figfp,"            } ifelse\n");
    fprintf(figfp,"            /lt  lx def\n");
    fprintf(figfp,"            /lx cth lt mul sth ly mul sub def\n");
    fprintf(figfp,"            /ly sth lt mul cth ly mul add def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"            ltx lty lx ly cr arcto\n");
    fprintf(figfp,"            pop pop pop pop\n");
    fprintf(figfp,"            lx ly lineto\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"            /dir dir neg def\n");
    fprintf(figfp,"        } repeat\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"    savematrix setmatrix\n");
    fprintf(figfp,"  end\n");
    fprintf(figfp,"} def\n");
    fprintf(figfp,"\n");
    fprintf(figfp,"end\n");
    fprintf(figfp, "%%%%EndProlog\n");
}

/*==================================================================*/
Public Void vfrcls(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(dbg)
        printf("vfrcls\n");

    if(!opened)
        return;
    fprintf(figfp, "end\n\n");
    if(prmode != EPS) {
        fprintf(figfp, "showpage\n\n");
    }
    fprintf(figfp, "%%%%Trailer\n");
    fprintf(figfp, "%%%%EOF\n");
    fclose(figfp);
    printf("Plotted in %s\n", figname);
    opened = False;
}

/*==================================================================*/
Public Void vfrdisp(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(dbg)
        printf("vfrdisp\n");

}

/*==================================================================*/
Public bool vfrclr(bool clr)
/*--------------------------------------------------------------------
*/
{
    if(dbg)
        printf("vfrclr\n");
    return True;
}

/*==================================================================*/
Public Void vfrnew(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(dbg)
        printf("vfrnew\n");
}


/*==================================================================*/
Public Void vfrend(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(dbg)
        printf("vfrend\n");
}
/*==================================================================*/
Public Void vfrline(lx0, ly0, lx1, ly1, color)
coord lx0, ly0;
coord lx1, ly1;
int color;
/*--------------------------------------------------------------------
*/
{
    int vfx, vfy, vfx1, vfy1, nw;
    coord r;

    if(dbg)
        printf("vfrline\n");

    /* Polyline */

    vfx = vfrxc(lx0);
    vfy = vfryc(ly0);
    vfx1 = vfrxc(lx1);
    vfy1 = vfryc(ly1);


    if(color == VR_MAGENTA) {
         r    = sqrt((lx0-lx1)*(lx0-lx1)+(ly0-ly1)*(ly0-ly1));
         nw   = ((int) (r / wwidth));
         nw   = maxd(nw, 3);

        fprintf(figfp,"gsave\n");
        fprintf(figfp, "%4.2f setlinewidth\n", LWIDTH);
        fprintf(figfp,"[] 0 setdash\n");
        fprintf(figfp,"newpath ");
        fprintf(figfp,"%d %d moveto ", vfx, vfy);
        fprintf(figfp,"%d %d %d grcwaveto ", vfx1, vfy1, nw);
        fprintf(figfp,"stroke\n");
        fprintf(figfp,"grestore\n");
    } else {
        fprintf(figfp,"gsave\n");
        vfrsetcol(color);
        fprintf(figfp,"newpath ");
        fprintf(figfp,"%d %d moveto ", vfx, vfy);
        fprintf(figfp,"%d %d lineto ", vfx1, vfy1);
        fprintf(figfp,"stroke\n");
        fprintf(figfp,"grestore\n");
    }
}

/*==================================================================*/
Public Void vfrcrcl(cx, cy, r, color)
coord cx, cy;
coord r;
int color;
/*--------------------------------------------------------------------
*/
{
    int  fcx, fcy, fr, fsx, fsy, nw;
    coord l;

    if(dbg)
        printf("vfrcrcl\n");

    fcx = vfrxc(cx);
    fcy = vfryc(cy);
    fr  = vfrrc(r);
    fsx = fcx + fr;
    fsy = fcy + fr;

    if(color == VR_MAGENTA) {
         l    = 2*pi*r;
         nw   = ((int) (l / (2*wwidth)))*2;
         nw   = maxd(nw, 8);

        fprintf(figfp,"gsave\n");
        fprintf(figfp, "%4.2f setlinewidth\n", LWIDTH);
        fprintf(figfp,"[] 0 setdash\n");
        fprintf(figfp,"newpath ");
        fprintf(figfp,"%d %d moveto ", fsx, fsy);
        fprintf(figfp,"%d %d %d 0 360 %d grcwarc ", fcx, fcy, fr, nw);
        fprintf(figfp,"stroke\n");
        fprintf(figfp,"grestore\n");
    } else {
        fprintf(figfp,"gsave\n");
        vfrsetcol(color);
        fprintf(figfp,"newpath ");
        fprintf(figfp,"%d %d moveto ", fsx, fsy);
        fprintf(figfp,"%d %d %d 0 360 arc ", fcx, fcy, fr);
        fprintf(figfp,"stroke\n");
        fprintf(figfp,"grestore\n");
    }
    
}

/*==================================================================*/
Private Void vfarc(lx0, ly0, cx, cy, cr, ths, the, color)
coord lx0, ly0;
coord cx, cy;
coord cr;
coord ths, the;
int color;
/*--------------------------------------------------------------------
*/
{
    int  fcx, fcy, fr, f0x, f0y, fs, fe, nw;
    coord l;

    if(dbg)
        printf("vfarc\n");

    fcx = vfrxc(cx);
    fcy = vfryc(cy);
    fr  = vfrrc(cr);
    f0x = vfrxc(lx0);
    f0y = vfryc(ly0);
    fs  = ths*180/pi;
    fe  = the*180/pi;
    
    if(color == VR_MAGENTA) {
         l    = cr*fabs(the - ths);
         nw   = ((int) (l / wwidth));
         nw   = maxd(nw, 3);

        fprintf(figfp,"gsave\n");
        fprintf(figfp, "%4.2f setlinewidth\n", LWIDTH);
        fprintf(figfp,"[] 0 setdash\n");
        fprintf(figfp,"newpath ");
        fprintf(figfp,"%d %d moveto ", f0x, f0y);
        fprintf(figfp,"%d %d %d %d %d %d grcwarc ", fcx, fcy, fr, fs, fe, nw);
        fprintf(figfp,"stroke\n");
        fprintf(figfp,"grestore\n");
    } else {
        fprintf(figfp,"gsave\n");
        vfrsetcol(color);
        fprintf(figfp,"newpath ");
        fprintf(figfp,"%d %d moveto ", f0x, f0y);
        fprintf(figfp,"%d %d %d %d %d arc ", fcx, fcy, fr, fs, fe);
        fprintf(figfp,"stroke\n");
        fprintf(figfp,"grestore\n");
    }
}

/*==================================================================*/
Public Void vfrarc(lx0, ly0, lx1, ly1, lx2, ly2, color)
coord lx0, ly0;
coord lx1, ly1;
coord lx2, ly2;
int color;
/*--------------------------------------------------------------------
*/
{
    coord d, cx, cy, r, xt, yt;
    coord ths, the;
    coord rs0, rs1, rs2;

    if(dbg)
        printf("vfrarc\n");

    d = ((lx0 - lx1)*(ly2 - ly1) - (lx2 - lx1)*(ly0 - ly1));
    if (fabs(d) < ARCEPS) {
        vfrline(lx0, ly0, lx1, ly1, color);
        vfrline(lx1, ly1, lx2, ly2, color);
    } else {
        rs0 = lx0*lx0 + ly0*ly0;
        rs1 = lx1*lx1 + ly1*ly1;
        rs2 = lx2*lx2 + ly2*ly2;
        cx  = (  (ly2 - ly1)*(rs0 - rs1) - (ly0 - ly1)*(rs2 - rs1))/
              (2.0*d);
        cy  = (- (lx2 - lx1)*(rs0 - rs1) + (lx0 - lx1)*(rs2 - rs1))/
              (2.0*d);
        r   = sqrt((lx0 - cx)*(lx0 - cx) + (ly0 - cy)*(ly0 - cy));
        d   = (lx0 - cx)*(ly1 - cy) - (lx1 - cx)*(ly0 - cy);
        if (d < 0) {
            xt  = lx2;  yt = ly2;
            lx2 = lx0;  ly2 = ly0;
            lx0 = xt;   ly0 = yt;
        }
        ths = atan2(ly0 - cy, lx0 - cx);
        the = atan2(ly2 - cy, lx2 - cx);
        if (the < ths)
            the += 2.0*pi;
        if (ths < 0) {
            the += 2.0*pi;
            ths += 2.0*pi;
        }

        vfarc(lx0, ly0, cx, cy, r, ths, the, color);
    }
}

/*==================================================================*/
Public Void vfrtext(x, y, str, color)
coord x, y;
string str;
int color;
/*--------------------------------------------------------------------
*/
{
    int vfx, vfy;

    if(dbg)
        printf("vfrtext\n");

    vfx = vfrxc(x);
    vfy = vfryc(y);
/*
    vfy = vfryc(y - 0.02);
 */

    if(str[0]=='G') {
        fprintf(figfp, "/Times-Roman findfont 12 scalefont setfont\n");
    } else {
        fprintf(figfp, "/Times-Roman findfont 10 scalefont setfont\n");
    }

    /* text */
    fprintf(figfp,"gsave\n");
    fprintf(figfp,"newpath ");
    fprintf(figfp,"%d %d moveto ", vfx, vfy);
    fprintf(figfp,"(%s) ", str);
    fprintf(figfp,"show\n");
    fprintf(figfp,"grestore\n");
}

/*==================================================================*/
Public Void vfrwvw(w)
coord  w;
/*--------------------------------------------------------------------
*/
{
    if(dbg)
        printf("vfrwvw\n");

    wwidth = w;
}

/*==================================================================*/
Private Void vfrsetcol(color)
int color;
/*--------------------------------------------------------------------
*/
{
    float width = LWIDTH;

    if(dbg)
        printf("vfrsetcol\n");

    if(color == VR_GREEN) {
        fprintf(figfp, "%4.2f setlinewidth\n", width);
        fprintf(figfp,"[2] 0 setdash\n");
    } else if(color == VR_BLUE) {
        fprintf(figfp, "%4.2f setlinewidth\n", width);
        fprintf(figfp,"[] 0 setdash\n");
    } else if(color == VR_MAGENTA) {
        fprintf(figfp, "%4.2f setlinewidth\n", width);
        fprintf(figfp,"[4 2] 0 setdash\n");
    } else if(color == VR_RED) {
        fprintf(figfp, "%4.2f setlinewidth\n", 0.5*width);
        fprintf(figfp,"[] 0 setdash\n");
    } else if(color == VR_BLACK) {
        fprintf(figfp, "%4.2f setlinewidth\n", 0.5*width);
        fprintf(figfp,"[] 0 setdash\n");
    } else {
        fprintf(figfp, "%4.2f setlinewidth\n", 0.5*width);
        fprintf(figfp,"[6 2 2 2] 0 setdash\n");
    }
}
