/*********************************************************************
 *  File : vrx.c
 *
 *  Graphic Interface package
 *
 *   Version for xlib.
 *                                                 by Kaneko Toshiaki
 */

#ifdef X11
#ifdef  TOOLKIT
#include <X11/Intrinsic.h> 
#include <X11/StringDefs.h>
#else
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif
#endif

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include "vr.h"
#include "vrx.h"
#include "dcx.h"

#ifdef X11
/*---------------------------------------------------------------------
 * Local types
 */
#ifndef Dimension
#define Dimension    long
#endif

/*---------------------------------------------------------------------
 *  Local functions
 */
Private Ulong vxmcolor  ARG((int));

/*---------------------------------------------------------------------
 *  Local variables
 */

Local Dimension width, height;

Local Font font;
Local GC gct;
Local int    direction, ascent, descent;
Local XCharStruct overall;

Local coord cx0=0, cx1=1;
Local coord cy0=0, cy1=1;
Local coord cr1=1;
Local double pi, pi2, rd;
Local bool   vopen;
Local coord csize;
Local Ulong dsize = 0;

Local bool vrdbg=False;

Local int  ncolors;

/*
static char *vrcnam[VR_CNUM] = {
    "black",
    "white",
    "blue4",
    "green4",
    "cyan4",
    "red4",
    "magenta4",
    "brown",
    "lightgray",
    "gray",
    "blue",
    "green",
    "cyan",
    "red",
    "magenta",
    "yellow",
};
*/

Local struct {
    string name;
    int    style;
    char   dash[2];
} vrcnam[VR_CNUM] = {
           {"black",          LineSolid,     {1, 0}}, /*   0   0   0 */
           {"white",          LineOnOffDash, {1, 2}}, /* 255 255 255 */
           {"navy blue",      LineSolid,     {1, 0}}, /*  34  34 139 */
           {"green",          LineOnOffDash, {2, 4}}, /*   0 255   0 */
           {"pale turquoise", LineOnOffDash, {4, 2}}, /* 173 238 238 */
           {"red",            LineSolid,     {1, 0}}, /* 255   0   0 */ 
           {"magenta",        LineOnOffDash, {4, 3}}, /* 255   0 255 */
           {"brown",          LineOnOffDash, {1, 1}}, /* 165  42  42 */ 
           {"light gray",     LineOnOffDash, {2, 2}}, /* 211 211 211 */
           {"gray",           LineOnOffDash, {1, 1}}, /* 192 192 192 */
           {"blue",           LineSolid,     {1, 0}}, /*   0   0 255 */
           {"forest green",   LineOnOffDash, {2, 2}}, /*  34 139  34 */
           {"cyan",           LineSolid,     {1, 0}}, /*   0 255 255 */
           {"light pink",     LineOnOffDash, {1, 1}}, /* 255 174 185 */
           {"violet",         LineOnOffDash, {1, 1}}, /* 238 130 238 */
           {"yellow",         LineOnOffDash, {1, 1}}, /* 255 255   0 */
};
Local GC gcs[VR_CNUM];
#define THICKNESS ((Ulong) 2)

Local char *foname[] = {
     "-adobe-helvetica-bold-r-*-*-10-*-*-*-*-*-*-*",
     "-adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*",
     "-adobe-helvetica-bold-r-*-*-14-*-*-*-*-*-*-*",
     "-adobe-helvetica-bold-r-*-*-18-*-*-*-*-*-*-*",
     "-adobe-helvetica-bold-r-*-*-24-*-*-*-*-*-*-*"
};
Local int fheight[] = {10, 12, 14, 18, 24};
Local int fsize = sizeof(foname)/sizeof(char*);
#define FSIZE 0.02

#define xnorm(x)  ((int) (cx0 + cx1 * (x)))
#define ynorm(y)  ((int) (cy0 + cy1 * (y)))
#define rnorm(r)  ((int) (cr1 * (r)))
#define todeg(s)  ((int) (rd * (s)))
#define ARCEPS    1.0e-8

#define OUTX      1
#define OUTFIG    2

/*====================================================================*/
Public Void vxrinit(NoArg)
/*--------------------------------------------------------------------
*/
{
    int i;
    double one = 1;

    if(vrdbg)
        printf("vxrint\n");

    pi   = 4*atan(one);
    pi2  = 2*pi;
    rd   = 180*64/pi;

    ncolors = vxrcolr();
    if(ncolors > VR_CNUM)
        ncolors = VR_CNUM;
    for(i = 0; i < ncolors; i++) {
        XColor closest, def;
        if(!XAllocNamedColor(disp, DefaultColormap(disp, 0),
                vrcnam[i].name, &closest, &def)) {
            ncolors = 2;
            break;
        }
        gcs[i] = XCreateGC(disp, RootWindow(disp, 0), 0, NULL);
        XSetLineAttributes(disp, gcs[i], THICKNESS, LineSolid,
                CapButt, JoinBevel);
        XSetForeground(disp, gcs[i], closest.pixel);
    }
    gct = XCreateGC(disp, RootWindow(disp, 0), 0, NULL);
    XSetForeground(disp, gct, BlackPixel(disp, 0));
}

/*====================================================================*/
Public Void vxropen(NoArg)
/*--------------------------------------------------------------------
*/
{
}

/*====================================================================*/
Public Void vxrcls(NoArg)
/*--------------------------------------------------------------------
*/
{
}

/*====================================================================*/
Public Void vxrdisp(NoArg)
/*--------------------------------------------------------------------
*/
{
    if(vrdbg)
        printf("vxrdisp\n");

    XFlush(disp);
}

/*====================================================================*/
Public bool vxrclr(bool clr)
/*--------------------------------------------------------------------
*/
{
    static int   lx=-1, ly=-1;
    static Uint  lw=0, lh=0, lb=0, ld=0;
    int           x, y;
    Uint          w, h, b, d;

    if(vrdbg)
        printf("vxrclr\n");

#ifdef TOOLKIT
    win = XtWindow(*wid);
#endif
    XGetGeometry(disp, win, &root, &x, &y, &w, &h, &b, &d);
    if(clr || x != lx || y != ly || w != lw || h != lh || b != lb || d != ld) {
        lx = x;
        ly = y;
        lw = w;
        lh = h;
        lb = b;
        ld = d;
        XClearWindow(disp, win);
        return True;
    }
    return False;
}

/*====================================================================*/
Public Void vxrnew(NoArg)
/*--------------------------------------------------------------------
*/
{
    coord lx0 = 0, lx1 = 1, ly0 = 0, ly1 = 1;
    coord dx = lx1 - lx0;
    coord dy = ly1 - ly0;
    coord fh;
    int   x, y;
    Uint  w, h, b, d;

    if(vrdbg)
        printf("vxrnew\n");

#ifdef TOOLKIT
/*
    XtVaGetValues(*wid, XtNwidth, &width, XtNheight, &height, NULL);
 */
    win = XtWindow(*wid);
#endif
    XGetGeometry(disp, win, &root, &x, &y, &w, &h, &b, &d);
    width  = w;
    height = h;

    if(width < 1)
        width = 100;
    if(height < 1)
        height = 100;

    cx1 = width/dx;
    cx0 = - cx1*lx0;

    cy1 = - height/dy;
    cy0 = - cy1*ly1;

    cr1 = 0.5*(fabs(cx1) + fabs(cy1));

    vxrchrh(FSIZE, &fh);
}

/*====================================================================*/
Public Void vxrend(NoArg)
/*--------------------------------------------------------------------
*/
{
}

/*********************************************************************
 * Plotting
 */
/*====================================================================*/
Public Void vxrline(lx0, ly0, lx1, ly1, color)
coord lx0, ly0;
coord lx1, ly1;
int color;
/*--------------------------------------------------------------------
*/
{
    int c  = vxmcolor(color);

    if(vrdbg)
        printf("vxrline(%f,%f,%f,%f,%d)\n", lx0, ly0, lx1, ly1, color);

    vopen = True;

    XDrawLine(disp, win, gcs[c], xnorm(lx0), ynorm(ly0),
              xnorm(lx1), ynorm(ly1));
}

/*====================================================================*/
Public Void vxrcrcl(cx, cy, r, color)
coord cx, cy;
coord r;
int color;
/*--------------------------------------------------------------------
*/
{
    int c = vxmcolor(color);

    if(vrdbg)
        printf("vxrcrcl\n");

    vopen = True;

    XDrawArc(disp, win, gcs[c], xnorm(cx-r), ynorm(cy+r),
             rnorm(2.0*r), rnorm(2.0*r), 0, 360*64);
}

/*====================================================================*/
Public Void vxrarc(lx0, ly0, lx1, ly1, lx2, ly2, color)
coord lx0, ly0;
coord lx1, ly1;
coord lx2, ly2;
int color;
/*--------------------------------------------------------------------
*/
{
    coord d, cx, cy, r;
    coord ths, the;
    coord rs0, rs1, rs2;
    int c;

    if(vrdbg)
        printf("vxrarc(%f,%f,%f,%f,%f,%f,%d)\n",
               lx0,ly0,lx1,ly1,lx2,ly2,color);

    vopen = True;

    d = ((lx0 - lx1)*(ly2 - ly1) - (lx2 - lx1)*(ly0 - ly1));
    if (fabs(d) < ARCEPS) {
        vrline(lx0, ly0, lx1, ly1, color);
        vrline(lx1, ly1, lx2, ly2, color);
    } else {
        rs0 = lx0*lx0 + ly0*ly0;
        rs1 = lx1*lx1 + ly1*ly1;
        rs2 = lx2*lx2 + ly2*ly2;
        cx  = (  (ly2 - ly1)*(rs0 - rs1) - (ly0 - ly1)*(rs2 - rs1))/
              (2.0*d);
        cy  = (- (lx2 - lx1)*(rs0 - rs1) + (lx0 - lx1)*(rs2 - rs1))/
              (2.0*d);
        r   = sqrt((lx0 - cx)*(lx0 - cx) + (ly0 - cy)*(ly0 - cy));
        d   = (lx0 - cx)*(ly1 - cy) - (lx1 - cx)*(ly0 - cy);
        if (d < 0) {
            ths = atan2(ly2 - cy, lx2 - cx);
            the = atan2(ly0 - cy, lx0 - cx);
        } else {
            ths = atan2(ly0 - cy, lx0 - cx);
            the = atan2(ly2 - cy, lx2 - cx);
        }

        if (the < ths)
            the += pi2;

        c  = vxmcolor(color);
        
        XDrawArc(disp, win, gcs[c], xnorm(cx-r), ynorm(cy+r),
              rnorm(2.0*r), rnorm(2.0*r), todeg(ths), todeg(the-ths));
    }
}

/*====================================================================*/
Public Void vxrtext(x, y, str, color)
coord x, y;
string str;
int color;
/*--------------------------------------------------------------------
 * draw text "str", whose lower left corner is placed at (x, y)
 */
{
    if(vrdbg)
        printf("vxrtext (%f, %f) %s\n", (float) x, (float) y, str);

    vopen = True;

    XDrawString(disp, win, gct, xnorm(x), ynorm(y-0.01),
                str, strlen(str));
}

/*====================================================================*/
Public  int vxrgnum(NoArg)
/*--------------------------------------------------------------------
 * resolustion.
 * retuns a value "n", which is recomended for division of display
 * by "n"*"n".
 */
{
    return 4;
}

/*====================================================================*/
Public  int vxrcolr(NoArg)
/*--------------------------------------------------------------------
*/
{
    int mcolor;

    mcolor = DisplayCells(disp, screen);
    return mcolor;
}

/*====================================================================*/
Private Ulong vxmcolor(color)
int     color;
/*--------------------------------------------------------------------
*/
{
    int s = vrcnam[color].style;

    if(vrdbg)
        printf("vxmcolor(%d)\n", color);

    if(ncolors < VR_CNUM) {
        int c = (int) VR_BLACK;
        XSetLineAttributes(disp, gcs[c], 2, s, CapButt, JoinMiter);
        if(s == LineOnOffDash)
            XSetDashes(disp, gcs[c], 0, vrcnam[color].dash, 2);
        return c;
    } else {
        return color;
    }
}

/*====================================================================*/
Public  int vxrback(color)
int     color;
/*--------------------------------------------------------------------
 * Change background color
 * Returns the resulting color
 */
{
    if(vrdbg)
        printf("vxrback(%d)\n", color);

    return VR_WHITE;
}

/*====================================================================*/
Public  Void  vxrchrw(str, ans)
string str;
coord  *ans;
/*--------------------------------------------------------------------
 * returns drawn width of "str".
 * This function does not really draw it.
 */
{
    if(vrdbg)
        printf("vxrchrw\n");

    if(dsize < 1) {
        *ans = 0.0;
        return;
    }

/*
    XQueryTextExtents(disp, font, str, strlen(str), &direction,
                      &ascent, &descent, &overall);
*/

    *ans = ((coord) overall.width) / cx1;
}

/*====================================================================*/
Public  Void  vxrchrh(size, ans)
coord size;
coord *ans;
/*--------------------------------------------------------------------
 * Require set character size to "size".
 * This function returns size to be used.
 *
 * In this implimentation, character size is fixed.
 */
{
    int    i;

    if(vrdbg)
        printf("vrxchrh\n");

    csize = size;
    dsize = rnorm(size);
    for (i = 0; i < fsize; i++) {
        if(dsize <= fheight[i])
            break;
    }
    if (i >= fsize)
        i = fsize - 1;

    font = XLoadFont(disp, foname[i]);

    XSetFont(disp, gct, font);

    XQueryTextExtents(disp, font, "H", 1, &direction,
                      &ascent, &descent, &overall);

    *ans = ascent / cx1;
}

#else
/*********************************************************************
 * No X11 system
 */
/*====================================================================*/
Public Void vxrinit(NoArg)
{
    fprintf(Stderr, "*** X Window system is not found\n");
    exit(1);
}
Public Void vxropen(NoArg)
{
}
Public Void vxrcls(NoArg)
{
}
Public Void vxrdisp(NoArg)
{
}
Public bool vxrclr(clr)
bool clr;
{
    return False; 
}
Public Void vxrnew(NoArg)
{
}
Public Void vxrend(NoArg)
{
}
Public Void vxrline(lx0, ly0, lx1, ly1, color)
coord lx0, ly0;
coord lx1, ly1;
int color;
{
}
Public Void vxrcrcl(cx, cy, r, color)
coord cx, cy;
coord r;
int color;
{
}
Public Void vxrarc(lx0, ly0, lx1, ly1, lx2, ly2, color)
coord lx0, ly0;
coord lx1, ly1;
coord lx2, ly2;
int color;
{
}
Public Void vxrtext(x, y, str, color)
coord x, y;
string str;
int color;
{
}
Public int  vxrgnum(NoArg)
{
    return 0; 
}
Public int  vxrcolr(NoArg)
{
    return 0;
}
Public int  vxrback(color)
int color;
{
    return 0;
}
Public  Void  vxrchrw(str, ans)
string str;
coord  *ans;
{
}
Public  Void  vxrchrh(size, ans)
coord size;
coord *ans;
{
}
#endif
