/*********************************************************************
 *  File : vrxfig.c
 *
 *  Graphic Interface package
 *
 *   Version for xfig.
 *                                                 by Kaneko Toshiaki
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include "vr.h"
#include "vrx.h"

/*---------------------------------------------------------------------
    Local functions
*/
Private Void vfrfopen   ARG((NoArg));
Private int  vfcolor    ARG((int));
Private Void vfarc
        ARG((coord,coord,coord,coord,coord,coord,coord,coord,int));

/*---------------------------------------------------------------------
    Local variables
*/
#define ARCEPS  1.0e-8
Local   char figname[128];
Local   FILE *figfp=NULL;
Local   int outcount=0;
Local   double pi, rd;

#define FSIZE 600
#define maxd(x,y) ((int)((x) > (y) ? (x) : (y)))
#define vfrrc(x) maxd(((int) FSIZE*(x)),0)
#define vfrxc(x) (vfrrc(x)+ 32)
#define vfryc(x) (FSIZE + 32 - vfrrc(x))

/**********************************************************************/
Public Void vfrnew(fnt)
int     fnt;
{
    figfp = NULL;

    pi   = 4.0*atan(1.0);
    rd   = 180.0*64.0/pi;
}

/**********************************************************************/
/**********************************************************************/
Public Void vfrline(x0, y0, x1, y1, w, color)
coord x0, y0;
coord x1, y1;
coord w;
int color;
{
    int vfx, vfy, vfx1, vfy1;

    vfrfopen();

    /* Polyline */

    vfx = vfrxc(x0);
    vfy = vfryc(y0);

    fprintf(figfp,"2");      /* object_code */
    fprintf(figfp," 1");     /* sub_type */
                             /* 1 : polyline
                              * 2 : box
                              * 3 : polygon
                              * 4 : arc-box
                              * 5 : imported-eps bounding-box
                              */
    if(color == VR_GREEN)
        fprintf(figfp," 2");     /* line_style : enumeration */
    else if(color == VR_MAGENTA)
        fprintf(figfp," 1");     /* line_style : enumeration */
    else
        fprintf(figfp," 0");     /* line_style : enumeration */

    fprintf(figfp," 1");     /* thickness  : pixels */
    fprintf(figfp," %d", 
            vfcolor(color)); /* color      : enumeration */
    fprintf(figfp," 0");     /* depth      : no unit */
    fprintf(figfp," 0");     /* pen        : not used */
    fprintf(figfp," 0");     /* area_fill  : enumeration */

    if(color == VR_GREEN)
        fprintf(figfp," 3.000"); /* style_val  : pixels */
    else if(color == VR_MAGENTA)
        fprintf(figfp," 4.000"); /* style_val  : pixels */
    else
        fprintf(figfp," 0.000"); /* style_val  : pixels */

    fprintf(figfp," -1");    /* radius     : pixels */
    fprintf(figfp," 0");     /* forward_arrow  : bool */
    fprintf(figfp," 0\n");   /* backward_arrow : bool */
    fprintf(figfp,"  %d %d", vfx, vfy);

    vfx1 = vfrxc(x1);
    vfy1 = vfryc(y1);

    fprintf(figfp," %d %d", vfx1, vfy1);
    fprintf(figfp," 9999 9999\n");

}

/**********************************************************************/
/**********************************************************************/
Private Void vfrfopen(NoArg)
{
    if(figfp != NULL)
        return;

    sprintf(figname, "f%d.fig", ++outcount);
    if((figfp = fopen(figname, "w")) == NULL) {
        fprintf(Stderr, "vfrfopen: open error %s\n", figname);
        exit(1);
    }
    fprintf(figfp, "#FIG 2.1\n");  /* the first line */
    fprintf(figfp, "80");          /* fir_resolution : pixels/inch */
    fprintf(figfp, " 2\n");        /* origin
                                    * 1 : lower left
                                    * 2 : upper left
                                    */
}

/**********************************************************************/
/**********************************************************************/
Public Void vfrfcls(NoArg)
{
    if(figfp != NULL)
        fclose(figfp);
    figfp = NULL;
}

/**********************************************************************/
/**********************************************************************/
Private int  vfcolor(color)
int color;
{
    int clr;

    clr = -1;
    return clr;
}

/**********************************************************************/
/**********************************************************************/
Public Void vfrcrcl(cx, cy, r, color)
coord cx, cy;
coord r;
int color;
{
    int  fcx, fcy, fr;

    fcx = vfrxc(cx);
    fcy = vfryc(cy);
    fr  = vfrrc(r);

    vfrfopen();

    /* ellipse */
    fprintf(figfp,"1");      /* object_code */
    fprintf(figfp," 3");     /* sub_type */
                             /* 1 : ellipse defined by radiuses
                              * 2 : ellipse defined by diameters
                              * 3 : circle defined by radius
                              * 4 : circle defined by diameter
                              */
    if(color == VR_GREEN)
        fprintf(figfp," 2");     /* line_style : enumeration */
    else if(color == VR_MAGENTA)
        fprintf(figfp," 1");     /* line_style : enumeration */
    else
        fprintf(figfp," 0");     /* line_style : enumeration */

/*    fprintf(figfp," 0");  */   /* line_style : enumeration */
    fprintf(figfp," 1");     /* thickness  : pixels */
    fprintf(figfp," %d",
            vfcolor(color)); /* color : enumeration */
    fprintf(figfp," 0");     /* depth      : no unit */
    fprintf(figfp," 0");     /* pen        : not used */
    fprintf(figfp," 0");     /* area_fill  : enumeration */

    if(color == VR_GREEN)
        fprintf(figfp," 3.000"); /* style_val  : pixels */
    else if(color == VR_MAGENTA)
        fprintf(figfp," 4.000"); /* style_val  : pixels */
    else
        fprintf(figfp," 0.000"); /* style_val  : pixels */

    fprintf(figfp," 1");     /* direction  : 1 */
    fprintf(figfp," 0.000"); /* angle      : radians */
    fprintf(figfp," %d %d",
            fcx, fcy);       /* center */
    fprintf(figfp," %d %d",   
            fr, fr);         /* radius */
    fprintf(figfp," %d %d",
            fcx, fcy);       /* start */
    fprintf(figfp," %d %d\n",
            fcx+fr, fcy);    /* end ??? */
}

/**********************************************************************/
/**********************************************************************/
Private Void vfarc(x0, y0, x1, y1, x2, y2, cx, cy, color)
coord x0, y0;
coord x1, y1;
coord x2, y2;
coord cx, cy;
int color;
{
    int  fcx, fcy;

    vfrfopen();

    fcx = vfrxc(cx);
    fcy = vfryc(cy);

    /* arc */
    fprintf(figfp,"5");      /* object_code */
    fprintf(figfp," 1");     /* sub_type   : 1 */

    if(color == VR_GREEN)
        fprintf(figfp," 2");     /* line_style : enumeration */
    else if(color == VR_MAGENTA)
        fprintf(figfp," 1");     /* line_style : enumeration */
    else
        fprintf(figfp," 0");     /* line_style : enumeration */

    fprintf(figfp," 1");     /* thickness  : pixels */
    fprintf(figfp," %d", 
            vfcolor(color)); /* color : enumeration */
    fprintf(figfp," 0");     /* depth      : no unit */
    fprintf(figfp," 0");     /* pen        : not used */
    fprintf(figfp," 0");     /* area_fill  : enumeration */

    if(color == VR_GREEN)
        fprintf(figfp," 3.000"); /* style_val  : pixels */
    else if(color == VR_MAGENTA)
        fprintf(figfp," 4.000"); /* style_val  : pixels */
    else
        fprintf(figfp," 0.000"); /* style_val  : pixels */

    fprintf(figfp," 1");     /* direction */
                             /* 0 : clockwise
                              * 1 : counter-clockwise
                              */
    fprintf(figfp," 0");     /* forward_arrow  : bool */
    fprintf(figfp," 0");     /* backward_arrow : bool */
    fprintf(figfp," %d.0 %d.0",
            fcx, fcy);       /* center ??? */
    fprintf(figfp," %d %d",
            vfrxc(x0), vfryc(y0));  /* 1st point */
    fprintf(figfp," %d %d",
            vfrxc(x1), vfryc(y1));  /* 2nd point */
    fprintf(figfp," %d %d\n",
            vfrxc(x2), vfryc(y2));  /* 3rd point */
}

Public Void vfrarc(x0, y0, x1, y1, x2, y2, color)
coord x0, y0;
coord x1, y1;
coord x2, y2;
int color;
{
    coord d, cx, cy, r, xt, yt;
    coord ths, the;
    coord rs0, rs1, rs2;

    d = ((x0 - x1)*(y2 - y1) - (x2 - x1)*(y0 - y1));
    if (fabs(d) < ARCEPS) {
        vfrline(x0, y0, x1, y1, color);
        vfrline(x1, y1, x2, y2, color);
    } else {
        rs0 = x0*x0 + y0*y0;
        rs1 = x1*x1 + y1*y1;
        rs2 = x2*x2 + y2*y2;
        cx  = (  (y2 - y1)*(rs0 - rs1) - (y0 - y1)*(rs2 - rs1))/
              (2.0*d);
        cy  = (- (x2 - x1)*(rs0 - rs1) + (x0 - x1)*(rs2 - rs1))/
              (2.0*d);
        r   = sqrt((x0 - cx)*(x0 - cx) + (y0 - cy)*(y0 - cy));
        d   = (x0 - cx)*(y1 - cy) - (x1 - cx)*(y0 - cy);
        if (d < 0) {
            xt = x2;  yt = y2;
            x2 = x0;  y2 = y0;
            x0 = xt;  y0 = yt;
        }
        ths = atan2(y0 - cy, x0 - cx);
        the = atan2(y2 - cy, x2 - cx);
        if (the < ths)
            the += 2.0*pi;

        vfarc(x0, y0, x1, y1, x2, y2, cx, cy, color);
    }
}

/*********************************************************************
 */
Public Void vfrtext(x, y, str, color)
coord x, y;
string str;
int color;
{
    int fsize = 12, vfx, vfy, vfl;

    vfrfopen();

    vfl = strlen(str)*fsize;
    vfx = vfrxc(x);
    vfy = vfryc(y) + fsize;

    if(str[0]=='G')
        fprintf(Stderr, "Output %s to f%d.fig\n", str, outcount);

    /* text */
    fprintf(figfp,"4");      /* object_code */
    fprintf(figfp," 1");     /* sub_type */
                             /* 1 : left justified
                              * 2 : center justified
                              * 3 : right justified
                              */
    fprintf(figfp," 0");     /* font       : enumeration */
    fprintf(figfp," %d", fsize);    /* font_size  : points  12 */
    fprintf(figfp," 0");     /* pen        : not used */
    fprintf(figfp," %d",
            vfcolor(color)); /* color      : enumeration */
    fprintf(figfp," 0");     /* depth      : no unit */
    fprintf(figfp," 0.000"); /* angle      : radians */
    fprintf(figfp," 4");     /* font_flags : bitvector ??? */
    fprintf(figfp," %d", fsize);  /* height     : pixels */
    fprintf(figfp," %d", vfl);    /* length     : pixels */
    fprintf(figfp," %d %d",
            vfx, vfy);       /* position */
    fprintf(figfp," %s\1\n", str);  /* text */
}

/*********************************************************************/
/*********************************************************************/
Public Void vfrdisp(NoArg)
{
}

/*********************************************************************/
/*********************************************************************/
Public Void vfrclr0(NoArg)
{
}

/*********************************************************************/
/*********************************************************************/
Public Void vfrwvw(w)
coord  w;
{
}

/**********************************************************************/
/**********************************************************************/
Public Void vfrend(NoArg)
{
    vfrfcls();
}

